#include "macro.h"

MODULE Mod_DM

!######################################################################
!#
!# FILENAME: mod_dm.f90
!#
!# DESCRIPTION: This module defines a data type for storing dark matter
!#  particles.
!#
!# DEPENDENCIES: Mod_Globals for defining data types
!#
!# HISTORY:
!#    3/5/14  - Peter Mendygral
!#
!######################################################################

!### require the needed modules ###
USE Mod_Globals
USE Mod_Error

IMPLICIT NONE

!### default private but make the data type available ###
PRIVATE
PUBLIC :: DM

!### define a data type for this module ###
TYPE :: DM

    !### default everything as private ###
    PRIVATE

    !### we'll store all position, velocity and acceleration for each particle.
    REAL(WRD), PUBLIC, ALLOCATABLE :: particles(:,:)

    !### we'll track which particles are actually in use ###
    LOGICAL(WIS), PUBLIC, ALLOCATABLE :: isactive(:), outofbounds(:)
    !### info = destination for out of bounds particles, 0 means in bounds, -1 is a special flag utilized by assign_dens in mod_particlemesh.f90
    INTEGER(WIS), PUBLIC, ALLOCATABLE :: info(:)
    !### then some accounting ###
    !### ntotslots = # of particles we can store, nopenslots = # empty slots in DM block (isactive == FALSE),
    !### isLive = object is constructed, hasSlots = object has allocated arrays and available slots ###
    INTEGER(WIS), PUBLIC :: ntotslots, nopenslots, nvars
    LOGICAL(WIS), PUBLIC :: isLive, hasSlots

    CONTAINS

    !### default everything as private ###
    PRIVATE

    PROCEDURE :: initdm
    PROCEDURE, PUBLIC :: build
    PROCEDURE, PUBLIC :: unbuild
    PROCEDURE, PUBLIC :: destroy

END TYPE DM

!### create an interface to the constructor ###
INTERFACE DM

    MODULE PROCEDURE constructor

END INTERFACE DM

!### object methods ###
CONTAINS

!### a constructor for the class ###
FUNCTION constructor(ndim, ntotslots)

    TYPE(DM) :: constructor
    INTEGER(WIS), INTENT(IN) :: ndim, ntotslots

    CALL constructor%initdm(ndim, ntotslots)
    RETURN

END FUNCTION constructor

!### actual initialization of object (not directly called by user) ###
SUBROUTINE initdm(self, ndim, ntotslots)

    CLASS(DM) :: self
    INTEGER(WIS), INTENT(IN) :: ndim, ntotslots

    !### store the settings ###
    self%ntotslots = ntotslots
    self%nvars     = 3*ndim        ! pos, vel, acc

    !### default that we are not yet open for business ###
    self%hasSlots    = .FALSE.
    self%isLive      = .TRUE.
    self%nopenslots  = 0

END SUBROUTINE initdm

!### actually allocate the block of slots ###
SUBROUTINE build(self)

    CLASS(DM) :: self

    !### allocate the block both state variables and open flag ###
    IF (self%isLive .AND. .NOT. self%hasSlots) THEN

        ALLOCATE(self%particles(self%ntotslots, self%nvars), self%isactive(self%ntotslots), self%outofbounds(self%ntotslots), self%info(self%ntotslots))
        self%particles   = 0
        self%isactive    = .FALSE.
        self%outofbounds = .FALSE.
        self%info = 0
        
        !### we're open for business ###
        self%hasSlots    = .TRUE.
        self%isLive      = .TRUE.
        self%nopenslots  = self%ntotslots

    END IF

END SUBROUTINE build

!### deallocate the arrays ###
SUBROUTINE unbuild(self)

    CLASS(DM) :: self

    IF (self%isLive .AND. ALLOCATED(self%particles)) THEN

        DEALLOCATE(self%particles)
        DEALLOCATE(self%isactive)
        DEALLOCATE(self%outofbounds)
        DEALLOCATE(self%info)
        self%hasSlots = .FALSE.
        self%nopenslots  = 0

    END IF

END SUBROUTINE unbuild

!### destructor for the class ###
SUBROUTINE destroy(self)

    CLASS(DM) :: self

    CALL self%unbuild()
    self%isLive = .FALSE.

END SUBROUTINE destroy

END MODULE Mod_DM
