#include "macro.h"

MODULE Mod_DMBlock_Manager

!######################################################################
!#
!# FILENAME: mod_dmblock_manager.f90
!#
!# DESCRIPTION: This module will manage blocks of DM objects.  It is responsible
!#  for placing particles, building/unbuilding DM objects and defragmenting them.
!#
!# DEPENDENCIES: Mod_Globals for defining data types
!#
!# HISTORY:
!#    4/9/14  - Peter Mendygral
!#
!######################################################################

!### require the needed modules ###
USE Mod_DM
USE Mod_Globals
USE Mod_Error
USE Mod_RankLocation
USE Mod_SimulationUnits

IMPLICIT NONE

!### default private but make the data type available ###
PRIVATE
PUBLIC :: DMBlock_Manager

!### define a data type for this module ###
TYPE :: DMBlock_Manager

    !### default everything as private ###
    PRIVATE

    !### we'll keep a list of DM objects to manage ###
    TYPE(DM), PUBLIC, ALLOCATABLE :: dms(:)

    !### lens = # of out-of-bounds elements (particles*ndim*3) to be passed to each neighbor across all active DMs
    INTEGER(WIS), PUBLIC, ALLOCATABLE :: lens(:)

    REAL(WRD), PUBLIC ::  xLow, xHi, yLow, yHi, zLow, zHi
    !### then some accounting ###
    !### ntotdms = max # of DM blocks that BM will manage, nactive = # of DMs built, nopenslots = # of slots open across all active DMs ###
    !### nslotspdm = # of slots each active DM has, nparticles = # of active particles across all DMs ###
    !### navailslots = # of particles that can be added until BM is full, isLive = object is constructed, hasDMs = object has allocated DMs list ###
    !### hasSlots = one or more active DMs has available slots, isFilled = all DMs are active and filled ###
    INTEGER(WIS), PUBLIC :: ndim, ntotdms, nactive, nopenslots, nslotspdm, navailslots
    LOGICAL(WIS), PUBLIC :: is1d, is2d, is3d, isLive, hasDMs, hasSlots, isFilled

    CONTAINS

    !### default everything as private ###
    PRIVATE

    PROCEDURE :: initdmblock_manager
    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: add_particles
    PROCEDURE, PUBLIC :: defragment
    PROCEDURE, PUBLIC :: find_out_of_bounds
    PROCEDURE, PUBLIC :: copy_MPI_buffer_out
    PROCEDURE, PUBLIC :: copy_MPI_buffer_in
    PROCEDURE, PUBLIC :: build
    PROCEDURE, PUBLIC :: unbuild

END TYPE DMBlock_Manager

!### create an interface to the constructor ###
INTERFACE DMBlock_Manager

    MODULE PROCEDURE constructor

END INTERFACE DMBlock_Manager

!### object methods ###
CONTAINS

!### a constructor for the class ###
FUNCTION constructor(ndim, ntotdms, nslotspdm, xLow, xHi, yLow, yHi, zLow, zHi)

    TYPE(DMBlock_Manager) :: constructor
    INTEGER(WIS), INTENT(IN) :: ndim, ntotdms, nslotspdm
    REAL(WRD), INTENT(IN) :: xLow, xHi, yLow, yHi, zLow, zHi

    CALL constructor%initdmblock_manager(ndim, ntotdms, nslotspdm, xLow, xHi, yLow, yHi, zLow, zHi)
    RETURN

END FUNCTION constructor

!### actual initialization of object (not directly called by user) ###
SUBROUTINE initdmblock_manager(self, ndim, ntotdms, nslotspdm, xLow, xHi, yLow, yHi, zLow, zHi)

    CLASS(DMBlock_Manager) :: self
    INTEGER(WIS), INTENT(IN) :: ntotdms, ndim, nslotspdm
    REAL(WRD), INTENT(IN) :: xLow, xHi, yLow, yHi, zLow, zHi

    self%ndim         = ndim
    self%ntotdms      = ntotdms
    self%nslotspdm    = nslotspdm
    self%nopenslots   = 0
    self%nactive      = 0
    self%navailslots  = ntotdms * nslotspdm
    self%xLow         = xLow
    self%xHi          = xHi
    self%yLow         = yLow
    self%yHi          = yHi
    self%zLow         = zLow
    self%zHi          = zHi
    self%is1d         = .FALSE.
    self%is2d         = .FALSE.
    self%is3d         = .FALSE.
    IF (self%ndim .EQ. 1) self%is1d = .TRUE.
    IF (self%ndim .EQ. 2) self%is2d = .TRUE.
    IF (self%ndim .EQ. 3) self%is3d = .TRUE.

    !### register as Live but no active DMs ###
    self%isLive     = .TRUE.
    self%hasDMs     = .FALSE.
    self%hasSlots   = .FALSE.
    self%isFilled   = .FALSE.


END SUBROUTINE initdmblock_manager

!### create our DM list but do not actually build up each DM block (deferred to as-needed) ###
SUBROUTINE build(self)

    CLASS(DMBlock_Manager)  :: self
    !### check that we are live but do not already have DMs, if so allocate DM list and set hasDMs flag true ###
    IF (self%isLive .AND. .NOT. self%hasDMs) THEN
        ALLOCATE(self%dms(self%ntotdms))
        self%hasDMs     = .TRUE.
    END IF

    IF (.NOT. ALLOCATED(self%lens)) THEN
        IF (self%is1d) THEN
            ALLOCATE(self%lens(2))
        ELSE IF (self%is2d) THEN
            ALLOCATE(self%lens(8))
        ELSE IF (self%is3d) THEN
            ALLOCATE(self%lens(26))
        END IF

        self%lens = 0
    END IF

END SUBROUTINE build

!### clean up all of our active DMs and deallocate the DM list ###
SUBROUTINE unbuild(self)

    CLASS(DMBlock_Manager) :: self

    IF (self%isLive .AND. ALLOCATED(self%dms)) THEN
        DEALLOCATE(self%dms)
        DEALLOCATE(self%lens)
        self%hasDMs  = .FALSE.
    END IF

END SUBROUTINE unbuild

!### destroy the object and flag that it is not constructed ###
SUBROUTINE destroy(self)

    CLASS(DMBlock_Manager) :: self

    CALL self%unbuild()
    self%isLive = .FALSE.

END SUBROUTINE destroy

!### for input paticle list, build new dms if necessary and copy particles into open slots. If this is the first pass we set the isactive flag for the
!    particle to TRUE, if this is not the first pass we set the keep isactive as FALSE and set info = -1 (this is done to distinguish
!    newly added particles during this pass for the assign_density procedure
SUBROUTINE add_particles(self, particlelist, ntotlist, first_pass)

    CLASS(DMBlock_Manager) :: self
    REAL(WRD), INTENT(IN) :: particlelist(:,:) !### indices: 1 - particle number ; 2 - pos,vel,acc ###
    INTEGER(WIS), INTENT(IN) :: ntotlist !### ntotlist - # of input particles ###
    LOGICAL(WIS), INTENT(IN) :: first_pass
    INTEGER(WIS) :: passes2go, newblocks !### passes2go - # still to be added to DM blocks ###
    INTEGER(WIS) :: ii,bi,si !### ii - input index, bi - block index, si - slot index ###

    passes2go  = ntotlist

    IF (ntotlist.GT.self%nopenslots) THEN !### new DM blocks necessary to handle incoming particles ###

       newblocks = CEILING( REAL(ntotlist - self%nopenslots) / REAL(self%nslotspdm) )

       !### construct and build new DM blocks ###
       DO bi = self%nactive + 1, self%nactive + newblocks
          self%dms(bi) = DM(self%ndim, self%nslotspdm)
          CALL self%dms(bi)%build()
       END DO

       !### update accounting for block manager ###
       self%nactive = self%nactive + newblocks
       self%nopenslots = self%nopenslots + newblocks * self%nslotspdm

    END IF

    ii = 1
    !### loop though active blocks, copying in new particles ###
    DO bi = 1, self%nactive

       IF (self%dms(bi)%hasSlots) THEN  !### DM block has open slots ###

           IF (passes2go .GE. self%dms(bi)%nopenslots) THEN !### input particles will fill current DM block ###

             !### loop through slots of current block, searching for empty ones ###
             DO si =1, self%nslotspdm

                IF (.NOT. self%dms(bi)%isactive(si)) THEN !### slot is empty, copy in particle and increment input index ###
                   self%dms(bi)%particles(si,:) = particlelist(ii,:)
                   IF (first_pass) THEN
                      self%dms(bi)%isactive(si) = .TRUE.
                   ELSE
                      self%dms(bi)%info(si) = -1
                   END IF
                   ii = ii + 1
                END IF
             END DO

             !### bookkeeping ###
             passes2go = passes2go - self%dms(bi)%nopenslots

             self%nopenslots         = self%nopenslots - self%dms(bi)%nopenslots
             self%navailslots        = self%navailslots - self%dms(bi)%nopenslots
             self%dms(bi)%nopenslots = 0
             self%dms(bi)%hasSlots   = .FALSE.

          ELSE IF(passes2go .GT. 0) THEN      !### still particles left to copy, but more slots in dms(bi) than particles left in input list ###
             si = 0

             !### bookkeeping ###
             self%nopenslots         = self%nopenslots - passes2go
             self%navailslots        = self%navailslots - passes2go
             self%dms(bi)%nopenslots = self%dms(bi)%nopenslots - passes2go

             DO WHILE (passes2go.GT.0) !### loop through slots of current block until no input particles remain
                si = si + 1
                IF (.NOT.self%dms(bi)%isactive(si)) THEN !### slot is empty, copy in particle, increment input index, and decrement passes2go ###
                   self%dms(bi)%particles(si,:) = particlelist(ii,:)
                   self%dms(bi)%isactive(si)    = .TRUE.
                   ii = ii + 1
                   passes2go  = passes2go - 1
                END IF
             END DO
          END IF
       END IF
    END DO

    IF (self%nopenslots .LE. 0) THEN
       self%hasSlots = .FALSE.
    END IF 

    IF (self%navailslots .LE. 0) THEN
       self%isFilled = .TRUE.
    END IF

END SUBROUTINE add_particles

!### move particles to the front of block manager to fill in empty slots, destroying empty dms
SUBROUTINE defragment(self)

    CLASS(DMBlock_Manager) :: self
    INTEGER(WIS) :: fbi,fsi,bbi,bsi       !### fbi - forward block index; fsi - forward slot index; bbi - backward block index; bsi - backward slot index ###

    fbi = 1
    bbi = self%nactive

    IF (self%hasSlots .AND. (self%nactive.GT.1)) THEN
blklp:   DO WHILE ( bbi.GT.fbi )   !### cycle from front and back, meet somewhere in the middle ###
            DO bsi = self%nslotspdm, 1, -1    !### loop from last slot in backward block ###
               IF (self%dms(bbi)%isactive(bsi)) THEN    !### if particle is active, to be moved to first open slot of forward block ###

                  DO WHILE (.NOT. self%dms(fbi)%hasSlots .AND. fbi.LT.self%nactive)   !### find first available DM block with open slots ###
                     fbi = fbi + 1
                     IF (fbi .EQ. bbi) EXIT blklp
                     fsi = 1
                  END DO

                  DO WHILE  (self%dms(fbi)%isactive(fsi) .AND. fsi.LT.self%nslotspdm)   !### loop through slots to find an open one ###
                     fsi = fsi + 1
                  END DO

                  !### copy particle from back to front, take care of accounting ###
                  self%dms(fbi)%particles(fsi,:) = self%dms(bbi)%particles(bsi,:)
                  self%dms(fbi)%isactive(fsi) = .TRUE.
                  self%dms(fbi)%outofbounds(fsi) = .FALSE.
                  self%dms(fbi)%info(fsi) = 0
                  self%dms(fbi)%nopenslots = self%dms(fbi)%nopenslots - 1
                  self%dms(bbi)%isactive(bsi) = .FALSE.
                  self%dms(bbi)%outofbounds(bsi) = .FALSE.
                  self%dms(bbi)%info(bsi) = 0
                  self%dms(bbi)%nopenslots = self%dms(bbi)%nopenslots + 1
                  !### Block is full, advance to next forward block ###
                  IF (self%dms(fbi)%nopenslots.LE.0) THEN
                     self%dms(fbi)%hasSlots = .FALSE.
                     fbi = fbi + 1
                     IF (fbi .EQ. bbi) EXIT blklp
                     fsi = 1
                  END IF
               END IF
            END DO
            bbi = bbi - 1

         END DO blklp

         !### destroy empty dm blocks
         DO bbi = self%nactive, 1, fbi
            IF (self%dms(bbi)%nopenslots .EQ. self%nslotspdm) THEN
               CALL self%dms(bbi)%destroy()
               self%nactive = self%nactive - 1
            END IF
         END DO

         !### all blocks in this dmbm will be filled now except maybe for the last one ###
         self%nopenslots = self%dms(self%nactive)%nopenslots
         self%hasSlots = self%dms(self%nactive)%hasSlots
    END IF

END SUBROUTINE defragment

!### search all active DMs for active particles out of this patch's bounds, save info for OOB particles, on exit total # of elements
!    to be sent to each neighbor from all dms in this dmbm is stored in self%lens() ###
SUBROUTINE find_out_of_bounds(self, rankloc, simunits)

    CLASS(DMBlock_Manager) :: self
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    INTEGER(WIS) :: noob(26)    !### - # out of bounds
    INTEGER(WIS) :: bi, si

    self%lens = 0
    IF (.NOT. self%hasDMs ) RETURN


    IF(self%is1d) THEN  !### 2 possible destinatons ###
       DO bi = 1, self%nactive
         noob = 0 
         DO si = 1, self%nslotspdm
             IF(self%dms(bi)%isactive(si)) THEN
                !### towards xLow: info = 1 ###
                IF(self%dms(bi)%particles(si,1).LT.self%xLow) THEN
                   noob(1) = noob(1) + 1
                   self%dms(bi)%info(si) = 1
                   self%dms(bi)%outofbounds(si) = .TRUE.
                !### towards xHi: info = 2 ###
                ELSE IF(self%dms(bi)%particles(si,1).GT.self%xHi) THEN
                   noob(2) = noob(2) + 1
                   self%dms(bi)%info(si) = 2
                   self%dms(bi)%outofbounds(si) = .TRUE.
                END IF
             END IF
          END DO
          self%lens = self%lens + noob(1:2) * 3
       END DO

    ELSE IF(self%is2d) THEN !### 8 possible destinatons
       DO bi = 1, self%nactive
          noob = 0
          DO si = 1, self%nslotspdm
            IF(self%dms(bi)%isactive(si)) THEN

                !### towards xLow ...###
                IF(self%dms(bi)%particles(si,1).LT.self%xLow) THEN
  
                   !### ... and yLow: info = 5 ###
                   IF(self%dms(bi)%particles(si,2).LT.self%yLow) THEN
                       noob(5) = noob(5) + 1
                       self%dms(bi)%info(si) = 5
                       self%dms(bi)%outofbounds(si) = .TRUE.
  
                   !### ... and yHi: info = 7 ###
                   ELSE IF(self%dms(bi)%particles(si,2).GT.self%yHi) THEN
                       noob(7) = noob(7) + 1
                       self%dms(bi)%info(si) = 7
                       self%dms(bi)%outofbounds(si) = .TRUE.
  
                   !### ... and only xLow: info = 1 ###
                   ELSE
                       noob(1) = noob(1) + 1
                       self%dms(bi)%info(si) = 1
                       self%dms(bi)%outofbounds(si) = .TRUE.

                   END IF
                !### towards xHi ###
                ELSE IF(self%dms(bi)%particles(si,1).GT.self%xHi) THEN

                   !### ... and yLow: info = 6 ###
                   IF(self%dms(bi)%particles(si,2).LT.self%yLow) THEN
                       noob(6) = noob(6) + 1
                       self%dms(bi)%info(si) = 6
                       self%dms(bi)%outofbounds(si) = .TRUE.

                   !### ... and yHi: info = 8 ###
                   ELSE IF(self%dms(bi)%particles(si,2).GT.self%yHi) THEN
                       noob(8) = noob(8) + 1
                       self%dms(bi)%info(si) = 8
                       self%dms(bi)%outofbounds(si) = .TRUE.

                   !### ... and only xHi: info = 2 ###
                   ELSE
                       noob(2) = noob(2) + 1
                       self%dms(bi)%info(si) = 2
                       self%dms(bi)%outofbounds(si) = .TRUE.

                   END IF
  
                ELSE !X!
                   !### towards yLow alone: info = 3 ###
                   IF(self%dms(bi)%particles(si,2).LT.self%yLow) THEN
                       noob(3) = noob(3) + 1
                       self%dms(bi)%info(si) = 3
                       self%dms(bi)%outofbounds(si) = .TRUE.
  
                   !### towards yHi alone: info = 4 ###
                   ELSE IF(self%dms(bi)%particles(si,2).GT.self%yHi) THEN
                       noob(4) = noob(4) + 1
                       self%dms(bi)%info(si) = 4
                       self%dms(bi)%outofbounds(si) = .TRUE.

                   END IF
                END IF
             ENDIF
          END DO
          self%lens = self%lens + noob(1:8) * 6
       END DO

    ELSE IF(self%is3d) THEN
       DO bi = 1, self%nactive
          noob = 0
          DO si = 1, self%nslotspdm
             IF (self%dms(bi)%isactive(si)) THEN

                !### towards xLo ... ###
                IF(self%dms(bi)%particles(si,1).LT.self%xLow) THEN

                   !### ..., yLo, ... ###
                   IF(self%dms(bi)%particles(si,2).LT.self%yLow) THEN

                      !### ... and zLo (-x-y-z): info 14 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(14) = noob(14) + 1
                         self%dms(bi)%info(si) = 14
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (-x-y+z): info 10  ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(10) = noob(10) + 1
                         self%dms(bi)%info(si) = 10
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (-x-y): info 26  ###
                      ELSE
                         noob(26) = noob(26) + 1
                         self%dms(bi)%info(si) = 26
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   !### ..., yHi, ... ###
                   ELSE IF(self%dms(bi)%particles(si,2).GT.self%yHi) THEN

                      !### ... and zLo (-x+y-z): info 7 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(7) = noob(7) + 1
                         self%dms(bi)%info(si) = 7
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (-x+y+z): info 11 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(11) = noob(11) + 1
                         self%dms(bi)%info(si) = 11
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (-x+y): info 23 ###
                      ELSE
                         noob(23) = noob(23) + 1
                         self%dms(bi)%info(si) = 23
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   ELSE !Y!

                      !### ... and zLo (-x-z): info 19 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(19) = noob(19) + 1
                         self%dms(bi)%info(si) = 19
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (-x+z): info 21 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(21) = noob(21) + 1
                         self%dms(bi)%info(si) = 21
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (-x): info 1 ###
                      ELSE
                         noob(1) = noob(1) + 1
                         self%dms(bi)%info(si) = 1
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   END IF

                !### towards xHi ... ###
                ELSE IF (self%dms(bi)%particles(si,1).GT.self%xHi) THEN

                   !### ..., yLo, ... ###
                   IF(self%dms(bi)%particles(si,2).LT.self%yLow) THEN

                      !### ... and zLo (+x-y-z): info 12 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(12) = noob(12) + 1
                         self%dms(bi)%info(si) = 12
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (+x-y+z): info 8 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(8) = noob(8) + 1
                         self%dms(bi)%info(si) = 8
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (+x-y):  info 24 ###
                      ELSE
                         noob(24) = noob(24) + 1
                         self%dms(bi)%info(si) = 24
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   !### ..., yHi, ... ###
                   ELSE IF(self%dms(bi)%particles(si,2).GT.self%yHi) THEN

                      !### ... and zLo (+x+y-z): info 9 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(9) = noob(9) + 1
                         self%dms(bi)%info(si) = 9
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi(+x+y+z): info 13 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(13) = noob(13) + 1
                         self%dms(bi)%info(si) = 13
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (+x+y): info 25 ###
                      ELSE
                         noob(25) = noob(25) + 1
                         self%dms(bi)%info(si) = 25
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   ELSE !Y!

                      !### ... and zLo (+x-z): info 22 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(22) = noob(22) + 1
                         self%dms(bi)%info(si) = 22
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (+x+z): info 20 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(20) = noob(20) + 1
                         self%dms(bi)%info(si) = 20
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (+x): info 2 ###
                      ELSE
                         noob(2) = noob(2) + 1
                         self%dms(bi)%info(si) = 2
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   END IF

                ELSE !X!

                   !### towards yLo ... ###
                   IF(self%dms(bi)%particles(si,2).LT.self%yLow) THEN

                      !### ... and zLo (-y-z): info  18 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(18) = noob(18) + 1
                         self%dms(bi)%info(si) = 18
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (-y+z): info 16 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(16) = noob(16) + 1
                         self%dms(bi)%info(si) = 16
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (-y): info 3 ###
                      ELSE
                         noob(3) = noob(3) + 1
                         self%dms(bi)%info(si) = 3
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   !### towards yHi ... ###
                   ELSE IF(self%dms(bi)%particles(si,2).GT.self%yHi) THEN

                      !### ... and zLo (+y-z): info 15 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(15) = noob(15) + 1
                         self%dms(bi)%info(si) = 15
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### ... and zHi (+y+z): info 17 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(17) = noob(17) + 1
                         self%dms(bi)%info(si) = 17
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### that's it (+y): info 4 ###
                      ELSE
                         noob(4) = noob(4) + 1
                         self%dms(bi)%info(si) = 4
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF

                   ELSE !Y!

                      !### towards zLo (-z): info 5 ###
                      IF(self%dms(bi)%particles(si,3).LT.self%zLow) THEN
                         noob(5) = noob(5) + 1
                         self%dms(bi)%info(si) = 5
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      !### towards zHi (+z): info 6 ###
                      ELSE IF(self%dms(bi)%particles(si,3).GT.self%zHi) THEN
                         noob(6) = noob(6) + 1
                         self%dms(bi)%info(si) = 6
                         self%dms(bi)%outofbounds(si) = .TRUE.

                      END IF
                   END IF
                END IF
             END IF
          END DO
          self%lens = self%lens + noob * 9
       END DO
    END IF

END SUBROUTINE find_out_of_bounds

!### copy_MPI_buffer_out - called by DMman object copy_MPI_buffer_out 
SUBROUTINE copy_MPI_buffer_out(self, dest, buffer, n2pass, curoff)

    CLASS(DMBlock_Manager) :: self
    REAL(WRD), INTENT(INOUT) :: buffer(:)
    INTEGER(WIS), INTENT(IN)  :: dest, n2pass, curoff   
!### dest - neighbor we are sending particles to; n2pass - total # being passed to this dest across all dmbms during current update_pass; curoff - current location in buffer
    INTEGER(WIS) :: nmoved, nmovedsofar, nmovedthisblock, nvars
    INTEGER(WIS) :: bi, si, vi, ii

    nvars = self%ndim * 3
    nmoved = 0

    !### move particles into buffer arranged as contiguous array by variable
    DO bi = 1, self%nactive
       nmovedsofar = nmoved
       DO si = 1, self%nslotspdm
          !### check that this particle is out-of-bounds and destined for the appropriate neighbor
          IF(nmoved .LT. n2pass .AND. self%dms(bi)%outofbounds(si) .AND. self%dms(bi)%info(si) .EQ. dest) THEN
             nmoved = nmoved + 1
             self%dms(bi)%isactive(si)    = .FALSE.
             self%dms(bi)%outofbounds(si) = .FALSE.
             DO vi = 1, nvars
                ii = curoff + nmoved + (vi-1)*n2pass
                buffer(ii) = self%dms(bi)%particles(si,vi)
             END DO
          END IF
       END DO
       nmovedthisblock = nmoved - nmovedsofar
       self%dms(bi)%nopenslots = self%dms(bi)%nopenslots + nmovedthisblock
       IF (self%dms(bi)%nopenslots .GT. 0) self%dms(bi)%hasSlots = .TRUE.
       IF (nmoved .EQ. n2pass) EXIT
    END DO

    self%nopenslots  = SUM(self%dms(:)%nopenslots)
    self%navailslots = self%nopenslots + (self%ntotdms - self%nactive) * self%nslotspdm
    self%lens(dest)  = self%lens(dest) - (nmoved * nvars)

    self%hasSlots = .FALSE.
    self%isFilled = .TRUE.

    IF (self%nopenslots .GT. 0) THEN
       self%hasSlots = .TRUE.
    END IF

    IF (self%navailslots .GT. 0) THEN
       self%isFilled = .FALSE.
    END IF

END SUBROUTINE copy_MPI_buffer_out

!### takes slice of contiguous MPI buffer (cast as a 2D array) and calls self%add_particles to store particles in slots
SUBROUTINE copy_MPI_buffer_in(self, buffer, totalpass, passedsofar, thispass, first_pass)
!### the reverse of copying out of an MPI buffer could re-use the add_particles routine.  the only difference might be whether or not the input to the first routine works on a 1d or multi-d input array.

    CLASS(DMBlock_Manager) :: self

    INTEGER(WIS), INTENT(IN) :: totalpass, passedsofar, thispass
    REAL(WRD),INTENT(IN) :: buffer(totalpass, 3*self%ndim)
    LOGICAL(WIS), INTENT(IN) :: first_pass
    INTEGER(WIS) :: start, end

    start = passedsofar + 1
    end = passedsofar + thispass

    CALL self%add_particles(buffer(start:end,:), totalpass, first_pass)

END SUBROUTINE copy_MPI_buffer_in
END MODULE Mod_DMBlock_Manager
