#include "macro.h"

MODULE Mod_Error

! FILENAME: mod_error.f90
!
! DESCRIPTION: This module provides routines for error messages
!
! DEPENDENCIES:
!
! HISTORY:
!    5/7/12  - Peter Mendygral
!    12/11/17 - Julius Donnert

USE Mod_Globals

IMPLICIT NONE

CONTAINS

SUBROUTINE Error(source, msg)

    CHARACTER(*) :: source
    CHARACTER(*) :: msg

    PRINT '("[", A, " ERROR] (Rank ", A, " | Thread ", A,") :: ", A)', &
            TRIM(source), I2A(WRANK), I2A(WTHREAD), TRIM(msg)
    STOP

END SUBROUTINE Error

SUBROUTINE Message(source, msg)

    CHARACTER(*) :: source
    CHARACTER(*) :: msg

#ifdef DEBUG
    PRINT '("DEBUG MODE [", A, " MESSAGE] (Rank ", A, " | Thread ", A, ") :: ", A)', &
            TRIM(source), I2A(WRANK),  I2A(WTHREAD), TRIM(msg)
#else
    PRINT '("[", A, " MESSAGE] (Rank ", A5, " | Thread ", A3, ") :: ", A)', &
            TRIM(source), I2A(WRANK),  I2A(WTHREAD), TRIM(msg)
#endif

END SUBROUTINE Message

SUBROUTINE Assert_Info(expr, msg, srcfile, line)
    
    LOGICAL(WIS) :: expr
    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    IF (expr) RETURN
        
    PRINT '(2A,"[",A5,":",A3,"] ASSERTION FAILED ", A,":", I4,"   ",A)', NEW_LINE('A'), &
           NEW_LINE('A'),I2A(WRANK),I2A(WTHREAD),TRIM(srcfile),line, NEW_LINE('A')
    PRINT '("               ",A)', TRIM(msg)
    PRINT '("ABORTING ...")'

    STOP
END SUBROUTINE

SUBROUTINE Assert_Info_Root(expr, msg, srcfile, line)
    
    LOGICAL(WIS) :: expr
    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    IF (expr) RETURN
        
    IF (WRANK .EQ. 0) THEN
        PRINT '(2A,"ASSERTION FAILED ON ROOT ", A,":", I4,"   ",A)', NEW_LINE('A'), NEW_LINE('A'), &
                                                          TRIM(srcfile), line, NEW_LINE('A')     
        PRINT '("               ",A)', TRIM(msg)
        PRINT '("ABORTING ...")'
    ENDIF

    STOP
END SUBROUTINE

! Print a warning, use I2A and ftoa to construct string
SUBROUTINE Warn_Info(expr, msg, srcfile, line)
    
    LOGICAL(WIS) :: expr
    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    IF (.NOT. expr) RETURN

    PRINT '(A1,"[",A,":",A,"] WARNING ", A,":", I4," - ", A,A1)', NEW_LINE('A'), &
            I2A(WRANK), I2A(WTHREAD), TRIM(srcfile), line, TRIM(msg), NEW_LINE('A')      
    
END SUBROUTINE

SUBROUTINE Warn_Info_Root(expr, msg, srcfile, line)
    
    LOGICAL(WIS) :: expr
    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    IF (.NOT. expr .OR. WRANK .NE. 0) RETURN

    PRINT '(A1,"[",A,":",A,"] WARNING ", A,":", I4," - ", A,A1)', NEW_LINE('A'), &
            I2A(WRANK), I2A(WTHREAD), TRIM(srcfile), line, TRIM(msg), NEW_LINE('A')      
    
END SUBROUTINE

! Print a message, use I2A and ftoa to construct string
SUBROUTINE Msg_Info(msg, srcfile, line)

    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    PRINT '("[",A5,":",A3,"] ", A,":", I4," - ", A)', &
            I2A(WRANK), I2A(WTHREAD), TRIM(srcfile), line, TRIM(msg)      

END SUBROUTINE

! Print a message on Rank 0 thread 0, use I2A and ftoa to construct string
SUBROUTINE Msg_Info_Root(msg, srcfile, line)
    
    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    IF (WRANK .EQ. 0 .AND. WTHREAD .EQ. 0) THEN
        PRINT '(A)', TRIM(msg)      
    END IF

END SUBROUTINE

SUBROUTINE Debugmsg_Info(msg, srcfile, line)
    
    CHARACTER(*) :: srcfile
    INTEGER(4) :: line
    CHARACTER(*) :: msg

    IF (WDEBUG .AND. WRANK .EQ. 0) THEN
        PRINT '("[",A,":",A,"] ", A,":", I4, " - " , A)', I2A(WRANK), &
              I2A(WTHREAD), TRIM(srcfile), line, TRIM(msg)  
    END IF

END SUBROUTINE

! conversion routines to character array, does not require to be wrapped into a TRIM() call

FUNCTION i2a(i) RESULT(str)

    INTEGER(WIS), INTENT(IN) :: i
    CHARACTER(WCHAR_BUF_SIZE) ::  str
    
    WRITE(str,'(1I11)') i

END FUNCTION i2a

FUNCTION l2a(i) RESULT(str)

    INTEGER(WID), INTENT(IN) :: i
    CHARACTER(WCHAR_BUF_SIZE) ::  str
    
    WRITE(str,'(1I21)') i

END FUNCTION l2a

! if you do not need it accurately
FUNCTION f2a(f) RESULT(str)

    REAL(WRD), INTENT(IN) :: f
    CHARACTER(WCHAR_BUF_SIZE) ::  str,fmtstr
    INTEGER(WID) :: n

    IF (f .EQ. 0.0_WRD) THEN
        str = "0.0"
    ELSE  
        n = ABS(FLOOR(LOG10(ABS(f))))

        IF (n .GT. 6) THEN
            fmtstr = "(1E12.4)"
        ELSE        
            IF (f .GT. 1) THEN
                WRITE(fmtstr,'("(1F12.",I1,")")') 9-n
            ELSE 
                WRITE(fmtstr,'("(1F12.",I1,")")') 9
            END IF
        END IF
        
        WRITE(str, fmtstr) f
    END IF

END FUNCTION f2a

! if you do need it accurately
FUNCTION e2a(e) RESULT(str)

    REAL(WRD), INTENT(IN) :: e
    CHARACTER(WCHAR_BUF_SIZE) :: tmp, str
    
    WRITE(str,'(1E23.16)') e

END FUNCTION e2a

END MODULE Mod_Error
