#include "macro.h" 

MODULE Mod_Globals

! FILENAME : globals.f90
!
! DESCRIPTION: This module defines global values for use throughout Wombat.
!
! DEPENDENCIES  : None
!
! HISTORY :  
!    07/08/2017  - Julius Donnert
!    07/29/2015  - Peter Mendygral

USE MPI ! need MPI_MAX_PROCESSOR_NAME

IMPLICIT NONE

REAL, PARAMETER :: WVERSION = 1.0

INTEGER, PARAMETER :: & 

     WRS = KIND(1.0),                                   &    ! define the precision for data 
     WRD = SELECTED_REAL_KIND(2*PRECISION(1.0_WRS)),    &    ! these may need to be augmented for compilers and/or systems
     WRQ = SELECTED_REAL_KIND(2*PRECISION(1.0_WRD)),    &    ! that do not support up to quad precision
     WIS = WRS,                                         &    ! in that case, you CAN change WRQ and WIQ (real and
     WID = WRD,                                         &    ! integer) types to be whatever at the cost of who knows?
     WIQ = WRQ,                                         &    !
#if (defined(__GFORTRAN__))
     WOMP_LOCK_SIZE = 4,                                &    !  GCC before version 6 uses INT(4) for an OMP lock ?
#else
     WOMP_LOCK_SIZE = 8,                                &    
#endif
     WADDRESS_KIND  = 8,                                &    ! we need a type that can be used for long indexing 
                                                             ! independent of precision elsewhere
     WLB_HEADER_SIZE              = 4,                  &    ! size of WOMBAT's internal MPI header 
                                                             ! (PatchID(3), flavor, more to come)
     WMPI_MAX_MESSAGE             = 1024,               &    ! the maximum MPI message size in MBytes
     WRESTART_IOBUFFER_SIZE       = 512,                &    ! the default IO buffer size for restarts in MBytes
     WCOMPRESSED_IOBUFFER_SIZE    = 64,                 &    ! the default IO buffer size for compressed files in MBytes
     WMOVIE_IOBUFFER_SIZE         = 16,                 &    ! the default IO buffer size for movie files in MBytes
     WNUM_AIO_OPS_INFLIGHT        = 1,                  &    ! the maximum number of AIO operations in flight per flavor
     WMAX_IO_FIELDS               = 99,                 &    ! the maximum number of output fields allowed 
     WMAX_MG_LEVELS               = 32,                 &    ! the maximum number of levels that can be used in the multigrid solver
     WCHAR_BUF_SIZE               = 256,                &    ! length of a standard character array

     ! solver flavors
     FL_NULL            = 0,          & ! NULL flavor
     FL_MHDTVD          = 1,          & ! MHD TVD solver first pass
     FL_MHDTVDCT        = 2,          & ! second pass EMF/CT scheme
     FL_PASSIVE         = 3,          & ! passive/color variables
     FL_COSMIC_RAYS     = 4,          & ! Jones et al (YEAR?) CR solver
     FL_DM_KICKDRIFT    = 5,          & ! First Kick & Drift of the DM leapfrog scheme
     FL_DM_PARTICLES    = 6,          & ! Particle variables, e.g. density assignment 
     FL_GRAV_FMG        = 7,          & ! Gravitational potential from full multi grid solver
     FL_DM_KICK2        = 8,          & ! Second Kick of particle leapfrog scheme
     FL_MHDWENO         = 9,          & ! MHD Weno5 solver - state vector update

     NFLAVORS           = FL_MHDWENO    ! This should always be the last in the list

CHARACTER(WCHAR_BUF_SIZE) :: USERNAME
CHARACTER(MPI_MAX_PROCESSOR_NAME) :: HOSTNAME
  
INTEGER(WIS) :: WRANK=0, WTHREAD = 0, WNTHREADS = 1              ! many modules need the rank & thread ID for a variety of reasons
CHARACTER(WCHAR_BUF_SIZE) :: WRANKSTR="0000", WTHREADSTR="00", WNTHREADSTR='1'   ! string that contains the MPI rank and thread ID of the process.
!$OMP THREADPRIVATE(WTHREADSTR, WTHREAD, WNTHREADSTR)            ! declare thread private globals as such

REAL(WRD), PARAMETER :: WSWEBY_BETA       = 1.08_WRD,   &    ! the beta parameter to be used in Sweby slope limiters
                        WSIGNAL_TOLERANCE = 0.5_WRD          ! since we pass REALs for signalling, we need some tolerance 
                                                             ! around them (less than 1) to be sure we don't falsely trigger on
                                                             ! on tiny FP differences
! set the debug mode flag
#ifdef DEBUG 
LOGICAL, PARAMETER :: WDEBUG = .TRUE.
#else
LOGICAL, PARAMETER :: WDEBUG = .FALSE.
#endif

! set the parameter output file mode flag
#ifdef GENPARFILE
LOGICAL, PARAMETER :: GENPARFILE_ON = .TRUE.        ! generate a valid parameter file
#else
LOGICAL, PARAMETER :: GENPARFILE_ON = .FALSE.       ! read a parameter file
#endif

! set some constants
REAL(8), PARAMETER :: PI = 2.0_WRD * ASIN(1.0_WRD)

CONTAINS

! NOTE, throughout the code I have used the !$OMP ... COLLAPSE clause.  This is relatively
! new, and it seems both Intel and CCE have real issues with it.  Intel actually produced
! garbarge indicies causing bus errors.  CCE just got wrong answers when using collapses even
! single threaded.  Thus, I am implementing my own collapse features here.

! given a 2d loop determine the i,j values given a single index and ranges 
SUBROUTINE collapse2d(n, nx, ny, i, j)

    INTEGER(WIS), INTENT(IN) :: n, nx, ny
    INTEGER(WIS), INTENT(OUT) :: i, j

    i = MOD(n, nx)
    IF (i .EQ. 0) i = nx
    j = CEILING(REAL(n, WRD) / REAL(nx, WRD))

END SUBROUTINE collapse2d

! given a 3d loop determine the i,j,k values given a single index and ranges 
SUBROUTINE collapse3d(n, nx, ny, nz, i, j, k)

    INTEGER(WIS), INTENT(IN) :: n, nx, ny, nz
    INTEGER(WIS), INTENT(OUT) :: i, j, k
    INTEGER(WIS) :: myn

    i = MOD(n, nx)
    IF (i .EQ. 0) i = nx
    k = CEILING(REAL(n, WRD) /  REAL(nx*ny, WRD))
    myn = n - (k-1)*nx*ny
    j = CEILING(REAL(myn, WRD) / REAL(nx, WRD))

END SUBROUTINE collapse3d

! handle optional arguments with a function call

FUNCTION handle_opt_logical(stdval, optval) RESULT(val)

    LOGICAL, INTENT(IN) :: stdval
    LOGICAL, INTENT(IN),OPTIONAL :: optval
    LOGICAL :: val

    IF (PRESENT(optval)) THEN
        val = optval
    ELSE
        val = stdval
    END IF

END FUNCTION handle_opt_logical

FUNCTION handle_opt_int8(stdval, optval) RESULT(val)

    INTEGER(WID), INTENT(IN) :: stdval
    INTEGER(WID), INTENT(IN),OPTIONAL :: optval
    INTEGER(WID) :: val

    IF (PRESENT(optval)) THEN
        val = optval
    ELSE
        val = stdval
    END IF

END FUNCTION handle_opt_int8

FUNCTION handle_opt_int4(stdval, optval) RESULT(val)

    INTEGER(WIS), INTENT(IN) :: stdval
    INTEGER(WIS), INTENT(IN),OPTIONAL :: optval
    INTEGER(WIS) :: val

    IF (PRESENT(optval)) THEN
        val = optval
    ELSE
        val = stdval
    END IF

END FUNCTION handle_opt_int4

FUNCTION handle_opt_real8(stdval, optval) RESULT(val)

    REAL(WRD), INTENT(IN) :: stdval
    REAL(WRD), INTENT(IN),OPTIONAL :: optval
    REAL(WRD) :: val

    IF (PRESENT(optval)) THEN
        val = optval
    ELSE
        val = stdval
    END IF

END FUNCTION handle_opt_real8

FUNCTION handle_opt_real4(stdval, optval) RESULT(val)

    REAL(WRS), INTENT(IN) :: stdval
    REAL(WRS), INTENT(IN),OPTIONAL :: optval
    REAL(WRS) :: val

    IF (PRESENT(optval)) THEN
        val = optval
    ELSE
        val = stdval
    END IF

END FUNCTION handle_opt_real4

FUNCTION handle_opt_character(stdval, optval) RESULT(val)

    CHARACTER(*), INTENT(IN) :: stdval
    CHARACTER(*), INTENT(IN),OPTIONAL :: optval
    CHARACTER(WRD) :: val

    IF (PRESENT(optval)) THEN
        val = optval
    ELSE
        val = stdval
    END IF

END FUNCTION handle_opt_character

END MODULE Mod_Globals
