#include "macro.h"

MODULE Mod_IO

! FILENAME: io.f90
!
! DESCRIPTION: Write out a snapshot using AIO or direct output given a 
!              dump description.
!
! DEPENDENCIES: io_dump.f90
!
! HISTORY:
!    7/8/17  - Julius Donnert
!    4/19/14  - Peter Mendygral

USE OMP_LIB

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Domain
USE Mod_IO_Dump

IMPLICIT NONE

PRIVATE

PUBLIC :: IO
 
TYPE :: IO ! data type for this module

    PRIVATE
    
    CHARACTER(WCHAR_BUF_SIZE), PUBLIC :: simulationTitle
    CHARACTER(2048)                   :: simulationDescription

    LOGICAL(WIS)                      :: aioOn, ioLead ! bookeeping
    INTEGER(WIS)                      :: aio_slot, n_patches_added, n_active_patches, patch_data_size, remaining_patches, &
                                         n_patches_in_write, fpart, nranks_x, nranks_y, nranks_z, my_coords(3)  
    INTEGER(WADDRESS_KIND)            :: manifest_offset, data_offset, buf_offset
    LOGICAL(WIS), ALLOCATABLE         :: patch_added1d(:), patch_added2d(:,:), patch_added3d(:,:,:)
    REAL(WRD), ALLOCATABLE            :: fieldbuf(:,:)

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: time_to_write
    PROCEDURE, PUBLIC :: write_dumpset
    PROCEDURE, PUBLIC :: wait_aio_slots
    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: allocate_field_buffer_patch_flags
    PROCEDURE, PUBLIC :: deallocate_field_buffer_patch_flags

    PROCEDURE :: find_patch_data_size
    PROCEDURE :: gen_manifest_header
    PROCEDURE :: add_pline_to_manifest
    PROCEDURE :: add_line
    PROCEDURE :: append_patch_data
    PROCEDURE :: find_aio_slot
    PROCEDURE :: get_manifest_pline_size
    PROCEDURE :: find_filenames
    PROCEDURE :: write_file
    PROCEDURE :: add_patches_1d
    PROCEDURE :: add_patches_2d
    PROCEDURE :: add_patches_3d

END TYPE IO

INTERFACE IO

    MODULE PROCEDURE constructor

END INTERFACE IO

CONTAINS ! object methods

FUNCTION constructor(param, iolead, naio, rankloc, simunits) RESULT(self)

    TYPE(IO) :: self

    CLASS(Parameters) :: param
    CLASS(RankLocation) :: rankloc
    CLASS(SimulationUnits) :: simunits

    LOGICAL(WIS), INTENT(IN) :: iolead
    INTEGER(WIS), INTENT(IN) :: naio
 
    ! store information about the world set of MPI ranks 
    self%ioLead    = ioLead
    self%nranks_x  = rankloc%nranks(1)
    self%nranks_y  = rankloc%nranks(2)
    self%nranks_z  = rankloc%nranks(3)
    self%my_coords = rankloc%my_coords

    CALL param%section("Input / Output options")

    self%simulationTitle        = param%getStr("simulationTitle", "MyTitle", "")
    self%simulationDescription  = param%getStr("simulationDescription", "MyDescription", "")

    CALL param%comment("")
    CALL param%comment("Precision Options:")
    CALL param%comment("byte1	- convert value into a 8 bit integer with the selected mapping")
    CALL param%comment("byte2	- convert value into a 16 bit integer with the selected mapping")
    CALL param%comment("real4	- convert value into a 32 bit floating point value")
    CALL param%comment("real8	- convert (if necessary) value into a 64 bit floating point value")
    CALL param%comment("")
    CALL param%comment("Mapping Options (value will be clipped to the range, log uses only use value absolute magnitudes):")
    CALL param%comment("")
    CALL param%comment("linear	- map like B * (value-varmin)/(varmax-varmin), where B = 255 for byte1 and 65535 for byte2")
    CALL param%comment("log_e	- map like B * ln(value/varmin)/ln(varmax/varmin), where B = 255 for byte1 and 65535 for byte2")
    CALL param%comment("natural	- (real4 and real8 only) do not map the value to any range, field ignored.")
    CALL param%comment("")
    CALL param%comment("Fields:")
    CALL param%comment("")
    CALL param%comment("[RHO,PX,PY,PZ,BX,BY,BZ,E] state variables")
    CALL param%comment("[PRS ]  pressure")
    CALL param%comment("[VMAG] magnitude of velocity")
    CALL param%comment("[BMAG] magnitude of magnetic field")
    CALL param%comment("[EKIN] kinetic energy density")
    CALL param%comment("[EMAG] magnetic energy density")
    CALL param%comment("[ETHE] thermal energy density")
    CALL param%comment("[PDM ] dark matter particles")
    CALL param%comment("[JCUZ ] current density in Z")
    CALL param%comment("[TEMP] gas temperature")
    CALL param%comment("[MACH] Mach numbers from shock detector")
    CALL param%comment("[RHOG] gravitating mass density (gas and DM)")
    CALL param%comment("[PA01:PA10] raw passive field")
    CALL param%comment("[VS,VY,VZ] velocity components vx, vy, vz")
    CALL param%comment("[BFX,BFY,BFZ] xyz face centered magnetic field")
    CALL param%comment("[GX,GY,GZ,PHI] gravitational force and potential")

    self%aioOn = .FALSE. ! set the flag if there are AIO servers

    IF (naio .GT. 0) self%aioOn = .TRUE.

    RETURN

END FUNCTION constructor

! Wait for all AIO slots and deallocate all dump types

SUBROUTINE destroy(self, dump1, dump2, dump3)


    CLASS(IO) :: self
    TYPE(IO_DUMP), INTENT(INOUT) :: dump1, dump2, dump3

    !$OMP MASTER
     
    CALL self%wait_aio_slots(dump1, 0, .TRUE.) ! finish up any outstanding IO
    CALL self%wait_aio_slots(dump2, 0, .TRUE.) 
    CALL self%wait_aio_slots(dump3, 0, .TRUE.) 

    CALL dump1%destroy()
    CALL dump2%destroy()
    CALL dump3%destroy()

    !$OMP END MASTER

    !$OMP BARRIER

END SUBROUTINE destroy

! check if it is time to write a dump set for a given flavor given the current time 
! pass back up updated dt if we're really close

SUBROUTINE time_to_write(self, dump, t, dt, flag, dt_mod)

    CLASS(IO) :: self

    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    REAL(WRD), INTENT(IN) :: t, dt
    LOGICAL(WIS), INTENT(OUT) :: flag
    REAL(WRD), INTENT(INOUT) :: dt_mod

    !$OMP SINGLE

    dt_mod = dt

    IF (ABS(t - dump%next_dump_time) .LE. 1.0E-14_WRD .AND. dump%On) THEN

        flag = .TRUE.

        dump%next_dump_time = dump%next_dump_time + dump%Interval

    ELSE

        flag = .FALSE.

    END IF

    IF (dump%On) THEN

        IF (dt+t .GT. dump%next_dump_time) THEN
            dt_mod = dump%next_dump_time - t
        END IF

    END IF

    !$OMP END SINGLE

END SUBROUTINE time_to_write

! Write out the files for the dump set 'dump' for domain 'workd'.
! 1) Find the relevant sizes of data and buffers
! 2) Start loop over patches until dump%buf is full
! 3) Wait on AIO buffer, if previous file hasn't been written
! 4) For every patch generate data to be written, fill it into self%fieldbuf.
!    Every patch gets a line in the manifest (pline).
! 5) convert & scale values in self%fieldbuf and write the result into dump%buf
! 6) Write / send off to AIO dump%buf and repeat
! The routine loops over patches to fill the 'self%fieldbuf' buffer with the (derived) values 
! from the patch grid using OpenMP. 
! The number of loops needed to pack and send data (patches->dump%buf) is the ratio of 
! how much we need to write and the size of our buffer. This becoming 1 will provide the 
! best performance since we can get nearly total IO overlap.

SUBROUTINE write_dumpset(self, workd, dump, mygamma)

    CLASS(IO) :: self

    TYPE(Domain), INTENT(IN) :: workd
    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WID) :: buf_size, slot_size, patch_transfer_size, manifest_head_size, patch_data_size, &
                    next_buf_offset

    INTEGER(WID) :: manifest_pline_size = 320 ! nbytes buffer space for the patch description in the manifest
    
    REAL(WRD), INTENT(IN) :: mygamma

    !$OMP MASTER

    dump%Number = dump%Number + 1

    RMSG("Starting "//TRIM(dump%Name)//" dump write '"//TRIM(dump%FileRoot)//"' #"//I2A(dump%Number))

    IF (workd%is1d) THEN ! reset the patch writing flags
        self%patch_added1d(:) = .FALSE.
    ELSE IF (workd%is2d) THEN
        self%patch_added2d(:,:) = .FALSE.
    ELSE IF (workd%is3d) THEN
        self%patch_added3d(:,:,:) = .FALSE.
    END IF

    !$OMP END MASTER

    !$OMP SINGLE 
    
    patch_data_size = self%find_patch_data_size(workd, dump)

    !$OMP END SINGLE COPYPRIVATE(patch_data_size)

    patch_transfer_size = patch_data_size + manifest_pline_size 
        
    slot_size  = MIN(INT(WMPI_MAX_MESSAGE,WID) * 1024**2, dump%buf_size)

    buf_size = dump%buf_size
     
    IF (self%ioLead) THEN ! set the manifest header length if we are the IO lead
        manifest_head_size = STORAGE_SIZE(self%simulationDescription)/8 + WMAX_IO_FIELDS * 64 + 1024
    ELSE
        manifest_head_size = 0
    END IF

    !$OMP MASTER 

    !CALL self%wait_aio_slots(self%compressed, 0, .TRUE.) ! AIO doesnt support multiple inflight messages yet
    !CALL self%wait_aio_slots(self%movie, 0, .TRUE.)
    !CALL self%wait_aio_slots(self%restart, 0, .TRUE.)

    CALL self%find_aio_slot(dump)

    IF ((self%buf_offset + manifest_head_size + manifest_pline_size + patch_data_size) .GE. buf_size) THEN

        CALL self%wait_aio_slots(dump, 0, .TRUE.) ! out of room in the buffer
                                                  ! wait on all operations to complete to free the buffer
        self%buf_offset = 1

    END IF
    
    self%data_offset        = self%buf_offset ! initialize the offsets and remaining patch counter
    self%remaining_patches  = self%n_active_patches
    self%fpart              = 0

    !$OMP END MASTER

    !$OMP BARRIER

    ! add a check that the associated buffer is big enough for at least one Patch otherwise the loop below will be 
    ! infinite.


    DO WHILE (self%remaining_patches .GT. 0) ! do the loop, but remember this call is thread safe in a parallel region. 

        ! we'll pack up all the buffer we have noting that the actual messages may need to be broken up limited by the 
        ! maximum MPI message size 

        DO WHILE (self%buf_offset .LT. buf_size)

            IF (self%fpart .GT. 0) THEN 

                !$OMP MASTER 

                CALL self%wait_aio_slots(dump, self%aio_slot, .FALSE.) ! wait for completion of the prev. write
                    
                dump%aio_ops%inflight(self%aio_slot) = .TRUE.

                !$OMP END MASTER

                !$OMP BARRIER

            END IF

            !$OMP MASTER

            self%n_patches_in_write = MIN(self%remaining_patches, FLOOR(REAL((slot_size - manifest_head_size),4) &
                                                               / REAL((manifest_pline_size + patch_data_size),4)))

            self%manifest_offset  = self%buf_offset + self%n_patches_in_write * patch_data_size ! space for plines

            CALL self%find_filenames(dump)

            self%n_patches_added = 0

            !$OMP END MASTER

            !$OMP BARRIER

            IF (self%fpart .GT. 0) THEN

                manifest_head_size = 0

            ELSE IF (self%fpart .EQ. 0 .AND. self%ioLead) THEN
                
                manifest_head_size = self%manifest_offset

                !$OMP SINGLE
                
                CALL self%gen_manifest_header(workd, dump)
                
                !$OMP END SINGLE
                
                manifest_head_size = self%manifest_offset - manifest_head_size

            END IF

            manifest_pline_size = self%get_manifest_pline_size(dump)

            SELECT CASE (workd%ndim)

                CASE (1)

                    CALL self%add_patches_1d(workd, dump, manifest_pline_size, patch_data_size, mygamma)                    

                CASE (2)

                    CALL self%add_patches_2d(workd, dump, manifest_pline_size, patch_data_size, mygamma)                    
                   
                CASE (3)
    
                    CALL self%add_patches_3d(workd, dump, manifest_pline_size, patch_data_size, mygamma)                    

            END SELECT

            !$OMP MASTER

            CALL self%write_file(dump, patch_data_size, manifest_head_size, manifest_pline_size)
            
            self%buf_offset        = self%data_offset
            self%remaining_patches = self%remaining_patches - self%n_patches_added
            self%fpart             = self%fpart + 1

            next_buf_offset = self%buf_offset + manifest_head_size + (self%n_patches_in_write + 1)*patch_transfer_size

            iF ((next_buf_offset .GT. buf_size) .OR. (self%remaining_patches .LE. 0)) THEN
                
                self%buf_offset = buf_size !  exit inner loop if we ran out of buffer space or wrote all patches

            END IF

            !$OMP END MASTER
            !$OMP BARRIER

        END DO ! WHILE (buf_offset < buf_size)

        !$OMP BARRIER

        !$OMP SINGLE
        
        self%buf_offset     = 1 ! reset the buf_offset in case there are more loops over our buffer needed
        self%data_offset = self%buf_offset

        !$OMP END SINGLE

    END DO ! WHILE (self%remaining_patches .NE. 0)

END SUBROUTINE write_dumpset

! determine the amount of memory needed to write the data of a single Patch, not considering the line in the manifest.
! This is not trivial, because, depending on dump type, we write some fields at some data type.

INTEGER FUNCTION find_patch_data_size(self, workd, dump) RESULT(data_size)

    CLASS(IO) :: self
    TYPE(Domain) :: workd
    TYPE(IO_DUMP), INTENT(IN) :: dump

    INTEGER(WIS) :: bpz, nx, ny, nz, nb, pi, pj, pk, ii
    INTEGER(WID) :: nzones

    self%n_active_patches = 0

    nx = 0
    ny = 0
    nz = 0
    nb = 0

    IF (workd%is1d) THEN

        DO pi = 1, workd%npx

            IF (workd%domain1d(pi)%hasZones) THEN
                
                self%n_active_patches = self%n_active_patches + 1

            END IF 

            IF (workd%domain1d(pi)%hasZones .AND. nx .EQ. 0) THEN

                nx = workd%domain1d(pi)%nx
                ny = workd%domain1d(pi)%ny
                nz = workd%domain1d(pi)%nz
                nb = workd%domain1d(pi)%nb

            END IF

        END DO

    ELSE IF (workd%is2d) THEN

        DO pj = 1, workd%npy

            DO pi = 1, workd%npx

                IF (workd%domain2d(pi,pj)%hasZones) THEN
                    
                    self%n_active_patches = self%n_active_patches + 1

                END IF

                IF (workd%domain2d(pi,pj)%hasZones .AND. nx .EQ. 0) THEN

                    nx = workd%domain2d(pi,pj)%nx
                    ny = workd%domain2d(pi,pj)%ny
                    nz = workd%domain2d(pi,pj)%nz
                    nb = workd%domain2d(pi,pj)%nb

                END IF

            END DO

        END DO

    ELSE IF (workd%is3d) THEN

        DO pk = 1, workd%npz
            
            DO pj = 1, workd%npy

                DO pi = 1, workd%npx

                    IF (workd%domain3d(pi,pj,pk)%hasZones) THEN
                        
                        self%n_active_patches = self%n_active_patches + 1  

                    END IF

                    IF (workd%domain3d(pi,pj,pk)%hasZones .AND. nx .EQ. 0) THEN
                        
                        nx = workd%domain3d(pi,pj,pk)%nx
                        ny = workd%domain3d(pi,pj,pk)%ny
                        nz = workd%domain3d(pi,pj,pk)%nz
                        nb = workd%domain3d(pi,pj,pk)%nb
                       
                    END IF

                END DO

            END DO

        END DO

    END IF

    bpz = 0 ! find bytes per zone depending on output field precisions

    DO ii = 1, dump%Nfields

        SELECT CASE (dump%fields%Prec(ii))
            
            CASE('byte1')
                bpz = bpz + 1
            CASE('byte2')
                bpz = bpz + 2
            CASE('real4')
                bpz = bpz + 4
            CASE('real8')
                bpz = bpz + 8
                
        END SELECT

    END DO

    IF (workd%is1d) nzones = (2*nb + nx)
    IF (workd%is2d) nzones = (2*nb + nx) * (2*nb + ny)
    IF (workd%is3d) nzones = (2*nb + nx) * (2*nb + ny) * (2*nb + nz)
    
    data_size = bpz * nzones

END FUNCTION find_patch_data_size

! generate the header portion of the manifest (excluding the Patch table) 

SUBROUTINE gen_manifest_header(self, workd, dump)

    CLASS(IO) :: self
    
    TYPE(Domain) :: workd
    TYPE(IO_DUMP) :: dump
    
    INTEGER :: date_time(8), slen, off
    INTEGER(WIS) :: i, nx, ny, nz, nb, pi, pj, pk, n, fn, formn
    
    REAL(WRD) :: llim, ulim

    CHARACTER(10) :: dt(3)
    CHARACTER(36) :: dateString, pdt, pt, pnsteps, vard, cubestr
    CHARACTER(48) :: limss
    CHARACTER(WCHAR_BUF_SIZE) :: strbuf
    CHARACTER(1) :: header(2 * STORAGE_SIZE(self%simulationDescription))

    ! title and description first 
    off                  = 1
    slen                 = 38
    
    CALL self%add_line('# SIMULATION DESCRIPTION AND STATUS #' // CHAR(10), header(off:off+slen-1), slen)
    
    off                  = off + slen
    slen                 = LEN_TRIM(self%simulationTitle) + 15

    CALL self%add_line('title         ' // TRIM(self%simulationTitle) // CHAR(10), header(off:off+slen-1), slen)
    
    off                  = off + slen
    slen                 = LEN_TRIM(self%simulationDescription) + 15
    
    CALL self%add_line('comment       ' // TRIM(self%simulationDescription) // CHAR(10), header(off:off+slen-1), slen)
    
    off                  = off + slen

    CALL DATE_AND_TIME(dt(1), dt(2), dt(3), date_time)

    WRITE(dateString, '("date          ", I4, "-", I2.2, "-", I2.2, " ", I2.2, ":", I2.2, ":", I2.2)')  &
                      date_time(1), date_time(2), date_time(3), date_time(5), date_time(6), date_time(7)

    ! add data, user and host
    slen                 = LEN_TRIM(dateString) + 1
    CALL self%add_line(TRIM(dateString) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    slen                 = LEN_TRIM(USERNAME) + 15
    CALL self%add_line('user          ' // TRIM(USERNAME) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    slen                 = LEN_TRIM(HOSTNAME) + 15
    CALL self%add_line('host          ' // TRIM(HOSTNAME) // CHAR(10), header(off:off+slen-1), slen)

    off                  = off + slen

    WRITE(pdt, '(ES22.15)') workd%dt
    slen                 = LEN_TRIM(ADJUSTL(pdt)) + 15
    CALL self%add_line('dtime         ' // TRIM(ADJUSTL(pdt)) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    WRITE(pt, '(ES22.15)') workd%t
    slen                 = LEN_TRIM(ADJUSTL(pt)) + 15
    CALL self%add_line('time          ' // TRIM(ADJUSTL(pt)) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    WRITE(pnsteps, '(I22)') workd%nsteps
    slen                 = LEN_TRIM(ADJUSTL(pnsteps)) + 15
    CALL self%add_line('step          ' // TRIM(ADJUSTL(pnsteps)) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    IF (workd%is1d) THEN

        DO pi = 1, workd%npx
            
            IF (workd%domain1d(pi)%hasZones) THEN
                nx = workd%domain1d(pi)%nx
                ny = 1
                nz = 1
                nb = workd%domain1d(pi)%nb
                EXIT
            END IF

        END DO

    ELSE IF(workd%is2d) THEN

        DO pj = 1, workd%npy
            DO pi = 1, workd%npx
            
                IF (workd%domain2d(pi,pj)%hasZones) THEN
                    nx = workd%domain2d(pi,pj)%nx
                    ny = workd%domain2d(pi,pj)%ny
                    nz = 1
                    nb = workd%domain2d(pi,pj)%nb
                    EXIT
                END IF

            END DO
        END DO
        
    ELSE IF(workd%is3d) THEN

        DO pk = 1, workd%npz
            DO pj = 1, workd%npy
                DO pi = 1, workd%npx
                
                    IF (workd%domain3d(pi,pj,pk)%hasZones) THEN
                        nx = workd%domain3d(pi,pj,pk)%nx
                        ny = workd%domain3d(pi,pj,pk)%ny
                        nz = workd%domain3d(pi,pj,pk)%nz
                        nb = workd%domain3d(pi,pj,pk)%nb
                        EXIT
                    END IF

                END DO
            END DO
        END DO

    END IF

    WRITE(cubestr, '(3(I10))') nx, ny, nz
    slen                 = LEN_TRIM(cubestr) + 15
    CALL self%add_line('ncubesxyz     ' // TRIM(cubestr) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    WRITE(cubestr, '(3(I10))') self%nranks_x*workd%npx, self%nranks_y*workd%npy, self%nranks_z*workd%npz
    slen                 = LEN_TRIM(cubestr) + 15
    CALL self%add_line('nbrickxyz     ' // TRIM(cubestr) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    
    IF (workd%is1d) THEN
        WRITE(cubestr, '(3(I10))') nb, 0, 0
    ELSE IF (workd%is2d) THEN
        WRITE(cubestr, '(3(I10))') nb, nb, 0
    ELSE IF (workd%is3d) THEN
        WRITE(cubestr, '(3(I10))') nb, nb, nb
    END IF
    
    slen                 = LEN_TRIM(cubestr) + 15
    CALL self%add_line('nbdry_xyz     ' // TRIM(cubestr) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    WRITE(limss, '(2(ES24.15))') -0.5_WRD*workd%dx*self%nranks_x*workd%npx*nx, 0.5_WRD*workd%dx*self%nranks_x*workd%npx*nx

    slen                 = LEN_TRIM(ADJUSTL(limss)) + 15
    CALL self%add_line('xlimits       ' // TRIM(ADJUSTL(limss)) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    WRITE(limss, '(2(ES24.15))') -0.5_WRD*workd%dx*self%nranks_y*workd%npy*ny, 0.5_WRD*workd%dx*self%nranks_y*workd%npy*ny
    
    slen                 = LEN_TRIM(ADJUSTL(limss)) + 15
    CALL self%add_line('ylimits       ' // TRIM(ADJUSTL(limss)) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    WRITE(limss, '(2(ES24.15))') -0.5_WRD*workd%dx*self%nranks_z*workd%npz*nz, 0.5_WRD*workd%dx*self%nranks_z*workd%npz*nz
    
    slen                 = LEN_TRIM(ADJUSTL(limss)) + 15
    CALL self%add_line('zlimits       ' // TRIM(ADJUSTL(limss)) // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    slen                 = 55
    CALL self%add_line('lobndry   continuation    continuation    continuation' // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    CALL self%add_line('hibndry   continuation    continuation    continuation' // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen

    slen                 = 37
    
    CALL self%add_line(CHAR(10) // '# VARIABLE FORMATTING INFORMATION #' // CHAR(10), header(off:off+slen-1), slen)

    off          = off + slen

    DO i = 1,dump%Nfields ! put fields

        strbuf = ' '
        
        WRITE(strbuf, '(A7,"   ",A5," ",A5," ",A8," ",1E14.7," ",1E14.7)') "field3d", &
                      dump%fields%Symbol(i), dump%fields%Prec(i), dump%fields%Map(i), dump%fields%Min(i), dump%fields%Max(i)

        slen = LEN_TRIM(strbuf) + 1

        CALL self%add_line(TRIM(strbuf) // CHAR(10), header(off:off+slen-1), slen)
        
        off = off + slen
    
    END DO

    slen                 = 14
    CALL self%add_line(CHAR(10) // '# MANIFEST #' // CHAR(10), header(off:off+slen-1), slen)
    off                  = off + slen
    slen                 = off - 1

    dump%buf(self%manifest_offset:self%manifest_offset+slen) = header(1:1+slen) ! write manifest into dump buffer

    self%manifest_offset = self%manifest_offset + off - 1

END SUBROUTINE gen_manifest_header

! Add a Patch at 'pi', 'pj', 'pk' to the manifest record in dump%buf
! A pline has this format : 
! boxb        1        6     3     2      KH2d0-0068-part00-000
! meaning:
! ID         level    px    py    pz      filename

SUBROUTINE add_pline_to_manifest(self, workd, dump, pi, pj, pk, offset)

    CLASS(IO) :: self
    
    TYPE(Domain), INTENT(IN) :: workd
    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WIS), INTENT(IN) :: pi, pj, pk, offset
 
    INTEGER(WIS) :: slen
    CHARACTER(320) :: record

    WRITE(record, '("boxb        1   ", 3(I6), "      ", A)')  &
          self%my_coords(1)*workd%npx+pi-1, self%my_coords(2)*workd%npy+pj-1, &
          self%my_coords(3)*workd%npz+pk-1, TRIM(ADJUSTL(dump%dataFile)) // CHAR(10)

    slen = LEN_TRIM(record)

    CALL self%add_line(TRIM(record), dump%buf(offset:offset+slen-1), slen)

END SUBROUTINE add_pline_to_manifest

! Get the storage size of a pline in the manifest file

INTEGER FUNCTION get_manifest_pline_size(self, dump)

    CLASS(IO) :: self
    TYPE(IO_DUMP), INTENT(IN) :: dump

    get_manifest_pline_size = LEN_TRIM(ADJUSTL(dump%dataFile)) + 16 + 18 + 6 + 1

    RETURN

END FUNCTION get_manifest_pline_size

SUBROUTINE add_line(self, string, buffer, slen)

    CLASS(IO) :: self
    INTEGER(WIS), INTENT(IN) :: slen
    CHARACTER(slen) :: string
    CHARACTER(1), INTENT(INOUT) :: buffer(slen)
    INTEGER(WIS) :: i
    
    DO i = 1, slen

        buffer(i) = string(i:i)

    END DO

END SUBROUTINE add_line

! allocate a field buffer for each thread to use 

SUBROUTINE allocate_field_buffer_patch_flags(self, workd)

    CLASS(IO) :: self
    TYPE(Domain) :: workd
    INTEGER(WIS) :: lb1, lb2, lb3, ub1, ub2, ub3, fblen, pi, pj, pk, ppi, ppj, ppk

    !$OMP SINGLE

    IF (workd%is1d) THEN ! allocate space to the field buffer

        DO ppi = 1, workd%npx

            IF (workd%domain1d(ppi)%hasZones) THEN
                pi = ppi
                EXIT
            END IF

        END DO
        
        lb1   = LBOUND(workd%domain1d(pi)%grid1d,1,WIS)
        ub1   = UBOUND(workd%domain1d(pi)%grid1d,1,WIS)
        fblen = ub1 - lb1 + 1
        ALLOCATE(self%fieldbuf(fblen, WNTHREADS))
        ALLOCATE(self%patch_added1d(workd%npx))

    ELSE IF (workd%is2d) THEN

        DO ppj = 1, workd%npy
            DO ppi = 1, workd%npx

                IF (workd%domain2d(ppi,ppj)%hasZones) THEN
                    pi = ppi
                    pj = ppj
                    EXIT
                END IF

            END DO
        END DO

        lb1   = LBOUND(workd%domain2d(pi,pj)%grid2d,1,WIS)
        lb2   = LBOUND(workd%domain2d(pi,pj)%grid2d,2,WIS)
        
        ub1   = UBOUND(workd%domain2d(pi,pj)%grid2d,1,WIS)
        ub2   = UBOUND(workd%domain2d(pi,pj)%grid2d,2,WIS)
        
        fblen = (ub2 - lb2 + 1) * (ub1 - lb1 + 1)
        
        ALLOCATE(self%fieldbuf(fblen, WNTHREADS))
        
        ALLOCATE(self%patch_added2d(workd%npx, workd%npy))

    ELSE IF (workd%is3d) THEN
        
        DO ppk = 1, workd%npz
            DO ppj = 1, workd%npy
                DO ppi = 1, workd%npx

                    IF (workd%domain3d(ppi,ppj,ppk)%hasZones) THEN
                        pi = ppi
                        pj = ppj
                        pk = ppk
                        EXIT
                    END IF

                END DO
            END DO
        END DO

        lb1   = LBOUND(workd%domain3d(pi,pj,pk)%grid3d,1,WIS)
        lb2   = LBOUND(workd%domain3d(pi,pj,pk)%grid3d,2,WIS)
        lb3   = LBOUND(workd%domain3d(pi,pj,pk)%grid3d,3,WIS)
        
        ub1   = UBOUND(workd%domain3d(pi,pj,pk)%grid3d,1,WIS)
        ub2   = UBOUND(workd%domain3d(pi,pj,pk)%grid3d,2,WIS)
        ub3   = UBOUND(workd%domain3d(pi,pj,pk)%grid3d,3,WIS)
        
        fblen = (ub3 - lb3 + 1) * (ub2 - lb2 + 1) * (ub1 - lb1 + 1)
        
        ALLOCATE(self%fieldbuf(fblen, WNTHREADS))
        ALLOCATE(self%patch_added3d(workd%npx, workd%npy, workd%npz))

    END IF

    !$OMP END SINGLE

END SUBROUTINE allocate_field_buffer_patch_flags

SUBROUTINE deallocate_field_buffer_patch_flags(self, workd)

    CLASS(IO) :: self
    TYPE(Domain) :: workd

    !$OMP SINGLE

    IF (ALLOCATED(self%fieldbuf)) DEALLOCATE(self%fieldbuf)

    IF (workd%is1d .AND. ALLOCATED(self%patch_added1d)) THEN
        
        DEALLOCATE(self%patch_added1d)

    ELSE IF (workd%is2d .AND. ALLOCATED(self%patch_added2d)) THEN

        DEALLOCATE(self%patch_added2d)

    ELSE IF (workd%is3d .AND. ALLOCATED(self%patch_added3d)) THEN
        
        DEALLOCATE(self%patch_added3d)

    END IF

    !$OMP END SINGLE

END SUBROUTINE deallocate_field_buffer_patch_flags

! For a single patch at work domain "workd" location "pi,pj,pk": 
! Calculate all output fields and place the result in the field buffer "self%fieldbuf". 
! Then convert the field buffer to its precision and write them in order into "dump%buf"
! starting at "offset".
! This is always one massive SELECT statement.

SUBROUTINE append_patch_data(self, workd, dump, pi, pj, pk, offset, mygamma)

    CLASS(IO) :: self
    
    TYPE(Domain), INTENT(IN) :: workd
    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WID), INTENT(IN) :: offset
    INTEGER(WIS), INTENT(IN) :: pi, pj, pk

    INTEGER(WIS) :: myoffset, f, nfields, lb1, lb2, lb3, ub1, ub2, ub3, fblen, i, j, k, ioff, di, &
                    mykind, mylen, r, c, idx

    INTEGER(WIS) :: tid 

    REAL(WRD), INTENT(IN) :: mygamma

    REAL(WRS) :: minv, maxv, scaling
    REAL(WRD) :: wrdc, gammam1, invdx
    REAL(8) :: dp_val
    REAL(4) :: sp_val
    INTEGER(2) :: tb_val
    INTEGER(1) :: sb_val
    CHARACTER(5) :: precs, sym
    CHARACTER(8) :: map

    gammam1 = mygamma - 1.0_WRD ! for pressure output we need (mygamma - 1)
    invdx = 1.0_WRD / workd%dx
    tid = WTHREAD + 1 

    myoffset = offset
    
    nfields   = dump%Nfields

    IF (workd%is1d) THEN ! find patch indices & size of fieldbuffer
        
        lb1   = LBOUND(workd%domain1d(pi)%grid1d,1,WIS)
        ub1   = UBOUND(workd%domain1d(pi)%grid1d,1,WIS)
        fblen = ub1 - lb1 + 1

    ELSE IF (workd%is2d) THEN

        lb1   = LBOUND(workd%domain2d(pi,pj)%grid2d,1,WIS)
        lb2   = LBOUND(workd%domain2d(pi,pj)%grid2d,2,WIS)
        ub1   = UBOUND(workd%domain2d(pi,pj)%grid2d,1,WIS)
        ub2   = UBOUND(workd%domain2d(pi,pj)%grid2d,2,WIS)
        fblen = (ub2 - lb2 + 1) * (ub1 - lb1 + 1)

    ELSE IF (workd%is3d) THEN
        
        lb1   = LBOUND(workd%domain3d(pi,pj,pk)%grid3d,1,WIS)
        lb2   = LBOUND(workd%domain3d(pi,pj,pk)%grid3d,2,WIS)
        lb3   = LBOUND(workd%domain3d(pi,pj,pk)%grid3d,3,WIS)
        ub1   = UBOUND(workd%domain3d(pi,pj,pk)%grid3d,1,WIS)
        ub2   = UBOUND(workd%domain3d(pi,pj,pk)%grid3d,2,WIS)
        ub3   = UBOUND(workd%domain3d(pi,pj,pk)%grid3d,3,WIS)
        fblen = (ub3 - lb3 + 1) * (ub2 - lb2 + 1) * (ub1 - lb1 + 1)

    END IF

    di = ub1 - lb1 ! # cells along x direction, length of implicit loop

    DO f = 1, nfields
        
        sym   = dump%fields%Symbol(f)
        precs = dump%fields%Prec(f)
        map   = dump%fields%Map(f)
        minv  = dump%fields%Min(f)
        maxv  = dump%fields%Max(f)

        SELECT CASE (sym) ! copy the field values of the patch into our intermediate buffer 'fieldbuf'

            CASE ("RHO","PX","PY","PZ","BX","BY","BZ","E") ! raw state vector fields 

                IF (sym .EQ. "RHO") c = 1
                IF (sym .EQ. "PX")  c = 2
                IF (sym .EQ. "PY")  c = 3
                IF (sym .EQ. "PZ")  c = 4
                IF (sym .EQ. "BX")  c = 5
                IF (sym .EQ. "BY")  c = 6
                IF (sym .EQ. "BZ")  c = 7
                IF (sym .EQ. "E")   c = 8

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%grid1d(:,c)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%grid2d(:,j,c)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%grid3d(:,j,k,c)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE ("PRS") ! fluid pressure

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = gammam1 * (workd%domain1d(pi)%grid1d(:,8) &
                                                      - (0.5_WRD / workd%domain1d(pi)%grid1d(:,1)) * &
                                                      (workd%domain1d(pi)%grid1d(:,2)**2 + &
                                                       workd%domain1d(pi)%grid1d(:,3)**2 + &
                                                       workd%domain1d(pi)%grid1d(:,4)**2) &
                                                      - 0.5_WRD * (workd%domain1d(pi)%grid1d(:,5)**2 &
                                                                  + workd%domain1d(pi)%grid1d(:,6)**2 &
                                                                  + workd%domain1d(pi)%grid1d(:,7)**2))
                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = gammam1 * (workd%domain2d(pi,pj)%grid2d(:,j,8) &
                                                              - (0.5_WRD / workd%domain2d(pi,pj)%grid2d(:,j,1)) * &
                                                                          (workd%domain2d(pi,pj)%grid2d(:,j,2)**2 &
                                                                         + workd%domain2d(pi,pj)%grid2d(:,j,3)**2 &
                                                                         + workd%domain2d(pi,pj)%grid2d(:,j,4)**2) &
                                                               - 0.5_WRD * (workd%domain2d(pi,pj)%grid2d(:,j,5)**2 &
                                                                          + workd%domain2d(pi,pj)%grid2d(:,j,6)**2 &
                                                                          + workd%domain2d(pi,pj)%grid2d(:,j,7)**2))
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = gammam1 * (workd%domain3d(pi,pj,pk)%grid3d(:,j,k,8) &
                                                                    - (0.5_WRD / workd%domain3d(pi,pj,pk)%grid3d(:,j,k,1)) * &
                                                                                (workd%domain3d(pi,pj,pk)%grid3d(:,j,k,2)**2 &
                                                                               + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,3)**2 &
                                                                               + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,4)**2) &
                                                                    - 0.5_WRD * (workd%domain3d(pi,pj,pk)%grid3d(:,j,k,5)**2 &
                                                                               + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,6)**2 &
                                                                               + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,7)**2))
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE("PA01", "PA02","PA03","PA04","PA05","PA06","PA07","PA08","PA09","PA10") ! a raw passive field

                 idx = SCAN(sym, "A")
                 READ(sym(idx+1:), "(I2)") c

                 IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%pass1d(:,c)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%pass2d(:,j,c)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%pass3d(:,j,k,c)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE ("VMAG") ! magnitude of the velocity

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = SQRT(workd%domain1d(pi)%grid1d(:,2)**2 + workd%domain1d(pi)%grid1d(:,3)**2 &
                                              + workd%domain1d(pi)%grid1d(:,4)**2)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = SQRT(workd%domain2d(pi,pj)%grid2d(:,j,2)**2 &
                                                                    + workd%domain2d(pi,pj)%grid2d(:,j,3)**2 &
                                                                    + workd%domain2d(pi,pj)%grid2d(:,j,4)**2)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = SQRT(workd%domain3d(pi,pj,pk)%grid3d(:,j,k,2)**2 &
                                                                        + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,3)**2 &
                                                                        + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,4)**2)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE ("BMAG") ! magnitude of the magentic field

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = SQRT(workd%domain1d(pi)%grid1d(:,5)**2 + workd%domain1d(pi)%grid1d(:,6)**2 &
                                              + workd%domain1d(pi)%grid1d(:,7)**2)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = SQRT(workd%domain2d(pi,pj)%grid2d(:,j,5)**2 &
                                                                    + workd%domain2d(pi,pj)%grid2d(:,j,6)**2 &
                                                                    + workd%domain2d(pi,pj)%grid2d(:,j,7)**2)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = SQRT(workd%domain3d(pi,pj,pk)%grid3d(:,j,k,5)**2 &
                                                                 + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,6)**2 &
                                                                 + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,7)**2)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE ("EKIN") ! kinetic energy density

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = 0.5_WRD * (workd%domain1d(pi)%grid1d(:,2)**2 &
                                                    + workd%domain1d(pi)%grid1d(:,3)**2 &
                                                    + workd%domain1d(pi)%grid1d(:,4)**2)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = 0.5_WRD * (workd%domain2d(pi,pj)%grid2d(:,j,2)**2 &
                                                                   + workd%domain2d(pi,pj)%grid2d(:,j,3)**2 &
                                                                   + workd%domain2d(pi,pj)%grid2d(:,j,4)**2)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = 0.5_WRD * (workd%domain3d(pi,pj,pk)%grid3d(:,j,k,2)**2 &
                                                                       + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,3)**2 &
                                                                       + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,4)**2)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE ("EMAG") ! magnetic energy density B^2

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = 0.5_WRD * (workd%domain1d(pi)%grid1d(:,5)**2 + workd%domain1d(pi)%grid1d(:,6)**2 &
                                                    + workd%domain1d(pi)%grid1d(:,7)**2)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = 0.5_WRD * (workd%domain2d(pi,pj)%grid2d(:,j,5)**2 &
                                                                   + workd%domain2d(pi,pj)%grid2d(:,j,6)**2 &
                                                                   + workd%domain2d(pi,pj)%grid2d(:,j,7)**2)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = 0.5_WRD * (workd%domain3d(pi,pj,pk)%grid3d(:,j,k,5)**2 &
                                                                       + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,6)**2 &
                                                                       + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,7)**2)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE("ETHE") ! thermal energy density

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%grid1d(:,8) &
                                         - (0.5_WRD / workd%domain1d(pi)%grid1d(:,1)) * (workd%domain1d(pi)%grid1d(:,2)**2  &
                                                                                       + workd%domain1d(pi)%grid1d(:,3)**2  &
                                                                                       + workd%domain1d(pi)%grid1d(:,4)**2) &
                                         - 0.5_WRD * (workd%domain1d(pi)%grid1d(:,5)**2 + workd%domain1d(pi)%grid1d(:,6)**2 &
                                                    + workd%domain1d(pi)%grid1d(:,7)**2)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%grid2d(:,j,8) &
                                                        - (0.5_WRD / workd%domain2d(pi,pj)%grid2d(:,j,1))   &
                                                        * ( workd%domain2d(pi,pj)%grid2d(:,j,2)**2  &
                                                          + workd%domain2d(pi,pj)%grid2d(:,j,3)**2  &
                                                          + workd%domain2d(pi,pj)%grid2d(:,j,4)**2) &
                                                        - 0.5_WRD * (workd%domain2d(pi,pj)%grid2d(:,j,5)**2 &
                                                                   + workd%domain2d(pi,pj)%grid2d(:,j,6)**2 &
                                                                   + workd%domain2d(pi,pj)%grid2d(:,j,7)**2)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%grid3d(:,j,k,8) &
                                                            - (0.5_WRD / workd%domain3d(pi,pj,pk)%grid3d(:,j,k,1))  &
                                                            * ( workd%domain3d(pi,pj,pk)%grid3d(:,j,k,2)**2   &
                                                              + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,3)**2   &
                                                              + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,4)**2 ) &
                                                            - 0.5_WRD * ( workd%domain3d(pi,pj,pk)%grid3d(:,j,k,5)**2 &
                                                                        + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,6)**2 &
                                                                        + workd%domain3d(pi,pj,pk)%grid3d(:,j,k,7)**2)   
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            !  more to come when gravity and dark matter are in 
             
            ! Gravitational Potential
            CASE ("PHI")

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%grav1d(:,2)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+(ub1-lb1),tid) = workd%domain2d(pi,pj)%grav2d(:,j,2)
                        ioff = ioff + (ub1-lb1) + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+(ub1-lb1),tid) = workd%domain3d(pi,pj,pk)%grav3d(:,j,k,2)
                            ioff = ioff + (ub1-lb1) + 1

                        END DO

                    END DO

                END IF

            !### Vx ###
            CASE ("VX")

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%grid1d(:,2) / workd%domain1d(pi)%grid1d(:,1)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%grid2d(:,j,2) &
                                                               / workd%domain2d(pi,pj)%grid2d(:,j,1)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%grid3d(:,j,k,2) &
                                                                   / workd%domain3d(pi,pj,pk)%grid3d(:,j,k,1)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF
             
            CASE ("VY") ! VY

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%grid1d(:,3) / workd%domain1d(pi)%grid1d(:,1)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%grid2d(:,j,3) &
                                                               / workd%domain2d(pi,pj)%grid2d(:,j,1)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%grid3d(:,j,k,3) &
                                                                   / workd%domain3d(pi,pj,pk)%grid3d(:,j,k,1)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

             
            CASE ("VZ") ! Vz

                IF (workd%is1d) THEN

                    self%fieldbuf(:,tid) = workd%domain1d(pi)%grid1d(:,4) / workd%domain1d(pi)%grid1d(:,1)

                ELSE IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%grid2d(:,j,4) &
                                                               / workd%domain2d(pi,pj)%grid2d(:,j,1)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%grid3d(:,j,k,4) &
                                                                   / workd%domain3d(pi,pj,pk)%grid3d(:,j,k,1)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

             
            CASE("BFX") ! Bx face

                IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%bface2d(:,j,1)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%bface3d(:,j,k,1)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE("BYF") ! By face

                IF (workd%is2d) THEN

                    ioff = 1
                    DO j = lb2, ub2

                        self%fieldbuf(ioff:ioff+di,tid) = workd%domain2d(pi,pj)%bface2d(:,j,2)
                        ioff = ioff + di + 1

                    END DO

                ELSE IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%bface3d(:,j,k,2)
                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE("BFZ") ! Bz face

                IF (workd%is3d) THEN

                    ioff = 1
                    DO k = lb3, ub3

                        DO j = lb2, ub2

                            self%fieldbuf(ioff:ioff+di,tid) = workd%domain3d(pi,pj,pk)%bface3d(:,j,k,3)

                            ioff = ioff + di + 1

                        END DO

                    END DO

                END IF

            CASE("JCUZ") ! current density in Z

                IF (workd%is2d) THEN  ! 2D only ?

                    ioff = 1
                    DO j = lb2, ub2

                        DO i = lb1, ub1

                            IF (j .GT. lb2 .AND. i .GT. lb1) THEN

                                self%fieldbuf(ioff,tid) = invdx * ((workd%domain2d(pi,pj)%bface2d(i+1,j,2) &
                                                                - workd%domain2d(pi,pj)%bface2d(i,j,2)) &
                                                                - (workd%domain2d(pi,pj)%bface2d(i,j+1,1) &
                                                                - workd%domain2d(pi,pj)%bface2d(i,j,1)))

                            END IF
                            ioff = ioff + 1

                        END DO

                    END DO

                END IF

            CASE DEFAULT
                
                ASSERT(.FALSE., "Field "//sym//" not implemented, yet.")

        END SELECT

        IF ((precs .EQ. 'real8' .AND. STORAGE_SIZE(wrdc) .EQ. 64) .OR. & ! if the precision is real8 and WRD is 64 bits or 
            (precs .EQ. 'real4' .AND. STORAGE_SIZE(wrdc) .EQ. 32)) THEN  ! real4 and WRD is 32 bit do a straight large copy

            mykind = 4 
            IF (precs .EQ. 'real8') mykind = 8

            mylen = mykind*fblen
        
            CALL memcpy(dump%buf(myoffset), self%fieldbuf(1,tid), mylen)

            myoffset = myoffset + mylen

        ELSE IF ((precs .EQ. 'real8' .AND. STORAGE_SIZE(wrdc) .NE. 64) .OR. &   ! if the precision is real8/4 and we're not
                 (precs .EQ. 'real4' .AND. STORAGE_SIZE(wrdc) .NE. 32))   THEN  ! 64/32 bits do a cast and an el by el copy

            IF (precs .EQ. 'real8') THEN
                mykind = 8
            ELSE 
                mykind = 4
            END IF

            mylen = mykind
                
            IF (precs .EQ. 'real8') THEN

                DO i = 1, fblen

                    dp_val = REAL(self%fieldbuf(i,tid), 8)

                    CALL memcpy(dump%buf(myoffset), dp_val, mylen)

                    myoffset = myoffset + mylen

                END DO
            ELSE
                DO i = 1, fblen

                    sp_val = REAL(self%fieldbuf(i,tid), 4)

                    CALL memcpy(dump%buf(myoffset), sp_val, mylen)

                    myoffset = myoffset + mylen

                END DO
            END IF

        ELSE IF (precs .EQ. 'byte2') THEN ! to do a byte2 quantity we scale to 65536 values

            IF (map .EQ. 'linear') THEN ! create the map avoiding divide by zero
                scaling = 65535.0_WRS / MAX(1.0E-6_WRS,(maxv - minv))
            ELSE IF (map .EQ. 'log_e') THEN
                scaling = 65535.0_WRS / MAX(1.0E-6_WRS,LOG(MAX(1.0E-36_WRS,maxv)) - LOG(MAX(1.0E-36_WRS,minv)))
            END IF

            IF (map .EQ. 'linear') THEN ! we'll clip values to within the min and max values given
                DO i = 1, fblen

                    tb_val = INT(scaling * MIN(maxv, MAX(minv, REAL(self%fieldbuf(i,tid),WRS))) - minv, 2)

                    CALL memcpy(dump%buf(myoffset), tb_val, 2)

                    myoffset = myoffset + 2

                END DO
            ELSE IF (map .EQ. 'log_e') THEN
                DO i = 1, fblen

                    tb_val = INT(scaling*MIN(maxv, MAX(minv, LOG( MAX( 1.0E-36_WRS, REAL(self%fieldbuf(i,tid),WRS))) - minv)), 2)

                    CALL memcpy(dump%buf(myoffset), tb_val, 2)

                    myoffset = myoffset + 2

                END DO
            END IF

        ELSE IF (precs .EQ. 'byte1') THEN ! similar to 2 byte mapping, but now over a more narrow range of buckets

            IF (map .EQ. 'linear') THEN ! create the map avoiding divide by zero 
                scaling = 255.0_WRS / MAX(1.0E-4_WRS,(maxv - minv))
            ELSE IF (map .EQ. 'log_e') THEN
                scaling = 255.0_WRS / MAX(1.0E-4_WRS,LOG(MAX(1.0E-36_WRS,maxv)) - LOG(MAX(1.0E-36_WRS,minv)))
            END IF

            IF (map .EQ. 'linear') THEN
                DO i = 1, fblen
   
                    sb_val = INT(scaling * MIN(maxv, MAX(minv, REAL(self%fieldbuf(i,tid), WRS) )) - minv, 1)

                    CALL memcpy(dump%buf(myoffset), sb_val, 1)

                    myoffset = myoffset + 1

                END DO
            ELSE IF (map .EQ. 'log_e') THEN
                DO i = 1, fblen

                    sb_val = INT(scaling*MIN(maxv, MAX(minv, LOG( MAX(1.0E-36_WRS, REAL(self%fieldbuf(i,tid),WRS)))-minv)), 1)

                    CALL memcpy(dump%buf(myoffset), sb_val, 1)

                    myoffset = myoffset + 1

                END DO
            END IF
        END IF
    END DO

END SUBROUTINE append_patch_data

! Wait on  AIO slot 'slot' to complete.  If slot = 0 this means wait on all inflight slots

SUBROUTINE wait_aio_slots(self, dump, slot, writeMeta)

    CLASS(IO) :: self

    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WIS), INTENT(IN) ::  slot
    INTEGER(WIS) :: op
    LOGICAL(WIS), INTENT(IN) :: writeMeta

    IF (self%aioOn) CALL AIO_PROGRESS()

    IF (slot .GT. 0) THEN ! wait on a specific slot

        IF (dump%aio_ops%inflight(slot)) THEN

            IF (self%aioOn) THEN

                CALL AIO_WAIT(dump%aio_ops%slotIDs(slot))

                IF (writeMeta) THEN

                    CALL AIO_WRITE_BUF_META(TRIM(dump%Path) // '/' // dump%manifestFile(slot),0, 0, &
                                                 dump%buf(dump%aio_ops%offsets(slot,1)), dump%aio_ops%slotIDs(slot), 2)

                    CALL AIO_WAIT(dump%aio_ops%slotIDs(slot))

                END IF

            END IF

            dump%aio_ops%inflight(slot)  = .FALSE.
            dump%aio_ops%offsets(slot,:) = 0
            dump%aio_ops%lengths(slot,:) = 0

        END IF

    ELSE ! wait on all inflight slots

        DO op = 1, WNUM_AIO_OPS_INFLIGHT

            IF (dump%aio_ops%inflight(op)) THEN

                IF (self%aioOn) THEN

                    CALL AIO_WAIT(dump%aio_ops%slotIDs(op))

                    IF (writeMeta) THEN

                        CALL AIO_WRITE_BUF_META(TRIM(dump%Path) // '/' // dump%manifestFile(op), 0, 0, &
                                                     dump%buf(dump%aio_ops%offsets(op,1)), dump%aio_ops%slotIDs(op), 2)

                        CALL AIO_WAIT(dump%aio_ops%slotIDs(op))

                    END IF

                END IF
                    
                dump%aio_ops%inflight(op)  = .FALSE.
                dump%aio_ops%offsets(op,:) = 0
                dump%aio_ops%lengths(op,:) = 0

            END IF

        END DO

    END IF

END SUBROUTINE wait_aio_slots

! Select an AIO slot, NOT threadsafe, has to be called in single or master.
! AIO actually does not support having multiple writes in flight while also gathering meta data 
! we would need a feature that keeps meta collection separate by AIO op handle 
! we can make a choice to force all data flavors into a single dumpset.  Do we do that? 
! Until that is decided we need all flavors in-flight to be complete 

SUBROUTINE find_aio_slot(self, dump)
   
    CLASS(IO) :: self

    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WIS) :: op

    self%aio_slot = 0

    DO op = 1, WNUM_AIO_OPS_INFLIGHT ! find an AIO slot we can use that is open

        IF (.NOT. dump%aio_ops%inflight(op)) THEN

            self%aio_slot = op

            dump%aio_ops%inflight(op) = .TRUE.

            EXIT

        END IF

    END DO

    IF (self%aio_slot .EQ. 0) THEN ! if none are open, wait on the oldest one to complete and use that

        self%aio_slot = 1

        CALL self%wait_aio_slots(dump, 0, .TRUE.)

        dump%aio_ops%inflight(self%aio_slot) = .TRUE.

    END IF

    self%buf_offset = 1 ! find the max buffer offset + size of any inflight IO operation, that will be our starting offset

    DO op = 1, WNUM_AIO_OPS_INFLIGHT 

        IF (dump%aio_ops%inflight(op)) THEN

            self%buf_offset = MAX(self%buf_offset, dump%aio_ops%offsets(op,2) + dump%aio_ops%lengths(op,2))

        END IF

    END DO

END SUBROUTINE find_aio_slot

! Find the filenames in the current dump: aio_root, manifestFile, dataFile.

SUBROUTINE find_filenames(self, dump)

    CLASS(IO) :: self

    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    IF (self%aioOn) THEN 

        WRITE(dump%manifestFile(self%aio_slot), '(A5, "-", I4.4, "-000")') dump%FileRoot, dump%Number

        WRITE(dump%aio_root, '(A5, "-", I4.4, "-part", I2.2)') dump%FileRoot, dump%Number, self%fpart

        CALL AIO_GEN_DATA_FILE_NAME(dump%aio_root, dump%dataFile)

    ELSE

        WRITE(dump%manifestFile(self%aio_slot), '(A, "-", I4.4, "-", I3.3)') TRIM(dump%FileRoot), dump%Number, WRANK

        WRITE(dump%dataFile, '(A, "-", I4.4, "-part", I2.2, "-", I3.3)') TRIM(dump%FileRoot), dump%Number, self%fpart, WRANK

    END IF

END SUBROUTINE find_filenames

! When this is called, we've packed a single send into self%buf, including the plines.
! We first increment our overall buffer offset to handle the next send, register the starting address and length 
! for the write and fire off the write. We either do an AIO write or a local write (AIO manifest completion is 
! done at the final wait on the slot)

SUBROUTINE write_file(self, dump, patch_data_size, manifest_head_size, manifest_pline_size)

    CLASS(IO) :: self

    TYPE(IO_DUMP), INTENT(INOUT) :: dump
    INTEGER(WID), INTENT(IN) :: patch_data_size, manifest_head_size, manifest_pline_size
    INTEGER(WID) :: i

    dump%aio_ops%offsets(self%aio_slot,1) = self%buf_offset
    dump%aio_ops%lengths(self%aio_slot,1) = self%n_patches_in_write * patch_data_size
    dump%aio_ops%offsets(self%aio_slot,2) = dump%aio_ops%lengths(self%aio_slot,1) + 1
    dump%aio_ops%lengths(self%aio_slot,2) = manifest_head_size + self%n_patches_in_write * manifest_pline_size

    IF (self%aioOn) THEN

        CALL AIO_WRITE_BUF_META(TRIM(dump%Path) // '/' // TRIM(dump%aio_root), &
                                dump%aio_ops%lengths(self%aio_slot,1), &
                                dump%aio_ops%lengths(self%aio_slot,2), &
                                dump%buf(dump%aio_ops%offsets(self%aio_slot,1)), &
                                dump%aio_ops%slotIDs(self%aio_slot), 0)

    ELSE

        OPEN(UNIT=12, FILE=TRIM(dump%Path) // '/' // dump%dataFile, FORM="UNFORMATTED", &
             ACCESS="STREAM", STATUS="REPLACE")

        WRITE(12) (dump%buf(i), i=dump%aio_ops%offsets(self%aio_slot,1), &
                   dump%aio_ops%offsets(self%aio_slot,1) + &
                   dump%aio_ops%lengths(self%aio_slot,1) - 1)

        CLOSE(12)
        
        IF (self%fpart .EQ. 0) THEN

            OPEN(UNIT=12, FILE=TRIM(dump%Path) // '/' // dump%manifestFile(self%aio_slot), &
                 FORM="UNFORMATTED", ACCESS="STREAM", STATUS="REPLACE")

        ELSE

            OPEN(UNIT=12, FILE=TRIM(dump%Path) // '/' // dump%manifestFile(self%aio_slot), &
                 FORM="UNFORMATTED", ACCESS="STREAM", STATUS="OLD", POSITION='APPEND')

        END IF

        ! TODO, add in MPI_GATHER for other ranks' manifests

        WRITE(12) (dump%buf(i), i=dump%aio_ops%offsets(self%aio_slot,2), &
                   dump%aio_ops%offsets(self%aio_slot,2)+dump%aio_ops%lengths(self%aio_slot,2)-1)

        CLOSE(12)

    END IF

END SUBROUTINE

! add the manifest record pline and and patch data to dump%buf until full or all done. 
! 1D, 2D & 3D version, essentially the same

SUBROUTINE add_patches_1d(self, workd, dump, manifest_pline_size, patch_data_size, mygamma)

    CLASS(IO) :: self

    TYPE(Domain), INTENT(IN) :: workd
    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WID), INTENT(IN) :: manifest_pline_size, patch_data_size
    REAL(WID), INTENT(IN) :: mygamma

    LOGICAL :: write_patch
    INTEGER(WIS) :: mymanifest_offset, pi
    INTEGER(WID) :: mydata_offset
    
    !$OMP DO SCHEDULE(DYNAMIC)
    DO pi = 1, workd%npx

        write_patch = .FALSE.

        !$OMP CRITICAL
        IF (workd%domain1d(pi)%hasZones  .AND. &
            .NOT. self%patch_added1d(pi) .AND. &
            self%n_patches_added .LT. self%n_patches_in_write) THEN
             
            mymanifest_offset    = self%manifest_offset ! reserve space by incrementing offsets
            mydata_offset        = self%data_offset

            self%manifest_offset = self%manifest_offset + manifest_pline_size
            self%data_offset     = self%data_offset + patch_data_size
            self%n_patches_added = self%n_patches_added + 1
            
            write_patch          = .TRUE.

        END IF
        !$OMP END CRITICAL

        IF (write_patch) THEN ! Not all threads might get an offset for small buffers

            CALL self%add_pline_to_manifest(workd, dump, pi, 1, 1, mymanifest_offset)

            CALL self%append_patch_data(workd, dump, pi, 1, 1, mydata_offset, mygamma)

            self%patch_added1d(pi) = .TRUE.

        END IF

    END DO
    !$OMP END DO

END SUBROUTINE add_patches_1d

SUBROUTINE add_patches_2d(self, workd, dump, manifest_pline_size, patch_data_size, mygamma)

    CLASS(IO) :: self

    TYPE(Domain), INTENT(IN) :: workd
    TYPE(IO_DUMP), INTENT(INOUT) :: dump

    INTEGER(WID), INTENT(IN) :: manifest_pline_size, patch_data_size
    REAL(WID), INTENT(IN) :: mygamma

    LOGICAL :: write_patch
    INTEGER(WIS) ::  mymanifest_offset, p, pi, pj
    INTEGER(WID) :: mydata_offset

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workd%npx*workd%npy

        CALL collapse2d(p, workd%npx, workd%npy, pi, pj) ! obtain the x & y location of the Patch pi,pj

        write_patch = .FALSE.

        !$OMP CRITICAL
        IF (workd%domain2d(pi,pj)%hasZones  .AND.  &
            .NOT. self%patch_added2d(pi,pj) .AND. &
            self%n_patches_added .LT. self%n_patches_in_write) THEN

            mymanifest_offset    = self%manifest_offset
            mydata_offset        = self%data_offset

            self%manifest_offset = self%manifest_offset + manifest_pline_size
            self%data_offset     = self%data_offset + patch_data_size
            self%n_patches_added = self%n_patches_added + 1
            
            write_patch          = .TRUE.

        END IF
        !$OMP END CRITICAL
            
        IF (write_patch) THEN
                
            CALL self%add_pline_to_manifest(workd, dump, pi, pj, 1,  mymanifest_offset)

            CALL self%append_patch_data(workd, dump, pi, pj, 1, mydata_offset, mygamma)

            self%patch_added2d(pi,pj) = .TRUE.

        END IF

    END DO
    !$OMP END DO

END SUBROUTINE add_patches_2d

SUBROUTINE add_patches_3d(self, workd, dump, manifest_pline_size, patch_data_size, mygamma)

    CLASS(IO) :: self

    TYPE(Domain), INTENT(IN) :: workd
    TYPE(IO_DUMP), INTENT(INOUT) :: dump
    INTEGER(WID), INTENT(IN) :: manifest_pline_size, patch_data_size
    REAL(WID), INTENT(IN) :: mygamma

    LOGICAL :: write_patch
    INTEGER(WIS) ::  mymanifest_offset, p, pi,pj,pk
    INTEGER(WID) :: mydata_offset

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workd%npx*workd%npy*workd%npz

        CALL collapse3d(p, workd%npx, workd%npy, workd%npz, pi, pj, pk)

        write_patch = .FALSE.
                
        !$OMP CRITICAL
        IF (workd%domain3d(pi,pj,pk)%hasZones  .AND. &
            .NOT. self%patch_added3d(pi,pj,pk) .AND. &
            self%n_patches_added .LT. self%n_patches_in_write) THEN
 
            mymanifest_offset    = self%manifest_offset
            mydata_offset        = self%data_offset

            self%manifest_offset = self%manifest_offset + manifest_pline_size
            self%data_offset     = self%data_offset + patch_data_size
            self%n_patches_added = self%n_patches_added + 1

            write_patch          = .TRUE.

        END IF
        !$OMP END CRITICAL
            
        IF (write_patch) THEN
                    
            CALL self%add_pline_to_manifest(workd, dump, pi, pj, pk,  mymanifest_offset)

            CALL self%append_patch_data(workd, dump, pi, pj, pk, mydata_offset, mygamma)

            self%patch_added3d(pi,pj,pk) = .TRUE.

        END IF

    END DO

    !$OMP END DO

END SUBROUTINE add_patches_3d

END MODULE Mod_IO
