#include "macro.h"

MODULE Mod_IO_Dump

! FILENAME: io_dump.f90

! DESCRIPTION: This module is a child module of Mod_IO and defines the datatype of a
!              dump / snapshot. It does little more than holding the derived variables,
!              read the parameters from file and instantiate them. The object is 'mutated' 
!              by Mod_IO that takes an IO_Dump as input and also handles the AIO or direct output.
!
! DEPENDENCIES: io.f90
!
! HISTORY: 
!    7/8/17 - Julius Donnert

USE OMP_LIB

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Domain

IMPLICIT NONE

PRIVATE

PUBLIC :: IO_DUMP

INTEGER(WIS), PUBLIC, PARAMETER :: IO_DUMP_COMPRESSED=1, IO_DUMP_MOVIE=2, IO_DUMP_RESTART=3

TYPE AIO_OPERATIONS ! type for an AIO operation

    INTEGER(WADDRESS_KIND) :: offsets(WNUM_AIO_OPS_INFLIGHT,2)  ! offset in dump%buf of this slot
    INTEGER(WADDRESS_KIND) :: lengths(WNUM_AIO_OPS_INFLIGHT,2)  ! length in dump%buf of this slot
    INTEGER(WIS)           :: slotIDs(WNUM_AIO_OPS_INFLIGHT)    ! unique AIO slot ID
    LOGICAL(WIS)           :: inflight(WNUM_AIO_OPS_INFLIGHT)   ! flag if a slot is currently being processed

END TYPE AIO_OPERATIONS

TYPE IO_FIELD ! Type of an AIO output field defined in the parameter file at column [X]
   
     CHARACTER(5)  :: Symbol(WMAX_IO_FIELDS) ! [3] unique 5 byte human readable ID
     CHARACTER(5)  :: Prec(WMAX_IO_FIELDS)   ! [5] precision: real8, real4, byte2, byte1
     CHARACTER(8)  :: Map(WMAX_IO_FIELDS)    ! [6] mapping options: "linear", "log_e", "natural"
     REAL(WRS)     :: Min(WMAX_IO_FIELDS)    ! [7] lower bound of mapping
     REAL(WRS)     :: Max(WMAX_IO_FIELDS)    ! [8] upper bound of mapping

END TYPE IO_FIELD

TYPE IO_DUMP
    
    LOGICAL(WIS)              :: On             ! is this type of dump enabled ?
    CHARACTER(WCHAR_BUF_SIZE) :: Name           ! name of this dump type for log messages
    CHARACTER(1), ALLOCATABLE :: buf(:)         ! 1D buffer to hold the converted dump data
    INTEGER(WID)              :: buf_size       ! size of that buffer
    INTEGER(WIS)              :: Number         ! number of this dump
    CHARACTER(WCHAR_BUF_SIZE) :: Host           ! write to Host not file visualization (not implemented)
    CHARACTER(WCHAR_BUF_SIZE) :: FileRoot       ! root file name if AIO is not used
    CHARACTER(WCHAR_BUF_SIZE) :: Path           ! patch to the file
    REAL(WRD)                 :: Interval       ! dump interval
    INTEGER(WIS)              :: Nfields        ! was _active_fields
    TYPE(IO_FIELD)            :: fields         ! fields defined in the parameter field
    TYPE(AIO_OPERATIONS)      :: aio_ops        ! AIO bookeeping ops
    CHARACTER(WCHAR_BUF_SIZE) :: aio_root       ! path for aio file name generation
    CHARACTER(WCHAR_BUF_SIZE) :: dataFile       ! current data file
    CHARACTER(WCHAR_BUF_SIZE) :: manifestFile(WNUM_AIO_OPS_INFLIGHT) ! manifest filename for each slot
    REAL(WRD)                 :: next_dump_time ! next time to write a snapshot 
    INTEGER(WIS)              :: slotID_offset  ! offset so AIO slot IDs are unique among dumpsets

    CONTAINS
    
    PRIVATE

    PROCEDURE, PUBLIC :: destroy 

END TYPE IO_DUMP

INTERFACE IO_DUMP

    MODULE PROCEDURE constructor

END INTERFACE IO_DUMP

CONTAINS

FUNCTION constructor(param, simunits, flavor) RESULT(self)
    
    TYPE(IO_DUMP) :: self

    TYPE(Parameters) :: param
    TYPE(SimulationUnits) :: simunits
    
    INTEGER(WIS), INTENT(IN) :: flavor
    INTEGER(WIS) :: i
    CHARACTER(WCHAR_BUF_SIZE) :: strbuf, prec, map

    SELECT CASE (flavor) ! Select which dump type we are initializing
 
        CASE(IO_DUMP_COMPRESSED)

            self%Name           = 'compr'
            self%slotID_offset  = 0
            WRITE(strbuf, '(I12)') WCOMPRESSED_IOBUFFER_SIZE
            prec = "real4"      ! default output precision 
            map = "natural"     ! default output mapping

        CASE(IO_DUMP_MOVIE)

            self%Name           = 'movie'
            self%slotID_offset  = WNUM_AIO_OPS_INFLIGHT
            WRITE(strbuf, '(I12)') WMOVIE_IOBUFFER_SIZE
            prec = "byte1"
            map = "linear"

        CASE(IO_DUMP_RESTART)
            
            self%Name           = 'restart'
            self%slotID_offset  = 2*WNUM_AIO_OPS_INFLIGHT
            WRITE(strbuf, '(I12)') WRESTART_IOBUFFER_SIZE
            prec = "real8"
            map = "natural"

        CASE DEFAULT
            ASSERT(.FALSE., 'Dumptype currently not supported')

    END SELECT

    CALL param%section("IO Dump Type '"//TRIM(self%Name)//"'")

    self%On       = param%getLog(TRIM(self%Name)//"On",       "false",     "# Enable output type "//self%Name)
    self%Interval = param%getFlt(TRIM(self%Name)//"Interval", "1.0",       "# Time interval between outputs in code units")
    self%Number   = param%getInt(TRIM(self%Name)//"Number",   "0",         "# [0] dump number to start from")

    self%buf_size = param%getInt(TRIM(self%Name)//"BufSize", TRIM(strbuf), "# Size of com buffer & thus data files in MBytes", &
                                 0, .TRUE.)
    self%FileRoot = param%getStr(TRIM(self%Name)//"FileRoot", self%Name,   "# File prefix of dump type", 0, .FALSE.)
    self%Path     = param%getStr(TRIM(self%Name)//"Path",     "./",        "# Path to the snapshots", 0, .TRUE.)
    self%Host     = param%getStr(TRIM(self%Name)//"Host",     "localhost", "# Output host (unused)", 0, .TRUE.)

    self%FileRoot = TRIM(self%FileRoot)
    self%Path     = TRIM(self%Path)
    self%Host     = TRIM(self%Host)

    ! read IO blocks

    IF (GENPARFILE_ON) THEN ! Default fields

        self%fields%Symbol(1) = param%getStr(TRIM(self%Name)//"(1)", "RHO "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# Density")
        self%fields%Symbol(2) = param%getStr(TRIM(self%Name)//"(2)", "VX  "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# x velocity")
        self%fields%Symbol(3) = param%getStr(TRIM(self%Name)//"(3)", "VY  "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# y velocity")
        self%fields%Symbol(4) = param%getStr(TRIM(self%Name)//"(4)", "VZ  "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# z velocity")
        self%fields%Symbol(5) = param%getStr(TRIM(self%Name)//"(5)", "BX  "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# x Mag Field")
        self%fields%Symbol(6) = param%getStr(TRIM(self%Name)//"(6)", "BY  "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# y Mag Field")
        self%fields%Symbol(7) = param%getStr(TRIM(self%Name)//"(7)", "BZ  "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# z Mag Field")
        self%fields%Symbol(8) = param%getStr(TRIM(self%Name)//"(8)", "PRS "//TRIM(prec)//" "//TRIM(map)//" 1  1 ", "# Pressure ")
    
        i = 9

    ELSE ! actual reading

        DO i = 1, WMAX_IO_FIELDS
            
            IF (I.LT.10) THEN
                WRITE(strbuf, '(A,"(",I1,")")') TRIM(self%Name), i ! e.g. compressed(1) ...
            ELSE
                WRITE(strbuf, '(A,"(",I2,")")') TRIM(self%Name), i ! e.g. movie(10) ...
            ENDIF

            self%fields%Symbol(i) = param%getStr(strbuf, "",    "", 1, .FALSE. )

            IF (self%fields%Symbol(i) .EQ. "") THEN 
                EXIT ! reached end of field list in parameter file
            END IF

            self%fields%Prec(i) = param%getStr(strbuf,   "",    "", 2, .FALSE. )
            self%fields%Map(i) = param%getStr(strbuf,    "",    "", 3, .FALSE. )
            self%fields%Min(i) = param%getFlt(strbuf,    "0.0", "", 4, .FALSE. )
            self%fields%Max(i) = param%getFlt(strbuf,    "0.0", "", 5, .FALSE. )

        END DO ! i

    ENDIF

    self%Nfields = i-1

    ASSERT(self%Nfields.GT.1 .OR. .NOT. self%On, "IO dump '"//TRIM(self%Name)//"' switched ON but output fields are missing")

    self%buf_size = self%buf_size * 1024**2    ! MByte to Byte

    ALLOCATE( self%buf(self%buf_size) ) ! TODO, replace with MPI_ALLOC_MEM

    DO i = 1, self%buf_size ! page this arrays now.  implied do loops here caused troubles for Intel and CCE
        self%buf(i) = CHAR(0)
    END DO

    self%Number = self%Number - 1 ! rewind the dump numbers by one so we can keep the dump counter consistent 
                                  ! (NOTE that when restarts are added this must be recalled)

    DO i = 1, WNUM_AIO_OPS_INFLIGHT ! initialize our AIO_OPS arrays

        self%aio_ops%slotIDs(i)   = self%slotID_offset + i
        self%aio_ops%inflight(i)  = .FALSE.
        self%aio_ops%offsets(i,:) = 0
        self%aio_ops%lengths(i,:) = 0

    END DO

    self%next_dump_time = simunits%t_beg ! default to writing out the next movie and compressed dumps at the initial time
    
    RETURN 

END FUNCTION constructor

SUBROUTINE destroy(self)

    CLASS(IO_DUMP) :: self

    if (ALLOCATED(self%buf)) DEALLOCATE(self%buf)

END SUBROUTINE destroy

END MODULE Mod_IO_DUMP
