#include "macro.h"

MODULE Mod_MHDWENO

!    FILENAME: mhdweno.f90
!                                 
!    DESCRIPTION: WENO 5th order implementation of CT-MHD.
!    This classical scheme is a standard 4-stage 4th order Runge-Kutta time integrator
!    with JW96 5th order WENO spatial discretization. We split the time integration into 8 passes,
!    4 for WENO state vector update and 4 for the CT+ES update, alternating between them. We  
!    are communicating 3 bnd zones between them. This is the optimal balance of memory 
!    requirements, flops and communication, but makes the RK4 scheme look a bit weird.
!
!    We use the J18 E/S scheme to conserve entropy  in cold flows 
!    (e.g. at high redshift in cosmological simulations). This requires saving the entropy S
!    in the state variable "q" as a ninth component. It is initialized at entry and generally not stored
!    on the global grid "gridXd". However, we need to carry it over from WENO to CT step. So we replace
!    the 5th component with S on even to odd cycles.  In 2D, the ES flag is recomputed on every substep.
!    In 3D, we store the ES flag in 'ishock' and save it across the 2*4 rk4 steps in SIGN(wenoq0(:,1)), 
!    as density is positive definite.
!    In cosmological simulations, the entropy switch is usually enabled to suppress artifical heating
!    from the inevitable particle noise (S10).    
!    
!    We use the CT scheme from R98. The "real" magnetic fields are defined on the faces 
!    "bfaceXd" and are updated after a boundary exchange of the corner fluxes "emfXd" 
!    (odd passes). The fields are interpolated back into the zones in 4th order.

!    The implementation is entirely 1D. 2D and 3D grids are flattened into 1D arrays on entry
!    to increase vector length. We completely separate memory operations and computations into 
!    separate subroutines, as the memory work can't be vectorized.
!    Multi-dimensions are treated by rotating the flattened data array and integrating along the first 
!    index again. In 2D, we rotate the state forward and the fluxes backwards when integrating along
!    the y direction. In 3D the rotation is cyclic, so we only rotate forward thrice. The CT fluxes 
!    however are rotated separately forward and backward.
!
!    In the computation heavy functions, all loops vectorize with an HPC compiler. 
!    To make life easier for the auto-vectorizer, we write loops with minimal content. The 
!    compiler will fuse them, if its metrics deem it beneficial considering SIMD loop overhead vers. 
!    cache  spilling. 
!    The eigenvector functions are further scalar optimized, i.e. similar components are grouped into on
!    loop and scalar "tmp" variables are introduced to reduce computation, most compilers do a good job 
!    here though.
!
!    If you're reading this for the first time, start with the variables & 1D routines and 
!    work your way up. The core drivers are in the RK4_integrator routine, the rest are memory 
!    operations.
!   
!    REFERENCES: Donnert+ 2018 (D18), Jang+ 2018 (J18), Jiang & Wu 1996 (JW96), 
!                Ryu+ 1993 (R93), Ryu+ 1998 (R98), Shu 2009, SIAM Review 51, 1 (S09)
!    HISTORY:
!       3/8/17  - Julius Donnert

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_Patch
USE Mod_MHDPool
USE, INTRINSIC :: IEEE_ARITHMETIC

IMPLICIT NONE

PRIVATE

PUBLIC :: MHDWENO

! Parameters

REAL(WRD), PARAMETER :: INVSQRT2 = 1.0_WRD/SQRT(2.0_WRD)

REAL(WRD), PARAMETER :: COURFAC=0.8_WRD, EPS = 1d-6, EPSZ = 1d-40 ! JW96, Borges+2011

! Protection Parameters. To make any scheme robust, we need to catch cases where B=0 and
! protect against negative densities and pressures.
REAL(WRD), PARAMETER :: ZERO = 0d0, DENSITY_FLOOR = 1d-16, PRESSURE_FLOOR = 1d-16

REAL(WRD), PARAMETER :: LIN_WEIGHTS(3) = (/ 1.0_WRD/16.0_WRD, 5.0_WRD/8.0_WRD, 5.0_WRD/16.0_WRD /)!S09 eq2.5+1

! Scratch/pool variables, threadprivate to prevent NUMA effects, 
! only one index in 2D and 3D because of flattening

REAL(WRD), ALLOCATABLE :: & 

            ! state vector of conserved quantities : q,q0 = [rho,vx*rho,vy*rho,vz*rho,Bx,By,Bz,E,S]
            q(:,:),                       & ! current running state vector      <-> workp%grid
            q0(:,:),                      & ! initial state vector              <-> workp%wenoq0

            qsave(:,:),                   & ! result buffer state vector (no S) <-> workp%wenoqsave
                
            ! the R98/L13 CT schemes track the magnetic field on the faces and edges
            b(:,:),                       & ! face centered magnetic field      <-> workp%bface
            b0(:,:),                      & ! initial face centered bfld        <-> workp%wenoB0
            bsave(:,:),                   & ! result buffer face centered bfld  <-> workp%wenoBsave
            omega(:,:),                   & ! corner fluxes                     <-> workp%emf2d
            gsx(:), gsz(:),               & ! edge fluxes
            fsy(:), fsz(:),               &
            hsx(:), hsy(:),               & 

            ! Riemann solver variables from running state vector q
            P(:),                         & ! pressure P(E)
            A(:,:),                       & ! cell centered eigenvalues 
            F(:,:),                       & ! cell centered fluxes
            L(:,:,:), R(:,:,:),           & ! eigenvectors, linear extrap. to interfaces
            Fs(:,:),                      & ! WENO5 interpolated fluxes on the interfaces, 8th comp is ES-flux

            ! True scratch arrays
            qbuf(:,:),                    & ! state vector buffer for rotations/multi-D etc, incl. S
            buf(:,:),                     & ! generic scratch buffer used at several places

            sqrt_rho(:),                  & ! square root of density
            vx(:),vy(:),vz(:),            & ! linearly interpolated velocity on faces
            cs(:), af(:), as(:),          & ! wave speeds 
            lf(:), la(:), ls(:),          & ! eigenvalues on faces
            bty(:),btz(:), dqsk(:,:)        

!$OMP THREADPRIVATE(q0,q,qsave,qbuf, P,b,b0,bsave, A,L,R, F,omega)
!$OMP THREADPRIVATE(sqrt_rho,vx,vy,vz,cs,af,as,lf,la,ls,bty,btz,Fs,buf,dqsk,gsx,gsz,fsy,fsz,hsx,hsy)

INTEGER(WIS), ALLOCATABLE :: signB(:),ishock(:) 
!$OMP THREADPRIVATE(signB,ishock)

REAL(WRD) :: dx, dt, id
!$OMP THREADPRIVATE(dx,dt,id)

! N is full 1D length of patch, incl bnd zones
! dj and dk are the index strides of (i,j+1) and (i,j,k+1) !
! Obviously, di = 1.

INTEGER(WIS) :: nx, ny, nz, nb, N, dj, dk, pcount
!$OMP THREADPRIVATE(N,nx,ny,nz,nb,dj,dk)

TYPE :: MHDWENO

    PRIVATE

    REAL(WRD) :: vmax, gamma, ndim
    REAL(WRD) :: es_Mthres, es_vel
    LOGICAL   :: high_order_CT

    CONTAINS 

    PRIVATE

    PROCEDURE, PUBLIC :: solve
    PROCEDURE, PUBLIC :: allocate_thread_local_variables
    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: boundary_flavor_req

    PROCEDURE :: flatten_1d
    PROCEDURE :: RK4_integrator_1d
    PROCEDURE :: copyback_1d
    PROCEDURE :: update_state_vector_1d
    PROCEDURE :: compute_interface_vmax_1d
    PROCEDURE :: switch_ES_1d

    PROCEDURE :: flatten_2d
    PROCEDURE :: RK4_integrator_2d
    PROCEDURE :: copyback_2d
    PROCEDURE :: update_state_vector_2d
    PROCEDURE :: transpose_state_fwd_2d
    PROCEDURE :: transpose_fluxes_bwd_2d
    PROCEDURE :: compute_CT_interface_flux_2d
    PROCEDURE :: compute_CT_corner_flux_2d
    PROCEDURE :: update_magnetic_fields_2d
    PROCEDURE :: interpolate_magnetic_field_b2q_2d
    PROCEDURE :: compute_magnetic_field_divergence_2d
    PROCEDURE :: save_WENO_step_2d
    PROCEDURE :: compute_interface_vmax_2d
    PROCEDURE :: switch_ES_2d
    
    PROCEDURE :: flatten_3d
    PROCEDURE :: RK4_integrator_3d
    PROCEDURE :: copyback_3d
    PROCEDURE :: update_state_vector_3d
    PROCEDURE :: rotate_states_fwd_3d
    PROCEDURE :: rotate_CT_flux_fwd_3d
    PROCEDURE :: rotate_CT_flux_bwd_3d
    PROCEDURE :: compute_CT_interface_flux_3d
    PROCEDURE :: compute_CT_corner_flux_3d
    PROCEDURE :: interpolate_magnetic_field_b2q_3d
    PROCEDURE :: update_magnetic_fields_3d
    PROCEDURE :: compute_magnetic_field_divergence_3d
    PROCEDURE :: save_WENO_step_3d
    PROCEDURE :: compute_interface_vmax_3d
    PROCEDURE :: compute_ES_switch_3d
    PROCEDURE :: apply_ES_switch_3d
    PROCEDURE :: check_state_3d

    PROCEDURE :: compute_pressure
    PROCEDURE :: compute_fluxes
    PROCEDURE :: compute_eigenvalues
    PROCEDURE :: compute_eigenvectors
    PROCEDURE :: compute_WENO5_fluxes
    PROCEDURE :: compute_WENO5_interpolation
    PROCEDURE :: compute_ES_fluxes
    PROCEDURE :: init_S_from_E
    PROCEDURE :: apply_protection_floors
    PROCEDURE :: S2E
    PROCEDURE :: E2S
    PROCEDURE :: E2P
    PROCEDURE :: signum

    PROCEDURE :: PQ3
    PROCEDURE :: PQ3RMBND
    PROCEDURE :: PB3
    PROCEDURE :: PQSX
    PROCEDURE :: PQSY
    PROCEDURE :: PQSZ
    PROCEDURE :: PISZ
    PROCEDURE :: PSSZ
    PROCEDURE :: PQSZNB
    PROCEDURE :: PQ
    PROCEDURE :: PQI
    PROCEDURE :: PQRMBND
    PROCEDURE :: PQINV
    PROCEDURE :: PS
    PROCEDURE :: PSINV
    PROCEDURE :: PQNB
    PROCEDURE :: PQ2
    PROCEDURE :: PQ2RMBND

END TYPE MHDWENO

INTERFACE MHDWENO

    MODULE PROCEDURE constructor

END INTERFACE

CONTAINS

FUNCTION constructor(param, gamma) RESULT(self)

    TYPE(MHDWENO) :: self
    TYPE(Parameters) :: param

    REAL(WRD), INTENT(IN) :: gamma

    self%gamma = gamma

    CALL param%section("MHD WENO5 solver (Donnert et al. 2018, Jang et al. 2018)", .FALSE.)
    
    self%es_vel = param%getFlt("es_vel", "0.05", "# [0.05] ES velocity criterion Jang+ 2018", 0, .FALSE.) 
    self%es_Mthres = param%getFlt("es_Mthres", "1.1" ,"# Mach threshold for E code,  (Springel 2010)", 0,.TRUE.)

    self%high_order_CT  = param%getLog("high_order_CT", "false", "#Enable Lee13 CT scheme (broken)", 0, .TRUE.) 

    RETURN 

END FUNCTION

SUBROUTINE destroy(self)

    CLASS(MHDWENO) :: self

    !$OMP PARALLEL
    DEALLOCATE(q0,q,qsave,qbuf,P,F,L,R,A,sqrt_rho,cs,vx,vy,vz,af,as,lf,la, &
               ls,bty,btz,signB,ishock,Fs,buf,dqsk,gsx,gsz,fsy,fsz,hsx,hsy)
    !$OMP END PARALLEL

END SUBROUTINE destroy

! Work arrays are all thread local to get the allocated memory close to the CPU.
! This could lead to memory constraints for very large thread counts.
! The storage is allocated only if mhdwenoOn .EQ. .TRUE.

SUBROUTINE allocate_thread_local_variables(self, ndim, nx_in, ny_in, nz_in, nb_in)

    CLASS(MHDWENO) :: self

    INTEGER(WIS), INTENT(IN) :: ndim, nx_in, ny_in, nz_in, nb_in
    REAL(WID) :: overhead
    
    RASSERT(nb_in .GE. 3, "Set patch_nb >= 3 for WENO5, current patch_nb = "//I2A(nb_in))

    !$OMP MASTER
    self%ndim = ndim
    !$OMP END MASTER

    nb = nb_in
 
    nx = nx_in
    N = nx + 2*nb

    IF (ndim .EQ. 2) THEN
        ny = ny_in
        N = N * (ny+2*nb)
    ELSE
        ny = 0
    END IF

    IF (ndim .EQ. 3) THEN

        ny = ny_in
        nz = nz_in

        N = (nx + 2*nb) * (ny+2*nb) * (nz+2*nb)
    ELSE
        nz = 0
    END IF

    ALLOCATE(q(N,9),qbuf(N,9),P(N), &
             F(N,7),L(N,7,7),R(N,7,7),A(N,7), &
             sqrt_rho(N),cs(N),vx(N),vy(N),vz(N),af(N),as(N),lf(N),la(N), &
             ls(N),bty(N),btz(N),signB(N), Fs(N,8), buf(N,9), dqsk(N,6), &
             gsx(N),gsz(N),fsz(N),fsy(N),hsx(N),hsy(N))

    ALLOCATE(q0(nx*MAX(1,ny)*MAX(1,nz),9),qsave(nx*MAX(1,ny)*MAX(1,nz),8),&
             ishock(nx*MAX(1,ny)*MAX(1,nz))) ! no boundary zones

    q0(:,:)    = 0.0_WRD ! page thread local arrays
    q(:,:)     = 0.0_WRD 
    qsave(:,:) = 0.0_WRD 
    qbuf(:,:)  = 0.0_WRD
    F(:,:)     = 0.0_WRD
    L(:,:,:)   = 0.0_WRD
    R(:,:,:)   = 0.0_WRD
    A(:,:)     = 0.0_WRD
    P(:)       = 0.0_WRD

    sqrt_rho(:) = 0.0_WRD
    vx(:) = 0.0_WRD
    vy(:) = 0.0_WRD
    vz(:) = 0.0_WRD
    cs(:) = 0.0_WRD
    af(:) = 0.0_WRD
    as(:) = 0.0_WRD
    lf(:) = 0.0_WRD
    la(:) = 0.0_WRD
    ls(:) = 0.0_WRD
    bty(:) = 0.0_WRD
    btz(:) = 0.0_WRD
    signB(:)  = 0_WIS
    ishock(:) = 0_WIS
    buf(:,:)  = 0.0_WRD
    Fs(:,:)   = 0.0_WRD
    dqsk(:,:) = 0.0_WRD
    gsx(:) = 0.0_WRD
    gsz(:) = 0.0_WRD
    fsy(:) = 0.0_WRD
    fsz(:) = 0.0_WRD
    hsx(:) = 0.0_WRD
    hsy(:) = 0.0_WRD

    IF (ndim .EQ. 2) THEN

        ALLOCATE(b(N,2),b0(N,2),bsave(N,2), omega(N,1))

        b(:,:) = 0.0_WRD
        b0(:,:) = 0.0_WRD
        bsave(:,:) = 0.0_WRD
        omega(:,:) = 0.0_WRD
                     
    ELSE IF (ndim .EQ. 3) THEN
       
        ALLOCATE(b(N,3),b0(N,3),bsave(N,3), omega(N,3))

        b(:,:) = 0.0_WRD
        b0(:,:) = 0.0_WRD
        bsave(:,:) = 0.0_WRD
        omega(:,:) = 0.0_WRD

    END IF

    !$OMP MASTER

    overhead = MEMSIZE(q0)   + MEMSIZE(q)    + MEMSIZE(qsave) + MEMSIZE(qbuf) &
             + MEMSIZE(A)    + MEMSIZE(F)    + MEMSIZE(L)     + MEMSIZE(R)   & 
             + MEMSIZE(b)    + MEMSIZE(omega)+ MEMSIZE(bsave) + MEMSIZE(ishock) &
             + MEMSIZE(b0)   + MEMSIZE(P)    + MEMSIZE(vx)    + MEMSIZE(sqrt_rho)&
             + MEMSIZE(vy)   + MEMSIZE(vz)   + MEMSIZE(cs)    + MEMSIZE(af) &
             + MEMSIZE(as)   + MEMSIZE(lf)   + MEMSIZE(ls)    + MEMSIZE(bty) &
             + MEMSIZE(btz)  + MEMSIZE(signB)+ MEMSIZE(la)    + MEMSIZE(buf) &
             + MEMSIZE(dqsk) + MEMSIZE(Fs)   + MEMSIZE(gsx)   + MEMSIZE(fsy) &
             + MEMSIZE(gsz)  + MEMSIZE(fsz)  + MEMSIZE(hsx)   + MEMSIZE(hsy) 

    RMSG("WENO Solver")
    RMSG("   Memory overhead per thread : "//F2A(overhead)//" MB")
    RMSG("   Memory overhead per rank   : "//F2A(overhead*WNTHREADS)//" MB")

#ifdef WENOZ
    RMSG("   Using WENO-Z weights (Borges et al. 2008)")
#else
    RMSG("   Using WENO-JW weights (Jiang & Wu 1998)")
#endif

#ifdef USE_PROTECTIONS
    RMSG("   Using protection fluxes and floors")
#endif

    !$OMP END MASTER

END SUBROUTINE allocate_thread_local_variables

! top-level routine that returns the next required boundary flavor type given the overall pass count
FUNCTION boundary_flavor_req(self, update_pass_count) RESULT(bflavor)

    CLASS(MHDWENO) :: self
    INTEGER(WIS), INTENT(IN) :: update_pass_count
    INTEGER(WIS) :: bflavor

    IF (MOD(update_pass_count, 2) .EQ. 0) THEN

        bflavor = FL_MHDTVD

    ELSE

        bflavor = FL_MHDTVD ! This need optimization
        
    END IF

    RETURN

END FUNCTION boundary_flavor_req

! The actual solve. This function is called 8 times asynchronously, 
! i.e. by different threads and in different patch order.

SUBROUTINE solve(self, workp, maxsignal, update_complete)

    CLASS(MHDWENO) :: self
    TYPE(Patch) :: workp
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    LOGICAL(WIS), INTENT(OUT) :: update_complete
    REAL(WRD) :: vmax(3)
    INTEGER(WID) :: i, j, k, idx,m, ii,jj

    dx  = workp%dx ! copy scalar patch properties into thread local variables
    dt  = workp%dt ! we use them like globals and don't pass them around for clarity
    nb  = workp%nb ! they are permutated accordingly when the cube is rotated in 2D and 3D
    nx  = workp%nx
    ny  = workp%ny
    nz  = workp%nz

    pcount = workp%update_pass_count

    IF (workp%is1d) THEN
    
        ASSERT(pcount .LT. 4, "update_pass_count has to be < 4 in 1D, is "//I2A(pcount))

        dj = 0
        dk = 0
   
        N = nx + 2*nb ! total number of zones
        CALL self%flatten_1d(workp%wenoq0,workp%grid1d,workp%wenoqsave, q0,q,qsave) 

        CALL self%RK4_integrator_1d(q0,q,qsave)

        CALL self%copyback_1d(q0,q,qsave, workp%wenoq0,workp%grid1d,workp%wenoqsave)
 
        IF (pcount .LT. 3) THEN
            update_complete = .FALSE. ! pass another time
        ELSE 
            vmax(:) = self%compute_interface_vmax_1d(workp%grid1d)

            maxsignal(1) = MAX(maxsignal(1), vmax(1))
            maxsignal(2) = MAX(maxsignal(2), 0.0_WRD)
            maxsignal(3) = MAX(maxsignal(3), 0.0_WRD)

            update_complete = .TRUE.  ! done

        END IF

    ELSE IF (workp%is2d) THEN
        
        dj = nx + 2*nb  ! q(i+dj) = q(i,j+1)
        dk = 0
        
        N = (nx + 2*nb) * (ny + 2*nb)

        CALL self%flatten_2d(workp%wenoq0, workp%grid2d,  workp%wenoqsave, &
                             workp%wenoB0, workp%bface2d, workp%wenoBsave, workp%emf2d, &
                             q0, q, qsave, b0, b, bsave, omega)

        CALL self%RK4_integrator_2d(q0,q,qsave, b0,b,bsave, omega)
        
        CALL self%copyback_2d(q0, q, qsave, b0, b, bsave, omega, &
                              workp%wenoq0, workp%grid2d,  workp%wenoqsave, &
                              workp%wenoB0, workp%bface2d, workp%wenoBsave, workp%emf2d)
        IF (pcount .EQ. 7) THEN

            CALL self%compute_magnetic_field_divergence_2d(workp%grid2d, workp%bface2d)

            vmax(:) = self%compute_interface_vmax_2d(workp%grid2d)

            maxsignal(1) = MAX(maxsignal(1), vmax(1))
            maxsignal(2) = MAX(maxsignal(2), vmax(2))
            maxsignal(3) = MAX(maxsignal(3), 0.0_WRD)

            update_complete = .TRUE. ! done
        ELSE
            update_complete = .FALSE. ! pass another time
        END IF

    ELSE IF (workp%is3d) THEN
        
        dj = nx + 2*nb        ! q(i+dj) = q(i,j+1)
        dk = (ny + 2*nb) * dj ! q(i+dk) = q(i,j,k+1)
        
        N = (nx + 2*nb) * (ny + 2*nb) * (nz + 2*nb)
    
        IF (pcount .EQ. 0) THEN
            CALL self%check_state_3d(workp) ! Garbage in the ICs ?
        END IF

        CALL self%flatten_3d(workp%wenoq0,  workp%grid3d, workp%wenoqsave, &
                             workp%wenoB0,  workp%bface3d,workp%wenoBsave, workp%emf3d, &
                             q0,q,qsave, b0,b,bsave, omega, ishock)

        CALL self%RK4_integrator_3d(q0,q,qsave, b0,b,bsave, omega, ishock)

        CALL self%copyback_3d(q0,q,qsave, b0,b,bsave, omega, ishock, & 
                              workp%wenoq0,workp%grid3d,  workp%wenoqsave, &
                              workp%wenoB0,workp%bface3d, workp%wenoBsave, workp%emf3d)

        IF (pcount .EQ. 7) THEN
            
            CALL self%check_state_3d(workp) ! No garbage out

            Call self%compute_magnetic_field_divergence_3d(workp%grid3d,workp%bface3d)

            vmax(1:3) = self%compute_interface_vmax_3d(workp%grid3d)

            maxsignal(1) = MAX(maxsignal(1), vmax(1))
            maxsignal(2) = MAX(maxsignal(2), vmax(2))
            maxsignal(3) = MAX(maxsignal(3), vmax(3))

            update_complete = .TRUE. ! done
        ELSE
            update_complete = .FALSE. ! pass another time
        END IF

    END IF 

END SUBROUTINE solve

! 3D routines

SUBROUTINE flatten_3d(self, wenoq0, grid3d, wenoqsave, wenob0, bface3d, wenobsave, emf3d, &
                                q0,      q,     qsave,     b0,       b,     bsave, omega, ishock)
    
    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)  :: wenoq0(:,:), grid3d(:,:,:,:), wenoqsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN)  :: wenob0(:,:), bface3d(:,:,:,:), wenobsave(:,:), emf3d(:,:,:,:)

    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: q0(:,:), q(:,:), qsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: b0(:,:), b(:,:), bsave(:,:), omega(:,:)

    INTEGER(WIS), CONTIGUOUS, INTENT(OUT) :: ishock(:)

    INTEGER(WID) :: i,j,k, idx
    

    IF (MOD(pcount,2) .EQ. 0) THEN ! Always flatten running grids

        DO k = 1,nz+2*nb 
            DO j = 1,ny+2*nb 
                DO i = 1,nx+2*nb

                    idx = (k-1)*dk + (j-1)*dj + i
                  
                    q(idx,1) = grid3d(i,j,k,1)
                    q(idx,2) = grid3d(i,j,k,2)
                    q(idx,3) = grid3d(i,j,k,3)
                    q(idx,4) = grid3d(i,j,k,4)
                    q(idx,5) = grid3d(i,j,k,5)
                    q(idx,6) = grid3d(i,j,k,6)
                    q(idx,7) = grid3d(i,j,k,7)
                    q(idx,8) = grid3d(i,j,k,8)
#ifdef DEBUG
                    b(idx,:) = bface3d(i,j,k,:)
#endif
                END DO
            END DO
        END DO

        CALL self%init_S_from_E(q) ! q[9] = S(E)
        CALL self%apply_protection_floors(buf, q)

    ELSE ! CT step

        DO k = 1,nz+2*nb 
            DO j = 1,ny+2*nb 
                DO i = 1,nx+2*nb

                    idx = (k-1)*dk + (j-1)*dj + i
                  
                    q(idx,1)   = grid3d(i,j,k,1)
                    q(idx,2)   = grid3d(i,j,k,2)
                    q(idx,3)   = grid3d(i,j,k,3)
                    q(idx,4)   = grid3d(i,j,k,4)
                    q(idx,7)   = grid3d(i,j,k,7)
                    q(idx,8)   = grid3d(i,j,k,8)

                    q(idx,9)   = grid3d(i,j,k,5) ! S for ES switch
                    
                    b(idx,1)   = bface3d(i,j,k,1)
                    b(idx,2)   = bface3d(i,j,k,2)
                    b(idx,3)   = bface3d(i,j,k,3)

                    omega(idx,1) = emf3d(i,j,k,1)
                    omega(idx,2) = emf3d(i,j,k,2)
                    omega(idx,3) = emf3d(i,j,k,3)
                END DO
            END DO
        END DO

    !    q(:,5:7) = 0.0_WRD ! will be updated now from CT scheme: b on interface & omega
    END IF
 
    ! set q0 & b0 or copy it in
    
    IF (pcount .EQ. 0) THEN  ! init q0, b0
        
        DO k = 1,nz 
            DO j = 1,ny 
                DO i = 1,nx
                
                    idx = (k-1)*nx*ny + (j-1)*nx + i

                    q0(idx,1) = grid3d(i+nb,j+nb,k+nb,1) ! remove bnd zones
                    q0(idx,2) = grid3d(i+nb,j+nb,k+nb,2)
                    q0(idx,3) = grid3d(i+nb,j+nb,k+nb,3)
                    q0(idx,4) = grid3d(i+nb,j+nb,k+nb,4)
                    q0(idx,5) = grid3d(i+nb,j+nb,k+nb,5)
                    q0(idx,6) = grid3d(i+nb,j+nb,k+nb,6)
                    q0(idx,7) = grid3d(i+nb,j+nb,k+nb,7)
                    q0(idx,8) = grid3d(i+nb,j+nb,k+nb,8)
                END DO
            END DO
        END DO
    
        CALL self%apply_protection_floors(buf, q0)
        
        DO k = 1,nz+2*nb 
            DO j = 1,ny+2*nb
                DO i = 1,nx+2*nb

                    idx = (k-1)*dk + (j-1)*dj + i
                
                    b0(idx,1) = bface3d(i,j,k,1) ! keep bnd zones
                    b0(idx,2) = bface3d(i,j,k,2) 
                    b0(idx,3) = bface3d(i,j,k,3) 
                END DO
            END DO
        END DO

        ishock(:)  = 1_WIS

    ELSE ! other passes, just copy

        q0(:,1:8)  = wenoq0(:,1:8)
        ishock(:)  = INT(SIGN(1.0_WRD,q0(:,1)),WIS) ! get ES flag from density sign
        q0(:,1)    = ABS(q0(:,1)) ! remove ES flag from rho

        b0(:,:)    = wenob0(:,:)

        qsave(:,:) = wenoqsave(:,:)
        bsave(:,:) = wenobsave(:,:)
    END IF

    CALL self%init_S_from_E(q0) ! Always generate q0[9] = S(E(q0))

END SUBROUTINE flatten_3d

SUBROUTINE copyback_3d(self, q0,      q,     qsave, b0,     b,       bsave,     omega, ishock, &
                         wenoq0, grid3d, wenoqsave, wenob0, bface3d, wenobsave, emf3d )

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q0(:,:), q(:,:), qsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: b0(:,:), b(:,:), bsave(:,:), omega(:,:)

    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: wenoq0(:,:), grid3d(:,:,:,:), wenoqsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: wenob0(:,:), bface3d(:,:,:,:), wenobsave(:,:), emf3d(:,:,:,:)

    INTEGER(WIS), CONTIGUOUS, INTENT(IN) :: ishock(:)

    INTEGER(WID) :: i,j,k, idx
 
    ! save initial state if first step

    IF (pcount .EQ. 0) THEN

        wenoq0(:,:) = q0(:,1:8) 
        wenob0(:,:) = b0(:,:)

        wenoq0(:,1) = wenoq0(:,1) * ishock(:) ! include ES flag in density
    END IF

    ! always expand and copy running grids

    IF (MOD(pcount,2) .EQ. 0) THEN ! WENO step

        IF (pcount .LT. 6) THEN ! intermediate result
            
            DO k = 1, nz+2*nb 
                DO j = 1,ny+2*nb  ! q & omega with boundary zones
                    DO i = 1,nx+2*nb
                    
                        idx = (k-1)*dk + (j-1)*dj + i
            
                        grid3d(i,j,k,1) = q(idx,1)
                        grid3d(i,j,k,2) = q(idx,2)
                        grid3d(i,j,k,3) = q(idx,3)
                        grid3d(i,j,k,4) = q(idx,4)

                        grid3d(i,j,k,6) = q(idx,6)
                        grid3d(i,j,k,7) = q(idx,7)
                        grid3d(i,j,k,8) = q(idx,8)

                        grid3d(i,j,k,5) = q(idx,9) ! q[5] = S for ES switch

                        emf3d(i,j,k,1)  = omega(idx,1)
                        emf3d(i,j,k,2)  = omega(idx,2)
                        emf3d(i,j,k,3)  = omega(idx,3)

                    END DO
                END DO
            END DO
            
            wenoqsave(:,:) = qsave(:,:)

        ELSE ! final WENO step

            DO k = 1,nz ! update grid with qsave
                DO j = 1,ny 
                    DO i = 1,nx
                    
                        idx = (k-1)*nx*ny + (j-1)*nx + i
        
                        grid3d(i+nb,j+nb,k+nb,1) = qsave(idx,1)
                        grid3d(i+nb,j+nb,k+nb,2) = qsave(idx,2)
                        grid3d(i+nb,j+nb,k+nb,3) = qsave(idx,3)
                        grid3d(i+nb,j+nb,k+nb,4) = qsave(idx,4)

                        grid3d(i+nb,j+nb,k+nb,5) = qsave(idx,5) ! q[5] = S still for ES switch

                        grid3d(i+nb,j+nb,k+nb,6) = qsave(idx,6)
                        grid3d(i+nb,j+nb,k+nb,7) = qsave(idx,7)
                        grid3d(i+nb,j+nb,k+nb,8) = qsave(idx,8)

                    END DO
                END DO
            END DO
            
            DO k = 1,nz+2*nb ! pass last omega to CT scheme
                DO j = 1,ny+2*nb 
                    DO i = 1,nx+2*nb
                    
                        idx = (k-1)*dk + (j-1)*dj + i
            
                        grid3d(i,j,k,1) = grid3d(i,j,k,1)

                        emf3d(i,j,k,1)  = omega(idx,1)
                        emf3d(i,j,k,2)  = omega(idx,2)
                        emf3d(i,j,k,3)  = omega(idx,3)

                    END DO
                END DO
            END DO

        END IF

    ELSE  ! CT step, pcount = 1,3,5,7

        IF (pcount .LT. 7) THEN 

            DO k = 1,nz+2*nb
                DO j = 1,ny+2*nb ! state vector with boundaries
                    DO i = 1,nx+2*nb
                        
                        idx = (k-1)*dk + (j-1)*dj + i
    
                        grid3d(i,j,k,1) = q(idx,1)
                        grid3d(i,j,k,2) = q(idx,2)
                        grid3d(i,j,k,3) = q(idx,3)
                        grid3d(i,j,k,4) = q(idx,4)
                        grid3d(i,j,k,5) = q(idx,5)
                        grid3d(i,j,k,6) = q(idx,6)
                        grid3d(i,j,k,7) = q(idx,7)
                        grid3d(i,j,k,8) = q(idx,8)

                        bface3d(i,j,k,1)  = b(idx,1)
                        bface3d(i,j,k,2)  = b(idx,2)
                        bface3d(i,j,k,3)  = b(idx,3)

                    END DO
                END DO
            END DO

            wenobsave(:,:) = bsave(:,:)

        ELSE  ! final result of this timestep, pcount = 7

            DO k = 1,nz+2*nb
                DO j = 1,ny+2*nb
                    DO i = 1,nx+2*nb
                        
                        idx = (k-1)*dk + (j-1)*dj + i
    
                        grid3d(i,j,k,1) = q(idx,1) ! protections

                       !grid3d(i,j,k,2) = q(idx,2) ! didn't update these 3
                       !grid3d(i,j,k,3) = q(idx,3)
                       !grid3d(i,j,k,4) = q(idx,4)
                        
                        grid3d(i,j,k,5) = q(idx,5) ! CT
                        grid3d(i,j,k,6) = q(idx,6)
                        grid3d(i,j,k,7) = q(idx,7)

                        grid3d(i,j,k,8) = q(idx,8) ! ES

                        bface3d(i,j,k,1) = bsave(idx,1) ! CT
                        bface3d(i,j,k,2) = bsave(idx,2)
                        bface3d(i,j,k,3) = bsave(idx,3)

                    END DO
                END DO
            END DO
        END IF
    END IF

END SUBROUTINE copyback_3d

! Advance the running state (q,b) from (q0,emf0) by one RK4 substep, 
! accumulate result in (qsave,bsave). We do the q update on even pcounts,
! the b update on odd pcounts. 
! At the final pcount = 7, the accumulation buffers qsave and bsave are 
! written back into q, followed by another ES update and protections.

SUBROUTINE RK4_integrator_3d(self, q0,q,qsave, b0,b,bsave, omega, ishock)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: q0(:,:), b0(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:), qsave(:,:), b(:,:), bsave(:,:), omega(:,:)

    INTEGER(WIS), CONTIGUOUS, INTENT(INOUT) :: ishock(:)

    REAL(WRD) :: rkfac
    INTEGER(WID) :: i,j,k,idx

    IF (pcount .EQ. 0) THEN
        CALL self%compute_ES_switch_3d(q, ishock, buf)
    END IF

    SELECT CASE (pcount) ! RK4 factors
        CASE (0:1)
            rkfac = -0.5_WRD*dt/dx
        CASE (2:3)
            rkfac = -0.5_WRD*dt/dx
        CASE (4:5)
            rkfac = -1.0_WRD*dt/dx
        CASE (6:7)
            rkfac = -1.0_WRD/6.0_WRD*dt/dx 
    END SELECT

    ! Advance running state - alternating q for WENO, b for CT 

    IF (MOD(pcount,2) .EQ. 0) THEN  

        CALL self%update_state_vector_3d(rkfac,q0, q,omega)       ! WENO step

    ELSE           

        CALL self%update_magnetic_fields_3d(rkfac, omega,b0, b,q) ! CT step

        CALL self%apply_ES_switch_3d(buf,ishock,q)                ! ES step

        CALL self%apply_protection_floors(buf, q)
        
    END IF
 
    ! init save buffer, if first WENO or first CT step

    IF (pcount .EQ. 0) THEN
        qsave(:,1:4) = -4.0_WRD/3.0_WRD * q0(:,1:4)
        qsave(:,5)   = -4.0_WRD/3.0_WRD * q0(:,9) ! S
        qsave(:,6:8) = -4.0_WRD/3.0_WRD * q0(:,6:8)
    END IF

    IF (pcount .EQ. 1) THEN
        bsave(:,:)   = -4.0_WRD/3.0_WRD * b0(:,:)
    END IF

    ! Accumulate the result in qsave: q^n+1 = -4q0/3 + q1/3 + 2q2/3 + q3/3 + q4

    SELECT CASE (pcount) 
        CASE (0:1)
            rkfac = 1.0_WRD/3.0_WRD 
        CASE (2:3)
            rkfac = 2.0_WRD/3.0_WRD
        CASE (4:5)
            rkfac = 1.0_WRD/3.0_WRD
        CASE(6:7)
            rkfac = 1.0_WRD
    END SELECT

    IF (MOD(pcount,2) .EQ. 0) THEN

        CALL self%save_WENO_step_3d(rkfac,q, qsave)

    ELSE
        bsave(:,:) = bsave(:,:) + rkfac * b(:,:)
    END IF

    ! Final update of magnetic field from bsave in zone centers,
    ! requires another run of the ES switch and protection floors.

    IF (pcount .EQ. 7) THEN
    
        CALL self%interpolate_magnetic_field_b2q_3d(bsave, q)

        CALL self%apply_ES_switch_3d(buf,ishock,q) ! final Energy-Entropy step with new b

        CALL self%apply_protection_floors(buf, q)
        
    END IF

END SUBROUTINE RK4_integrator_3d

! Do one RK4 substep. Different from 2D, we are rotating the running state 
! vector q AND the result buffer qbuf forward twice, so we end up in
! x-direction again. Hence the updates to qbuf index a bit differently.
! That leaves us to rotate the CT fluxes forward and backward in the end.

SUBROUTINE update_state_vector_3d(self, rkfac, q0, q, omega)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: q0(:,:) 
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT)   :: omega(:,:)

    REAL(WRD), INTENT(IN) :: rkfac

    INTEGER(WIS) :: i,j,k, m, iMin, iMax, i0,i1, idx
    
    ! init q buffer with q0

    DO k = 1,nz
        DO j = 1,ny
        
            iMin = (k+nb-1)*dk + (j+nb-1)*dj + nb + 1 ! bnd zones present in q
            iMax = iMin + nx - 1

            i0 = (k-1)*nx*ny + (j-1)*nx + 1 ! bnd zones not present in q0
            i1 = i0 + nx - 1

            qbuf(iMin:iMax,1) = q0(i0:i1,1)
            qbuf(iMin:iMax,2) = q0(i0:i1,2)
            qbuf(iMin:iMax,3) = q0(i0:i1,3)
            qbuf(iMin:iMax,4) = q0(i0:i1,4)
            qbuf(iMin:iMax,5) = q0(i0:i1,5)
            qbuf(iMin:iMax,6) = q0(i0:i1,6)
            qbuf(iMin:iMax,7) = q0(i0:i1,7)
            qbuf(iMin:iMax,8) = q0(i0:i1,8)
            qbuf(iMin:iMax,9) = q0(i0:i1,9)

        END DO
    END DO

    ! x-direction

    CALL self%compute_pressure(q, P)
    CALL self%compute_eigenvalues(q,P, A)
    CALL self%compute_fluxes(q, F)
    CALL self%compute_eigenvectors(buf, q,P,F, L,R)
    CALL self%compute_WENO5_fluxes(buf, q,A,F,L,R, Fs)
    CALL self%compute_ES_fluxes(q,P, Fs)
    CALL self%compute_CT_interface_flux_3d(q,Fs, fsy,fsz)
 
    buf(:,:) = qbuf(:,:) ! save for protection fluxes

    DO i = 2,N 
       qbuf(i,1) = qbuf(i,1) + rkfac * (Fs(i,1) - Fs(i-1,1))
       qbuf(i,2) = qbuf(i,2) + rkfac * (Fs(i,2) - Fs(i-1,2))
       qbuf(i,3) = qbuf(i,3) + rkfac * (Fs(i,3) - Fs(i-1,3))
       qbuf(i,4) = qbuf(i,4) + rkfac * (Fs(i,4) - Fs(i-1,4))

       qbuf(i,6) = qbuf(i,6) + rkfac * (Fs(i,5) - Fs(i-1,5))
       qbuf(i,7) = qbuf(i,7) + rkfac * (Fs(i,6) - Fs(i-1,6))
       qbuf(i,8) = qbuf(i,8) + rkfac * (Fs(i,7) - Fs(i-1,7))

       qbuf(i,9) = qbuf(i,9) + rkfac * (Fs(i,8) - Fs(i-1,8))   ! S 
    END DO 

    CALL self%apply_protection_floors(buf, qbuf)
    
    ! y-direction

    CALL self%rotate_states_fwd_3d(q,qbuf,P,buf) ! also changes nx,ny,nz,dj,dk

    CALL self%compute_eigenvalues(q,P, A)
    CALL self%compute_fluxes(q, F)
    CALL self%compute_eigenvectors(buf, q,P,F, L,R)
    CALL self%compute_WENO5_fluxes(buf, q,A,F,L,R, Fs)
    CALL self%compute_ES_fluxes(q,P, Fs)
    CALL self%compute_CT_interface_flux_3d(q,Fs, gsz,gsx) ! inverted order of CT fluxes

    CALL self%rotate_CT_flux_bwd_3d(gsx,gsz, buf)

    buf(:,:) = qbuf(:,:) ! save for protection fluxes

    DO i = 2,N 
       qbuf(i,1) = qbuf(i,1) + rkfac * (Fs(i,1) - Fs(i-1,1))
       qbuf(i,2) = qbuf(i,2) + rkfac * (Fs(i,2) - Fs(i-1,2))
       qbuf(i,3) = qbuf(i,3) + rkfac * (Fs(i,3) - Fs(i-1,3))
       qbuf(i,4) = qbuf(i,4) + rkfac * (Fs(i,4) - Fs(i-1,4))
               
       qbuf(i,6) = qbuf(i,6) + rkfac * (Fs(i,5) - Fs(i-1,5))
       qbuf(i,7) = qbuf(i,7) + rkfac * (Fs(i,6) - Fs(i-1,6))
       qbuf(i,8) = qbuf(i,8) + rkfac * (Fs(i,7) - Fs(i-1,7))

       qbuf(i,9) = qbuf(i,9) + rkfac * (Fs(i,8) - Fs(i-1,8))   ! S 
    END DO
    
    CALL self%apply_protection_floors(buf, qbuf)

    ! z-direction

    CALL self%rotate_states_fwd_3d(q,qbuf,P,buf) ! also changes nx,ny,nz,dj,dk

    CALL self%compute_eigenvalues(q,P, A)
    CALL self%compute_fluxes(q, F)
    CALL self%compute_eigenvectors(buf, q,P,F, L,R)
    CALL self%compute_WENO5_fluxes(buf, q,A,F,L,R, Fs)
    CALL self%compute_ES_fluxes(q,P, Fs)
    CALL self%compute_CT_interface_flux_3d(q,Fs, hsx,hsy)
    
    buf(:,:) = qbuf(:,:) ! save for protection fluxes

    DO i = 2,N 
       qbuf(i,1) = qbuf(i,1) + rkfac * (Fs(i,1) - Fs(i-1,1))
       qbuf(i,2) = qbuf(i,2) + rkfac * (Fs(i,2) - Fs(i-1,2))
       qbuf(i,3) = qbuf(i,3) + rkfac * (Fs(i,3) - Fs(i-1,3))
       qbuf(i,4) = qbuf(i,4) + rkfac * (Fs(i,4) - Fs(i-1,4))
               
       qbuf(i,6) = qbuf(i,6) + rkfac * (Fs(i,5) - Fs(i-1,5))
       qbuf(i,7) = qbuf(i,7) + rkfac * (Fs(i,6) - Fs(i-1,6))
       qbuf(i,8) = qbuf(i,8) + rkfac * (Fs(i,7) - Fs(i-1,7))

       qbuf(i,9) = qbuf(i,9) + rkfac * (Fs(i,8) - Fs(i-1,8))   ! S 
    END DO
    
    CALL self%apply_protection_floors(buf, qbuf)

    ! Rotate q, hsx & hsy forward into x-direction and compute omega

    CALL self%rotate_CT_flux_fwd_3d(hsx,hsy, buf)

    CALL self%rotate_states_fwd_3d(q,qbuf,P, buf)

    q(:,:) = qbuf(:,:)

    CALL self%apply_protection_floors(buf, q)

    CALL self%compute_CT_corner_flux_3d(fsz,fsy,gsx,gsz,hsx,hsy, omega)

END SUBROUTINE update_state_vector_3d

! The CT RK4 update.

! We do not update b over the boundary, it'll violate CT.

SUBROUTINE update_magnetic_fields_3d(self, rkfac, omega,b0, b,q)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:), b(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: omega(:,:), b0(:,:)
    REAL(WRD), INTENT(IN) :: rkfac

    INTEGER(WID) ::  i,j,k,idx

    IF (self%high_order_CT) THEN

    ELSE ! R98 CT scheme

        DO k = 2,nz+2*nb-1 ! R98 eq 16 & 17
            DO j = 2,ny+2*nb-1
                DO i = 2,nx+2*nb-1

                idx = (k - 1) * dk + (j - 1) * dj + i

                b(idx,1) = b0(idx,1) + rkfac * (omega(idx,1) - omega(idx-dj,1)) &
                                     - rkfac * (omega(idx,3) - omega(idx-dk,3))

                b(idx,2) = b0(idx,2) + rkfac * (omega(idx,2) - omega(idx-dk,2)) & 
                                     - rkfac * (omega(idx,1) - omega(idx-1 ,1))

                b(idx,3) = b0(idx,3) + rkfac * (omega(idx,3) - omega(idx-1, 3)) &
                                     - rkfac * (omega(idx,2) - omega(idx-dj,2))
                END DO
            END DO
        END DO
    END IF
    
    CALL self%interpolate_magnetic_field_b2q_3d(b, q)
    
END SUBROUTINE

! get the cell centered magnetic field from the face centered
! magnetic field.

SUBROUTINE interpolate_magnetic_field_b2q_3d(self, b, q)
    
    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: b(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:)

    REAL(WRD) :: tmp(3)

    INTEGER(WIS) :: i, j, k, iMin, iMax, i0,i1

    ASSERT(SIZE(buf,1) .GE. N, "Buffer not large enough")
    ASSERT(SIZE(buf,2) .GE. 2, "Buffer not large enough")

    IF (self%high_order_CT) THEN
    

    ELSE ! R98 CT scheme

       DO i = 1+2*dk,N-dk ! R98 eq 3 & 4

           q(i,5) = (-b(i-2   ,1) + 9.0_WRD*b(i-1 ,1) + 9.0_WRD*b(i,1) - b(i+1 ,1))/16.0_WRD
           q(i,6) = (-b(i-2*dj,2) + 9.0_WRD*b(i-dj,2) + 9.0_WRD*b(i,2) - b(i+dj,2))/16.0_WRD
           q(i,7) = (-b(i-2*dk,3) + 9.0_WRD*b(i-dk,3) + 9.0_WRD*b(i,3) - b(i+dk,3))/16.0_WRD

       END DO

   END IF

END SUBROUTINE interpolate_magnetic_field_b2q_3d

! CT fluxes cannot be integrated over the boundary in our scheme.
! DivB will grow pretty quickly.

SUBROUTINE compute_CT_interface_flux_3d(self, q,Fs, fgh1, fgh2)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: Fs(:,:), q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT)   :: fgh1(:), fgh2(:)

    INTEGER(WID) :: i,j,k,idx

    IF (self%high_order_CT) THEN
 
    ELSE ! R98 CT scheme

        DO k = 1,nz+2*nb-1
            DO j = 1,ny+2*nb-1
                DO i = 1,nx+2*nb-1

                    idx = (k - 1) * dk + (j - 1) * dj + i

                    fgh1(idx) = Fs(idx,5) + 0.5_WRD*(q(idx,  5)*q(idx,  3)/q(idx,  1) &
                                                   + q(idx+1,5)*q(idx+1,3)/q(idx+1,1))
                    fgh2(idx) = Fs(idx,6) + 0.5_WRD*(q(idx,  5)*q(idx,  4)/q(idx,  1) &
                                                   + q(idx+1,5)*q(idx+1,4)/q(idx+1,1))

                END DO
            END DO
        END DO
        
    END IF
          
END SUBROUTINE compute_CT_interface_flux_3d

SUBROUTINE compute_CT_corner_flux_3d(self, fsz,fsy,gsx,gsz,hsx,hsy, omega)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: fsz(:),fsy(:),gsx(:),gsz(:),hsx(:),hsy(:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: omega(:,:)

    INTEGER(WID) :: i,j,k, idx

    IF (self%high_order_CT) THEN

    ELSE ! R98 CT scheme
        
        DO i = 1,N-dk ! J18

             omega(i,1) = 0.5_WRD * (gsx(i) + gsx(i+1 )) - 0.5_WRD * (fsy(i) + fsy(i+dj))
             omega(i,2) = 0.5_WRD * (hsy(i) + hsy(i+dj)) - 0.5_WRD * (gsz(i) + gsz(i+dk))
             omega(i,3) = 0.5_WRD * (fsz(i) + fsz(i+dk)) - 0.5_WRD * (hsx(i) + hsx(i+1 ))

        END DO    

    END IF

END SUBROUTINE compute_CT_corner_flux_3d

! Rotate the CT fluxes of the z-direction step forward into 
! the x-direction. New index is the std one: i,j,k & nx,ny,nz 
! old index was j,i,k & ny,nx,nz before rotation.

SUBROUTINE rotate_CT_flux_fwd_3d(self, hsx,hsy, buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: hsx(:), hsy(:)
    REAL(WRD), CONTIGUOUS :: buf(:,:)

    INTEGER(WID) :: i,j,k, iold,inew

    buf(:,1) = hsx(:)
    buf(:,2) = hsy(:)

    DO k = 1,nz+2*nb
        DO j = 1,ny+2*nb
            DO i = 1,nx+2*nb

                inew = (k-1) * dk + (j-1) * dj + i 
                iold = (j-1) * (nx+2*nb)*(nz+2*nb) + (i-1) * (nz+2*nb) + k

                hsx(inew) = buf(iold,1) ! forward
                hsy(inew) = buf(iold,2)

            END DO
        END DO
    END DO

END SUBROUTINE rotate_CT_flux_fwd_3d

! Transform the CT fluxes of the y-direction step backward into 
! the x-direction. Current index: i,j,k & nx,ny,nz 
! was j,k,i & ny,nz,nx before rotation.

SUBROUTINE rotate_CT_flux_bwd_3d(self, gsx,gsz, buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: gsx(:), gsz(:)
    REAL(WRD), CONTIGUOUS :: buf(:,:)

    INTEGER(WID) :: i,j,k, iold,inew

    buf(:,1) = gsx
    buf(:,2) = gsz

    DO k = 1,nz+2*nb
        DO j = 1,ny+2*nb
            DO i = 1,nx+2*nb

                inew = (k-1) * dk + (j-1) * dj + i
                iold =  (j-1)*(nz+2*nb)*(nx+2*nb) + (i-1)*(nz+2*nb) + k

                gsx(iold) = buf(inew,1)
                gsz(iold) = buf(inew,2)

            END DO
        END DO
    END DO

END SUBROUTINE rotate_CT_flux_bwd_3d

! Reorder the zones in the 1D arrays so 
!     q(idx+1) = q(i+1,j,k) -> q(idx+1) = q(i,j+1,k) ,
! which means rotating the state: 
!     q(i,j,k) -> q(j,k,i)
! We also treat the vector quantities and permutate nx,ny,nz and dj,dk so 
! everything stays consistent. In 3D this is cyclic, which makes life easier.
! Nothing here can vectorize, so we might as well do it in 3D. This routine
! is around 10% of the compute time, thus carries the bulk of memory movement.

SUBROUTINE rotate_states_fwd_3d(self, q,qbuf,P, buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:),qbuf(:,:),P(:)
    REAL(WRD), CONTIGUOUS :: buf(:,:)

    INTEGER(WID) :: i,j,k, iold,inew
    INTEGER(WID) :: tmp(3)

    ASSERT(SIZE(buf,1) .GE. N, "Rotate buffer size 1 too small")
    ASSERT(SIZE(buf,2) .GE. 8, "Rotate buffer size 2 too small")

    buf(:,:) = q(:,:)

    DO k = 1,nz+2*nb
        DO j = 1,ny+2*nb
            DO i = 1,nx+2*nb
 
                iold = (k-1) * dk + (j-1) * dj + i 
                inew = (i-1) * (ny+2*nb)*(nz+2*nb) + (k-1) * (ny+2*nb) + j

                q(inew,1) = buf(iold,1)
                q(inew,2) = buf(iold,3)
                q(inew,3) = buf(iold,4)
                q(inew,4) = buf(iold,2)
                q(inew,5) = buf(iold,6)
                q(inew,6) = buf(iold,7)
                q(inew,7) = buf(iold,5)
                q(inew,8) = buf(iold,8)
                q(inew,9) = buf(iold,9)

            END DO
        END DO
    END DO
    
    buf(:,:) = qbuf(:,:)

    DO k = 1,nz+2*nb
        DO j = 1,ny+2*nb
            DO i = 1,nx+2*nb
 
                iold = (k-1) * dk + (j-1) * dj + i 
                inew = (i-1) * (ny+2*nb)*(nz+2*nb) + (k-1) * (ny+2*nb) + j

                qbuf(inew,1) = buf(iold,1)
                qbuf(inew,2) = buf(iold,3)
                qbuf(inew,3) = buf(iold,4)
                qbuf(inew,4) = buf(iold,2)
                qbuf(inew,5) = buf(iold,6)
                qbuf(inew,6) = buf(iold,7)
                qbuf(inew,7) = buf(iold,5)
                qbuf(inew,8) = buf(iold,8)
                qbuf(inew,9) = buf(iold,9)

            END DO
        END DO
    END DO

    buf(:,1) = P(:)

    DO k = 1,nz+2*nb
        DO j = 1,ny+2*nb
            DO i = 1,nx+2*nb
 
                iold = (k-1) * dk + (j-1) * dj + i 
                inew = (i-1) * (ny+2*nb)*(nz+2*nb) + (k-1) * (ny+2*nb) + j

                P(inew) = buf(iold,1)
            END DO
        END DO
    END DO

    tmp(:) = (/ nx, ny, nz /)

    nx = tmp(2)
    ny = tmp(3)
    nz = tmp(1)

    dj = nx+2*nb             ! because we changed nx,ny,nz
    dk = (ny+2*nb)*(nx+2*nb) ! these now look as usual

END SUBROUTINE rotate_states_fwd_3d

! Accumulate q from the intermediate steps into qsave with RK factor rkfac.
! Save Entropy S into the 5th component, so we have it available at the final 
! ES step at pcount = 7.

SUBROUTINE save_WENO_step_3d(self, rkfac,q, qsave)
    
    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)  :: q(:,:)
    REAL(WRD),             INTENT(IN)  :: rkfac

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: qsave(:,:)

    INTEGER(WIS) :: k, j, iMin, iMax, i0, i1
 
    DO k = 1,nz
        DO j = 1,ny
            
            iMin = (k+nb-1)*dk + (j+nb-1)*dj + nb+1 ! bnd zones Pent
            iMax = iMin + nx - 1

            i0 = (k-1) * nx*ny + (j-1)*nx + 1       ! bnd zones not Pent
            i1 = i0 + nx - 1

            qsave(i0:i1,1) = qsave(i0:i1,1) + rkfac * q(iMin:iMax,1)
            qsave(i0:i1,2) = qsave(i0:i1,2) + rkfac * q(iMin:iMax,2)
            qsave(i0:i1,3) = qsave(i0:i1,3) + rkfac * q(iMin:iMax,3)
            qsave(i0:i1,4) = qsave(i0:i1,4) + rkfac * q(iMin:iMax,4)

            qsave(i0:i1,5) = qsave(i0:i1,5) + rkfac * q(iMin:iMax,9)  ! qsave[5] = S

            qsave(i0:i1,7) = qsave(i0:i1,7) + rkfac * q(iMin:iMax,7)
            qsave(i0:i1,8) = qsave(i0:i1,8) + rkfac * q(iMin:iMax,8)
        END DO
    END DO

END SUBROUTINE save_WENO_step_3d

SUBROUTINE  compute_magnetic_field_divergence_3d(self, q, b)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:,:,:), b(:,:,:,:)
    REAL(WRD) :: max_divB, divB, max_B2, B2
    INTEGER(WIS) :: i,j,k
    
    B2       = 0.0_WRD
    divB     = 0.0_WRD
    max_divB = 0.0_WRD
    max_B2   = 0.0_WRD

    DO k = nb,nz+nb
        DO j = nb,ny+nb
            DO i = nb,nx+nb

                divB = (b(i,j,k,1)-b(i-1,j,k,1))/dx + (b(i,j,k,2)-b(i,j-1,k,2))/dx &
                     + (b(i,j,k,3)-b(i,j,k-1,3))/dx

                B2 = q(i,j,k,5)**2 + q(i,j,k,6)**2 + q(i,j,k,7)**2

                max_divB = MAX(max_divB, divB)
                max_B2 = MAX(max_B2, B2)

            END DO
        END DO
    END DO
    
    MSG("max(divB) = "//E2A(max_divB)//", max(B^2) = "//E2A(max_B2))

    ASSERT(max_divB .LT. 1e-6, "divB = "//E2A(max_divB)//". ICs or CT broken ?")

END SUBROUTINE  compute_magnetic_field_divergence_3d


! 2D routines

! Copy and flatten current grid and bfld. I.e. the running 2D array is converted into
! a 1D array so we maximize vector length. The computationally heavy loops later 
! will then just integrate over the boundary. The RK4 buffers qsave,q0, bsave & b0 are
! flattened on the patch already, so this is a simple memcopy really.

SUBROUTINE flatten_2d(self, wenoq0, grid2d, wenoqsave, wenob0, bface2d, wenobsave, emf2d, &
                                q0,      q,     qsave,     b0,       b,     bsave,    omega)
    
    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: wenoq0(:,:), grid2d(:,:,:), wenoqsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: wenob0(:,:), bface2d(:,:,:), wenobsave(:,:), emf2d(:,:,:)

    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: q0(:,:), q(:,:), qsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: b0(:,:), b(:,:), bsave(:,:), omega(:,:)

    INTEGER(WID) :: i,j, idx
    
    ! Clean target arrays

    q(:,:)     = 0.0_WRD
    q0(:,:)    = 0.0_WRD
    qsave(:,:) = 0.0_WRD
    b(:,:)     = 0.0_WRD
    b0(:,:)    = 0.0_WRD
    bsave(:,:) = 0.0_WRD
    omega(:,:) = 0.0_WRD

    ! Always flatten running grids
    
    IF (MOD(pcount,2) .EQ. 0) THEN ! WENO step

        DO j = 1, ny + 2*nb 
            DO i = 1, nx + 2*nb

                idx = (j-1)*dj + i
                  
                q(idx,1:8) = grid2d(i,j,:)
                b(idx,:)     = bface2d(i,j,:)
                
            END DO
        END DO
    
        CALL self%init_S_from_E(q) ! q[9] = S(E)

    ELSE ! CT step

        DO j = 1, ny + 2*nb 
            DO i = 1, nx + 2*nb

                idx = (j-1)*dj + i
                  
                q(idx,1:4) = grid2d(i,j,1:4)
                q(idx,7:8) = grid2d(i,j,7:8)
                q(idx,9)   = grid2d(i,j,5)    ! evolved S in 5th component for ES switch 

                b(idx,:)     = bface2d(i,j,:)
                omega(idx,:) = emf2d(i,j,:)
            END DO
        END DO
                
        q(:,5:6) = 0.0_WRD ! update now from CT scheme & bface
    END IF
 
    ! initially set q0 & b0
    
    IF (pcount .EQ. 0) THEN
     
        DO j = 1,ny 
            DO i = 1,nx
                
                idx = (j-1)*nx + i
                    
                q0(idx,1:8) = grid2d(i+nb,j+nb,:) ! remove bnd zones

            END DO
        END DO
        
        CALL self%apply_protection_floors(buf, q0)

        DO j = 1, ny + 2*nb
            DO i = 1, nx + 2*nb

                idx = (j-1)*dj + i
                
                b0(idx,:) = bface2d(i,j,:) ! keep bnd zones
            END DO
        END DO

    ELSE ! other passes, just copy buffers
    
        q0(:,1:8) = wenoq0(:,:)
        b0(:,:)   = wenob0(:,:)

        qsave(:,:) = wenoqsave(:,:)
        bsave(:,:) = wenobsave(:,:)
    END IF

    CALL self%init_S_from_E(q0) ! Always generate q0[9] = S(E(q0))

END SUBROUTINE flatten_2d

! Move q0, q, qsave, b, b0, bsave and emf/omega onto the patch. q and b,b0 are expanded into a full
! 2D array, q0, qsave and bsave are copied directly. They don't need communication,
! thus carry no boundary zones and are just carried along.
! After a CT step, save the magnetic field on faces b and in zones q(:,5:6) onto the
! patch. If end of the stepping is reached, copy from accumulated state qsave & bsave.

SUBROUTINE copyback_2d(self, q0,      q,     qsave, b0,     b,       bsave,     omega, &
                         wenoq0, grid2d, wenoqsave, wenob0, bface2d, wenobsave, emf2d )

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q0(:,:), q(:,:), qsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: b0(:,:), b(:,:), bsave(:,:), omega(:,:)

    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: wenoq0(:,:), grid2d(:,:,:), wenoqsave(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: wenob0(:,:), bface2d(:,:,:), wenobsave(:,:), emf2d(:,:,:)

    INTEGER(WID) :: i,j,idx

    IF (pcount .EQ. 0) THEN ! save initial state
        wenoq0(:,:) = q0(:,1:8) 
        wenob0(:,:) = b0(:,:)
    END IF

    ! expand and copy running grids
    
    IF (MOD(pcount,2) .EQ. 0) THEN ! WENO step

        IF (pcount .LT. 6) THEN ! intermediate result
            
            DO j = 1,ny + 2*nb  ! q & omega with boundary zones
                DO i = 1,nx + 2*nb
                    
                    idx = (j-1)*dj + i
            
                    grid2d(i,j,1:4) = q(idx,1:4)
                    grid2d(i,j,6:8) = q(idx,6:8)
                    emf2d(i,j,:)    = omega(idx,:)
                    grid2d(i,j,5)   = q(idx,9)    ! q[9] = S for ES switch 

                END DO
            END DO
            
            wenoqsave(:,:) = qsave(:,:)

        ELSE ! final WENO step
    
            DO j = 1,ny ! update grid with qsave
                DO i = 1,nx
                    
                    idx = (j-1)*nx + i
        
                    grid2d(i+nb,j+nb,1:4) = qsave(idx,1:4)
                    grid2d(i+nb,j+nb,5)   = qsave(idx,5)  ! qsave[5] = S for ES switch
                    grid2d(i+nb,j+nb,6:8) = qsave(idx,6:8)
                END DO
            END DO
            
            DO j = 1,ny + 2*nb ! pass last omega to CT scheme
                DO i = 1,nx + 2*nb
                    
                    idx = (j-1)*dj + i
            
                    emf2d(i,j,:) = omega(idx,:)
                END DO
            END DO
        END IF

    ELSE  ! CT step

        IF (pcount .LT. 7) THEN 

            DO j = 1,ny + 2*nb ! state vector q has boundaries
                DO i = 1,nx + 2*nb
                        
                    idx = (j-1)*dj + i
    
                    grid2d(i,j,1:8) = q(idx,1:8)
                    bface2d(i,j,:)  = b(idx,:)
                END DO
            END DO

            wenobsave(:,:) = bsave(:,:)

        ELSE  ! final result
            
            DO j = 1,ny + 2*nb
                DO i = 1,nx + 2*nb

                    idx = (j-1)*dj + i
                    
                    grid2d(i,j,:)  = q(idx,1:8)
                    bface2d(i,j,:) = bsave(idx,:)
                END DO
            END DO

        END IF
    END IF

END SUBROUTINE copyback_2d

! Re-order a flattened state vector q and the pressure, 
! so that linear indexing leads to access along the y-direction of the 
! 2d grid. Uses last arg as a scratch buffer.
! Note that we also swap nx and ny ! 

SUBROUTINE transpose_state_fwd_2d(self, q,P, nx,ny,dj,buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:),P(:)
    REAL(WRD), CONTIGUOUS :: buf(:,:)

    INTEGER(WIS), INTENT(INOUT) :: nx, ny, dj
    INTEGER(WID) :: i,j,iold,inew, tmp

    ASSERT(SIZE(buf,1) .EQ. SIZE(q,1), "Transpose buffer size 1 doesnt match")
    ASSERT(SIZE(buf,2) .EQ. SIZE(q,2), "Transpose buffer size 2 doesnt match")

    buf(:,:) = q(:,:)
    
    DO j = 1,ny + 2*nb
        DO i = 1,nx + 2*nb
 
            iold = (j-1) * (nx + 2*nb) + i
            inew = (i-1) * (ny + 2*nb) + j

            q(inew,1) = buf(iold,1)
            q(inew,2) = buf(iold,3)
            q(inew,3) = buf(iold,4)
            q(inew,4) = buf(iold,2)
            q(inew,5) = buf(iold,6)
            q(inew,6) = buf(iold,7)
            q(inew,7) = buf(iold,5)
            q(inew,8) = buf(iold,8)
            q(inew,9) = buf(iold,9)

        END DO
    END DO

    buf(:,8) = P(:)

    DO j = 1,ny + 2*nb
        DO i = 1,nx + 2*nb
 
            iold = (j-1) * (nx + 2*nb) + i
            inew = (i-1) * (ny + 2*nb) + j

            P(inew) = buf(iold,8)

        END DO
    END DO

   tmp = nx     ! swap nx & ny
   nx = ny
   ny = tmp
   dj = nx+2*nb ! update dj

END SUBROUTINE transpose_state_fwd_2d

! transpose the WENO and CT fluxes back, swap nx & ny, reset dj.

SUBROUTINE transpose_fluxes_bwd_2d(self, Fs, gsx, nx,ny,dj,buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: Fs(:,:), gsx(:)
    REAL(WRD), CONTIGUOUS :: buf(:,:)

    INTEGER(WIS), INTENT(INOUT) :: nx, ny, dj
    INTEGER(WID) :: i,j, inew, iold, tmp

    ASSERT(SIZE(buf,1) .GE. SIZE(Fs,1), "Transpose buffer size 1 not N")
    ASSERT(SIZE(buf,2) .GE. SIZE(Fs,2)+1, "Transpose buffer size 2 not >= 9")

    buf(:,1:8) = Fs(:,1:8)
    buf(:,9)   = gsx(:)

    tmp = nx     ! swap back nx & ny
    nx  = ny
    ny  = tmp
    dj  = nx+2*nb ! reset dj

    DO j = 1,ny + 2*nb
        DO i = 1,nx + 2*nb
    
            inew = (j-1) * (nx + 2*nb) + i
            iold = (i-1) * (ny + 2*nb) + j

            Fs(inew,1) = buf(iold,1)
            Fs(inew,3) = buf(iold,2)
            Fs(inew,4) = buf(iold,3)
            Fs(inew,2) = buf(iold,4)
            Fs(inew,6) = buf(iold,5)
            Fs(inew,5) = buf(iold,6)
            Fs(inew,7) = buf(iold,7)
            Fs(inew,8) = buf(iold,8)

            gsx(inew) = buf(iold,9)
        END DO
    END DO
    
END SUBROUTINE transpose_fluxes_bwd_2d

! Advance the running state (q,b) from (q0,emf0) by one RK4 substep, 
! accumulate result in (qsave,bsave).

SUBROUTINE RK4_integrator_2d(self, q0,q,qsave, b0,b,bsave, omega)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: q0(:,:), b0(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:), qsave(:,:), &
                                            b(:,:), bsave(:,:), omega(:,:)
    REAL(WRD) :: rkfac
    INTEGER :: i,j,idx, idx2, iMin,iMax,i0,i1

    SELECT CASE (pcount) ! RK4 factors
        CASE (0:1)
            rkfac = -0.5_WRD*dt/dx
        CASE (2:3)
            rkfac = -0.5_WRD*dt/dx
        CASE (4:5)
            rkfac = -1.0_WRD*dt/dx
        CASE (6:7)
            rkfac = -1.0_WRD/6.0_WRD*dt/dx 
    END SELECT

    ! Advance running state - alternating WENO & CT+ES 

    IF (MOD(pcount,2) .EQ. 0) THEN  

        CALL self%update_state_vector_2d(rkfac,q0, q,omega)       ! WENO step

    ELSE                           
        CALL self%update_magnetic_fields_2d(rkfac, omega,b0, b,q) ! CT step

        CALL self%switch_ES_2d(q, buf)                            ! ES step
        
        CALL self%apply_protection_floors(buf, q)
    END IF

    ! init save buffer, if first WENO or CT step

    IF (pcount .EQ. 0) THEN
        qsave(:,1:4) = -4.0_WRD/3.0_WRD * q0(:,1:4) 
        qsave(:,5)   = -4.0_WRD/3.0_WRD * q0(:,9)     ! q[5] = S for ES switch
        qsave(:,6:8) = -4.0_WRD/3.0_WRD * q0(:,6:8) 
    END IF

    IF (pcount .EQ. 1) THEN
        bsave(:,:)   = -4.0_WRD/3.0_WRD * b0(:,:)
    END IF

    ! Accumulate the result in qsave: q^n+1 = -4q0/3 + q1/3 + 2q2/3 + q3/3 + q4

    SELECT CASE (pcount) 
        CASE (0:1)
            rkfac = 1.0_WRD/3.0_WRD 
        CASE (2:3)
            rkfac = 2.0_WRD/3.0_WRD
        CASE (4:5)
            rkfac = 1.0_WRD/3.0_WRD
        CASE(6:7)
            rkfac = 1.0_WRD
    END SELECT

    IF (MOD(pcount,2) .EQ. 0) THEN

        CALL self%save_WENO_step_2d(rkfac,q, qsave) ! WENO step
    ELSE
        bsave(:,:) = bsave(:,:) + rkfac * b(:,:)    ! CT step
    END IF

    ! Final update of state vector q, including S @ q[9]

    IF (pcount .EQ. 7) THEN
    
        CALL self%interpolate_magnetic_field_b2q_2d(bsave, q)

        CALL self%switch_ES_2d(q, buf) ! final Energy-Entropy step with new b
        
        CALL self%apply_protection_floors(buf, q)
    END IF

END SUBROUTINE RK4_integrator_2d

! Write the running state q into the save buffer qsave with factor rkfac.
! As qsave doesn't have boundaries, this is mostly an index translation.
! q(:,5:6) are treated with the CT scheme via the corner fluxes omega.
! qsave(:,5) thus contains S.

SUBROUTINE save_WENO_step_2d(self, rkfac,q, qsave)
    
    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)  :: q(:,:)
    REAL(WRD), INTENT(IN)  :: rkfac

    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: qsave(:,:)

    INTEGER(WIS) :: j, iMin, iMax, i0, i1
 
    DO j = 1,ny
            
        iMin = (j+nb-1)*dj+nb+1 ! bnd zones present
        iMax = iMin + nx - 1

        i0 = (j-1)*nx + 1       ! bnd zones not present
        i1 = i0 + nx - 1

        qsave(i0:i1,1:4) = qsave(i0:i1,1:4) + rkfac * q(iMin:iMax,1:4)
        qsave(i0:i1,5)   = qsave(i0:i1,5)   + rkfac * q(iMin:iMax,9)   ! qsave[5] = S
        qsave(i0:i1,7:8) = qsave(i0:i1,7:8) + rkfac * q(iMin:iMax,7:8)
    END DO

END SUBROUTINE save_WENO_step_2d

! Do a single RK4 substep. Return updated running state vector 
! q = q0 + rkfac * (F(q_i+1/2) - F(q_i-1/2)) and corner fluxes omega for CT step.

SUBROUTINE update_state_vector_2d(self, rkfac, q0, q, omega)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: q0(:,:) 
    REAL(WRD), CONTIGUOUS, INTENT(OUT)   :: omega(:,:)

    REAL(WRD), INTENT(IN) :: rkfac

    INTEGER(WIS) :: i,j, ii,jj,iMin, iMax, i0,i1, iidx,idx, m, B

    qbuf(:,:) = 0.0_WRD

    DO j = 1,ny ! initialize qbuf with flattened state from first substep q0
            
        iMin = (j+nb-1)*dj + nb + 1 ! bnd zones present in qbuf
        iMax = iMin + nx - 1

        i0 = (j-1)*nx + 1 ! bnd zones not present in q0
        i1 = i0 + nx - 1

        qbuf(iMin:iMax,:) = q0(i0:i1,:)
    END DO

    CALL self%apply_protection_floors(buf, q)

    CALL self%compute_pressure(q, P)
    CALL self%compute_eigenvalues(q,P, A)
    CALL self%compute_fluxes(q, F)
    CALL self%compute_eigenvectors(buf, q,P,F, L,R)

    CALL self%compute_WENO5_fluxes(buf, q,A,F,L,R, Fs)
    CALL self%compute_ES_fluxes(q,P, Fs)

    CALL self%compute_CT_interface_flux_2d("X",Fs,q,buf, fsy)

    DO i=2,N ! RK4 x-direction step
        qbuf(i,1:4) = qbuf(i,1:4) + rkfac * (Fs(i,1:4) - Fs(i-1,1:4))
        qbuf(i,6:8) = qbuf(i,6:8) + rkfac * (Fs(i,5:7) - Fs(i-1,5:7))
        qbuf(i,9)   = qbuf(i,9)   + rkfac * (Fs(i,8)   - Fs(i-1,8))   ! S 
    END DO
    
    CALL self%apply_protection_floors(buf, qbuf)

    CALL self%transpose_state_fwd_2d(q,P, nx,ny,dj,buf)

    CALL self%compute_eigenvalues(q,P, A)
    CALL self%compute_fluxes(q, F)
    CALL self%compute_eigenvectors(buf, q,P,F, L,R)

    CALL self%compute_WENO5_fluxes(buf, q,A,F,L,R, Fs)
    CALL self%compute_ES_fluxes(q,P, Fs)

    CALL self%compute_CT_interface_flux_2d("Y",Fs,q,buf, gsx)

    CALL self%transpose_fluxes_bwd_2d(Fs,gsx, nx,ny,dj,buf)

    DO i=1+dj,N-dj ! RK4 y-direction step
        q(i,1:4) = qbuf(i,1:4) + rkfac * (Fs(i,1:4) - Fs(i-dj,1:4))
        q(i,5)   = 0.0_WRD
        q(i,6:8) = qbuf(i,6:8) + rkfac * (Fs(i,5:7) - Fs(i-dj,5:7))
        q(i,9)   = qbuf(i,9)   + rkfac * (Fs(i,8)   - Fs(i-dj,8))   ! S 
    END DO
    
    CALL self%apply_protection_floors(buf, q)
 
    CALL self%compute_CT_corner_flux_2d(q, fsy,gsx,buf, omega)

ENDSUBROUTINE update_state_vector_2d

! A CT update consists of doing a RK step on the face centered magnetic
! field b and then interpolating it back onto the state vector q(:,5:6).
SUBROUTINE update_magnetic_fields_2d(self, rkfac, omega,b0, b,q)
    
    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:), b(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: omega(:,:), b0(:,:)
    REAL(WRD),             INTENT(IN)    :: rkfac

    INTEGER(WID) ::  i, j, idx

    b(:,:) = b0(:,:) ! b0 carries bnd zones because of this

    IF (self%high_order_CT) THEN ! L13 CT scheme & WENO interpolation

    ELSE ! R98 CT scheme

        DO j = -1,ny+nb-1
            DO i = -1,nx+nb-1

                idx = (j - 1 + nb) * dj + (i + nb)

                b(idx,1) = b(idx,1) + rkfac * (omega(idx,  1) - omega(idx-dj,1))
                b(idx,2) = b(idx,2) + rkfac * (omega(idx-1,1) - omega(idx,   1))
            
            END DO
        END DO
    END IF

    CALL self%interpolate_magnetic_field_b2q_2d(b, q) ! b(:,1:2) -> q(:,5:6)

END SUBROUTINE update_magnetic_fields_2d

! Interpolation from the face centered magnetic fields b(:,1:2) onto
! the zone centered fields q(:,5:6).

SUBROUTINE interpolate_magnetic_field_b2q_2d(self, b, q)
    
    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)  :: b(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: q(:,:)

    INTEGER(WIS) :: i

    q(:,5:6) = 0.0_WRD

    IF (self%high_order_CT) THEN ! L13 CT scheme & WENO interpolation

    ELSE ! R98 CT scheme

        DO i = 1+2*dj,N-dj ! R98 eq 3 & 4

            q(i,5) = (-b(i-2,1)    + 9.0_WRD*b(i-1,1)  + 9.0_WRD*b(i,1) - b(i+1,1) )/16.0_WRD
            q(i,6) = (-b(i-2*dj,2) + 9.0_WRD*b(i-dj,2) + 9.0_WRD*b(i,2) - b(i+dj,2))/16.0_WRD

        END DO

    END IF

END SUBROUTINE interpolate_magnetic_field_b2q_2d

! Compute the CT flux through the interface for direction "dir" of the 
! WENO flux "Fs". These fluxes don't like to be integrated over the
! boundary.
! For the R98 scheme, this is the transport part of 
! upwind fluxes, e.g. Toth 2000. See Balsara & Spicer 99 for the 
! flux -> E duality.

SUBROUTINE compute_CT_interface_flux_2d(self, dir,Fs,q, buf, fgh)

    CLASS(MHDWENO) :: self

    CHARACTER, INTENT(IN) :: dir

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: Fs(:,:), q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: buf(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT)   :: fgh(:)

    INTEGER(WID) :: i,j,idx

    fgh(:) = 0.0_WRD

    IF (self%high_order_CT) THEN
        
    ELSE ! R98 CT scheme
        
        IF (dir .EQ. "X") THEN 

            DO j = -2,ny+nb-1
                DO i = -2,nx+nb-1

                    idx = (j - 1 + nb) * dj + (i + nb)
    
                    fgh(idx) = Fs(idx,5) + 0.5_WRD*(q(idx,5)  *q(idx,3)  /q(idx,1) &
                                                  + q(idx+1,5)*q(idx+1,3)/q(idx+1,1))
                END DO
            END DO

        ELSE IF (dir .EQ. "Y") THEN

            DO j = -2,ny+nb-1
                DO i = -2,nx+nb-1

                    idx = (j - 1 + nb) * dj + (i + nb)

                    fgh(idx) = Fs(idx,6) + 0.5_WRD*(q(idx,5)  *q(idx,4)  /q(idx,1) &
                                                  + q(idx+1,5)*q(idx+1,4)/q(idx+1,1))
                END DO
            END DO
        ELSE 
            ASSERT(.FALSE., "Direction not supported: "//dir)
        END IF

    END IF

END SUBROUTINE compute_CT_interface_flux_2d

! Corner fluxes Omega of the CT scheme.
SUBROUTINE compute_CT_corner_flux_2d(self, q,fsy,gsx, buf, omega)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: q(:,:), fsy(:), gsx(:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: buf(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT)   :: omega(:,:)

    INTEGER(WID) :: i,j,idx

    omega(:,:) = 0.0_WRD

    IF (self%high_order_CT) THEN 
        
    ELSE  ! second order CT scheme R93

        DO i = 1,N-dj
            omega(i,1) = 0.5_WRD * (gsx(i) + gsx(i+1) - fsy(i) - fsy(i+dj)) ! R98 eq. 15
        END DO

    END IF
    
END SUBROUTINE compute_CT_corner_flux_2d


SUBROUTINE  compute_magnetic_field_divergence_2d(self, q, b)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:,:), b(:,:,:)
    REAL(WRD) :: max_divB, divB, max_B2, B2
    INTEGER(WIS) :: i,j, ii,jj

    IF (WDEBUG) THEN
        
        B2       = 0.0_WRD
        divB     = 0.0_WRD
        max_divB = 0.0_WRD
        max_B2   = 0.0_WRD
        ii = 0
        jj = 0

        DO j = nb,ny+nb
            DO i = nb,nx+nb

                divB = (b(i,j,1)-b(i-1,j,1))/dx + (b(i,j,2)-b(i,j-1,2))/dx

                B2 = q(i,j,5)**2 + q(i,j,6)**2 + q(i,j,7)**2

                IF (B2 .GT. max_B2) THEN
                    ii = i
                    jj = j
                END IF

                max_divB = MAX(max_divB, divB)
                max_B2 = MAX(max_B2, B2)

            END DO
        END DO
        
        MSG("max divB = "//E2A(max_divB)//" at "//I2A(ii)//","//I2A(jj))
        MSG("max B^2 = "//E2A(max_B2))
        MSG("MAX(B) = "//E2A(q(ii,jj,5))//","//E2A(q(ii,jj,6))//","//E2A(q(ii,jj,7)))
        MSG("MAX(bxb) = "//E2A(b(ii,jj,1))//","//E2A(b(ii,jj,2)))

        ASSERT(max_divB .LT. 1e-6, "divB > 1e-6 or not finite")
    END IF

END SUBROUTINE  compute_magnetic_field_divergence_2d

! 1D ROUTINES

! Move initial patch grid "wenoq0" and running buffer "grid1d" 
! and save buffer "wenosave" into thread local memory. 
! "wenosave" and "wenoq0" do not carry boundary zones.

SUBROUTINE flatten_1d(self, wenoq0,grid1d,wenoqsave,  q0,q,qsave)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: grid1d(:,:), wenoqsave(:,:), wenoq0(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: q0(:,:), q(:,:), qsave(:,:)
    
    SELECT CASE (pcount)
        CASE(0)
            q0(:,1:8)  = grid1d(nb+1:N-nb, :) ! no bnd zones
            q(:,1:8)   = grid1d(:,:)    ! is q0
            qsave(:,:) = 0.0_WRD
        CASE(1)
            q0(:,1:8)  = wenoq0(:,:)
            q(:,1:8)   = grid1d(:,:)    ! is q1
            qsave(:,:) = wenoqsave(:,:) ! contains -4/3 q0 + 1/3 q1
        CASE(2)
            q0(:,1:8)  = wenoq0(:,:)
            q(:,1:8)   = grid1d(:,:)    ! is q2
            qsave(:,:) = wenoqsave(:,:) ! contains -4/3 q0 + 1/3 q1 + 2/3 q2
        CASE(3)
            q0(:,1:8)  = wenoq0(:,:)
            q(:,1:8)   = grid1d(:,:)    ! is q3
            qsave(:,:) = wenoqsave(:,:) ! contains -4/3 q0 + 1/3 q1 + 2/3 q2 + 1/3 q3
        CASE DEFAULT
            ASSERT(.FALSE., "pass_count has to be within 0:3, is "//I2A(pcount))
    END SELECT

    CALL self%init_S_from_E(q)
    CALL self%init_S_from_E(q0)

END SUBROUTINE flatten_1d

! Move thread local solved patch data back into global memory arrays inside
! the patch object.

SUBROUTINE copyback_1d(self, q0, q, qsave, wenoq0, grid1d, wenoqsave)
    
    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: grid1d(:,:), wenoqsave(:,:), wenoq0(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q0(:,:), q(:,:), qsave(:,:)

    SELECT CASE (pcount)
        CASE(0)
            wenoq0(:,:)    = q0(:,1:8)  ! save initial state 
            grid1d(:,:)    = q(:,1:8)   ! running state q1, get new boundaries
            wenoqsave(:,:) = qsave(:,:) ! accumulate result here
        CASE(1)
            grid1d(:,:)    = q(:,1:8)   ! running state q2, get new boundaries
            wenoqsave(:,:) = qsave(:,:) ! accumulate result here
        CASE(2)
            grid1d(:,:)    = q(:,1:8)   ! running state q3 get new boundaries
            wenoqsave(:,:) = qsave(:,:) ! accumulate result here
        CASE(3)
            grid1d(nb+1:N-nb,:) = qsave(:,:) ! final result
            grid1d(1:nb,:)      = 0.0_WRD    ! clean boundaries
            grid1d(N-nb+1:N,:)  = 0.0_WRD 
        CASE DEFAULT
            ASSERT(.FALSE., "pass_count has to be within 0:3, is "//I2A(pcount))
    END SELECT

END SUBROUTINE

! 4th order Runge-Kutta (Method-Of-Lines) time stepping with ES_Switch.
! The scheme is formally not TVD, as mentioned in JW99. The original 
! scheme is from Shu SIAM 1073 1988. We do 2*4 (WENO,CT,WENO, ...) sweeps 
! and accumulate the result successively in qsave.

SUBROUTINE RK4_integrator_1d(self, q0, q, qsave)

    CLASS(MHDWENO) :: self
    REAL(WRD), INTENT(INOUT) :: q0(:,:), q(:,:), qsave(:,:)

    REAL(WRD) :: rkfac
    INTEGER(WID) :: i


    SELECT CASE (pcount)
        CASE (0)
            rkfac = -0.5_WRD*dt/dx ! RK4 Step 1
        CASE (1)
            rkfac = -0.5_WRD*dt/dx ! RK4 Step 2
        CASE (2)
            rkfac = -1.0_WRD*dt/dx ! RK4 Step 3
        CASE (3)
            rkfac = -1.0_WRD/6.0_WRD*dt/dx ! RK4 Step 4
        CASE DEFAULT
            ASSERT(.FALSE., "pass_count has to be within 0:3, is "//I2A(pcount))
    END SELECT

    CALL self%update_state_vector_1d(q0,rkfac, q)

    SELECT CASE (pcount) 
        CASE (0)
            rkfac = 1.0_WRD/3.0_WRD
        CASE (1)
            rkfac = 2.0_WRD/3.0_WRD
        CASE (2)
            rkfac = 1.0_WRD/3.0_WRD
        CASE(3)
            rkfac = 1.0_WRD
        CASE DEFAULT
            ASSERT(.FALSE., "pass_count has to be within 0:3, is "//I2A(pcount))
    END SELECT

    if (pcount .EQ. 0) THEN
        qsave(:,:) = -4.0_WRD/3.0_WRD * q0(:,:)
    END IF

    qsave(:,:) = qsave(:,:) + rkfac * q(nb+1:N-nb,:) ! accumulate the result in qsave

END SUBROUTINE RK4_integrator_1d

! The next function returns the new rk4 stepped state vector. 
! I.e. it returns q0 + rkfac * F(q).
! This is really where the flops are in this scheme. Everything above
! are just memory operations.

SUBROUTINE update_state_vector_1d(self, q0, rkfac, q)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q0(:,:) ! first q
    REAL(WRD), INTENT(IN) :: rkfac

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:) ! state vector for update

    INTEGER(WIS) :: i,j

    CALL self%compute_pressure(q, P)
    CALL self%compute_eigenvalues(q, P, A)
    CALL self%compute_fluxes(q, F)
    CALL self%compute_eigenvectors(buf, q,P,F, L,R)
    CALL self%compute_WENO5_fluxes(buf, q,A,F,L, R, Fs)
    CALL self%compute_ES_fluxes(q,P, Fs)

    DO i=1, nx ! RK4  step
        q(nb+i,1:4) = q0(i,1:4) + rkfac * (Fs(nb+i,1:4) - Fs(nb+i-1,1:4)) ! q0(1:nx,8)
        q(nb+i,5)   = q0(i,5)
        q(nb+i,6:8) = q0(i,6:8) + rkfac * (Fs(nb+i,5:7) - Fs(nb+i-1,5:7)) 
        q(nb+i,9)   = q0(i,9)   + rkfac * (Fs(nb+i,8)   - Fs(nb+i-1,8))   ! S 
    END DO

    CALL self%switch_ES_1d(q, buf)

END SUBROUTINE update_state_vector_1d

! common routines all dimensions

! Zone centered thermal pressure from density and entropy.
! We do not use the total energy E here: For advection
! dominated flows (large v), this would lead to the subtraction
! of two numbers of similar size, reducing numerical accuracy.
! P = (E - 0.5*(rho*v^2 + B^2)) * (gamma-1)

SUBROUTINE compute_pressure(self, q, P)
    
    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: P(:)
    INTEGER(WIS) :: i

    P(:) = 0.0_WRD

    DO i = 1,SIZE(q,1)
        P(i) = q(i,9) * q(i,1)**(self%gamma-1.0_WRD)
    END DO

END SUBROUTINE compute_pressure

! No solver guarantees positivity in 3D. As a production
! code should never fail (Pete's lemma), protection floors 
! are the last resort - they break conservation ! 

SUBROUTINE apply_protection_floors(self, buf, q)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: buf(:,:)

#ifdef USE_PROTECTIONS

    INTEGER(WIS) :: i, c

    REAL(WRD) :: mom2, B2,  gm1

    gm1 = self%gamma - 1.0_WRD

    c = 0
    
    DO i = 1,SIZE(q,1) ! Need to calculate these anyway
        
        q(i,1) = MAX( q(i,1), DENSITY_FLOOR )    

        buf(i,1) = q(i,2)**2 + q(i,3)**2 + q(i,4)**2 ! mom2
        buf(i,2) = q(i,5)**2 + q(i,6)**2 + q(i,7)**2 ! B2
        
        buf(i,3) = (q(i,8) - 0.5_WRD * (buf(i,1)/q(i,1) + buf(i,2))) * gm1 ! P
    END DO

    DO i = 1,SIZE(q,1)

        IF (buf(i,3) .LT. PRESSURE_FLOOR) THEN
    
            q(i,8) = PRESSURE_FLOOR / gm1 + 0.5_WRD * (buf(i,1)/q(i,1) + buf(i,2)) ! E(P)
            q(i,9) = PRESSURE_FLOOR / q(i,1)**gm1  ! S(E)
    
            c = c + 1
        END IF    
    END DO

    WARN(c .GT. N/4.0_WRD, "Applied protection floors in "//I2A(c)//" zones.")

#endif
    
END SUBROUTINE apply_protection_floors

! Check a 3D state vector for negative density and pressure.

SUBROUTINE check_state_3d(self, workp)

    CLASS(MHDWENO) :: self
    TYPE(Patch) :: workp
    LOGICAL      :: show
    INTEGER(WIS) :: i,j,k, xi, yi, zi
    REAL(WRD)    :: rho, P,  B2, mom2
    
    show = .FALSE.

    DO k = 1,workp%nz
        DO j = 1,workp%ny
            DO i = 1,workp%nx
                
                rho  = workp%grid3d(i,j,k,1)
                mom2 = workp%grid3d(i,j,k,2)**2 + workp%grid3d(i,j,k,3)**2 + workp%grid3d(i,j,k,4)**2
                B2   = workp%grid3d(i,j,k,5)**2 + workp%grid3d(i,j,k,6)**2 + workp%grid3d(i,j,k,7)**2
                P    = (workp%grid3d(i,j,k,8) - 0.5_WRD * ( mom2/rho + B2 ) )  * (self%gamma - 1.0_WRD)

                IF (rho .LE. 0.0_WRD) show = .TRUE.
                IF (P .LE. 0.0_WRD)   show = .TRUE.

                IF (show) THEN
                
                    xi = workp%xi0 + i
                    yi = workp%yi0 + j
                    zi = workp%zi0 + k

                    MSG("Found bad state at zone index ( "//I2A(xi)//", "//I2A(yi)//", "//I2A(zi)//" ) :")
                    MSG("   rho = "//E2A(workp%grid3d(i,j,k,1)))
                    MSG("   vx  = "//E2A(workp%grid3d(i,j,k,2)))
                    MSG("   vy  = "//E2A(workp%grid3d(i,j,k,3)))
                    MSG("   vz  = "//E2A(workp%grid3d(i,j,k,4)))
                    MSG("   Bx  = "//E2A(workp%grid3d(i,j,k,5)))
                    MSG("   By  = "//E2A(workp%grid3d(i,j,k,6)))
                    MSG("   Bz  = "//E2A(workp%grid3d(i,j,k,7)))
                    MSG("   E   = "//E2A(workp%grid3d(i,j,k,8)))
                    MSG("   P   = "//E2A(P))
                    MSG("   bx   = "//E2A(workp%bface3d(i,j,k,1)))
                    MSG("   by   = "//E2A(workp%bface3d(i,j,k,2)))
                    MSG("   bz   = "//E2A(workp%bface3d(i,j,k,3)))
                    MSG("pcount was "//I2A(pcount))

                    ASSERT(.FALSE., "Solution unphysical.")

                END IF

            END DO 
        END DO 
    END DO 

END SUBROUTINE check_state_3d

! Initialize entropy in q : q(:,9) = S(q)

SUBROUTINE init_S_from_E(self, q)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:)

    REAL(WRD) :: mom2, B2, gm1
    INTEGER(WIS) :: i

    gm1 = self%gamma - 1.0_WRD

    DO i = 1,SIZE(q,1)

        mom2 = q(i,2)**2 + q(i,3)**2 + q(i,4)**2
        B2   = q(i,5)**2 + q(i,6)**2 + q(i,7)**2

        q(i,9) = (q(i,8) - 0.5_WRD * (mom2/q(i,1) + B2))* gm1 / q(i,1)**gm1 ! S(E)  
    END DO

END SUBROUTINE init_S_from_E

! Now some conversion routines. 
! I assume that every reasonable compiler will inline those.
! Turns out ifort18 is not a reasonable compiler ...

FUNCTION S2E(self, q) RESULT(E)

    CLASS(MHDWENO) :: self

    REAL(WRD), INTENT(IN) :: q(:)
    REAL(WRD) :: E

    REAL(WRD) :: mom2, B2, gm1

    gm1 = self%gamma - 1.0_WRD

    mom2 = q(2)**2 + q(3)**2 + q(4)**2
    B2   = q(5)**2 + q(6)**2 + q(7)**2

    E = q(9)/gm1 * q(1)**gm1 + 0.5_WRD * (mom2/q(1) + B2)

END FUNCTION

FUNCTION E2S(self, q) RESULT(S)

    CLASS(MHDWENO) :: self

    REAL(WRD), INTENT(IN) :: q(:)
    REAL(WRD) :: S

    REAL(WRD) :: mom2, B2, gm1

    gm1 = self%gamma - 1.0_WRD

    mom2 = q(2)**2 + q(3)**2 + q(4)**2
    B2   = q(5)**2 + q(6)**2 + q(7)**2

    ! This minus can be an issue if q(8) is small
    ! (cold flow) and mom2 is large (fast flow)
    S = (q(8) - 0.5_WRD * (mom2/q(1) + B2))* gm1 / q(1)**gm1

END FUNCTION

FUNCTION E2P(self, q) RESULT(P)

    CLASS(MHDWENO) :: self

    REAL(WRD), INTENT(IN) :: q(:)
    REAL(WRD) :: P

    REAL(WRD) :: mom2, B2, gm1

    gm1 = self%gamma - 1.0_WRD

    mom2 = q(2)**2 + q(3)**2 + q(4)**2
    B2   = q(5)**2 + q(6)**2 + q(7)**2
    
    ! This minus can be an issue if q(8) is small
    ! (cold flow) and mom2 is large (fast flow)
    P = (q(8) - 0.5_WRD * (mom2/q(1) + B2))* gm1 

END FUNCTION

! these are the current cell centered fluxes
SUBROUTINE compute_fluxes(self, q, F)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: F(:,:)

    REAL(WRD) :: pt, vx, vB
    INTEGER(WID) :: i,j,k, ks,m,idx

!   F(:,:) = 0.0_WRD

    DO i = 1,N
 
        vx = q(i,2)/q(i,1)
        pt = P(i) + 0.5_WRD * (q(i,5)**2 + q(i,6)**2 + q(i,7)**2)       ! thermal + mag. pressure
        vB = (q(i,2)*q(i,5) + q(i,3)*q(i,6) + q(i,4)*q(i,7)) / q(i,1)   ! v dot B

        F(i,1) = q(i,2)                                                 ! rho*vx
        F(i,2) = q(i,2)*vx + pt - q(i,5)**2                             ! mvx*vx + Ptot - Bx^2
        F(i,3) = q(i,3)*vx - q(i,5)*q(i,6)                              ! mvy*vx - Bx*B 
        F(i,4) = q(i,4)*vx - q(i,5)*q(i,7)                              ! mvy*vx - Bx*B
        F(i,5) = q(i,6)*vx - q(i,5)*q(i,3)/q(i,1)                       ! By*vx - Bx*vy
        F(i,6) = q(i,7)*vx - q(i,5)*q(i,4)/q(i,1)                       ! Bz*vx - Bx*vz
        F(i,7) = (q(i,8)+pt) * vx - q(i,5)*vB                           ! (E+Ptot)*vx - Bx * v*B
    END DO

END SUBROUTINE compute_fluxes

! Scalar optimized eigenvectors for the E-system, 
! from JW99, see also Jefrey & Taniuti 1969, Brio & Wu 1988.
! These are minimal loops, every good compiler will fuse them
! depending on its internal metrics that balance loop overhead 
! and cache spilling.

SUBROUTINE compute_eigenvectors(self, buf, q, P, a, L, R)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)    :: q(:,:), a(:,:), P(:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT)   :: L(:,:,:), R(:,:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: buf(:,:)

    REAL(WRD) :: rho, dRho, Bx, By,Bz, bbn2, bnx2, root, Borth,  dl, E, pg, v2, B2
    REAL(WRD) :: gam, gam0, gam1, gam2, gam3, tmp1, tmp2, tmp3, tmp4, tmp5

    INTEGER(WIS) :: i,m

    gam = self%gamma
    gam0 = 1.0_WRD - gam
    gam1 = 0.5_WRD * (gam - 1.0_WRD)
    gam2 = (gam - 2.0_WRD)/(gam - 1.0_WRD)
    gam3 = 1.0_WRD/gam

    !L(:,:,:) = 0.0_WRD
    !R(:,:,:) = 0.0_WRD

    DO i = 1,N-1

        rho = 0.5_WRD * (q(i,1) + q(i+1,1))
        sqrt_rho(i) = SQRT(rho)
    
        vx(i) = 0.5_WRD * (q(i,2)/q(i,1) + q(i+1,2)/q(i+1,1))
        vy(i) = 0.5_WRD * (q(i,3)/q(i,1) + q(i+1,3)/q(i+1,1))
        vz(i) = 0.5_WRD * (q(i,4)/q(i,1) + q(i+1,4)/q(i+1,1))
    
        Bx = 0.5_WRD * (q(i,5) + q(i+1,5))
        By = 0.5_WRD * (q(i,6) + q(i+1,6))
        Bz = 0.5_WRD * (q(i,7) + q(i+1,7))

        pg = 0.5_WRD * (P(i) + P(i+1))
        
        v2 = vx(i)**2  + vy(i)**2 + vz(i)**2
        B2 = Bx**2 + By**2 + Bz**2

        cs(i) = SQRT( MAX(ZERO, gam * ABS(pg/rho)) )
        
        bbn2 = B2/rho
        bnx2 = Bx**2/rho

        root  = SQRT( MAX(ZERO, (bbn2 + cs(i)**2)**2 - 4.0_WRD*bnx2 * cs(i)**2) ) 

        lf(i) = SQRT( MAX(ZERO, 0.5_WRD * (bbn2 + cs(i)**2 + root)) ) ! wave speeds
        ls(i) = SQRT( MAX(ZERO, 0.5_WRD * (bbn2 + cs(i)**2 - root)) ) ! shouldn't be zero
        la(i) = SQRT( MAX(ZERO, bnx2) )
        
        signB(i) = SIGN(1.0_WRD, Bx)
    END DO

    DO i = 1,N-1 ! resolve degeneracies
        
        By = 0.5_WRD * (q(i,6) + q(i+1,6))
        Bz = 0.5_WRD * (q(i,7) + q(i+1,7))

        Borth = SQRT(By**2 + Bz**2)

        IF (Borth**2 .GT. ZERO) THEN
            bty(i) = By/Borth
            btz(i) = Bz/Borth
        ELSE
            bty(i) = INVSQRT2
            btz(i) = INVSQRT2
        END IF
    END DO

    DO i = 1,N-1

        IF (lf(i)**2 - ls(i)**2 .GE. ZERO) THEN

            ! Numerical stability: Storing intermediate results or compiler
            ! optimizations of SQRT might lead to numerical noise here.
            ! For B = 0 we explicitly set af = 0 & as = 0.

            IF (cs(i) - ls(i) .LE. 0.0_WRD) THEN
                af(i) = ZERO
            ELSE 
                af(i) = SQRT( (cs(i)**2 - ls(i)**2) / (lf(i)**2 - ls(i)**2) )
            END IF

            IF (lf(i) - cs(i) .LE. 0.0_WRD) THEN
                as(i) = ZERO
            ELSE
                as(i) = SQRT( (lf(i)**2 - cs(i)**2) / (lf(i)**2 - ls(i)**2) ) 
            END IF

        ELSE
            af(i) = 1.0_WRD
            as(i) = 1.0_WRD
        END IF

    END DO

    ! start actual ev calculation

    DO i = 1,N-1 

        buf(i,1) = 0.5_WRD/cs(i)**2

        v2 = vx(i)**2  + vy(i)**2 + vz(i)**2

        tmp2 = (bty(i) * vy(i) + btz(i) * vz(i)) * signB(i)
        tmp3 = 0.5_WRD * v2 - gam2 * cs(i)**2

        L(i,1,1) = buf(i,1) * (af(i) * (gam1 * v2 + lf(i) * vx(i)) - as(i) * ls(i) * tmp2)
        L(i,7,1) = buf(i,1) * (af(i) * (gam1 * v2 - lf(i) * vx(i)) + as(i) * ls(i) * tmp2)
        L(i,3,1) = buf(i,1) * (as(i) * (gam1 * v2 + ls(i) * vx(i)) + af(i) * lf(i) * tmp2)
        L(i,5,1) = buf(i,1) * (as(i) * (gam1 * v2 - ls(i) * vx(i)) - af(i) * lf(i) * tmp2)

        R(i,7,1) = af(i) * (lf(i)**2 - lf(i) * vx(i) + tmp3) + as(i) * ls(i) * tmp2
        R(i,7,7) = af(i) * (lf(i)**2 + lf(i) * vx(i) + tmp3) - as(i) * ls(i) * tmp2
        R(i,7,3) = as(i) * (ls(i)**2 - ls(i) * vx(i) + tmp3) - af(i) * lf(i) * tmp2
        R(i,7,5) = as(i) * (ls(i)**2 + ls(i) * vx(i) + tmp3) + af(i) * lf(i) * tmp2

        L(i,4,1) = 1.0_WRD + buf(i,1)*v2*gam0
        R(i,7,4) = 0.5_WRD*v2
    END DO

    DO i = 1,N-1 
        
        L(i,1,7) = -buf(i,1) * gam0 * af(i)
        L(i,7,7) = -buf(i,1) * gam0 * af(i)
        L(i,3,7) = -buf(i,1) * gam0 * as(i)
        L(i,5,7) = -buf(i,1) * gam0 * as(i)

        R(i,1,1) = af(i)
        R(i,1,7) = af(i)
        R(i,1,3) = as(i)
        R(i,1,5) = as(i)

    END DO

    DO i = 1,N-1 

        R(i,4,2) =  bty(i)
        R(i,4,6) =  bty(i)
        R(i,3,2) = -btz(i)
        R(i,3,6) = -btz(i)
        L(i,2,4) =  0.5_WRD * bty(i)
        L(i,6,4) =  0.5_WRD * bty(i)
        L(i,2,3) = -0.5_WRD * btz(i)
        L(i,6,3) = -0.5_WRD * btz(i)
 
        tmp2 = bty(i) * vz(i) - btz(i) * vy(i)

        R(i,7,2) = tmp2
        R(i,7,6) = tmp2
        L(i,2,1) = -0.5_WRD * tmp2
        L(i,6,1) = -0.5_WRD * tmp2

        R(i,1,4) = 1.0_WRD
    END DO

    DO i = 1,N-1 
        
        L(i,1,2) = buf(i,1) * af(i) * (gam0 * vx(i) - lf(i))
        L(i,3,2) = buf(i,1) * as(i) * (gam0 * vx(i) - ls(i))
        R(i,2,1) =            af(i) *        (vx(i) - lf(i))
        R(i,2,3) =            as(i) *        (vx(i) - ls(i))

        L(i,7,2) = buf(i,1) * af(i) * (gam0 * vx(i) + lf(i))
        R(i,2,7) =            af(i) *        (vx(i) + lf(i))
        L(i,5,2) = buf(i,1) * as(i) * (gam0 * vx(i) + ls(i))
        R(i,2,5) =            as(i) *        (vx(i) + ls(i))

        tmp2 = as(i) * ls(i) * bty(i) * signB(i)
        tmp3 = af(i) * vy(i)

        R(i,3,7) =                    tmp3 - tmp2
        R(i,3,1) =                    tmp3 + tmp2
        L(i,1,3) = buf(i,1) * (gam0 * tmp3 + tmp2)
        L(i,7,3) = buf(i,1) * (gam0 * tmp3 - tmp2)

        tmp4 = as(i) * ls(i) * btz(i) * signB(i)
        tmp5 = af(i) * vz(i)

        R(i,4,1) =                    tmp5 + tmp4
        R(i,4,7) =                    tmp5 - tmp4
        L(i,1,4) = buf(i,1) * (gam0 * tmp5 + tmp4)
        L(i,7,4) = buf(i,1) * (gam0 * tmp5 - tmp4)

    END DO

    DO i = 1,N-1 

        tmp2 = af(i) * lf(i) * bty(i) * signB(i)
        tmp3 = as(i) * vy(i)

        R(i,3,3) =                    tmp3 - tmp2
        R(i,3,5) =                    tmp3 + tmp2
        L(i,3,3) = buf(i,1) * (gam0 * tmp3 - tmp2)
        L(i,5,3) = buf(i,1) * (gam0 * tmp3 + tmp2)
        
        tmp4 = af(i) * lf(i) * btz(i) * signB(i)
        tmp5 = as(i) * vz(i)

        R(i,4,3) =                    tmp5 - tmp4 
        R(i,4,5) =                    tmp5 + tmp4 
        L(i,3,4) = buf(i,1) * (gam0 * tmp5 - tmp4)
        L(i,5,4) = buf(i,1) * (gam0 * tmp5 + tmp4)
        
    END DO

    DO i = 1,N-1 

        tmp2 = signB(i) / sqrt_rho(i)

        R(i,6,2) =  bty(i) * tmp2
        R(i,6,6) = -bty(i) * tmp2
        R(i,5,6) =  btz(i) * tmp2
        R(i,5,2) = -btz(i) * tmp2

        tmp3 = signB(i) * sqrt_rho(i)

        L(i,2,6) =  0.5_WRD * bty(i) * tmp3
        L(i,6,6) = -0.5_WRD * bty(i) * tmp3
        L(i,2,5) = -0.5_WRD * btz(i) * tmp3
        L(i,6,5) =  0.5_WRD * btz(i) * tmp3

    END DO

    DO i = 1,N-1 

        tmp2 = cs(i) * bty(i) / sqrt_rho(i)

        R(i,5,1) =  as(i) * tmp2
        R(i,5,7) =  as(i) * tmp2
        R(i,5,3) = -af(i) * tmp2
        R(i,5,5) = -af(i) * tmp2

        tmp3 = cs(i) * btz(i) / sqrt_rho(i)

        R(i,6,1) =  as(i) * tmp3
        R(i,6,7) =  as(i) * tmp3
        R(i,6,3) = -af(i) * tmp3
        R(i,6,5) = -af(i) * tmp3

    END DO

    DO i = 1,N-1 

        tmp2 = -gam0/cs(i)**2

        L(i,4,7) = -tmp2
        L(i,4,2) =  tmp2 * vx(i)
        L(i,4,3) =  tmp2 * vy(i)
        L(i,4,4) =  tmp2 * vz(i)
        
        R(i,3,4) = vy(i)
        R(i,2,4) = vx(i)
        R(i,4,4) = vz(i)
        R(i,3,4) = vy(i)
    END DO

    DO i = 1,N-1 

        Bz = 0.5_WRD * (q(i,7) + q(i+1,7))

        tmp2 = cs(i) * btz(i) * sqrt_rho(i)

        L(i,1,6) = buf(i,1) * (gam0 * af(i) * Bz + as(i) * tmp2)
        L(i,7,6) = buf(i,1) * (gam0 * af(i) * Bz + as(i) * tmp2)
        L(i,3,6) = buf(i,1) * (gam0 * as(i) * Bz - af(i) * tmp2)
        L(i,5,6) = buf(i,1) * (gam0 * as(i) * Bz - af(i) * tmp2)

        L(i,4,6) = -gam0 / cs(i)**2 * Bz

        By = 0.5_WRD * (q(i,6) + q(i+1,6))

        tmp3 = cs(i) * bty(i) * sqrt_rho(i)

        L(i,1,5) = buf(i,1) * (gam0 * af(i) * By + as(i) * tmp3)
        L(i,7,5) = buf(i,1) * (gam0 * af(i) * By + as(i) * tmp3)
        L(i,3,5) = buf(i,1) * (gam0 * as(i) * By - af(i) * tmp3)
        L(i,5,5) = buf(i,1) * (gam0 * as(i) * By - af(i) * tmp3)

        L(i,4,5) = -gam0 / cs(i)**2 * By

    END DO

    DO i = 1,N-1 

        Bx = 0.5_WRD * (q(i,5) + q(i+1,5))
        By = 0.5_WRD * (q(i,6) + q(i+1,6))
        Bz = 0.5_WRD * (q(i,7) + q(i+1,7))

        IF (By .NE. 0.0_WRD) THEN
            signB(i) = SIGN(1.0_WRD, By)
        ELSE
            signB(i) = SIGN(1.0_WRD, Bz)
        END IF
    END DO

    DO m = 1,7 ! enforce continuity
        DO i = 1, N-1
            IF (cs(i) .GE. la(i)) THEN
                L(i,3,m) = L(i,3,m) * signB(i) 
                L(i,5,m) = L(i,5,m) * signB(i) 
                R(i,m,3) = R(i,m,3) * signB(i)
                R(i,m,5) = R(i,m,5) * signB(i)
            ELSE
                L(i,1,m) = L(i,1,m) * signB(i) 
                L(i,7,m) = L(i,7,m) * signB(i) 
                R(i,m,1) = R(i,m,1) * signB(i) 
                R(i,m,7) = R(i,m,7) * signB(i)
            END IF
        END DO
    END DO

END SUBROUTINE compute_eigenvectors

! Return the WENO 5th order extrapolated fluxes at the cell interface: "Fs".
! To save variable space we use buf as a generic buffer "buf", this can be
! a bit confusing. This is the most expensive function of the solver, 
! usually 40% of CPU time.

SUBROUTINE compute_WENO5_fluxes(self, buf, q, a, F, L, R, Fs)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:), a(:,:), F(:,:), L(:,:,:), R(:,:,:) 
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: Fs(:,:)
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: buf(:,:)

    REAL(WRD) :: second, third, aterm, bterm, cterm, dterm, IS0, IS1, IS2, alpha0, tau5
    REAL(WRD) :: alpha1, alpha2, omega0, omega2, amax
    INTEGER(WIS) :: i, j,k, m, ks, iMin, iMax, idx

    iMin = 3
    iMax = N-3

    DO  m = 1,7 ! compute F^_{i+1/2}, 5th component == 0 omitted, JW96 eq. 2.17

          DO i = iMin,iMax ! calc Fsk and qsk first, reuse variables, naming confusing
             DO ks = 1,6

                buf(i,ks)  = L(i,m,1)*F(i-3+ks,1) + L(i,m,2)*F(i-3+ks,2) &  ! = Fsk
                           + L(i,m,3)*F(i-3+ks,3) + L(i,m,4)*F(i-3+ks,4) &  ! = Fsk
                           + L(i,m,5)*F(i-3+ks,5) + L(i,m,6)*F(i-3+ks,6) &  ! = Fsk
                           + L(i,m,7)*F(i-3+ks,7)

                dqsk(i,ks) = L(i,m,1)*q(i-3+ks,1) + L(i,m,2)*q(i-3+ks,2) + & ! = qsk
                             L(i,m,3)*q(i-3+ks,3) + L(i,m,4)*q(i-3+ks,4) + &
                             L(i,m,5)*q(i-3+ks,6) + L(i,m,6)*q(i-3+ks,7) + &
                             L(i,m,7)*q(i-3+ks,8)
             END DO

             ! set Fs(i,m) = first, buf = Fsk,  JW96 eq 2.11
             Fs(i,m) = (-buf(i,2) + 7.0_WRD*buf(i,3) + 7.0_WRD*buf(i,4) - buf(i,5)) / 12.0_WRD 
         END DO

        DO ks = 1,5 ! now transform buf -> dFsk & qsk -> dqsk
            DO i = 1,N
                buf(i,ks)  = buf(i,ks+1) - buf(i,ks)
                dqsk(i,ks) = dqsk(i,ks+1) - dqsk(i,ks)
            END DO
        END DO ! i

        DO i = iMin,iMax

            amax    = MAX( ABS(a(i,m)), ABS(a(i+1,m)) )

            aterm   = 0.5_WRD * (buf(i,1) + amax*dqsk(i,1)) ! Lax-Friedrichs 
            bterm   = 0.5_WRD * (buf(i,2) + amax*dqsk(i,2)) ! J&W eq. 2.10 & 2.16+
            cterm   = 0.5_WRD * (buf(i,3) + amax*dqsk(i,3))
            dterm   = 0.5_WRD * (buf(i,4) + amax*dqsk(i,4))

            IS0     = 13.0_WRD*(aterm-bterm)**2 + 3.0_WRD*(aterm-3.0_WRD*bterm)**2
            IS1     = 13.0_WRD*(bterm-cterm)**2 + 3.0_WRD*(bterm+cterm)**2
            IS2     = 13.0_WRD*(cterm-dterm)**2 + 3.0_WRD*(3.0_WRD*cterm-dterm)**2

#ifdef WENOZ
            tau5    = ABS(IS0-IS2)

            alpha0 = 1.0_WRD * (1.0_WRD + tau5/(EPSZ+IS0)**2)
            alpha1 = 6.0_WRD * (1.0_WRD + tau5/(EPSZ+IS1)**2)
            alpha2 = 3.0_WRD * (1.0_WRD + tau5/(EPSZ+IS2)**2)
#else
            alpha0  = 1.0_WRD/(EPS+IS0)**2
            alpha1  = 6.0_WRD/(EPS+IS1)**2
            alpha2  = 3.0_WRD/(EPS+IS2)**2
#endif
            omega0  = alpha0/(alpha0+alpha1+alpha2)
            omega2  = alpha2/(alpha0+alpha1+alpha2)

            second  = omega0*(aterm - 2.0_WRD*bterm + cterm)/3.0_WRD & ! phi_N(f+)
                    +(omega2 - 0.5_WRD)*(bterm - 2.0_WRD*cterm + dterm)/6.0_WRD

            Fs(i,m) = Fs(i,m) - second ! JW96 eq 2.3 + 1

        END DO ! i

        DO i = iMin,iMax

            amax   = MAX( ABS(a(i,m)), ABS(a(i+1,m)) )

            aterm  = 0.5_WRD * (buf(i,5) - amax*dqsk(i,5)) ! Lax-Friedrichs
            bterm  = 0.5_WRD * (buf(i,4) - amax*dqsk(i,4)) ! JW96 eq. 2.10 & 2.16+
            cterm  = 0.5_WRD * (buf(i,3) - amax*dqsk(i,3))  
            dterm  = 0.5_WRD * (buf(i,2) - amax*dqsk(i,2))

            IS0    = 13.0_WRD*(aterm-bterm)**2 + 3.0_WRD*(aterm - 3.0_WRD*bterm)**2
            IS1    = 13.0_WRD*(bterm-cterm)**2 + 3.0_WRD*(bterm + cterm)**2
            IS2    = 13.0_WRD*(cterm-dterm)**2 + 3.0_WRD*(3.0_WRD*cterm - dterm)**2
#ifdef WENOZ
            tau5    = ABS(IS0-IS2)

            alpha0 = 1.0_WRD * (1.0_WRD + tau5/(EPSZ+IS0)**2)
            alpha1 = 6.0_WRD * (1.0_WRD + tau5/(EPSZ+IS1)**2)
            alpha2 = 3.0_WRD * (1.0_WRD + tau5/(EPSZ+IS2)**2)
#else

            alpha0 = 1.0_WRD/(EPS + IS0)**2
            alpha1 = 6.0_WRD/(EPS + IS1)**2
            alpha2 = 3.0_WRD/(EPS + IS2)**2
#endif
            omega0 = alpha0/(alpha0 + alpha1 + alpha2)
            omega2 = alpha2/(alpha0 + alpha1 + alpha2)

            third  = omega0*(aterm - 2.0_WRD*bterm + cterm) / 3.0_WRD & ! phi_N(f-)
                   + (omega2 - 0.5_WRD)*(bterm - 2.0_WRD*cterm + dterm) / 6.0_WRD
    
            Fs(i,m) = Fs(i,m) + third ! JW96 eq. 2.16 + 1

        END DO ! i
    END DO ! m
    
    DO m = 1,7
        DO i = 1,N

            buf(i,m) = Fs(i,1)*R(i,m,1) + Fs(i,2)*R(i,m,2) + & ! JW96 eq. 2.17
                       Fs(i,3)*R(i,m,3) + Fs(i,4)*R(i,m,4) + &
                       Fs(i,5)*R(i,m,5) + Fs(i,6)*R(i,m,6) + &
                       Fs(i,7)*R(i,m,7)
        END DO
    END DO

    Fs(:,1:7) = buf(:,1:7)
    
END SUBROUTINE compute_WENO5_fluxes

! Compute the 8th component of Fs, the S flux via splitting (J18)
SUBROUTINE compute_ES_fluxes(self, q,P, Fs)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:), P(:) 
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: Fs(:,:)

    REAL(WRD) :: rho,pg, vx,vy,vz, Bx,By,Bz, v2,B2,vdB, fac

    INTEGER(WID) :: i
    
    DO i = 1,N-1

         rho = 0.5_WRD * (q(i,1) + q(i+1,1))
         vx  = 0.5_WRD * (q(i,2)/q(i,1) + q(i+1,2)/q(i+1,1))
         vy  = 0.5_WRD * (q(i,3)/q(i,1) + q(i+1,3)/q(i+1,1))
         vz  = 0.5_WRD * (q(i,4)/q(i,1) + q(i+1,4)/q(i+1,1))
         Bx  = 0.5_WRD * (q(i,5) + q(i+1,5))
         By  = 0.5_WRD * (q(i,6) + q(i+1,6))
         Bz  = 0.5_WRD * (q(i,7) + q(i+1,7))
         pg  = 0.5_WRD * (P(i)   + P(i+1)) 
        
         v2  = vx**2 + vy**2 + vz**2
         B2  = Bx**2 + By**2 + Bz**2
         vdB = vx*Bx + vy*By + vz*Bz

         fac = (1.0_WRD-self%gamma)/rho**(self%gamma-1.0_WRD)
        
         Fs(i,8) = fac * ( Fs(i,1)*(pg/rho - 0.5_WRD*v2)          &
                         + Fs(i,2)*vx + Fs(i,3)*vy + Fs(i,4)*vz &
                         + Fs(i,5)*By + Fs(i,6)*Bz - Fs(i,7) )  &
                 + fac * ( Bx*vdB - pg*vx - 0.5_WRD*B2*vx)
    END DO

END SUBROUTINE  compute_ES_Fluxes

! Replace E with E(S) to ensure adiabatic energy evolution.
! This is the J18 switch, comparing fast mode speed and velocity divergence.
! We use the available buffer space to maximize vectorization
! This is one inch away from being a shock finder. !
! The boundaries are not updated at this point!

SUBROUTINE switch_ES_1d(self, q, buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:), buf(:,:)

    REAL(WRD) :: rho, Bx,By,Bz, p, cs2,B2, bbn2,bnx2, root, &
                      vx, lfx, dvx, Max_lfx

    INTEGER(WID) :: i

    DO i = 1,N
        buf(i,2) = self%S2E(q(i,:)) ! Get Energy from Entropy
    END DO

    DO i = 1,N  ! fast mode speed and velocity

        rho = q(i,1) 
        vx  = q(i,2)/rho
        vy  = q(i,3)/rho
        vz  = q(i,4)/rho
        Bx  = q(i,5)
        By  = q(i,6)
        Bz  = q(i,7)
        p   = q(i,9)*rho**(self%gamma-1.0_WRD) ! P(S)

        cs2 = MAX(0.0_WRD, self%gamma * ABS(p/rho) )
        B2  = Bx**2 + By**2 + Bz**2

        bbn2 = B2/rho
        
        bnx2 = Bx**2/rho

        root = SQRT( MAX(0.0_WRD, (bbn2 + cs2)**2 - 4.0_WRD*bnx2*cs2) )
        lfx = SQRT( MAX(0.0_WRD, 0.5_WRD * (bbn2 + cs2 + root)))

        buf(i,3) = vx

        buf(i,6) = lfx
    END DO

    DO i = 2,N-1
    
        dvx     = ABS(buf(i-1,3) - buf(i+1,3)) ! divergence of advection velocity
        Max_lfx = MAX(buf(i-1,6), buf(i+1,6))  ! max of fast mode

        buf(i,1) = dvx - self%es_vel*Max_lfx
    END DO

    DO i = 2,N-1

        IF (buf(i,1) .LE. 0.0_WRD) THEN ! Smooth flow ?
            
            q(i,8) = buf(i,2) ! replace E with E(S)

        END IF
    END DO

END SUBROUTINE switch_ES_1d

SUBROUTINE switch_ES_2d(self, q, buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: q(:,:), buf(:,:)

    REAL(WRD) :: rho, Bx,By,Bz, p, cs2,B2, bbn2,bnx2, bny2, root, &
                      vx,vy, lfx,lfy, dvx, dvy, Max_lfx, Max_lfy
    REAL(WRD) :: mom2, gm1

    INTEGER(WID) :: i

    gm1 = self%gamma - 1.0_WRD

    DO i = 1,SIZE(q,1)  ! fast mode speed and velocity

        rho = q(i,1) 
        vx  = q(i,2)/rho
        vy  = q(i,3)/rho
        vz  = q(i,4)/rho
        Bx  = q(i,5)
        By  = q(i,6)
        Bz  = q(i,7)
        p   = q(i,9)*rho**(self%gamma-1.0_WRD) ! P(S)

        cs2 = MAX(0.0_WRD, self%gamma * ABS(p/rho) )
        B2  = Bx**2 + By**2 + Bz**2

        bbn2 = B2/rho
        
        bnx2 = Bx**2/rho
        root = SQRT( MAX(0.0_WRD, (bbn2 + cs2)**2 - 4.0_WRD*bnx2*cs2))
        lfx  = SQRT( MAX(0.0_WRD, 0.5_WRD * (bbn2 + cs2 + root)))

        bny2 = By**2/rho
        root = SQRT( MAX(0.0_WRD, (bbn2 + cs2)**2 - 4.0_WRD*bny2*cs2))
        lfy  = SQRT( MAX(0.0_WRD, 0.5_WRD * (bbn2 + cs2 + root)))
        
        buf(i,3) = vx
        buf(i,4) = vy

        buf(i,6) = lfx
        buf(i,7) = lfy
    END DO

    DO i = 1+dj,SIZE(q,1)-dj

         dvx = ABS(buf(i-1, 3) - buf(i+1, 3))
         dvy = ABS(buf(i-dj,4) - buf(i+dj,4))

         Max_lfx = MAX(buf(i-1, 6), buf(i+1, 6)) ! max of fast mode
         Max_lfy = MAX(buf(i-dj,7), buf(i+dj,7))

         buf(i,1) = ABS(dvx+dvy)/dx
         buf(i,2) = (Max_lfx + Max_lfy)/dx
    END DO

    DO i = 1+dj,SIZE(q,1)-dj
        
        mom2 = q(i,2)**2 + q(i,3)**2 + q(i,4)**2
        B2   = q(i,5)**2 + q(i,6)**2 + q(i,7)**2
        
        buf(i,3) = (q(i,8) - 0.5_WRD * (mom2/q(i,1) + B2))* gm1 / q(i,1)**gm1 ! S(E)
        buf(i,4) =  q(i,9)/gm1 * q(i,1)**gm1 + 0.5_WRD * (mom2/q(i,1) + B2) ! E(S)

    END DO

    DO i = 1+dj,SIZE(q,1)-dj

        IF (buf(i,1) .GE. self%es_vel *  buf(i,2)) THEN ! Smooth flow ?
            
            q(i,9) = buf(i,3) ! replace S with S(E)

        ELSE

            q(i,8) = buf(i,4) ! replace E with E(S)

        END IF
    END DO

END SUBROUTINE switch_ES_2d

SUBROUTINE compute_ES_switch_3d(self, q, ishock, buf)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT)  :: q(:,:), buf(:,:)
    INTEGER(WIS), CONTIGUOUS, INTENT(OUT) :: ishock(:)  

    REAL(WRD) :: rho, Bx,By,Bz, p, cs2,B2, bbn2,bnx2, bny2, bnz2, root,&
                      vx,vy,vz, lfx,lfy,lfz, dvx, dvy,dvz, Max_lfx, Max_lfy, Max_lfz

    INTEGER(WID) :: i,j,k, idx, ii

    DO i = 1,N  ! fast mode speed and velocity

        rho = q(i,1) 
        vx  = q(i,2)/q(i,1)
        vy  = q(i,3)/q(i,1)
        vz  = q(i,4)/q(i,1)
        Bx  = q(i,5)
        By  = q(i,6)
        Bz  = q(i,7)
        p   = q(i,9)*q(i,1)**(self%gamma-1.0_WRD) ! P(S)

        cs2 = MAX(0.0_WRD, self%gamma * ABS(p/rho) )
        B2  = Bx**2 + By**2 + Bz**2

        bbn2 = B2/rho
        
        bnx2 = Bx**2/rho
        bny2 = By**2/rho
        bnz2 = Bz**2/rho

        root = SQRT( MAX(0.0_WRD, (bbn2 + cs2)**2 - 4.0_WRD*bnx2*cs2))
        lfx = SQRT( MAX(0.0_WRD, 0.5_WRD * (bbn2 + cs2 + root)))

        root = SQRT( MAX(0.0_WRD, (bbn2 + cs2)**2 - 4.0_WRD*bny2*cs2))
        lfy = SQRT( MAX(0.0_WRD, 0.5_WRD * (bbn2 + cs2 + root)))
        
        root = SQRT( MAX(0.0_WRD, (bbn2 + cs2)**2 - 4.0_WRD*bnz2*cs2))
        lfz = SQRT( MAX(0.0_WRD, 0.5_WRD * (bbn2 + cs2 + root)))

        buf(i,3) = vx
        buf(i,4) = vy
        buf(i,5) = vz

        buf(i,6) = lfx
        buf(i,7) = lfy
        buf(i,8) = lfz
    END DO

    DO i = 1+dk,N-dk

        dvx = ABS(buf(i+1, 3) - buf(i-1, 3))
        dvy = ABS(buf(i+dj,4) - buf(i-dj,4))
        dvz = ABS(buf(i+dk,5) - buf(i-dk,5))

        Max_lfx = MAX(buf(i-1, 6), buf(i+1, 6)) ! max of fast mode
        Max_lfy = MAX(buf(i-dj,7), buf(i+dj,7))
        Max_lfz = MAX(buf(i-dk,8), buf(i+dk,8))

        buf(i,1) = ABS(dvx+dvy+dvz)/dx
        buf(i,2) = (Max_lfx+Max_lfy+Max_lfz)/dx
    END DO
    
    DO i = 1,nx
        DO j = 1,ny
            DO k = 1,nz

                idx = (k-1+nb)*dk + (j-1+nb)*dj + i+nb
                ii  = (k-1)*nx*ny + (j-1)*nx    + i 

                IF (buf(idx,1) .GE. self%es_vel *  buf(idx,2)) THEN ! Shock ?

                    ishock(ii) = 1 ! replace S with S(E)

                ELSE

                    ishock(ii) = -1 ! ! replace E with E(S)

                END IF
            END DO
        END DO
    END DO

END SUBROUTINE compute_ES_switch_3d

SUBROUTINE apply_ES_switch_3d(self, buf, ishock, q)

    CLASS(MHDWENO) :: self

    REAL(WRD), CONTIGUOUS, INTENT(INOUT)  :: q(:,:), buf(:,:)
    INTEGER(WIS), CONTIGUOUS, INTENT(IN) :: ishock(:)  

    REAL(WRD) :: B2, gm1, mom2

    INTEGER(WID) :: i,j,k, idx, ii

    gm1 = self%gamma - 1.0_WRD

    DO i = 1,N
        
        mom2 = q(i,2)**2 + q(i,3)**2 + q(i,4)**2
        B2   = q(i,5)**2 + q(i,6)**2 + q(i,7)**2
        
        buf(i,1) = (q(i,8) - 0.5_WRD * (mom2/q(i,1) + B2))* gm1 / q(i,1)**gm1 ! S(E)
        buf(i,2) =  q(i,9)/gm1 * q(i,1)**gm1 + 0.5_WRD * (mom2/q(i,1) + B2)   ! E(S)

    END DO

    ! Apply ES switch outside of boundaries

    DO i = 1,nx
        DO j = 1,ny
            DO k = 1,nz

                idx = (k-1+nb)*dk + (j-1+nb)*dj + i+nb
                ii  = (k-1)*nx*ny + (j-1)*nx    + i 

                IF (ishock(ii) .GT. 0) THEN
        
                    q(idx,9) = buf(idx,1) ! replace S with S(E)
            
                ELSE

                    q(idx,8) = buf(idx,2) ! replace E with E(S)
        
                END IF
            END DO
        END DO
    END DO

END SUBROUTINE apply_ES_switch_3d

! Compute the WENO5 interpolation (not the reconstruction) at index 3+1/2
! of u and put it in v. See Shu 2009, Chapter 2.1, eq. 2.6.

SUBROUTINE compute_WENO5_interpolation(self, di, u, v)

    CLASS(MHDWENO) :: self

    INTEGER(WIS), INTENT(IN) :: di

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: u(:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: v(:)

    INTEGER(WIS) :: i
    REAL(WRD) :: beta1, beta2, beta3, w1, w2, w3, p1, p2, p3, norm

    DO i = 1+2*di, N-2*di 
       
        ! third order polynomials, eq. 2.1 - 2.3
        p1 =  3.0_WRD/8.0_WRD * u(i-2*di) - 5.0_WRD/4.0_WRD * u(i-1*di) + 15.0_WRD/8.0_WRD * u(i)
        p2 = -1.0_WRD/8.0_WRD * u(i-1*di) + 3.0_WRD/4.0_WRD * u(i)      +  3.0_WRD/8.0_WRD * u(i+1*di)
        p3 =  3.0_WRD/8.0_WRD * u(i)      + 3.0_WRD/4.0_WRD * u(i+1*di) -  1.0_WRD/8.0_WRD * u(i+2*di)

        ! smoothness indicators eq. 2.9
        beta1 =  4.0_WRD * u(i-2*di)**2   - 19.0_WRD * u(i-2*di)*u(i-1*di) + 25.0_WRD * u(i-1*di)**2 & 
             + 11.0_WRD * u(i-2*di)*u(i)  - 31.0_WRD * u(i-1*di)*u(i)      + 10.0_WRD * u(i)**2

        beta2 =  4.0_WRD * u(i-1*di)**2   - 13.0_WRD * u(i-1*di)*u(i)      + 13.0_WRD * u(i)**2      &
             +  5.0_WRD * u(i-1*di)*u(i+1*di) - 13.0_WRD * u(i+1*di)*u(i)      +  4.0_WRD * u(i+1*di)**2

        beta3 = 10.0_WRD * u(i)**2        - 31.0_WRD * u(i+1*di)*u(i)      + 25.0_WRD * u(i+1*di)**2 & 
             + 11.0_WRD * u(i+2*di)*u(i)  - 19.0_WRD * u(i+1*di)*u(i+2*di) +  4.0_WRD * u(i+2*di)**2

        beta1 = beta1/3.0_WRD
        beta2 = beta2/3.0_WRD
        beta3 = beta3/3.0_WRD

       !p1 =  1.0_WRD/3.0_WRD * u(i-2*di) - 7.0_WRD/6.0_WRD * u(i-1*di) + 11.0_WRD/6.0_WRD * u(i)      ! eq. 2.11
       !p2 = -1.0_WRD/6.0_WRD * u(i-1*di) + 5.0_WRD/6.0_WRD * u(i)      +  1.0_WRD/3.0_WRD * u(i+1*di) ! eq. 2.12
       !p3 =  1.0_WRD/3.0_WRD * u(i)      + 5.0_WRD/6.0_WRD * u(i+1*di) -  1.0_WRD/6.0_WRD * u(i+2*di) ! eq. 2.13

       !beta1 = 13.0_WRD / 12.0_WRD * (u(i-2*di) -2.0_WRD*u(i-1*di) + u(i)     )**2      
       !beta2 = 13.0_WRD / 12.0_WRD * (u(i-1*di) -2.0_WRD*u(i)      + u(i+1*di))**2 
       !beta3 = 13.0_WRD / 12.0_WRD * (u(i)      -2.0_WRD*u(i+1*di) + u(i+2*di))**2 

       !beta1 = beta1 + 1.0_WRD/4.0_WRD * (u(i-2*di)    - 4.0_WRD*u(i-1*di) + 3.0_WRD*u(i))**2
       !beta2 = beta2 + 1.0_WRD/4.0_WRD * (u(i-1*di)    - u(i+1*di))**2
       !beta3 = beta3 + 1.0_WRD/4.0_WRD * (3.0_WRD*u(i) - 4.0_WRD*u(i+1*di) + u(i+2*di))**2

        ! non-linear weights eq. 2.10
        w1 = LIN_WEIGHTS(1) / (EPS + beta1)**2 
        w2 = LIN_WEIGHTS(2) / (EPS + beta2)**2 
        w3 = LIN_WEIGHTS(3) / (EPS + beta3)**2 
        
        norm = w1 + w2 + w3

        w1 = w1 / norm
        w2 = w2 / norm
        w3 = w3 / norm

        ! interpolant eq. 2.6
        v(i+di) = w1*p1 + w2*p2 + w3*p3 

    END DO

END SUBROUTINE compute_WENO5_interpolation


! Eigenvalues "a" at cell centers

SUBROUTINE compute_eigenvalues(self, q, P, a)

    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:), P(:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: a(:,:)

    INTEGER(WID) :: i
    REAL(WRD) :: rho, vx, vy, vz, Bx, By, Bz
    REAL(WRD) :: lf, ls, la, B2, cs, bnx2, bbn2, root
    
    DO i = 1,N
    
        rho = q(i,1)

        vx = q(i,2)/rho
        vy = q(i,3)/rho
        vz = q(i,4)/rho 

        Bx = q(i,5)
        By = q(i,6)
        Bz = q(i,7)

        B2 = Bx**2 + By**2 + Bz**2
        
        ! we don't want cs2 for consistency
        cs = SQRT(MAX(ZERO, self%gamma*ABS(P(i)/rho)))
        
        bbn2 = B2/rho
        bnx2 = Bx**2/rho

        root = SQRT(MAX(ZERO, (bbn2+cs**2)**2 - 4.0_WRD*bnx2*cs**2))

        lf = SQRT(MAX(ZERO, 0.5_WRD*(bbn2 + cs**2 + root) ) ) ! fast mode
        ls = SQRT(MAX(ZERO, 0.5_WRD*(bbn2 + cs**2 - root) ) ) ! slow mode
        la = SQRT(MAX(ZERO, bnx2)) ! Alfven mode

        a(i,1) = vx - lf
        a(i,2) = vx - la
        a(i,3) = vx - ls
        a(i,4) = vx
        a(i,5) = vx + ls
        a(i,6) = vx + la
        a(i,7) = vx + lf

    END DO
    
END SUBROUTINE compute_eigenvalues

! find maximal signal velocity at the interfaces for the new timestep.

FUNCTION compute_interface_vmax_1d(self, q) RESULT(vmax)

    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    REAL(WRD) :: vmax(3), vxmax

    REAL(WRD) :: gam, gm1, rho, mom2, B2, Bx, By, Bz, vx, E
    REAL(WRD) :: pg, cs2, bnx2, bny2, bnz2, bbn2, root, lf, pg1, pg2

    INTEGER(WID) :: i
    
    gam = self%gamma
    gm1 = gam-1

    vxmax = 0.0_WRD
    
    DO i = 1+nb,nb+nx -1
    
        rho = 0.5_WRD * (q(i,1) + q(i+1,1)) 
    
        vx = 0.5_WRD * (q(i,2)/q(i,1) + q(i+1,2)/q(i+1,1)) 

        Bx = 0.5_WRD * (q(i,5) + q(i+1,5))
        By = 0.5_WRD * (q(i,6) + q(i+1,6))
        Bz = 0.5_WRD * (q(i,7) + q(i+1,7))

        pg1 = self%E2P(q(i,:))
        pg2 = self%E2P(q(i+1,:))

        pg = 0.5_WRD * (pg1 + pg2)

        B2 = Bx**2 + By**2 + Bz**2

        cs2 = MAX(0.0_WRD, ABS(gam * pg/rho )) 
        
        bbn2 = B2 / rho
        
        bnx2 = Bx**2 / rho 
        
        root = SQRT(MAX(0.0_WRD,(bbn2+cs2)**2 - 4.0_WRD*bnx2*cs2))
        lf = SQRT( MAX(0.0_WRD, 0.5_WRD*(bbn2 + cs2 + root) ) ) ! x fast mode
        vxmax = MAX(vxmax, ABS(vx) + lf)

    END DO

    vmax(1) = vxmax
    vmax(2) = 0.0_WRD
    vmax(3) = 0.0_WRD

    RETURN

END FUNCTION compute_interface_vmax_1d

FUNCTION compute_interface_vmax_2d(self, q) RESULT(vmax)

    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:,:)
    REAL(WRD) :: vmax(3)

    REAL(WRD) :: gam, gm1, rho, v2, B2, Bx, By, Bz, vx, vy, pg1, pg2
    REAL(WRD) :: pg, cs2, bnx2, bny2, bnz2, bbn2, root, lf, vxmax, vymax

    INTEGER(WID) :: i,j
    
    gam = self%gamma
    gm1 = gam - 1.0_WRD

    vxmax = 0.0_WRD
    vymax = 0.0_WRD

    DO j = 1+nb, ny+nb

        DO i = 1+nb, nx+nb

            rho = 0.5_WRD * (q(i,j,1) + q(i+1,j,1)) ! always check right interface
    
            vx = 0.5_WRD * (q(i,j,2)/q(i,j,1) + q(i+1,j,2)/q(i+1,j,1)) 
            vy = 0.5_WRD * (q(i,j,3)/q(i,j,1) + q(i+1,j,3)/q(i+1,j,1))

            Bx = 0.5_WRD * (q(i,j,5) + q(i+1,j,5))
            By = 0.5_WRD * (q(i,j,6) + q(i+1,j,6))
            Bz = 0.5_WRD * (q(i,j,7) + q(i+1,j,7))

            B2 = Bx**2 + By**2 + Bz**2

            pg1 = self%E2P(q(i,j,:))
            pg2 = self%E2P(q(i+1,j,:))
            pg = 0.5_WRD * (pg1 + pg2)

            cs2 = MAX(0.0_WRD, ABS(gam * pg/rho )) 
            
            bbn2 = B2 / rho
            bnx2 = Bx**2 / rho 
            bny2 = By**2 / rho 
        
            root  = SQRT(MAX(0.0_WRD, (bbn2+cs2)**2 - 4.0_WRD*bnx2*cs2))
            lf    = SQRT(MAX(0.0_WRD, 0.5_WRD*(bbn2 + cs2 + root) ))  ! x fast mode
            vxmax = MAX(vxmax, ABS(vx) + lf)

            root  = SQRT(MAX(0.0_WRD, (bbn2+cs2)**2 - 4.0_WRD*bny2*cs2))
            lf    = SQRT(MAX(0.0_WRD, 0.5_WRD*(bbn2 + cs2 + root) ))  ! y fast mode
            vymax = MAX(vymax, ABS(vy) + lf)
    
        END DO
    END DO

    vmax(1) = vxmax
    vmax(2) = vymax
    vmax(3) = 0.0_WRD

    RETURN

END FUNCTION compute_interface_vmax_2d

FUNCTION compute_interface_vmax_3d(self, q) RESULT(vmax)

    CLASS(MHDWENO) :: self
    
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:,:,:)
    REAL(WRD) :: vmax(3)

    REAL(WRD) :: gam, gm1, rho, v2, B2, Bx, By, Bz, vx, vy, vz, E, pg1, pg2
    REAL(WRD) :: pg, cs2, bnx2, bny2, bnz2, bbn2, rootx, rooty, rootz  
    REAL(WRD) :: vxmax, vymax, vzmax,lfx, lfy, lfz

    INTEGER(WID) :: i,j,k
    
    gam = self%gamma
    gm1 = gam - 1.0_WRD

    vxmax = 0.0_WRD
    vymax = 0.0_WRD
    vzmax = 0.0_WRD

    DO k = 1+nb, nz+nb
        DO j = 1+nb, ny+nb
            DO i = 1+nb, nx+nb

                rho = 0.5_WRD * (q(i,j,k,1) + q(i+1,j,k,1)) 
    
                vx = 0.5_WRD * (q(i,j,k,2)/q(i,j,k,1) + q(i+1,j,k,2)/q(i+1,j,k,1)) 
                vy = 0.5_WRD * (q(i,j,k,3)/q(i,j,k,1) + q(i+1,j,k,3)/q(i+1,j,k,1))
                vz = 0.5_WRD * (q(i,j,k,4)/q(i,j,k,1) + q(i+1,j,k,4)/q(i+1,j,k,1))
    
                Bx = 0.5_WRD * (q(i,j,k,5) + q(i+1,j,k,5))
                By = 0.5_WRD * (q(i,j,k,6) + q(i+1,j,k,6))
                Bz = 0.5_WRD * (q(i,j,k,7) + q(i+1,j,k,7))

                B2 = Bx**2 + By**2 + Bz**2

                pg1 = self%E2P(q(i,j,k,:))
                pg2 = self%E2P(q(i+1,j,k,:))

                pg = 0.5_WRD * (pg1 + pg2)

                cs2 = MAX(0.0_WRD, ABS(gam * pg/rho )) 
            
                bbn2 = B2 / rho
        
                bnx2 = Bx**2 / rho 
                bny2 = By**2 / rho 
                bnz2 = Bz**2 / rho 
        
                rootx = SQRT((bbn2+cs2)**2 - 4.0_WRD*bnx2*cs2)
                rooty = SQRT((bbn2+cs2)**2 - 4.0_WRD*bny2*cs2)
                rootz = SQRT((bbn2+cs2)**2 - 4.0_WRD*bnz2*cs2)

                lfx = SQRT(0.5_WRD*(bbn2 + cs2 + rootx)) ! x fast mode
                lfy = SQRT(0.5_WRD*(bbn2 + cs2 + rooty)) ! y fast mode
                lfz = SQRT(0.5_WRD*(bbn2 + cs2 + rootz)) ! z fast mode

                vxmax = MAX(vxmax, ABS(vx) + lfx)
                vymax = MAX(vymax, ABS(vy) + lfy)
                vzmax = MAX(vzmax, ABS(vz) + lfz)

            END DO
        END DO
    END DO

    vmax(1) = vxmax
    vmax(2) = vymax
    vmax(3) = vzmax

    RETURN

END FUNCTION compute_interface_vmax_3d

FUNCTION signum(self, x) RESULT(s)
 
    CLASS(MHDWENO) :: self
    REAL(WRD), INTENT(IN) :: x
    REAL(WRD) :: s

    IF (x .GT. 0.0_WRD) THEN
        s = 1.0_WRD
    ELSE IF (x .LT. 0.0_WRD) THEN
        s = -1.0_WRD
    ELSE 
        s = 0.0_WRD
    END IF

END FUNCTION signum

!Print routines for 3D

SUBROUTINE PQSX(self, q, s, c) ! slice
    
    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c,s 
    INTEGER(WIS) :: i,j,k,idx

    PRINT*, "PQSX s = "//L2A(s)//" c = "//L2A(c)

     DO k=1, nz+2*nb
        WRITE(*,fmt='(1I3)',advance="no") k-nb
        DO j = 1, ny+2*nb
            idx = (k-1)*dk + (j-1)*(nx+2*nb) + s
            WRITE(*, fmt='(1f10.6)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

    PRINT*, " "
END SUBROUTINE PQSX

SUBROUTINE PQSY(self, q, s, c) ! slice
    
    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c,s 
    INTEGER(WIS) :: i,j,k,idx
    
    PRINT*, "PQSY s = "//L2A(s)//" c = "//L2A(c)

    DO i=1, nx+2*nb

        WRITE(*,fmt='(1I3)',advance="no") i-nb
    
        DO k = 1, nz+2*nb
            idx = (k-1)*dk + (s-1)*dj + i
            WRITE(*, fmt='(1f10.6)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

    PRINT*, " "
END SUBROUTINE PQSY

SUBROUTINE PQSZ(self, q, s, c) ! slice
    
    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c,s 
    INTEGER(WIS) :: i,j,k,idx

    PRINT*, "PQSZ s = "//L2A(s)//" c = "//L2A(c)//" pc = "//I2A(pcount)

    ASSERT(s .LE. nz+2*nb, "slice out of bounds")
    
     DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (s-1)*dk + (j-1)*(nx+2*nb) + i
            !WRITE(*, fmt='(1f.4)', advance="no") q(idx,c)
            WRITE(*, fmt='(1f23.16)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

    PRINT*, " "

END SUBROUTINE PQSZ

SUBROUTINE PISZ(self, q, s) ! slice
    
    CLASS(MHDWENO) :: self
 
    INTEGER(WIS), CONTIGUOUS, INTENT(IN) :: q(:)
    INTEGER(WID), INTENT(IN) :: s 
    INTEGER(WIS) :: i,j,k,idx

    PRINT*, "PISZ s = "//L2A(s)//" pc = "//I2A(pcount)

    ASSERT(s .LE. nz+2*nb, "slice out of bounds")
    
     DO i=1,nx
        WRITE(*,fmt='(1I3)',advance="no") i
        DO j = 1,ny
            idx = (s-1)*nx*ny + (j-1)*nx + i
            WRITE(*, fmt='(1I3)', advance="no") q(idx)
        END DO
        WRITE(*,*)
    END DO

    PRINT*, " "

END SUBROUTINE PISZ

SUBROUTINE PQI(self, q) ! slice
    
    CLASS(MHDWENO) :: self
 
    INTEGER(WIS), CONTIGUOUS, INTENT(IN) :: q(:)
    INTEGER(WIS) :: i,j,k,idx

    PRINT*, "PQI  pc = "//I2A(pcount)

    DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (j-1)*(nx+2*nb) + i
            WRITE(*, fmt='(1I2)', advance="no") q(idx)
        END DO
        WRITE(*,*)
    END DO

    PRINT*, " "

END SUBROUTINE PQI

SUBROUTINE PSSZ(self, q, s) ! slice
    
    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:)
    INTEGER(WID), INTENT(IN) :: s 
    INTEGER(WIS) :: i,j,k,idx

    PRINT*, "PSSZ s = "//L2A(s)//" pc = "//I2A(pcount)

    ASSERT(s .LE. nz+2*nb, "slice out of bounds")
    
     DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (s-1)*dk + (j-1)*(nx+2*nb) + i
            WRITE(*, fmt='(1f23.16)', advance="no") q(idx)
        END DO
        WRITE(*,*)
    END DO

    PRINT*, " "

END SUBROUTINE PSSZ

SUBROUTINE PQSZNB(self, q, s, c) ! slice
    
    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c,s 
    INTEGER(WIS) :: i,j,k,idx

    ASSERT(s .LE. nz, "slice out of bounds, nz = "//I2A(nz))
    
    PRINT*, "PQSZNB s = "//L2A(s)//" c = "//L2A(c)

     DO i=1, nx
        WRITE(*,fmt='(1I3)',advance="no") i
        DO j = 1, ny
            idx = (s-1)*nx*ny + (j-1)*nx + i
            WRITE(*, fmt='(1f23.16)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PQSZNB


SUBROUTINE PB3(self, b)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: b(:,:)

    INTEGER(WID) :: i,j,k,c,idx

    open (unit=20,file="../julius.txt",action="write",status="replace")

    DO k = 1,nz+2*nb
        DO j = 1,ny+2*nb
            DO i = 1,nx+2*nb
            
            idx = (k-1)*dk + (j-1)*dj + i
            
            WRITE (20,'(3I3, 3F23.16)') i,j,k, b(idx,1),b(idx,2), b(idx,3)
            
            END DO
        END DO
    END DO

    close (20)
    ASSERT(.FALSE., "STOP")

END SUBROUTINE PB3

SUBROUTINE PQ3(self, q)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)

    INTEGER(WID) :: i,j,k,c,idx

    MSG("PQ3 pcount = "//I2A(pcount))
    open (unit=20,file="../julius.txt",action="write",status="replace")

    DO c = 1,8
        DO k = 1,nz+2*nb
            DO j = 1,ny+2*nb
                DO i = 1,nx+2*nb
                
                idx = (k-1)*dk + (j-1)*dj + i
                
                WRITE (20,'(4I3, 1F23.16)') c, i,j,k, q(idx,c)
                
                END DO
            END DO
        END DO
    END DO

    close (20)

    ASSERT(.FALSE., "STOP")

END SUBROUTINE PQ3

SUBROUTINE PQ3RMBND(self, q, b)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:), b(:,:)

    INTEGER(WID) :: i,j,k,c,idx

    PRINT*, "PQ3RMBND OUTPUT TO : ", "../julius.txt"

    open (unit=20,file="../julius.txt",action="write",status="replace")

    DO c = 1,8
        DO k = 1+nb,nz+nb
            DO j = 1+nb,ny+nb
                DO i = 1+nb,nx+nb
                
                idx = (k-1)*dk + (j-1)*dj + i
                
                WRITE (20,'(4I3, 1F23.16)') c, i,j,k, q(idx,c)
                
                END DO
            END DO
        END DO
    END DO

    DO c = 1,3
        DO k = 1+nb,nz+nb
            DO j = 1+nb,ny+nb
                DO i = 1+nb,nx+nb
                
                idx = (k-1)*dk + (j-1)*dj + i
                
                WRITE (20,'(4I3, 1F23.16)') -1*c, i,j,k, b(idx,c)
            
                END DO
            END DO
        END DO
    END DO

    close (20)
    !ASSERT(.FALSE., "STOP")

END SUBROUTINE PQ3RMBND

! Print routines for 2d and flattened 2d data
SUBROUTINE PQ(self, q, c)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c
    INTEGER(WID) :: i,j,idx

    PRINT '("PQ, c = ",1I3)',c

    PRINT '(10A13)', "-2", "-1", "0", "1","2","3","4", " 4+1", " 4+2", " 4+3"
    
    DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (j-1)*(nx+2*nb) + i
            !WRITE(*, fmt='(1E11.3)', advance="no") q(idx,c)
            WRITE(*, fmt='(1F20.16)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PQ

SUBROUTINE PQRMBND(self, q, c)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c
    INTEGER(WID) :: i,j,idx

    PRINT '("PQRMBND, c = ",1I3)',c

    PRINT '(4A17)',  "1","2","3","4"
    
    DO i=nb+1, nx+nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = nb+1, ny+nb
            idx = (j-1)*(nx+2*nb) + i
            WRITE(*, fmt='(1F20.16)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PQRMBND

SUBROUTINE PQINV(self, q, c)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c
    INTEGER(WID) :: i,j,idx

    WRITE(*,fmt='(A1)',advance="no") " "
    DO j = 1, ny+2*nb
    WRITE(*, fmt='(1I9)', advance="no") j
    END DO
    WRITE(*,*)
    DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (j-1)*(nx+2*nb) + i
            WRITE(*, fmt='(1F12.6)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PQINV
SUBROUTINE PSINV(self, q)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:)
    INTEGER(WID) :: i,j,idx

    WRITE(*,fmt='(A1)',advance="no") " "
    DO j = 1, ny+2*nb
    WRITE(*, fmt='(1I8)', advance="no") j
    END DO
    WRITE(*,*)
    DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (j-1)*(nx+2*nb) + i
            WRITE(*, fmt='(1F11.6)', advance="no") q(idx)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PSINV
SUBROUTINE PQNB(self, q, c)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:)
    INTEGER(WID), INTENT(IN) :: c
    INTEGER(WID) :: i,j,idx

    PRINT '("PQNB, c = ",1I3)',c

    PRINT '(4A17)', "1","2","3","4"
    
    DO i=1, nx
        WRITE(*,fmt='(1I3)',advance="no") i
        DO j = 1, ny
            idx = (j-1)*(nx) + i
            WRITE(*, fmt='(1F25.20)', advance="no") q(idx,c)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PQNB

SUBROUTINE PQ2(self, q, c)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:,:)
    INTEGER(WID), INTENT(IN) :: c
    INTEGER(WID) :: i,j

    PRINT '("PQ2, c = ",1I3)',c

    PRINT '(10A17)', "-2", "-1", "0", "1","2","3","4", " 4+1", " 4+2", " 4+3"

    DO i = 1, SIZE(q,1)

        WRITE(*,fmt='(1I3)',advance="no") i-nb
        
        DO j = 1, SIZE(q,2)
            WRITE(*, fmt='(1F23.16)', advance="no") q(i,j,c)
        END DO
        
        WRITE(*,*)
    END DO

END SUBROUTINE PQ2

SUBROUTINE PQ2RMBND(self, q, c)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:,:)
    INTEGER(WID), INTENT(IN) :: c
    INTEGER(WID) :: i,j

    PRINT '("PQ2RMBND, c = ",1I3)',c

    PRINT '(4A17)', "1","2","3","4"

    DO i = 1+nb, nb+nx

        WRITE(*,fmt='(1I3)',advance="no") i-nb
        
        DO j = 1+nb, ny+nb
            WRITE(*, fmt='(1F26.20)', advance="no") q(i,j,c)
        END DO
        
        WRITE(*,*)
    END DO

END SUBROUTINE PQ2RMBND

SUBROUTINE PS(self, q)

    CLASS(MHDWENO) :: self
 
    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:)
    INTEGER(WID) :: i,j,idx

    PRINT '(10A11)', "-2", "-1", "0", "1","2","3","4", " 4+1", " 4+2", " 4+3"
    
    DO i=1, nx+2*nb
        WRITE(*,fmt='(1I3)',advance="no") i-nb
        DO j = 1, ny+2*nb
            idx = (j-1)*(nx+2*nb) + i
            WRITE(*, fmt='(1E11.3)', advance="no") q(idx)
            !WRITE(*, fmt='(1F23.16)', advance="no") q(idx)
        END DO
        WRITE(*,*)
    END DO

END SUBROUTINE PS

END MODULE Mod_MHDWENO
