#include "macro.h"

MODULE Mod_MultiGrid

! FILENAME: mod_MultiGrid.f90
!
! DESCRIPTION: This module provides a class for a Multi-Grid solver
!
! DEPENDENCIES: Mod_Globals for defining data types
!
! HISTORY:
!    8/12/2017  - Julius Donnert
!    3/20/2015  - Peter Mendygral

!### require the needed modules ###
USE MPI
USE Mod_Globals
USE Mod_Parameters
USE Mod_Error
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_MultiGrid_Domain
USE Mod_Domain
USE Mod_DomainSolver
USE Mod_Decomposition
USE Mod_Problem

IMPLICIT NONE

PRIVATE
PUBLIC :: MultiGrid 

TYPE :: MultiGrid 

    PRIVATE

    ! internal tuning variables
    INTEGER(WIS) :: npresmooth, npostsmooth, nmaxsmooth, nvcycles, nsmooth_block, dsolve_flavor
    REAL(WRD) :: convergence_criteria, maxResid, maxResidLocal, maxResidNext

    ! solve_class identifies which smoothing/residual solver to use 
    ! 1 = Gravity 
    INTEGER(WIS), PUBLIC :: solve_class

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: fmg_solve
    PROCEDURE, PUBLIC :: destroy

    PROCEDURE :: extract_source
    PROCEDURE :: insert_solution
    PROCEDURE :: restrict_levels
    PROCEDURE :: prolongate_levels
    PROCEDURE :: get_maxresid
    PROCEDURE :: prolongate1d
    PROCEDURE :: prolongate2d
    PROCEDURE :: prolongate3d
    PROCEDURE :: restrict1d
    PROCEDURE :: restrict2d
    PROCEDURE :: restrict3d
    PROCEDURE :: get_maxresid1d
    PROCEDURE :: get_maxresid2d
    PROCEDURE :: get_maxresid3d
    PROCEDURE :: sweep_level
    PROCEDURE :: zero_solution
    PROCEDURE :: do_vcycle

END TYPE MultiGrid

INTERFACE MultiGrid

    MODULE PROCEDURE constructor

END INTERFACE MultiGrid

CONTAINS

FUNCTION constructor(param, solve_class) RESULT(self)

    TYPE(MultiGrid) :: self
    CLASS(Parameters) :: param

    INTEGER(WIS), INTENT(IN) :: solve_class

    CALL param%section("Full Multigrid Solver", .TRUE.)

    self%convergence_criteria = param%getFlt("convergence_criteria", "1d-14", "# Maximum error tolerance for convergence",0,.TRUE.)
    self%npresmooth           = param%getInt("npresmooth_iters", "2", "# Number of smoothings going down a V cycle",0,.TRUE.)
    self%npostsmooth          = param%getInt("npostsmooth_iters",  "2", "# number of smoothings going up a V cycle",0,.TRUE.)
    self%nmaxsmooth           = param%getInt("nmaxsmooth_iters",  "100000", "# max number of smoothings  at top level",0,.TRUE.)
    self%nvcycles             = param%getInt("nvcycles", "1", "# number of V cycles", 0, .TRUE.) 
    self%nsmooth_block        = param%getInt("nsmooth_block", "4", "# number of full multigrid cycles", 0, .TRUE.) 

    self%solve_class          = solve_class

    RETURN

END FUNCTION constructor

! cleanup the object 
SUBROUTINE destroy(self)

    CLASS(MultiGrid) :: self
    ! nothing to do for now as we allocate nothing

END SUBROUTINE destroy

! perform a full multigrid solve on the given MGDomain starting from data from the given Domain 
SUBROUTINE fmg_solve(self, workd, workmgd, dsolve, rankloc, simunits, decomp, prob)

    CLASS(MultiGrid) :: self
    TYPE(Domain) :: workd
    TYPE(MGDomain) :: workmgd
    TYPE(DomainSolver) :: dsolve
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    TYPE(Decomposition) :: decomp
    TYPE(Problem) :: prob
    INTEGER(WIS) :: l, v

    ! first we must extract the source term (gravitating mass) from all Patches in the given grid domain and 
    ! place them into the root level Patches in the given MGDomain 
    CALL self%extract_source(workd, workmgd, dsolve)

    ! next restrict the source term and initial solution guess through each level
    IF (workmgd%nlevels .GT. 1) THEN

        CALL self%restrict_levels(workmgd, 0, workmgd%nlevels-2, 1, MG_SOLN)
        CALL self%restrict_levels(workmgd, 0, workmgd%nlevels-2, 3, MG_SRC)

    END IF

    ! solve the lowest resolution grid to convergence
    CALL self%sweep_level(workmgd, dsolve, rankloc, simunits, decomp, prob, workmgd%nlevels-1, self%nmaxsmooth, .TRUE., MG_SRC)

    ! now loop through each level and perform a V-cycle starting at that level to levels above at lower resolution 
    IF (workmgd%nlevels .GT. 1) THEN

        DO l = workmgd%nlevels-2, 0, -1

            ! first prolongate the solution from the next coarser grid to this grid 
            IF (l .LT. workmgd%nlevels-1) CALL self%prolongate_levels(workmgd, l+1, l+1, .FALSE.)

            ! now perform a loop of V-cycles starting from this grid
            DO v = 1, self%nvcycles

                CALL self%do_vcycle(workmgd, dsolve, rankloc, simunits, decomp, prob, l, MG_SRC)
                IF (self%maxResid .LE. self%convergence_criteria) EXIT

            END DO

        END DO

    END IF
    IF (decomp%myrank .EQ. 0 .AND. WTHREAD .EQ. 0) PRINT '("  FMG Solve :: Final error =", ES15.8,  " (converged = ", L, ")")', self%maxResid, (self%maxResid .LE. self%convergence_criteria)

    ! now place the solution back into the real domain 
    CALL self%insert_solution(workd, workmgd, dsolve)

END SUBROUTINE fmg_solve

! based on the flavor of FMG solve, extract the source term from the given Domain and place it into the root level domain 
SUBROUTINE extract_source(self, workd, workmgd, dsolve)

    CLASS(MultiGrid) :: self
    TYPE(Domain) :: workd
    TYPE(MGDomain) :: workmgd
    TYPE(DomainSolver) :: dsolve
    INTEGER(WIS) :: pi, pj, pk, p

    IF (workd%is1d) THEN

        ! gravity 
        IF (self%solve_class .EQ. 1) THEN

            !$OMP DO SCHEDULE(DYNAMIC)
            DO pi = 1, workd%npx

                IF (workd%domain1d(pi)%hasZones) THEN
                    CALL dsolve%mggrav%extract_patch(workd%domain1d(pi), workmgd%levels(0)%domain1d(pi))
                END IF

            END DO
            !$OMP END DO

        END IF

    ELSE IF (workd%is2d) THEN

        ! gravity 
        IF (self%solve_class .EQ. 1) THEN

            !$OMP DO SCHEDULE(DYNAMIC)
            DO p = 1, workd%npx*workd%npy

                ! obtain the x/y location of the Patch 
                CALL collapse2d(p, workd%npx, workd%npy, pi, pj)

                IF (workd%domain2d(pi,pj)%hasZones) THEN
                    CALL dsolve%mggrav%extract_patch(workd%domain2d(pi,pj), workmgd%levels(0)%domain2d(pi,pj))
                END IF

            END DO
            !$OMP END DO

        END IF

    ELSE IF (workd%is3d) THEN

        ! gravity 
        IF (self%solve_class .EQ. 1) THEN

            !$OMP DO SCHEDULE(DYNAMIC)
            DO p = 1, workd%npx*workd%npy*workd%npz

                ! obtain the x/y/z location of the Patch 
                CALL collapse3d(p, workd%npx, workd%npy, workd%npz, pi, pj, pk)

                IF (workd%domain3d(pi,pj,pk)%hasZones) THEN
                    CALL dsolve%mggrav%extract_patch(workd%domain3d(pi,pj,pk), workmgd%levels(0)%domain3d(pi,pj,pk))
                END IF

            END DO
            !$OMP END DO

        END IF

    END IF
    
END SUBROUTINE extract_source

! based on the flavor of FMG solve, place the solution into the actual Patch 
SUBROUTINE insert_solution(self, workd, workmgd, dsolve)

    CLASS(MultiGrid) :: self
    TYPE(Domain) :: workd
    TYPE(MGDomain) :: workmgd
    TYPE(DomainSolver) :: dsolve
    INTEGER(WIS) :: pi, pj, pk, p

    IF (workd%is1d) THEN

        ! gravity 
        IF (self%solve_class .EQ. 1) THEN

            !$OMP DO SCHEDULE(DYNAMIC)
            DO pi = 1, workd%npx

                IF (workd%domain1d(pi)%hasZones) THEN
                    CALL dsolve%mggrav%insert_patch(workd%domain1d(pi), workmgd%levels(0)%domain1d(pi))
                END IF

            END DO
            !$OMP END DO

        END IF

    ELSE IF (workd%is2d) THEN

        ! gravity 
        IF (self%solve_class .EQ. 1) THEN

            !$OMP DO SCHEDULE(DYNAMIC)
            DO p = 1, workd%npx*workd%npy

                ! obtain the x/y location of the Patch 
                CALL collapse2d(p, workd%npx, workd%npy, pi, pj)

                IF (workd%domain2d(pi,pj)%hasZones) THEN
                    CALL dsolve%mggrav%insert_patch(workd%domain2d(pi,pj), workmgd%levels(0)%domain2d(pi,pj))
                END IF

            END DO
            !$OMP END DO

        END IF

    ELSE IF (workd%is3d) THEN

        ! gravity 
        IF (self%solve_class .EQ. 1) THEN

            !$OMP DO SCHEDULE(DYNAMIC)
            DO p = 1, workd%npx*workd%npy*workd%npz

                ! obtain the x/y/z location of the Patch 
                CALL collapse3d(p, workd%npx, workd%npy, workd%npz, pi, pj, pk)

                IF (workd%domain3d(pi,pj,pk)%hasZones) THEN
                    CALL dsolve%mggrav%insert_patch(workd%domain3d(pi,pj,pk), workmgd%levels(0)%domain3d(pi,pj,pk))
                END IF

            END DO
            !$OMP END DO

        END IF

    END IF
    
END SUBROUTINE insert_solution

! restrict through the given range of levels
SUBROUTINE restrict_levels(self, workmgd, llow, lhi, lv, rv)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: llow, lhi, lv, rv
    INTEGER(WIS) :: l

    ! loop through the given range of levels and route to the correct restriction 
    DO l = llow, lhi

        IF (workmgd%ndim .EQ. 1) THEN

            CALL self%restrict1d(workmgd, l, lv, rv)

        ELSE IF (workmgd%ndim .EQ. 2) THEN

            CALL self%restrict2d(workmgd, l, lv, rv)

        ELSE IF (workmgd%ndim .EQ. 3) THEN

            CALL self%restrict3d(workmgd, l, lv)

        END IF

    END DO

END SUBROUTINE restrict_levels

! restriction in 1d from the given level l to l+1 ###
! this version does a simple averaging across all zones covered by the coarse zone
SUBROUTINE restrict1d(self, workmgd, l, lv, rv)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l, lv, rv
    INTEGER(WIS) :: pi, pci, pbase, pend, pnx, pnxc, i, ir, ileft, ilow, ihi
    REAL(WRD) :: cavg, icoarse

    pnx     = workmgd%levels(l)%domain1d(1)%nx
    pnxc    = workmgd%levels(l+1)%domain1d(1)%nx
    icoarse = 1.0_WRD / REAL(workmgd%coarse_grid(l), WRD)

    !$OMP DO SCHEDULE(DYNAMIC)
    DO pci = 1, workmgd%levels(l+1)%npx

        ! loop over the each coarse level zone and average up all fine level zones it covers 
        DO i = 1, pnxc

            cavg  = 0.0_WRD
            pbase = 1 + FLOOR(REAL(workmgd%coarse_grid(l) * ((pci-1)*pnxc + i - 1),WRD) / REAL(pnx,WRD))
            pend  = CEILING(REAL(workmgd%coarse_grid(l) * ((pci-1)*pnxc + i),WRD) / REAL(pnx,WRD))
            ileft = workmgd%coarse_grid(l)

            ! loop over each fine level Patch and average the zones we need 
            DO pi = pbase, pend

                ilow  = ((pci-1) * pnxc + i) * workmgd%coarse_grid(l) - (pi - 1) * pnx - ileft + 1
                ihi   = MIN(pnx, ilow + ileft - 1)
                ileft = ileft - (ihi - ilow + 1)

                DO ir = ilow, ihi

                    cavg = cavg + workmgd%levels(l)%domain1d(pi)%grid1d(ir,rv)

                END DO

            END DO

            workmgd%levels(l+1)%domain1d(pci)%grid1d(i,lv) = cavg * icoarse

        END DO

    END DO
    !$OMP END DO

END SUBROUTINE restrict1d

! restriction in 2d from the given level l to l+1
! this version does a simple averaging across all zones covered by the coarse zone
SUBROUTINE restrict2d(self, workmgd, l, lv, rv)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l, lv, rv
    INTEGER(WIS) :: p, pi, pj, pci, pcj, pibase, piend, pjbase, pjend, pnx, pnxc, pny, pnyc, i, &
         j, ir, jr, ileft, jleft, ilow, ihi, jlow, jhi
    REAL(WRD) :: cavg, icoarse

    pnx     = workmgd%levels(l)%domain2d(1,1)%nx
    pnxc    = workmgd%levels(l+1)%domain2d(1,1)%nx
    pny     = workmgd%levels(l)%domain2d(1,1)%ny
    pnyc    = workmgd%levels(l+1)%domain2d(1,1)%ny
    icoarse = 1.0_WRD / REAL(workmgd%coarse_grid(l)**2, WRD)

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workmgd%levels(l+1)%npx * workmgd%levels(l+1)%npy

        CALL collapse2d(p, workmgd%levels(l+1)%npx, workmgd%levels(l+1)%npy, pci, pcj)

        ! loop over the each coarse level zone and average up all fine level zones it covers 
        DO j = 1, pnyc

            pjbase = 1 + FLOOR(REAL(workmgd%coarse_grid(l) * ((pcj-1)*pnyc + j - 1),WRD) / REAL(pny,WRD))
            pjend  = CEILING(REAL(workmgd%coarse_grid(l) * ((pcj-1)*pnyc + j),WRD) / REAL(pny,WRD))

            DO i = 1, pnxc

                cavg   = 0.0_WRD
                pibase = 1 + FLOOR(REAL(workmgd%coarse_grid(l) * ((pci-1)*pnxc + i - 1),WRD) / REAL(pnx,WRD))
                piend  = CEILING(REAL(workmgd%coarse_grid(l) * ((pci-1)*pnxc + i),WRD) / REAL(pnx,WRD))
                ileft  = workmgd%coarse_grid(l)
                jleft  = workmgd%coarse_grid(l)

                ! loop over each fine level Patch and average the zones we need 
                DO pj = pjbase, pjend

                    DO pi = pibase, piend

                        jlow  = ((pcj-1) * pnyc + j) * workmgd%coarse_grid(l) - (pj - 1) * pny - jleft + 1
                        jhi   = MIN(pny, jlow + jleft - 1)
                        jleft = jleft - (jhi - jlow + 1)
                        DO jr = jlow, jhi

                            ilow  = ((pci-1) * pnxc + i) * workmgd%coarse_grid(l) - (pi - 1) * pnx - ileft + 1
                            ihi   = MIN(pnx, ilow + ileft - 1)
                            ileft = ileft - (ihi - ilow + 1)
                            DO ir = ilow, ihi

                                cavg = cavg + workmgd%levels(l)%domain2d(pi,pj)%grid2d(ir,jr,rv)

                            END DO

                        END DO

                    END DO

                END DO

                workmgd%levels(l+1)%domain2d(pci,pcj)%grid2d(i,j,lv) = cavg * icoarse

            END DO

        END DO

    END DO
    !$OMP END DO

END SUBROUTINE restrict2d

! restriction in 3d from the given level l to l+1 
! this version does a simple averaging across all zones covered by the coarse zone 
SUBROUTINE restrict3d(self, workmgd, l, v)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l, v
    INTEGER(WIS) :: p, pi, pj, pk, pci, pcj, pck, pibase, piend, pjbase, pjend, pkbase, pkend, &
         pnx, pnxc, pny, pnyc, pnz, pnzc, i, j, k, &
         ir, jr, kr, ileft, jleft, kleft, ilow, ihi, jlow, jhi, klow, khi
    REAL(WRD) :: cavg, icoarse

    pnx     = workmgd%levels(l)%domain1d(1)%nx
    pnxc    = workmgd%levels(l+1)%domain1d(1)%nx
    pny     = workmgd%levels(l)%domain1d(1)%ny
    pnyc    = workmgd%levels(l+1)%domain1d(1)%ny
    pnz     = workmgd%levels(l)%domain1d(1)%nz
    pnzc    = workmgd%levels(l+1)%domain1d(1)%nz
    icoarse = 1.0_WRD / REAL(workmgd%coarse_grid(l)**2, WRD)

    ! this looks nasty with so many nested loops, but this should not be a walltime consumer, and it is simplest that 
    ! this averaging be done in a single routine 

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workmgd%levels(l+1)%npx * workmgd%levels(l+1)%npy * workmgd%levels(l+1)%npz

        CALL collapse3d(p, workmgd%levels(l+1)%npx, workmgd%levels(l+1)%npy, workmgd%levels(l+1)%npz, pci, pcj, pck)

        ! loop over the each coarse level zone and average up all fine level zones it covers 
        DO k = 1, pnzc

            pkbase = FLOOR(REAL(workmgd%coarse_grid(l) * ((pck-1)*pnzc + k),WRD) / REAL(workmgd%levels(l)%npz,WRD))
            pkend  = FLOOR(REAL(workmgd%coarse_grid(l) * ((pck-1)*pnzc + k + 1),WRD) / REAL(workmgd%levels(l)%npz,WRD))

            DO j = 1, pnyc

                pjbase = FLOOR(REAL(workmgd%coarse_grid(l) * ((pcj-1)*pnyc + j),WRD) / REAL(workmgd%levels(l)%npy,WRD))
                pjend  = FLOOR(REAL(workmgd%coarse_grid(l) * ((pcj-1)*pnyc + j + 1),WRD) / REAL(workmgd%levels(l)%npy,WRD))

                DO i = 1, pnxc

                    cavg   = 0.0_WRD
                    pibase = FLOOR(REAL(workmgd%coarse_grid(l) * ((pci-1)*pnxc + i),WRD) / REAL(workmgd%levels(l)%npx,WRD))
                    piend  = FLOOR(REAL(workmgd%coarse_grid(l) * ((pci-1)*pnxc + i + 1),WRD) / REAL(workmgd%levels(l)%npx,WRD))
                    ileft  = workmgd%coarse_grid(l)
                    jleft  = workmgd%coarse_grid(l)
                    kleft  = workmgd%coarse_grid(l)

                    ! loop over each fine level Patch and average the zones we need 
                    DO pk = pkbase, pkend

                        DO pj = pjbase, pjend

                            DO pi = pibase, piend

                                klow  = ((pck-1) * pnzc + j) * workmgd%coarse_grid(l) - (pk - 1) * pnz - kleft + 1
                                khi   = MIN(pnz, klow + kleft - 1)
                                kleft = kleft - (khi - klow + 1)
                                DO kr = klow, khi

                                    jlow  = ((pcj-1) * pnyc + j) * workmgd%coarse_grid(l) - (pj - 1) * pny - jleft + 1
                                    jhi   = MIN(pny, jlow + jleft - 1)
                                    jleft = jleft - (jhi - jlow + 1)
                                    DO jr = jlow, jhi

                                        ilow  = ((pci-1) * pnxc + i) * workmgd%coarse_grid(l) - (pi - 1) * pnx - ileft + 1
                                        ihi   = MIN(pnx, ilow + ileft - 1)
                                        ileft = ileft - (ihi - ilow + 1)
                                        DO ir = ilow, ihi

                                            cavg = cavg + workmgd%levels(l)%domain3d(pi,pj,pk)%grid3d(ir,jr,kr,v)

                                        END DO

                                    END DO

                                END DO

                            END DO

                        END DO

                    END DO

                    workmgd%levels(l+1)%domain3d(pci,pcj,pck)%grid3d(i,j,k,v) = cavg * icoarse

                END DO

            END DO

        END DO

    END DO
    !$OMP END DO

END SUBROUTINE restrict3d

! prolongate through the given range of levels 
SUBROUTINE prolongate_levels(self, workmgd, llow, lhi, withAdd)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: llow, lhi
    LOGICAL(WIS), INTENT(IN) :: withAdd
    INTEGER(WIS) :: l

    ! loop through the given range of levels and route to the correct prolongation 
    DO l = llow, lhi

        IF (workmgd%ndim .EQ. 1) THEN

            CALL self%prolongate1d(workmgd, l, withAdd)

        ELSE IF (workmgd%ndim .EQ. 2) THEN

            CALL self%prolongate2d(workmgd, l, withAdd)

        ELSE IF (workmgd%ndim .EQ. 3) THEN

            CALL self%prolongate3d(workmgd, l, withAdd)

        END IF

    END DO

END SUBROUTINE prolongate_levels

! prolongation in 1d from the given level l to l-1 
! this version does a simple injection from the coarse zone to all covered fine zones 
SUBROUTINE prolongate1d(self, workmgd, l, withAdd)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    LOGICAL(WIS), INTENT(IN) :: withAdd
    INTEGER(WIS) :: pnx, pnxc, pi, i, pci, ip, ip_l, ip_r
    REAL(WRD) :: addOrig

    pnx     = workmgd%levels(l-1)%domain1d(1)%nx
    pnxc    = workmgd%levels(l)%domain1d(1)%nx
    addOrig = 0.0_WRD
    IF (withAdd) addOrig = 1.0_WRD

! TODO, I BELIEVE THERE IS A BUG BELOW, CHECK THE ZONES THAT GET USED AGAIN

    !$OMP DO SCHEDULE(DYNAMIC)
    DO pi = 1, workmgd%levels(l-1)%npx

        ! loop over the each fine level zone and inject the overlapping coarse zone value into it 
        DO i = 1, pnx

            pci   = CEILING(REAL((pi-1) * pnx + i, WRD) / REAL(workmgd%coarse_grid(l-1) * pnxc, WRD))
            
            ip    = CEILING(REAL((pi-1) * pnx + i, WRD) / REAL(workmgd%coarse_grid(l-1), WRD)) - (pci-1) * pnxc
            ip_l  = CEILING(REAL((pi-1) * pnx + i - 1, WRD) / REAL(workmgd%coarse_grid(l-1), WRD)) - (pci-1) * pnxc
            ip_r  = CEILING(REAL((pi-1) * pnx + i + 1, WRD) / REAL(workmgd%coarse_grid(l-1), WRD)) - (pci-1) * pnxc

            workmgd%levels(l-1)%domain1d(pi)%grid1d(i,MG_SOLN) = 0.5_WRD * workmgd%levels(l)%domain1d(pci)%grid1d(ip,MG_SOLN) + &
                0.25_WRD * workmgd%levels(l)%domain1d(pci)%grid1d(ip_l,MG_SOLN) + &
                0.25_WRD * workmgd%levels(l)%domain1d(pci)%grid1d(ip_r,MG_SOLN) + &
                addOrig * workmgd%levels(l-1)%domain1d(pi)%grid1d(i,MG_SOLN)
                
        END DO

    END DO
    !$OMP END DO

END SUBROUTINE prolongate1d

! prolongation in 2d from the given level l to l-1 
! this version does a simple injection from the coarse zone to all covered fine zones 
SUBROUTINE prolongate2d(self, workmgd, l, withAdd)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    LOGICAL(WIS), INTENT(IN) :: withAdd
    INTEGER(WIS) :: pnx, pnxc, pny, pnyc, p, pi, pj, i, j, pci, pcj, ip, jp
    REAL(WRD) :: addSrc

    pnx  = workmgd%levels(l-1)%domain2d(1,1)%nx
    pnxc = workmgd%levels(l)%domain2d(1,1)%nx
    pny  = workmgd%levels(l-1)%domain2d(1,1)%ny
    pnyc = workmgd%levels(l)%domain2d(1,1)%ny

    addSrc = 0.0_WRD
    IF (withAdd) addSrc = 1.0_WRD

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workmgd%levels(l-1)%npx * workmgd%levels(l-1)%npy

        CALL collapse2d(p, workmgd%levels(l-1)%npx, workmgd%levels(l-1)%npy, pi, pj)

        ! loop over the each fine level zone and inject the overlapping coarse zone value into it 
        DO j = 1, pny

            pcj = CEILING(REAL((pj-1) * pny + j, WRD) / REAL(workmgd%coarse_grid(l-1) * pnyc, WRD))
            jp  = CEILING(REAL((pj-1) * pny + j, WRD) / REAL(workmgd%coarse_grid(l-1), WRD)) - (pcj-1) * pnyc

            DO i = 1, pnx

                pci = CEILING(REAL((pi-1) * pnx + i, WRD) / REAL(workmgd%coarse_grid(l-1) * pnxc, WRD))
                ip  = CEILING(REAL((pi-1) * pnx + i, WRD) / REAL(workmgd%coarse_grid(l-1), WRD)) - (pci-1) * pnxc

                workmgd%levels(l-1)%domain2d(pi,pj)%grid2d(i,j,1) = workmgd%levels(l)%domain2d(pci,pcj)%grid2d(ip,jp,1) + &
                    addSrc * workmgd%levels(l-1)%domain2d(pi,pj)%grid2d(i,j,1)

            END DO

        END DO

    END DO
    !$OMP END DO

END SUBROUTINE prolongate2d

! prolongation in 3d from the given level l to l-1 
! this version does a simple injection from the coarse zone to all covered fine zones 
SUBROUTINE prolongate3d(self, workmgd, l, withAdd)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    LOGICAL(WIS), INTENT(IN) :: withAdd
    INTEGER(WIS) :: pnx, pnxc, pny, pnyc, pnz, pnzc, p, pi, pj, pk, i, j, k, pci, pcj, pck, ip, jp, kp
    REAL(WRD) :: addSrc

    pnx  = workmgd%levels(l-1)%domain1d(1)%nx
    pnxc = workmgd%levels(l)%domain1d(1)%nx
    pny  = workmgd%levels(l-1)%domain1d(1)%ny
    pnyc = workmgd%levels(l)%domain1d(1)%ny
    pnz  = workmgd%levels(l-1)%domain1d(1)%nz
    pnzc = workmgd%levels(l)%domain1d(1)%nz

    addSrc = 0.0_WRD
    IF (withAdd) addSrc = 1.0_WRD

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workmgd%levels(l-1)%npx * workmgd%levels(l-1)%npy

        CALL collapse3d(p, workmgd%levels(l-1)%npx, workmgd%levels(l-1)%npy, workmgd%levels(l-1)%npz, pi, pj, pk)

        ! loop over the each fine level zone and inject the overlapping coarse zone value into it 
        DO k = 1, pnz

            pck = FLOOR(REAL((pk-1) * pnz + k, WRD) / REAL(workmgd%coarse_grid(l) * pnzc, WRD))
            kp  = FLOOR(REAL((pk-1) * pnz + k, WRD) / REAL(workmgd%coarse_grid(l), WRD)) - (pck-1) * pnzc

            DO j = 1, pny

                pcj = FLOOR(REAL((pj-1) * pny + j, WRD) / REAL(workmgd%coarse_grid(l) * pnyc, WRD))
                jp  = FLOOR(REAL((pj-1) * pny + j, WRD) / REAL(workmgd%coarse_grid(l), WRD)) - (pcj-1) * pnyc

                DO i = 1, pnx

                    pci = FLOOR(REAL((pi-1) * pnx + i, WRD) / REAL(workmgd%coarse_grid(l) * pnxc, WRD))
                    ip  = FLOOR(REAL((pi-1) * pnx + i, WRD) / REAL(workmgd%coarse_grid(l), WRD)) - (pci-1) * pnxc

                    workmgd%levels(l-1)%domain3d(pi,pj,pk)%grid3d(i,j,k,1) = workmgd%levels(l)%domain3d(pci,pcj,pck)%grid3d(ip,jp,kp,1) + &
                        addSrc * workmgd%levels(l-1)%domain3d(pi,pj,pk)%grid3d(i,j,k,1)

                END DO

            END DO

        END DO

    END DO
    !$OMP END DO

END SUBROUTINE prolongate3d

! iterate on the given level either the given number of times or to convergence 
SUBROUTINE sweep_level(self, workmgd, dsolve, rankloc, simunits, decomp, prob, l, niters, toConvergence, v)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    TYPE(DomainSolver) :: dsolve
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    TYPE(Decomposition) :: decomp
    TYPE(Problem) :: prob
    INTEGER(WIS), INTENT(IN) :: l, niters, v
    LOGICAL(WIS), INTENT(IN) :: toConvergence
    INTEGER(WIS) :: redi, nb, totiters

    IF (workmgd%levels(l)%ndim .EQ. 1) THEN
        nb = workmgd%levels(l)%domain1d(1)%nb
    ELSE IF (workmgd%levels(l)%ndim .EQ. 2) THEN
        nb = workmgd%levels(l)%domain2d(1,1)%nb
    ELSE IF (workmgd%levels(l)%ndim .EQ. 3) THEN
        nb = workmgd%levels(l)%domain3d(1,1,1)%nb
    END IF
    totiters = 0

    !$OMP MASTER
    self%maxResid = 1.0E15_WRD
    redi          = MPI_REQUEST_NULL

    ! if we're not going to convergence we simply forward along the number of iterations we want to the proper smoothing operator 
    SELECT CASE(self%solve_class)

        ! gravity 
        CASE(1)

            IF (.NOT. toConvergence) THEN
                dsolve%mggrav%niters = niters
            ELSE
                dsolve%mggrav%niters = self%nsmooth_block
            END IF

            ! also set the variable we will be smoothing on and solve type 
            dsolve%mggrav%srci = v
            self%dsolve_flavor = FL_GRAV_FMG

    END SELECT
    !$OMP END MASTER
    !$OMP BARRIER

    ! loop on this level until we have done at least the requested number of sweeps or to convergence
    DO WHILE ((.NOT. toConvergence .AND. totiters .LT. niters) .OR. (self%maxResid .GT. self%convergence_criteria .AND. totiters .LT. niters))

        ! solve this level
        CALL dsolve%solve(workmgd%levels(l), prob, decomp, rankloc, simunits, self%dsolve_flavor)

        ! determine the maximum residual on the domain 
        CALL self%get_maxresid(workmgd, l)

        ! if we have already posted an allreduce lets complete it 
        !$OMP MASTER
        IF (redi .NE. MPI_REQUEST_NULL) THEN

            CALL decomp%end_allreduce(redi)
            self%maxResid = self%maxResidNext

        END IF

        ! do the reduction on the current block 
        CALL decomp%start_allreduce_max(self%maxResidLocal, self%maxResidNext, redi)
        !$OMP END MASTER
        !$OMP BARRIER

        ! a single solve performs multiple sweep based on nb
        totiters = totiters + dsolve%mggrav%niters*FLOOR(REAL(nb)*0.5)

    END DO

    ! there is still a reduction in flight we should complete it 
    !$OMP MASTER
    IF (redi .NE. MPI_REQUEST_NULL) THEN

        CALL decomp%end_allreduce(redi)
        self%maxResid = self%maxResidNext

    END IF
    !$OMP END MASTER
    !$OMP BARRIER

END SUBROUTINE sweep_level

! route to the correct max residual calculation 
SUBROUTINE get_maxresid(self, workmgd, l)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l

    IF (workmgd%ndim .EQ. 1) THEN
        
        CALL self%get_maxresid1d(workmgd, l)
        
    ELSE IF (workmgd%ndim .EQ. 2) THEN
        
        CALL self%get_maxresid2d(workmgd, l)
        
    ELSE IF (workmgd%ndim .EQ. 3) THEN
        
        CALL self%get_maxresid3d(workmgd, l)
        
    END IF

END SUBROUTINE get_maxresid

! determine the maximum residual on the grid in 1d 
SUBROUTINE get_maxresid1d(self, workmgd, l)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    REAL(WRD) :: mymaxresid
    INTEGER(WIS) :: pi, ilow, ihi

    !$OMP SINGLE
    self%maxResidLocal = -1.0E15_WRD
    !$OMP END SINGLE
    mymaxresid = -1.0E15_WRD
    ilow       = 1
    ihi        = workmgd%levels(l)%domain1d(1)%nx

    !$OMP DO SCHEDULE(DYNAMIC)
    DO pi = 1, workmgd%levels(l)%npx

        mymaxresid = MAX(mymaxresid, MAXVAL(ABS(workmgd%levels(l)%domain1d(pi)%grid1d(ilow:ihi,MG_ERROR))))

    END DO
    !$OMP END DO NOWAIT

    !$OMP CRITICAL
    self%maxResidLocal = MAX(self%maxResidLocal, mymaxresid)
    !$OMP END CRITICAL
    !$OMP BARRIER

END SUBROUTINE get_maxresid1d

! determine the maximum residual on the grid in 2d 
SUBROUTINE get_maxresid2d(self, workmgd, l)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    REAL(WRD) :: mymaxresid
    INTEGER(WIS) :: p, pi, pj, ilow, ihi, jlow, jhi

    !$OMP SINGLE
    self%maxResidLocal = -1.0E15_WRD
    !$OMP END SINGLE
    mymaxresid = -1.0E15_WRD
    ilow       = 1
    ihi        = workmgd%levels(l)%domain2d(1,1)%nx
    jlow       = 1
    jhi        = workmgd%levels(l)%domain2d(1,1)%ny

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workmgd%levels(l)%npx * workmgd%levels(l)%npy

        CALL collapse2d(p, workmgd%levels(l)%npx, workmgd%levels(l)%npy, pi, pj)
        mymaxresid = MAX(mymaxresid, MAXVAL(ABS(workmgd%levels(l)%domain2d(pi,pj)%grid2d(ilow:ihi,jlow:jhi,2))))

    END DO
    !$OMP END DO NOWAIT

    !$OMP CRITICAL
    self%maxResidLocal = MAX(self%maxResidLocal, mymaxresid)
    !$OMP END CRITICAL
    !$OMP BARRIER

END SUBROUTINE get_maxresid2d

! determine the maximum residual on the grid in 3d 
SUBROUTINE get_maxresid3d(self, workmgd, l)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    REAL(WRD) :: mymaxresid
    INTEGER(WIS) :: p, pi, pj, pk

    !$OMP SINGLE
    self%maxResidLocal = -1.0E15_WRD
    !$OMP END SINGLE
    mymaxresid = -1.0E15_WRD

    !$OMP DO SCHEDULE(DYNAMIC)
    DO p = 1, workmgd%levels(l)%npx * workmgd%levels(l)%npy * workmgd%levels(l)%npz

        CALL collapse3d(p, workmgd%levels(l)%npx, workmgd%levels(l)%npy, workmgd%levels(l)%npz, pi, pj, pk)
        mymaxresid = MAX(mymaxresid, MAXVAL(ABS(workmgd%levels(l)%domain3d(pi,pj,pk)%grid3d(:,:,:,2))))

    END DO
    !$OMP END DO NOWAIT

    !$OMP CRITICAL
    self%maxResidLocal = MAX(self%maxResidLocal, mymaxresid)
    !$OMP END CRITICAL
    !$OMP BARRIER

END SUBROUTINE get_maxresid3d

! zero soultion will wipe clean the solution at the given level 
SUBROUTINE zero_solution(self, workmgd, l)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    INTEGER(WIS), INTENT(IN) :: l
    INTEGER(WIS) :: p, pi, pj, pk

    IF (workmgd%levels(l)%ndim .EQ. 1) THEN

        !$OMP DO SCHEDULE(DYNAMIC)
        DO pi = 1, workmgd%levels(l)%npx

            workmgd%levels(l)%domain1d(pi)%grid1d(:,1) = 0.0_WRD

        END DO
        !$OMP END DO

    ELSE IF (workmgd%levels(l)%ndim .EQ. 2) THEN

        !$OMP DO SCHEDULE(DYNAMIC)
        DO p = 1, workmgd%levels(l)%npx * workmgd%levels(l)%npy

            CALL collapse2d(p, workmgd%levels(l)%npx, workmgd%levels(l)%npy, pi, pj)
            workmgd%levels(l)%domain2d(pi,pj)%grid2d(:,:,1) = 0.0_WRD

        END DO
        !$OMP END DO

    ELSE IF (workmgd%levels(l)%ndim .EQ. 3) THEN
        
        !$OMP DO SCHEDULE(DYNAMIC)
        DO p = 1, workmgd%levels(l)%npx * workmgd%levels(l)%npy * workmgd%levels(l)%npz

            CALL collapse3d(p, workmgd%levels(l)%npx, workmgd%levels(l)%npy, workmgd%levels(l)%npz, pi, pj, pk)
            workmgd%levels(l)%domain3d(pi,pj,pk)%grid3d(:,:,:,1) = 0.0_WRD

        END DO
        !$OMP END DO

    END IF

END SUBROUTINE zero_solution

! the V cycle scheduler 
RECURSIVE SUBROUTINE do_vcycle(self, workmgd, dsolve, rankloc, simunits, decomp, prob, l, v)

    CLASS(MultiGrid) :: self
    TYPE(MGDomain) :: workmgd
    TYPE(DomainSolver) :: dsolve
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    TYPE(Decomposition) :: decomp
    TYPE(Problem) :: prob
    INTEGER(WIS), INTENT(IN) :: l, v

    ! if we're at the coarsest grid we need to solve to convergence (using the residual as the source)
    IF (l .EQ. workmgd%nlevels-1) THEN

        CALL self%sweep_level(workmgd, dsolve, rankloc, simunits, decomp, prob, l, self%nmaxsmooth, .TRUE., v)

    ! otherwise we work through the V cycle 
    ELSE

        ! do a round of presmoothing 
        CALL self%sweep_level(workmgd, dsolve, rankloc, simunits, decomp, prob, l, self%npresmooth, .FALSE., v)

        ! coarsen the residuals to the source term of the next level
        CALL self%restrict_levels(workmgd, l, l, MG_RESID_SRC, MG_RESID)

        ! reset the solution at the next coarse level to 0 
        CALL self%zero_solution(workmgd, l+1)

        ! do a V cycle starting from the next more coarse domain using the new source term
        CALL self%do_vcycle(workmgd, dsolve, rankloc, simunits, decomp, prob, l+1, MG_RESID_SRC)

        ! interpolate the solution to our finer grid (adding it to our current solution from the TRUE source) 
        CALL self%prolongate_levels(workmgd, l+1, l+1, .TRUE.)

        ! do a round of postsmoothing 
        CALL self%sweep_level(workmgd, dsolve, rankloc, simunits, decomp, prob, l, self%npostsmooth, .FALSE., v)

    END IF

END SUBROUTINE do_vcycle

END MODULE Mod_MultiGrid
