#include "macro.h"

MODULE Mod_MultiGrid_Domain

! FILENAME: mod_multigrid_domain.f90
!
! DESCRIPTION: This module defines a data type that can be used to house MG levels
!  and operations needed to pack/unpack boundaries from them.
!
! DEPENDENCIES: Mod_Globals for defining data types
!
! HISTORY:
!    8/14/17 - Julius Donnert
!    1/5/16  - Peter Mendygral

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Domain
USE Mod_Problem

IMPLICIT NONE

PRIVATE

INTEGER(WIS), PUBLIC :: MG_NVARS, MG_SOLN, MG_SRC, MG_RESID, MG_RESID_SRC, MG_ERROR
PARAMETER(MG_NVARS     = 5, &
          MG_SOLN      = 1, &
          MG_RESID     = 2, &
          MG_SRC       = 3, &
          MG_RESID_SRC = 4, &
          MG_ERROR     = 5)

PUBLIC :: MGDomain

TYPE :: MGDomain

    PRIVATE

    ! we need an array of levels and data about its sizes 
    TYPE(Domain), PUBLIC, ALLOCATABLE :: levels(:)

    !### hold onto information about the base domain and other info we feed to the other domains ###
    INTEGER(WIS), PUBLIC :: tag, ndim, nlevels, nthreads, base_npx, base_npy, base_npz, xi0, yi0, zi0
    REAL(WRD), PUBLIC :: x0, y0, z0, base_dx, dt, t
    LOGICAL(WIS), PUBLIC :: hasDomains, isLive, is1d, is2d, is3d
    LOGICAL(WIS) :: lx_bound, hx_bound, ly_bound, hy_bound, lz_bound, hz_bound

    ! these arrays store how we coarsen between levels and adjust the decomposition 
    INTEGER(WIS), PUBLIC, ALLOCATABLE :: coarse_patches(:), coarse_grid(:)

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: build
    PROCEDURE, PUBLIC :: unbuild
    PROCEDURE, PUBLIC :: build_all_patches
    PROCEDURE, PUBLIC :: unbuild_all_patches

END TYPE MGDomain

INTERFACE MGDomain

    MODULE PROCEDURE constructor

END INTERFACE MGDomain

CONTAINS

FUNCTION constructor(tag, param, ndim, npx, npy, npz, rankloc, simunits) RESULT(self)

    TYPE(MGDomain) :: self
    
    CLASS(Parameters) :: param
    CLASS(RankLocation) :: rankloc
    CLASS(SimulationUnits) :: simunits
    INTEGER(WIS) :: nlevels, coarse_factor_patches(0:WMAX_MG_LEVELS-1), coarse_factor_grid(0:WMAX_MG_LEVELS-1), n
    INTEGER(WIS), INTENT(IN) :: tag, ndim, npx, npy, npz

    !### assign the settings ###
    self%tag       = tag
    self%ndim      = ndim
    self%base_npx  = npx
    self%base_npy  = npy
    self%base_npz  = npz
    self%xi0       = rankloc%izone_origin(1)
    self%yi0       = rankloc%izone_origin(2)
    self%zi0       = rankloc%izone_origin(3)
    self%x0        = rankloc%rzone_origin(1)
    self%y0        = rankloc%rzone_origin(2)
    self%z0        = rankloc%rzone_origin(3)
    self%base_dx   = simunits%dx
    self%dt        = simunits%dt
    self%t         = simunits%t
    self%lx_bound  = rankloc%is_boundary(1,1)
    self%hx_bound  = rankloc%is_boundary(2,1)
    self%ly_bound  = rankloc%is_boundary(1,2)
    self%hy_bound  = rankloc%is_boundary(2,2)
    self%lz_bound  = rankloc%is_boundary(1,3)
    self%hz_bound  = rankloc%is_boundary(2,3)
    self%is1d      = .FALSE.
    self%is2d      = .FALSE.
    self%is3d      = .FALSE.

    IF (ndim .EQ. 1) THEN
        self%base_npy  = 1
        self%base_npz  = 1
        self%is1d      = .TRUE.
    ELSE IF (ndim .EQ. 2) THEN
        self%base_npz  = 1
        self%is2d      = .TRUE.
    ELSE IF (ndim .EQ. 3) THEN
        self%is3d = .TRUE.
    ELSE
        ASSERT(.FALSE.,"An unsupported number of dimensions was requested.  Only 1-3d is supported.")
    END IF

    CALL param%section("Multi Grid Domain", .TRUE.) ! Parameter file
    CALL param%comment("Set the number of levels and the coarsening factor of patches and cells", .TRUE.)
    CALL param%comment("As the number of levels includes the base level there are one less coarse factors than levels", .TRUE.)
    CALL param%comment("", .TRUE.)

    self%nlevels = param%getInt("nlevels", "3", "# total number of levels including base",0, .TRUE.)

    ASSERT(self%nlevels .LE. WMAX_MG_LEVELS, "More than the max allowed MG levels was requested.")

    IF (self%nlevels .GT. 1) THEN ! allocate 1 less than nlevels of the coarsen factors 
                                  ! because the interpretation is from level N to the next, i.e. include base layer

        ALLOCATE(self%coarse_patches(0:self%nlevels-2), self%coarse_grid(0:self%nlevels-2))

        DO n = 0, self%nlevels-2
            self%coarse_patches(n) = param%getInt("coarse_factor_patches", "1", "", n+1, .TRUE.)
        END DO
            
        DO n = 0, self%nlevels-2
            self%coarse_grid(n) = param%getInt("coarse_factor_grid", "1", "", n+1, .TRUE.)
        END DO

    END IF
     
    self%hasDomains = .FALSE. ! default to inactive
    self%isLive     = .FALSE.

END FUNCTION constructor

! build up all child domain levels 
SUBROUTINE build(self, myrank, neighbors, base_nx, base_ny, base_nz, nb, rankloc, simunits)

    CLASS(MGDomain) :: self
    INTEGER(WIS), INTENT(IN) :: neighbors(:)
    INTEGER(WIS), INTENT(IN) :: myrank, base_nx, base_ny, base_nz, nb
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    INTEGER(WIS) :: n, npx, npy, npz, nx, ny, nz, ndx, ndy, ndz, c
    REAL(WRD) :: dx
    LOGICAL(WIS) :: isErr

    ! allocate our Domain levels 
    IF (.NOT. ALLOCATED(self%levels)) THEN

        !$OMP BARRIER
        !$OMP SINGLE

        !### before we do anything, we need to validate that the coarsening at each level does not violate the prolongation/restriction requirement that each level is evenly divisible by the coarsening amount ###
        DO n = 0, self%nlevels-1

            npx = self%base_npx
            npy = self%base_npy
            npz = self%base_npz
            nx  = base_nx
            ny  = base_ny
            nz  = base_nz
            ndx = nx * npx
            ndy = ny * npy
            ndz = nz * npz

            IF (n .LT. self%nlevels-1) THEN

                DO c = 0, n-1
                    
                    isErr = .FALSE.
                    IF (MOD(npx, self%coarse_patches(c)) .GT. 0                       ) isErr = .TRUE.
                    IF (self%ndim .GT. 1 .AND. MOD(npy, self%coarse_patches(c)) .GT. 0) isErr = .TRUE.
                    IF (self%ndim .GT. 2 .AND. MOD(npz, self%coarse_patches(c)) .GT. 0) isErr = .TRUE.
                    
                    RASSERT(.NOT. isErr,"The requested Patch count coarsening is invalid. Be certain that the number of Patches in each physical dimension is evenly divisible by the Patch coarsening factor from one level to the next." )
                    
                    npx = npx / self%coarse_patches(c)
                    IF (self%ndim .GT. 1) npy = npy / self%coarse_patches(c)
                    IF (self%ndim .GT. 2) npz = npz / self%coarse_patches(c)
                    
                    isErr = .FALSE.
                    IF (MOD(ndx, npx * self%coarse_grid(c)) .GT. 0                       ) isErr = .TRUE.
                    IF (self%ndim .GT. 1 .AND. MOD(ndy, npy * self%coarse_grid(c)) .GT. 0) isErr = .TRUE.
                    IF (self%ndim .GT. 2 .AND. MOD(ndz, npz * self%coarse_grid(c)) .GT. 0) isErr = .TRUE.
                    
                    RASSERT(.NOT. isErr,"The requested zone coarsening is invalid.  The total number of zones across the full domain in each physical dimension must be evenly divisible by the the product of the Patch count and grid coarsening factor from one level to the next.")
                    ! we also must check that we have at least nb zones in the resulting Patch 
                    isErr = .FALSE.
                    IF (nx .LT. nb                       ) isErr = .TRUE.
                    IF (self%ndim .GT. 1 .AND. ny .LT. nb) isErr = .TRUE.
                    IF (self%ndim .GT. 2 .AND. nz .LT. nb) isErr = .TRUE.
                    
                    RASSERT(.NOT. isErr,"The requested zone coarsening is invalid.  Be certain the the resulting Patches at each level have at least as many zones as there are boundary zones.")
                    
                    nx = ndx / (npx * self%coarse_grid(c))
                    IF (self%ndim .GT. 1) ny = ndy / (npy * self%coarse_grid(c))
                    IF (self%ndim .GT. 2) nz = ndz / (npz * self%coarse_grid(c))
                    
                    ndx = nx * npx
                    ndy = ny * npy
                    ndz = nz * npz
                
                END DO

            END IF
                
        END DO

        ALLOCATE(self%levels(0:self%nlevels-1))

        ! now loop through and build each level
        DO n = 0, self%nlevels-1

            ! determine the Patch counts for this level and the zone size 
            npx = self%base_npx
            npy = self%base_npy
            npz = self%base_npz
            ndx = base_nx * self%base_npx
            dx  = self%base_dx

            DO c = 0, n-1

                npx = npx / self%coarse_patches(c)
                IF (self%ndim .GT. 1) npy = npy / self%coarse_patches(c)
                IF (self%ndim .GT. 2) npz = npz / self%coarse_patches(c)

                ndx = ndx / self%coarse_grid(c)

            END DO
            dx = self%base_dx * REAL(base_nx * self%base_npx / ndx,WRD)

            self%levels(n)     = Domain(self%tag, self%ndim, npx, npy, npz, rankloc, simunits)
            self%levels(n)%dx  = dx
            self%levels(n)%npx = npx
            self%levels(n)%npy = npy
            self%levels(n)%npz = npz

        END DO
        !$OMP END SINGLE

        !### now go back through and build each of those domains ###
        DO n = 0, self%nlevels-1

            ! determine the zone counts on this level 
            npx = self%base_npx
            npy = self%base_npy
            npz = self%base_npz
            nx  = base_nx
            ny  = 1
            nz  = 1
            IF (self%ndim .GT. 1) ny = base_ny
            IF (self%ndim .GT. 2) nz = base_nz
            ndx = nx * npx
            ndy = ny * npy
            ndz = nz * npz

            DO c = 0, n-1

                npx = npx / self%coarse_patches(c)
                IF (self%ndim .GT. 1) npy = npy / self%coarse_patches(c)
                IF (self%ndim .GT. 2) npz = npz / self%coarse_patches(c)

                nx = ndx / (npx * self%coarse_grid(c))
                IF (self%ndim .GT. 1) ny = ndy / (npy * self%coarse_grid(c))
                IF (self%ndim .GT. 2) nz = ndz / (npz * self%coarse_grid(c))

                ndx = nx * npx
                ndy = ny * npy
                ndz = nz * npz

            END DO

            CALL self%levels(n)%build(myrank, neighbors, MG_NVARS, 0, 0, .FALSE., .FALSE., .FALSE., .FALSE., .FALSE., .FALSE., &
                                      .FALSE., nx, ny, nz, nb, 0, 0, 0, 0, 0.0_WRD)

        END DO

    END IF

    ! we are now active 
    self%hasDomains = .TRUE.
    self%isLive     = .TRUE.

END SUBROUTINE build

! actually allocate all Patches in all levels (later we will want a way to do this one level at a time for AMR upper levels) 
SUBROUTINE build_all_patches(self, prob, rankloc, simunits)

    CLASS(MGDomain) :: self
    TYPE(Problem) :: prob
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    INTEGER(WIS) :: n

    DO n = 0, self%nlevels-1

        CALL self%levels(n)%build_all_patches(prob, rankloc, simunits, .FALSE.)

    END DO

END SUBROUTINE build_all_patches

! free up the Patches in all levels 
SUBROUTINE unbuild_all_patches(self)

    CLASS(MGDomain) :: self
    INTEGER(WIS) :: n

    DO n = 0, self%nlevels-1

        CALL self%levels(n)%unbuild_all_patches()

    END DO

END SUBROUTINE unbuild_all_patches

! tear down the domains and Patches within 
SUBROUTINE unbuild(self)

    CLASS(MGDomain) :: self
    INTEGER(WIS) :: n

    IF (ALLOCATED(self%levels)) THEN

        DO n = 0, self%nlevels-1

            CALL self%levels(n)%unbuild()

        END DO

        !$OMP BARRIER
        !$OMP SINGLE
        DEALLOCATE(self%levels)

        ! we are now inactive 
        self%hasDomains = .FALSE.
        self%isLive     = .FALSE.
        !$OMP END SINGLE

    END IF

END SUBROUTINE unbuild

SUBROUTINE destroy(self)

    CLASS(MGDomain) :: self

    CALL self%unbuild()

    IF (ALLOCATED(self%coarse_patches)) THEN

        !$OMP BARRIER
        !$OMP SINGLE
        DEALLOCATE(self%coarse_patches)
        DEALLOCATE(self%coarse_grid)
        !$OMP END SINGLE

    END IF

    ! default to inactive 
    self%hasDomains = .FALSE.
    self%isLive     = .FALSE.

END SUBROUTINE destroy

END MODULE Mod_MultiGrid_Domain
