#include "macro.h"

MODULE Mod_MultiGrid_Gravity

! FILENAME: mod_MultiGrid_Gravity.f90
!
! DESCRIPTION: This module provides a class for a the smoothing and residuals for solving gravity
!
! DEPENDENCIES: Mod_Globals for defining data types
!
! HISTORY:
!    8/12/2017 - Julius Donnert
!    1/6/2016  - Peter Mendygral

USE Mod_Globals
USE Mod_Error
USE Mod_Patch
USE Mod_MultiGrid_Domain, ONLY : MG_NVARS, MG_SOLN, MG_SRC, MG_RESID, MG_ERROR

IMPLICIT NONE

PRIVATE

PUBLIC :: MultiGrid_Gravity

INTEGER(WID) :: bound_flavor, substep
!$OMP THREADPRIVATE(bound_flavor, substep)

TYPE :: MultiGrid_Gravity

    PRIVATE

    INTEGER(WIS), PUBLIC :: niters, srci
    REAL(WRD), PUBLIC :: gconst, rhofac

    CONTAINS

    PRIVATE

    PROCEDURE :: smooth1d
    PROCEDURE :: smooth2d
    PROCEDURE :: smooth3d
    PROCEDURE :: get_residuals1d
    PROCEDURE :: get_residuals2d
    PROCEDURE :: get_residuals3d

    PROCEDURE, PUBLIC :: solve
    PROCEDURE, PUBLIC :: boundary_flavor_req
    PROCEDURE, PUBLIC :: extract_patch
    PROCEDURE, PUBLIC :: insert_patch

END TYPE MultiGrid_Gravity

INTERFACE MultiGrid_Gravity

    MODULE PROCEDURE constructor

END INTERFACE MultiGrid_Gravity

CONTAINS

FUNCTION constructor(grav_const) RESULT(self)

    TYPE(MultiGrid_Gravity) :: self

    REAL(WRD), INTENT(IN) :: grav_const

    self%gconst = grav_const ! in code units !
    self%rhofac = -(8.0_WRD * ASIN(1.0_WRD) * self%gconst)

END FUNCTION constructor

! extract the gravitating mass and guess at the potential from the given Patch and place it into the source term of 
! the other given Patch 
SUBROUTINE extract_patch(self, workp, mgworkp)

    CLASS(MultiGrid_Gravity) :: self
    TYPE(Patch) :: workp, mgworkp
    INTEGER(WIS) :: i, j, k

    ! this will need to be enhanced later to also add the DM mass

    IF (workp%ndim .EQ. 1) THEN

        DO i = 1-workp%nb, workp%nx+workp%nb

            mgworkp%grid1d(i,MG_SOLN) = workp%grav1d(i,3) ! use the extrapolation to t = n+1/2 (1/2 dt away from the current solve)
            mgworkp%grid1d(i,MG_SRC)  = workp%grid1d(i,1) ! later also include DM mass if enabled

        END DO

    ELSE IF (workp%ndim .EQ. 2) THEN

        DO j = 1-workp%nb, workp%ny+workp%nb

            DO i = 1-workp%nb, workp%nx+workp%nb

                mgworkp%grid2d(i,j,MG_SOLN) = workp%grav2d(i,j,3) ! use the extrapolation to t = n+1/2 (1/2 dt away from the current solve)
                mgworkp%grid2d(i,j,MG_SRC)  = workp%grid2d(i,j,1) ! later also include DM mass if enabled

            END DO

        END DO

    ELSE IF (workp%ndim .EQ. 3) THEN

        DO k = 1-workp%nb, workp%nz+workp%nb

            DO j = 1-workp%nb, workp%ny+workp%nb

                DO i = 1-workp%nb, workp%nx+workp%nb

                    mgworkp%grid3d(i,j,k,MG_SOLN) = workp%grav3d(i,j,k,3) ! use the extrapolation to t = n+1/2 (1/2 dt away from the current solve)
                    mgworkp%grid3d(i,j,k,MG_SRC)  = workp%grid3d(i,j,k,1) ! later also include DM mass if enabled

                END DO

            END DO

        END DO
            
    END IF

END SUBROUTINE extract_patch

! place the solution for the potential from the given Patch into the other given Patch. 
SUBROUTINE insert_patch(self, workp, mgworkp)

    CLASS(MultiGrid_Gravity) :: self
    TYPE(Patch) :: workp, mgworkp
    INTEGER(WIS) :: i, j, k

    IF (workp%ndim .EQ. 1) THEN

        DO i = 1-workp%nb, workp%nx+workp%nb

            workp%grav1d(i,1) = workp%grav1d(i,2)
            workp%grav1d(i,2) = mgworkp%grid1d(i,MG_SOLN)

        END DO

    ELSE IF (workp%ndim .EQ. 2) THEN

        DO j = 1-workp%nb, workp%ny+workp%nb

            DO i = 1-workp%nb, workp%nx+workp%nb

                workp%grav2d(i,j,1) = workp%grav2d(i,j,2)
                workp%grav2d(i,j,2) = mgworkp%grid2d(i,j,MG_SOLN)

            END DO

        END DO

    ELSE IF (workp%ndim .EQ. 3) THEN

        DO k = 1-workp%nb, workp%nz+workp%nb

            DO j = 1-workp%nb, workp%ny+workp%nb

                DO i = 1-workp%nb, workp%nx+workp%nb

                    workp%grav3d(i,j,k,1) = workp%grav3d(i,j,k,2)
                    workp%grav3d(i,j,k,2) = mgworkp%grid3d(i,j,k,MG_SOLN)

                END DO

            END DO

        END DO
            
    END IF

END SUBROUTINE insert_patch

! top-level routine that returns the next required boundary flavor type given the overall pass count
FUNCTION boundary_flavor_req(self, update_pass_count) RESULT(bflavor)

    CLASS(MultiGrid_Gravity) :: self
    INTEGER(WIS), INTENT(IN) :: update_pass_count
    INTEGER(WIS) :: bflavor

    IF (update_pass_count .EQ. 1) THEN
        bflavor = FL_MHDTVD
        substep = 1
    ELSE
        bflavor = bound_flavor
        substep = substep + 1
    END IF

    RETURN

END FUNCTION boundary_flavor_req

!### route to the appropriate solver given the number of dimensions in the Patch we are handed ###
SUBROUTINE solve(self, workp, maxsignal, update_cmplt)

    CLASS(MultiGrid_Gravity) :: self
    TYPE(Patch) :: workp
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    LOGICAL(WIS), INTENT(OUT) :: update_cmplt
    INTEGER(WIS) :: ilow, ihi, jlow, jhi, klow, khi, pxoff, pyoff, pzoff
    LOGICAL(WIS) :: substeps_cmplt

    ! assume we will not complete the update in this pass 
    update_cmplt = .FALSE.

    ! determine the offset based on world grid Patch location
    pxoff = 1 - MOD(workp%xi0, 2)

    ilow = 1
    ihi  = workp%nx + 2*workp%nb

    IF (workp%ndim .GT. 1) THEN

        pyoff = 1 - MOD(workp%yi0, 2)
        
        jlow = 1
        jhi  = workp%ny + 2*workp%nb

    END IF

    !### note that we have to measures residuals first since we only now have fully updated solutions in the boundaries ###
    IF (workp%ndim .EQ. 1) THEN

        IF (workp%update_pass_count .LT. self%niters) CALL self%smooth1d(workp%grid1d, workp%nb, ilow, ihi, workp%dx, pxoff)
        IF (workp%update_pass_count .EQ. self%niters) CALL self%get_residuals1d(workp%grid1d, workp%nb, ilow, ihi, workp%dx)

    ELSE IF (workp%ndim .EQ. 2) THEN

        IF (workp%update_pass_count .LT. self%niters) CALL self%smooth2d(workp%grid2d, workp%nb, ilow, ihi, jlow, jhi, workp%dx, pxoff, pyoff)
        IF (workp%update_pass_count .EQ. self%niters) CALL self%get_residuals2d(workp%grid2d, workp%nb, ilow, ihi, jlow, jhi, workp%dx)

    ELSE IF (workp%ndim .EQ. 3) THEN

        pzoff = 1 - MOD(workp%zi0, 2)

        klow = 1
        khi  = workp%nz + 2*workp%nb
        IF (workp%update_pass_count .LT. self%niters) CALL self%smooth3d(workp%grid3d, workp%nb, ilow, ihi, jlow, jhi, klow, khi, workp%dx, pxoff, pyoff, pzoff)
        IF (workp%update_pass_count .EQ. self%niters) CALL self%get_residuals3d(workp%grid3d, workp%nb, ilow, ihi, jlow, jhi, klow, khi, workp%dx)

    END IF

    !### once we have reached the number of iterations we're suppose to hit we can call this Patch complete ###
    IF (workp%update_pass_count .GE. self%niters) update_cmplt = .TRUE.

END SUBROUTINE solve

!### the 1d smoothing routine ###
SUBROUTINE smooth1d(self, grid, nb, ilow, ihi, dx, pxoff)

    CLASS(MultiGrid_Gravity) :: self
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: grid(:,:)
    INTEGER(WIS), INTENT(IN) :: nb, ilow, ihi, pxoff
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD) :: rhofac
    INTEGER(WIS) :: i, n, s, ifloor, iceil, myilow, mynb, modstep

    !### we use Gauss Seidel for all of this, which means to do red+black in a single pass we need multiples of 2 boundary zones ###
    mynb    = nb - MOD(nb, 2)
    modstep = MOD(substep, 2)

    rhofac = self%rhofac * dx**2
    s      = self%srci

    ! determine the offset this substep should start at working inward from the full boundary zones
    n = mynb-2 - FLOOR(REAL(substep, WRD) * 0.5) * 2
    IF (n .EQ. 0 .AND. modstep .EQ. 1) THEN
        bound_flavor = FL_MHDTVD
        substep      = 1
    END IF

    IF (modstep .EQ. 0) THEN
    
        myilow = ilow - pxoff
        ifloor = myilow+nb-n
        iceil  = ihi-nb+n+1

        !### do the red zones first ###
        DO i = ifloor, iceil, 2

            grid(i,MG_SOLN) = 0.5_WRD * (grid(i+1,MG_SOLN) + grid(i-1,MG_SOLN) - rhofac * grid(i,s))

        END DO

    ELSE

        myilow = ilow - pxoff + 1
        ifloor = myilow+nb-n
        iceil  = ihi-nb+n

        !### now the black sweeps ###
        DO i = ifloor, iceil, 2

            grid(i,MG_SOLN) = 0.5_WRD * (grid(i+1,MG_SOLN) + grid(i-1,MG_SOLN) - rhofac * grid(i,s))

        END DO

    END IF

END SUBROUTINE smooth1d

!### the 2d smoothing routine ###
SUBROUTINE smooth2d(self, grid, nb, ilow, ihi, jlow, jhi, dx, pxoff, pyoff)

    CLASS(MultiGrid_Gravity) :: self
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: grid(:,:,:)
    INTEGER(WIS), INTENT(IN) :: nb, ilow, ihi, jlow, jhi, pxoff, pyoff
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD) :: rhofac
    INTEGER(WIS) :: i, j, n, s, ifloor, iceil, jfloor, jceil, myilow, mynb

    !### we use Gauss Seidel for all of this, which means to do red+black in a single pass we need multiples of 2 boundary zones ###
    mynb = nb - MOD(nb, 2)

    rhofac = self%rhofac * dx**2
    s      = self%srci

    !### we can do nb/2 sweeps total (need at least nb = 2) ###
    DO n = mynb-2, 0, -2

        jfloor = jlow+nb-n-1
        jceil  = jhi-nb+n

        !IF (patch_bound(1,2)) jfloor = jlow+nb
        !IF (patch_bound(2,2)) jceil = jhi-nb
!print *, "RED for n = ", n, "jfloor = ", jfloor, "jceil = ", jceil, patch_bound
        !### do the red zones first ###
        DO j = jfloor, jceil

            !### we cannot update into the world grid boundaries, and also account for the world grid location of this grid for the offset ###
            myilow = ilow + MOD(pyoff + j, 2) - pxoff - 1
            ifloor = myilow+nb-n
            !IF (patch_bound(1,1)) ifloor = ilow+nb+1 - MOD(pyoff + j, 2)
            iceil  = ihi-nb+n+1
            !IF (patch_bound(2,1)) iceil = ihi-nb
!print *, "RED",j,ifloor, iceil, grid(1:6,j,1)
            !!dir$ prefervector
            DO i = ifloor, iceil, 2

                grid(i,j,1) = 0.25_WRD * (grid(i+1,j,1) + grid(i-1,j,1) + grid(i,j+1,1) + grid(i,j-1,1) - rhofac * grid(i,j,s))

            END DO

        END DO

        !IF (.NOT. patch_bound(1,2)) jfloor = jfloor + 1
!print *, "BLACK for n = ", n, "jfloor = ", jfloor, "jceil = ", jceil

        !### now the black sweeps ###
        DO j = jfloor, jceil

            !### we cannot update into the world grid boundaries, and also account for the world grid location of this grid for the offset ###
            myilow = ilow + MOD(pyoff + j, 2) - pxoff
            ifloor = myilow+nb-n
            !IF (patch_bound(1,1)) ifloor = ilow+nb + MOD(pyoff + j, 2)
            iceil  = ihi-nb+n  !+1  this was here and i think it is a bug
            !IF (patch_bound(2,1)) iceil = ihi-nb
!print *, "BLACK",j,ifloor, iceil, grid(1:6,j,1)

            !!dir$ prefervector
            DO i = ifloor, iceil, 2

                grid(i,j,1) = 0.25_WRD * (grid(i+1,j,1) + grid(i-1,j,1) + grid(i,j+1,1) + grid(i,j-1,1) - rhofac * grid(i,j,s))

            END DO

        END DO

    END DO

END SUBROUTINE smooth2d

! the 3d smoothing routine
SUBROUTINE smooth3d(self, grid, nb, ilow, ihi, jlow, jhi, klow, khi, dx, pxoff, pyoff, pzoff)

    CLASS(MultiGrid_Gravity) :: self
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: grid(:,:,:,:)
    INTEGER(WIS), INTENT(IN) :: nb, ilow, ihi, jlow, jhi, klow, khi, pxoff, pyoff, pzoff
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD) :: rhofac, sixth
    INTEGER(WIS) :: i, j, k, ioff, n, s, ifloor, iceil, jfloor, jceil, kfloor, kceil, myilow, mynb

    !### we use Gauss Seidel for all of this, which means to do red+black in a single pass we need multiples of 2 boundary zones ###
    mynb = nb - MOD(nb, 2)

    rhofac = self%rhofac * dx**2
    sixth  = 1.0_WRD / 6.0_WRD
    s      = self%srci

    ! we can do nb/2 sweeps total (need at least nb = 2)
    DO n = mynb-2, 0, -2

        !### do the red zones first ###
        kfloor = klow+nb-n-1
        kceil  = khi-nb+n

        !IF (patch_bound(1,3)) kfloor = klow+nb
        !IF (patch_bound(2,3)) kceil = khi-nb
        DO k = kfloor, kceil

            jfloor = jlow+nb-n-1
            jceil  = jhi-nb+n

            !IF (patch_bound(1,2)) jfloor = jlow+nb
            !IF (patch_bound(2,2)) jceil = jhi-nb
            !print *, "RED for n = ", n, "jfloor = ", jfloor, "jceil = ", jceil, patch_bound
            !### do the red zones first ###
            DO j = jfloor, jceil

                !### we cannot update into the world grid boundaries, and also account for the world grid location of this grid for the offset ###
                myilow = ilow + MOD(pzoff + k, 2) + MOD(pyoff + j, 2) - pxoff - 1
                ifloor = myilow+nb-n
                !IF (patch_bound(1,1)) ifloor = ilow+nb+1 - MOD(pyoff + j, 2)
                iceil  = ihi-nb+n+1
                !IF (patch_bound(2,1)) iceil = ihi-nb
                !print *, "RED",j,ifloor, iceil, grid(1:6,j,1)
                !!dir$ prefervector
                DO i = ifloor, iceil, 2

                    !grid(i,j,1) = 0.25_WRD * (grid(i+1,j,1) + grid(i-1,j,1) + grid(i,j+1,1) + grid(i,j-1,1) - rhofac * grid(i,j,s))

                END DO

            END DO

        END DO

        ! now the black sweeps 
        DO k = klow+n, khi-n

            DO j = jlow+n, jhi-n

                ioff = (1 - MOD(k,2)) * (MOD(j,2) + 1) + MOD(k,2) * (2 - MOD(j,2)) - 1
                !dir$ PREFVEC
                DO i = ilow+n+ioff, ihi-n, 2

                    !grid(i,j,k,1) = sixth * (grid(i+1,j,k,1) + grid(i-1,j,k,1) + grid(i,j+1,k,1) + grid(i,j-1,k,1) + &
                    !     grid(i,j,k+1,1) + grid(i,j,k-1,1) - rhofac * grid(i,j,k,s))

                END DO

            END DO

        END DO

    END DO

END SUBROUTINE smooth3d

! the 1d residuals routine 
SUBROUTINE get_residuals1d(self, grid, nb, ilow, ihi, dx)

    CLASS(MultiGrid_Gravity) :: self
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: grid(:,:)
    INTEGER(WIS), INTENT(IN) :: nb, ilow, ihi
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD) :: irhofac, resid
    INTEGER(WIS) :: i, s

    irhofac = 1.0_WRD / (self%rhofac * dx**2)
    s       = self%srci

    ! sweep the real grid 
    DO i = ilow+nb, ihi-nb

        resid            = grid(i,s) - irhofac * (grid(i+1,MG_SOLN) - 2.0_WRD * grid(i,MG_SOLN) + grid(i-1,MG_SOLN))
        grid(i,MG_ERROR) = resid - grid(i,MG_RESID)
        grid(i,MG_RESID) = resid
      
    END DO

    !### at the boundaries we'll use 0 as the residual (these are ignored except at world boundaries) ###
    grid(ilow:ilow+nb-1,MG_RESID) = 0.0_WRD
    grid(ihi-nb+1:ihi,MG_RESID)   = 0.0_WRD

END SUBROUTINE get_residuals1d

! the 2d residuals routine 
SUBROUTINE get_residuals2d(self, grid, nb, ilow, ihi, jlow, jhi, dx)

    CLASS(MultiGrid_Gravity) :: self
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: grid(:,:,:)
    INTEGER(WIS), INTENT(IN) :: nb, ilow, ihi, jlow, jhi
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD) :: irhofac
    INTEGER(WIS) :: i, j, s

    irhofac = 1.0_WRD / (self%rhofac * dx**2)
    s       = self%srci

    ! sweep the real grid 
    DO j = jlow+nb, jhi-nb

        DO i = ilow+nb, ihi-nb

            grid(i,j,2) = grid(i,j,s) - irhofac * (grid(i+1,j,1) + grid(i-1,j,1) + &
                grid(i,j+1,1) + grid(i,j-1,1) - 4.0_WRD * grid(i,j,1))
        
        END DO
!print '(I4, 8(ES12.3))', j, grid(ilow+nb:ihi-nb,j,2)

    END DO

    !### at the boundaries we'll use 0 as the residual (these are ignored except at world boundaries) ###
    grid(ilow:ilow+nb-1,:,2) = 0.0_WRD
    grid(ihi-nb+1:ihi,:,2)   = 0.0_WRD
    grid(:,jlow:jlow+nb-1,2) = 0.0_WRD
    grid(:,jhi-nb+1:jhi,2)   = 0.0_WRD

END SUBROUTINE get_residuals2d

! the 3d residuals routine 
SUBROUTINE get_residuals3d(self, grid, nb, ilow, ihi, jlow, jhi, klow, khi, dx)

    CLASS(MultiGrid_Gravity) :: self
    REAL(WRD), CONTIGUOUS, INTENT(INOUT) :: grid(:,:,:,:)
    INTEGER(WIS), INTENT(IN) :: nb, ilow, ihi, jlow, jhi, klow, khi
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD) :: irhofac
    INTEGER(WIS) :: i, j, k, s

    irhofac = 1.0_WRD / (self%rhofac * dx**2)
    s       = self%srci

    ! sweep the real grid 
    DO k = klow+nb, khi-nb

        DO j = jlow+nb, jhi-nb

            DO i = ilow+nb, ihi-nb

                grid(i,j,k,2) = grid(i,j,k,s) - irhofac * (grid(i+1,j,k,1) + grid(i-1,j,k,1) + &
                    grid(i,j+1,k,1) + grid(i,j-1,k,1) + grid(i,j,k+1,1) + grid(i,j,k-1,1) - &
                    6.0_WRD * grid(i,j,k,1))
        
            END DO

        END DO

    END DO

    ! at the boundaries we'll use 0 as the residual 
    grid(ilow:ilow+nb-1,:,:,2) = 0.0_WRD
    grid(ihi-nb+1:ihi,:,:,2)   = 0.0_WRD
    grid(:,jlow:jlow+nb-1,:,2) = 0.0_WRD
    grid(:,jhi-nb+1:jhi,:,2)   = 0.0_WRD
    grid(:,:,klow:klow+nb-1,2) = 0.0_WRD
    grid(:,:,khi-nb+1:khi,2)   = 0.0_WRD

END SUBROUTINE get_residuals3d

END MODULE Mod_MultiGrid_Gravity
