#include "macro.h"

MODULE Mod_Parameters

! FILENAME: parameters.f90
!
! DESCRIPTION: Handle Parameter files. We instantiate an object that
! holds the parameter file in two string buffers, the key and the val buffer. 
! A module can then use
!
!   var = param%getX("NAME", "DEFAULT", "# COMMENT", COLUMN, ADV_FLAG)  
!
! to get a parameter, where X=(Log,Int,Flt,Str) denotes the type of the parameter. 
! That call includes, the name in the parameterfile, a default value and a comment. 
! If the code is compiled with GENPARFILE , this information is used to write a valid
! parameter file to disk, given the compile time parameters. Some
! parameters may be be hidden in "standard mode" and are available only with
! ADVANCED_PARAMETER_INTERFACE. The GetX function includes this as a logical flag ADV_FLAG.
! We also allow multi-column parameters in the file. The column number COLUMN
! is another optional parameter. Set it to 0 for no multi column. 
! The return value will always be 8 bytes, Strings will always be WCHAR_BUF_SIZE
! characters long.
! A section or comment in the parameter file is set with 
! 
!   CALL param%section("TEXT")
!   CALL param%comment("TEXT")
!
! EXAMPLE:
!   CALL param%section("MHD WENO5 parameters")
!   self%ndims = param%getInt("ndims", "3", "# Number of dimensions", 0, .FALSE.)
!
! DEPENDENCIES: Globals and Error functions
!
! HISTORY:
!    7/8/17  - Julius Donnert

USE Mod_Globals
USE Mod_Error

IMPLICIT NONE

PRIVATE

PUBLIC :: Parameters

INTEGER, PARAMETER :: FP = 2

INTEGER, PARAMETER :: MAX_LINES = 1024 ! excluding blank & comment lines

#ifdef ADVANCED_PARAMETER_INTERFACE
LOGICAL, PARAMETER :: ADVANCED_MODE_ON = .TRUE.     ! expose advanced parameters
#else
LOGICAL, PARAMETER :: ADVANCED_MODE_ON = .FALSE.    ! hide advance parameters 
#endif

TYPE Parameters

    PRIVATE

    CHARACTER(WCHAR_BUF_SIZE), ALLOCATABLE :: keybuf(:), valbuf(:) ! holds trimmed parameter file
    INTEGER(4) :: nLines
    INTEGER(WID) :: last_column

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: getLog
    PROCEDURE, PUBLIC :: getInt
    PROCEDURE, PUBLIC :: getFlt
    PROCEDURE, PUBLIC :: getStr
    PROCEDURE, PUBLIC :: section
    PROCEDURE, PUBLIC :: comment

    PROCEDURE :: select_and_print
    PROCEDURE :: write_line
    PROCEDURE :: find_val_from_key
    PROCEDURE :: string2integer
    PROCEDURE :: string2logical
    PROCEDURE :: string2real

END TYPE Parameters

INTERFACE Parameters

    MODULE PROCEDURE constructor

END INTERFACE

CONTAINS

FUNCTION constructor() RESULT(self)
    
    TYPE(Parameters) :: self

    CHARACTER(WCHAR_BUF_SIZE) :: fname
    CHARACTER(WCHAR_BUF_SIZE) :: buf, fmt
    INTEGER(WID) :: line, stat, idx, myrank, n_arg
    INTEGER(WIS) :: ierr, nBytes
    LOGICAL :: file_exists


    n_arg = COMMAND_ARGUMENT_COUNT()

    ASSERT(n_arg .EQ. 1, "Wombat requires a single argument specifying the parameter file ! ")

    CALL GET_COMMAND_ARGUMENT(1, fname)

    ALLOCATE(self%keybuf(MAX_LINES), self%valbuf(MAX_LINES))
    
    self%last_column = 0

    IF (GENPARFILE_ON) THEN

        IF (WRANK .EQ. 0) THEN

            INQUIRE(FILE=fname, EXIST=file_exists)

            ASSERT(.NOT. file_exists, "Parameter file '"//TRIM(fname)//"' already exists. Won't overwrite it ...")

            MSG("Generating a valid parameter file in "//TRIM(fname))

            OPEN(UNIT=FP, FILE=fname)

            WRITE(FP,'(A)') "### WOMBAT-2 auto-generated parameter file ###"
            WRITE(FP,'(A)') "### All parameter values in code units (mass_cgs,length_cgs,velocity_cgs) ###"//NEW_LINE('A')
        
        END IF
            
        RETURN ! skip reading, values will be filled with defaults
    END IF

    RMSG("Reading parameter file '"//TRIM(fname)//"'")

    IF (WRANK .EQ. 0) THEN

        line = 0
        stat = 0

        INQUIRE(FILE=fname, EXIST=file_exists)

        ASSERT(file_exists, "Parameter file '"//TRIM(fname)//"' does not exist ! ")

        OPEN(UNIT=FP, FILE=fname)
        
        WRITE(fmt, '("(A", I3,")")') WCHAR_BUF_SIZE ! format string of a line in file

        DO ! read line by line and split into 2 buffers

            READ(FP, fmt, IOSTAT=stat) buf

            IF (stat .NE. 0 .OR. line .EQ. MAX_LINES) THEN

                EXIT ! end of file
            END IF

            buf = ADJUSTL(ADJUSTR(buf))
            
            IF (buf(1:1) .EQ. "#" .OR. buf(1:1) .EQ. " ") THEN
                CYCLE ! ignore line
            END IF

            line = line + 1

            idx = SCAN(buf, " ") ! cut on first white space

            self%keybuf(line) = buf(1:idx-1)         ! already trimmed
            self%valbuf(line) = ADJUSTL(buf(idx+1:)) ! remove left spaces

        END DO

        self%nLines = line

        IF (.NOT. GENPARFILE_ON) THEN
            CLOSE(FP)
        END IF
    
        ASSERT(self%nLines .LE. MAX_LINES, "Too many valid lines in parameter file. Increase MAX_LINES")

    END IF

    CALL MPI_BCAST(self%nLines, 1, MPI_INT, 0, MPI_COMM_WORLD, ierr) ! make available on all ranks
    
    nBytes = self%nLines*WCHAR_BUF_SIZE  
    
    CALL MPI_BCAST(self%keybuf, nBytes, MPI_CHAR, 0, MPI_COMM_WORLD,ierr) 
    CALL MPI_BCAST(self%valbuf, nBytes, MPI_CHAR, 0, MPI_COMM_WORLD,ierr) 
    
END FUNCTION constructor

! close file if we are in write mode.

SUBROUTINE destroy(self)

    CLASS(Parameters) :: self

    IF (GENPARFILE_ON) THEN
        
        WRITE(FP,'(A)') NEW_LINE('A')
        WRITE(FP,'(A)') "# Parameter file end"
        
        CLOSE(FP)
    END IF

    DEALLOCATE(self%keybuf,self%valbuf)

END SUBROUTINE destroy

! Get a logical parameter from the parameter file buffer

FUNCTION getLog(self, key, stdval, comment, column_arg, adv_arg) RESULT(val)

    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: key, comment, stdval
    INTEGER(WIS), OPTIONAL, INTENT(IN) :: column_arg
    LOGICAL, OPTIONAL, INTENT(IN) :: adv_arg

    INTEGER(WID) :: column
    LOGICAL :: adv_only, val

    CHARACTER(WCHAR_BUF_SIZE) :: charval, msg
    
    column = handle_opt_int4(0, column_arg)
    adv_only = handle_opt_logical(.FALSE., adv_arg)

    charval = self%select_and_print(key, stdval, comment, column, adv_only, "LOG")

    val = self%string2logical(charval, key)

    RETURN

END FUNCTION getLog

! Get an integer parameter from the parameter file buffer

FUNCTION getInt(self, key, stdval, comment, column_arg, adv_arg) RESULT(val)

    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: key, comment, stdval
    INTEGER(WIS), OPTIONAL, INTENT(IN) :: column_arg
    LOGICAL, OPTIONAL, INTENT(IN) :: adv_arg

    INTEGER(WID) :: column
    LOGICAL :: adv_only
    INTEGER(WID) :: val

    CHARACTER(WCHAR_BUF_SIZE) :: charval, msg

    column = handle_opt_int4(0, column_arg)
    adv_only = handle_opt_logical(.FALSE., adv_arg)

    charval = self%select_and_print(key, stdval, comment, column, adv_only, "INT ")

    val = self%string2integer(charval, key)
        
    RETURN ! val
END FUNCTION getInt

! Get a floating point parameter from the parameter file buffer

FUNCTION getFlt(self, key, stdval, comment, column_arg, adv_arg) RESULT(val)

    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: key, comment, stdval
    INTEGER(WIS), OPTIONAL, INTENT(IN) :: column_arg
    LOGICAL, OPTIONAL, INTENT(IN) :: adv_arg
 
    INTEGER(WID) :: column
    LOGICAL :: adv_only
    REAL(WRD) :: val

    CHARACTER(WCHAR_BUF_SIZE) :: charval, msg

    column = handle_opt_int4(0, column_arg)
    adv_only = handle_opt_logical(.FALSE., adv_arg)

    charval = self%select_and_print(key, stdval, comment, column, adv_only, "REAL")
            
    val = self%string2real(charval, key)

    RETURN
END FUNCTION getFlt

! Get a string parameter from the parameter file buffer

FUNCTION getStr(self, key, stdval, comment, column_arg, adv_arg) RESULT(charval)
        
    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: key, comment, stdval  
    INTEGER(WIS), OPTIONAL, INTENT(IN) :: column_arg
    LOGICAL, OPTIONAL, INTENT(IN) :: adv_arg
    
    CHARACTER(WCHAR_BUF_SIZE) :: charval, msg
    
    INTEGER(WID) :: column
    LOGICAL :: adv_only
    
    column = handle_opt_int4(0, column_arg)
    adv_only = handle_opt_logical(.FALSE., adv_arg)

    charval = self%select_and_print(key, stdval, comment, column, adv_only, "CHAR")
    
    RETURN ! charval
END FUNCTION getStr

! Write a comment in the parameter file

SUBROUTINE comment(self, comment_arg, adv_arg)
     
    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) ::  comment_arg
    LOGICAL, INTENT(IN), OPTIONAL :: adv_arg
    LOGICAL :: adv_only
 
    adv_only = handle_opt_logical(.FALSE., adv_arg)

    IF (GENPARFILE_ON .AND. self%last_column .GT. 0) THEN
        WRITE(FP,'(" ")')
    END IF

    IF (GENPARFILE_ON .AND. (.NOT. adv_only .OR. ADVANCED_MODE_ON)) THEN
        WRITE(FP,'("# ",A)') TRIM(comment_arg)
    END IF

END SUBROUTINE comment

! Write a section into the parameter file

SUBROUTINE section(self, comment_arg, adv_arg)
     
    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) ::  comment_arg
    LOGICAL, INTENT(IN), OPTIONAL :: adv_arg
    LOGICAL :: adv_only
 
    adv_only = handle_opt_logical(.FALSE., adv_arg)

    IF (GENPARFILE_ON .AND. self%last_column .GT. 0) THEN
        WRITE(FP,'(" ")')
    END IF

    IF (GENPARFILE_ON .AND. (.NOT. adv_only .OR. ADVANCED_MODE_ON)) THEN

        WRITE(FP,'(A,"## ",A," ##",A)') NEW_LINE('A'), TRIM(comment_arg),NEW_LINE('A')
    ELSE
        RMSG(" ")
        RMSG("## "//comment_arg//" ##") ! show the section in the output
        RMSG(" ")
    END IF
    
END SUBROUTINE section

! internal routines

! select the value to return, no conversion is done here. 
! Also print the value which turns out to be a pain because of multi-column lines

FUNCTION select_and_print(self, key, stdval, comment, column, adv_only, typestr_arg) RESULT(charval)
        
    CLASS(Parameters) :: self

    LOGICAL, INTENT(IN) :: adv_only

    CHARACTER(*), INTENT(IN) :: key, stdval, typestr_arg, comment
    CHARACTER(WCHAR_BUF_SIZE) :: charval, msg, typestr

    INTEGER(WID), INTENT(IN) :: column
    INTEGER(WID) :: i, idx, stat

    IF (GENPARFILE_ON) THEN ! write line and return the stdval string

        charval = stdval

        CALL self%write_line(key, stdval, comment, column, adv_only)

        RETURN
    ENDIF

    IF (column .EQ. 0) THEN
        typestr = typestr_arg
    ELSE
        typestr = "MULTI"
    END IF

    IF (column .LT. self%last_column) THEN
        RPRINT '(A)', " " ! insert a new line
    END IF
    
    self%last_column = column

    IF (adv_only .AND. .NOT. ADVANCED_MODE_ON) THEN ! Not in adv mode but param is adv only ..
        
        charval = stdval

        IF (column.EQ.0 .AND. WRANK .EQ. 0) THEN
            WRITE(*,'("Parameter, adv only,  ",A4,": ", A24," = ", A32)') TRIM(typestr), TRIM(key), ADJUSTL(charval)
        ELSE IF (column.EQ.1 .AND. WRANK .EQ. 0) THEN
            WRITE(*,'("Parameter, adv only,  ",A4,": ", A24," = ", A)',ADVANCE='NO') TRIM(typestr), TRIM(key), TRIM(charval)
        ELSE IF (WRANK .EQ. 0) THEN
            WRITE(*,'(" ", A8)',ADVANCE='NO') TRIM(stdval)
        END IF

        RETURN
    END IF

    charval = self%find_val_from_key(key, stdval, column)

    IF (charval .EQ. "") THEN ! not found in file

        charval = stdval

        IF (column.EQ.0 .AND. WRANK .EQ. 0) THEN
            WRITE(*,'("Parameter, default,   ",A4,": ", A24," = ", A32)') TRIM(typestr), TRIM(key), ADJUSTL(charval)
        END IF
  
        RETURN
    ELSE
        IF (column.EQ.0 .AND. WRANK .EQ. 0) THEN
            WRITE(*,'("Parameter, from file, ",A4,": ", A24," = ", A32)') TRIM(typestr), TRIM(key), ADJUSTL(charval)
        ELSE IF (column.EQ.1 .AND. WRANK .EQ. 0) THEN
            WRITE(*,'("Parameter, from file, ",A4,": ", A24," = ", A)',ADVANCE='NO') TRIM(typestr), TRIM(key), TRIM(charval)
        ELSE IF (WRANK .EQ. 0) THEN
            WRITE(*,'(" ", A)',ADVANCE='NO') TRIM(charval)
        ENDIF
    END IF
    

    RETURN
END FUNCTION


! find the index in keybuf that belongs to key and return the substring at index "column"

FUNCTION find_val_from_key(self, key, stdval, column) RESULT(charval)

    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: key, stdval
    INTEGER(WID), INTENT(IN) :: column

    CHARACTER(WCHAR_BUF_SIZE) :: charval
    INTEGER(WID) :: i, idx, stat

    DO i = 1,self%nLines  ! find index belonging to key

        IF (TRIM(self%keybuf(i)) .EQ. TRIM(key)) THEN
            
            charval = self%valbuf(i)
            
            EXIT
        END IF
    END DO
    
    IF (i .GT. self%nLines) THEN

        charval = ""

        RETURN ! key not found
    END IF

    DO i = 1,column-1  ! find column by stripping from the left
        
        idx = SCAN(TRIM(charval), " ")    

        charval = ADJUSTL(TRIM(charval(idx:))) ! remove left chunk

    END DO
        
    IF (charval .NE. "") THEN ! return empty if not found

        idx = SCAN(charval, " ")    

        charval = ADJUSTL(TRIM(charval(1:idx))) ! strip the right
    END IF

    RETURN

END FUNCTION find_val_from_key

SUBROUTINE write_line(self, key, stdval, comment, column, adv_only)

    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: key, comment, stdval
    INTEGER(WID), INTENT(IN) :: column
    
    CHARACTER(WCHAR_BUF_SIZE) :: buf
    LOGICAL, INTENT(IN) :: adv_only

    IF (.NOT. adv_only .OR. ADVANCED_MODE_ON) THEN

        IF (self%last_column .GT. column) THEN

            WRITE(FP,'(A)') NEW_LINE('A') ! add newline if last line was multi parameter

        END IF

        buf = stdval 

        IF (column .EQ. 0) THEN ! single val parameter

            WRITE(FP,'(A30,"  ",A30," ",A)') TRIM(key), ADJUSTL(buf), TRIM(comment)
        
        ELSE IF (column .EQ. 1) THEN ! multi val parameter first column with key
        
            WRITE(FP,'(A30,"  ",A," ")',ADVANCE="NO") TRIM(key), TRIM(buf)
        
        ELSE ! multi val parameter 2nd to last column
        
            WRITE(FP,'(A,"  ",A)',ADVANCE="NO") TRIM(buf), TRIM(comment)
        
        END IF

        self%last_column = column
    END IF

END SUBROUTINE write_line

! conversion routines, string to logical,integer,float

FUNCTION string2logical(self, string, key) RESULT(val)
    
    CLASS(Parameters) :: self

    CHARACTER(*), INTENT(IN) :: string, key
    LOGICAL :: val
    INTEGER(4) :: stat

    val = .FALSE.

    READ(string, '(L1)', IOSTAT=stat) val

    ASSERT(stat == 0, "String conversion to LOGICAL failed: '"//TRIM(string)//"', key: '"//key//"'")

    RETURN
END FUNCTION string2logical

FUNCTION string2integer(self, string, key) RESULT(val)
    
    CLASS(Parameters) :: self
    
    CHARACTER(*) :: string, key
    INTEGER(WID) :: val
    INTEGER(4) :: stat

    val = 0

    READ(string, '(1I12)', IOSTAT=stat) val

    ASSERT(stat == 0, "String conversion to INTEGER failed: '"//TRIM(string)//"', key: '"//key//"'")

    RETURN
END FUNCTION string2integer

FUNCTION string2real(self, string, key) RESULT(val)
    
    CLASS(Parameters) :: self
    
    CHARACTER(*) :: string, key
    REAL(WRD) :: val

    INTEGER(4) :: stat

    val = 0

    READ(string, *, IOSTAT=stat) val

    RASSERT(stat == 0, "String conversion to REAL failed: '"//TRIM(string)//"', key: "//key//"'")

    RETURN
END FUNCTION string2real

END MODULE Mod_Parameters
