#include "macro.h"

MODULE Mod_Particle_Mesh

! FILENAME: mod_particle_mesh.f90
!
! DESCRIPTION: This module will update a given Patch's DM objects with a particle mesh method.
! It contains the Kick-Drift-Kick algorithm (Hockney & Eastwood, Springel 2006)
!
! DEPENDENCIES: Mod_Patch for defining data types.
!
! HISTORY:
!    08/14/17 - Julius Donnert
!    3/19/14  - Peter Mendygral

USE Mod_Globals
USE Mod_Error
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Patch
USE Mod_DM
USE Mod_PMPool

IMPLICIT NONE

PRIVATE
PUBLIC :: ParticleMesh

TYPE :: ParticleMesh

    PRIVATE

    REAL(WRD) :: mass
    INTEGER(WIS), PUBLIC :: ndim, nslotspdm

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: boundary_flavor_req_kd
    PROCEDURE, PUBLIC :: boundary_flavor_req_k
    PROCEDURE, PUBLIC :: solve_kickdrift
    PROCEDURE, PUBLIC :: solve_kick
    PROCEDURE :: assign_density
    PROCEDURE :: kick_dm1d
    PROCEDURE :: kick_dm2d
    PROCEDURE :: kick_dm3d
    PROCEDURE :: drift_dm1d
    PROCEDURE :: drift_dm2d
    PROCEDURE :: drift_dm3d
    PROCEDURE, PUBLIC :: compute_accel1d
    PROCEDURE, PUBLIC :: compute_accel2d
    PROCEDURE, PUBLIC :: compute_accel3d
    PROCEDURE :: assign_dens1d
    PROCEDURE :: assign_dens2d
    PROCEDURE :: assign_dens3d

END TYPE ParticleMesh

INTERFACE ParticleMesh

    MODULE PROCEDURE constructor

END INTERFACE ParticleMesh

CONTAINS

FUNCTION constructor(ndim, nslotspdm, mass) RESULT(self)

    TYPE(ParticleMesh) :: self
    INTEGER(WIS), INTENT(IN) :: ndim, nslotspdm
    REAL(WRD), INTENT(IN) :: mass

    self%ndim           = ndim
    self%nslotspdm      = nslotspdm
    self%mass           = mass

    RETURN

END FUNCTION

! top-level routine that returns the next required boundary flavor type given the overall pass count
FUNCTION boundary_flavor_req_kd(self, update_pass_count) RESULT(bflavor)

    CLASS(ParticleMesh) :: self
    INTEGER(WIS), INTENT(IN) :: update_pass_count
    INTEGER(WIS) :: bflavor

    IF (update_pass_count .EQ. 0) THEN

        bflavor = FL_NULL

    ELSE

        bflavor = FL_DM_PARTICLES

    END IF
        
    RETURN

END FUNCTION boundary_flavor_req_kd

! first part of the gravity solve step, performs the initial kick (assuming particle 
! accelerations are already known), then drift to particle positions. afterwards, 
! find an flag any out-of-bounds particles. (see e.g. Springel 2006)

SUBROUTINE solve_kickdrift(self, workp, pool, rankloc, simunits, maxsignal, update_cmplt)

    CLASS(ParticleMesh) :: self
    TYPE(Patch) :: workp
    TYPE(PMPool) :: pool
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    LOGICAL(WIS), INTENT(OUT) :: update_cmplt
    INTEGER(WIS) :: mi, bi

    update_cmplt = .FALSE.

    ! on the first pass perform a kick
    IF (workp%update_pass_count .EQ. 0) THEN
    
        DO mi = 1, workp%dmMan%nactive ! loop over block managers
       
            DO bi = 1, workp%dmMan%dmbm(mi)%nactive ! loop over blocks in block manager

                SELECT CASE(self%ndim)

                CASE(1)
                    CALL self%kick_dm1d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                         workp%dt, maxsignal )
                    CALL self%drift_dm1d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, & 
                         workp%dt )
                CASE(2)
                    CALL self%kick_dm2d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                         workp%dt, maxsignal)
                    CALL self%drift_dm2d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive,&
                         workp%dt )
                CASE(3)
                    CALL self%kick_dm3d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                         workp%dt, maxsignal )
                    CALL self%drift_dm3d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                         workp%dt )
                END SELECT
                
            END DO
            
        END DO

        CALL workp%dmMan%find_out_of_bounds(rankloc, simunits)

    ! on all others we cycle assigning density and processing OOB particles
    ELSE

        CALL self%assign_density(workp, pool, update_cmplt)

    END IF

END SUBROUTINE solve_kickdrift

! top-level routine that returns the next required boundary flavor type given the overall pass count
FUNCTION boundary_flavor_req_k(self, update_pass_count) RESULT(bflavor)

    CLASS(ParticleMesh) :: self
    INTEGER(WIS), INTENT(IN) :: update_pass_count
    INTEGER(WIS) :: bflavor

    bflavor = FL_GRAV_FMG
    RETURN

END FUNCTION boundary_flavor_req_k

! 4th gravity solve step computes the acceleration on each particle based on the newly caluculated gravity potential, then
! performs the final kick

SUBROUTINE solve_kick(self, workp, pool, maxsignal, update_cmplt)

    CLASS(ParticleMesh) :: self
    TYPE(Patch) :: workp
    TYPE(PMPool) :: pool
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    LOGICAL(WIS), INTENT(OUT) :: update_cmplt
    INTEGER(WIS) :: mi, bi

    workp%dmMan%nstepsuntildefrag = workp%dmMan%nstepsuntildefrag - 1 ! check if it is time to defragment the DM particles

    IF (workp%dmMan%nstepsuntildefrag .EQ. 0) THEN

        CALL workp%dmMan%defragment()
        
        workp%dmMan%nstepsuntildefrag = workp%dmMan%nstepspdefrag ! reset defrag count
    END IF

    
    DO mi = 1, workp%dmMan%nactive ! loop over block managers
       
       DO bi = 1, workp%dmMan%dmbm(mi)%nactive ! loop over blocks in block manager

          SELECT CASE(self%ndim)

             CASE(1)
                CALL  self%compute_accel1d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                           workp%grav1d, workp%x0, workp%dx, workp%nx, workp%nb, pool%ixs, pool%x0s, &
                                           pool%Axs, pool%Bxs, pool%grad1d )
                CALL  self%kick_dm1d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                     workp%dt, maxsignal )
             CASE(2)
                CALL  self%compute_accel2d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                           workp%grav2d, workp%x0, workp%y0, workp%dx, workp%nx, workp%ny, workp%nb, &
                                           pool%ixs, pool%x0s, pool%Axs, pool%Bxs, pool%grad2d )
                CALL  self%kick_dm2d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                     workp%dt, maxsignal )
             CASE(3)
                CALL  self%compute_accel3d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                           workp%grav3d, workp%x0, workp%y0, workp%z0, workp%dx, workp%nx, workp%ny, &
                                           workp%nz, workp%nb, pool%ixs, pool%x0s, pool%Axs, pool%Bxs, pool%grad3d )
                CALL  self%kick_dm3d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                     workp%dt, maxsignal )
          END SELECT
       END DO
    END DO

    update_cmplt = .TRUE.

END SUBROUTINE solve_kick

! PRIVATE METHODS

! 2nd gravity solve procedure - the communication that precedes this step sends out of bounds particles to their neighbors, 
! which must be repeated until all particles have been transferred to their neighbors. this step distributes the mass of 
! the particles, upon completion of this step the mass of all particles will be assigned to the spatial grid of this patch, 
! or to dmMan%OOBmass which will be communicated to neighbors prior to the poisson solver step

SUBROUTINE assign_density(self, workp, pool, update_cmplt)

    CLASS(ParticleMesh) :: self
    TYPE(Patch) :: workp
    TYPE(PMPool) :: pool
    LOGICAL(WIS) :: first_pass
    LOGICAL(WIS), INTENT(OUT) :: update_cmplt
    INTEGER(WIS) :: mi, bi

    CALL workp%dmMan%set_hasOOBslots_flag()

    ! assume update will not be completed
    update_cmplt = .FALSE.

    ! assume this is the first pass
    first_pass = .TRUE.
    IF (workp%update_pass_count .GT. 0) THEN
        first_pass = .FALSE.
    END IF

    ! need two methods to copy particles in: when first_pass == T, copyin will add the first slate of particles to patch 
    ! and this step will assign the mass of all particles currently in the patch (is_active == T); when first_pass == F, 
    ! copyin will add new particles and flag them with isactive = F AND info = -1, and assign_dens will distribute the 
    ! mass of only these new particles
    
    IF (first_pass) THEN
        workp%dmMan%OOBmass(:) = 0.0_WRD ! reset out-of-bounds mass array if first pass
    END IF

    ! distribute mass of particles, always done on first_pass; for subsequent updates, only if particles have been 
    ! copied in on the current update_pass
    IF (first_pass .OR. (.NOT. first_pass .AND. workp%dmMan%gotParticles)) THEN
        
        DO mi = 1, workp%dmMan%nactive ! loop over block managers
           
           DO bi = 1, workp%dmMan%dmbm(mi)%nactive ! loop over blocks in block manager

              SELECT CASE(self%ndim)

                 CASE(1)
                    CALL self%assign_dens1d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                            workp%dmMan%dmbm(mi)%dms(bi)%info, workp%grid1d(:,1), workp%x0, workp%dx, workp%nx, &
                                            workp%nb, pool%ixs, pool%x0s, pool%Axs, pool%Bxs, workp%dmMan%OOBmass, &
                                            workp%dmMan%massoffs, first_pass)
                 CASE(2)
                    CALL self%assign_dens2d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                            workp%dmMan%dmbm(mi)%dms(bi)%info, workp%grid2d(:,:,1), workp%x0, workp%y0, workp%dx, &
                                            workp%nx, workp%ny, workp%nb, pool%ixs, pool%x0s, pool%Axs, pool%Bxs, &
                                            workp%dmMan%OOBmass, workp%dmMan%massoffs, first_pass)
                 CASE(3)
                    CALL self%assign_dens3d(workp%dmMan%dmbm(mi)%dms(bi)%particles, workp%dmMan%dmbm(mi)%dms(bi)%isactive, &
                                            workp%dmMan%dmbm(mi)%dms(bi)%info, workp%grid3d(:,:,:,1), workp%x0, workp%y0, &
                                            workp%z0, workp%dx, workp%nx, workp%ny, workp%nz, workp%nb, pool%ixs, pool%x0s, &
                                            pool%Axs, pool%Bxs, workp%dmMan%OOBmass, workp%dmMan%massoffs, first_pass)
              END SELECT
           END DO
        END DO
    END IF

    ! this needs to repeat until this patch no longer has OOB particles AND all neighbors have passed particles, 
    ! last pass through will be essentially a no-op

    IF (.NOT. workp%dmMan%hasOOBslots .AND. .NOT. workp%dmMan%gotParticles) THEN
        update_cmplt = .TRUE.
    END IF

    ! reset gotParticles flag
    ! PJM - I don't quite follow this yet.  will need debugging

    workp%dmMan%gotParticles = .FALSE.

END SUBROUTINE assign_density

! for DKD leapfrog, factor of .5 must be on drifts 
! this routine will update the given DM object via a kick 

SUBROUTINE kick_dm1d(self, particles, isactive, dt, maxsignal)

    CLASS(ParticleMesh) :: self
    REAL(WRD), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: particles
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    REAL(WRD), INTENT(IN) :: dt
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    REAL(WRD) :: mymax
    INTEGER(WIS) :: i

    DO i = 1, self%nslotspdm
        IF (isactive(i)) THEN
            particles(i,2) = particles(i,2) + 0.5_WRD * dt * particles(i,3)
        END IF
    END DO

    mymax = MAXVAL(ABS(particles(:,2)), MASK=isactive)
    maxsignal(1) = MAX(maxsignal(1), mymax/3)    
    maxsignal(2) = MAX(maxsignal(2), mymax/3)    
    maxsignal(3) = MAX(maxsignal(3), mymax/3)    

END SUBROUTINE kick_dm1d

! this routine will update the given DM object via a kick 

SUBROUTINE kick_dm2d(self, particles, isactive, dt, maxsignal)

    CLASS(ParticleMesh) :: self
    REAL(WRD), DIMENSION(self%nslotspdm, 6), INTENT(INOUT) :: particles
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    REAL(WRD), INTENT(IN) :: dt
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    REAL(WRD) :: mymax
    INTEGER(WIS) :: i

    DO i = 1, self%nslotspdm
        IF (isactive(i)) THEN
            particles(i,3:4) = particles(i,3:4) + 0.5_WRD * dt * particles(i,5:6)
        END IF
    END DO

    mymax = MAXVAL(SQRT(SUM(particles(:,3:4)**2,DIM=2)), MASK=isactive)
    maxsignal(1) = MAX(maxsignal(1), mymax/3)
    maxsignal(2) = MAX(maxsignal(2), mymax/3)
    maxsignal(3) = MAX(maxsignal(3), mymax/3)

END SUBROUTINE kick_dm2d

! this routine will update the given DM object via a kick 
SUBROUTINE kick_dm3d(self, particles, isactive, dt, maxsignal)

    CLASS(ParticleMesh) :: self
    REAL(WRD), DIMENSION(self%nslotspdm, 9), INTENT(INOUT) :: particles
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    REAL(WRD), INTENT(IN) :: dt
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    REAL(WRD) :: mymax
    INTEGER(WIS) :: i


    DO i = 1, self%nslotspdm
        IF (isactive(i)) THEN
            particles(i,4:6) = particles(i,4:6) + 0.5_WRD * dt * particles(i,7:9)
        END IF
    END DO

    mymax = MAXVAL(SQRT(SUM(particles(:,4:6)**2,DIM=2)), MASK=isactive)
    maxsignal(1) = MAX(maxsignal(1), mymax/3)
    maxsignal(2) = MAX(maxsignal(2), mymax/3)
    maxsignal(3) = MAX(maxsignal(3), mymax/3)

    ! The timestep estimates here are now WRONG! This needs fixing.
    ! For symplectic integration we want :  dt = param%alpha * sqrt(2 * softening / acc_phys)
    ! where alpha is something like 0.1 and softening is the cell size

END SUBROUTINE kick_dm3d

! this routine will update the given DM object via a drift 
SUBROUTINE drift_dm1d(self, particles, isactive, dt)

    CLASS(ParticleMesh) :: self
    REAL(WRD), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: particles
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    REAL(WRD), INTENT(IN) :: dt
    INTEGER(WIS) :: i

    DO i = 1, self%nslotspdm
        IF (isactive(i)) THEN
            particles(i,1) = particles(i,1) + dt * particles(i,2)
        END IF
    END DO

END SUBROUTINE drift_dm1d

! this routine will update the given DM object via a drift 
SUBROUTINE drift_dm2d(self, particles, isactive, dt)

    CLASS(ParticleMesh) :: self
    REAL(WRD), DIMENSION(self%nslotspdm, 6), INTENT(INOUT) :: particles
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    REAL(WRD), INTENT(IN) :: dt
    INTEGER(WIS) :: i


    DO i = 1, self%nslotspdm
        IF (isactive(i)) THEN
            particles(i,1:2) = particles(i,1:2) + dt * particles(i,3:4)
        END IF
    END DO

END SUBROUTINE drift_dm2d

! this routine will update the given DM object via a drift 
SUBROUTINE drift_dm3d(self, particles, isactive, dt)

    CLASS(ParticleMesh) :: self
    REAL(WRD), DIMENSION(self%nslotspdm, 9), INTENT(INOUT) :: particles
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    REAL(WRD), INTENT(IN) :: dt
    INTEGER(WIS) :: i

    DO i = 1, self%nslotspdm
        IF (isactive(i)) THEN
            particles(i,1:3) = particles(i,1:3) + dt * particles(i,4:6)
        END IF
    END DO

END SUBROUTINE drift_dm3d

SUBROUTINE assign_dens1d(self, particles, isactive, info, dens1d, x0, dx, nx, nb, ixs, x0s, Axs, Bxs, &
                                OOBmass, massoffs, first_pass)

    CLASS(ParticleMesh) :: self
    
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(INOUT) :: isactive
    LOGICAL(WIS), INTENT(IN) :: first_pass
    LOGICAL(WIS) :: inbounds(2)

    INTEGER(WIS), DIMENSION(self%nslotspdm), INTENT(INOUT) :: info
    INTEGER(WIS), INTENT(IN) :: nx, nb
    INTEGER(WIS), DIMENSION(self%nslotspdm, 1), INTENT(INOUT) :: ixs
    INTEGER(WIS), DIMENSION(2), INTENT(IN) :: massoffs
    INTEGER(WIS) :: idx(2)
    INTEGER(WIS) :: si, wi
    
    REAL(WRD), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: particles
    REAL(WRD), DIMENSION(1-nb:nx+nb), INTENT(INOUT) :: dens1d
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD), DIMENSION(self%nslotspdm, 1), INTENT(INOUT) :: x0s, Axs, Bxs
    REAL(WRD), DIMENSION(:), INTENT(INOUT) :: OOBmass
    REAL(WRD), INTENT(IN) :: x0
    REAL(WRD) :: rhofac, invdx

    invdx = 1.0_WRD / dx
    rhofac = self%mass * invdx**self%ndim

    DO si = 1, self%nslotspdm
       ! check flag for whether particle slot is active
       IF ((first_pass .AND. isactive(si)) .OR. (.NOT. first_pass .AND. .NOT. isactive(si) .AND. info(si) .EQ. -1)) THEN
          ! index of closest cell center to left of particle
          ixs(si,1) = CEILING( (particles(si,1) - x0) * invdx )

          ! coord of that cell center
          x0s(si,1) = x0 + (ixs(si,1) - 1) * dx

          ! Weighting factors
          Axs(si,1) = (particles(si,1) - x0s(si,1)) * invdx
          Bxs(si,1) = 1.0_WRD - Axs(si,1)

       END IF
    END DO
    DO si = 1, self%nslotspdm
       IF ((first_pass .AND. isactive(si)) .OR. (.NOT. first_pass .AND. .NOT. isactive(si) .AND. info(si) .EQ. -1)) THEN
          
          inbounds = .TRUE. ! assume all nodes are inbounds

          
          idx = (/ ixs(si,1), ixs(si,1)+1 /) ! indices of nodes that will accept mass from this particle

          ! determine if any nodes are outside the bounds of this patch, and where in buffer to add mass if so (buffer 
          ! index stored in idx(wi,1))

          DO wi = 1, 2
             IF ( idx(wi) .LT. 1 ) THEN
                idx(wi) = idx(wi) + nb
                inbounds(wi) = .FALSE.
             ELSE IF ( idx(wi) .GT. nx ) THEN
                idx(wi) = nb + idx(wi) - nx
                inbounds(wi) = .FALSE.
             END IF
          END DO

          ! distribute mass
          IF ( inbounds(1) ) THEN
             dens1d(idx(1))      = dens1d(idx(1))  + rhofac*Bxs(si,1)
          ELSE 
             OOBmass(idx(1))     = OOBmass(idx(1)) + rhofac*Bxs(si,1)
          END IF
          IF ( inbounds(2) ) THEN
             dens1d(idx(2))      = dens1d(idx(2))  + rhofac*Axs(si,1)
          ELSE
             OOBmass(idx(2))     = OOBmass(idx(2)) + rhofac*Axs(si,1)
          END IF

          IF (.NOT. first_pass) THEN
             isactive(si) = .TRUE.
             info(si) = 0
          END IF
       END IF
    END DO

END SUBROUTINE assign_dens1d

SUBROUTINE assign_dens2d(self, particles, isactive, info, dens2d, x0, y0, dx, nx, ny, nb, ixs, x0s, Axs, Bxs, OOBmass, massoffs, first_pass)

    CLASS(ParticleMesh) :: self
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(INOUT) :: isactive
    LOGICAL(WIS), INTENT(IN) :: first_pass
    LOGICAL(WIS) :: inbounds(4)
    INTEGER(WIS), DIMENSION(self%nslotspdm), INTENT(INOUT) :: info
    INTEGER(WIS), INTENT(IN) :: nx, ny, nb
    INTEGER(WIS), DIMENSION(8), INTENT(IN) :: massoffs
    INTEGER(WIS), DIMENSION(self%nslotspdm, 2), INTENT(INOUT) :: ixs
    INTEGER(WIS) :: idx(4,2)
    INTEGER(WIS) :: si, wi, i, j
    REAL(WRD), DIMENSION(1-nb:nx+nb,1-nb:ny+nb), INTENT(INOUT) :: dens2d
    REAL(WRD), DIMENSION(self%nslotspdm, 6), INTENT(INOUT) :: particles
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD), DIMENSION(self%nslotspdm, 2), INTENT(INOUT) :: x0s, Axs, Bxs
    REAL(WRD), DIMENSION(:), INTENT(INOUT) :: OOBmass
    REAL(WRD), INTENT(IN) :: x0, y0
    REAL(WRD) :: rhofac, invdx, mfac(4)

    invdx = 1.0_WRD / dx
    rhofac = self%mass * invdx**self%ndim

    DO si = 1, self%nslotspdm
       ! check flag for whether particle slot is active
       IF ((first_pass .AND. isactive(si)) .OR. (.NOT. first_pass .AND. .NOT. isactive(si) .AND. info(si) .EQ. -1)) THEN

          ! index of closest cell center to left of particle
          ixs(si,1) = CEILING( (particles(si,1) - x0) * invdx )
          ixs(si,2) = CEILING( (particles(si,2) - y0) * invdx )

          ! coords of that cell center
          x0s(si,1) = x0 + (ixs(si,1) - 1) * dx
          x0s(si,2) = y0 + (ixs(si,2) - 1) * dx

          ! Weighting factors
          Axs(si,1) = (particles(si,1) - x0s(si,1)) * invdx
          Axs(si,2) = (particles(si,2) - x0s(si,2)) * invdx

          Bxs(si,1) = 1.0_WRD - Axs(si,1)
          Bxs(si,2) = 1.0_WRD - Axs(si,2)

       END IF
    END DO
    DO si = 1, self%nslotspdm
       IF ((first_pass .AND. isactive(si)) .OR. (.NOT. first_pass .AND. .NOT. isactive(si) .AND. info(si) .EQ. -1)) THEN

          ! assume all nodes are inbounds
          inbounds(:) = .TRUE.

          ! indices of nodes that will accept mass from this particle
          DO j = 0, 1
             DO i = 0, 1
                idx(1+i+2*j,:) = (/ ixs(si,1)+i, ixs(si,2)+j /)
             END DO
          END DO

          ! determine if any nodes are outside the bounds of this patch, and where in buffer to add mass if so (buffer indx stored in idx(wi,1))
          DO wi = 1, 4
             IF (idx(wi,2) .LT. 1) THEN
                IF (idx(wi,1) .LT. 1) THEN          !info = 5
                   idx(wi,1)    = massoffs(5) + idx(wi,1)+nb + (idx(wi,2)+nb-1)*nb
                   inbounds(wi) = .FALSE.
                ELSE IF (idx(wi,1) .GT. nx) THEN    !info = 8
                   idx(wi,1)    = massoffs(8) + idx(wi,1)-nx + (idx(wi,2)+nb-1)*nb
                   inbounds(wi) = .FALSE.
                ELSE                                !info = 3
                   idx(wi,1)    = massoffs(3) + idx(wi,1) + (idx(wi,2)+nb-1)*nx
                   inbounds(wi) = .FALSE.
                END IF
             ELSE IF (idx(wi,2) .GT. ny) THEN
                IF (idx(wi,1) .LT. 1) THEN          !info = 7
                   idx(wi,1)    = massoffs(7) + idx(wi,1)+nb + (idx(wi,2)-ny-1)*nb
                   inbounds(wi) = .FALSE.
                ELSE IF (idx(wi,1) .GT. nx) THEN    !info = 6
                   idx(wi,1)    = massoffs(6) + idx(wi,1)-nx + (idx(wi,2)-ny-1)*nb
                   inbounds(wi) = .FALSE.
                ELSE                                !info = 4
                   idx(wi,1)    = massoffs(4) + idx(wi,1) + (idx(wi,2)-ny-1)*nx
                   inbounds(wi) = .FALSE.
                END IF
             ELSE
                IF (idx(wi,1) .LT. 1) THEN          !info = 1
                   idx(wi,1)    = massoffs(1) + idx(wi,1)+nb + (idx(wi,2)-1)*nb
                   inbounds(wi) = .FALSE.
                ELSE IF (idx(wi,1) .GT. nx) THEN    !info = 2
                   idx(wi,1)    = massoffs(2) + idx(wi,1)-nx + (idx(wi,2)-1)*nb
                   inbounds(wi) = .FALSE.
                END IF
             END IF
          END DO

          mfac(:) = rhofac*(/ Bxs(si,1)*Bxs(si,2), Axs(si,1)*Bxs(si,2), Bxs(si,1)*Axs(si,2), Axs(si,1)*Axs(si,2) /)
          ! distribute mass
          DO wi = 1, 4
             IF (inbounds(wi)) THEN
                dens2d(idx(wi,1), idx(wi,2)) = dens2d(idx(wi,1), idx(wi,2)) + mfac(wi)
             ELSE
                OOBmass(idx(wi,1))           = OOBmass(idx(wi,1)) + mfac(wi)
             END IF
          END DO

          IF (.NOT. first_pass) THEN
             isactive(si) = .TRUE.
             info(si) = 0
          END IF

       END IF
    END DO

END SUBROUTINE assign_dens2d

SUBROUTINE assign_dens3d(self, particles, isactive, info, dens3d, x0, y0, z0, dx, nx, ny, nz, nb, ixs, x0s, Axs, Bxs, OOBmass, massoffs, first_pass)

    CLASS(ParticleMesh) :: self
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(INOUT) :: isactive
    LOGICAL(WIS), INTENT(IN) :: first_pass
    LOGICAL(WIS) :: inbounds(8)
    INTEGER(WIS), DIMENSION(self%nslotspdm), INTENT(INOUT) :: info
    INTEGER(WIS), INTENT(IN) :: nx, ny, nz, nb
    INTEGER(WIS), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: ixs
    INTEGER(WIS), DIMENSION(26), INTENT(IN) :: massoffs
    INTEGER(WIS) :: idx(8,3)
    INTEGER(WIS) :: si, wi, i, j, k
    REAL(WRD), DIMENSION(self%nslotspdm, 9), INTENT(INOUT) :: particles
    REAL(WRD), DIMENSION(1-nb:nx+nb,1-nb:ny+nb, 1-nb:nz+nb), INTENT(INOUT) :: dens3d
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: x0s, Axs, Bxs
    REAL(WRD), DIMENSION(:), INTENT(INOUT) :: OOBmass
    REAL(WRD), INTENT(IN) :: x0, y0, z0
    REAL(WRD) :: rhofac, invdx, mfac(8)

    invdx = 1.0_WRD / dx
    rhofac = self%mass * invdx**self%ndim

    DO si = 1, self%nslotspdm

        ! check flag for whether particle slot is active
        IF ((first_pass .AND. isactive(si)) .OR. (.NOT. first_pass .AND. .NOT. isactive(si) .AND. info(si) .EQ. -1)) THEN

            
            ixs(si,1) = CEILING( (particles(si,1) - x0) * invdx ) ! index of closest cell center to left of particle
            ixs(si,2) = CEILING( (particles(si,2) - y0) * invdx )
            ixs(si,3) = CEILING( (particles(si,3) - z0) * invdx )

            x0s(si,1) = x0 + (ixs(si,1) - 1) * dx ! coords of that cell center
            x0s(si,2) = y0 + (ixs(si,2) - 1) * dx
            x0s(si,3) = z0 + (ixs(si,3) - 1) * dx

            Axs(si,1) = (particles(si,1) - x0s(si,1)) * invdx ! Weighting factors
            Axs(si,2) = (particles(si,2) - x0s(si,2)) * invdx
            Axs(si,3) = (particles(si,3) - x0s(si,3)) * invdx

            Bxs(si,1) = 1.0_WRD - Axs(si,1)
            Bxs(si,2) = 1.0_WRD - Axs(si,2)
            Bxs(si,3) = 1.0_WRD - Axs(si,3)

        END IF
    END DO

    DO si = 1, self%nslotspdm
        
        IF ((first_pass .AND. isactive(si)) .OR. (.NOT. first_pass .AND. .NOT. isactive(si) .AND. info(si) .EQ. -1)) THEN

            inbounds(:) = .TRUE. ! assume all nodes are inbounds

          
            DO k = 0, 1 ! indices of nodes that will accept mass from this particle
                DO j = 0, 1
                    DO i = 0, 1
                    idx(1+i+2*j+4*k,:) = (/ ixs(si,1)+i, ixs(si,2)+j, ixs(si,3)+k /)
                    END DO
                END DO
            END DO

            ! determine if any nodes are outside the bounds of this patch, and where in buffer to add mass if so 
            ! (buffer indx stored in idx(wi,1))

            DO wi = 1, 8   ! see mod_decomp line 1167 for destination key

                IF (idx(wi,3) .LT. 1) THEN
          
                    IF (idx(wi,2) .LT. 1) THEN
                   
                        IF (idx(wi,1) .LT. 1) THEN           !info = 14
                            idx(wi,1)    = massoffs(14) + idx(wi,1)+nb + (idx(wi,2)+nb-1)*nb + (idx(wi,3)+nb-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 12
                            idx(wi,1)    = massoffs(12) + idx(wi,1)-nx + (idx(wi,2)+nb-1)*nb + (idx(wi,3)+nb-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 18
                            idx(wi,1)    = massoffs(18) + idx(wi,1) + (idx(wi,2)+nb-1)*nx + (idx(wi,3)+nb-1)*nx*nb
                            inbounds(wi) = .FALSE.
                        END IF

                    ELSE IF (idx(wi,2) .GT. ny) THEN

                        IF (idx(wi,1) .LT. 1) THEN           !info = 7
                            idx(wi,1)    = massoffs(7) + idx(wi,1)+nb + (idx(wi,2)-ny-1)*nb + (idx(wi,3)+nb-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 9
                            idx(wi,1)    = massoffs(9) + idx(wi,1)-nx + (idx(wi,2)-ny-1)*nb + (idx(wi,3)+nb-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 15
                            idx(wi,1)    = massoffs(15) + idx(wi,1) + (idx(wi,2)-ny-1)*nx + (idx(wi,3)+nb-1)*nx*nb
                            inbounds(wi) = .FALSE.
                        END IF
                    ELSE
                   
                        IF (idx(wi,1) .LT. 1) THEN           !info = 19
                            idx(wi,1)    = massoffs(19) + idx(wi,1)+nb + (idx(wi,2)-1)*nb + (idx(wi,3)+nb-1)*ny*nb
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 22
                            idx(wi,1)    = massoffs(22) + idx(wi,1)-nx + (idx(wi,2)-1)*nb + (idx(wi,3)+nb-1)*ny*nb
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 5
                            idx(wi,1)    = massoffs(5) + idx(wi,1) + (idx(wi,2)-1)*nx + (idx(wi,3)+nb-1)*nx*ny
                            inbounds(wi) = .FALSE.
                        END IF
                    END IF

                ELSE IF (idx(wi,3) .GT. nz) THEN

                    IF (idx(wi,2) .LT. 1) THEN
                        
                        IF (idx(wi,1) .LT. 1) THEN           !info = 10
                            idx(wi,1)    = massoffs(10) + idx(wi,1)+nb + (idx(wi,2)+nb-1)*nb + (idx(wi,3)-nz-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 8
                            idx(wi,1)    = massoffs(8) + idx(wi,1)-nx + (idx(wi,2)+nb-1)*nb + (idx(wi,3)-nz-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 16
                            idx(wi,1)    = massoffs(16) + idx(wi,1) + (idx(wi,2)+nb-1)*nx + (idx(wi,3)-nz-1)*nx*nb
                            inbounds(wi) = .FALSE.
                        END IF
                    ELSE IF (idx(wi,2) .GT. ny) THEN
                        IF (idx(wi,1) .LT. 1) THEN           !info = 11
                            idx(wi,1)    = massoffs(11) + idx(wi,1)+nb + (idx(wi,2)-ny-1)*nb + (idx(wi,3)-nz-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 13
                            idx(wi,1)    = massoffs(13) + idx(wi,1)-nx + (idx(wi,2)-ny-1)*nb + (idx(wi,3)-nz-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 17
                            idx(wi,1)    = massoffs(17) + idx(wi,1) + (idx(wi,2)-ny-1)*nx + (idx(wi,3)-nz-1)*nx*nb
                            inbounds(wi) = .FALSE.
                        END IF
                    ELSE
                        IF (idx(wi,1) .LT. 1) THEN           !info = 21
                            idx(wi,1)    = massoffs(21) + idx(wi,1)+nb + (idx(wi,2)-1)*nb + (idx(wi,3)-nz-1)*ny*nb
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 20
                            idx(wi,1)    = massoffs(20) + idx(wi,1)-nx + (idx(wi,2)-1)*nb + (idx(wi,3)-nz-1)*ny*nb
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 6
                            idx(wi,1)    = massoffs(6) + idx(wi,1) + (idx(wi,2)-1)*nx + (idx(wi,3)-nz-1)*nx*ny
                            inbounds(wi) = .FALSE.
                        END IF
                    END IF
                ELSE
                    IF (idx(wi,2) .LT. 1) THEN
                        IF (idx(wi,1) .LT. 1) THEN           !info = 26
                            idx(wi,1)    = massoffs(26) + idx(wi,1)+nb + (idx(wi,2)+nb-1)*nb + (idx(wi,3)-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 24
                            idx(wi,1)    = massoffs(24) + idx(wi,1)-nx + (idx(wi,2)+nb-1)*nb + (idx(wi,3)-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 3
                            idx(wi,1)    = massoffs(3) + idx(wi,1) + (idx(wi,2)+nb-1)*nx + (idx(wi,3)-1)*nx*nb
                            inbounds(wi) = .FALSE.
                        END IF
                    ELSE IF (idx(wi,2) .GT. ny) THEN
                        IF (idx(wi,1) .LT. 1) THEN           !info = 23
                            idx(wi,1)    = massoffs(23) + idx(wi,1)+nb + (idx(wi,2)-ny-1)*nb + (idx(wi,3)-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 25
                            idx(wi,1)    = massoffs(25) + idx(wi,1)-nx + (idx(wi,2)-ny-1)*nb + (idx(wi,3)-1)*nb**2
                            inbounds(wi) = .FALSE.
                        ELSE                                 !info = 4
                            idx(wi,1)    = massoffs(4) + idx(wi,1) + (idx(wi,2)-ny-1)*nx + (idx(wi,3)-1)*nx*nb
                            inbounds(wi) = .FALSE.
                        END IF
                    ELSE
                        IF (idx(wi,1) .LT. 1) THEN           !info = 1
                            idx(wi,1)    = massoffs(1) + idx(wi,1)+nb + (idx(wi,2)-1)*nb + (idx(wi,3)-1)*ny*nb
                            inbounds(wi) = .FALSE.
                        ELSE IF (idx(wi,1) .GT. nx) THEN     !info = 2
                            idx(wi,1)    = massoffs(2) + idx(wi,1)-nx + (idx(wi,2)-1)*nb + (idx(wi,3)-1)*ny*nb
                            inbounds(wi) = .FALSE.
                        END IF
                    END IF
                END IF
            END DO

            mfac(:) = rhofac*(/ Bxs(si,1)*Bxs(si,2)*Bxs(si,3), Axs(si,1)*Bxs(si,2)*Bxs(si,3), Bxs(si,1)*Axs(si,2)*Bxs(si,3), &
                                Axs(si,1)*Axs(si,2)*Bxs(si,3), Bxs(si,1)*Bxs(si,2)*Axs(si,3), Axs(si,1)*Bxs(si,2)*Axs(si,3), &
                                Bxs(si,1)*Axs(si,2)*Axs(si,3), Axs(si,1)*Axs(si,2)*Axs(si,3)   /)

          ! distribute mass

            DO wi = 1, 8
             
                IF (inbounds(wi)) THEN
                    dens3d(idx(wi,1), idx(wi,2), idx(wi,3)) = dens3d(idx(wi,1), idx(wi,2), idx(wi,3)) + mfac(wi)
                 ELSE
                    OOBmass(idx(wi,1))                      = OOBmass(idx(wi,1)) + mfac(wi)
                 END IF
            END DO

            IF (.NOT. first_pass) THEN
                isactive(si) = .TRUE.
                info(si) = 0
            END IF
       END IF
    END DO

END SUBROUTINE assign_dens3d

SUBROUTINE compute_accel1d(self, particles, isactive, grav1d, x0, dx, nx, nb, ixs, x0s, Axs, Bxs, grad1d)

    CLASS(ParticleMesh) :: self
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    INTEGER(WIS), INTENT(IN) :: nx, nb
    INTEGER(WIS), DIMENSION(self%nslotspdm, 1), INTENT(INOUT) :: ixs
    INTEGER(WIS) :: si
    REAL(WRD), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: particles
    REAL(WRD), DIMENSION(1-nb:nx+nb,3), INTENT(IN) :: grav1d
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD), DIMENSION(self%nslotspdm, 1), INTENT(INOUT) :: x0s, Axs, Bxs
    REAL(WRD), DIMENSION(self%nslotspdm, 2), INTENT(INOUT) :: grad1d
    REAL(WRD), INTENT(IN) :: x0
    REAL(WRD) :: twothirds,twelfth,invdx

    twothirds = 2.0_WRD / 3.0_WRD
    twelfth = 1.0_WRD / 12.0_WRD
    invdx = 1.0_WRD / dx

    DO si = 1, self%nslotspdm
       IF ( isactive(si) ) THEN

           ! index of closest cell center to left of particle
           ixs(si,1) = CEILING( (particles(si,1) - x0) * invdx )

           ! coord of that cell center
           x0s(si,1) = x0 + (ixs(si,1) - 1) * dx

           ! weighting factors
           Axs(si,1) = (particles(si,1) - x0s(si,1)) * invdx

           Bxs(si,1) = 1.0_WRD - Axs(si,1)
       END IF
    END DO

    DO si = 1, self%nslotspdm

       IF ( isactive(si) ) THEN
           ! values of gradient at cell centers
           grad1d(si,1) = twothirds * ( grav1d(ixs(si,1)+1,2) - grav1d(ixs(si,1)-1,2) ) * invdx &
                        - twelfth   * ( grav1d(ixs(si,1)+2,2) - grav1d(ixs(si,1)-2,2) ) * invdx
           grad1d(si,2) = twothirds * ( grav1d(ixs(si,1)+2,2) - grav1d(ixs(si,1)  ,2) ) * invdx &
                        - twelfth   * ( grav1d(ixs(si,1)+3,2) - grav1d(ixs(si,1)-1,2) ) * invdx
       END IF
    END DO


    DO si = 1, self%nslotspdm
       IF ( isactive(si) ) THEN

           ! acceleration at particle position
           particles(si,3) = -( Bxs(si,1) * grad1d(si,1) + Axs(si,1) * grad1d(si,2) )

       END IF
    END DO

END SUBROUTINE compute_accel1d

SUBROUTINE compute_accel2d(self, particles, isactive, grav2d, x0, y0, dx, nx, ny, nb, ixs, x0s, Axs, Bxs, grad2d)

    CLASS(ParticleMesh) :: self
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    INTEGER(WIS), INTENT(IN) :: nx, ny, nb
    INTEGER(WIS), DIMENSION(self%nslotspdm, 2), INTENT(INOUT) :: ixs
    INTEGER(WIS) :: si
    REAL(WRD), DIMENSION(self%nslotspdm, 6), INTENT(INOUT) :: particles
    REAL(WRD), DIMENSION(1-nb:nx+nb,1-nb:ny+nb,3), INTENT(IN) :: grav2d
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD), DIMENSION(self%nslotspdm, 2), INTENT(INOUT) :: x0s, Axs, Bxs
    REAL(WRD), DIMENSION(self%nslotspdm,2,2,2), INTENT(INOUT) :: grad2d
    REAL(WRD), INTENT(IN) :: x0, y0
    REAL(WRD) :: twothirds,twelfth,invdx


    twothirds = 2.0_WRD / 3.0_WRD
    twelfth = 1.0_WRD / 12.0_WRD
    invdx = 1.0_WRD / dx


    DO si = 1, self%nslotspdm
       IF ( isactive(si) ) THEN

           ! indices of closest cell center to left of particle
           ixs(si,1) = CEILING( (particles(si,1) - x0) * invdx )
           ixs(si,2) = CEILING( (particles(si,2) - y0) * invdx )

           ! coords of that cell center
           x0s(si,1) = x0 + (ixs(si,1) - 1) * dx
           x0s(si,2) = y0 + (ixs(si,2) - 1) * dx

           ! Weighting factors
           Axs(si,1) = (particles(si,1) - x0s(si,1)) * invdx
           Axs(si,2) = (particles(si,2) - x0s(si,2)) * invdx

           Bxs(si,1) = 1.0_WRD - Axs(si,1)
           Bxs(si,2) = 1.0_WRD - Axs(si,2)
       END IF
    END DO

    DO si = 1, self%nslotspdm
       IF ( isactive(si) ) THEN

           ! values of gradient at cell centers
           grad2d(si,1,1,1) = twothirds * ( grav2d(ixs(si,1)+1,ixs(si,2)  ,2) - grav2d(ixs(si,1)-1,ixs(si,2)  ,2)) * invdx &
                            - twelfth   * ( grav2d(ixs(si,1)+2,ixs(si,2)  ,2) - grav2d(ixs(si,1)-2,ixs(si,2)  ,2)) * invdx
           grad2d(si,1,2,1) = twothirds * ( grav2d(ixs(si,1)+2,ixs(si,2)  ,2) - grav2d(ixs(si,1)  ,ixs(si,2)  ,2)) * invdx &
                            - twelfth   * ( grav2d(ixs(si,1)+3,ixs(si,2)  ,2) - grav2d(ixs(si,1)-1,ixs(si,2)  ,2)) * invdx
           grad2d(si,1,1,2) = twothirds * ( grav2d(ixs(si,1)+1,ixs(si,2)+1,2) - grav2d(ixs(si,1)-1,ixs(si,2)+1,2)) * invdx &
                            - twelfth   * ( grav2d(ixs(si,1)+2,ixs(si,2)+1,2) - grav2d(ixs(si,1)-2,ixs(si,2)+1,2)) * invdx
           grad2d(si,1,2,2) = twothirds * ( grav2d(ixs(si,1)+2,ixs(si,2)+1,2) - grav2d(ixs(si,1)  ,ixs(si,2)+1,2)) * invdx &
                            - twelfth   * ( grav2d(ixs(si,1)+3,ixs(si,2)+1,2) - grav2d(ixs(si,1)-1,ixs(si,2)+1,2)) * invdx
           grad2d(si,2,1,1) = twothirds * ( grav2d(ixs(si,1)  ,ixs(si,2)+1,2) - grav2d(ixs(si,1)  ,ixs(si,2)-1,2)) * invdx & 
                            - twelfth   * ( grav2d(ixs(si,1)  ,ixs(si,2)+2,2) - grav2d(ixs(si,1)  ,ixs(si,2)-2,2)) * invdx 
           grad2d(si,2,2,1) = twothirds * ( grav2d(ixs(si,1)+1,ixs(si,2)+1,2) - grav2d(ixs(si,1)+1,ixs(si,2)-1,2)) * invdx &  
                            - twelfth   * ( grav2d(ixs(si,1)+1,ixs(si,2)+2,2) - grav2d(ixs(si,1)+1,ixs(si,2)-2,2)) * invdx
           grad2d(si,2,1,2) = twothirds * ( grav2d(ixs(si,1)  ,ixs(si,2)+2,2) - grav2d(ixs(si,1)  ,ixs(si,2)  ,2)) * invdx &  
                            - twelfth   * ( grav2d(ixs(si,1)  ,ixs(si,2)+3,2) - grav2d(ixs(si,1)  ,ixs(si,2)-1,2)) * invdx
           grad2d(si,2,2,2) = twothirds * ( grav2d(ixs(si,1)+1,ixs(si,2)+2,2) - grav2d(ixs(si,1)+1,ixs(si,2)  ,2)) * invdx &
                            - twelfth   * ( grav2d(ixs(si,1)+1,ixs(si,2)+3,2) - grav2d(ixs(si,1)+1,ixs(si,2)-1,2)) * invdx

       END IF
    END DO

    DO si = 1, self%nslotspdm
       IF ( isactive(si) ) THEN

           ! components of acceleration at particle position
           particles(si,5) = - ( Bxs(si,1) * Bxs(si,2) * grad2d(si,1,1,1) + Axs(si,1) * Bxs(si,2) * grad2d(si,1,2,1) &
                               + Bxs(si,1) * Axs(si,2) * grad2d(si,1,1,2) + Axs(si,1) * Axs(si,2) * grad2d(si,1,2,2) )
           particles(si,6) = - ( Bxs(si,1) * Bxs(si,2) * grad2d(si,2,1,1) + Axs(si,1) * Bxs(si,2) * grad2d(si,2,2,1) &
                               + Bxs(si,1) * Axs(si,2) * grad2d(si,2,1,2) + Axs(si,1) * Axs(si,2) * grad2d(si,2,2,2) )
       END IF
    END DO


END SUBROUTINE compute_accel2d

SUBROUTINE compute_accel3d(self, particles, isactive, grav3d, x0, y0, z0, dx, nx, ny, nz, nb, ixs, x0s, Axs, Bxs, grad3d)

    CLASS(ParticleMesh) :: self
    LOGICAL(WIS), DIMENSION(self%nslotspdm), INTENT(IN) :: isactive
    INTEGER(WIS), INTENT(IN) :: nx, ny, nz, nb
    INTEGER(WIS), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: ixs
    INTEGER(WIS) :: si
    REAL(WRD), DIMENSION(self%nslotspdm, 9), INTENT(INOUT) :: particles
    REAL(WRD), DIMENSION(1-nb:nx+nb,1-nb:ny+nb,1-nb:nz+nb,3), INTENT(IN) :: grav3d
    REAL(WRD), INTENT(IN) :: dx
    REAL(WRD), DIMENSION(self%nslotspdm, 3), INTENT(INOUT) :: x0s, Axs, Bxs
    REAL(WRD), DIMENSION(self%nslotspdm,3,2,2,2), INTENT(INOUT) :: grad3d
    REAL(WRD), INTENT(IN) :: x0, y0, z0
    REAL(WRD) :: twothirds,twelfth,invdx


    twothirds = 2.0_WRD / 3.0_WRD
    twelfth = 1.0_WRD / 12.0_WRD
    invdx = 1.0_WRD / dx

    DO si = 1, self%nslotspdm

       IF ( isactive(si) ) THEN
           
           ixs(si,1) = CEILING( (particles(si,1) - x0) * invdx ) ! indices of closest cell center to left of particle
           ixs(si,2) = CEILING( (particles(si,2) - y0) * invdx )
           ixs(si,3) = CEILING( (particles(si,3) - z0) * invdx )

           x0s(si,1) = x0 + (ixs(si,1) - 1) * dx ! coords of that cell center
           x0s(si,2) = y0 + (ixs(si,2) - 1) * dx
           x0s(si,3) = z0 + (ixs(si,3) - 1) * dx

           Axs(si,1) = (particles(si,1) - x0s(si,1)) * invdx ! Weighting factors
           Axs(si,2) = (particles(si,2) - x0s(si,2)) * invdx
           Axs(si,3) = (particles(si,3) - x0s(si,3)) * invdx

           Bxs(si,1) = 1.0_WRD - Axs(si,1)
           Bxs(si,2) = 1.0_WRD - Axs(si,2)
           Bxs(si,3) = 1.0_WRD - Axs(si,3)

       END IF
    END DO

    DO si = 1, self%nslotspdm

       IF ( isactive(si) ) THEN

           grad3d(si,1,1,1,1) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+2,ixs(si,2)  ,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)-2,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx
           grad3d(si,1,2,1,1) = twothirds * ( grav3d(ixs(si,1)+2,ixs(si,2)  ,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+3,ixs(si,2)  ,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx
           grad3d(si,1,1,2,1) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)+1,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+2,ixs(si,2)+1,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)-2,ixs(si,2)+1,ixs(si,3)  ,2)) * invdx
           grad3d(si,1,2,2,1) = twothirds * ( grav3d(ixs(si,1)+2,ixs(si,2)+1,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+3,ixs(si,2)+1,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)+1,ixs(si,3)  ,2)) * invdx
           grad3d(si,1,1,1,2) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)  ,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+2,ixs(si,2)  ,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)-2,ixs(si,2)  ,ixs(si,3)+1,2)) * invdx
           grad3d(si,1,2,1,2) = twothirds * ( grav3d(ixs(si,1)+2,ixs(si,2)  ,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+3,ixs(si,2)  ,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)  ,ixs(si,3)+1,2)) * invdx
           grad3d(si,1,1,2,2) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)+1,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+2,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)-2,ixs(si,2)+1,ixs(si,3)+1,2)) * invdx
           grad3d(si,1,2,2,2) = twothirds * ( grav3d(ixs(si,1)+2,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+3,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)-1,ixs(si,2)+1,ixs(si,3)+1,2)) * invdx

           grad3d(si,2,1,1,1) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)-1,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)+2,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)-2,ixs(si,3)  ,2)) * invdx
           grad3d(si,2,2,1,1) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)-1,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)+2,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)-2,ixs(si,3)  ,2)) * invdx
           grad3d(si,2,1,2,1) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)+2,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)+3,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)-1,ixs(si,3)  ,2)) * invdx
           grad3d(si,2,2,2,1) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+2,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)+3,ixs(si,3)  ,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)-1,ixs(si,3)  ,2)) * invdx
           grad3d(si,2,1,1,2) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)-1,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)+2,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)-2,ixs(si,3)+1,2)) * invdx
           grad3d(si,2,2,1,2) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)-1,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)+2,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)-2,ixs(si,3)+1,2)) * invdx
           grad3d(si,2,1,2,2) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)+2,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)+3,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)-1,ixs(si,3)+1,2)) * invdx
           grad3d(si,2,2,2,2) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+2,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)+1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)+3,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)-1,ixs(si,3)+1,2)) * invdx

           grad3d(si,3,1,1,1) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)-1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)-2,2)) * invdx
           grad3d(si,3,2,1,1) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)-1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)-2,2)) * invdx
           grad3d(si,3,1,2,1) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)-1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)-2,2)) * invdx
           grad3d(si,3,2,2,1) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)+1,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)-1,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)-2,2)) * invdx
           grad3d(si,3,1,1,2) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)+3,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)  ,ixs(si,3)-1,2)) * invdx
           grad3d(si,3,2,1,2) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)+3,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)  ,ixs(si,3)-1,2)) * invdx
           grad3d(si,3,1,2,2) = twothirds * ( grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)+3,2) &
                                            - grav3d(ixs(si,1)  ,ixs(si,2)+1,ixs(si,3)-1,2)) * invdx
           grad3d(si,3,2,2,2) = twothirds * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)+2,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)  ,2)) * invdx &
                              - twelfth   * ( grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)+3,2) &
                                            - grav3d(ixs(si,1)+1,ixs(si,2)+1,ixs(si,3)-1,2)) * invdx

       END IF
    END DO

    DO si = 1, self%nslotspdm

       IF ( isactive(si) ) THEN

           ! components of acceleration at particle position
           particles(si,7) = - ( Bxs(si,1) * Bxs(si,2) * Bxs(si,3) * grad3d(si,1,1,1,1) &
                               + Axs(si,1) * Bxs(si,2) * Bxs(si,3) * grad3d(si,1,2,1,1) &
                               + Bxs(si,1) * Axs(si,2) * Bxs(si,3) * grad3d(si,1,1,2,1) &
                               + Axs(si,1) * Axs(si,2) * Bxs(si,3) * grad3d(si,1,2,2,1) &
                               + Bxs(si,1) * Bxs(si,2) * Axs(si,3) * grad3d(si,1,1,1,2) &
                               + Axs(si,1) * Bxs(si,2) * Axs(si,3) * grad3d(si,1,2,1,2) &
                               + Bxs(si,1) * Axs(si,2) * Axs(si,3) * grad3d(si,1,1,2,2) &
                               + Axs(si,1) * Axs(si,2) * Axs(si,3) * grad3d(si,1,2,2,2) )
           particles(si,8) = - ( Bxs(si,1) * Bxs(si,2) * Bxs(si,3) * grad3d(si,2,1,1,1) &
                               + Axs(si,1) * Bxs(si,2) * Bxs(si,3) * grad3d(si,2,2,1,1) &
                               + Bxs(si,1) * Axs(si,2) * Bxs(si,3) * grad3d(si,2,1,2,1) &
                               + Axs(si,1) * Axs(si,2) * Bxs(si,3) * grad3d(si,2,2,2,1) &
                               + Bxs(si,1) * Bxs(si,2) * Axs(si,3) * grad3d(si,2,1,1,2) &
                               + Axs(si,1) * Bxs(si,2) * Axs(si,3) * grad3d(si,2,2,1,2) &
                               + Bxs(si,1) * Axs(si,2) * Axs(si,3) * grad3d(si,2,1,2,2) &
                               + Axs(si,1) * Axs(si,2) * Axs(si,3) * grad3d(si,2,2,2,2) )
           particles(si,9) = - ( Bxs(si,1) * Bxs(si,2) * Bxs(si,3) * grad3d(si,3,1,1,1) &
                               + Axs(si,1) * Bxs(si,2) * Bxs(si,3) * grad3d(si,3,2,1,1) &
                               + Bxs(si,1) * Axs(si,2) * Bxs(si,3) * grad3d(si,3,1,2,1) &
                               + Axs(si,1) * Axs(si,2) * Bxs(si,3) * grad3d(si,3,2,2,1) &
                               + Bxs(si,1) * Bxs(si,2) * Axs(si,3) * grad3d(si,3,1,1,2) &
                               + Axs(si,1) * Bxs(si,2) * Axs(si,3) * grad3d(si,3,2,1,2) &
                               + Bxs(si,1) * Axs(si,2) * Axs(si,3) * grad3d(si,3,1,2,2) &
                               + Axs(si,1) * Axs(si,2) * Axs(si,3) * grad3d(si,3,2,2,2) )

       END IF
    END DO

END SUBROUTINE compute_accel3d

END MODULE Mod_Particle_Mesh
