#include "macro.h"

Module Mod_PMPool

!######################################################################
!#
!# FILENAME: mod_pmpool.f90
!#
!# DESCRIPTION: This module defines an object for work arrays for use
!#              by the ParticleMesh object.
!#
!# DEPENDENCIES: Mod_Globals for defining data types
!#
!# HISTORY:
!#    5/15/14 - Brian O'Neill
!#
!######################################################################

USE Mod_Globals
USE Mod_DataPool

IMPLICIT NONE

!### default private but make the data type available ###
PRIVATE
PUBLIC :: PMPool

!### define a data type for this module ###
TYPE, EXTENDS(DataPool) :: PMPool

    !### default everything as private ###
    PRIVATE

    INTEGER(WIS), PUBLIC, ALLOCATABLE :: ixs(:,:)                                    ! indices of reference cells for density assignment and acceleration computation
    REAL(WRD), PUBLIC, ALLOCATABLE :: x0s(:,:), &                                   ! coords of reference cells
                                      Axs(:,:), Bxs(:,:), &                         ! dimensional weighting factors, Bx's represent 1(cell center .LT. particle pos.) and Ax's represent 2(cell center .GT. particle pos.), counter-intuitive, yes but Bx's are calculated from Ax's
                                      grad1d(:,:),grad2d(:,:,:,:),grad3d(:,:,:,:,:) ! gradients of gravitational potential ; grad1d - 1st index is particle number, 2nd index is the value of the gradient at the 2 nearest cell centers ; grad2d - 1st index is particle number, 2nd index is gradient direction, 3rd & 4th indices represent the 4 nearest cell centers ; grad3d - 1st index is particle number, 2nd index is gradient direction, 3rd, 4th, & 5th indices represent the 8 nearest cell centers

    CONTAINS

    !### default everything as private ###
    PRIVATE

    PROCEDURE :: initpmpool
    PROCEDURE , PUBLIC :: destroy

END TYPE PMPool

INTERFACE PMPool

    MODULE PROCEDURE constructor

END INTERFACE

CONTAINS

!### constructor call
FUNCTION constructor(ndim, nslotspdm)

    TYPE(PMPool) :: constructor
    INTEGER(WIS), INTENT(IN) :: ndim, nslotspdm

    CALL constructor%initpmpool(ndim, nslotspdm )
    RETURN

END FUNCTION constructor

SUBROUTINE initpmpool(self, ndim, nslotspdm)
    CLASS(PMPool) :: self
    INTEGER(WIS), INTENT(IN) :: ndim, nslotspdm

    SELECT CASE(ndim)

    CASE (1)
        ALLOCATE(self%ixs(nslotspdm,1))
        ALLOCATE(self%x0s(nslotspdm,1))
        ALLOCATE(self%Axs(nslotspdm,1), self%Bxs(nslotspdm,1))
        ALLOCATE(self%grad1d(nslotspdm,2))
    CASE (2)
        ALLOCATE(self%ixs(nslotspdm,2))
        ALLOCATE(self%x0s(nslotspdm,2))
        ALLOCATE(self%Axs(nslotspdm,2), self%Bxs(nslotspdm,2))
        ALLOCATE(self%grad2d(nslotspdm,2,2,2))
    CASE (3)
        ALLOCATE(self%ixs(nslotspdm,3))
        ALLOCATE(self%x0s(nslotspdm,3))
        ALLOCATE(self%Axs(nslotspdm,3), self%Bxs(nslotspdm,3))
        ALLOCATE(self%grad3d(nslotspdm,3,2,2,2))
    END SELECT

END SUBROUTINE initpmpool

SUBROUTINE destroy(self)
    CLASS(PMPool) :: self

    IF (ALLOCATED(self%ixs)) THEN

        DEALLOCATE(self%ixs, self%x0s, self%Axs, self%Bxs)
        IF (ALLOCATED(self%grad1d)) DEALLOCATE(self%grad1d)
        IF (ALLOCATED(self%grad2d)) DEALLOCATE(self%grad2d)
        IF (ALLOCATED(self%grad3d)) DEALLOCATE(self%grad3d)

    END IF


END SUBROUTINE destroy
END MODULE
