#include "../macro.h"

#ifdef GRAVITY_SUITE

MODULE Mod_Problem

!  FILENAME: Gravity_Suite.f90
! 
!  DESCRIPTION: This module provides a class for initializing a domain
!    and setting boundary values a variety of tests of the gravity solver.
! 
!  DEPENDENCIES: Mod_Globals for defining data types
! 
!  HISTORY:
!     8/16/2017  - Julius Donnert
!     1/6/2017  - Pete Mendygral

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_ProblemBase
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Patch

IMPLICIT NONE

PRIVATE
PUBLIC :: Problem

TYPE, EXTENDS(ProblemBase) :: Problem

    PRIVATE

    CHARACTER*30 :: init_routine, bounds_routine
    REAL(WRD), PUBLIC :: gamma, a1,a3,e,rho

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: patch_init
    PROCEDURE, PUBLIC :: patch_bounds
    !PROCEDURE :: grav_init
    !PROCEDURE :: phi_in
    !PROCEDURE :: phi_out
    !PROCEDURE :: bA1_calc
    !PROCEDURE :: bA3_calc
    !PROCEDURE :: h_calc
    !PROCEDURE :: e_calc
    !PROCEDURE :: lambda_calc

END TYPE Problem

INTERFACE Problem

    MODULE PROCEDURE constructor

END INTERFACE Problem

CONTAINS

! constructor for the class 
!   arguments are: the namelist file contents for grabbing parameters out of 
FUNCTION constructor(param, rankloc, simunits) RESULT(self)

    TYPE(Problem) :: self
    TYPE(Parameters) :: param
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits

    CALL param%section("Initial Conditions: Gravity Suite")

    self%gamma          = param%getFlt("gamma", "1.6666666666667", "# Adiabatic Index")
    self%init_routine   = param%getStr("init_routine", "Gauss", "# Available: 'Gauss'")
    self%a1             = param%getFlt("a1", "1.0", "#spheroid radius")
    self%a3             = param%getFlt("a3","1.0", "#spheroid radius")
    self%e              = param%getFlt("e", "1.0", "#spheroid eccentricity")
    self%rho            = param%getFlt("rho","1.0", "#spheroid density")
    self%bounds_routine = param%getStr("bounds_routine", "", "# ")

    RETURN

END FUNCTION constructor

SUBROUTINE destroy(self)

    CLASS(Problem) :: self
    
    RETURN

END SUBROUTINE destroy

! the routine called to initialize every Patch 
SUBROUTINE patch_init(self, workp, rankloc, simunits, is_boundary, tid, ierr)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
    INTEGER(WIS), INTENT(IN) :: tid
    INTEGER(WIS), INTENT(OUT) :: ierr

    ! route to the correct initialization routine 
    SELECT CASE (TRIM(self%init_routine))

        CASE ('Gauss')

            CALL grav_init(self, workp, simunits)
            ierr = 0

        CASE('Spheroid')

            CALL grav_spheriod_init(self, workp, simunits)
            ierr = 0

        CASE DEFAULT

            CALL Error('Mod_Problem', 'Unknown initialization routine requested => "' // TRIM(self%init_routine) // '"')
            ierr = 1

    END SELECT

END SUBROUTINE patch_init

! the routine called to set boundary values (called for every Patch even if it is not a world boundary) 
SUBROUTINE patch_bounds(self, workp, rankloc, simunits, is_boundary, maxsignal, tid, ierr)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    INTEGER(WIS), INTENT(OUT) :: ierr
    INTEGER(WIS), INTENT(IN) :: tid

    maxsignal(1) = MAX(maxsignal(1), 0.0_WRD)
    maxsignal(2) = MAX(maxsignal(2), 0.0_WRD)
    maxsignal(3) = MAX(maxsignal(3), 0.0_WRD)

    ! route to the correct initialization routine 
    SELECT CASE (TRIM(self%bounds_routine))

       CASE('Gauss')

            CALL grav_bounds(self,workp, is_boundary)    
            ierr = 0

       CASE('Spheriod')
 
            CALL grav_spheriod_bounds(self,workp, is_boundary)
            ierr = 0

       CASE DEFAULT

            ELSE IF (workp%is2d) THEN

                   !### lower X ###
                   IF (is_boundary(1,1)) THEN

                       DO i = 1-workp%nb, 0
                           workp%grid2d(i,:,:) = 0.0_WRD
                       END DO

                   END IF

                   !### upper X ###
                   IF (is_boundary(2,1)) THEN
                       
                       DO i = workp%nx+1, workp%nx+workp%nb
                           workp%grid2d(i,:,:) = 0.0_WRD
                       END DO

                   END IF

                   !### lower Y ###
                   IF (is_boundary(1,2)) THEN

                       DO j = 1-workp%nb, 0
                           workp%grid2d(:,j,:) = 0.0_WRD
                       END DO

                   END IF

                   !### upper Y ###
                   IF (is_boundary(2,2)) THEN
                       
                       DO j = workp%ny+1, workp%ny+workp%nb
                           workp%grid2d(:,j,:) = 0.0_WRD
                       END DO

                   END IF

               END IF

           END IF

           !### we're periodic in all directions so do nothing ###
           ierr = 0

    END SELECT

END SUBROUTINE patch_bounds

! setup 
SUBROUTINE grav_init(self, workp, simunits)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    TYPE(SimulationUnits) :: simunits
    REAL(WRD) :: pos(2)
    INTEGER(WIS) :: i, j, k

    IF (workp%tag .NE. self%stateTag) RETURN

    IF (workp%is1d) THEN

        DO i = 1-workp%nb, workp%nx+workp%nb
   
            pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx

            !workp%grid1d(i,1) = 10.0_WRD * EXP(-(pos(1)**2) / (0.4_WRD**2))
            !workp%grid1d(i,1) = 10.0_WRD * EXP(-(pos(1)**2) / (0.01_WRD**2))
            workp%grid1d(i,1) = 0.0_WRD
            IF (ABS(pos(1)) .LE. workp%dx) workp%grid1d(i,1) = 10.0_WRD
            workp%grid1d(i,2) = 0.0_WRD
            workp%grid1d(i,3) = 0.0_WRD
            workp%grid1d(i,4) = 0.0_WRD
            workp%grid1d(i,5) = 0.0_WRD
            workp%grid1d(i,6) = 0.0_WRD
            workp%grid1d(i,7) = 0.0_WRD
            workp%grid1d(i,8) = 1.0_WRD

        END DO

    ELSE IF (workp%is2d) THEN

        ! vary though the grid along X only 
        DO j = 1-workp%nb, workp%ny+workp%nb
   
            DO i = 1-workp%nb, workp%nx+workp%nb
   
                pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                
                workp%grid2d(i,j,1)  = EXP(-(SQRT(pos(1)**2 + pos(2)**2)) / (0.25_WRD**2))
                workp%grid2d(i,j,2)  = 0.0_WRD
                workp%grid2d(i,j,3)  = 0.0_WRD
                workp%grid2d(i,j,4)  = 0.0_WRD
                workp%bface2d(i,j,1) = 0.0_WRD
                workp%bface2d(i,j,2) = 0.0_WRD
                workp%grid2d(i,j,7)  = 0.0_WRD
                IF (workp%passOn .AND. workp%npass .GT. 0) THEN

                    workp%pass2d(i,j,:) = 0.0_WRD
                    
                END IF
                
            END DO
            
        END DO
        
        ! loop back through and set Bx/y and energy 
        DO j = 1, workp%ny
            
            DO i = 1, workp%nx
                
                workp%grid2d(i,j,5) = 0.5_WRD * (workp%bface2d(i,j,1) + workp%bface2d(i-1,j,1))
                workp%grid2d(i,j,6) = 0.5_WRD * (workp%bface2d(i,j,2) + workp%bface2d(i,j-1,2))
                
                workp%grid2d(i,j,8) = 0.6_WRD / (self%gamma - 1.0_WRD) + (0.5_WRD / workp%grid2d(i,j,1)) &
                                    * (workp%grid2d(i,j,2)**2 + workp%grid2d(i,j,3)**2 + workp%grid2d(i,j,4)**2) &
                                    + 0.5_WRD * (workp%grid2d(i,j,5)**2 + workp%grid2d(i,j,6)**2 + workp%grid2d(i,j,7)**2)
                
            END DO
            
        END DO
        
    END IF

END SUBROUTINE grav_init


SUBROUTINE grav_spheriod_init(self, workp, simunits)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    TYPE(SimulationUnits) :: simunits
    REAL(WRD) :: pos(3),  p
    INTEGER(WIS) :: i, j, k

    self%a3 = self%a1*SQRT(1.0_WRD - self%e**2) 

    IF (workp%tag .NE. self%stateTag) RETURN

        ! vary though the grid  
        DO k = 1-workp%nb, workp%nz+workp%nb

            DO j = 1-workp%nb, workp%ny+workp%nb
        
                DO i = 1-workp%nb, workp%nx+workp%nb
   
                        pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                        pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                        pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                        ! find the parameter position, lessthan/equalto 1 is inside the
                        ! spheriod, greater than one is outside it
                        p = (pos(1)**2 + pos(2)**2)/self%a1**2 + pos(3)**2/self%a3**2

                        IF (p .LE. 1.0_WRD) THEN

                                workp%grid3d(i,j,k,1)  = self%rho
                
                        ELSE
                                workp%grid3d(i,j,k,1)   = 0.0_WRD

                        END IF

                        workp%grid3d(i,j,k,2)  = 0.0_WRD
                        workp%grid3d(i,j,k,3)  = 0.0_WRD
                        workp%grid3d(i,j,k,4)  = 0.0_WRD
                        workp%bface3d(i,j,k,1) = 0.0_WRD
                        workp%bface3d(i,j,k,2) = 0.0_WRD
                        workp%bface3d(i,j,k,3) = 0.0_WRD
                
                END DO
            
            END DO
       
        END DO

        ! loop back through and set Bx/y/z and energy 
        DO k =1, workp%nz

            DO j = 1, workp%ny
            
                DO i = 1, workp%nx
                
                    workp%grid3d(i,j,k,5) = 0.5_WRD * (workp%bface3d(i,j,k,1) + workp%bface3d(i-1,j,k,1))
                    workp%grid3d(i,j,k,6) = 0.5_WRD * (workp%bface3d(i,j,k,2) + workp%bface3d(i,j-1,k,2))
                    workp%grid3d(i,j,k,7) = 0.5_WRD * (workp%bface3d(i,j,k,3) + workp%bface3d(i,j,k-1,3))

                    workp%grid3d(i,j,k,8) = 0.6_WRD / (self%gamma - 1.0_WRD) + (0.5_WRD / workp%grid3d(i,j,k,1)) &
                                    * (workp%grid3d(i,j,k,2)**2 + workp%grid3d(i,j,k,3)**2 + workp%grid3d(i,j,k,4)**2) &
                                    + 0.5_WRD * (workp%grid3d(i,j,k,5)**2 + workp%grid3d(i,j,k,6)**2 + workp%grid3d(i,j,k,7)**2)
                
                END DO

            END DO

        END DO
        

END SUBROUTINE grav_spheriod_init

SUBROUTINE grav_bounds(self,workp,is_boundary)

        CLASS(Problem) :: self
        TYPE(Patch) :: workp
        LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
        LOGICAL(WIS) :: lx_bound, hx_bound, ly_bound, hy_bound, lz_bound, hz_bound
        INTEGER(WIS) :: i,j,k

        lx_bound = is_boundary(1,1)
        hx_bound = is_boundary(2,1)
        ly_bound = is_boundary(1,2)
        hy_bound = is_boundary(2,2)
        lz_bound = is_boundary(1,3)
        hz_bound = is_boundary(2,3)

         ! we're periodic in all directions so do nothing 
           IF (workp%tag .EQ. self%mggravTag) THEN

               IF (workp%is1d) THEN

                   !### lower X ###
                   IF (lx_bound) THEN

                       DO i = 1-workp%nb, 0
                           workp%grid1d(i,:) = 0.0_WRD
                       END DO

                   END IF

                   !### upper X ###
                   IF (hx_bound) THEN

                       DO i = workp%nx+1, workp%nx+workp%nb
                           workp%grid1d(i,:) = 0.0_WRD
                       END DO

                   END IF

               ELSE IF (workp%is2d) THEN

                   !### lower X ###
                   IF (lx_bound) THEN

                       DO i = 1-workp%nb, 0
                           workp%grid2d(i,:,:) = 0.0_WRD
                       END DO

                   END IF

                   !### upper X ###
                   IF (hx_bound) THEN

                       DO i = workp%nx+1, workp%nx+workp%nb
                           workp%grid2d(i,:,:) = 0.0_WRD
                       END DO

                   END IF

                   !### lower Y ###
                   IF (ly_bound) THEN

                       DO j = 1-workp%nb, 0
                           workp%grid2d(:,j,:) = 0.0_WRD
                       END DO

                   END IF

                   !### upper Y ###
                   IF (hy_bound) THEN

                       DO j = workp%ny+1, workp%ny+workp%nb
                           workp%grid2d(:,j,:) = 0.0_WRD
                       END DO

                   END IF

               END IF

           END IF
            

END SUBROUTINE grav_bounds


SUBROUTINE grav_spheriod_bounds(self,workp,is_boundary)

        CLASS(Problem) :: self
        TYPE(Patch) :: workp
        LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
        LOGICAL(WIS) :: lx_bound, hx_bound, ly_bound, hy_bound, lz_bound, hz_bound
        REAL(WRD) :: pos(3),x,y,z, p, analyticalPhi, calculatedPhi, error
        INTEGER(WIS) :: i, j, k



        lx_bound = is_boundary(1,1)
        hx_bound = is_boundary(2,1)
        ly_bound = is_boundary(1,2)
        hy_bound = is_boundary(2,2)
        lz_bound = is_boundary(1,3)
        hz_bound = is_boundary(2,3)

       IF (workp%tag .EQ. self%mggravTag) THEN



                   !### lower X ###
                   IF (lx_bound) THEN

                       DO k = 1-workp%nb, workp%nz+workp%nb
                           DO j = 1-workp%nb, workp%ny+workp%nb
                                DO i = 1-workp%nb, 0

                                    pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                                    pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                                    pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                    workp%grid3d(i,j,k,1) = phi_out(self%a1,self%a3,self%e,self%rho,pos(1),pos(2),pos(3))        
                                END DO
                            END DO
                        END DO

                   END IF

                   !### upper X ###
                   IF (hx_bound) THEN

                       DO k = 1-workp%nb, workp%nz+workp%nb
                           DO j = 1-workp%nb, workp%ny+workp%nb
                                DO i = workp%nx+1, workp%nx+workp%nb

                                    pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                                    pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                                    pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                    workp%grid3d(i,j,k,1) = phi_out(self%a1,self%a3,self%e,self%rho,pos(1),pos(2),pos(3))        
                                END DO
                            END DO
                        END DO

                   END IF

                   !### lower Y ###
                   IF (ly_bound) THEN


                       DO k = 1-workp%nb, workp%nz+workp%nb
                           DO j = 1-workp%nb, 0
                                DO i = 1-workp%nb, workp%nx+workp%nb

                                    pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                                    pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                                    pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                    workp%grid3d(i,j,k,1) = phi_out(self%a1,self%a3,self%e,self%rho,pos(1),pos(2),pos(3))        
                                END DO
                            END DO
                        END DO

                   END IF

                   !### upper Y ###
                   IF (hy_bound) THEN

                       DO k = 1-workp%nb, workp%nz+workp%nb
                           DO j = workp%ny+1, workp%ny+workp%nb
                                DO i = 1-workp%nb, workp%nx+workp%nb

                                    pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                                    pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                                    pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                    workp%grid3d(i,j,k,1) = phi_out(self%a1,self%a3,self%e,self%rho,pos(1),pos(2),pos(3))        
                                END DO
                            END DO
                        END DO

                   END IF

                   !### lower Z ###
                   IF (lz_bound) THEN

                       DO k = 1-workp%nb, 0
                           DO j = 1-workp%nb, workp%ny+workp%nb
                                DO i = 1-workp%nb, workp%nx+workp%nb

                                    pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                                    pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                                    pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                    workp%grid3d(i,j,k,1) = phi_out(self%a1,self%a3,self%e,self%rho,pos(1),pos(2),pos(3))        
                                END DO
                            END DO
                        END DO

                   END IF

                   !### upper Z ###
                   IF (hz_bound) THEN

                       DO k = workp%nz+1, workp%nz+workp%nb
                           DO j = 1-workp%nb, workp%ny+workp%nb
                                DO i = 1-workp%nb, workp%nx+workp%nb

                                    pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
                                    pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
                                    pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                    workp%grid3d(i,j,k,1) = phi_out(self%a1,self%a3,self%e,self%rho,pos(1),pos(2),pos(3))        
                                END DO
                            END DO
                        END DO

                   END IF

                !### loop through the grid and find out the difference between
                !the calculated potential and the analytical one

!                DO k = 1 - workp%nb,workp%nz+workp%nb

!                        DO j = 1 - workp%nb,workp%ny+workp%nb

!                                DO i = 1 - workp%nb,workp%nx+workp%nb
       
!                                        pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx
!                                        pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx
!                                        pos(3) = workp%z0 + REAL((k-1),WRD)*workp%dx

                                        
                                        ! find the parameter position, lessthan/equalto 1 is inside the
                                        ! spheriod, greater than one is outside it
!                                        p = (pos(1)**2 + pos(2)**2)/self%a1**2 + pos(3)**2/self%a3**2

!                                        IF (p .LE. 1.0_WRD) THEN
                                        
!                                                analyticalPhi = phi_in(self%a1,self%a3,self%e,self%rho,pos(1), pos(2), pos(3))
!                                                calculatedPhi = workp%grid3d(i,j,k,1)

!                                        ELSE

!                                                analyticalPhi = phi_out(self%a1,self%a3,self%e,self%rho,pos(1), pos(2), pos(3))
!                                                calculatedPhi = workp%grid3d(i,j,k,1)
                                                
!                                        ENDIF

!                                        error = calculatedPhi - analyticalPhi

!                                END DO
                
!                        END DO

!               END DO

         END IF


END SUBROUTINE grav_spheriod_bounds

    REAL(WRD) FUNCTION phi_in(a1,a3,e,rho,x,y,z)
        REAL(WRD), INTENT(IN) :: a1, a3, rho,e,x,y,z
        REAL(WRD) :: pi, G, bA1, bA3
        G = 1.0_WRD
        pi = 2.0_WRD*DASIN(1.0_WRD)

        bA1 = bA1_calc(e)
        bA3 = bA3_calc(e)

        phi_in = -pi*G*rho*(bA1*(a1**2 - x**2) + bA1*(a1**2 - y**2)&
                        + bA3*(a3**2 - z**2))

        RETURN
    END FUNCTION

    REAL(WRD) FUNCTION phi_out(a1,a3,e,rho,x,y,z)
        REAL(WRD), INTENT(IN) :: a1,a3,e,rho,x,y,z
        REAL(WRD) :: pi, G, h, lambda
        
        G = 1.0_WRD
        pi = 2.0_WRD*DASIN(1.0_WRD)

        lambda = lambda_calc(a1,a3,x,y,z)

        h = h_calc(a1,e,a3,lambda)

        phi_out = 2.0_WRD*pi*G*rho*a1*a3/e * (DATAN(h) - 1.0_WRD/(2.0_WRD*a1**2*e**2)*((x**2+y**2)*(DATAN(h)-h/(1.0_WRD+h**2)) &
                 + 2.0_WRD*z**2*(h-DATAN(h))))

        RETURN     

    END FUNCTION


    REAL(WRD) FUNCTION lambda_calc(a1,a3,x,y,z)
        REAL(WRD), INTENT(IN) :: a1,a3,x,y,z
        REAL(WRD) :: b, c

        b = (a1**2+a3**2-x**2-y**2-z**2)
        c = a1**2*a3**2 - a3**2*(x**2+y**2) - a1**2*z**2

        lambda_calc =  -0.5_WRD*b+0.5_WRD*SQRT(b**2 - 4.0_WRD*c)
        
        RETURN
    END FUNCTION


    REAL(WRD) FUNCTION h_calc(a1,e,a3,lambda)
        REAL(WRD), INTENT(IN) :: a1, e, a3, lambda

        h_calc = a1*e/SQRT(a3**2+lambda)
        RETURN
    END FUNCTION


    REAL(WRD) FUNCTION bA1_calc(e)
        REAL(WRD), INTENT(IN) :: e

        bA1_calc = SQRT(1.0_WRD-e**2)/e**3 * DASIN(e) - (1.0_WRD-e**2)/e**2
        RETURN        

    END FUNCTION

    REAL(WRD) FUNCTION bA3_calc(e)
        REAL(WRD), INTENT(IN) :: e

        bA3_calc = 2.0_WRD/e**2 - 2.0_WRD*SQRT(1-e**2)*DASIN(e)/e**3
        RETURN
    END FUNCTION

    REAL(WRD) FUNCTION e_calc(a1,a3)
        REAL(WRD), INTENT(IN) :: a1, a3

        e_calc = SQRT(1.0_WRD-(a3/a1)**2)
        RETURN
    END FUNCTION

END MODULE Mod_Problem

#endif 
! GRAVITY_SUITE
