#include "../macro.h"

#ifdef KELVIN_HELMHOLTZ

MODULE Mod_Problem

! FILENAME: mod_KelvinHelmholtz_tests.f90
!
! DESCRIPTION: Initial Conditions for the Kelving Helmholtz test in 2D,
! see Lecoanet et al 2015 for details.
!
! DEPENDENCIES: Mod_Globals for defining data types
!
! HISTORY:
!    8/16/2017  - Julius Donnert
!    9/14/2015  - Pete Mendygral

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_ProblemBase
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Patch

IMPLICIT NONE

PRIVATE
PUBLIC :: Problem

TYPE, EXTENDS(ProblemBase) :: Problem

    PRIVATE

    REAL(WRD), PUBLIC :: gamma 
    REAL(WRD)         :: amplitude, pressure, a, sigma, z1, z2, ux, density_ratio, vx_boost

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: patch_init
    PROCEDURE, PUBLIC :: patch_bounds

END TYPE Problem

INTERFACE Problem

    MODULE PROCEDURE constructor

END INTERFACE Problem

CONTAINS

FUNCTION constructor(param, rankloc, simunits) RESULT(self)

    TYPE(Problem) :: self
    TYPE(Parameters) :: param
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits

    CALL param%section("Initial Conditions:  Kelving Helmholtz test (Lecoanet et al 2015)")

    self%gamma         = param%getFlt("gamma",          "1.66666666666667d0", "# Adiabatic index")
    self%amplitude     = param%getFlt("amplitude",      "0.01d0",             "# eq. 8c")
    self%pressure      = param%getFlt("pressure",       "10d0",               "# eq. 8d")
    self%a             = param%getFlt("a",              "0.05d0",             "# resolution dep. eq. 8a,b,c,e")
    self%sigma         = param%getFlt("sigma",          "0.2d0",              "# resolution dep. eq. 8c")
    self%z1            = param%getFlt("z1",             "-0.5d0",             "# eq. 8a,b,c,e")
    self%z2            = param%getFlt("z2",             "0.5d0",              "# eq. 8a,b,c,e")
    self%ux            = param%getFlt("uflow",          "1.0d0",              "# velocity parameter eq. 8b")
    self%density_ratio = param%getFlt("density_ratio",  "1.0d0",              "# Stratification parameter ")
    self%vx_boost      = param%getFlt("vx_boost",       "0.0d0",              "# boost velocity in x direction")

    RWARN(simunits%boxsize(1) .NE. 1.0_WRD, "boxsize(1) = 1 recommended")
    RWARN(simunits%boxsize(2)/simunits%boxsize(1) .NE. 2.0_WRD, "Boxsize(2) = 2 * Boxsize(1) recommended")

    RETURN

END FUNCTION constructor

SUBROUTINE destroy(self)

    CLASS(Problem) :: self
    
    RETURN

END SUBROUTINE destroy

! the routine called to initialize every Patch 
SUBROUTINE patch_init(self, workp, rankloc, simunits, is_boundary, tid, ierr)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
    INTEGER(WIS), INTENT(IN) :: tid
    INTEGER(WIS), INTENT(OUT) :: ierr
    REAL(WRD)    :: pos(2), a, z1, z2, sigma
    INTEGER(WIS) :: i, j, k

    a  = self%a
    z1 = self%z1
    z2 = self%z2
    sigma = self%sigma

    ASSERT(workp%is2d .EQ. .TRUE., "Kelvin Helmholtz ICs need ndim = 2 in parameter file.")


    ! vary though the grid along X only 
       DO j = 1-workp%nb, workp%ny+workp%nb
   
           DO i = 1-workp%nb, workp%nx+workp%nb
   
               pos(1) = workp%x0 + REAL((i-1),WRD)*workp%dx + 0.5_WRD * simunits%boxsize(1)
               pos(2) = workp%y0 + REAL((j-1),WRD)*workp%dx + 0.5_WRD * simunits%boxsize(2)

               workp%grid2d(i,j,1)  = 1.0_WRD + self%density_ratio * 0.5_WRD * (TANH((pos(2) - z1)/ a) & !eq 8a
                                    - TANH((pos(2) - z2) / a))    
               workp%grid2d(i,j,2)  = workp%grid2d(i,j,1) * self%ux * (TANH((pos(2) - z1) / a) &         !eq 8b
                                    - TANH((pos(2) - z2) / a) - 1.0_WRD)  &
                                    + workp%grid2d(i,j,1) * self%vx_boost                                ! boost
               workp%grid2d(i,j,3)  = workp%grid2d(i,j,1) * self%amplitude * SIN(2.0_WRD * simunits%pi * pos(1)) & !eq 8c
                                    * (EXP(-(pos(2) - z1)**2 / sigma**2) &
                                     + EXP(-(pos(2) - z2)**2 / sigma**2))
               workp%grid2d(i,j,4)  = 0.0_WRD
               workp%bface2d(i,j,1) = 0.0_WRD
               workp%bface2d(i,j,2) = 0.0_WRD
               workp%grid2d(i,j,7)  = 0.0_WRD

               IF (workp%passOn .AND. workp%npass .GT. 0) THEN

                   workp%pass2d(i,j,:) = 0.5_WRD * (TANH((pos(2) - self%z2) / self%a) &
                                                  - TANH((pos(2) - self%z1) / self%a) + 2.0_WRD) ! eq 8e

               END IF
   
           END DO
   
       END DO
   
       ! loop back through and set Bx/y and energy 
       DO j = 1, workp%ny
   
           DO i = 1, workp%nx

               workp%grid2d(i,j,5) = 0.5_WRD * (workp%bface2d(i,j,1) + workp%bface2d(i-1,j,1))
               workp%grid2d(i,j,6) = 0.5_WRD * (workp%bface2d(i,j,2) + workp%bface2d(i,j-1,2))

               workp%grid2d(i,j,8) = self%pressure / (self%gamma - 1.0_WRD) &
                                   + (0.5_WRD / workp%grid2d(i,j,1)) * (workp%grid2d(i,j,2)**2 + workp%grid2d(i,j,3)**2 &
                                                                      + workp%grid2d(i,j,4)**2) + &
                                      0.5_WRD * (workp%grid2d(i,j,5)**2 + workp%grid2d(i,j,6)**2 + workp%grid2d(i,j,7)**2)
           END DO

       END DO

    ierr = 0

END SUBROUTINE patch_init

! the routine called to set boundary values (called for every Patch even if it is not a world boundary) 
SUBROUTINE patch_bounds(self, workp, rankloc, simunits, is_boundary, maxsignal, tid, ierr)

    CLASS(Problem) :: self

    TYPE(Patch) :: workp
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits

    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
    REAL(WRD), INTENT(INOUT) :: maxsignal(:)
    INTEGER(WIS), INTENT(IN) :: tid
    INTEGER(WIS), INTENT(OUT) :: ierr

    maxsignal(1) = MAX(maxsignal(1), 0.0_WRD)
    maxsignal(2) = MAX(maxsignal(2), 0.0_WRD)
    maxsignal(3) = MAX(maxsignal(3), 0.0_WRD)

    ierr = 0

END SUBROUTINE patch_bounds

END MODULE Mod_Problem

#endif 
! KELVIN_HELMHOLTZ
