#include "../macro.h"

#ifdef WENO_SUITE

MODULE Mod_Problem

!   FILENAME: Weno_Suite.f90
!
!   DESCRIPTION: This module defines a number of test problems
!   used in the Weno-Wombat paper.
!
!   DEPENDENCIES: Mod_Globals for defining data types
! 
!   REFERENCES: Donnert et al. 2018, Yang et al. 2018
! 
!   HISTORY:
!       18/12/2017 - Julius Donnert

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Patch
USE Mod_Decomposition
USE Mod_ProblemBase

IMPLICIT NONE

PRIVATE

PUBLIC :: Problem

TYPE, EXTENDS(ProblemBase) :: Problem

    PRIVATE

    REAL(WRD), PUBLIC :: gamma
    REAL(WRD) :: boxsize(3), vboost(3)
    
    CHARACTER(WCHAR_BUF_SIZE) :: init_routine, bounds_routine

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: destroy
    PROCEDURE, PUBLIC :: patch_init
    PROCEDURE, PUBLIC :: patch_bounds

    PROCEDURE :: rj1a_1d
    PROCEDURE :: shock_dwave_1d
    PROCEDURE :: blob_test_2d
    PROCEDURE :: blob_test_3d
    PROCEDURE :: blob_test_Agertz_3d
    PROCEDURE :: square_adv_2d
    PROCEDURE :: implosion_2d
    PROCEDURE :: gresho_vortex_2d ! Liska & Wendroff 2003

    PROCEDURE :: interpolateB_center2face_2d
    PROCEDURE :: interpolateB_center2face_3d

END TYPE Problem

INTERFACE Problem

    MODULE PROCEDURE constructor

END INTERFACE Problem

CONTAINS ! methods
    
FUNCTION constructor(param, rankloc, simunits) RESULT(self)

    TYPE(Problem) :: self

    TYPE(Parameters) :: param
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
     
    RMSG("Problem: WENO Suite")
    RMSG(" ")

    CALL param%section("Problem: WENO Suite")
    CALL param%comment("Available init routines: Blob_Test Blob_Test_Agertz ES RJ1a ABC GreshoVortex ")
    CALL param%comment("                         Square_Advection ShockDWave ")

    self%init_routine = param%getStr("init_routine ", "Blob_Test", "# Name of problem")
    self%gamma        = param%getFlt("gamma ", "1.666666666666667d0", "# Adiabatic index", 0, .FALSE.)
    self%vboost(1)    = param%getFlt("vxboost", "0d0", "# Boost velocity in x-direction", 0, .FALSE.)
    self%vboost(2)    = param%getFlt("vyboost", "0d0", "# Boost velocity in y-direction", 0, .FALSE.)
    self%vboost(3)    = param%getFlt("vzboost", "0d0", "# Boost velocity in z-direction", 0, .FALSE.)

    SELECT CASE (TRIM(self%init_routine))
        
        CASE("Blob_Test_Agertz")

            ASSERT(simunits%boxsize(1).EQ.2000, "boxsize(1) = 2000 required, with ny=nz=nx/4.")
            ASSERT(simunits%boxsize(2).EQ.2000, "boxsize(2) = 2000 required, with ny=nz=nx/2.")
            ASSERT(simunits%boxsize(3).EQ.8000, "boxsize(3) = 8000 required, with ny=nz=nx/2.")

            RMSG("Problem: Blob test from Agertz 2007, Sijacki 2012")

        CASE("Blob_Test")

            WARN(simunits%boxsize(1).NE.4, "boxsize = 4 recommended, with ny=nz=nx/4.")

            RMSG("Problem: Blob test from Jang et al. 2018")

        CASE ("RJ1a")

            RMSG("Problem: 1D Shock Tube test no 1a from Ryu & Jones 1994")

        CASE ("ES")
            
            RMSG("Problem: 1D ES test Ryu 1998")

        CASE ("ABC")

            RMSG("Problem: ABC test from Gamer paper.")

        CASE ("Square_Advection")
            
            RMSG("Problem: 2D Square Advection test from Schaal+ 2015.")

            ASSERT(simunits%boxsize(1) .EQ. 1.0d0, "boxsize(1) = 1 required.")
            ASSERT(simunits%boxsize(2) .EQ. 1.0d0, "boxsize(2) = 1 required.")

            RMSG("Setting gamma = 7/5, ignoring parameter file value.")
            self%gamma = 7.0_WRD/5.0_WRD

        CASE ("Implosion")

            RMSG("Problem: 2D Implosion test from Sijacki+ 2012.")

            WARN(simunits%boxsize(1) .NE. 0.3d0, "boxsize(1) = 0.3 required.")
            WARN(simunits%boxsize(2) .NE. 0.3d0, "boxsize(2) = 0.3 required.")

            RMSG("Setting gamma = 1.4, ignoring parameter file value.")
            self%gamma = 1.4_WRD

        CASE ("ShockDWave")

            RMSG(" ")
            RMSG("Problem: 2D Density wave shock interaction.")

            WARN(simunits%boxsize(1) .NE. 10.0d0, "boxsize(1) = 10.0 required.")

            RMSG("Setting gamma = 1.4, ignoring parameter file value.")
            self%gamma = 1.4_WRD

        CASE ("GreshoVortex")

            RMSG(" ")
            RMSG("Problem: 2D Gresho Vortex test from Liska & Wendroff 2003.")
           
            ASSERT(simunits%boxsize(1) .GE. 1.0d0, "boxsize(1) >= 1 required.")
            ASSERT(simunits%boxsize(2) .GE. 1.0d0, "boxsize(2) >= 1 required.")

            RMSG("Setting gamma = 5/3, ignoring parameter file value.")
            self%gamma = 5.0_WRD / 3.0_WRD

        CASE DEFAULT
            RASSERT(.FALSE., "Problem "//TRIM(self%init_routine)//" not found. See parameter file.")

    END SELECT
    
    self%boxsize(:) = simunits%boxsize(:)
    

    RETURN

END FUNCTION constructor

SUBROUTINE destroy(self)

    CLASS(Problem) :: self
    
    RETURN

END SUBROUTINE destroy

! the routine called to initialize every Patch 
SUBROUTINE patch_init(self, workp, rankloc, simunits, is_boundary, tid, ierr)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits
    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
    INTEGER(WIS), INTENT(IN) :: tid
    INTEGER(WIS), INTENT(OUT) :: ierr

    SELECT CASE (TRIM(self%init_routine))
        
        CASE("Blob_Test")
            
            ASSERT(workp%ndim .GE. 2, "Blob_Test requires ndim >= 2 in the parameter file.")

            IF (workp%ndim .EQ. 2) CALL self%blob_test_2d(workp)
            IF (workp%ndim .EQ. 3) CALL self%blob_test_3d(workp)

        CASE("Blob_Test_Agertz")
            
            ASSERT(workp%ndim .EQ. 3, "Blob_Test_Agertz requires ndim = 3 in the parameter file.")

            CALL self%blob_test_Agertz_3d(workp)
                
        CASE ("RJ1a")
            
            ASSERT(workp%ndim .EQ. 1, "RJ1a requires ndim = 1 in the parameter file")

            CALL self%rj1a_1d(workp)

        CASE ("ABC")


        CASE ("Square_Advection")
            
            ASSERT(workp%ndim .EQ. 2, "Square_Advection requires ndim = 2 in the parameter file")

            CALL self%square_adv_2d(workp)

        CASE ("Implosion")

            ASSERT(workp%ndim .EQ. 2, "Implosion requires ndim=2 in the parameter file.")

            CALL self%implosion_2d(workp)

        CASE ("ShockDWave")

            ASSERT(workp%ndim .EQ. 1, "ShockDWave requires ndim=1 in the parameter file.")

            CALL self%shock_dwave_1d(workp)

        CASE ("GreshoVortex")

            ASSERT(workp%ndim .EQ. 2, "ShockDWave requires ndim=1 in the parameter file.")

            CALL self%gresho_vortex_2d(workp)

        CASE DEFAULT
            RASSERT(.FALSE., "Problem name "//TRIM(self%init_routine)//" not found.")

     END SELECT

    ierr = 0

END SUBROUTINE patch_init

! the routine called to set boundary values (called for every Patch even if it is not a world boundary) 
SUBROUTINE patch_bounds(self, workp, rankloc, simunits, is_boundary, maxsignal, tid, ierr)

    CLASS(Problem) :: self
    
    TYPE(Patch) :: workp
    TYPE(RankLocation) :: rankloc
    TYPE(SimulationUnits) :: simunits

    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
    INTEGER(WIS), INTENT(IN) :: tid

    INTEGER(WIS), INTENT(OUT) :: ierr

    REAL(WRD), INTENT(INOUT) :: maxsignal(:)

    maxsignal(1) = MAX(maxsignal(1), 0.0_WRD)
    maxsignal(2) = MAX(maxsignal(2), 0.0_WRD)
    maxsignal(3) = MAX(maxsignal(3), 0.0_WRD)

    SELECT CASE (TRIM(self%init_routine))

        CASE("Blob_Test")
            CALL self%open_bounds(workp, is_boundary) ! blob test treats emf differently
    
        CASE ("RJ1a")
            CALL self%open_bounds(workp, is_boundary)

        CASE ("ShockDWave")
            CALL self%open_bounds(workp, is_boundary)

        CASE("Blob_Test_Agertz")
            ! periodic
        CASE ("Square_Advection")
            ! periodic
        CASE ("ABC")
            ! periodic
        CASE ("Implosion")
            ! periodic

        CASE ("GreshoVortex")
            ! periodic

        CASE DEFAULT
            RASSERT(.FALSE., "Boundary routine for "//TRIM(self%init_routine)//" not found.")

    END SELECT

    ierr = 0

END SUBROUTINE patch_bounds

SUBROUTINE rj1a_1d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp

    INTEGER(WID) :: i

    REAL(WRD), PARAMETER :: UL(8) = (/ 1.0_WRD, 10.0_WRD, 0.0_WRD, 0.0_WRD, & 
                                       5.0_WRD/SQRT(4.0_WRD*PI),  5.0_WRD/sqrt(4.0_WRD*PI), 0.0_WRD, 20.0_WRD/)
                                
    REAL(WRD), PARAMETER :: UR(8) = (/ 1.0_WRD, -10.0_WRD, 0.0_WRD, 0.0_WRD, &
                                       5.0_WRD/SQRT(4.0_WRD*PI),  5.0_WRD/sqrt(4.0_WRD*PI), 0.0_WRD, 1.0_WRD /)

    DO i = 1-workp%nb, workp%nx+workp%nb
            
        IF (i .le. workp%nx/2+1) THEN
            workp%grid1d(i,:) = UL
        ELSE
            workp%grid1d(i,:) = UR
        END IF
    END DO

    CALL self%state2conserved_1d(self%gamma, workp%grid1d)

END SUBROUTINE rj1a_1d

! Borges et al. 2008 test for WENO-Z
SUBROUTINE shock_dwave_1d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp

    INTEGER(WID) :: i
    REAL(WRD)    :: x

    REAL(WRD), PARAMETER :: UL(8) = (/ 3.857143_WRD, 2.629369_WRD,0.0_WRD,0.0_WRD, &
                                       0.0_WRD,0.0_WRD,0.0_WRD, 31.0_WRD/3.0_WRD  /)
    REAL(WRD), PARAMETER :: UR(8) = (/ 1.0_WRD, 0.0_WRD,0.0_WRD,0.0_WRD, 0.0_WRD,0.0_WRD,0.0_WRD, 1.0_WRD /)
    REAL(WRD), PARAMETER :: k = 5

    DO i = 1-workp%nb, workp%nx+workp%nb

        x = workp%x0 + REAL(i-1,WRD)*workp%dx

        IF (x .LT. -4.0_WRD) THEN
            workp%grid1d(i,:) = UL(:)
        ELSE
            workp%grid1d(i,:) = UR(:)

            workp%grid1d(i,1) = workp%grid1d(i,1) + 0.2_WRD * SIN(k*x)
        END IF

    END DO

    CALL self%state2conserved_1d(self%gamma, workp%grid1d)

END SUBROUTINE shock_dwave_1d

! Liska & Wendroff 2003, tests angular momentum conservation
SUBROUTINE gresho_vortex_2d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp

    REAL(WRD) :: x, y, r, v, p, phi

    INTEGER(WID) :: i, j

    DO j = 1-workp%nb, workp%ny+workp%nb

        y = workp%y0 + REAL(j-1,WRD)*workp%dx

        DO i = 1-workp%nb, workp%nx+workp%nb

            x = workp%x0 + REAL(i-1,WRD)*workp%dx

            r = SQRT(x**2 + y**2)

            IF (r .LE. 0.2_WRD) THEN

                v = 5.0_WRD * r
                p = 5.0_WRD + 25.0_WRD/2.0_WRD * r**2

            ELSE IF (r .LT. 0.4_WRD) THEN

                v = 2.0_WRD - 5.0_WRD * r
                p = 9.0_WRD + 25.0_WRD/2.0_WRD * r**2 - 20.0_WRD * r &
                            + 4.0_WRD * LOG(r/0.2_WRD)

            ELSE
 
                v = 0.0_WRD
                p = 3.0_WRD + 4.0_WRD * LOG(2.0_WRD)
            END IF

            phi = ATAN2(x,y) ! make arctan inversion

            workp%grid2d(i,j,2) = -v * COS(phi) + self%vboost(1)
            workp%grid2d(i,j,3) = v * SIN(phi) + self%vboost(2)
            workp%grid2d(i,j,8) = p
        END DO

    END DO

    workp%grid2d(:,:,1)   = 1.0_WRD ! constant density everywhere
    workp%grid2d(:,:,4:7) = 0.0_WRD ! 2D hydro problem
    workp%bface2d(:,:,:)  = 0.0_WRD

    CALL self%state2conserved_2d(self%gamma, workp%grid2d)

END SUBROUTINE gresho_vortex_2d

! Blob test from Jang et al 2018
SUBROUTINE blob_test_2d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    
    REAL(WRD) :: x, y, r, v2,B2

    INTEGER(WID) :: i, j

    ! blob test states: [rho,vx,vy,vz,Bx,By,Bz,Pressure]

    REAL(WRD), PARAMETER :: UL(8) = (/2.11820_WRD,13.69836_WRD,0.0_WRD,0.0_WRD, &
                                      0.0_WRD,19.33647_WRD,19.33647_WRD,65.88816_WRD/)
    REAL(WRD), PARAMETER :: UR(8) = (/1.0_WRD,0.0_WRD,0.0_WRD,0.0_WRD,0.0_WRD, &
                                      9.12871_WRD,9.12871_WRD,1.0_WRD/)
    REAL(WRD), PARAMETER :: R0 = 0.12_WRD, R1 = 0.19_WRD

    DO j = 1-workp%nb, workp%ny+workp%nb
        
        y = workp%y0 + REAL(j-0.5_WRD,WRD)*workp%dx + 0.5_WRD * self%boxsize(2)
        
        DO i = 1-workp%nb, workp%nx+workp%nb
        
            x = workp%x0 + REAL(i-0.5_WRD,WRD)*workp%dx + 0.5_WRD * self%boxsize(1)

            r = SQRT((x-1.5_WRD)**2 + (y-0.5_WRD)**2) ! place a circle @ 1.5,0.5

            IF (x .LE. 1.2_WRD) THEN
                workp%grid2d(i,j,:) = UL(:)
            ELSE
                workp%grid2d(i,j,:) = UR(:)
            END IF

            IF (r .LE. R0) THEN
                workp%grid2d(i,j,1) = 10.0_WRD
            END IF
            
            IF ((r .GT. R0) .AND. (r .LE. R1)) THEN
                workp%grid2d(i,j,1) = 1.0_WRD + 9.0_WRD * (R1 - r) / (R1 - R0)
            END IF

        END DO
    END DO

    ! face centered fields

    DO j = 1-workp%nb, workp%ny+workp%nb
        
        y = workp%y0 + REAL(j-1+0.5,WRD)*workp%dx + 0.5_WRD * self%boxsize(2)
        
        DO i = 1-workp%nb, workp%nx+workp%nb

            x = workp%x0 + REAL(i-1+0.5,WRD)*workp%dx + 0.5_WRD * self%boxsize(1)

            r = SQRT((x-1.5_WRD)**2 + (y-0.5_WRD)**2)

            IF (x .LE. 1.2_WRD) THEN
                workp%bface2d(i,j,1:2) = UL(5:6)
            ELSE
                workp%bface2d(i,j,1:2) = UR(5:6)
            END IF

        END DO

    END DO
            
    CALL self%state2conserved_2d(self%gamma, workp%grid2d)

    !CALL self%interpolateB_center2face_2d(workp%grid2d, workp%bface2d)

END SUBROUTINE blob_test_2d

SUBROUTINE blob_test_3d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    
    REAL(WRD) :: x, y, z, r, v2,B2

    INTEGER(WID) :: i, j, k

    REAL(WRD), PARAMETER :: UL(8) = (/3.86859_WRD,11.2536_WRD,0.0_WRD,0.0_WRD, &
                                      0.0_WRD,2.1826182_WRD,-2.1826182_WRD,167.345_WRD/)
    REAL(WRD), PARAMETER :: UR(8) = (/1.0_WRD,0.0_WRD,0.0_WRD,0.0_WRD,0.0_WRD, &
                                      0.56418958_WRD,0.56418958_WRD,1.0_WRD/)

    REAL(WRD), PARAMETER :: R0 = 0.15_WRD, R1 = 0.19_WRD

    DO k = 1-workp%nb, workp%nz+workp%nb
        
        z = workp%z0 + REAL(k-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(3)

        DO j = 1-workp%nb, workp%ny+workp%nb
        
            y = workp%y0 + REAL(j-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(2)
        
            DO i = 1-workp%nb, workp%nx+workp%nb
        
                x = workp%x0 + REAL(i-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(1)

                r = SQRT((x-1.5_WRD)**2 + (y-0.5_WRD)**2 + (z-0.5_WRD)**2) ! place a circle @ 1.5,0.5
                
                IF (x .LE. 1.2_WRD) THEN
                    workp%grid3d(i,j,k,:) = UL(:)
                ELSE
                    workp%grid3d(i,j,k,:) = UR(:)
                END IF

                IF (r .LE. R0) THEN
                    workp%grid3d(i,j,k,1) = 10.0_WRD
                END IF
            
                IF ((r .GT. R0) .AND. (r .LE. R1)) THEN
                     workp%grid3d(i,j,k,1) = 1.0_WRD + 9.0_WRD * (R1 - r) / (R1 - R0)
                END IF

            END DO
        END DO
    END DO

    ! face centered fields

    DO k = 1-workp%nb, workp%nz+workp%nb
            
        z = workp%z0 + (REAL(k,WRD) - 1.5_WRD) *workp%dx + 0.5_WRD * self%boxsize(3)
    
        DO j = 1-workp%nb, workp%ny+workp%nb
        
            y = workp%y0 + (REAL(j,WRD) - 1.5_WRD)*workp%dx + 0.5_WRD * self%boxsize(2)
        
            DO i = 1-workp%nb, workp%nx+workp%nb

                x = workp%x0 + (REAL(i,WRD) - 1.5_WRD)*workp%dx + 0.5_WRD * self%boxsize(1)
    
                r = SQRT((x-1.5_WRD)**2 + (y-0.5_WRD)**2 + (z-0.5_WRD)**2)

                IF (x .LE. 1.2_WRD) THEN
                    workp%bface3d(i,j,k,1) = UL(5)
                    workp%bface3d(i,j,k,2) = UL(6)
                    workp%bface3d(i,j,k,3) = UL(7)
                ELSE
                    workp%bface3d(i,j,k,1) = UR(5)
                    workp%bface3d(i,j,k,2) = UR(6)
                    workp%bface3d(i,j,k,3) = UR(7)
                END IF

            END DO
        END DO

    END DO
            
    CALL self%state2conserved_3d(self%gamma, workp%grid3d)

END SUBROUTINE blob_test_3d

! Blob test following Agertz 2007
SUBROUTINE blob_test_Agertz_3d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    
    REAL(WRD) :: x, y, z, r, v2, B2, R1

    INTEGER(WID) :: i, j, k

    REAL(WRD), PARAMETER :: U(8) = (/ 3.13d-2,0.0d0,0.0d0,1000.0d0, &
                                      0.0d0,0.0d0, 0.0d0,4392.1829333333335d0 /)

    REAL(WRD), PARAMETER :: R0 = 197.0_WRD

    R1 = R0 + workp%dx ! soften by one zone

    DO k = 1-workp%nb, workp%nz+workp%nb
        
        z = workp%z0 + REAL(k-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(3)

        DO j = 1-workp%nb, workp%ny+workp%nb
        
            y = workp%y0 + REAL(j-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(2)
        
            DO i = 1-workp%nb, workp%nx+workp%nb
        
                x = workp%x0 + REAL(i-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(1)

                workp%grid3d(i,j,k,:) = U(:)

                r = SQRT((x-1000.0_WRD)**2 + (y-1000.0_WRD)**2 + (z-1000.0_WRD)**2) ! place a circle @  1000

                IF (r .LE. R0) THEN
    
                    workp%grid3d(i,j,k,1) = U(1) * 10.0_WRD
                    workp%grid3d(i,j,k,4) = 0.0_WRD

                ELSE IF ((r .GT. R0) .AND. (r .LE. R1)) THEN

                    workp%grid3d(i,j,k,1) = U(1) * (1.0_WRD + 9.0_WRD * (R1 - r) / (R1 - R0))
                    workp%grid3d(i,j,k,4)  = U(4) * (R1 - r) / (R1 - R0)
                END IF

                workp%bface3d(i,j,k,1) = U(5)
                workp%bface3d(i,j,k,2) = U(6)
                workp%bface3d(i,j,k,3) = U(7)
            END DO
        END DO
    END DO
            
    CALL self%state2conserved_3d(self%gamma, workp%grid3d)

END SUBROUTINE blob_test_Agertz_3d

SUBROUTINE square_adv_2d(self, workp)

    CLASS(Problem) :: self
    TYPE(Patch) :: workp
    
    REAL(WRD), PARAMETER :: Q(8)   = (/ 1.0_WRD, 100.0_WRD, 50.0_WRD, 0.0_WRD, &
                                        0.0_WRD, 0.0_WRD, 0.0_WRD, 2.5_WRD /)
    REAL(WRD), PARAMETER :: RHO_IN = 4.0_WRD  

    REAL(WRD) :: x, y
    INTEGER(WIS) :: i,j

    DO j = 1-workp%nb, workp%ny+workp%nb
    
        y = workp%y0 + REAL(j-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(2)
    
        DO i = 1-workp%nb, workp%nx+workp%nb

            x = workp%x0 + REAL(i-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(1)

            workp%grid2d(i,j,:) =  Q(:)

            IF (x.GT.0.25_WRD .AND. x.LT.0.75_WRD) THEN

                IF (y.GT.0.25_WRD .AND. y.LT.0.75_WRD) THEN
            
                    workp%grid2d(i,j,1) =  RHO_IN

                END IF
            END IF
        END DO
    END DO

    CALL self%state2conserved_2d(self%gamma, workp%grid2d)

    workp%bface2d(:,:,:) = 0.0_WRD

END SUBROUTINE square_adv_2d


! Interacting shock wave test with periodic boundaries 
! from e.g. Sijacki+ 2012. Highly non-linear, exposes any
! kind of noise.

SUBROUTINE implosion_2d(self, workp)

    CLASS(Problem) :: self

    TYPE(Patch) :: workp
    
    REAL(WRD), PARAMETER :: U1(8) = (/ 1.0_WRD, 0.0_WRD, 0.0_WRD, 0.0_WRD, &
                                        0.0_WRD, 0.0_WRD, 0.0_WRD, 1.0_WRD /)
    REAL(WRD), PARAMETER :: U2(8) = (/ 0.125_WRD, 0.0_WRD, 0.0_WRD, 0.0_WRD, &
                                        0.0_WRD, 0.0_WRD, 0.0_WRD, 0.14_WRD /)  
    REAL(WRD)    :: x, y
    INTEGER(WIS) :: i,j

    DO j = 1-workp%nb, workp%ny+workp%nb
    
        y = workp%y0 + REAL(j-1,WRD) * workp%dx + 0.5_WRD * self%boxsize(2)
    
        DO i = 1-workp%nb, workp%nx+workp%nb

            x = workp%x0 + REAL(i-1,WRD)*workp%dx + 0.5_WRD * self%boxsize(1)

            IF (REAL(x+y,4) .GT. 0.15_WRD) THEN 
                workp%grid2d(i,j,:) =  U1(:)
            ELSE
                workp%grid2d(i,j,:) =  U2(:)
            END IF

        END DO
    END DO

    CALL self%state2conserved_2d(self%gamma, workp%grid2d)

    workp%bface2d(:,:,:) = 0.0_WRD

END SUBROUTINE implosion_2d

! Fourth order interpolation of cell centered magnetic fields in q
! onto face centered magnetic fields emf(nx,ny,2)= bxb,byb
! This is NOT the correct way to set the magnetic field on the faces,
! they should be set directly by position into bface.

SUBROUTINE interpolateB_center2face_2d(self, q, b)

    CLASS(Problem) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN)  :: q(:,:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: b(:,:,:)

    INTEGER(WID) :: i,j

    b(:,:,:) = 0.0_WRD
    
    DO j = 2, SIZE(q,2)-2
        DO i = 2, SIZE(q,1)-2

            b(i,j,1) = (-q(i-1,j,5) + 9.0_WRD*q(i,j,5) + 9.0_WRD*q(i+1,j,5) - q(i+2,j,5))/16.0_WRD
            b(i,j,2) = (-q(i,j-1,6) + 9.0_WRD*q(i,j,6) + 9.0_WRD*q(i,j+1,6) - q(i,j+2,6))/16.0_WRD

        END DO
    END DO

END SUBROUTINE interpolateB_center2face_2d

SUBROUTINE interpolateB_center2face_3d(self, q, b)

    CLASS(Problem) :: self

    REAL(WRD), CONTIGUOUS, INTENT(IN) :: q(:,:,:,:)
    REAL(WRD), CONTIGUOUS, INTENT(OUT) :: b(:,:,:,:)

    INTEGER(WID) :: i,j,k

    b(:,:,:,:) = 0.0_WRD
    
    DO k = 2, SIZE(b,3)-2
        DO j = 2, SIZE(b,2)-2
            DO i = 2, SIZE(b,1)-2

            b(i,j,k,1) = (-q(i-1,j  ,k  ,5) + 9.0_WRD*q(i,j,k,5) + 9.0_WRD*q(i+1,j  ,k  ,5) - q(i+2,j  ,k  ,5))/16.0_WRD
            b(i,j,k,2) = (-q(i  ,j-1,k  ,6) + 9.0_WRD*q(i,j,k,6) + 9.0_WRD*q(i  ,j+1,k  ,6) - q(i  ,j+2,k  ,6))/16.0_WRD
            b(i,j,k,3) = (-q(i  ,j  ,k-1,7) + 9.0_WRD*q(i,j,k,7) + 9.0_WRD*q(i  ,j  ,k+1,7) - q(i  ,j  ,k+2,7))/16.0_WRD

            END DO
        END DO
    END DO

END SUBROUTINE interpolateB_center2face_3d


END MODULE Mod_Problem

#endif

