#include "macro.h"

MODULE Mod_RankLocation

! FILENAME: rank_location.f90
!
! DESCRIPTION: Calculate rank location on the root grid
!
! DEPENDENCIES: Mod_Globals for defining data types
!
! HISTORY:
!    5/19/16  - Peter Mendygral

! require the needed modules 
USE Mod_Globals
USE Mod_Error

IMPLICIT NONE

PRIVATE
PUBLIC :: RankLocation

TYPE :: RankLocation

    PRIVATE

    INTEGER(WIS), PUBLIC :: izone_origin(3), &   ! world grid zone integer location of this rank's root domain origin
                            my_coords(3),    &   ! this rank's location in the MPI Cartesian communicator
                            nranks(3),       &   ! number of ranks in each dimension  
                            npatches(3),     &   ! number of root domain Patches in each dimension within a MPI rank
                            patch_size(3)        ! number of zones in each dimension for a root domain Patch
    REAL(WRD), PUBLIC ::    rzone_origin(3)      ! world grid zone real location of this rank's root domain origin (code units)
    LOGICAL(WIS), PUBLIC :: is_boundary(2,3)     ! true if for given dimension and edge this rank is a world grid boundary

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: calc_location

END TYPE RankLocation

INTERFACE RankLocation
   
   MODULE PROCEDURE constructor
   
END INTERFACE

CONTAINS ! object methods
  
FUNCTION constructor(my_coords, is_boundary, nranks, npatches, patch_size) RESULT(self)
  
    TYPE(RankLocation) ::  self
    INTEGER(WIS), INTENT(IN) :: my_coords(3), nranks(3), npatches(3), patch_size(3)
    LOGICAL(WIS), INTENT(IN) :: is_boundary(2,3)
     
    self%my_coords   = my_coords ! store these values as we will need them to initialize
    self%is_boundary = is_boundary
    self%nranks      = nranks
    self%npatches    = npatches
    self%patch_size  = patch_size

    RETURN
  
END FUNCTION constructor

! given dx compute everything about this rank's world grid location 
SUBROUTINE calc_location(self, dx)

    CLASS(RankLocation) :: self
    REAL(WRD), INTENT(IN) :: dx
    INTEGER(WIS) :: i

    DO i = 1, 3
        
        self%izone_origin(i) = self%my_coords(i) * self%npatches(i) * self%patch_size(i) + 1

        self%rzone_origin(i) = dx * self%my_coords(i) * self%npatches(i) * self%patch_size(i) &
                               - 0.5_WRD * dx * (self%nranks(i) * self%npatches(i) * self%patch_size(i)) &
                               + 0.5_WRD * dx 
    END DO

END SUBROUTINE calc_location

END MODULE Mod_RankLocation
