#include "macro.h"

MODULE Mod_SimulationUnits

! FILENAME: simulation_units.f90
!
! DESCRIPTION: Define constants & units relative to CGSM. 
! Note that we set the boxsize only along the x-direction and infer the other boxsizes from
! the number of cells in the other directions, i.e. cells are assumed always cubic.
!
! DEPENDENCIES: Mod_Globals for defining data types
!
! HISTORY:
!    11/08/2017  - Julius Donnert
!    05/19/16  - Peter Mendygral

USE Mod_Globals
USE Mod_Error
USE Mod_Parameters
USE Mod_RankLocation

IMPLICIT NONE

REAL(WID), PARAMETER :: GRAV = 6.67408d-8 ! Newtonian constant of gravitation [cm^3/g/s^2] (physics.nist.gov)

PRIVATE

PUBLIC :: SimulationUnits

TYPE :: SimulationUnits

    PRIVATE

    REAL(WRD), PUBLIC :: boxsize(3),          &       ! length of the world grid along each dimension in code units
                         dx,                  &       ! zone size in code units for the root domain 
                         dt,                  &       ! time step size in code units for the root domain 
                         t,                   &       ! current time in code units
                         t_beg,               &       ! initial simulation time in code units
                         t_end,               &       ! ending time in code units

                         boxsize_cgs(3),      &       ! length of the world grid along each dimension in CGS units
                         dx_cgs,              &       ! zone size in CGS units for the root domain
                         dt_cgs,              &       ! time step size in CGS units
                         t_cgs,               &       ! time in CGS units
                         t_beg_cgs,           &       ! initial time in CGS units
                         t_end_cgs,           &       ! ending time in CGS units

                         length_cgs,          &       ! conversion factor to CGS units for length
                         mass_cgs,            &       ! conversion factor to CGS units for mass
                         velocity_cgs,        &       ! conversion factor to CGS units for velocity
                         time_cgs,            &       ! conversion factor to CGS units for time
                         mass_density_cgs,    &       ! conversion factor to CGS units for mass density
                         energy_cgs,          &       ! conversion factor to CGS units for energy
                         energy_density_cgs,  &       ! conversion factor to CGS units for energy density
                         magnetic_cgs,        &       ! conversion factor to CGS units for the magnetic field

                         cfl,                 &       ! the CFL parameter to maintain when determining the time step

                         pi,                  &       ! PI
                         grav                         ! gravitational constant in code units

    INTEGER(WIS), PUBLIC :: max_steps                 ! maximum number of time steps to take on the root grid

    CONTAINS

    PRIVATE

    PROCEDURE, PUBLIC :: update_code2cgs
    PROCEDURE, PUBLIC :: propose_timestep

END TYPE SimulationUnits

INTERFACE SimulationUnits
   
   MODULE PROCEDURE constructor
   
END INTERFACE

! object methods 
CONTAINS
  
FUNCTION constructor(param, rankloc) RESULT(self)

    TYPE(SimulationUnits) :: self
    CLASS(RankLocation) :: rankloc
    CLASS(Parameters) :: param

    CALL param%section("Simulation units and dimensions")

    self%mass_cgs      = param%getFlt("mass_cgs", "1.989d43", "# [10^10 Msol] Unit mass in CGSM units")
    self%length_cgs    = param%getFlt("length_cgs", "3.085677581d21", "# [kpc] Unit length in CGSM units")
    self%velocity_cgs  = param%getFlt("velocity_cgs", "1d5", "# [km/s] Unit length in CGSM units")

    self%boxsize(1) = param%getFlt("boxsize", "1d0", "# In x direction. y & z direction inferred from number of zones.")

    CALL param%section("Time integration parameters")

    self%t_beg      = param%getFlt("t_beg",     "0",         "# Initial time in code units")
    self%t_end      = param%getFlt("t_end",     "1",         "# Final time in code units")
    self%dt         = param%getFlt("dt0",       "1e-8",      "# Intial timestep size", 0, .TRUE.)
    self%max_steps  = param%getFlt("max_steps", "100000000", "# Maximum number of steps", 0, .TRUE.)
    self%cfl        = param%getFlt("cfl", "0.8",             "# Fraction of the CFL condition to maintain", 0, .TRUE.)

    ! constants
    self%pi   = 2.0_WRD * ASIN(1.0_WRD)
    
    ! derived parameters
    self%time_cgs           = self%length_cgs / self%velocity_cgs
    self%mass_density_cgs   = self%mass_cgs / self%length_cgs**3
    self%energy_cgs         = self%mass_cgs * self%length_cgs**2 / self%time_cgs**2
    self%energy_density_cgs = self%energy_cgs / self%length_cgs**3
    self%magnetic_cgs       = SQRT(4.0_WRD * self%pi * self%energy_density_cgs)

    self%dx = self%boxsize(1) / (rankloc%nranks(1) * rankloc%npatches(1) * rankloc%patch_size(1))
    self%boxsize(2) = self%dx * rankloc%nranks(2) * rankloc%npatches(2) * rankloc%patch_size(2)
    self%boxsize(3) = self%dx * rankloc%nranks(3) * rankloc%npatches(3) * rankloc%patch_size(3)

    RMSG(" ")
    RMSG("dx         = "//E2A(self%dx))
    RMSG("boxsize(1) = "//E2A(self%boxsize(1)))
    RMSG("boxsize(2) = "//E2A(self%boxsize(2)))
    RMSG("boxsize(3) = "//E2A(self%boxsize(3)))
    RMSG(" ")


    self%boxsize_cgs(1) = self%boxsize(1) * self%length_cgs
    self%boxsize_cgs(2) = self%boxsize(2) * self%length_cgs
    self%boxsize_cgs(3) = self%boxsize(3) * self%length_cgs

    self%t          = self%t_beg

    self%dx_cgs      = self%dx * self%length_cgs
    self%t_beg_cgs   = self%t_beg * self%time_cgs
    self%t_cgs       = self%t * self%time_cgs
    self%t_end_cgs   = self%t_end * self%time_cgs
 
    ! call our conversion for all variables that change with the simulation timestep
    CALL self%update_code2cgs()

    ! constants
    self%grav = GRAV / self%length_cgs**3 * self%mass_cgs * self%time_cgs

#ifdef WMBT_COMM_BENCHMARK
    RPRINT '("COMM BENCHMARK ENABLED: Running for ", I8, " total iterations")', simunits%max_steps
    simunits%t_end = 1e30 ! run only max_steps, t_end is ignored
#endif

    RETURN
  
END FUNCTION constructor

! update any CGS values we have with code values that vary 
SUBROUTINE update_code2cgs(self)

    CLASS(SimulationUnits) :: self

    self%t_cgs  = self%t * self%time_cgs
    self%dt_cgs = self%dt * self%time_cgs

END SUBROUTINE update_code2cgs

! Given an input signal speed, compute a proposed time step size 
! For the first two steps of the simulations, do an exponential ramp 
! from the base timestep to the calculated timestep as a precaution.

FUNCTION propose_timestep(self, nsteps, maxsignal) RESULT(newdt)

    CLASS(SimulationUnits) :: self
    INTEGER(WIS), INTENT(IN) :: nsteps
    REAL(WRD), INTENT(IN) :: maxsignal(:)
    REAL(WRD) :: dt_CFL, newdt

    dt_CFL = self%cfl * self%dx / (maxsignal(1) + maxsignal(2) + maxsignal(3)) ! Full 3D CFL criterion

    DEBUGMSG("CFL dt ="//E2A(dt_CFL)//" maxsignal = ("//F2A(maxsignal(1))//", " &
                       //F2A(maxsignal(2))//", "//F2A(maxsignal(3))//")")

    WARN(dt_CFL .LE. 1.0E-8_WRD, "Found tiny timestep from CFL criterion: "//E2A(dt_CFL))

    IF (nsteps .LE. 4) THEN

        newdt = dt_CFL * EXP(REAL(nsteps-5,WRD))

        DEBUGMSG("    Increasing dt to "//E2A(newdt))

    ELSE IF (dt_CFL .LT. self%dt) THEN

        newdt = MAX(1.0E-8_WRD, dt_CFL + (dt_CFL - self%dt) * 2)

        DEBUGMSG("    Decreasing dt to "//E2A(newdt))
    ELSE
        newdt = self%dt + 0.1_WRD * (dt_CFL - self%dt)

        DEBUGMSG("    Increasing dt to "//E2A(newdt))
    END IF

#ifdef WMBT_COMM_BENCHMARK
    newdt = self%dt
#endif

END FUNCTION propose_timestep

END MODULE Mod_SimulationUnits
