#include "macro.h"

PROGRAM Wombat

! FILENAME: wombat.f90
!
! DESCRIPTION: This is the main driver for Wombat-2,
!              0) Setup MPI & OpenMP, set global rank and thread variables
!              1) Instantiate parameter file object
!              2) Set top level parameters
!              3) Instantiate Decomposition. This includes MPI & RMA setup and initial rank location 
!                 on the world domain. AIO servers remain in this call.
!              4) Set MPI worker rank location
!              3) Instantiate all other classes (Units, ICs, solvers, IO)
!              4) Enter OpenMP region, construct patches, page pool 
!                 variables for memory locality.
!              5) Open RMA locks to neighbouring ranks (neighbourhood size can be > 1)
!              6) Enter simulation loop:
!                  6a) Run all solvers for this problem for this time
!                  6b) Adjust Timestep
!                  6c) Check/Do output
!              7) Destroy all objects
!
! DEPENDENCIES: The rest of Wombat
!
! HISTORY:
!    11/08/2017  - Julius Donnert - Restructure & new parameter files
!    07/29/2015  - Peter Mendygral - Adding this header

USE OMP_LIB                 !require these external libraries/modules

#ifndef GENPARFILE
USE MPI
#endif

USE Mod_Globals             ! pull in all of the required wombat modules
USE Mod_Error
USE Mod_Parameters
USE Mod_RankLocation
USE Mod_SimulationUnits
USE Mod_Domain
USE Mod_Problem
USE Mod_DomainSolver
USE Mod_Decomposition
USE Mod_MultiGrid_Domain
USE Mod_MultiGrid
USE Mod_IO
USE Mod_IO_DUMP

IMPLICIT NONE

TYPE(Parameters) :: param
TYPE(RankLocation) :: rankloc
TYPE(SimulationUnits) :: simunits
TYPE(Domain) :: rootd
TYPE(MGDomain) :: mgd
TYPE(Problem) :: prob
TYPE(DomainSolver) :: dsolve
TYPE(MultiGrid) :: mg
TYPE(Decomposition) :: decomp

TYPE(IO) :: ioo
TYPE(IO_DUMP) :: dump_compressed, dump_movie, dump_restart

REAL(WRD) :: dt_mod_comp, dt_mod_movie, dtproposed, dtreduced
INTEGER(WIS) :: n, dtid, npass_vars, ncr_momentum_bins, n_xpatches, n_ypatches, n_zpatches, &
                patch_nx, patch_ny, patch_nz, patch_nb
INTEGER(WIS) :: npm_maxblockmgrs, npm_partcls_per_dm, npm_dms_per_blkmgr, npm_steps_per_defrag
LOGICAL(WIS) :: mhdtvdOn, mhdwenoOn, passiveOn, crsOn, static_gravityOn, variable_gravityOn, dark_matterOn

REAL(WRD) :: timer1, timer2, timer3, DM_buff_fac, particle_mass

! START of simulation

CALL Startup()

CALL InitSimulation()

decomp = Decomposition(param) ! AIO servers will remain here

CALL InitMyLocation()

CALL InitClasses()

!$OMP PARALLEL DEFAULT(SHARED)
 
CALL ConstructDomain() ! domains can be built and paged in now that we are in the threaded region

!$OMP MASTER

dtid   = 0
n      = 0
timer1 = decomp%global_time()

!$OMP END MASTER
!$OMP BARRIER

CALL decomp%rma_lock_neighbors() ! open up the RMA locks to our neighbors

CALL DoWrites() ! Check for output of the ICs

RMSG("Starting simulation"" t = "//F2A(simunits%t)//", dt = "//E2A(simunits%dt)//", t_end ="//F2A(simunits%t_end)//", max steps = "//I2A(simunits%max_steps) )

DO WHILE (simunits%t .LT. simunits%t_end .AND. n .LT. simunits%max_steps)

    !$OMP MASTER

    timer3  = decomp%global_time() ! start the loop timer and init the maxsignal back to 0
    
    rootd%maxsignal(:) = 0.0_WRD

    !$OMP END MASTER
	!$OMP BARRIER

    CALL DoSolves()
    
    CALL TimeStepManagement()
    
    CALL DoWrites()

    !$OMP MASTER

    timer2 = decomp%global_time()

    !$OMP END MASTER
	!$OMP BARRIER

    RMSG("Step "//I2A(n)//", t = "//F2A(simunits%t)//", dt = "//F2A(simunits%dt)//", StepWall = "//F2A((timer2 - timer3))//", SimWall = "//F2A((timer2 - timer1)))

END DO

!$OMP SINGLE
CALL decomp%end_allreduce(dtid)
!$OMP END SINGLE

CALL decomp%rma_unlock_neighbors()

!$OMP END PARALLEL

CALL DestroyClasses()

! END of simulation

CONTAINS ! Internal routines

! Define the world location of this MPI rank. This makes sense only for homogenous load distributions.
! For inhomogeneous load distributions, the code will redistribute patch among ranks in a domain
! neighbourhood until balance is reached. In this case, a rank can hold any patch which was originally 
! inside its domain neighborhood.

SUBROUTINE InitMyLocation()

    CHARACTER(WCHAR_BUF_SIZE) :: strbuf1, strbuf2, strbuf3
    INTEGER(WIS) :: nranks(3), npatches(3), patch_size(3)
    LOGICAL(WIS) :: is_boundary(2,3)
    INTEGER(8) :: N

#ifndef GENPARFILE
    CALL decomp%initRMA(patch_nx, patch_ny, patch_nz, patch_nb, MAX(8, npass_vars, ncr_momentum_bins))
#endif

    nranks     = (/decomp%nranks_x, decomp%nranks_y, decomp%nranks_z/)
    npatches   = (/n_xpatches, n_ypatches, n_zpatches/)
    patch_size = (/patch_nx, patch_ny, patch_nz/)

    is_boundary(:,1) = (/decomp%lx_bound, decomp%hx_bound/)
    is_boundary(:,2) = (/decomp%ly_bound, decomp%hy_bound/)
    is_boundary(:,3) = (/decomp%lz_bound, decomp%hz_bound/)

    rankloc = RankLocation(decomp%my_coords, is_boundary, nranks, npatches, patch_size)
    
    RPRINT '("")'
    RMSG("Number of patches x direction: "//I2A(npatches(1)*nranks(1)))
    RMSG("Number of patches y direction: "//I2A(npatches(2)*nranks(2)))
    RMSG("Number of patches z direction: "//I2A(npatches(3)*nranks(3)))
    RPRINT '("")'
    RMSG("Number of zones   x direction: "//I2A(nranks(1)*npatches(1)*patch_size(1)))
    RMSG("Number of zones   y direction: "//I2A(nranks(2)*npatches(2)*patch_size(2)))
    RMSG("Number of zones   z direction: "//I2A(nranks(3)*npatches(3)*patch_size(3)))
    RPRINT '("")'

    N = npatches(1)*nranks(1)*patch_size(1)
    N = N * npatches(2)*nranks(2)*patch_size(2)
    N = N * npatches(3)*nranks(3)*patch_size(3)

    RMSG("Total number of zones: "//L2A(N))
    RPRINT '("")'

END SUBROUTINE InitMyLocation

! Instantiate and initialize all the classes we need for the run. 
SUBROUTINE InitClasses()

    RMSG("Initializing classes ...")

    simunits = SimulationUnits(param, rankloc)

    CALL rankloc%calc_location(simunits%dx)

    prob           = Problem(param, rankloc, simunits)
    prob%stateTag  = 1
    prob%mggravTag = 0

    rootd  = Domain(prob%stateTag, decomp%ndims, n_xpatches, n_ypatches, n_zpatches, rankloc, simunits)

    dsolve = DomainSolver(param, simunits, decomp%ndims, prob%gamma, npass_vars, npm_partcls_per_dm, particle_mass)

    IF (variable_gravityOn .OR. GENPARFILE_ON) THEN

        mgd = MGDomain(prob%mggravTag, param, decomp%ndims, n_xpatches, n_ypatches, n_zpatches, rankloc, simunits)

        mg  = MultiGrid(param, 1)

    END IF

    dump_compressed = IO_DUMP(param, simunits, IO_DUMP_COMPRESSED) ! dump types get their own objects
    dump_movie      = IO_DUMP(param, simunits, IO_DUMP_MOVIE)
    !dump_restart   = IO_DUMP(param, simunits, IO_DUMP_RESTART)
    ioo = IO(param, decomp%iolead, decomp%naio, rankloc, simunits) ! parent class

    IF (GENPARFILE_ON) THEN

        CALL param%destroy()
        RMSG( "Parameter file generation done.")
        STOP
    END IF

END SUBROUTINE InitClasses

SUBROUTINE DestroyClasses()
    
    CALL ioo%destroy(dump_compressed, dump_movie, dump_restart)
    CALL rootd%destroy()
    CALL mgd%destroy()
    CALL prob%destroy()
    CALL dsolve%destroy()
    CALL mg%destroy()
    CALL decomp%destroy()
    CALL param%destroy()

END SUBROUTINE DestroyClasses

! Build up our domains and patches : 
! First create the shells of the Patches within the root domain (Patch arrays are allocated in a separate step)
! Then actually construct the Patches once that they are prepared. This step does first touches etc. and is 
! separate as we will be dynamically building and un-building patches once we have AMR 
! We are also indicating that the user Init routine (init the data on the grid) should be called on each Patch.
! We then initialize the RMA Pipeline and have each thread create its own workpool variable to ensure good
! memory locality.

SUBROUTINE ConstructDomain()

    RMSG("Constructing domains ...")

    IF (decomp%ndims .EQ. 1) THEN

        CALL rootd%build(decomp%myrank, decomp%domain_neighbors1d(0,:), 8, npass_vars, ncr_momentum_bins, mhdtvdOn, &
                         mhdwenoOn, passiveOn, crsOn, static_gravityOn, variable_gravityOn, dark_matterOn, &
                         patch_nx, patch_ny, patch_nz, patch_nb, &
                         npm_maxblockmgrs, npm_partcls_per_dm, npm_dms_per_blkmgr, npm_steps_per_defrag, DM_buff_fac)

        IF (variable_gravityOn) THEN

            CALL mgd%build(decomp%myrank, decomp%domain_neighbors1d(0,:), patch_nx, patch_ny, patch_nz, patch_nb, &
                           rankloc, simunits)

        END IF

    ELSE IF (decomp%ndims .EQ. 2) THEN

        CALL rootd%build(decomp%myrank, decomp%domain_neighbors2d(0,0,:), 8, npass_vars, ncr_momentum_bins, mhdtvdOn, &
                         mhdwenoOn, passiveOn, crsOn, static_gravityOn, variable_gravityOn, dark_matterOn, &
                         patch_nx, patch_ny, patch_nz, patch_nb, &
                         npm_maxblockmgrs, npm_partcls_per_dm, npm_dms_per_blkmgr, npm_steps_per_defrag, DM_buff_fac)

        IF (variable_gravityOn) THEN

            CALL mgd%build(decomp%myrank, decomp%domain_neighbors2d(0,0,:), patch_nx, patch_ny, patch_nz, patch_nb, &
                           rankloc, simunits)
        
        END IF 

    ELSE IF (decomp%ndims .EQ. 3) THEN
        
        CALL rootd%build(decomp%myrank, decomp%domain_neighbors3d(0,0,0,:), 8, npass_vars, ncr_momentum_bins, mhdtvdOn, &
                         mhdwenoOn, passiveOn, crsOn, static_gravityOn, variable_gravityOn, dark_matterOn, &
                         patch_nx, patch_ny, patch_nz, patch_nb, &
                         npm_maxblockmgrs, npm_partcls_per_dm, npm_dms_per_blkmgr, npm_steps_per_defrag, DM_buff_fac)

        IF (variable_gravityOn) THEN 

            CALL mgd%build(decomp%myrank, decomp%domain_neighbors3d(0,0,0,:), patch_nx, patch_ny, patch_nz, &
                           patch_nb, rankloc, simunits)
        END IF

    END IF

    ! actually construct the Patches now that they are prepared 
    ! this step does first touches etc. and is separate as we will be dynamically building and un-building patches 
    ! once we have AMR we are also indicating that the user Init routine (init the data on the grid) should be 
    ! called on each Patch 
    
    RMSG("Building Patches ...")

    CALL rootd%build_all_patches(prob, rankloc, simunits, .TRUE.)

    IF (variable_gravityOn) THEN

        CALL mgd%build_all_patches(prob, rankloc, simunits)

    ENDIF 

    CALL decomp%initRMA_pipeline() ! initialize the RMA pipeline for solves of any type

    ! have each thread create its own workpools 
    CALL dsolve%init_workpools(mhdtvdOn, passiveOn, crsOn, variable_gravityOn, dark_matterOn, mhdwenoOn, &
                               patch_nx, patch_ny, patch_nz, patch_nb, npm_partcls_per_dm)

END SUBROUTINE ConstructDomain

! Call the solvers to advance the system. Flavors :
! 1 - MHD solver
! 2 - CT algorithm (2 and 3d only)
! 3 - Van Leer passive solver
! 4 - Cosmic Ray solver
! 5 - Dark Matter particles: 1st Kick and Drift (no boundary exchange)
! 6 - Particle Mesh solver : Assign density (exchange particles, repeat until done)
! 7 - Full Multigrid Solver: Solve for the gravitational potential. Solver is different.
! 8 - Dark Matter particles: 2nd Kick (exchange grav potential)
!
! DO NOT SHUFFLE THE LIST BELOW !

SUBROUTINE DoSolves()

    IF (mhdtvdOn          ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_MHDTVD)
    IF (mhdwenoOn         ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_MHDWENO)                   
    IF (passiveOn         ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_PASSIVE)
    !IF (crsOn            ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_COSMIC_RAYS)
    IF (dark_matterOn     ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_DM_KICKDRIFT)
    IF (dark_matterOn     ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_DM_PARTICLES)
    IF (variable_gravityOn) CALL mg%fmg_solve(rootd, mgd, dsolve, rankloc, simunits, decomp, prob)
    IF (dark_matterOn     ) CALL dsolve%solve(rootd, prob, decomp, rankloc, simunits, FL_DM_KICK2)

END SUBROUTINE DoSolves

! Go through each dumpset type and perform the AIO write if it is time.  
! We also track adjustments to dt to get dumps at precisely the desired times 

SUBROUTINE DoWrites()

    LOGICAL(WIS) :: writeCompressed=.FALSE., writeMovie=.FALSE., writeRestart=.FALSE.

    CALL ioo%time_to_write(dump_movie, simunits%t, simunits%dt, writeMovie, dt_mod_movie)
    CALL ioo%time_to_write(dump_compressed, simunits%t, simunits%dt, writeCompressed, dt_mod_comp)
    !CALL ioo%time_to_write(dump_restart, simunits%t, simunits%dt, writeRestart, dt_mod_restart)

    !$OMP SINGLE

    simunits%dt = MIN(simunits%dt, dt_mod_comp, dt_mod_movie) ! set dt based on when our next write will occur

    CALL simunits%update_code2cgs()
    
    !$OMP END SINGLE

    ! inform all domains and Patches of the current time and time step 
    CALL rootd%broadcast_t_and_dt(simunits%t, simunits%dt)

    IF (writeCompressed .OR. writeMovie .OR. writeRestart) THEN
        
        CALL ioo%allocate_field_buffer_patch_flags(rootd)
    END IF

    IF (writeMovie)       THEN

        !$OMP MASTER
        CALL ioo%wait_aio_slots(dump_movie, 0, .TRUE.) ! AIO doesnt support multiple inflight messages yet
        CALL ioo%wait_aio_slots(dump_compressed, 0, .TRUE.) 
        !CALL ioo%wait_aio_slots(dump_restart, 0, .TRUE.)
        !$OMP END MASTER
        !$OMP BARRIER

        CALL ioo%write_dumpset(rootd, dump_movie, prob%gamma)

    END IF

    IF (writeCompressed) THEN

        !$OMP MASTER
        CALL ioo%wait_aio_slots(dump_movie, 0, .TRUE.) 
        CALL ioo%wait_aio_slots(dump_compressed, 0, .TRUE.) 
        !CALL ioo%wait_aio_slots(dump_restart, 0, .TRUE.)
        !$OMP END MASTER
        !$OMP BARRIER

        CALL ioo%write_dumpset(rootd, dump_compressed, prob%gamma)

    END IF

    IF (writeRestart) THEN
        
        !$OMP MASTER
        CALL ioo%wait_aio_slots(dump_movie, 0, .TRUE.)
        CALL ioo%wait_aio_slots(dump_compressed, 0, .TRUE.) 
        !CALL ioo%wait_aio_slots(dump_restart, 0, .TRUE.)
        !$OMP END MASTER
        !$OMP BARRIER

        CALL ioo%write_dumpset(rootd, dump_compressed, prob%gamma)

    END IF

    IF (writeCompressed .OR. writeMovie .OR. writeRestart) THEN
        
        CALL ioo%deallocate_field_buffer_patch_flags(rootd)
    END IF

END SUBROUTINE DoWrites

! Time step management including the global reduction. 
! This will get its own module once we do cosmology as well.

SUBROUTINE TimeStepManagement()

    !$OMP SINGLE
     
    simunits%t = simunits%t + simunits%dt ! increment time and the step counter
    n          = n + 1
     
    IF (dtid .NE. 0) THEN ! if we already had a time step reduction in-flight, wait for it and apply the new step

        CALL decomp%end_allreduce(dtid)
        
        simunits%dt = dtreduced

        IF ((simunits%t+simunits%dt) .GT. simunits%t_end) THEN

            simunits%dt = simunits%t_end - simunits%t
        END IF
    END IF
    
    ! determine a locally proposed new time step size then start a reduction across ranks of that value 
    dtproposed = simunits%propose_timestep(n, rootd%maxsignal)

    CALL decomp%start_allreduce_min(dtproposed, dtreduced, dtid)

    !$OMP END SINGLE

END SUBROUTINE TimeStepManagement

SUBROUTINE Startup()

    INTEGER(WIS) :: ierr, thread_support, myierr, str_len, nranks
    CHARACTER(MPI_MAX_ERROR_STRING) :: err_string
    CHARACTER(8) :: date
    CHARACTER(10) :: time
    CHARACTER(5) :: zone
    INTEGER(4), DIMENSION(8) :: values

    ! initialize MPI with thread support.

#ifndef GENPARFILE
    CALL MPI_INIT_THREAD(MPI_THREAD_MULTIPLE, thread_support, ierr)

    IF (ierr .NE. MPI_SUCCESS) THEN ! Catch MPI errors

        CALL MPI_ERROR_STRING(ierr, err_string, str_len, myierr)

        MSG('MPI Library Error => ' // err_string)

        STOP
    END IF

    CALL MPI_COMM_RANK(MPI_COMM_WORLD, WRANK, ierr)

    DEBUGMSG('MPI_COMM_WORLD created')

    RASSERT(thread_support .EQ. MPI_THREAD_MULTIPLE, 'MPI_THREAD_MULTIPLE is NOT available but is required to run WOMBAT.')
#endif

    param = Parameters()

    ! initialize OpenMP
    
    WNTHREADS = param%getInt("WNTHREADS", "1", "# Global number of OpenMP threads per MPI rank")
    
    WRITE(WNTHREADSTR, '(I3)') WNTHREADS

    CALL OMP_SET_NUM_THREADS(WNTHREADS) 

!$OMP PARALLEL DEFAULT(SHARED) 
    WTHREAD   = OMP_GET_THREAD_NUM() ! vars are consistent between parallel regions. OpenMP standard 4.5, section 2.15.
    
    WRITE(WTHREADSTR, '(I3)') WTHREAD

!$OMP END PARALLEL

    CALL GET_ENVIRONMENT_VARIABLE("USER", USERNAME)
    CALL DATE_AND_TIME(date, time, zone, values)
    
#ifndef GENPARFILE
    CALL MPI_Get_processor_name(HOSTNAME, str_len, ierr)
    CALL MPI_COMM_SIZE(MPI_COMM_WORLD, nranks, ierr) ! no global, because we don't want to confuse anyone with AIO
#else
    nranks = 1
#endif

    RPRINT '("+-----------------------------------------------------------+")'
    RPRINT '("| Wombat - Scalable astrophysical MHD, DM & Cosmology       |")'
    RPRINT '("| Version: ", F3.1, "                                              |")', WVERSION
    RPRINT '("|                                                           |")'
    RPRINT '("|              ", I2,"/",I2,"/",I4," - ",I2,":",I2,":",I2," UTC",1F4.1,"                |")', &
            values(2), values(3), values(1), values(5), values(6), values(7), values(4)/60.0
    RPRINT '("|                                                           |")'
    RPRINT '("|                   System Name: ", A24, "   |")', HOSTNAME
    RPRINT '("|                     User Name: ", A24, "   |")', USERNAME
    RPRINT '("|     Total Number of MPI Ranks: ", I24, "   |")', nranks
    RPRINT '("|    Number of Threads per Rank: ", I24, "   |")', WNTHREADS
    RPRINT '("|    Total Number Compute Units: ", I24, "   |")', WNTHREADS * nranks
    RPRINT '("|                                                           |")'
    RPRINT '("| Cray Inc.                                                 |")'
    RPRINT '("| University of Minnesota                                   |")'
    RPRINT '("|                                                           |")'
    RPRINT '("| Mendygral et al. 2017, ApJS (228)                         |")'
    RPRINT '("|                                                           |")'
    RPRINT '("| bitbucket.org/pmendygral/wombat-public.git                |")'
    RPRINT '("| MIT license                                               |")'
    RPRINT '("+-----------------------------------------------------------+")'
    RPRINT '(" ")'

    DEBUGMSG('Startup done')

END SUBROUTINE Startup

! Do a bunch of sanity checks on the parameters from the file

SUBROUTINE InitSimulation()

    CALL param%section("Solver Component Controls")
    mhdtvdOn            = param%getLog("mhdtvdOn",          "false", "# MHD TVD solver")
    mhdwenoOn           = param%getLog("mhdwenoOn",         "false", "# MHD WENO solver")
    passiveOn           = param%getLog("passiveOn",         "false", "# Passive/Color variables")
    crsOn               = param%getLog("crsOn",             "false", "# Cosmic Ray solver")                        
    static_gravityOn    = param%getLog("static_gravityOn",  "false", "# Static gravitation potential ?")                        
    variable_gravityOn  = param%getLog("variable_gravityOn","false", "# Multi grid gravity")                        
    dark_matterOn       = param%getLog("dark_matterOn",     "false", "# Dark Matter collisionless particles")
    !comovingOn          = param%getLog("comovingOn",        "false", "# Comoving (cosmological) time integration"

    CALL param%section("Patch decomposition of simulation domain")
    n_xpatches          = param%getInt("n_xpatches", "1", "# Number of patches in x-direction per MPI rank")
    n_ypatches          = param%getInt("n_ypatches", "1", "# Number of patches in y-direction per MPI rank")
    n_zpatches          = param%getInt("n_zpatches", "1", "# Number of patches in z-direction per MPI rank")
    patch_nx            = param%getInt("patch_nx",   "1", "# Number of zones in x-direction per patch")
    patch_ny            = param%getInt("patch_ny",   "1", "# Number of zones in y-direction per patch")
    patch_nz            = param%getInt("patch_nz",   "1", "# Number of zones in z-direction per patch")
    patch_nb            = param%getInt("patch_nb",   "3", "# Number of boundary zones per boundary")

    CALL param%section("DM particles") ! Can we change the variable names to something easier ?
    npm_maxblockmgrs    = param%getInt("npm_maxblockmgrs",      "8",        &
                                       "# Number of dark matter block managers (first level in DM tree)")
    npm_partcls_per_dm  = param%getInt("npm_partcls_per_dm",    "8",        &
                                       "# Number of dark matter segments per block manager (second level in DM tree)")
    npm_dms_per_blkmgr  = param%getInt("npm_dms_per_blkmgr",    "1000000",  &
                                       "# Number of dark matter particle slots per DM segment")
    npm_steps_per_defrag= param%getInt("npm_steps_per_defrag",  "8",        &
                                       "# Number of time steps between DM tree defragmentation operations", 0, .true.)

    particle_mass       = param%getFlt("particle_mass", "1.0", &
        "# dark matter particle mass in code units") ! will be set by ICs, no parameter needed
     
    CALL param%section("Various") ! move those somewhere else ?
    npass_vars          = param%getInt("npass_vars",        "1",    "# Number of passive variables")
    ncr_momentum_bins   = param%getInt("ncr_momentum_bins", "1",    "# Number of CR momentum bins")


    ! consistency check basic simulation parameters
    
    RASSERT(mhdtvdOn .OR. .NOT. passiveOn, "MHD must be enabled to solve for passive scalars")
    RASSERT(mhdtvdOn .OR. .NOT. crsOn, "MHD must be enabled to solve for cosmic rays")
    RASSERT(.NOT. mhdtvdOn .OR. patch_nb .GE. 5, "mhdtvd requires patch_nb>=5")
    RASSERT(npass_vars.GT.1 .OR. .NOT. passiveOn, "passiveOn requires npass_vars >= 1")
    RASSERT(ncr_momentum_bins.GT.1 .OR. .NOT. crsOn, "crsOn requires ncr_momentum_bins >= 1")
    RASSERT(.NOT. (static_gravityOn .AND. variable_gravityOn), "static_gravityOn and variable_gravityOn cannot both be enabled")
    RASSERT(n_xpatches.GE.1 .AND. n_ypatches.GE.1 .AND. n_zpatches.GE.1, "There must be at least 1 Patch per dimension on the root domain")
    RASSERT(patch_nx .GE. 1 .AND. patch_ny .GE. 1 .AND. patch_nz .GE. 1, "There must be at least 1 zone in each dimension")

    IF (dark_matterOn) THEN
        RASSERT(npm_maxblockmgrs .GE. 1,    "npm_maxblockmgrs must be at least 1 when dark matter is enabled")
        RASSERT(npm_partcls_per_dm .GT. 1,  "npm_partcls_per_dm must be at least 1 when dark matter is enabled")
        RASSERT(npm_dms_per_blkmgr .GT. 1,  "npm_dms_per_blkmgrmust be at least 1 when dark matter is enabled")
        RASSERT(npm_steps_per_defrag .GE. 1,"npm_steps_per_defrag must be at least 1 when dark matter is enabled")
    END IF

END SUBROUTINE InitSimulation


END PROGRAM Wombat
