
module JColMaps

using Colors
using ColorTypes
using FixedPointNumbers
import Plots
using Measures
import Interpolations 
using StaticArrays
using StringConversion

export ColMap, Arr2Img, ShowColorMaps, ColBar

"""
Show all available perceptual color maps.
"""
function ShowColorMaps()

	arr = zeros(Float64, 1, 256)
	
	for i=1:1
		arr[i,:] = linspace(0,1, 256)
	end

	img = Array{RGB{N0f8}}( 256, 1*51)

	for i=1:51

		cmap = ColMap(i)

		imin = (i-1)*1+1
		imax = i*1
			
		img[:,imin:imax] = Arr2Img(arr, cmap, range=[0,1])
	end

    Plots.plot(img, show=true, size=(512,512), 
               top_margin=1mm, yticks=nothing,
               ticks=((0,50,100,150,200,250), (0,10,20,30,40,50)))

	return
end

"""
These are the Perceptual Colour Maps from 
http://peterkovesi.com/projects/colourmaps/index.html

The native Julia module is broken so we take them from a big table.

	cmap = ColMap(10, debug=true)

	cmap = ColMap(0) # show all maps available

cmap is then a 256-element Array of RGB{Float64} triplets
"""
function ColMap(i::Integer; debug=false)

	if debug == true
		println("Colormap $i : $(names[i])")
	end

	if i == 0
		
		println("Available perceptual colormaps: ")
		
		for j = 1:size(names,1)
			println("   $j - $(names[j])")
		end

		return -1
	end

	tmp = bigtable[i] # get colmap

	N = size(tmp, 1)

	#cmap_rgb = Array{RGB24}(N)
	cmap_rgb = Array{RGB{N0f8}}(N)
	
	for j=1:N
		cmap_rgb[j] = RGB(tmp[j][1],tmp[j][2],tmp[j][3])
	end

	return cmap_rgb
end

"""
Plot a color bar from color map "cmap" with range "range" in subplot
"subplot". If log=true, use log scaling. Use "N" ticks in the color bar
    
    ColBar(cmap, range, subplot; log=false, nticks=4)

"""
function ColBar(cmap, range, subplot; log=false, nticks=4, npix=256)
    
    arr = zeros(Float64,npix,npix)

    if log == true
        for i = 1:npix
            arr[:,i] = logspace(log10(range[2]), log10(range[1]), npix)
        end
    else
        for i = 1:256
            arr[:,i] = linspace(range[2], range[1], npix)
        end
    end

    N = nticks

    if log == true
        vals  = logspace(1,log10(256),N)
        lvals = logspace(log10(256),1,N)./256 .* (range[2]-range[1]) + range[1]
    else
        vals  = linspace(1,256,N)
        lvals = linspace(256,1,N)./256 .* (range[2]-range[1]) + range[1]
    end

    labels= Array{String}(N)
    for i = 1:N
        labels[i] = gtoa(lvals[i])
    end

    yticks = (vals,labels)

    img = Arr2Img(arr,cmap, range=range)

    Plots.plot!(img, subplot=subplot, xticks=false, xflip=false, 
          mirror=true, yticks=yticks, rotation=45)

    return
end


"""
Convert a 2D array into an image using a color map and an optional 
range of scaling values and types. The type of the image is the type of the 
color map. The packages Colors and ColorType provide the conversions
and type definitions. Giving a range speeds up this function.
	
Example: 

	cmap = ColMap(1)
	img = Arr2Img(arr::Array{<:Real,2}, cmap; range=nothing, log=false, sqrt=false)

"""
function Arr2Img(arr::Array{<:Real,2}, cmap; range=nothing, log=false, sqrt=false)

	img = (arr)  # correct alignment

    if log == true
   		bad = find(img .<= 0)
   		good = find(img .> 0)
        img[bad] = minimum(img[good])
    end

	if sqrt == true
		bad = find(img .< 0)
        img[bad] = 0
    end

	if range == nothing # auto range

		good = find(isfinite.(img))
		
		range = [ minimum(img[good]), maximum(img[good]) ]

		println("Autorange: $range")
	end

	if log == true
		img = log10.(img)
		range = log10.(range)
	end

	if sqrt == true
		img = sqrt.(img)
		range = sqrt.(range)
	end

	ncolours = length(cmap)

    rgbimg = Array{typeof(cmap[1])}(size(img))
    
    @. img[isnan(img)] = range[1]

    N = length(img)

    range2colidx = (ncolours-1)/(range[2]-range[1]) 

    idx = Array{Int64}(N)
    @. idx = clamp(round(Int64,(img[:]-range[1]) * range2colidx + 1), 1, ncolours)

    cmap_static = SVector{size(cmap,1), typeof(cmap[1])}(cmap)

    Threads.@threads for i = 1:N
        @inbounds rgbimg[i] = cmap_static[idx[i]]
    end
   
    return rgbimg
end

const names = [
	"_bgyr_35-85_c73_n256"	,					# 1
   	"cyclic_grey_15-85_c0_n256",				# 2
   	"cyclic_grey_15-85_c0_n256_s25",
    "cyclic_mrybm_35-75_c68_n256",
    "cyclic_mrybm_35-75_c68_n256_s25",
    "cyclic_mygbm_30-95_c78_n256",
    "cyclic_mygbm_30-95_c78_n256_s25",
	"cyclic_wrwbw_40-90_c42_n256",
    "cyclic_wrwbw_40-90_c42_n256_s25",
    "diverging-isoluminant_cjm_75_c23_n256",	# 10
    "diverging-isoluminant_cjm_75_c24_n256",
    "diverging-isoluminant_cjo_70_c25_n256",
    "diverging-linear_bjr_30-55_c53_n256",
    "diverging-linear_bjy_30-90_c45_n256",
	"diverging-rainbow_bgymr_45-85_c67_n256",
	"diverging_bkr_55-10_c35_n256",
    "diverging_bky_60-10_c30_n256",
    "diverging_bwr_40-95_c42_n256",
    "diverging_bwr_55-98_c37_n256",
    "diverging_cwm_80-100_c22_n256",
    "diverging_gkr_60-10_c40_n256",
    "diverging_gwr_55-95_c38_n256",
    "diverging_gwv_55-95_c39_n256",				# 23
    "isoluminant_cgo_70_c39_n256",
    "isoluminant_cgo_80_c38_n256",
    "isoluminant_cm_70_c39_n256",				# 26
    "linear_bgy_10-95_c74_n256",
    "linear_bgyw_15-100_c67_n256",
    "linear_bgyw_15-100_c68_n256",
    "linear_blue_5-95_c73_n256",
    "linear_blue_95-50_c20_n256",
    "linear_bmw_5-95_c86_n256",
    "linear_bmw_5-95_c89_n256",
    "linear_bmy_10-95_c71_n256",
    "linear_bmy_10-95_c78_n256",
    "linear_gow_60-85_c27_n256",
    "linear_gow_65-90_c35_n256",
    "linear_green_5-95_c69_n256",
    "linear_grey_0-100_c0_n256",
    "linear_grey_10-95_c0_n256",
    "linear_kry_5-95_c72_n256",
    "linear_kry_5-98_c75_n256",
    "linear_kryw_5-100_c64_n256",
    "linear_kryw_5-100_c67_n256",
    "linear_ternary-blue_0-44_c57_n256",
    "linear_ternary-green_0-46_c42_n256",
	"linear_ternary-red_0-50_c52_n256",	
	"rainbow_bgyr_35-85_c72_n256",				# 48
	"rainbow_bgyr_35-85_c73_n256",
	"rainbow_bgyrm_35-85_c69_n256",
 	"rainbow_bgyrm_35-85_c71_n256" ]			# 51

const bigtable = [ # It's not stupid if it works ;-)
[
[0, 0.207843, 0.976471],[0, 0.219608, 0.964706],[0, 0.227451, 0.952941],[0, 0.239216, 0.941176],
[0, 0.247059, 0.929412],[0, 0.258824, 0.917647],[0, 0.266667, 0.905882],[0, 0.278431, 0.894118],
[0, 0.286275, 0.882353],[0, 0.294118, 0.87451],[0, 0.301961, 0.862745],[0, 0.309804, 0.85098],
[0, 0.317647, 0.839216],[0, 0.32549, 0.827451],[0, 0.333333, 0.815686],[0, 0.341176, 0.803922],
[0, 0.34902, 0.792157],[0, 0.356863, 0.780392],[0, 0.360784, 0.768627],[0, 0.368627, 0.760784],
[0, 0.376471, 0.74902],[0, 0.384314, 0.737255],[0, 0.388235, 0.72549],[0, 0.396078, 0.713725],
[0, 0.403922, 0.701961],[0, 0.407843, 0.690196],[0, 0.415686, 0.682353],[0, 0.423529, 0.670588],
[0, 0.427451, 0.658824],[0, 0.435294, 0.647059],[0, 0.439216, 0.639216],[0, 0.443137, 0.627451],
[0, 0.45098, 0.615686],[0, 0.454902, 0.607843],[0, 0.458824, 0.596078],[0, 0.462745, 0.588235],
[0.027451, 0.466667, 0.576471],[0.054902, 0.470588, 0.568627],[0.0784314, 0.478431, 0.556863],[0.0941176, 0.482353, 0.54902],
[0.109804, 0.486275, 0.537255],[0.12549, 0.490196, 0.529412],[0.137255, 0.494118, 0.521569],[0.14902, 0.498039, 0.509804],
[0.160784, 0.501961, 0.501961],[0.168627, 0.505882, 0.494118],[0.176471, 0.509804, 0.482353],[0.184314, 0.513725, 0.47451],
[0.192157, 0.517647, 0.462745],[0.2, 0.521569, 0.454902],[0.203922, 0.52549, 0.447059],[0.207843, 0.529412, 0.435294],
[0.215686, 0.533333, 0.427451],[0.219608, 0.537255, 0.415686],[0.223529, 0.541176, 0.407843],[0.227451, 0.545098, 0.396078],
[0.231373, 0.54902, 0.388235],[0.231373, 0.552941, 0.376471],[0.235294, 0.556863, 0.368627],[0.239216, 0.560784, 0.356863],
[0.239216, 0.564706, 0.345098],[0.243137, 0.568627, 0.337255],[0.243137, 0.572549, 0.32549],[0.243137, 0.576471, 0.313725],
[0.247059, 0.580392, 0.305882],[0.247059, 0.584314, 0.294118],[0.247059, 0.588235, 0.282353],[0.247059, 0.596078, 0.270588],
[0.247059, 0.6, 0.258824],[0.247059, 0.603922, 0.247059],[0.247059, 0.607843, 0.235294],[0.247059, 0.611765, 0.223529],
[0.247059, 0.615686, 0.207843],[0.247059, 0.619608, 0.196078],[0.247059, 0.623529, 0.184314],[0.247059, 0.627451, 0.168627],
[0.247059, 0.631373, 0.156863],[0.25098, 0.635294, 0.141176],[0.25098, 0.639216, 0.129412],[0.254902, 0.643137, 0.117647],
[0.258824, 0.647059, 0.105882],[0.266667, 0.65098, 0.0941176],[0.27451, 0.65098, 0.0823529],[0.282353, 0.654902, 0.0745098],
[0.290196, 0.658824, 0.0666667],[0.301961, 0.662745, 0.0627451],[0.309804, 0.662745, 0.0588235],[0.321569, 0.666667, 0.054902],
[0.333333, 0.670588, 0.0509804],[0.341176, 0.670588, 0.0509804],[0.352941, 0.67451, 0.0509804],[0.364706, 0.67451, 0.054902],
[0.376471, 0.678431, 0.054902],[0.388235, 0.682353, 0.054902],[0.396078, 0.682353, 0.054902],[0.407843, 0.686275, 0.0588235],
[0.415686, 0.686275, 0.0588235],[0.427451, 0.690196, 0.0627451],[0.439216, 0.694118, 0.0627451],[0.447059, 0.694118, 0.0627451],
[0.458824, 0.698039, 0.0666667],[0.466667, 0.698039, 0.0666667],[0.478431, 0.701961, 0.0666667],[0.486275, 0.705882, 0.0705882],
[0.494118, 0.705882, 0.0705882],[0.505882, 0.709804, 0.0745098],[0.513725, 0.709804, 0.0745098],[0.52549, 0.713725, 0.0745098],
[0.533333, 0.713725, 0.0784314],[0.541176, 0.717647, 0.0784314],[0.552941, 0.717647, 0.0784314],[0.560784, 0.721569, 0.0823529],
[0.568627, 0.72549, 0.0823529],[0.580392, 0.72549, 0.0823529],[0.588235, 0.729412, 0.0862745],[0.596078, 0.729412, 0.0862745],
[0.603922, 0.733333, 0.0901961],[0.615686, 0.733333, 0.0901961],[0.623529, 0.737255, 0.0901961],[0.631373, 0.737255, 0.0941176],
[0.639216, 0.741176, 0.0941176],[0.65098, 0.741176, 0.0941176],[0.658824, 0.745098, 0.0980392],[0.666667, 0.74902, 0.0980392],
[0.67451, 0.74902, 0.101961],[0.686275, 0.752941, 0.101961],[0.694118, 0.752941, 0.101961],[0.701961, 0.756863, 0.105882],
[0.709804, 0.756863, 0.105882],[0.717647, 0.760784, 0.105882],[0.729412, 0.760784, 0.109804],[0.737255, 0.764706, 0.109804],
[0.745098, 0.764706, 0.109804],[0.752941, 0.768627, 0.113725],[0.760784, 0.768627, 0.113725],[0.768627, 0.772549, 0.117647],
[0.780392, 0.772549, 0.117647],[0.788235, 0.776471, 0.117647],[0.796078, 0.776471, 0.121569],[0.803922, 0.780392, 0.121569],
[0.811765, 0.780392, 0.121569],[0.819608, 0.784314, 0.12549],[0.827451, 0.784314, 0.12549],[0.839216, 0.788235, 0.129412],
[0.847059, 0.788235, 0.129412],[0.854902, 0.792157, 0.129412],[0.862745, 0.792157, 0.133333],[0.870588, 0.796078, 0.133333],
[0.878431, 0.796078, 0.133333],[0.886275, 0.796078, 0.137255],[0.898039, 0.8, 0.137255],[0.905882, 0.8, 0.137255],
[0.913725, 0.803922, 0.141176],[0.921569, 0.803922, 0.141176],[0.929412, 0.803922, 0.141176],[0.937255, 0.803922, 0.145098],
[0.945098, 0.803922, 0.145098],[0.94902, 0.803922, 0.145098],[0.956863, 0.803922, 0.145098],[0.960784, 0.803922, 0.145098],
[0.968627, 0.8, 0.145098],[0.972549, 0.8, 0.141176],[0.976471, 0.796078, 0.141176],[0.980392, 0.792157, 0.141176],
[0.984314, 0.788235, 0.137255],[0.984314, 0.784314, 0.137255],[0.988235, 0.780392, 0.133333],[0.988235, 0.772549, 0.133333],
[0.992157, 0.768627, 0.129412],[0.992157, 0.764706, 0.129412],[0.992157, 0.756863, 0.12549],[0.992157, 0.752941, 0.12549],
[0.996078, 0.74902, 0.121569],[0.996078, 0.741176, 0.117647],[0.996078, 0.737255, 0.117647],[0.996078, 0.733333, 0.113725],
[0.996078, 0.72549, 0.113725],[1, 0.721569, 0.109804],[1, 0.713725, 0.105882],[1, 0.709804, 0.105882],
[1, 0.705882, 0.101961],[1, 0.698039, 0.0980392],[1, 0.694118, 0.0980392],[1, 0.690196, 0.0941176],
[1, 0.682353, 0.0941176],[1, 0.678431, 0.0901961],[1, 0.670588, 0.0862745],[1, 0.666667, 0.0862745],
[1, 0.658824, 0.0823529],[1, 0.654902, 0.0784314],[1, 0.65098, 0.0784314],[1, 0.643137, 0.0745098],
[1, 0.639216, 0.0705882],[1, 0.631373, 0.0705882],[1, 0.627451, 0.0666667],[1, 0.619608, 0.0627451],
[1, 0.615686, 0.0627451],[1, 0.611765, 0.0588235],[1, 0.603922, 0.054902],[1, 0.6, 0.0509804],
[1, 0.592157, 0.0509804],[1, 0.588235, 0.0470588],[1, 0.580392, 0.0431373],[1, 0.576471, 0.0392157],
[1, 0.568627, 0.0392157],[1, 0.564706, 0.0352941],[1, 0.556863, 0.0313725],[1, 0.552941, 0.027451],
[1, 0.545098, 0.027451],[1, 0.541176, 0.0235294],[1, 0.533333, 0.0196078],[1, 0.52549, 0.0196078],
[1, 0.521569, 0.0156863],[1, 0.513725, 0.0117647],[1, 0.509804, 0.0117647],[1, 0.501961, 0.00784314],
[1, 0.498039, 0.00784314],[1, 0.490196, 0.00392157],[1, 0.482353, 0.00392157],[1, 0.478431, 0],
[1, 0.470588, 0],[1, 0.462745, 0],[1, 0.458824, 0],[1, 0.45098, 0],
[1, 0.443137, 0],[1, 0.439216, 0],[1, 0.431373, 0],[1, 0.423529, 0],
[1, 0.415686, 0],[1, 0.407843, 0],[1, 0.403922, 0],[1, 0.396078, 0],
[1, 0.388235, 0],[1, 0.380392, 0],[1, 0.372549, 0],[1, 0.364706, 0],
[1, 0.356863, 0],[1, 0.34902, 0],[1, 0.341176, 0],[1, 0.333333, 0],
[1, 0.32549, 0],[1, 0.317647, 0],[1, 0.309804, 0],[1, 0.298039, 0],
[1, 0.290196, 0],[1, 0.282353, 0],[1, 0.270588, 0],[1, 0.262745, 0],
[1, 0.25098, 0],[1, 0.239216, 0],[1, 0.231373, 0],[1, 0.219608, 0],
[1, 0.207843, 0],[1, 0.192157, 0],[1, 0.180392, 0],[1, 0.164706, 0] 
], 
[
[0.470588, 0.470588, 0.470588],[0.47451, 0.47451, 0.47451],[0.478431, 0.478431, 0.478431],
[0.482353, 0.482353, 0.482353],[0.490196, 0.490196, 0.490196],[0.494118, 0.494118, 0.494118],[0.498039, 0.498039, 0.498039],
[0.505882, 0.505882, 0.505882],[0.509804, 0.509804, 0.509804],[0.513725, 0.513725, 0.513725],[0.521569, 0.521569, 0.521569],
[0.52549, 0.52549, 0.52549],[0.529412, 0.529412, 0.529412],[0.537255, 0.537255, 0.537255],[0.541176, 0.541176, 0.541176],
[0.54902, 0.54902, 0.54902],[0.552941, 0.552941, 0.552941],[0.556863, 0.556863, 0.556863],[0.564706, 0.564706, 0.564706],
[0.568627, 0.568627, 0.568627],[0.576471, 0.576471, 0.576471],[0.580392, 0.580392, 0.580392],[0.584314, 0.584314, 0.584314],
[0.592157, 0.592157, 0.592157],[0.596078, 0.596078, 0.596078],[0.603922, 0.603922, 0.603922],[0.607843, 0.607843, 0.607843],
[0.615686, 0.615686, 0.615686],[0.619608, 0.619608, 0.619608],[0.623529, 0.623529, 0.623529],[0.631373, 0.631373, 0.631373],
[0.635294, 0.635294, 0.635294],[0.643137, 0.643137, 0.643137],[0.647059, 0.647059, 0.647059],[0.654902, 0.654902, 0.654902],
[0.658824, 0.658824, 0.658824],[0.666667, 0.666667, 0.666667],[0.670588, 0.670588, 0.670588],[0.67451, 0.67451, 0.67451],
[0.682353, 0.682353, 0.682353],[0.686275, 0.686275, 0.686275],[0.694118, 0.694118, 0.694118],[0.698039, 0.698039, 0.698039],
[0.705882, 0.705882, 0.705882],[0.709804, 0.709804, 0.709804],[0.717647, 0.717647, 0.717647],[0.721569, 0.721569, 0.721569],
[0.729412, 0.729412, 0.729412],[0.733333, 0.733333, 0.733333],[0.737255, 0.741176, 0.741176],[0.745098, 0.745098, 0.745098],
[0.74902, 0.74902, 0.74902],[0.756863, 0.756863, 0.756863],[0.760784, 0.760784, 0.760784],[0.764706, 0.764706, 0.764706],
[0.768627, 0.768627, 0.768627],[0.776471, 0.776471, 0.776471],[0.780392, 0.780392, 0.780392],[0.784314, 0.784314, 0.784314],
[0.784314, 0.784314, 0.784314],[0.788235, 0.788235, 0.788235],[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],
[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],
[0.788235, 0.788235, 0.788235],[0.788235, 0.788235, 0.788235],[0.784314, 0.784314, 0.784314],[0.780392, 0.780392, 0.780392],
[0.776471, 0.776471, 0.776471],[0.772549, 0.772549, 0.772549],[0.768627, 0.768627, 0.768627],[0.760784, 0.760784, 0.760784],
[0.756863, 0.756863, 0.756863],[0.752941, 0.752941, 0.752941],[0.745098, 0.745098, 0.745098],[0.741176, 0.741176, 0.741176],
[0.733333, 0.733333, 0.733333],[0.729412, 0.729412, 0.729412],[0.721569, 0.72549, 0.72549],[0.717647, 0.717647, 0.717647],
[0.713725, 0.713725, 0.713725],[0.705882, 0.705882, 0.705882],[0.701961, 0.701961, 0.701961],[0.694118, 0.694118, 0.694118],
[0.690196, 0.690196, 0.690196],[0.682353, 0.682353, 0.682353],[0.678431, 0.678431, 0.678431],[0.670588, 0.670588, 0.670588],
[0.666667, 0.666667, 0.666667],[0.658824, 0.658824, 0.658824],[0.654902, 0.654902, 0.654902],[0.65098, 0.65098, 0.65098],
[0.643137, 0.643137, 0.643137],[0.639216, 0.639216, 0.639216],[0.631373, 0.631373, 0.631373],[0.627451, 0.627451, 0.627451],
[0.619608, 0.619608, 0.619608],[0.615686, 0.615686, 0.615686],[0.607843, 0.611765, 0.611765],[0.603922, 0.603922, 0.603922],
[0.6, 0.6, 0.6],[0.592157, 0.592157, 0.592157],[0.588235, 0.588235, 0.588235],[0.580392, 0.580392, 0.580392],
[0.576471, 0.576471, 0.576471],[0.572549, 0.572549, 0.572549],[0.564706, 0.564706, 0.564706],[0.560784, 0.560784, 0.560784],
[0.552941, 0.552941, 0.552941],[0.54902, 0.54902, 0.54902],[0.545098, 0.545098, 0.545098],[0.537255, 0.537255, 0.537255],
[0.533333, 0.533333, 0.533333],[0.52549, 0.52549, 0.52549],[0.521569, 0.521569, 0.521569],[0.517647, 0.517647, 0.517647],
[0.509804, 0.509804, 0.509804],[0.505882, 0.505882, 0.505882],[0.498039, 0.498039, 0.498039],[0.494118, 0.494118, 0.494118],
[0.490196, 0.490196, 0.490196],[0.482353, 0.482353, 0.482353],[0.478431, 0.478431, 0.478431],[0.47451, 0.47451, 0.47451],
[0.466667, 0.466667, 0.466667],[0.462745, 0.462745, 0.462745],[0.458824, 0.458824, 0.458824],[0.45098, 0.45098, 0.45098],
[0.447059, 0.447059, 0.447059],[0.439216, 0.439216, 0.439216],[0.435294, 0.435294, 0.435294],[0.431373, 0.431373, 0.431373],
[0.423529, 0.423529, 0.423529],[0.419608, 0.419608, 0.419608],[0.415686, 0.415686, 0.415686],[0.407843, 0.407843, 0.407843],
[0.403922, 0.403922, 0.403922],[0.4, 0.4, 0.4],[0.392157, 0.392157, 0.392157],[0.388235, 0.388235, 0.388235],
[0.384314, 0.384314, 0.384314],[0.380392, 0.380392, 0.380392],[0.372549, 0.372549, 0.372549],[0.368627, 0.368627, 0.368627],
[0.364706, 0.364706, 0.364706],[0.356863, 0.356863, 0.356863],[0.352941, 0.352941, 0.352941],[0.34902, 0.34902, 0.34902],
[0.341176, 0.341176, 0.341176],[0.337255, 0.337255, 0.337255],[0.333333, 0.333333, 0.333333],[0.329412, 0.329412, 0.329412],
[0.321569, 0.321569, 0.321569],[0.317647, 0.317647, 0.317647],[0.313725, 0.313725, 0.313725],[0.305882, 0.305882, 0.305882],
[0.301961, 0.301961, 0.301961],[0.298039, 0.298039, 0.298039],[0.294118, 0.294118, 0.294118],[0.286275, 0.286275, 0.286275],
[0.282353, 0.282353, 0.282353],[0.278431, 0.278431, 0.278431],[0.27451, 0.27451, 0.27451],[0.266667, 0.266667, 0.266667],
[0.262745, 0.262745, 0.262745],[0.258824, 0.258824, 0.258824],[0.254902, 0.254902, 0.254902],[0.25098, 0.25098, 0.25098],
[0.243137, 0.243137, 0.243137],[0.239216, 0.239216, 0.239216],[0.235294, 0.235294, 0.235294],[0.231373, 0.231373, 0.231373],
[0.227451, 0.227451, 0.227451],[0.219608, 0.219608, 0.219608],[0.215686, 0.215686, 0.215686],[0.211765, 0.211765, 0.211765],
[0.207843, 0.207843, 0.207843],[0.203922, 0.203922, 0.203922],[0.2, 0.2, 0.2],[0.196078, 0.196078, 0.196078],
[0.192157, 0.192157, 0.192157],[0.188235, 0.188235, 0.188235],[0.184314, 0.184314, 0.184314],[0.184314, 0.184314, 0.184314],
[0.180392, 0.180392, 0.180392],[0.180392, 0.180392, 0.180392],[0.176471, 0.176471, 0.176471],[0.176471, 0.176471, 0.176471],
[0.176471, 0.176471, 0.176471],[0.176471, 0.176471, 0.176471],[0.176471, 0.176471, 0.176471],[0.180392, 0.180392, 0.180392],
[0.180392, 0.180392, 0.180392],[0.184314, 0.184314, 0.184314],[0.184314, 0.184314, 0.184314],[0.188235, 0.188235, 0.188235],
[0.192157, 0.192157, 0.192157],[0.196078, 0.196078, 0.196078],[0.2, 0.2, 0.2],[0.203922, 0.203922, 0.203922],
[0.207843, 0.207843, 0.207843],[0.211765, 0.211765, 0.211765],[0.215686, 0.215686, 0.215686],[0.219608, 0.219608, 0.219608],
[0.223529, 0.223529, 0.223529],[0.227451, 0.227451, 0.227451],[0.235294, 0.235294, 0.235294],[0.239216, 0.239216, 0.239216],
[0.243137, 0.243137, 0.243137],[0.247059, 0.247059, 0.247059],[0.25098, 0.254902, 0.25098],[0.258824, 0.258824, 0.258824],
[0.262745, 0.262745, 0.262745],[0.266667, 0.266667, 0.266667],[0.270588, 0.270588, 0.270588],[0.278431, 0.278431, 0.278431],
[0.282353, 0.282353, 0.282353],[0.286275, 0.286275, 0.286275],[0.290196, 0.290196, 0.290196],[0.298039, 0.298039, 0.298039],
[0.301961, 0.301961, 0.301961],[0.305882, 0.305882, 0.305882],[0.309804, 0.309804, 0.309804],[0.317647, 0.317647, 0.317647],
[0.321569, 0.321569, 0.321569],[0.32549, 0.32549, 0.32549],[0.333333, 0.333333, 0.333333],[0.337255, 0.337255, 0.337255],
[0.341176, 0.341176, 0.341176],[0.345098, 0.345098, 0.345098],[0.352941, 0.352941, 0.352941],[0.356863, 0.356863, 0.356863],
[0.360784, 0.360784, 0.360784],[0.368627, 0.368627, 0.368627],[0.372549, 0.372549, 0.372549],[0.376471, 0.376471, 0.376471],
[0.380392, 0.384314, 0.384314],[0.388235, 0.388235, 0.388235],[0.392157, 0.392157, 0.392157],[0.396078, 0.396078, 0.396078],
[0.403922, 0.403922, 0.403922],[0.407843, 0.407843, 0.407843],[0.411765, 0.411765, 0.411765],[0.419608, 0.419608, 0.419608],
[0.423529, 0.423529, 0.423529],[0.427451, 0.427451, 0.427451],[0.435294, 0.435294, 0.435294],[0.439216, 0.439216, 0.439216],
[0.443137, 0.443137, 0.443137],[0.447059, 0.447059, 0.447059],[0.454902, 0.454902, 0.454902],[0.458824, 0.458824, 0.458824],
[0.462745, 0.462745, 0.462745] 
],
[
[0.176471, 0.176471, 0.176471],[0.176471, 0.176471, 0.176471],
[0.180392, 0.180392, 0.180392],[0.180392, 0.180392, 0.180392],[0.184314, 0.184314, 0.184314],[0.184314, 0.184314, 0.184314],
[0.188235, 0.188235, 0.188235],[0.192157, 0.192157, 0.192157],[0.196078, 0.196078, 0.196078],[0.2, 0.2, 0.2],
[0.203922, 0.203922, 0.203922],[0.207843, 0.207843, 0.207843],[0.211765, 0.211765, 0.211765],[0.215686, 0.215686, 0.215686],
[0.219608, 0.219608, 0.219608],[0.223529, 0.223529, 0.223529],[0.227451, 0.227451, 0.227451],[0.235294, 0.235294, 0.235294],
[0.239216, 0.239216, 0.239216],[0.243137, 0.243137, 0.243137],[0.247059, 0.247059, 0.247059],[0.25098, 0.254902, 0.25098],
[0.258824, 0.258824, 0.258824],[0.262745, 0.262745, 0.262745],[0.266667, 0.266667, 0.266667],[0.270588, 0.270588, 0.270588],
[0.278431, 0.278431, 0.278431],[0.282353, 0.282353, 0.282353],[0.286275, 0.286275, 0.286275],[0.290196, 0.290196, 0.290196],
[0.298039, 0.298039, 0.298039],[0.301961, 0.301961, 0.301961],[0.305882, 0.305882, 0.305882],[0.309804, 0.309804, 0.309804],
[0.317647, 0.317647, 0.317647],[0.321569, 0.321569, 0.321569],[0.32549, 0.32549, 0.32549],[0.333333, 0.333333, 0.333333],
[0.337255, 0.337255, 0.337255],[0.341176, 0.341176, 0.341176],[0.345098, 0.345098, 0.345098],[0.352941, 0.352941, 0.352941],
[0.356863, 0.356863, 0.356863],[0.360784, 0.360784, 0.360784],[0.368627, 0.368627, 0.368627],[0.372549, 0.372549, 0.372549],
[0.376471, 0.376471, 0.376471],[0.380392, 0.384314, 0.384314],[0.388235, 0.388235, 0.388235],[0.392157, 0.392157, 0.392157],
[0.396078, 0.396078, 0.396078],[0.403922, 0.403922, 0.403922],[0.407843, 0.407843, 0.407843],[0.411765, 0.411765, 0.411765],
[0.419608, 0.419608, 0.419608],[0.423529, 0.423529, 0.423529],[0.427451, 0.427451, 0.427451],[0.435294, 0.435294, 0.435294],
[0.439216, 0.439216, 0.439216],[0.443137, 0.443137, 0.443137],[0.447059, 0.447059, 0.447059],[0.454902, 0.454902, 0.454902],
[0.458824, 0.458824, 0.458824],[0.462745, 0.462745, 0.462745],[0.470588, 0.470588, 0.470588],[0.47451, 0.47451, 0.47451],
[0.478431, 0.478431, 0.478431],[0.482353, 0.482353, 0.482353],[0.490196, 0.490196, 0.490196],[0.494118, 0.494118, 0.494118],
[0.498039, 0.498039, 0.498039],[0.505882, 0.505882, 0.505882],[0.509804, 0.509804, 0.509804],[0.513725, 0.513725, 0.513725],
[0.521569, 0.521569, 0.521569],[0.52549, 0.52549, 0.52549],[0.529412, 0.529412, 0.529412],[0.537255, 0.537255, 0.537255],
[0.541176, 0.541176, 0.541176],[0.54902, 0.54902, 0.54902],[0.552941, 0.552941, 0.552941],[0.556863, 0.556863, 0.556863],
[0.564706, 0.564706, 0.564706],[0.568627, 0.568627, 0.568627],[0.576471, 0.576471, 0.576471],[0.580392, 0.580392, 0.580392],
[0.584314, 0.584314, 0.584314],[0.592157, 0.592157, 0.592157],[0.596078, 0.596078, 0.596078],[0.603922, 0.603922, 0.603922],
[0.607843, 0.607843, 0.607843],[0.615686, 0.615686, 0.615686],[0.619608, 0.619608, 0.619608],[0.623529, 0.623529, 0.623529],
[0.631373, 0.631373, 0.631373],[0.635294, 0.635294, 0.635294],[0.643137, 0.643137, 0.643137],[0.647059, 0.647059, 0.647059],
[0.654902, 0.654902, 0.654902],[0.658824, 0.658824, 0.658824],[0.666667, 0.666667, 0.666667],[0.670588, 0.670588, 0.670588],
[0.67451, 0.67451, 0.67451],[0.682353, 0.682353, 0.682353],[0.686275, 0.686275, 0.686275],[0.694118, 0.694118, 0.694118],
[0.698039, 0.698039, 0.698039],[0.705882, 0.705882, 0.705882],[0.709804, 0.709804, 0.709804],[0.717647, 0.717647, 0.717647],
[0.721569, 0.721569, 0.721569],[0.729412, 0.729412, 0.729412],[0.733333, 0.733333, 0.733333],[0.737255, 0.741176, 0.741176],
[0.745098, 0.745098, 0.745098],[0.74902, 0.74902, 0.74902],[0.756863, 0.756863, 0.756863],[0.760784, 0.760784, 0.760784],
[0.764706, 0.764706, 0.764706],[0.768627, 0.768627, 0.768627],[0.776471, 0.776471, 0.776471],[0.780392, 0.780392, 0.780392],
[0.784314, 0.784314, 0.784314],[0.784314, 0.784314, 0.784314],[0.788235, 0.788235, 0.788235],[0.792157, 0.792157, 0.792157],
[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],[0.792157, 0.792157, 0.792157],
[0.792157, 0.792157, 0.792157],[0.788235, 0.788235, 0.788235],[0.788235, 0.788235, 0.788235],[0.784314, 0.784314, 0.784314],
[0.780392, 0.780392, 0.780392],[0.776471, 0.776471, 0.776471],[0.772549, 0.772549, 0.772549],[0.768627, 0.768627, 0.768627],
[0.760784, 0.760784, 0.760784],[0.756863, 0.756863, 0.756863],[0.752941, 0.752941, 0.752941],[0.745098, 0.745098, 0.745098],
[0.741176, 0.741176, 0.741176],[0.733333, 0.733333, 0.733333],[0.729412, 0.729412, 0.729412],[0.721569, 0.72549, 0.72549],
[0.717647, 0.717647, 0.717647],[0.713725, 0.713725, 0.713725],[0.705882, 0.705882, 0.705882],[0.701961, 0.701961, 0.701961],
[0.694118, 0.694118, 0.694118],[0.690196, 0.690196, 0.690196],[0.682353, 0.682353, 0.682353],[0.678431, 0.678431, 0.678431],
[0.670588, 0.670588, 0.670588],[0.666667, 0.666667, 0.666667],[0.658824, 0.658824, 0.658824],[0.654902, 0.654902, 0.654902],
[0.65098, 0.65098, 0.65098],[0.643137, 0.643137, 0.643137],[0.639216, 0.639216, 0.639216],[0.631373, 0.631373, 0.631373],
[0.627451, 0.627451, 0.627451],[0.619608, 0.619608, 0.619608],[0.615686, 0.615686, 0.615686],[0.607843, 0.611765, 0.611765],
[0.603922, 0.603922, 0.603922],[0.6, 0.6, 0.6],[0.592157, 0.592157, 0.592157],[0.588235, 0.588235, 0.588235],
[0.580392, 0.580392, 0.580392],[0.576471, 0.576471, 0.576471],[0.572549, 0.572549, 0.572549],[0.564706, 0.564706, 0.564706],
[0.560784, 0.560784, 0.560784],[0.552941, 0.552941, 0.552941],[0.54902, 0.54902, 0.54902],[0.545098, 0.545098, 0.545098],
[0.537255, 0.537255, 0.537255],[0.533333, 0.533333, 0.533333],[0.52549, 0.52549, 0.52549],[0.521569, 0.521569, 0.521569],
[0.517647, 0.517647, 0.517647],[0.509804, 0.509804, 0.509804],[0.505882, 0.505882, 0.505882],[0.498039, 0.498039, 0.498039],
[0.494118, 0.494118, 0.494118],[0.490196, 0.490196, 0.490196],[0.482353, 0.482353, 0.482353],[0.478431, 0.478431, 0.478431],
[0.47451, 0.47451, 0.47451],[0.466667, 0.466667, 0.466667],[0.462745, 0.462745, 0.462745],[0.458824, 0.458824, 0.458824],
[0.45098, 0.45098, 0.45098],[0.447059, 0.447059, 0.447059],[0.439216, 0.439216, 0.439216],[0.435294, 0.435294, 0.435294],
[0.431373, 0.431373, 0.431373],[0.423529, 0.423529, 0.423529],[0.419608, 0.419608, 0.419608],[0.415686, 0.415686, 0.415686],
[0.407843, 0.407843, 0.407843],[0.403922, 0.403922, 0.403922],[0.4, 0.4, 0.4],[0.392157, 0.392157, 0.392157],
[0.388235, 0.388235, 0.388235],[0.384314, 0.384314, 0.384314],[0.380392, 0.380392, 0.380392],[0.372549, 0.372549, 0.372549],
[0.368627, 0.368627, 0.368627],[0.364706, 0.364706, 0.364706],[0.356863, 0.356863, 0.356863],[0.352941, 0.352941, 0.352941],
[0.34902, 0.34902, 0.34902],[0.341176, 0.341176, 0.341176],[0.337255, 0.337255, 0.337255],[0.333333, 0.333333, 0.333333],
[0.329412, 0.329412, 0.329412],[0.321569, 0.321569, 0.321569],[0.317647, 0.317647, 0.317647],[0.313725, 0.313725, 0.313725],
[0.305882, 0.305882, 0.305882],[0.301961, 0.301961, 0.301961],[0.298039, 0.298039, 0.298039],[0.294118, 0.294118, 0.294118],
[0.286275, 0.286275, 0.286275],[0.282353, 0.282353, 0.282353],[0.278431, 0.278431, 0.278431],[0.27451, 0.27451, 0.27451],
[0.266667, 0.266667, 0.266667],[0.262745, 0.262745, 0.262745],[0.258824, 0.258824, 0.258824],[0.254902, 0.254902, 0.254902],
[0.25098, 0.25098, 0.25098],[0.243137, 0.243137, 0.243137],[0.239216, 0.239216, 0.239216],[0.235294, 0.235294, 0.235294],
[0.231373, 0.231373, 0.231373],[0.227451, 0.227451, 0.227451],[0.219608, 0.219608, 0.219608],[0.215686, 0.215686, 0.215686],
[0.211765, 0.211765, 0.211765],[0.207843, 0.207843, 0.207843],[0.203922, 0.203922, 0.203922],[0.2, 0.2, 0.2],
[0.196078, 0.196078, 0.196078],[0.192157, 0.192157, 0.192157],[0.188235, 0.188235, 0.188235],[0.184314, 0.184314, 0.184314],
[0.184314, 0.184314, 0.184314],[0.180392, 0.180392, 0.180392],[0.180392, 0.180392, 0.180392],[0.176471, 0.176471, 0.176471],
[0.176471, 0.176471, 0.176471],[0.176471, 0.176471, 0.176471] 
], 
[
[0.976471, 0.521569, 0.972549],
[0.980392, 0.517647, 0.964706],[0.984314, 0.513725, 0.952941],[0.988235, 0.509804, 0.945098],[0.988235, 0.505882, 0.933333],
[0.992157, 0.498039, 0.921569],[0.992157, 0.494118, 0.909804],[0.992157, 0.486275, 0.898039],[0.992157, 0.478431, 0.886275],
[0.992157, 0.470588, 0.870588],[0.992157, 0.462745, 0.858824],[0.988235, 0.454902, 0.843137],[0.988235, 0.447059, 0.831373],
[0.984314, 0.439216, 0.815686],[0.984314, 0.427451, 0.8],[0.980392, 0.419608, 0.784314],[0.980392, 0.411765, 0.772549],
[0.976471, 0.403922, 0.756863],[0.972549, 0.392157, 0.741176],[0.968627, 0.384314, 0.72549],[0.968627, 0.372549, 0.709804],
[0.964706, 0.364706, 0.698039],[0.960784, 0.356863, 0.682353],[0.956863, 0.345098, 0.666667],[0.952941, 0.337255, 0.65098],
[0.94902, 0.329412, 0.639216],[0.945098, 0.317647, 0.623529],[0.941176, 0.309804, 0.607843],[0.937255, 0.298039, 0.592157],
[0.933333, 0.290196, 0.580392],[0.929412, 0.282353, 0.564706],[0.92549, 0.27451, 0.54902],[0.917647, 0.262745, 0.533333],
[0.913725, 0.254902, 0.517647],[0.909804, 0.247059, 0.505882],[0.901961, 0.239216, 0.490196],[0.898039, 0.231373, 0.47451],
[0.890196, 0.223529, 0.458824],[0.882353, 0.215686, 0.443137],[0.878431, 0.207843, 0.427451],[0.870588, 0.2, 0.415686],
[0.862745, 0.192157, 0.4],[0.854902, 0.184314, 0.384314],[0.85098, 0.176471, 0.368627],[0.843137, 0.172549, 0.352941],
[0.835294, 0.164706, 0.337255],[0.827451, 0.156863, 0.32549],[0.819608, 0.14902, 0.309804],[0.811765, 0.141176, 0.294118],
[0.803922, 0.137255, 0.278431],[0.796078, 0.129412, 0.266667],[0.788235, 0.121569, 0.25098],[0.780392, 0.117647, 0.235294],
[0.776471, 0.109804, 0.223529],[0.768627, 0.105882, 0.207843],[0.760784, 0.0980392, 0.196078],[0.752941, 0.0941176, 0.180392],
[0.745098, 0.0901961, 0.168627],[0.741176, 0.0901961, 0.152941],[0.733333, 0.0901961, 0.141176],[0.729412, 0.0901961, 0.129412],
[0.72549, 0.0901961, 0.117647],[0.721569, 0.0941176, 0.105882],[0.717647, 0.0980392, 0.0980392],[0.713725, 0.105882, 0.0862745],
[0.709804, 0.113725, 0.0784314],[0.709804, 0.121569, 0.0666667],[0.709804, 0.129412, 0.0588235],[0.709804, 0.137255, 0.0509804],
[0.709804, 0.14902, 0.0470588],[0.709804, 0.160784, 0.0392157],[0.713725, 0.168627, 0.0313725],[0.713725, 0.180392, 0.027451],
[0.717647, 0.192157, 0.0235294],[0.721569, 0.2, 0.0235294],[0.721569, 0.211765, 0.0196078],[0.72549, 0.223529, 0.0196078],
[0.729412, 0.231373, 0.0196078],[0.733333, 0.243137, 0.0156863],[0.737255, 0.254902, 0.0156863],[0.741176, 0.262745, 0.0156863],
[0.745098, 0.27451, 0.0156863],[0.74902, 0.282353, 0.0156863],[0.752941, 0.294118, 0.0156863],[0.756863, 0.301961, 0.0156863],
[0.760784, 0.313725, 0.0156863],[0.764706, 0.321569, 0.0156863],[0.768627, 0.329412, 0.0156863],[0.772549, 0.341176, 0.0156863],
[0.776471, 0.34902, 0.0196078],[0.780392, 0.360784, 0.0196078],[0.784314, 0.368627, 0.0196078],[0.788235, 0.376471, 0.0196078],
[0.792157, 0.388235, 0.0196078],[0.792157, 0.396078, 0.0196078],[0.796078, 0.403922, 0.0196078],[0.8, 0.415686, 0.0196078],
[0.803922, 0.423529, 0.0196078],[0.807843, 0.435294, 0.0196078],[0.807843, 0.443137, 0.0196078],[0.811765, 0.45098, 0.0196078],
[0.815686, 0.462745, 0.0196078],[0.815686, 0.470588, 0.0196078],[0.819608, 0.478431, 0.0196078],[0.819608, 0.490196, 0.0196078],
[0.823529, 0.498039, 0.0156863],[0.823529, 0.505882, 0.0156863],[0.827451, 0.517647, 0.0156863],[0.827451, 0.52549, 0.0156863],
[0.831373, 0.537255, 0.0156863],[0.831373, 0.545098, 0.0156863],[0.835294, 0.552941, 0.0156863],[0.835294, 0.560784, 0.0156863],
[0.839216, 0.572549, 0.0156863],[0.839216, 0.580392, 0.0196078],[0.839216, 0.588235, 0.0196078],[0.839216, 0.596078, 0.0235294],
[0.843137, 0.607843, 0.027451],[0.843137, 0.615686, 0.0313725],[0.843137, 0.623529, 0.0392157],[0.843137, 0.627451, 0.0509804],
[0.839216, 0.635294, 0.0588235],[0.839216, 0.643137, 0.0705882],[0.835294, 0.647059, 0.0823529],[0.835294, 0.654902, 0.0941176],
[0.831373, 0.658824, 0.109804],[0.823529, 0.662745, 0.121569],[0.819608, 0.666667, 0.137255],[0.815686, 0.666667, 0.152941],
[0.807843, 0.666667, 0.168627],[0.8, 0.670588, 0.184314],[0.792157, 0.670588, 0.2],[0.780392, 0.666667, 0.215686],
[0.772549, 0.666667, 0.231373],[0.760784, 0.662745, 0.247059],[0.74902, 0.662745, 0.262745],[0.737255, 0.658824, 0.27451],
[0.72549, 0.654902, 0.290196],[0.713725, 0.65098, 0.305882],[0.701961, 0.647059, 0.321569],[0.686275, 0.643137, 0.333333],
[0.67451, 0.639216, 0.34902],[0.662745, 0.631373, 0.360784],[0.647059, 0.627451, 0.376471],[0.631373, 0.623529, 0.388235],
[0.615686, 0.619608, 0.4],[0.603922, 0.615686, 0.411765],[0.588235, 0.607843, 0.427451],[0.572549, 0.603922, 0.439216],
[0.556863, 0.6, 0.45098],[0.537255, 0.596078, 0.462745],[0.521569, 0.588235, 0.47451],[0.505882, 0.584314, 0.486275],
[0.486275, 0.580392, 0.498039],[0.466667, 0.576471, 0.509804],[0.45098, 0.568627, 0.521569],[0.431373, 0.564706, 0.533333],
[0.411765, 0.560784, 0.545098],[0.388235, 0.552941, 0.556863],[0.368627, 0.54902, 0.568627],[0.34902, 0.545098, 0.580392],
[0.32549, 0.537255, 0.592157],[0.305882, 0.533333, 0.603922],[0.282353, 0.52549, 0.615686],[0.262745, 0.517647, 0.627451],
[0.243137, 0.513725, 0.639216],[0.223529, 0.505882, 0.65098],[0.203922, 0.498039, 0.662745],[0.184314, 0.490196, 0.67451],
[0.168627, 0.482353, 0.686275],[0.156863, 0.47451, 0.698039],[0.145098, 0.462745, 0.709804],[0.137255, 0.454902, 0.721569],
[0.133333, 0.443137, 0.733333],[0.133333, 0.435294, 0.745098],[0.133333, 0.423529, 0.756863],[0.133333, 0.411765, 0.772549],
[0.141176, 0.4, 0.784314],[0.145098, 0.388235, 0.796078],[0.14902, 0.376471, 0.807843],[0.156863, 0.364706, 0.819608],
[0.160784, 0.352941, 0.831373],[0.168627, 0.341176, 0.843137],[0.176471, 0.329412, 0.854902],[0.180392, 0.317647, 0.866667],
[0.188235, 0.305882, 0.878431],[0.192157, 0.294118, 0.890196],[0.2, 0.286275, 0.898039],[0.207843, 0.27451, 0.909804],
[0.215686, 0.266667, 0.917647],[0.223529, 0.258824, 0.92549],[0.235294, 0.25098, 0.933333],[0.243137, 0.247059, 0.941176],
[0.254902, 0.243137, 0.94902],[0.266667, 0.239216, 0.952941],[0.278431, 0.239216, 0.956863],[0.286275, 0.239216, 0.960784],
[0.298039, 0.239216, 0.964706],[0.309804, 0.243137, 0.968627],[0.321569, 0.247059, 0.968627],[0.333333, 0.25098, 0.972549],
[0.345098, 0.254902, 0.972549],[0.356863, 0.262745, 0.976471],[0.368627, 0.270588, 0.976471],[0.380392, 0.278431, 0.976471],
[0.388235, 0.286275, 0.976471],[0.4, 0.294118, 0.976471],[0.411765, 0.301961, 0.976471],[0.419608, 0.309804, 0.976471],
[0.431373, 0.317647, 0.976471],[0.439216, 0.32549, 0.976471],[0.45098, 0.333333, 0.976471],[0.458824, 0.341176, 0.976471],
[0.466667, 0.34902, 0.976471],[0.478431, 0.356863, 0.980392],[0.486275, 0.364706, 0.980392],[0.498039, 0.368627, 0.980392],
[0.509804, 0.376471, 0.980392],[0.517647, 0.384314, 0.980392],[0.529412, 0.392157, 0.980392],[0.541176, 0.396078, 0.980392],
[0.552941, 0.403922, 0.980392],[0.564706, 0.411765, 0.980392],[0.576471, 0.415686, 0.980392],[0.588235, 0.419608, 0.984314],
[0.6, 0.427451, 0.984314],[0.611765, 0.431373, 0.984314],[0.627451, 0.435294, 0.984314],[0.639216, 0.443137, 0.984314],
[0.654902, 0.447059, 0.988235],[0.666667, 0.45098, 0.988235],[0.682353, 0.454902, 0.988235],[0.694118, 0.458824, 0.988235],
[0.709804, 0.462745, 0.992157],[0.72549, 0.466667, 0.992157],[0.737255, 0.470588, 0.992157],[0.752941, 0.47451, 0.996078],
[0.764706, 0.478431, 0.996078],[0.780392, 0.482353, 0.996078],[0.792157, 0.486275, 1],[0.807843, 0.490196, 1],
[0.819608, 0.494118, 1],[0.835294, 0.498039, 1],[0.847059, 0.501961, 1],[0.858824, 0.505882, 1],
[0.87451, 0.509804, 1],[0.886275, 0.513725, 1],[0.898039, 0.517647, 1],[0.909804, 0.517647, 1],
[0.917647, 0.521569, 1],[0.929412, 0.521569, 1],[0.941176, 0.52549, 0.996078],[0.94902, 0.52549, 0.992157],
[0.956863, 0.52549, 0.988235],[0.964706, 0.52549, 0.984314],[0.968627, 0.521569, 0.976471] 
], 
[
[0.243137, 0.247059, 0.941176],[0.254902, 0.243137, 0.94902],[0.266667, 0.239216, 0.952941],[0.278431, 0.239216, 0.956863],
[0.286275, 0.239216, 0.960784],[0.298039, 0.239216, 0.964706],[0.309804, 0.243137, 0.968627],[0.321569, 0.247059, 0.968627],
[0.333333, 0.25098, 0.972549],[0.345098, 0.254902, 0.972549],[0.356863, 0.262745, 0.976471],[0.368627, 0.270588, 0.976471],
[0.380392, 0.278431, 0.976471],[0.388235, 0.286275, 0.976471],[0.4, 0.294118, 0.976471],[0.411765, 0.301961, 0.976471],
[0.419608, 0.309804, 0.976471],[0.431373, 0.317647, 0.976471],[0.439216, 0.32549, 0.976471],[0.45098, 0.333333, 0.976471],
[0.458824, 0.341176, 0.976471],[0.466667, 0.34902, 0.976471],[0.478431, 0.356863, 0.980392],[0.486275, 0.364706, 0.980392],
[0.498039, 0.368627, 0.980392],[0.509804, 0.376471, 0.980392],[0.517647, 0.384314, 0.980392],[0.529412, 0.392157, 0.980392],
[0.541176, 0.396078, 0.980392],[0.552941, 0.403922, 0.980392],[0.564706, 0.411765, 0.980392],[0.576471, 0.415686, 0.980392],
[0.588235, 0.419608, 0.984314],[0.6, 0.427451, 0.984314],[0.611765, 0.431373, 0.984314],[0.627451, 0.435294, 0.984314],
[0.639216, 0.443137, 0.984314],[0.654902, 0.447059, 0.988235],[0.666667, 0.45098, 0.988235],[0.682353, 0.454902, 0.988235],
[0.694118, 0.458824, 0.988235],[0.709804, 0.462745, 0.992157],[0.72549, 0.466667, 0.992157],[0.737255, 0.470588, 0.992157],
[0.752941, 0.47451, 0.996078],[0.764706, 0.478431, 0.996078],[0.780392, 0.482353, 0.996078],[0.792157, 0.486275, 1],
[0.807843, 0.490196, 1],[0.819608, 0.494118, 1],[0.835294, 0.498039, 1],[0.847059, 0.501961, 1],
[0.858824, 0.505882, 1],[0.87451, 0.509804, 1],[0.886275, 0.513725, 1],[0.898039, 0.517647, 1],
[0.909804, 0.517647, 1],[0.917647, 0.521569, 1],[0.929412, 0.521569, 1],[0.941176, 0.52549, 0.996078],
[0.94902, 0.52549, 0.992157],[0.956863, 0.52549, 0.988235],[0.964706, 0.52549, 0.984314],[0.968627, 0.521569, 0.976471],
[0.976471, 0.521569, 0.972549],[0.980392, 0.517647, 0.964706],[0.984314, 0.513725, 0.952941],[0.988235, 0.509804, 0.945098],
[0.988235, 0.505882, 0.933333],[0.992157, 0.498039, 0.921569],[0.992157, 0.494118, 0.909804],[0.992157, 0.486275, 0.898039],
[0.992157, 0.478431, 0.886275],[0.992157, 0.470588, 0.870588],[0.992157, 0.462745, 0.858824],[0.988235, 0.454902, 0.843137],
[0.988235, 0.447059, 0.831373],[0.984314, 0.439216, 0.815686],[0.984314, 0.427451, 0.8],[0.980392, 0.419608, 0.784314],
[0.980392, 0.411765, 0.772549],[0.976471, 0.403922, 0.756863],[0.972549, 0.392157, 0.741176],[0.968627, 0.384314, 0.72549],
[0.968627, 0.372549, 0.709804],[0.964706, 0.364706, 0.698039],[0.960784, 0.356863, 0.682353],[0.956863, 0.345098, 0.666667],
[0.952941, 0.337255, 0.65098],[0.94902, 0.329412, 0.639216],[0.945098, 0.317647, 0.623529],[0.941176, 0.309804, 0.607843],
[0.937255, 0.298039, 0.592157],[0.933333, 0.290196, 0.580392],[0.929412, 0.282353, 0.564706],[0.92549, 0.27451, 0.54902],
[0.917647, 0.262745, 0.533333],[0.913725, 0.254902, 0.517647],[0.909804, 0.247059, 0.505882],[0.901961, 0.239216, 0.490196],
[0.898039, 0.231373, 0.47451],[0.890196, 0.223529, 0.458824],[0.882353, 0.215686, 0.443137],[0.878431, 0.207843, 0.427451],
[0.870588, 0.2, 0.415686],[0.862745, 0.192157, 0.4],[0.854902, 0.184314, 0.384314],[0.85098, 0.176471, 0.368627],
[0.843137, 0.172549, 0.352941],[0.835294, 0.164706, 0.337255],[0.827451, 0.156863, 0.32549],[0.819608, 0.14902, 0.309804],
[0.811765, 0.141176, 0.294118],[0.803922, 0.137255, 0.278431],[0.796078, 0.129412, 0.266667],[0.788235, 0.121569, 0.25098],
[0.780392, 0.117647, 0.235294],[0.776471, 0.109804, 0.223529],[0.768627, 0.105882, 0.207843],[0.760784, 0.0980392, 0.196078],
[0.752941, 0.0941176, 0.180392],[0.745098, 0.0901961, 0.168627],[0.741176, 0.0901961, 0.152941],[0.733333, 0.0901961, 0.141176],
[0.729412, 0.0901961, 0.129412],[0.72549, 0.0901961, 0.117647],[0.721569, 0.0941176, 0.105882],[0.717647, 0.0980392, 0.0980392],
[0.713725, 0.105882, 0.0862745],[0.709804, 0.113725, 0.0784314],[0.709804, 0.121569, 0.0666667],[0.709804, 0.129412, 0.0588235],
[0.709804, 0.137255, 0.0509804],[0.709804, 0.14902, 0.0470588],[0.709804, 0.160784, 0.0392157],[0.713725, 0.168627, 0.0313725],
[0.713725, 0.180392, 0.027451],[0.717647, 0.192157, 0.0235294],[0.721569, 0.2, 0.0235294],[0.721569, 0.211765, 0.0196078],
[0.72549, 0.223529, 0.0196078],[0.729412, 0.231373, 0.0196078],[0.733333, 0.243137, 0.0156863],[0.737255, 0.254902, 0.0156863],
[0.741176, 0.262745, 0.0156863],[0.745098, 0.27451, 0.0156863],[0.74902, 0.282353, 0.0156863],[0.752941, 0.294118, 0.0156863],
[0.756863, 0.301961, 0.0156863],[0.760784, 0.313725, 0.0156863],[0.764706, 0.321569, 0.0156863],[0.768627, 0.329412, 0.0156863],
[0.772549, 0.341176, 0.0156863],[0.776471, 0.34902, 0.0196078],[0.780392, 0.360784, 0.0196078],[0.784314, 0.368627, 0.0196078],
[0.788235, 0.376471, 0.0196078],[0.792157, 0.388235, 0.0196078],[0.792157, 0.396078, 0.0196078],[0.796078, 0.403922, 0.0196078],
[0.8, 0.415686, 0.0196078],[0.803922, 0.423529, 0.0196078],[0.807843, 0.435294, 0.0196078],[0.807843, 0.443137, 0.0196078],
[0.811765, 0.45098, 0.0196078],[0.815686, 0.462745, 0.0196078],[0.815686, 0.470588, 0.0196078],[0.819608, 0.478431, 0.0196078],
[0.819608, 0.490196, 0.0196078],[0.823529, 0.498039, 0.0156863],[0.823529, 0.505882, 0.0156863],[0.827451, 0.517647, 0.0156863],
[0.827451, 0.52549, 0.0156863],[0.831373, 0.537255, 0.0156863],[0.831373, 0.545098, 0.0156863],[0.835294, 0.552941, 0.0156863],
[0.835294, 0.560784, 0.0156863],[0.839216, 0.572549, 0.0156863],[0.839216, 0.580392, 0.0196078],[0.839216, 0.588235, 0.0196078],
[0.839216, 0.596078, 0.0235294],[0.843137, 0.607843, 0.027451],[0.843137, 0.615686, 0.0313725],[0.843137, 0.623529, 0.0392157],
[0.843137, 0.627451, 0.0509804],[0.839216, 0.635294, 0.0588235],[0.839216, 0.643137, 0.0705882],[0.835294, 0.647059, 0.0823529],
[0.835294, 0.654902, 0.0941176],[0.831373, 0.658824, 0.109804],[0.823529, 0.662745, 0.121569],[0.819608, 0.666667, 0.137255],
[0.815686, 0.666667, 0.152941],[0.807843, 0.666667, 0.168627],[0.8, 0.670588, 0.184314],[0.792157, 0.670588, 0.2],
[0.780392, 0.666667, 0.215686],[0.772549, 0.666667, 0.231373],[0.760784, 0.662745, 0.247059],[0.74902, 0.662745, 0.262745],
[0.737255, 0.658824, 0.27451],[0.72549, 0.654902, 0.290196],[0.713725, 0.65098, 0.305882],[0.701961, 0.647059, 0.321569],
[0.686275, 0.643137, 0.333333],[0.67451, 0.639216, 0.34902],[0.662745, 0.631373, 0.360784],[0.647059, 0.627451, 0.376471],
[0.631373, 0.623529, 0.388235],[0.615686, 0.619608, 0.4],[0.603922, 0.615686, 0.411765],[0.588235, 0.607843, 0.427451],
[0.572549, 0.603922, 0.439216],[0.556863, 0.6, 0.45098],[0.537255, 0.596078, 0.462745],[0.521569, 0.588235, 0.47451],
[0.505882, 0.584314, 0.486275],[0.486275, 0.580392, 0.498039],[0.466667, 0.576471, 0.509804],[0.45098, 0.568627, 0.521569],
[0.431373, 0.564706, 0.533333],[0.411765, 0.560784, 0.545098],[0.388235, 0.552941, 0.556863],[0.368627, 0.54902, 0.568627],
[0.34902, 0.545098, 0.580392],[0.32549, 0.537255, 0.592157],[0.305882, 0.533333, 0.603922],[0.282353, 0.52549, 0.615686],
[0.262745, 0.517647, 0.627451],[0.243137, 0.513725, 0.639216],[0.223529, 0.505882, 0.65098],[0.203922, 0.498039, 0.662745],
[0.184314, 0.490196, 0.67451],[0.168627, 0.482353, 0.686275],[0.156863, 0.47451, 0.698039],[0.145098, 0.462745, 0.709804],
[0.137255, 0.454902, 0.721569],[0.133333, 0.443137, 0.733333],[0.133333, 0.435294, 0.745098],[0.133333, 0.423529, 0.756863],
[0.133333, 0.411765, 0.772549],[0.141176, 0.4, 0.784314],[0.145098, 0.388235, 0.796078],[0.14902, 0.376471, 0.807843],
[0.156863, 0.364706, 0.819608],[0.160784, 0.352941, 0.831373],[0.168627, 0.341176, 0.843137],[0.176471, 0.329412, 0.854902],
[0.180392, 0.317647, 0.866667],[0.188235, 0.305882, 0.878431],[0.192157, 0.294118, 0.890196],[0.2, 0.286275, 0.898039],
[0.207843, 0.27451, 0.909804],[0.215686, 0.266667, 0.917647],[0.223529, 0.258824, 0.92549],[0.235294, 0.25098, 0.933333] ], 
[[0.937255, 0.333333, 0.94902],[0.945098, 0.341176, 0.941176],[0.94902, 0.352941, 0.937255],
[0.952941, 0.364706, 0.929412],[0.960784, 0.372549, 0.921569],[0.964706, 0.384314, 0.913725],[0.964706, 0.396078, 0.905882],
[0.968627, 0.407843, 0.898039],[0.972549, 0.419608, 0.886275],[0.976471, 0.435294, 0.878431],[0.976471, 0.447059, 0.870588],
[0.980392, 0.458824, 0.858824],[0.980392, 0.470588, 0.85098],[0.980392, 0.482353, 0.839216],[0.984314, 0.494118, 0.827451],
[0.984314, 0.505882, 0.819608],[0.984314, 0.517647, 0.807843],[0.984314, 0.529412, 0.796078],[0.984314, 0.541176, 0.788235],
[0.988235, 0.552941, 0.776471],[0.988235, 0.564706, 0.764706],[0.988235, 0.576471, 0.752941],[0.988235, 0.584314, 0.745098],
[0.988235, 0.596078, 0.733333],[0.988235, 0.607843, 0.721569],[0.988235, 0.619608, 0.709804],[0.988235, 0.627451, 0.701961],
[0.988235, 0.639216, 0.690196],[0.988235, 0.65098, 0.678431],[0.988235, 0.658824, 0.666667],[0.988235, 0.670588, 0.654902],
[0.988235, 0.678431, 0.647059],[0.988235, 0.690196, 0.635294],[0.984314, 0.698039, 0.623529],[0.984314, 0.709804, 0.611765],
[0.984314, 0.717647, 0.6],[0.984314, 0.72549, 0.588235],[0.988235, 0.737255, 0.576471],[0.988235, 0.745098, 0.564706],
[0.988235, 0.752941, 0.552941],[0.988235, 0.764706, 0.541176],[0.988235, 0.772549, 0.529412],[0.988235, 0.780392, 0.517647],
[0.988235, 0.788235, 0.505882],[0.988235, 0.8, 0.494118],[0.988235, 0.807843, 0.482353],[0.988235, 0.815686, 0.470588],
[0.988235, 0.823529, 0.458824],[0.988235, 0.831373, 0.443137],[0.988235, 0.839216, 0.431373],[0.988235, 0.847059, 0.419608],
[0.988235, 0.854902, 0.403922],[0.988235, 0.866667, 0.392157],[0.984314, 0.87451, 0.376471],[0.984314, 0.878431, 0.360784],
[0.984314, 0.886275, 0.345098],[0.980392, 0.894118, 0.333333],[0.980392, 0.901961, 0.317647],[0.976471, 0.905882, 0.301961],
[0.972549, 0.913725, 0.286275],[0.968627, 0.917647, 0.270588],[0.960784, 0.921569, 0.254902],[0.956863, 0.92549, 0.239216],
[0.94902, 0.929412, 0.223529],[0.945098, 0.933333, 0.207843],[0.937255, 0.933333, 0.196078],[0.929412, 0.933333, 0.180392],
[0.917647, 0.937255, 0.168627],[0.909804, 0.933333, 0.156863],[0.901961, 0.933333, 0.145098],[0.890196, 0.933333, 0.137255],
[0.882353, 0.929412, 0.129412],[0.870588, 0.929412, 0.121569],[0.858824, 0.92549, 0.113725],[0.847059, 0.921569, 0.105882],
[0.839216, 0.917647, 0.101961],[0.827451, 0.913725, 0.0980392],[0.815686, 0.909804, 0.0941176],[0.803922, 0.905882, 0.0941176],
[0.792157, 0.901961, 0.0901961],[0.780392, 0.898039, 0.0862745],[0.768627, 0.894118, 0.0862745],[0.756863, 0.890196, 0.0823529],
[0.745098, 0.886275, 0.0823529],[0.733333, 0.882353, 0.0823529],[0.721569, 0.878431, 0.0784314],[0.709804, 0.87451, 0.0784314],
[0.698039, 0.870588, 0.0745098],[0.686275, 0.866667, 0.0745098],[0.67451, 0.862745, 0.0745098],[0.662745, 0.854902, 0.0705882],
[0.65098, 0.85098, 0.0705882],[0.639216, 0.847059, 0.0705882],[0.627451, 0.843137, 0.0666667],[0.615686, 0.839216, 0.0666667],
[0.603922, 0.835294, 0.0666667],[0.592157, 0.831373, 0.0627451],[0.580392, 0.827451, 0.0627451],[0.568627, 0.823529, 0.0627451],
[0.552941, 0.815686, 0.0588235],[0.541176, 0.811765, 0.0588235],[0.529412, 0.807843, 0.0588235],[0.517647, 0.803922, 0.054902],
[0.505882, 0.8, 0.054902],[0.490196, 0.796078, 0.054902],[0.478431, 0.792157, 0.0509804],[0.466667, 0.784314, 0.0509804],
[0.45098, 0.780392, 0.0509804],[0.439216, 0.776471, 0.0509804],[0.423529, 0.772549, 0.0470588],[0.411765, 0.768627, 0.0470588],
[0.396078, 0.764706, 0.0470588],[0.384314, 0.760784, 0.0470588],[0.368627, 0.752941, 0.0470588],[0.352941, 0.74902, 0.0470588],
[0.337255, 0.745098, 0.0470588],[0.32549, 0.741176, 0.0509804],[0.309804, 0.737255, 0.054902],[0.294118, 0.729412, 0.0588235],
[0.282353, 0.72549, 0.0627451],[0.266667, 0.721569, 0.0705882],[0.254902, 0.717647, 0.0784314],[0.243137, 0.709804, 0.0862745],
[0.231373, 0.705882, 0.0980392],[0.219608, 0.701961, 0.105882],[0.211765, 0.694118, 0.117647],[0.203922, 0.690196, 0.129412],
[0.2, 0.682353, 0.145098],[0.196078, 0.678431, 0.156863],[0.196078, 0.670588, 0.172549],[0.196078, 0.662745, 0.184314],
[0.196078, 0.658824, 0.2],[0.2, 0.65098, 0.215686],[0.207843, 0.643137, 0.231373],[0.211765, 0.639216, 0.247059],
[0.219608, 0.631373, 0.258824],[0.223529, 0.623529, 0.27451],[0.231373, 0.615686, 0.290196],[0.235294, 0.607843, 0.305882],
[0.243137, 0.6, 0.321569],[0.247059, 0.596078, 0.337255],[0.25098, 0.588235, 0.34902],[0.254902, 0.580392, 0.364706],
[0.258824, 0.572549, 0.380392],[0.262745, 0.564706, 0.392157],[0.266667, 0.556863, 0.407843],[0.266667, 0.54902, 0.419608],
[0.270588, 0.541176, 0.435294],[0.270588, 0.537255, 0.447059],[0.270588, 0.529412, 0.462745],[0.270588, 0.521569, 0.47451],
[0.270588, 0.513725, 0.490196],[0.270588, 0.505882, 0.501961],[0.266667, 0.498039, 0.517647],[0.266667, 0.490196, 0.529412],
[0.262745, 0.482353, 0.541176],[0.258824, 0.47451, 0.556863],[0.254902, 0.470588, 0.568627],[0.25098, 0.462745, 0.580392],
[0.247059, 0.454902, 0.592157],[0.243137, 0.447059, 0.607843],[0.235294, 0.439216, 0.619608],[0.231373, 0.431373, 0.631373],
[0.223529, 0.423529, 0.643137],[0.219608, 0.415686, 0.654902],[0.211765, 0.407843, 0.666667],[0.203922, 0.4, 0.678431],
[0.2, 0.392157, 0.690196],[0.192157, 0.380392, 0.701961],[0.188235, 0.372549, 0.713725],[0.184314, 0.364706, 0.72549],
[0.176471, 0.356863, 0.733333],[0.172549, 0.345098, 0.745098],[0.168627, 0.337255, 0.756863],[0.164706, 0.32549, 0.764706],
[0.164706, 0.317647, 0.776471],[0.160784, 0.305882, 0.784314],[0.156863, 0.298039, 0.796078],[0.152941, 0.286275, 0.803922],
[0.152941, 0.27451, 0.815686],[0.14902, 0.262745, 0.823529],[0.14902, 0.25098, 0.835294],[0.145098, 0.243137, 0.843137],
[0.145098, 0.231373, 0.85098],[0.141176, 0.219608, 0.858824],[0.141176, 0.207843, 0.870588],[0.145098, 0.196078, 0.878431],
[0.145098, 0.184314, 0.886275],[0.14902, 0.172549, 0.894118],[0.156863, 0.160784, 0.898039],[0.160784, 0.152941, 0.905882],
[0.172549, 0.141176, 0.913725],[0.180392, 0.133333, 0.917647],[0.192157, 0.12549, 0.92549],[0.203922, 0.117647, 0.929412],
[0.215686, 0.113725, 0.933333],[0.231373, 0.109804, 0.937255],[0.243137, 0.105882, 0.941176],[0.258824, 0.105882, 0.945098],
[0.270588, 0.105882, 0.94902],[0.286275, 0.105882, 0.94902],[0.301961, 0.109804, 0.952941],[0.313725, 0.109804, 0.952941],
[0.329412, 0.113725, 0.956863],[0.341176, 0.121569, 0.956863],[0.352941, 0.12549, 0.960784],[0.368627, 0.129412, 0.960784],
[0.380392, 0.137255, 0.964706],[0.392157, 0.141176, 0.964706],[0.403922, 0.14902, 0.964706],[0.415686, 0.152941, 0.968627],
[0.427451, 0.160784, 0.968627],[0.439216, 0.164706, 0.968627],[0.45098, 0.172549, 0.972549],[0.458824, 0.176471, 0.972549],
[0.470588, 0.180392, 0.972549],[0.482353, 0.188235, 0.976471],[0.494118, 0.192157, 0.976471],[0.501961, 0.196078, 0.976471],
[0.513725, 0.203922, 0.980392],[0.52549, 0.207843, 0.980392],[0.537255, 0.211765, 0.980392],[0.54902, 0.215686, 0.980392],
[0.560784, 0.219608, 0.984314],[0.568627, 0.223529, 0.984314],[0.580392, 0.227451, 0.984314],[0.596078, 0.231373, 0.984314],
[0.607843, 0.235294, 0.984314],[0.619608, 0.239216, 0.984314],[0.631373, 0.243137, 0.984314],[0.643137, 0.243137, 0.984314],
[0.654902, 0.247059, 0.984314],[0.670588, 0.25098, 0.984314],[0.682353, 0.25098, 0.984314],[0.694118, 0.254902, 0.984314],
[0.709804, 0.254902, 0.984314],[0.721569, 0.258824, 0.984314],[0.733333, 0.258824, 0.984314],[0.74902, 0.262745, 0.984314],
[0.760784, 0.262745, 0.984314],[0.772549, 0.266667, 0.984314],[0.784314, 0.266667, 0.980392],[0.796078, 0.270588, 0.980392],
[0.811765, 0.270588, 0.980392],[0.823529, 0.27451, 0.980392],[0.835294, 0.27451, 0.980392],[0.847059, 0.278431, 0.976471],
[0.858824, 0.282353, 0.976471],[0.866667, 0.282353, 0.976471],[0.878431, 0.286275, 0.972549],[0.890196, 0.294118, 0.972549],
[0.898039, 0.298039, 0.968627],[0.905882, 0.301961, 0.964706],[0.913725, 0.309804, 0.960784],[0.92549, 0.317647, 0.956863],
[0.929412, 0.32549, 0.952941] ], [[0.180392, 0.133333, 0.917647],[0.192157, 0.12549, 0.92549],
[0.203922, 0.117647, 0.929412],[0.215686, 0.113725, 0.933333],[0.231373, 0.109804, 0.937255],[0.243137, 0.105882, 0.941176],
[0.258824, 0.105882, 0.945098],[0.270588, 0.105882, 0.94902],[0.286275, 0.105882, 0.94902],[0.301961, 0.109804, 0.952941],
[0.313725, 0.109804, 0.952941],[0.329412, 0.113725, 0.956863],[0.341176, 0.121569, 0.956863],[0.352941, 0.12549, 0.960784],
[0.368627, 0.129412, 0.960784],[0.380392, 0.137255, 0.964706],[0.392157, 0.141176, 0.964706],[0.403922, 0.14902, 0.964706],
[0.415686, 0.152941, 0.968627],[0.427451, 0.160784, 0.968627],[0.439216, 0.164706, 0.968627],[0.45098, 0.172549, 0.972549],
[0.458824, 0.176471, 0.972549],[0.470588, 0.180392, 0.972549],[0.482353, 0.188235, 0.976471],[0.494118, 0.192157, 0.976471],
[0.501961, 0.196078, 0.976471],[0.513725, 0.203922, 0.980392],[0.52549, 0.207843, 0.980392],[0.537255, 0.211765, 0.980392],
[0.54902, 0.215686, 0.980392],[0.560784, 0.219608, 0.984314],[0.568627, 0.223529, 0.984314],[0.580392, 0.227451, 0.984314],
[0.596078, 0.231373, 0.984314],[0.607843, 0.235294, 0.984314],[0.619608, 0.239216, 0.984314],[0.631373, 0.243137, 0.984314],
[0.643137, 0.243137, 0.984314],[0.654902, 0.247059, 0.984314],[0.670588, 0.25098, 0.984314],[0.682353, 0.25098, 0.984314],
[0.694118, 0.254902, 0.984314],[0.709804, 0.254902, 0.984314],[0.721569, 0.258824, 0.984314],[0.733333, 0.258824, 0.984314],
[0.74902, 0.262745, 0.984314],[0.760784, 0.262745, 0.984314],[0.772549, 0.266667, 0.984314],[0.784314, 0.266667, 0.980392],
[0.796078, 0.270588, 0.980392],[0.811765, 0.270588, 0.980392],[0.823529, 0.27451, 0.980392],[0.835294, 0.27451, 0.980392],
[0.847059, 0.278431, 0.976471],[0.858824, 0.282353, 0.976471],[0.866667, 0.282353, 0.976471],[0.878431, 0.286275, 0.972549],
[0.890196, 0.294118, 0.972549],[0.898039, 0.298039, 0.968627],[0.905882, 0.301961, 0.964706],[0.913725, 0.309804, 0.960784],
[0.92549, 0.317647, 0.956863],[0.929412, 0.32549, 0.952941],[0.937255, 0.333333, 0.94902],[0.945098, 0.341176, 0.941176],
[0.94902, 0.352941, 0.937255],[0.952941, 0.364706, 0.929412],[0.960784, 0.372549, 0.921569],[0.964706, 0.384314, 0.913725],
[0.964706, 0.396078, 0.905882],[0.968627, 0.407843, 0.898039],[0.972549, 0.419608, 0.886275],[0.976471, 0.435294, 0.878431],
[0.976471, 0.447059, 0.870588],[0.980392, 0.458824, 0.858824],[0.980392, 0.470588, 0.85098],[0.980392, 0.482353, 0.839216],
[0.984314, 0.494118, 0.827451],[0.984314, 0.505882, 0.819608],[0.984314, 0.517647, 0.807843],[0.984314, 0.529412, 0.796078],
[0.984314, 0.541176, 0.788235],[0.988235, 0.552941, 0.776471],[0.988235, 0.564706, 0.764706],[0.988235, 0.576471, 0.752941],
[0.988235, 0.584314, 0.745098],[0.988235, 0.596078, 0.733333],[0.988235, 0.607843, 0.721569],[0.988235, 0.619608, 0.709804],
[0.988235, 0.627451, 0.701961],[0.988235, 0.639216, 0.690196],[0.988235, 0.65098, 0.678431],[0.988235, 0.658824, 0.666667],
[0.988235, 0.670588, 0.654902],[0.988235, 0.678431, 0.647059],[0.988235, 0.690196, 0.635294],[0.984314, 0.698039, 0.623529],
[0.984314, 0.709804, 0.611765],[0.984314, 0.717647, 0.6],[0.984314, 0.72549, 0.588235],[0.988235, 0.737255, 0.576471],
[0.988235, 0.745098, 0.564706],[0.988235, 0.752941, 0.552941],[0.988235, 0.764706, 0.541176],[0.988235, 0.772549, 0.529412],
[0.988235, 0.780392, 0.517647],[0.988235, 0.788235, 0.505882],[0.988235, 0.8, 0.494118],[0.988235, 0.807843, 0.482353],
[0.988235, 0.815686, 0.470588],[0.988235, 0.823529, 0.458824],[0.988235, 0.831373, 0.443137],[0.988235, 0.839216, 0.431373],
[0.988235, 0.847059, 0.419608],[0.988235, 0.854902, 0.403922],[0.988235, 0.866667, 0.392157],[0.984314, 0.87451, 0.376471],
[0.984314, 0.878431, 0.360784],[0.984314, 0.886275, 0.345098],[0.980392, 0.894118, 0.333333],[0.980392, 0.901961, 0.317647],
[0.976471, 0.905882, 0.301961],[0.972549, 0.913725, 0.286275],[0.968627, 0.917647, 0.270588],[0.960784, 0.921569, 0.254902],
[0.956863, 0.92549, 0.239216],[0.94902, 0.929412, 0.223529],[0.945098, 0.933333, 0.207843],[0.937255, 0.933333, 0.196078],
[0.929412, 0.933333, 0.180392],[0.917647, 0.937255, 0.168627],[0.909804, 0.933333, 0.156863],[0.901961, 0.933333, 0.145098],
[0.890196, 0.933333, 0.137255],[0.882353, 0.929412, 0.129412],[0.870588, 0.929412, 0.121569],[0.858824, 0.92549, 0.113725],
[0.847059, 0.921569, 0.105882],[0.839216, 0.917647, 0.101961],[0.827451, 0.913725, 0.0980392],[0.815686, 0.909804, 0.0941176],
[0.803922, 0.905882, 0.0941176],[0.792157, 0.901961, 0.0901961],[0.780392, 0.898039, 0.0862745],[0.768627, 0.894118, 0.0862745],
[0.756863, 0.890196, 0.0823529],[0.745098, 0.886275, 0.0823529],[0.733333, 0.882353, 0.0823529],[0.721569, 0.878431, 0.0784314],
[0.709804, 0.87451, 0.0784314],[0.698039, 0.870588, 0.0745098],[0.686275, 0.866667, 0.0745098],[0.67451, 0.862745, 0.0745098],
[0.662745, 0.854902, 0.0705882],[0.65098, 0.85098, 0.0705882],[0.639216, 0.847059, 0.0705882],[0.627451, 0.843137, 0.0666667],
[0.615686, 0.839216, 0.0666667],[0.603922, 0.835294, 0.0666667],[0.592157, 0.831373, 0.0627451],[0.580392, 0.827451, 0.0627451],
[0.568627, 0.823529, 0.0627451],[0.552941, 0.815686, 0.0588235],[0.541176, 0.811765, 0.0588235],[0.529412, 0.807843, 0.0588235],
[0.517647, 0.803922, 0.054902],[0.505882, 0.8, 0.054902],[0.490196, 0.796078, 0.054902],[0.478431, 0.792157, 0.0509804],
[0.466667, 0.784314, 0.0509804],[0.45098, 0.780392, 0.0509804],[0.439216, 0.776471, 0.0509804],[0.423529, 0.772549, 0.0470588],
[0.411765, 0.768627, 0.0470588],[0.396078, 0.764706, 0.0470588],[0.384314, 0.760784, 0.0470588],[0.368627, 0.752941, 0.0470588],
[0.352941, 0.74902, 0.0470588],[0.337255, 0.745098, 0.0470588],[0.32549, 0.741176, 0.0509804],[0.309804, 0.737255, 0.054902],
[0.294118, 0.729412, 0.0588235],[0.282353, 0.72549, 0.0627451],[0.266667, 0.721569, 0.0705882],[0.254902, 0.717647, 0.0784314],
[0.243137, 0.709804, 0.0862745],[0.231373, 0.705882, 0.0980392],[0.219608, 0.701961, 0.105882],[0.211765, 0.694118, 0.117647],
[0.203922, 0.690196, 0.129412],[0.2, 0.682353, 0.145098],[0.196078, 0.678431, 0.156863],[0.196078, 0.670588, 0.172549],
[0.196078, 0.662745, 0.184314],[0.196078, 0.658824, 0.2],[0.2, 0.65098, 0.215686],[0.207843, 0.643137, 0.231373],
[0.211765, 0.639216, 0.247059],[0.219608, 0.631373, 0.258824],[0.223529, 0.623529, 0.27451],[0.231373, 0.615686, 0.290196],
[0.235294, 0.607843, 0.305882],[0.243137, 0.6, 0.321569],[0.247059, 0.596078, 0.337255],[0.25098, 0.588235, 0.34902],
[0.254902, 0.580392, 0.364706],[0.258824, 0.572549, 0.380392],[0.262745, 0.564706, 0.392157],[0.266667, 0.556863, 0.407843],
[0.266667, 0.54902, 0.419608],[0.270588, 0.541176, 0.435294],[0.270588, 0.537255, 0.447059],[0.270588, 0.529412, 0.462745],
[0.270588, 0.521569, 0.47451],[0.270588, 0.513725, 0.490196],[0.270588, 0.505882, 0.501961],[0.266667, 0.498039, 0.517647],
[0.266667, 0.490196, 0.529412],[0.262745, 0.482353, 0.541176],[0.258824, 0.47451, 0.556863],[0.254902, 0.470588, 0.568627],
[0.25098, 0.462745, 0.580392],[0.247059, 0.454902, 0.592157],[0.243137, 0.447059, 0.607843],[0.235294, 0.439216, 0.619608],
[0.231373, 0.431373, 0.631373],[0.223529, 0.423529, 0.643137],[0.219608, 0.415686, 0.654902],[0.211765, 0.407843, 0.666667],
[0.203922, 0.4, 0.678431],[0.2, 0.392157, 0.690196],[0.192157, 0.380392, 0.701961],[0.188235, 0.372549, 0.713725],
[0.184314, 0.364706, 0.72549],[0.176471, 0.356863, 0.733333],[0.172549, 0.345098, 0.745098],[0.168627, 0.337255, 0.756863],
[0.164706, 0.32549, 0.764706],[0.164706, 0.317647, 0.776471],[0.160784, 0.305882, 0.784314],[0.156863, 0.298039, 0.796078],
[0.152941, 0.286275, 0.803922],[0.152941, 0.27451, 0.815686],[0.14902, 0.262745, 0.823529],[0.14902, 0.25098, 0.835294],
[0.145098, 0.243137, 0.843137],[0.145098, 0.231373, 0.85098],[0.141176, 0.219608, 0.858824],[0.141176, 0.207843, 0.870588],
[0.145098, 0.196078, 0.878431],[0.145098, 0.184314, 0.886275],[0.14902, 0.172549, 0.894118],[0.156863, 0.160784, 0.898039],
[0.160784, 0.152941, 0.905882],[0.172549, 0.141176, 0.913725] ], [[0.87451, 0.835294, 0.847059],
[0.878431, 0.835294, 0.839216],[0.882353, 0.831373, 0.831373],[0.886275, 0.827451, 0.823529],[0.890196, 0.819608, 0.811765],
[0.890196, 0.815686, 0.8],[0.894118, 0.807843, 0.788235],[0.898039, 0.8, 0.776471],[0.898039, 0.792157, 0.764706],
[0.898039, 0.784314, 0.752941],[0.901961, 0.772549, 0.741176],[0.901961, 0.764706, 0.72549],[0.901961, 0.752941, 0.713725],
[0.901961, 0.741176, 0.698039],[0.901961, 0.733333, 0.686275],[0.901961, 0.721569, 0.670588],[0.901961, 0.709804, 0.658824],
[0.901961, 0.698039, 0.643137],[0.901961, 0.690196, 0.631373],[0.901961, 0.678431, 0.615686],[0.901961, 0.666667, 0.603922],
[0.901961, 0.654902, 0.588235],[0.898039, 0.643137, 0.576471],[0.898039, 0.635294, 0.560784],[0.898039, 0.623529, 0.54902],
[0.898039, 0.611765, 0.533333],[0.894118, 0.6, 0.521569],[0.894118, 0.588235, 0.505882],[0.890196, 0.576471, 0.494118],
[0.890196, 0.568627, 0.482353],[0.886275, 0.556863, 0.466667],[0.886275, 0.545098, 0.454902],[0.882353, 0.533333, 0.439216],
[0.882353, 0.521569, 0.427451],[0.878431, 0.509804, 0.415686],[0.878431, 0.498039, 0.4],[0.87451, 0.486275, 0.388235],
[0.870588, 0.478431, 0.376471],[0.866667, 0.466667, 0.360784],[0.866667, 0.454902, 0.34902],[0.862745, 0.443137, 0.337255],
[0.858824, 0.431373, 0.321569],[0.854902, 0.419608, 0.309804],[0.854902, 0.407843, 0.298039],[0.85098, 0.396078, 0.282353],
[0.847059, 0.384314, 0.270588],[0.843137, 0.368627, 0.258824],[0.839216, 0.356863, 0.247059],[0.835294, 0.345098, 0.235294],
[0.831373, 0.333333, 0.219608],[0.827451, 0.321569, 0.207843],[0.823529, 0.309804, 0.196078],[0.819608, 0.298039, 0.184314],
[0.815686, 0.282353, 0.172549],[0.811765, 0.270588, 0.160784],[0.807843, 0.258824, 0.14902],[0.803922, 0.247059, 0.141176],
[0.803922, 0.235294, 0.129412],[0.8, 0.227451, 0.121569],[0.796078, 0.215686, 0.113725],[0.792157, 0.207843, 0.105882],
[0.792157, 0.2, 0.0980392],[0.792157, 0.196078, 0.0941176],[0.788235, 0.192157, 0.0901961],[0.788235, 0.192157, 0.0901961],
[0.788235, 0.192157, 0.0901961],[0.788235, 0.192157, 0.0941176],[0.792157, 0.196078, 0.0941176],[0.792157, 0.203922, 0.101961],
[0.796078, 0.211765, 0.105882],[0.796078, 0.219608, 0.113725],[0.8, 0.227451, 0.121569],[0.803922, 0.239216, 0.133333],
[0.807843, 0.25098, 0.141176],[0.811765, 0.262745, 0.152941],[0.811765, 0.27451, 0.164706],[0.815686, 0.286275, 0.176471],
[0.819608, 0.298039, 0.188235],[0.823529, 0.313725, 0.2],[0.827451, 0.32549, 0.211765],[0.831373, 0.337255, 0.223529],
[0.835294, 0.34902, 0.235294],[0.839216, 0.360784, 0.25098],[0.843137, 0.372549, 0.262745],[0.847059, 0.384314, 0.27451],
[0.85098, 0.396078, 0.286275],[0.854902, 0.411765, 0.301961],[0.858824, 0.423529, 0.313725],[0.858824, 0.435294, 0.32549],
[0.862745, 0.447059, 0.337255],[0.866667, 0.454902, 0.352941],[0.870588, 0.466667, 0.364706],[0.870588, 0.478431, 0.376471],
[0.87451, 0.490196, 0.392157],[0.878431, 0.501961, 0.403922],[0.878431, 0.513725, 0.415686],[0.882353, 0.52549, 0.431373],
[0.882353, 0.537255, 0.443137],[0.886275, 0.54902, 0.458824],[0.890196, 0.560784, 0.470588],[0.890196, 0.568627, 0.482353],
[0.890196, 0.580392, 0.498039],[0.894118, 0.592157, 0.509804],[0.894118, 0.603922, 0.52549],[0.898039, 0.615686, 0.537255],
[0.898039, 0.627451, 0.552941],[0.898039, 0.635294, 0.564706],[0.901961, 0.647059, 0.580392],[0.901961, 0.658824, 0.592157],
[0.901961, 0.670588, 0.607843],[0.901961, 0.682353, 0.619608],[0.901961, 0.690196, 0.635294],[0.901961, 0.701961, 0.647059],
[0.901961, 0.713725, 0.662745],[0.901961, 0.72549, 0.67451],[0.901961, 0.733333, 0.690196],[0.901961, 0.745098, 0.701961],
[0.901961, 0.752941, 0.717647],[0.901961, 0.764706, 0.729412],[0.898039, 0.772549, 0.745098],[0.898039, 0.780392, 0.756863],
[0.894118, 0.792157, 0.768627],[0.890196, 0.796078, 0.780392],[0.886275, 0.803922, 0.796078],[0.882353, 0.811765, 0.803922],
[0.878431, 0.815686, 0.815686],[0.87451, 0.819608, 0.827451],[0.870588, 0.819608, 0.835294],[0.862745, 0.823529, 0.843137],
[0.858824, 0.823529, 0.85098],[0.85098, 0.819608, 0.858824],[0.843137, 0.819608, 0.866667],[0.835294, 0.815686, 0.870588],
[0.827451, 0.811765, 0.87451],[0.819608, 0.807843, 0.878431],[0.811765, 0.8, 0.882353],[0.803922, 0.796078, 0.886275],
[0.792157, 0.788235, 0.886275],[0.784314, 0.780392, 0.890196],[0.776471, 0.772549, 0.890196],[0.764706, 0.764706, 0.894118],
[0.756863, 0.756863, 0.894118],[0.745098, 0.74902, 0.894118],[0.737255, 0.741176, 0.894118],[0.72549, 0.733333, 0.894118],
[0.717647, 0.721569, 0.894118],[0.705882, 0.713725, 0.898039],[0.698039, 0.705882, 0.898039],[0.686275, 0.698039, 0.898039],
[0.678431, 0.686275, 0.898039],[0.666667, 0.678431, 0.898039],[0.654902, 0.670588, 0.898039],[0.647059, 0.662745, 0.898039],
[0.635294, 0.654902, 0.898039],[0.623529, 0.643137, 0.898039],[0.615686, 0.635294, 0.898039],[0.603922, 0.627451, 0.898039],
[0.592157, 0.619608, 0.898039],[0.580392, 0.611765, 0.898039],[0.568627, 0.6, 0.898039],[0.560784, 0.592157, 0.898039],
[0.54902, 0.584314, 0.898039],[0.537255, 0.576471, 0.898039],[0.52549, 0.568627, 0.898039],[0.513725, 0.560784, 0.898039],
[0.498039, 0.552941, 0.898039],[0.486275, 0.541176, 0.898039],[0.47451, 0.533333, 0.898039],[0.462745, 0.52549, 0.898039],
[0.447059, 0.517647, 0.898039],[0.435294, 0.509804, 0.898039],[0.419608, 0.501961, 0.898039],[0.407843, 0.494118, 0.898039],
[0.392157, 0.486275, 0.898039],[0.376471, 0.478431, 0.898039],[0.360784, 0.470588, 0.898039],[0.345098, 0.462745, 0.898039],
[0.329412, 0.454902, 0.898039],[0.309804, 0.447059, 0.898039],[0.294118, 0.439216, 0.898039],[0.27451, 0.431373, 0.898039],
[0.254902, 0.427451, 0.898039],[0.235294, 0.419608, 0.898039],[0.215686, 0.415686, 0.898039],[0.196078, 0.407843, 0.898039],
[0.176471, 0.403922, 0.898039],[0.156863, 0.4, 0.898039],[0.137255, 0.396078, 0.898039],[0.121569, 0.392157, 0.898039],
[0.109804, 0.392157, 0.898039],[0.101961, 0.388235, 0.898039],[0.0980392, 0.388235, 0.898039],[0.101961, 0.388235, 0.898039],
[0.113725, 0.392157, 0.898039],[0.12549, 0.392157, 0.898039],[0.141176, 0.396078, 0.898039],[0.160784, 0.4, 0.898039],
[0.180392, 0.403922, 0.898039],[0.2, 0.407843, 0.898039],[0.219608, 0.415686, 0.898039],[0.239216, 0.419608, 0.898039],
[0.258824, 0.427451, 0.898039],[0.278431, 0.435294, 0.898039],[0.298039, 0.443137, 0.898039],[0.313725, 0.447059, 0.898039],
[0.333333, 0.454902, 0.898039],[0.34902, 0.462745, 0.898039],[0.364706, 0.470588, 0.898039],[0.380392, 0.478431, 0.898039],
[0.396078, 0.486275, 0.898039],[0.407843, 0.494118, 0.898039],[0.423529, 0.501961, 0.898039],[0.439216, 0.513725, 0.898039],
[0.45098, 0.521569, 0.898039],[0.462745, 0.529412, 0.898039],[0.478431, 0.537255, 0.898039],[0.490196, 0.545098, 0.898039],
[0.501961, 0.552941, 0.898039],[0.513725, 0.560784, 0.898039],[0.52549, 0.568627, 0.898039],[0.537255, 0.576471, 0.898039],
[0.54902, 0.588235, 0.898039],[0.560784, 0.596078, 0.898039],[0.572549, 0.603922, 0.898039],[0.584314, 0.611765, 0.898039],
[0.596078, 0.619608, 0.898039],[0.607843, 0.627451, 0.898039],[0.615686, 0.639216, 0.898039],[0.627451, 0.647059, 0.898039],
[0.639216, 0.654902, 0.898039],[0.647059, 0.662745, 0.898039],[0.658824, 0.670588, 0.898039],[0.670588, 0.682353, 0.898039],
[0.678431, 0.690196, 0.898039],[0.690196, 0.698039, 0.898039],[0.698039, 0.705882, 0.898039],[0.709804, 0.717647, 0.898039],
[0.717647, 0.72549, 0.894118],[0.729412, 0.733333, 0.894118],[0.737255, 0.741176, 0.894118],[0.74902, 0.752941, 0.894118],
[0.756863, 0.760784, 0.894118],[0.768627, 0.768627, 0.894118],[0.776471, 0.776471, 0.894118],[0.784314, 0.784314, 0.890196],
[0.796078, 0.792157, 0.890196],[0.803922, 0.8, 0.890196],[0.811765, 0.807843, 0.886275],[0.819608, 0.815686, 0.886275],
[0.827451, 0.819608, 0.882353],[0.835294, 0.823529, 0.878431],[0.843137, 0.827451, 0.878431],[0.85098, 0.831373, 0.87451],
[0.854902, 0.835294, 0.866667],[0.862745, 0.835294, 0.862745],[0.866667, 0.835294, 0.854902] ], [
[0.101961, 0.388235, 0.898039],[0.113725, 0.392157, 0.898039],[0.12549, 0.392157, 0.898039],[0.141176, 0.396078, 0.898039],
[0.160784, 0.4, 0.898039],[0.180392, 0.403922, 0.898039],[0.2, 0.407843, 0.898039],[0.219608, 0.415686, 0.898039],
[0.239216, 0.419608, 0.898039],[0.258824, 0.427451, 0.898039],[0.278431, 0.435294, 0.898039],[0.298039, 0.443137, 0.898039],
[0.313725, 0.447059, 0.898039],[0.333333, 0.454902, 0.898039],[0.34902, 0.462745, 0.898039],[0.364706, 0.470588, 0.898039],
[0.380392, 0.478431, 0.898039],[0.396078, 0.486275, 0.898039],[0.407843, 0.494118, 0.898039],[0.423529, 0.501961, 0.898039],
[0.439216, 0.513725, 0.898039],[0.45098, 0.521569, 0.898039],[0.462745, 0.529412, 0.898039],[0.478431, 0.537255, 0.898039],
[0.490196, 0.545098, 0.898039],[0.501961, 0.552941, 0.898039],[0.513725, 0.560784, 0.898039],[0.52549, 0.568627, 0.898039],
[0.537255, 0.576471, 0.898039],[0.54902, 0.588235, 0.898039],[0.560784, 0.596078, 0.898039],[0.572549, 0.603922, 0.898039],
[0.584314, 0.611765, 0.898039],[0.596078, 0.619608, 0.898039],[0.607843, 0.627451, 0.898039],[0.615686, 0.639216, 0.898039],
[0.627451, 0.647059, 0.898039],[0.639216, 0.654902, 0.898039],[0.647059, 0.662745, 0.898039],[0.658824, 0.670588, 0.898039],
[0.670588, 0.682353, 0.898039],[0.678431, 0.690196, 0.898039],[0.690196, 0.698039, 0.898039],[0.698039, 0.705882, 0.898039],
[0.709804, 0.717647, 0.898039],[0.717647, 0.72549, 0.894118],[0.729412, 0.733333, 0.894118],[0.737255, 0.741176, 0.894118],
[0.74902, 0.752941, 0.894118],[0.756863, 0.760784, 0.894118],[0.768627, 0.768627, 0.894118],[0.776471, 0.776471, 0.894118],
[0.784314, 0.784314, 0.890196],[0.796078, 0.792157, 0.890196],[0.803922, 0.8, 0.890196],[0.811765, 0.807843, 0.886275],
[0.819608, 0.815686, 0.886275],[0.827451, 0.819608, 0.882353],[0.835294, 0.823529, 0.878431],[0.843137, 0.827451, 0.878431],
[0.85098, 0.831373, 0.87451],[0.854902, 0.835294, 0.866667],[0.862745, 0.835294, 0.862745],[0.866667, 0.835294, 0.854902],
[0.87451, 0.835294, 0.847059],[0.878431, 0.835294, 0.839216],[0.882353, 0.831373, 0.831373],[0.886275, 0.827451, 0.823529],
[0.890196, 0.819608, 0.811765],[0.890196, 0.815686, 0.8],[0.894118, 0.807843, 0.788235],[0.898039, 0.8, 0.776471],
[0.898039, 0.792157, 0.764706],[0.898039, 0.784314, 0.752941],[0.901961, 0.772549, 0.741176],[0.901961, 0.764706, 0.72549],
[0.901961, 0.752941, 0.713725],[0.901961, 0.741176, 0.698039],[0.901961, 0.733333, 0.686275],[0.901961, 0.721569, 0.670588],
[0.901961, 0.709804, 0.658824],[0.901961, 0.698039, 0.643137],[0.901961, 0.690196, 0.631373],[0.901961, 0.678431, 0.615686],
[0.901961, 0.666667, 0.603922],[0.901961, 0.654902, 0.588235],[0.898039, 0.643137, 0.576471],[0.898039, 0.635294, 0.560784],
[0.898039, 0.623529, 0.54902],[0.898039, 0.611765, 0.533333],[0.894118, 0.6, 0.521569],[0.894118, 0.588235, 0.505882],
[0.890196, 0.576471, 0.494118],[0.890196, 0.568627, 0.482353],[0.886275, 0.556863, 0.466667],[0.886275, 0.545098, 0.454902],
[0.882353, 0.533333, 0.439216],[0.882353, 0.521569, 0.427451],[0.878431, 0.509804, 0.415686],[0.878431, 0.498039, 0.4],
[0.87451, 0.486275, 0.388235],[0.870588, 0.478431, 0.376471],[0.866667, 0.466667, 0.360784],[0.866667, 0.454902, 0.34902],
[0.862745, 0.443137, 0.337255],[0.858824, 0.431373, 0.321569],[0.854902, 0.419608, 0.309804],[0.854902, 0.407843, 0.298039],
[0.85098, 0.396078, 0.282353],[0.847059, 0.384314, 0.270588],[0.843137, 0.368627, 0.258824],[0.839216, 0.356863, 0.247059],
[0.835294, 0.345098, 0.235294],[0.831373, 0.333333, 0.219608],[0.827451, 0.321569, 0.207843],[0.823529, 0.309804, 0.196078],
[0.819608, 0.298039, 0.184314],[0.815686, 0.282353, 0.172549],[0.811765, 0.270588, 0.160784],[0.807843, 0.258824, 0.14902],
[0.803922, 0.247059, 0.141176],[0.803922, 0.235294, 0.129412],[0.8, 0.227451, 0.121569],[0.796078, 0.215686, 0.113725],
[0.792157, 0.207843, 0.105882],[0.792157, 0.2, 0.0980392],[0.792157, 0.196078, 0.0941176],[0.788235, 0.192157, 0.0901961],
[0.788235, 0.192157, 0.0901961],[0.788235, 0.192157, 0.0901961],[0.788235, 0.192157, 0.0941176],[0.792157, 0.196078, 0.0941176],
[0.792157, 0.203922, 0.101961],[0.796078, 0.211765, 0.105882],[0.796078, 0.219608, 0.113725],[0.8, 0.227451, 0.121569],
[0.803922, 0.239216, 0.133333],[0.807843, 0.25098, 0.141176],[0.811765, 0.262745, 0.152941],[0.811765, 0.27451, 0.164706],
[0.815686, 0.286275, 0.176471],[0.819608, 0.298039, 0.188235],[0.823529, 0.313725, 0.2],[0.827451, 0.32549, 0.211765],
[0.831373, 0.337255, 0.223529],[0.835294, 0.34902, 0.235294],[0.839216, 0.360784, 0.25098],[0.843137, 0.372549, 0.262745],
[0.847059, 0.384314, 0.27451],[0.85098, 0.396078, 0.286275],[0.854902, 0.411765, 0.301961],[0.858824, 0.423529, 0.313725],
[0.858824, 0.435294, 0.32549],[0.862745, 0.447059, 0.337255],[0.866667, 0.454902, 0.352941],[0.870588, 0.466667, 0.364706],
[0.870588, 0.478431, 0.376471],[0.87451, 0.490196, 0.392157],[0.878431, 0.501961, 0.403922],[0.878431, 0.513725, 0.415686],
[0.882353, 0.52549, 0.431373],[0.882353, 0.537255, 0.443137],[0.886275, 0.54902, 0.458824],[0.890196, 0.560784, 0.470588],
[0.890196, 0.568627, 0.482353],[0.890196, 0.580392, 0.498039],[0.894118, 0.592157, 0.509804],[0.894118, 0.603922, 0.52549],
[0.898039, 0.615686, 0.537255],[0.898039, 0.627451, 0.552941],[0.898039, 0.635294, 0.564706],[0.901961, 0.647059, 0.580392],
[0.901961, 0.658824, 0.592157],[0.901961, 0.670588, 0.607843],[0.901961, 0.682353, 0.619608],[0.901961, 0.690196, 0.635294],
[0.901961, 0.701961, 0.647059],[0.901961, 0.713725, 0.662745],[0.901961, 0.72549, 0.67451],[0.901961, 0.733333, 0.690196],
[0.901961, 0.745098, 0.701961],[0.901961, 0.752941, 0.717647],[0.901961, 0.764706, 0.729412],[0.898039, 0.772549, 0.745098],
[0.898039, 0.780392, 0.756863],[0.894118, 0.792157, 0.768627],[0.890196, 0.796078, 0.780392],[0.886275, 0.803922, 0.796078],
[0.882353, 0.811765, 0.803922],[0.878431, 0.815686, 0.815686],[0.87451, 0.819608, 0.827451],[0.870588, 0.819608, 0.835294],
[0.862745, 0.823529, 0.843137],[0.858824, 0.823529, 0.85098],[0.85098, 0.819608, 0.858824],[0.843137, 0.819608, 0.866667],
[0.835294, 0.815686, 0.870588],[0.827451, 0.811765, 0.87451],[0.819608, 0.807843, 0.878431],[0.811765, 0.8, 0.882353],
[0.803922, 0.796078, 0.886275],[0.792157, 0.788235, 0.886275],[0.784314, 0.780392, 0.890196],[0.776471, 0.772549, 0.890196],
[0.764706, 0.764706, 0.894118],[0.756863, 0.756863, 0.894118],[0.745098, 0.74902, 0.894118],[0.737255, 0.741176, 0.894118],
[0.72549, 0.733333, 0.894118],[0.717647, 0.721569, 0.894118],[0.705882, 0.713725, 0.898039],[0.698039, 0.705882, 0.898039],
[0.686275, 0.698039, 0.898039],[0.678431, 0.686275, 0.898039],[0.666667, 0.678431, 0.898039],[0.654902, 0.670588, 0.898039],
[0.647059, 0.662745, 0.898039],[0.635294, 0.654902, 0.898039],[0.623529, 0.643137, 0.898039],[0.615686, 0.635294, 0.898039],
[0.603922, 0.627451, 0.898039],[0.592157, 0.619608, 0.898039],[0.580392, 0.611765, 0.898039],[0.568627, 0.6, 0.898039],
[0.560784, 0.592157, 0.898039],[0.54902, 0.584314, 0.898039],[0.537255, 0.576471, 0.898039],[0.52549, 0.568627, 0.898039],
[0.513725, 0.560784, 0.898039],[0.498039, 0.552941, 0.898039],[0.486275, 0.541176, 0.898039],[0.47451, 0.533333, 0.898039],
[0.462745, 0.52549, 0.898039],[0.447059, 0.517647, 0.898039],[0.435294, 0.509804, 0.898039],[0.419608, 0.501961, 0.898039],
[0.407843, 0.494118, 0.898039],[0.392157, 0.486275, 0.898039],[0.376471, 0.478431, 0.898039],[0.360784, 0.470588, 0.898039],
[0.345098, 0.462745, 0.898039],[0.329412, 0.454902, 0.898039],[0.309804, 0.447059, 0.898039],[0.294118, 0.439216, 0.898039],
[0.27451, 0.431373, 0.898039],[0.254902, 0.427451, 0.898039],[0.235294, 0.419608, 0.898039],[0.215686, 0.415686, 0.898039],
[0.196078, 0.407843, 0.898039],[0.176471, 0.403922, 0.898039],[0.156863, 0.4, 0.898039],[0.137255, 0.396078, 0.898039],
[0.121569, 0.392157, 0.898039],[0.109804, 0.392157, 0.898039],[0.101961, 0.388235, 0.898039],[0.0980392, 0.388235, 0.898039] ], 
[[0, 0.788235, 1],[0, 0.784314, 1],[0, 0.784314, 0.996078],
[0, 0.784314, 0.996078],[0, 0.784314, 0.992157],[0, 0.784314, 0.992157],[0, 0.784314, 0.988235],
[0.0509804, 0.784314, 0.988235],[0.0901961, 0.784314, 0.984314],[0.113725, 0.784314, 0.980392],[0.137255, 0.780392, 0.980392],
[0.156863, 0.780392, 0.976471],[0.172549, 0.780392, 0.976471],[0.188235, 0.780392, 0.972549],[0.2, 0.780392, 0.972549],
[0.215686, 0.780392, 0.968627],[0.227451, 0.780392, 0.968627],[0.239216, 0.780392, 0.964706],[0.25098, 0.776471, 0.960784],
[0.258824, 0.776471, 0.960784],[0.270588, 0.776471, 0.956863],[0.278431, 0.776471, 0.956863],[0.286275, 0.776471, 0.952941],
[0.298039, 0.776471, 0.952941],[0.305882, 0.776471, 0.94902],[0.313725, 0.776471, 0.945098],[0.321569, 0.772549, 0.945098],
[0.329412, 0.772549, 0.941176],[0.337255, 0.772549, 0.941176],[0.341176, 0.772549, 0.937255],[0.34902, 0.772549, 0.937255],
[0.356863, 0.772549, 0.933333],[0.364706, 0.772549, 0.933333],[0.368627, 0.772549, 0.929412],[0.376471, 0.772549, 0.92549],
[0.384314, 0.768627, 0.92549],[0.388235, 0.768627, 0.921569],[0.396078, 0.768627, 0.921569],[0.4, 0.768627, 0.917647],
[0.407843, 0.768627, 0.917647],[0.411765, 0.768627, 0.913725],[0.419608, 0.768627, 0.913725],[0.423529, 0.768627, 0.909804],
[0.427451, 0.764706, 0.905882],[0.435294, 0.764706, 0.905882],[0.439216, 0.764706, 0.901961],[0.443137, 0.764706, 0.901961],
[0.45098, 0.764706, 0.898039],[0.454902, 0.764706, 0.898039],[0.458824, 0.764706, 0.894118],[0.462745, 0.764706, 0.894118],
[0.470588, 0.760784, 0.890196],[0.47451, 0.760784, 0.886275],[0.478431, 0.760784, 0.886275],[0.482353, 0.760784, 0.882353],
[0.486275, 0.760784, 0.882353],[0.490196, 0.760784, 0.878431],[0.494118, 0.760784, 0.878431],[0.501961, 0.760784, 0.87451],
[0.505882, 0.756863, 0.87451],[0.509804, 0.756863, 0.870588],[0.513725, 0.756863, 0.866667],[0.517647, 0.756863, 0.866667],
[0.521569, 0.756863, 0.862745],[0.52549, 0.756863, 0.862745],[0.529412, 0.756863, 0.858824],[0.533333, 0.756863, 0.858824],
[0.537255, 0.752941, 0.854902],[0.541176, 0.752941, 0.854902],[0.545098, 0.752941, 0.85098],[0.54902, 0.752941, 0.847059],
[0.552941, 0.752941, 0.847059],[0.556863, 0.752941, 0.843137],[0.560784, 0.752941, 0.843137],[0.564706, 0.752941, 0.839216],
[0.568627, 0.74902, 0.839216],[0.572549, 0.74902, 0.835294],[0.572549, 0.74902, 0.835294],[0.576471, 0.74902, 0.831373],
[0.580392, 0.74902, 0.827451],[0.584314, 0.74902, 0.827451],[0.588235, 0.74902, 0.823529],[0.592157, 0.74902, 0.823529],
[0.596078, 0.745098, 0.819608],[0.6, 0.745098, 0.819608],[0.6, 0.745098, 0.815686],[0.603922, 0.745098, 0.815686],
[0.607843, 0.745098, 0.811765],[0.611765, 0.745098, 0.811765],[0.615686, 0.745098, 0.807843],[0.619608, 0.745098, 0.803922],
[0.619608, 0.741176, 0.803922],[0.623529, 0.741176, 0.8],[0.627451, 0.741176, 0.8],[0.631373, 0.741176, 0.796078],
[0.635294, 0.741176, 0.796078],[0.635294, 0.741176, 0.792157],[0.639216, 0.741176, 0.792157],[0.643137, 0.741176, 0.788235],
[0.647059, 0.737255, 0.784314],[0.647059, 0.737255, 0.784314],[0.65098, 0.737255, 0.780392],[0.654902, 0.737255, 0.780392],
[0.658824, 0.737255, 0.776471],[0.658824, 0.737255, 0.776471],[0.662745, 0.737255, 0.772549],[0.666667, 0.733333, 0.772549],
[0.670588, 0.733333, 0.768627],[0.670588, 0.733333, 0.764706],[0.67451, 0.733333, 0.764706],[0.678431, 0.733333, 0.760784],
[0.678431, 0.733333, 0.760784],[0.682353, 0.733333, 0.756863],[0.686275, 0.733333, 0.756863],[0.686275, 0.729412, 0.752941],
[0.690196, 0.729412, 0.752941],[0.694118, 0.729412, 0.74902],[0.698039, 0.729412, 0.74902],[0.698039, 0.729412, 0.745098],
[0.701961, 0.729412, 0.745098],[0.701961, 0.729412, 0.741176],[0.705882, 0.72549, 0.741176],[0.709804, 0.72549, 0.741176],
[0.709804, 0.72549, 0.737255],[0.713725, 0.72549, 0.737255],[0.717647, 0.72549, 0.737255],[0.717647, 0.72549, 0.733333],
[0.721569, 0.72549, 0.733333],[0.721569, 0.721569, 0.733333],[0.72549, 0.721569, 0.733333],[0.72549, 0.721569, 0.733333],
[0.729412, 0.721569, 0.733333],[0.729412, 0.721569, 0.733333],[0.733333, 0.721569, 0.733333],[0.733333, 0.717647, 0.733333],
[0.737255, 0.717647, 0.737255],[0.737255, 0.717647, 0.737255],[0.741176, 0.717647, 0.737255],[0.741176, 0.717647, 0.737255],
[0.745098, 0.717647, 0.741176],[0.745098, 0.713725, 0.741176],[0.74902, 0.713725, 0.741176],[0.74902, 0.713725, 0.745098],
[0.752941, 0.713725, 0.745098],[0.752941, 0.713725, 0.745098],[0.752941, 0.713725, 0.74902],[0.756863, 0.709804, 0.74902],
[0.756863, 0.709804, 0.74902],[0.760784, 0.709804, 0.752941],[0.760784, 0.709804, 0.752941],[0.764706, 0.709804, 0.752941],
[0.764706, 0.705882, 0.756863],[0.768627, 0.705882, 0.756863],[0.768627, 0.705882, 0.756863],[0.768627, 0.705882, 0.760784],
[0.772549, 0.705882, 0.760784],[0.772549, 0.705882, 0.760784],[0.776471, 0.701961, 0.764706],[0.776471, 0.701961, 0.764706],
[0.780392, 0.701961, 0.764706],[0.780392, 0.701961, 0.768627],[0.780392, 0.701961, 0.768627],[0.784314, 0.698039, 0.768627],
[0.784314, 0.698039, 0.772549],[0.788235, 0.698039, 0.772549],[0.788235, 0.698039, 0.772549],[0.792157, 0.698039, 0.776471],
[0.792157, 0.698039, 0.776471],[0.792157, 0.694118, 0.776471],[0.796078, 0.694118, 0.780392],[0.796078, 0.694118, 0.780392],
[0.8, 0.694118, 0.780392],[0.8, 0.694118, 0.784314],[0.803922, 0.690196, 0.784314],[0.803922, 0.690196, 0.784314],
[0.803922, 0.690196, 0.788235],[0.807843, 0.690196, 0.788235],[0.807843, 0.690196, 0.788235],[0.811765, 0.686275, 0.792157],
[0.811765, 0.686275, 0.792157],[0.811765, 0.686275, 0.792157],[0.815686, 0.686275, 0.796078],[0.815686, 0.686275, 0.796078],
[0.819608, 0.682353, 0.796078],[0.819608, 0.682353, 0.8],[0.823529, 0.682353, 0.8],[0.823529, 0.682353, 0.8],
[0.823529, 0.682353, 0.803922],[0.827451, 0.678431, 0.803922],[0.827451, 0.678431, 0.803922],[0.831373, 0.678431, 0.807843],
[0.831373, 0.678431, 0.807843],[0.831373, 0.678431, 0.807843],[0.835294, 0.67451, 0.811765],[0.835294, 0.67451, 0.811765],
[0.839216, 0.67451, 0.811765],[0.839216, 0.67451, 0.815686],[0.839216, 0.67451, 0.815686],[0.843137, 0.670588, 0.815686],
[0.843137, 0.670588, 0.819608],[0.847059, 0.670588, 0.819608],[0.847059, 0.670588, 0.819608],[0.85098, 0.670588, 0.823529],
[0.85098, 0.666667, 0.823529],[0.85098, 0.666667, 0.823529],[0.854902, 0.666667, 0.827451],[0.854902, 0.666667, 0.827451],
[0.858824, 0.662745, 0.831373],[0.858824, 0.662745, 0.831373],[0.858824, 0.662745, 0.831373],[0.862745, 0.662745, 0.835294],
[0.862745, 0.662745, 0.835294],[0.866667, 0.658824, 0.835294],[0.866667, 0.658824, 0.839216],[0.866667, 0.658824, 0.839216],
[0.870588, 0.658824, 0.839216],[0.870588, 0.658824, 0.843137],[0.87451, 0.654902, 0.843137],[0.87451, 0.654902, 0.843137],
[0.87451, 0.654902, 0.847059],[0.878431, 0.654902, 0.847059],[0.878431, 0.65098, 0.847059],[0.882353, 0.65098, 0.85098],
[0.882353, 0.65098, 0.85098],[0.882353, 0.65098, 0.85098],[0.886275, 0.65098, 0.854902],[0.886275, 0.647059, 0.854902],
[0.890196, 0.647059, 0.854902],[0.890196, 0.647059, 0.858824],[0.890196, 0.647059, 0.858824],[0.894118, 0.643137, 0.858824],
[0.894118, 0.643137, 0.862745],[0.894118, 0.643137, 0.862745],[0.898039, 0.643137, 0.862745],[0.898039, 0.643137, 0.866667],
[0.901961, 0.639216, 0.866667],[0.901961, 0.639216, 0.866667],[0.901961, 0.639216, 0.870588],[0.905882, 0.639216, 0.870588],
[0.905882, 0.635294, 0.870588],[0.909804, 0.635294, 0.87451],[0.909804, 0.635294, 0.87451],[0.909804, 0.635294, 0.87451],
[0.913725, 0.631373, 0.878431],[0.913725, 0.631373, 0.878431],[0.917647, 0.631373, 0.878431],[0.917647, 0.631373, 0.882353],
[0.917647, 0.627451, 0.882353],[0.921569, 0.627451, 0.882353],[0.921569, 0.627451, 0.886275],[0.921569, 0.627451, 0.886275],
[0.92549, 0.623529, 0.886275],[0.92549, 0.623529, 0.890196],[0.929412, 0.623529, 0.890196],[0.929412, 0.623529, 0.890196],
[0.929412, 0.619608, 0.894118] ], [[0, 0.796078, 0.996078],[0, 0.792157, 0.996078],
[0, 0.792157, 0.992157],[0, 0.792157, 0.992157],[0, 0.792157, 0.988235],[0, 0.792157, 0.984314],
[0, 0.792157, 0.984314],[0, 0.792157, 0.980392],[0, 0.792157, 0.980392],[0, 0.788235, 0.976471],
[0, 0.788235, 0.976471],[0, 0.788235, 0.972549],[0, 0.788235, 0.972549],[0, 0.788235, 0.968627],
[0, 0.788235, 0.964706],[0.0313725, 0.788235, 0.964706],[0.0784314, 0.784314, 0.960784],[0.109804, 0.784314, 0.960784],
[0.133333, 0.784314, 0.956863],[0.152941, 0.784314, 0.956863],[0.172549, 0.784314, 0.952941],[0.188235, 0.784314, 0.952941],
[0.2, 0.784314, 0.94902],[0.215686, 0.784314, 0.94902],[0.227451, 0.780392, 0.945098],[0.239216, 0.780392, 0.941176],
[0.25098, 0.780392, 0.941176],[0.262745, 0.780392, 0.937255],[0.270588, 0.780392, 0.937255],[0.282353, 0.780392, 0.933333],
[0.290196, 0.780392, 0.933333],[0.301961, 0.776471, 0.929412],[0.309804, 0.776471, 0.929412],[0.317647, 0.776471, 0.92549],
[0.32549, 0.776471, 0.921569],[0.333333, 0.776471, 0.921569],[0.341176, 0.776471, 0.917647],[0.34902, 0.776471, 0.917647],
[0.356863, 0.776471, 0.913725],[0.364706, 0.772549, 0.913725],[0.368627, 0.772549, 0.909804],[0.376471, 0.772549, 0.909804],
[0.384314, 0.772549, 0.905882],[0.388235, 0.772549, 0.905882],[0.396078, 0.772549, 0.901961],[0.403922, 0.772549, 0.898039],
[0.407843, 0.768627, 0.898039],[0.415686, 0.768627, 0.894118],[0.419608, 0.768627, 0.894118],[0.427451, 0.768627, 0.890196],
[0.431373, 0.768627, 0.890196],[0.439216, 0.768627, 0.886275],[0.443137, 0.768627, 0.886275],[0.447059, 0.764706, 0.882353],
[0.454902, 0.764706, 0.882353],[0.458824, 0.764706, 0.878431],[0.462745, 0.764706, 0.87451],[0.470588, 0.764706, 0.87451],
[0.47451, 0.764706, 0.870588],[0.478431, 0.764706, 0.870588],[0.482353, 0.760784, 0.866667],[0.490196, 0.760784, 0.866667],
[0.494118, 0.760784, 0.862745],[0.498039, 0.760784, 0.862745],[0.501961, 0.760784, 0.858824],[0.505882, 0.760784, 0.858824],
[0.513725, 0.760784, 0.854902],[0.517647, 0.760784, 0.85098],[0.521569, 0.756863, 0.85098],[0.52549, 0.756863, 0.847059],
[0.529412, 0.756863, 0.847059],[0.533333, 0.756863, 0.843137],[0.537255, 0.756863, 0.843137],[0.541176, 0.756863, 0.839216],
[0.545098, 0.756863, 0.839216],[0.54902, 0.752941, 0.835294],[0.552941, 0.752941, 0.835294],[0.556863, 0.752941, 0.831373],
[0.560784, 0.752941, 0.827451],[0.564706, 0.752941, 0.827451],[0.568627, 0.752941, 0.823529],[0.572549, 0.752941, 0.823529],
[0.576471, 0.74902, 0.819608],[0.580392, 0.74902, 0.819608],[0.584314, 0.74902, 0.815686],[0.588235, 0.74902, 0.815686],
[0.592157, 0.74902, 0.811765],[0.596078, 0.74902, 0.811765],[0.6, 0.74902, 0.807843],[0.603922, 0.745098, 0.807843],
[0.607843, 0.745098, 0.803922],[0.611765, 0.745098, 0.8],[0.615686, 0.745098, 0.8],[0.615686, 0.745098, 0.796078],
[0.619608, 0.745098, 0.796078],[0.623529, 0.745098, 0.792157],[0.627451, 0.741176, 0.792157],[0.631373, 0.741176, 0.788235],
[0.635294, 0.741176, 0.788235],[0.639216, 0.741176, 0.784314],[0.639216, 0.741176, 0.784314],[0.643137, 0.741176, 0.780392],
[0.647059, 0.737255, 0.776471],[0.65098, 0.737255, 0.776471],[0.654902, 0.737255, 0.772549],[0.658824, 0.737255, 0.772549],
[0.658824, 0.737255, 0.768627],[0.662745, 0.737255, 0.768627],[0.666667, 0.737255, 0.764706],[0.670588, 0.733333, 0.764706],
[0.67451, 0.733333, 0.760784],[0.67451, 0.733333, 0.760784],[0.678431, 0.733333, 0.756863],[0.682353, 0.733333, 0.756863],
[0.686275, 0.733333, 0.752941],[0.686275, 0.733333, 0.752941],[0.690196, 0.729412, 0.74902],[0.694118, 0.729412, 0.74902],
[0.698039, 0.729412, 0.745098],[0.698039, 0.729412, 0.745098],[0.701961, 0.729412, 0.741176],[0.705882, 0.729412, 0.741176],
[0.705882, 0.72549, 0.737255],[0.709804, 0.72549, 0.737255],[0.713725, 0.72549, 0.737255],[0.713725, 0.72549, 0.737255],
[0.717647, 0.72549, 0.733333],[0.721569, 0.72549, 0.733333],[0.721569, 0.721569, 0.733333],[0.72549, 0.721569, 0.733333],
[0.72549, 0.721569, 0.733333],[0.729412, 0.721569, 0.733333],[0.729412, 0.721569, 0.733333],[0.733333, 0.721569, 0.733333],
[0.737255, 0.717647, 0.737255],[0.737255, 0.717647, 0.737255],[0.741176, 0.717647, 0.737255],[0.741176, 0.717647, 0.737255],
[0.745098, 0.717647, 0.741176],[0.745098, 0.717647, 0.741176],[0.745098, 0.713725, 0.741176],[0.74902, 0.713725, 0.741176],
[0.74902, 0.713725, 0.745098],[0.752941, 0.713725, 0.745098],[0.752941, 0.713725, 0.74902],[0.756863, 0.709804, 0.74902],
[0.756863, 0.709804, 0.74902],[0.760784, 0.709804, 0.752941],[0.760784, 0.709804, 0.752941],[0.764706, 0.709804, 0.752941],
[0.764706, 0.705882, 0.756863],[0.768627, 0.705882, 0.756863],[0.768627, 0.705882, 0.756863],[0.768627, 0.705882, 0.760784],
[0.772549, 0.705882, 0.760784],[0.772549, 0.705882, 0.760784],[0.776471, 0.701961, 0.764706],[0.776471, 0.701961, 0.764706],
[0.780392, 0.701961, 0.764706],[0.780392, 0.701961, 0.768627],[0.784314, 0.701961, 0.768627],[0.784314, 0.698039, 0.768627],
[0.788235, 0.698039, 0.772549],[0.788235, 0.698039, 0.772549],[0.788235, 0.698039, 0.776471],[0.792157, 0.698039, 0.776471],
[0.792157, 0.694118, 0.776471],[0.796078, 0.694118, 0.780392],[0.796078, 0.694118, 0.780392],[0.8, 0.694118, 0.780392],
[0.8, 0.694118, 0.784314],[0.8, 0.690196, 0.784314],[0.803922, 0.690196, 0.784314],[0.803922, 0.690196, 0.788235],
[0.807843, 0.690196, 0.788235],[0.807843, 0.690196, 0.788235],[0.811765, 0.686275, 0.792157],[0.811765, 0.686275, 0.792157],
[0.815686, 0.686275, 0.792157],[0.815686, 0.686275, 0.796078],[0.815686, 0.686275, 0.796078],[0.819608, 0.682353, 0.8],
[0.819608, 0.682353, 0.8],[0.823529, 0.682353, 0.8],[0.823529, 0.682353, 0.803922],[0.827451, 0.682353, 0.803922],
[0.827451, 0.678431, 0.803922],[0.827451, 0.678431, 0.807843],[0.831373, 0.678431, 0.807843],[0.831373, 0.678431, 0.807843],
[0.835294, 0.67451, 0.811765],[0.835294, 0.67451, 0.811765],[0.839216, 0.67451, 0.811765],[0.839216, 0.67451, 0.815686],
[0.839216, 0.67451, 0.815686],[0.843137, 0.670588, 0.815686],[0.843137, 0.670588, 0.819608],[0.847059, 0.670588, 0.819608],
[0.847059, 0.670588, 0.823529],[0.85098, 0.670588, 0.823529],[0.85098, 0.666667, 0.823529],[0.85098, 0.666667, 0.827451],
[0.854902, 0.666667, 0.827451],[0.854902, 0.666667, 0.827451],[0.858824, 0.662745, 0.831373],[0.858824, 0.662745, 0.831373],
[0.862745, 0.662745, 0.831373],[0.862745, 0.662745, 0.835294],[0.862745, 0.662745, 0.835294],[0.866667, 0.658824, 0.835294],
[0.866667, 0.658824, 0.839216],[0.870588, 0.658824, 0.839216],[0.870588, 0.658824, 0.839216],[0.870588, 0.654902, 0.843137],
[0.87451, 0.654902, 0.843137],[0.87451, 0.654902, 0.843137],[0.878431, 0.654902, 0.847059],[0.878431, 0.65098, 0.847059],
[0.882353, 0.65098, 0.85098],[0.882353, 0.65098, 0.85098],[0.882353, 0.65098, 0.85098],[0.886275, 0.65098, 0.854902],
[0.886275, 0.647059, 0.854902],[0.890196, 0.647059, 0.854902],[0.890196, 0.647059, 0.858824],[0.890196, 0.647059, 0.858824],
[0.894118, 0.643137, 0.858824],[0.894118, 0.643137, 0.862745],[0.898039, 0.643137, 0.862745],[0.898039, 0.643137, 0.862745],
[0.898039, 0.639216, 0.866667],[0.901961, 0.639216, 0.866667],[0.901961, 0.639216, 0.866667],[0.905882, 0.639216, 0.870588],
[0.905882, 0.635294, 0.870588],[0.905882, 0.635294, 0.87451],[0.909804, 0.635294, 0.87451],[0.909804, 0.635294, 0.87451],
[0.913725, 0.631373, 0.878431],[0.913725, 0.631373, 0.878431],[0.917647, 0.631373, 0.878431],[0.917647, 0.631373, 0.882353],
[0.917647, 0.627451, 0.882353],[0.921569, 0.627451, 0.882353],[0.921569, 0.627451, 0.886275],[0.92549, 0.627451, 0.886275],
[0.92549, 0.623529, 0.886275],[0.92549, 0.623529, 0.890196],[0.929412, 0.623529, 0.890196],[0.929412, 0.623529, 0.890196],
[0.933333, 0.619608, 0.894118],[0.933333, 0.619608, 0.894118],[0.933333, 0.619608, 0.894118],[0.937255, 0.619608, 0.898039],
[0.937255, 0.615686, 0.898039],[0.941176, 0.615686, 0.901961] ], [[0, 0.713725, 1],
[0, 0.713725, 1],[0, 0.713725, 1],[0, 0.713725, 1],[0.0235294, 0.713725, 1],
[0.0666667, 0.713725, 1],[0.0980392, 0.713725, 0.996078],[0.121569, 0.713725, 0.992157],[0.141176, 0.713725, 0.992157],
[0.156863, 0.713725, 0.988235],[0.172549, 0.709804, 0.984314],[0.184314, 0.709804, 0.984314],[0.2, 0.709804, 0.980392],
[0.211765, 0.709804, 0.976471],[0.223529, 0.709804, 0.976471],[0.231373, 0.709804, 0.972549],[0.243137, 0.709804, 0.968627],
[0.25098, 0.709804, 0.964706],[0.262745, 0.709804, 0.964706],[0.270588, 0.709804, 0.960784],[0.278431, 0.709804, 0.956863],
[0.286275, 0.705882, 0.956863],[0.294118, 0.705882, 0.952941],[0.301961, 0.705882, 0.94902],[0.309804, 0.705882, 0.94902],
[0.317647, 0.705882, 0.945098],[0.321569, 0.705882, 0.941176],[0.329412, 0.705882, 0.941176],[0.337255, 0.705882, 0.937255],
[0.341176, 0.705882, 0.933333],[0.34902, 0.705882, 0.929412],[0.352941, 0.705882, 0.929412],[0.360784, 0.701961, 0.92549],
[0.364706, 0.701961, 0.921569],[0.372549, 0.701961, 0.921569],[0.376471, 0.701961, 0.917647],[0.384314, 0.701961, 0.913725],
[0.388235, 0.701961, 0.913725],[0.392157, 0.701961, 0.909804],[0.4, 0.701961, 0.905882],[0.403922, 0.701961, 0.905882],
[0.407843, 0.701961, 0.901961],[0.411765, 0.701961, 0.898039],[0.419608, 0.698039, 0.898039],[0.423529, 0.698039, 0.894118],
[0.427451, 0.698039, 0.890196],[0.431373, 0.698039, 0.890196],[0.435294, 0.698039, 0.886275],[0.439216, 0.698039, 0.882353],
[0.443137, 0.698039, 0.878431],[0.45098, 0.698039, 0.878431],[0.454902, 0.698039, 0.87451],[0.458824, 0.698039, 0.870588],
[0.462745, 0.698039, 0.870588],[0.466667, 0.694118, 0.866667],[0.470588, 0.694118, 0.862745],[0.47451, 0.694118, 0.862745],
[0.478431, 0.694118, 0.858824],[0.482353, 0.694118, 0.854902],[0.486275, 0.694118, 0.854902],[0.490196, 0.694118, 0.85098],
[0.494118, 0.694118, 0.847059],[0.494118, 0.694118, 0.847059],[0.498039, 0.694118, 0.843137],[0.501961, 0.694118, 0.839216],
[0.505882, 0.690196, 0.839216],[0.509804, 0.690196, 0.835294],[0.513725, 0.690196, 0.831373],[0.517647, 0.690196, 0.827451],
[0.521569, 0.690196, 0.827451],[0.521569, 0.690196, 0.823529],[0.52549, 0.690196, 0.819608],[0.529412, 0.690196, 0.819608],
[0.533333, 0.690196, 0.815686],[0.537255, 0.690196, 0.811765],[0.541176, 0.690196, 0.811765],[0.541176, 0.690196, 0.807843],
[0.545098, 0.686275, 0.803922],[0.54902, 0.686275, 0.803922],[0.552941, 0.686275, 0.8],[0.552941, 0.686275, 0.796078],
[0.556863, 0.686275, 0.796078],[0.560784, 0.686275, 0.792157],[0.564706, 0.686275, 0.788235],[0.564706, 0.686275, 0.788235],
[0.568627, 0.686275, 0.784314],[0.572549, 0.686275, 0.780392],[0.576471, 0.686275, 0.780392],[0.576471, 0.682353, 0.776471],
[0.580392, 0.682353, 0.772549],[0.584314, 0.682353, 0.768627],[0.584314, 0.682353, 0.768627],[0.588235, 0.682353, 0.764706],
[0.592157, 0.682353, 0.760784],[0.592157, 0.682353, 0.760784],[0.596078, 0.682353, 0.756863],[0.6, 0.682353, 0.752941],
[0.6, 0.682353, 0.752941],[0.603922, 0.682353, 0.74902],[0.607843, 0.682353, 0.745098],[0.607843, 0.678431, 0.745098],
[0.611765, 0.678431, 0.741176],[0.615686, 0.678431, 0.737255],[0.615686, 0.678431, 0.737255],[0.619608, 0.678431, 0.733333],
[0.619608, 0.678431, 0.729412],[0.623529, 0.678431, 0.729412],[0.627451, 0.678431, 0.72549],[0.627451, 0.678431, 0.721569],
[0.631373, 0.678431, 0.721569],[0.631373, 0.678431, 0.717647],[0.635294, 0.67451, 0.713725],[0.639216, 0.67451, 0.713725],
[0.639216, 0.67451, 0.709804],[0.643137, 0.67451, 0.705882],[0.643137, 0.67451, 0.705882],[0.647059, 0.67451, 0.701961],
[0.647059, 0.67451, 0.698039],[0.65098, 0.67451, 0.698039],[0.654902, 0.67451, 0.694118],[0.654902, 0.67451, 0.690196],
[0.658824, 0.67451, 0.690196],[0.658824, 0.670588, 0.686275],[0.662745, 0.670588, 0.682353],[0.666667, 0.670588, 0.682353],
[0.666667, 0.670588, 0.678431],[0.670588, 0.670588, 0.67451],[0.670588, 0.670588, 0.67451],[0.67451, 0.670588, 0.670588],
[0.678431, 0.670588, 0.670588],[0.678431, 0.666667, 0.666667],[0.682353, 0.666667, 0.666667],[0.686275, 0.666667, 0.662745],
[0.686275, 0.666667, 0.662745],[0.690196, 0.666667, 0.658824],[0.694118, 0.666667, 0.654902],[0.698039, 0.662745, 0.654902],
[0.698039, 0.662745, 0.65098],[0.701961, 0.662745, 0.65098],[0.705882, 0.662745, 0.647059],[0.705882, 0.662745, 0.647059],
[0.709804, 0.662745, 0.643137],[0.713725, 0.658824, 0.643137],[0.713725, 0.658824, 0.639216],[0.717647, 0.658824, 0.639216],
[0.721569, 0.658824, 0.635294],[0.72549, 0.658824, 0.635294],[0.72549, 0.654902, 0.635294],[0.729412, 0.654902, 0.631373],
[0.733333, 0.654902, 0.631373],[0.733333, 0.654902, 0.627451],[0.737255, 0.654902, 0.627451],[0.741176, 0.654902, 0.623529],
[0.741176, 0.65098, 0.623529],[0.745098, 0.65098, 0.619608],[0.74902, 0.65098, 0.619608],[0.74902, 0.65098, 0.615686],
[0.752941, 0.65098, 0.615686],[0.756863, 0.647059, 0.611765],[0.756863, 0.647059, 0.611765],[0.760784, 0.647059, 0.607843],
[0.764706, 0.647059, 0.607843],[0.764706, 0.647059, 0.603922],[0.768627, 0.643137, 0.603922],[0.772549, 0.643137, 0.6],
[0.772549, 0.643137, 0.6],[0.776471, 0.643137, 0.596078],[0.776471, 0.643137, 0.596078],[0.780392, 0.639216, 0.592157],
[0.784314, 0.639216, 0.592157],[0.784314, 0.639216, 0.588235],[0.788235, 0.639216, 0.588235],[0.792157, 0.639216, 0.584314],
[0.792157, 0.635294, 0.584314],[0.796078, 0.635294, 0.580392],[0.796078, 0.635294, 0.580392],[0.8, 0.635294, 0.576471],
[0.803922, 0.635294, 0.576471],[0.803922, 0.631373, 0.572549],[0.807843, 0.631373, 0.572549],[0.807843, 0.631373, 0.568627],
[0.811765, 0.631373, 0.568627],[0.815686, 0.631373, 0.564706],[0.815686, 0.627451, 0.564706],[0.819608, 0.627451, 0.564706],
[0.819608, 0.627451, 0.560784],[0.823529, 0.627451, 0.560784],[0.827451, 0.627451, 0.556863],[0.827451, 0.623529, 0.556863],
[0.831373, 0.623529, 0.552941],[0.831373, 0.623529, 0.552941],[0.835294, 0.623529, 0.54902],[0.835294, 0.623529, 0.54902],
[0.839216, 0.619608, 0.545098],[0.843137, 0.619608, 0.545098],[0.843137, 0.619608, 0.541176],[0.847059, 0.619608, 0.541176],
[0.847059, 0.619608, 0.537255],[0.85098, 0.615686, 0.537255],[0.85098, 0.615686, 0.533333],[0.854902, 0.615686, 0.533333],
[0.858824, 0.615686, 0.529412],[0.858824, 0.611765, 0.529412],[0.862745, 0.611765, 0.52549],[0.862745, 0.611765, 0.52549],
[0.866667, 0.611765, 0.521569],[0.866667, 0.611765, 0.521569],[0.870588, 0.607843, 0.517647],[0.870588, 0.607843, 0.517647],
[0.87451, 0.607843, 0.513725],[0.878431, 0.607843, 0.513725],[0.878431, 0.603922, 0.509804],[0.882353, 0.603922, 0.509804],
[0.882353, 0.603922, 0.505882],[0.886275, 0.603922, 0.505882],[0.886275, 0.603922, 0.501961],[0.890196, 0.6, 0.501961],
[0.890196, 0.6, 0.498039],[0.894118, 0.6, 0.498039],[0.894118, 0.6, 0.494118],[0.898039, 0.596078, 0.494118],
[0.898039, 0.596078, 0.490196],[0.901961, 0.596078, 0.490196],[0.901961, 0.596078, 0.490196],[0.905882, 0.596078, 0.486275],
[0.905882, 0.592157, 0.486275],[0.909804, 0.592157, 0.482353],[0.913725, 0.592157, 0.482353],[0.913725, 0.592157, 0.478431],
[0.917647, 0.588235, 0.478431],[0.917647, 0.588235, 0.47451],[0.921569, 0.588235, 0.47451],[0.921569, 0.588235, 0.470588],
[0.92549, 0.584314, 0.470588],[0.92549, 0.584314, 0.466667],[0.929412, 0.584314, 0.466667],[0.929412, 0.584314, 0.462745],
[0.933333, 0.584314, 0.462745],[0.933333, 0.580392, 0.458824],[0.937255, 0.580392, 0.458824],[0.937255, 0.580392, 0.454902],
[0.941176, 0.580392, 0.454902],[0.941176, 0.576471, 0.45098],[0.945098, 0.576471, 0.45098],[0.945098, 0.576471, 0.447059],
[0.94902, 0.576471, 0.447059],[0.94902, 0.572549, 0.443137],[0.952941, 0.572549, 0.443137],[0.952941, 0.572549, 0.439216],
[0.956863, 0.572549, 0.439216],[0.956863, 0.568627, 0.435294],[0.960784, 0.568627, 0.435294],[0.960784, 0.568627, 0.431373],
[0.960784, 0.568627, 0.431373],[0.964706, 0.564706, 0.427451],[0.964706, 0.564706, 0.427451] ], [
[0, 0.164706, 0.843137],[0, 0.168627, 0.839216],[0, 0.168627, 0.835294],[0, 0.172549, 0.831373],
[0, 0.172549, 0.831373],[0, 0.176471, 0.827451],[0, 0.176471, 0.823529],[0, 0.180392, 0.819608],
[0.027451, 0.180392, 0.815686],[0.0509804, 0.184314, 0.811765],[0.0745098, 0.184314, 0.807843],[0.0901961, 0.188235, 0.803922],
[0.105882, 0.188235, 0.8],[0.117647, 0.192157, 0.796078],[0.129412, 0.192157, 0.792157],[0.137255, 0.196078, 0.788235],
[0.14902, 0.196078, 0.788235],[0.156863, 0.2, 0.784314],[0.164706, 0.2, 0.780392],[0.172549, 0.203922, 0.776471],
[0.180392, 0.203922, 0.772549],[0.184314, 0.207843, 0.768627],[0.192157, 0.207843, 0.764706],[0.2, 0.211765, 0.760784],
[0.203922, 0.211765, 0.756863],[0.211765, 0.215686, 0.752941],[0.215686, 0.215686, 0.74902],[0.219608, 0.219608, 0.74902],
[0.227451, 0.219608, 0.745098],[0.231373, 0.223529, 0.741176],[0.235294, 0.223529, 0.737255],[0.239216, 0.223529, 0.733333],
[0.243137, 0.227451, 0.729412],[0.25098, 0.227451, 0.72549],[0.254902, 0.231373, 0.721569],[0.258824, 0.231373, 0.717647],
[0.262745, 0.235294, 0.713725],[0.266667, 0.235294, 0.709804],[0.270588, 0.239216, 0.709804],[0.270588, 0.239216, 0.705882],
[0.27451, 0.243137, 0.701961],[0.278431, 0.243137, 0.698039],[0.282353, 0.247059, 0.694118],[0.286275, 0.247059, 0.690196],
[0.290196, 0.25098, 0.686275],[0.290196, 0.25098, 0.682353],[0.294118, 0.25098, 0.678431],[0.298039, 0.254902, 0.67451],
[0.301961, 0.254902, 0.67451],[0.301961, 0.258824, 0.670588],[0.305882, 0.258824, 0.666667],[0.309804, 0.262745, 0.662745],
[0.309804, 0.262745, 0.658824],[0.313725, 0.266667, 0.654902],[0.317647, 0.266667, 0.65098],[0.317647, 0.270588, 0.647059],
[0.321569, 0.270588, 0.643137],[0.321569, 0.27451, 0.639216],[0.32549, 0.27451, 0.639216],[0.329412, 0.27451, 0.635294],
[0.329412, 0.278431, 0.631373],[0.333333, 0.278431, 0.627451],[0.333333, 0.282353, 0.623529],[0.337255, 0.282353, 0.619608],
[0.337255, 0.286275, 0.615686],[0.341176, 0.286275, 0.611765],[0.341176, 0.290196, 0.607843],[0.345098, 0.290196, 0.607843],
[0.345098, 0.290196, 0.603922],[0.345098, 0.294118, 0.6],[0.34902, 0.294118, 0.596078],[0.34902, 0.298039, 0.592157],
[0.352941, 0.298039, 0.588235],[0.352941, 0.301961, 0.584314],[0.356863, 0.301961, 0.580392],[0.356863, 0.305882, 0.576471],
[0.356863, 0.305882, 0.572549],[0.360784, 0.305882, 0.572549],[0.360784, 0.309804, 0.568627],[0.360784, 0.309804, 0.564706],
[0.364706, 0.313725, 0.560784],[0.364706, 0.313725, 0.556863],[0.364706, 0.317647, 0.552941],[0.368627, 0.317647, 0.54902],
[0.368627, 0.321569, 0.545098],[0.368627, 0.321569, 0.541176],[0.368627, 0.321569, 0.541176],[0.372549, 0.32549, 0.537255],
[0.372549, 0.32549, 0.533333],[0.372549, 0.329412, 0.529412],[0.376471, 0.329412, 0.52549],[0.376471, 0.333333, 0.521569],
[0.376471, 0.333333, 0.517647],[0.376471, 0.337255, 0.513725],[0.376471, 0.337255, 0.509804],[0.380392, 0.337255, 0.509804],
[0.380392, 0.341176, 0.505882],[0.380392, 0.341176, 0.501961],[0.380392, 0.345098, 0.498039],[0.384314, 0.345098, 0.494118],
[0.384314, 0.34902, 0.490196],[0.384314, 0.34902, 0.486275],[0.384314, 0.34902, 0.482353],[0.384314, 0.352941, 0.478431],
[0.384314, 0.352941, 0.478431],[0.388235, 0.356863, 0.47451],[0.388235, 0.356863, 0.470588],[0.388235, 0.360784, 0.466667],
[0.388235, 0.360784, 0.462745],[0.388235, 0.360784, 0.458824],[0.388235, 0.364706, 0.454902],[0.388235, 0.364706, 0.45098],
[0.388235, 0.368627, 0.447059],[0.388235, 0.368627, 0.443137],[0.392157, 0.372549, 0.443137],[0.392157, 0.372549, 0.439216],
[0.392157, 0.372549, 0.435294],[0.392157, 0.376471, 0.431373],[0.392157, 0.376471, 0.427451],[0.392157, 0.380392, 0.423529],
[0.392157, 0.380392, 0.419608],[0.392157, 0.384314, 0.415686],[0.392157, 0.384314, 0.411765],[0.392157, 0.384314, 0.407843],
[0.392157, 0.388235, 0.403922],[0.392157, 0.388235, 0.403922],[0.392157, 0.392157, 0.4],[0.392157, 0.392157, 0.396078],
[0.396078, 0.392157, 0.392157],[0.403922, 0.392157, 0.388235],[0.411765, 0.392157, 0.388235],[0.415686, 0.392157, 0.384314],
[0.423529, 0.392157, 0.380392],[0.431373, 0.392157, 0.380392],[0.435294, 0.388235, 0.376471],[0.443137, 0.388235, 0.376471],
[0.447059, 0.388235, 0.372549],[0.454902, 0.388235, 0.368627],[0.458824, 0.388235, 0.368627],[0.466667, 0.388235, 0.364706],
[0.470588, 0.388235, 0.360784],[0.478431, 0.384314, 0.360784],[0.482353, 0.384314, 0.356863],[0.490196, 0.384314, 0.356863],
[0.494118, 0.384314, 0.352941],[0.501961, 0.384314, 0.34902],[0.505882, 0.384314, 0.34902],[0.513725, 0.380392, 0.345098],
[0.517647, 0.380392, 0.341176],[0.52549, 0.380392, 0.341176],[0.529412, 0.380392, 0.337255],[0.533333, 0.380392, 0.337255],
[0.541176, 0.376471, 0.333333],[0.545098, 0.376471, 0.329412],[0.54902, 0.376471, 0.329412],[0.556863, 0.376471, 0.32549],
[0.560784, 0.376471, 0.321569],[0.564706, 0.372549, 0.321569],[0.572549, 0.372549, 0.317647],[0.576471, 0.372549, 0.317647],
[0.580392, 0.372549, 0.313725],[0.588235, 0.368627, 0.309804],[0.592157, 0.368627, 0.309804],[0.596078, 0.368627, 0.305882],
[0.603922, 0.368627, 0.301961],[0.607843, 0.364706, 0.301961],[0.611765, 0.364706, 0.298039],[0.619608, 0.364706, 0.294118],
[0.623529, 0.364706, 0.294118],[0.627451, 0.360784, 0.290196],[0.631373, 0.360784, 0.286275],[0.639216, 0.360784, 0.286275],
[0.643137, 0.356863, 0.282353],[0.647059, 0.356863, 0.282353],[0.65098, 0.356863, 0.278431],[0.658824, 0.356863, 0.27451],
[0.662745, 0.352941, 0.27451],[0.666667, 0.352941, 0.270588],[0.670588, 0.352941, 0.266667],[0.678431, 0.34902, 0.266667],
[0.682353, 0.34902, 0.262745],[0.686275, 0.34902, 0.258824],[0.690196, 0.345098, 0.258824],[0.694118, 0.345098, 0.254902],
[0.701961, 0.345098, 0.25098],[0.705882, 0.341176, 0.25098],[0.709804, 0.341176, 0.247059],[0.713725, 0.337255, 0.243137],
[0.717647, 0.337255, 0.243137],[0.72549, 0.337255, 0.239216],[0.729412, 0.333333, 0.235294],[0.733333, 0.333333, 0.235294],
[0.737255, 0.329412, 0.231373],[0.741176, 0.329412, 0.227451],[0.74902, 0.329412, 0.227451],[0.752941, 0.32549, 0.223529],
[0.756863, 0.32549, 0.219608],[0.760784, 0.321569, 0.219608],[0.764706, 0.321569, 0.215686],[0.772549, 0.317647, 0.211765],
[0.776471, 0.317647, 0.207843],[0.780392, 0.313725, 0.207843],[0.784314, 0.313725, 0.203922],[0.788235, 0.309804, 0.2],
[0.792157, 0.309804, 0.2],[0.8, 0.305882, 0.196078],[0.803922, 0.305882, 0.192157],[0.807843, 0.301961, 0.188235],
[0.811765, 0.301961, 0.188235],[0.815686, 0.298039, 0.184314],[0.819608, 0.298039, 0.180392],[0.823529, 0.294118, 0.180392],
[0.831373, 0.290196, 0.176471],[0.835294, 0.290196, 0.172549],[0.839216, 0.286275, 0.168627],[0.843137, 0.286275, 0.168627],
[0.847059, 0.282353, 0.164706],[0.85098, 0.278431, 0.160784],[0.854902, 0.278431, 0.156863],[0.862745, 0.27451, 0.152941],
[0.866667, 0.270588, 0.152941],[0.870588, 0.270588, 0.14902],[0.87451, 0.266667, 0.145098],[0.878431, 0.262745, 0.141176],
[0.882353, 0.258824, 0.137255],[0.886275, 0.254902, 0.137255],[0.894118, 0.254902, 0.133333],[0.898039, 0.25098, 0.129412],
[0.901961, 0.247059, 0.12549],[0.905882, 0.243137, 0.121569],[0.909804, 0.239216, 0.117647],[0.913725, 0.235294, 0.113725],
[0.917647, 0.231373, 0.109804],[0.921569, 0.227451, 0.105882],[0.929412, 0.227451, 0.105882],[0.933333, 0.219608, 0.101961],
[0.937255, 0.215686, 0.0980392],[0.941176, 0.211765, 0.0941176],[0.945098, 0.207843, 0.0862745],[0.94902, 0.203922, 0.0823529],
[0.952941, 0.2, 0.0784314],[0.956863, 0.196078, 0.0745098],[0.964706, 0.188235, 0.0705882],[0.968627, 0.184314, 0.0666667],
[0.972549, 0.180392, 0.0588235],[0.976471, 0.172549, 0.054902],[0.980392, 0.168627, 0.0509804],[0.984314, 0.160784, 0.0431373],
[0.988235, 0.156863, 0.0352941],[0.992157, 0.14902, 0.0313725],[1, 0.141176, 0.0235294],[1, 0.133333, 0.0196078],
[1, 0.12549, 0.0117647],[1, 0.117647, 0.00392157],[1, 0.105882, 0],[1, 0.0980392, 0] ], 
[[0.0784314, 0.192157, 0.756863],[0.0941176, 0.196078, 0.756863],[0.109804, 0.2, 0.756863],
[0.121569, 0.2, 0.752941],[0.133333, 0.203922, 0.752941],[0.145098, 0.207843, 0.752941],[0.156863, 0.211765, 0.74902],
[0.164706, 0.211765, 0.74902],[0.176471, 0.215686, 0.74902],[0.184314, 0.219608, 0.745098],[0.192157, 0.223529, 0.745098],
[0.2, 0.223529, 0.745098],[0.207843, 0.227451, 0.741176],[0.211765, 0.231373, 0.741176],[0.219608, 0.231373, 0.737255],
[0.227451, 0.235294, 0.737255],[0.231373, 0.239216, 0.737255],[0.239216, 0.243137, 0.733333],[0.247059, 0.243137, 0.733333],
[0.25098, 0.247059, 0.733333],[0.258824, 0.25098, 0.729412],[0.262745, 0.254902, 0.729412],[0.266667, 0.254902, 0.729412],
[0.27451, 0.258824, 0.72549],[0.278431, 0.262745, 0.72549],[0.282353, 0.266667, 0.72549],[0.290196, 0.266667, 0.721569],
[0.294118, 0.270588, 0.721569],[0.298039, 0.27451, 0.717647],[0.301961, 0.27451, 0.717647],[0.305882, 0.278431, 0.717647],
[0.309804, 0.282353, 0.713725],[0.317647, 0.286275, 0.713725],[0.321569, 0.286275, 0.713725],[0.32549, 0.290196, 0.709804],
[0.329412, 0.294118, 0.709804],[0.333333, 0.298039, 0.709804],[0.337255, 0.298039, 0.705882],[0.341176, 0.301961, 0.705882],
[0.345098, 0.305882, 0.705882],[0.34902, 0.309804, 0.701961],[0.352941, 0.309804, 0.701961],[0.356863, 0.313725, 0.698039],
[0.360784, 0.317647, 0.698039],[0.364706, 0.317647, 0.698039],[0.368627, 0.321569, 0.694118],[0.368627, 0.32549, 0.694118],
[0.372549, 0.329412, 0.694118],[0.376471, 0.329412, 0.690196],[0.380392, 0.333333, 0.690196],[0.384314, 0.337255, 0.686275],
[0.388235, 0.341176, 0.686275],[0.392157, 0.341176, 0.686275],[0.392157, 0.345098, 0.682353],[0.396078, 0.34902, 0.682353],
[0.4, 0.352941, 0.682353],[0.403922, 0.352941, 0.678431],[0.407843, 0.356863, 0.678431],[0.407843, 0.360784, 0.678431],
[0.411765, 0.364706, 0.67451],[0.415686, 0.364706, 0.67451],[0.419608, 0.368627, 0.670588],[0.419608, 0.372549, 0.670588],
[0.423529, 0.376471, 0.670588],[0.427451, 0.376471, 0.666667],[0.431373, 0.380392, 0.666667],[0.431373, 0.384314, 0.662745],
[0.435294, 0.388235, 0.662745],[0.439216, 0.388235, 0.662745],[0.439216, 0.392157, 0.658824],[0.443137, 0.396078, 0.658824],
[0.447059, 0.4, 0.658824],[0.45098, 0.4, 0.654902],[0.45098, 0.403922, 0.654902],[0.454902, 0.407843, 0.65098],
[0.458824, 0.411765, 0.65098],[0.458824, 0.411765, 0.65098],[0.462745, 0.415686, 0.647059],[0.462745, 0.419608, 0.647059],
[0.466667, 0.423529, 0.647059],[0.470588, 0.423529, 0.643137],[0.470588, 0.427451, 0.643137],[0.47451, 0.431373, 0.639216],
[0.478431, 0.435294, 0.639216],[0.478431, 0.435294, 0.639216],[0.482353, 0.439216, 0.635294],[0.482353, 0.443137, 0.635294],
[0.486275, 0.447059, 0.631373],[0.490196, 0.447059, 0.631373],[0.490196, 0.45098, 0.631373],[0.494118, 0.454902, 0.627451],
[0.494118, 0.458824, 0.627451],[0.498039, 0.458824, 0.623529],[0.498039, 0.462745, 0.623529],[0.501961, 0.466667, 0.623529],
[0.505882, 0.470588, 0.619608],[0.505882, 0.470588, 0.619608],[0.509804, 0.47451, 0.615686],[0.509804, 0.478431, 0.615686],
[0.513725, 0.482353, 0.615686],[0.513725, 0.482353, 0.611765],[0.517647, 0.486275, 0.611765],[0.517647, 0.490196, 0.607843],
[0.521569, 0.494118, 0.607843],[0.521569, 0.498039, 0.607843],[0.52549, 0.498039, 0.603922],[0.52549, 0.501961, 0.603922],
[0.529412, 0.505882, 0.6],[0.529412, 0.509804, 0.6],[0.533333, 0.509804, 0.596078],[0.533333, 0.513725, 0.596078],
[0.537255, 0.517647, 0.596078],[0.537255, 0.521569, 0.592157],[0.541176, 0.521569, 0.592157],[0.541176, 0.52549, 0.588235],
[0.545098, 0.529412, 0.588235],[0.545098, 0.533333, 0.588235],[0.54902, 0.533333, 0.584314],[0.54902, 0.537255, 0.584314],
[0.552941, 0.541176, 0.580392],[0.552941, 0.545098, 0.580392],[0.556863, 0.54902, 0.576471],[0.556863, 0.54902, 0.576471],
[0.560784, 0.552941, 0.576471],[0.560784, 0.556863, 0.572549],[0.560784, 0.560784, 0.572549],[0.564706, 0.560784, 0.568627],
[0.564706, 0.564706, 0.568627],[0.568627, 0.568627, 0.564706],[0.572549, 0.568627, 0.564706],[0.576471, 0.572549, 0.560784],
[0.580392, 0.576471, 0.556863],[0.584314, 0.576471, 0.556863],[0.588235, 0.580392, 0.552941],[0.592157, 0.584314, 0.552941],
[0.596078, 0.584314, 0.54902],[0.6, 0.588235, 0.545098],[0.603922, 0.588235, 0.545098],[0.607843, 0.592157, 0.541176],
[0.611765, 0.596078, 0.541176],[0.615686, 0.596078, 0.537255],[0.619608, 0.6, 0.533333],[0.623529, 0.6, 0.533333],
[0.627451, 0.603922, 0.529412],[0.631373, 0.607843, 0.52549],[0.635294, 0.607843, 0.52549],[0.639216, 0.611765, 0.521569],
[0.643137, 0.615686, 0.521569],[0.647059, 0.615686, 0.517647],[0.65098, 0.619608, 0.513725],[0.654902, 0.619608, 0.513725],
[0.658824, 0.623529, 0.509804],[0.662745, 0.627451, 0.505882],[0.666667, 0.627451, 0.505882],[0.666667, 0.631373, 0.501961],
[0.670588, 0.635294, 0.498039],[0.67451, 0.635294, 0.498039],[0.678431, 0.639216, 0.494118],[0.682353, 0.639216, 0.490196],
[0.686275, 0.643137, 0.490196],[0.690196, 0.647059, 0.486275],[0.694118, 0.647059, 0.482353],[0.698039, 0.65098, 0.482353],
[0.701961, 0.654902, 0.478431],[0.701961, 0.654902, 0.47451],[0.705882, 0.658824, 0.47451],[0.709804, 0.662745, 0.470588],
[0.713725, 0.662745, 0.466667],[0.717647, 0.666667, 0.466667],[0.721569, 0.666667, 0.462745],[0.72549, 0.670588, 0.458824],
[0.729412, 0.67451, 0.458824],[0.729412, 0.67451, 0.454902],[0.733333, 0.678431, 0.45098],[0.737255, 0.682353, 0.45098],
[0.741176, 0.682353, 0.447059],[0.745098, 0.686275, 0.443137],[0.74902, 0.690196, 0.439216],[0.752941, 0.690196, 0.439216],
[0.752941, 0.694118, 0.435294],[0.756863, 0.698039, 0.431373],[0.760784, 0.698039, 0.427451],[0.764706, 0.701961, 0.427451],
[0.768627, 0.705882, 0.423529],[0.772549, 0.705882, 0.419608],[0.776471, 0.709804, 0.415686],[0.776471, 0.709804, 0.415686],
[0.780392, 0.713725, 0.411765],[0.784314, 0.717647, 0.407843],[0.788235, 0.717647, 0.403922],[0.792157, 0.721569, 0.403922],
[0.796078, 0.72549, 0.4],[0.796078, 0.72549, 0.396078],[0.8, 0.729412, 0.392157],[0.803922, 0.733333, 0.388235],
[0.807843, 0.733333, 0.388235],[0.811765, 0.737255, 0.384314],[0.811765, 0.741176, 0.380392],[0.815686, 0.741176, 0.376471],
[0.819608, 0.745098, 0.372549],[0.823529, 0.74902, 0.368627],[0.827451, 0.74902, 0.368627],[0.831373, 0.752941, 0.364706],
[0.831373, 0.756863, 0.360784],[0.835294, 0.756863, 0.356863],[0.839216, 0.760784, 0.352941],[0.843137, 0.764706, 0.34902],
[0.847059, 0.764706, 0.345098],[0.847059, 0.768627, 0.341176],[0.85098, 0.772549, 0.337255],[0.854902, 0.772549, 0.337255],
[0.858824, 0.776471, 0.333333],[0.862745, 0.780392, 0.329412],[0.862745, 0.780392, 0.32549],[0.866667, 0.784314, 0.321569],
[0.870588, 0.788235, 0.317647],[0.87451, 0.788235, 0.313725],[0.87451, 0.792157, 0.309804],[0.878431, 0.796078, 0.305882],
[0.882353, 0.796078, 0.301961],[0.886275, 0.8, 0.298039],[0.890196, 0.803922, 0.294118],[0.890196, 0.803922, 0.286275],
[0.894118, 0.807843, 0.282353],[0.898039, 0.811765, 0.278431],[0.901961, 0.811765, 0.27451],[0.905882, 0.815686, 0.270588],
[0.905882, 0.819608, 0.266667],[0.909804, 0.819608, 0.262745],[0.913725, 0.823529, 0.254902],[0.917647, 0.827451, 0.25098],
[0.917647, 0.827451, 0.247059],[0.921569, 0.831373, 0.243137],[0.92549, 0.835294, 0.235294],[0.929412, 0.839216, 0.231373],
[0.929412, 0.839216, 0.223529],[0.933333, 0.843137, 0.219608],[0.937255, 0.847059, 0.215686],[0.941176, 0.847059, 0.207843],
[0.945098, 0.85098, 0.2],[0.945098, 0.854902, 0.196078],[0.94902, 0.854902, 0.188235],[0.952941, 0.858824, 0.180392],
[0.956863, 0.862745, 0.176471],[0.956863, 0.862745, 0.168627],[0.960784, 0.866667, 0.160784],[0.964706, 0.870588, 0.152941],
[0.968627, 0.870588, 0.145098],[0.968627, 0.87451, 0.133333],[0.972549, 0.878431, 0.12549],[0.976471, 0.878431, 0.113725],
[0.980392, 0.882353, 0.101961],[0.980392, 0.886275, 0.0901961],[0.984314, 0.890196, 0.0745098],[0.988235, 0.890196, 0.0588235],
[0.992157, 0.894118, 0.0352941] ], [[0.0313725, 0.360784, 0.972549],[0.0588235, 0.372549, 0.956863],
[0.0745098, 0.380392, 0.945098],[0.0901961, 0.388235, 0.929412],[0.0980392, 0.396078, 0.913725],[0.105882, 0.403922, 0.898039],
[0.109804, 0.415686, 0.882353],[0.113725, 0.423529, 0.870588],[0.113725, 0.431373, 0.854902],[0.117647, 0.439216, 0.839216],
[0.113725, 0.447059, 0.823529],[0.113725, 0.454902, 0.807843],[0.109804, 0.458824, 0.796078],[0.105882, 0.466667, 0.780392],
[0.0980392, 0.47451, 0.764706],[0.0941176, 0.482353, 0.74902],[0.0862745, 0.490196, 0.733333],[0.0784314, 0.494118, 0.721569],
[0.0745098, 0.501961, 0.705882],[0.0705882, 0.509804, 0.690196],[0.0705882, 0.513725, 0.67451],[0.0745098, 0.521569, 0.658824],
[0.0784314, 0.52549, 0.643137],[0.0901961, 0.533333, 0.627451],[0.101961, 0.537255, 0.611765],[0.113725, 0.541176, 0.596078],
[0.12549, 0.54902, 0.576471],[0.137255, 0.552941, 0.560784],[0.152941, 0.556863, 0.545098],[0.164706, 0.560784, 0.52549],
[0.176471, 0.564706, 0.509804],[0.184314, 0.572549, 0.494118],[0.192157, 0.576471, 0.47451],[0.203922, 0.580392, 0.458824],
[0.207843, 0.584314, 0.439216],[0.215686, 0.588235, 0.423529],[0.219608, 0.592157, 0.403922],[0.223529, 0.596078, 0.384314],
[0.227451, 0.603922, 0.368627],[0.231373, 0.607843, 0.34902],[0.231373, 0.611765, 0.329412],[0.231373, 0.615686, 0.313725],
[0.235294, 0.619608, 0.294118],[0.235294, 0.623529, 0.27451],[0.235294, 0.627451, 0.258824],[0.239216, 0.631373, 0.239216],
[0.239216, 0.635294, 0.223529],[0.243137, 0.639216, 0.207843],[0.247059, 0.643137, 0.192157],[0.254902, 0.647059, 0.180392],
[0.258824, 0.65098, 0.168627],[0.266667, 0.654902, 0.156863],[0.278431, 0.658824, 0.14902],[0.286275, 0.658824, 0.141176],
[0.298039, 0.662745, 0.133333],[0.305882, 0.666667, 0.129412],[0.317647, 0.666667, 0.12549],[0.329412, 0.670588, 0.12549],
[0.341176, 0.67451, 0.121569],[0.352941, 0.67451, 0.121569],[0.360784, 0.678431, 0.117647],[0.372549, 0.682353, 0.117647],
[0.384314, 0.682353, 0.117647],[0.396078, 0.686275, 0.117647],[0.403922, 0.686275, 0.113725],[0.415686, 0.690196, 0.113725],
[0.423529, 0.694118, 0.113725],[0.435294, 0.694118, 0.113725],[0.447059, 0.698039, 0.113725],[0.454902, 0.698039, 0.113725],
[0.466667, 0.701961, 0.109804],[0.47451, 0.705882, 0.109804],[0.482353, 0.705882, 0.109804],[0.494118, 0.709804, 0.109804],
[0.501961, 0.709804, 0.105882],[0.513725, 0.713725, 0.105882],[0.521569, 0.713725, 0.105882],[0.529412, 0.717647, 0.105882],
[0.541176, 0.721569, 0.105882],[0.54902, 0.721569, 0.101961],[0.556863, 0.72549, 0.101961],[0.568627, 0.72549, 0.101961],
[0.576471, 0.729412, 0.101961],[0.584314, 0.729412, 0.0980392],[0.592157, 0.733333, 0.0980392],[0.603922, 0.733333, 0.0980392],
[0.611765, 0.737255, 0.0980392],[0.619608, 0.741176, 0.0941176],[0.627451, 0.741176, 0.0941176],[0.639216, 0.745098, 0.0941176],
[0.647059, 0.745098, 0.0941176],[0.654902, 0.74902, 0.0901961],[0.662745, 0.74902, 0.0901961],[0.670588, 0.752941, 0.0901961],
[0.682353, 0.752941, 0.0862745],[0.690196, 0.756863, 0.0862745],[0.698039, 0.756863, 0.0862745],[0.705882, 0.760784, 0.0823529],
[0.713725, 0.760784, 0.0823529],[0.721569, 0.764706, 0.0823529],[0.733333, 0.764706, 0.0784314],[0.741176, 0.768627, 0.0784314],
[0.74902, 0.772549, 0.0784314],[0.756863, 0.772549, 0.0745098],[0.764706, 0.776471, 0.0745098],[0.772549, 0.776471, 0.0745098],
[0.780392, 0.780392, 0.0705882],[0.792157, 0.780392, 0.0705882],[0.8, 0.784314, 0.0705882],[0.807843, 0.784314, 0.0666667],
[0.815686, 0.788235, 0.0666667],[0.823529, 0.788235, 0.0627451],[0.831373, 0.792157, 0.0627451],[0.839216, 0.792157, 0.0588235],
[0.847059, 0.796078, 0.0588235],[0.858824, 0.796078, 0.054902],[0.866667, 0.796078, 0.054902],[0.87451, 0.8, 0.054902],
[0.882353, 0.8, 0.054902],[0.890196, 0.803922, 0.054902],[0.898039, 0.803922, 0.054902],[0.905882, 0.803922, 0.0588235],
[0.913725, 0.807843, 0.0627451],[0.921569, 0.807843, 0.0705882],[0.929412, 0.807843, 0.0784314],[0.937255, 0.803922, 0.0862745],
[0.945098, 0.803922, 0.0980392],[0.952941, 0.803922, 0.109804],[0.956863, 0.8, 0.121569],[0.964706, 0.796078, 0.137255],
[0.968627, 0.792157, 0.14902],[0.972549, 0.788235, 0.160784],[0.976471, 0.784314, 0.176471],[0.976471, 0.780392, 0.188235],
[0.980392, 0.776471, 0.2],[0.980392, 0.772549, 0.211765],[0.984314, 0.764706, 0.223529],[0.984314, 0.760784, 0.235294],
[0.988235, 0.752941, 0.247059],[0.988235, 0.74902, 0.254902],[0.988235, 0.745098, 0.266667],[0.988235, 0.737255, 0.27451],
[0.988235, 0.733333, 0.282353],[0.992157, 0.72549, 0.294118],[0.992157, 0.721569, 0.301961],[0.992157, 0.713725, 0.309804],
[0.992157, 0.709804, 0.317647],[0.992157, 0.701961, 0.32549],[0.992157, 0.698039, 0.333333],[0.996078, 0.690196, 0.341176],
[0.996078, 0.686275, 0.34902],[0.996078, 0.678431, 0.356863],[0.996078, 0.67451, 0.364706],[0.996078, 0.666667, 0.372549],
[0.996078, 0.662745, 0.380392],[0.996078, 0.654902, 0.388235],[0.996078, 0.65098, 0.396078],[0.996078, 0.643137, 0.4],
[0.996078, 0.639216, 0.407843],[0.996078, 0.631373, 0.415686],[0.996078, 0.627451, 0.423529],[1, 0.619608, 0.427451],
[1, 0.615686, 0.435294],[1, 0.607843, 0.443137],[1, 0.603922, 0.447059],[1, 0.596078, 0.454902],
[1, 0.592157, 0.458824],[1, 0.584314, 0.466667],[1, 0.576471, 0.47451],[1, 0.572549, 0.478431],
[1, 0.564706, 0.486275],[1, 0.560784, 0.490196],[1, 0.552941, 0.498039],[0.996078, 0.545098, 0.501961],
[0.996078, 0.541176, 0.509804],[0.996078, 0.533333, 0.517647],[0.996078, 0.52549, 0.521569],[0.996078, 0.521569, 0.529412],
[0.996078, 0.513725, 0.533333],[0.996078, 0.505882, 0.541176],[0.996078, 0.501961, 0.545098],[0.996078, 0.494118, 0.54902],
[0.996078, 0.486275, 0.556863],[0.996078, 0.478431, 0.560784],[0.992157, 0.47451, 0.568627],[0.992157, 0.466667, 0.572549],
[0.992157, 0.458824, 0.580392],[0.992157, 0.45098, 0.584314],[0.992157, 0.443137, 0.592157],[0.992157, 0.439216, 0.596078],
[0.992157, 0.431373, 0.6],[0.988235, 0.423529, 0.607843],[0.988235, 0.415686, 0.611765],[0.988235, 0.407843, 0.619608],
[0.988235, 0.4, 0.623529],[0.988235, 0.392157, 0.627451],[0.984314, 0.384314, 0.635294],[0.984314, 0.376471, 0.639216],
[0.984314, 0.368627, 0.643137],[0.984314, 0.360784, 0.647059],[0.980392, 0.352941, 0.65098],[0.980392, 0.345098, 0.654902],
[0.980392, 0.337255, 0.658824],[0.980392, 0.329412, 0.658824],[0.976471, 0.321569, 0.658824],[0.976471, 0.313725, 0.658824],
[0.976471, 0.305882, 0.654902],[0.972549, 0.298039, 0.65098],[0.972549, 0.290196, 0.647059],[0.972549, 0.286275, 0.639216],
[0.968627, 0.278431, 0.631373],[0.968627, 0.27451, 0.619608],[0.968627, 0.266667, 0.611765],[0.964706, 0.262745, 0.6],
[0.964706, 0.258824, 0.588235],[0.964706, 0.25098, 0.572549],[0.960784, 0.247059, 0.560784],[0.960784, 0.243137, 0.54902],
[0.956863, 0.239216, 0.533333],[0.956863, 0.235294, 0.521569],[0.952941, 0.227451, 0.509804],[0.952941, 0.223529, 0.494118],
[0.94902, 0.219608, 0.482353],[0.945098, 0.215686, 0.470588],[0.945098, 0.211765, 0.454902],[0.941176, 0.203922, 0.443137],
[0.941176, 0.2, 0.427451],[0.937255, 0.196078, 0.415686],[0.933333, 0.192157, 0.403922],[0.933333, 0.184314, 0.388235],
[0.929412, 0.180392, 0.376471],[0.92549, 0.176471, 0.364706],[0.921569, 0.172549, 0.34902],[0.921569, 0.164706, 0.337255],
[0.917647, 0.160784, 0.32549],[0.913725, 0.156863, 0.309804],[0.909804, 0.14902, 0.298039],[0.909804, 0.145098, 0.286275],
[0.905882, 0.141176, 0.270588],[0.901961, 0.133333, 0.258824],[0.898039, 0.129412, 0.247059],[0.894118, 0.12549, 0.231373],
[0.890196, 0.117647, 0.219608],[0.886275, 0.113725, 0.207843],[0.886275, 0.105882, 0.192157],[0.882353, 0.101961, 0.180392],
[0.878431, 0.0941176, 0.164706],[0.87451, 0.0862745, 0.14902],[0.870588, 0.0784314, 0.137255],[0.866667, 0.0745098, 0.121569],
[0.862745, 0.0666667, 0.105882],[0.858824, 0.0588235, 0.0901961],[0.854902, 0.0470588, 0.0705882],[0.85098, 0.0392157, 0.0470588],
[0.847059, 0.027451, 0.0235294],[0.843137, 0.0196078, 0] ], [[0.0980392, 0.505882, 0.980392],
[0.101961, 0.501961, 0.972549],[0.109804, 0.501961, 0.964706],[0.113725, 0.498039, 0.956863],[0.117647, 0.494118, 0.952941],
[0.121569, 0.490196, 0.945098],[0.12549, 0.486275, 0.937255],[0.129412, 0.482353, 0.929412],[0.133333, 0.478431, 0.921569],
[0.137255, 0.47451, 0.913725],[0.141176, 0.47451, 0.905882],[0.145098, 0.470588, 0.898039],[0.14902, 0.466667, 0.890196],
[0.14902, 0.462745, 0.882353],[0.152941, 0.458824, 0.87451],[0.156863, 0.454902, 0.866667],[0.156863, 0.45098, 0.858824],
[0.160784, 0.447059, 0.85098],[0.164706, 0.447059, 0.843137],[0.164706, 0.443137, 0.835294],[0.168627, 0.439216, 0.827451],
[0.168627, 0.435294, 0.819608],[0.172549, 0.431373, 0.815686],[0.172549, 0.427451, 0.807843],[0.172549, 0.423529, 0.8],
[0.176471, 0.423529, 0.792157],[0.176471, 0.419608, 0.784314],[0.180392, 0.415686, 0.776471],[0.180392, 0.411765, 0.768627],
[0.180392, 0.407843, 0.760784],[0.184314, 0.403922, 0.752941],[0.184314, 0.4, 0.74902],[0.184314, 0.4, 0.741176],
[0.184314, 0.396078, 0.733333],[0.188235, 0.392157, 0.72549],[0.188235, 0.388235, 0.717647],[0.188235, 0.384314, 0.709804],
[0.188235, 0.380392, 0.701961],[0.188235, 0.380392, 0.694118],[0.192157, 0.376471, 0.690196],[0.192157, 0.372549, 0.682353],
[0.192157, 0.368627, 0.67451],[0.192157, 0.364706, 0.666667],[0.192157, 0.360784, 0.658824],[0.192157, 0.360784, 0.65098],
[0.192157, 0.356863, 0.643137],[0.192157, 0.352941, 0.639216],[0.196078, 0.34902, 0.631373],[0.196078, 0.345098, 0.623529],
[0.196078, 0.341176, 0.615686],[0.196078, 0.341176, 0.607843],[0.196078, 0.337255, 0.603922],[0.196078, 0.333333, 0.596078],
[0.196078, 0.329412, 0.588235],[0.196078, 0.32549, 0.580392],[0.196078, 0.321569, 0.572549],[0.196078, 0.321569, 0.568627],
[0.196078, 0.317647, 0.560784],[0.196078, 0.313725, 0.552941],[0.196078, 0.309804, 0.545098],[0.196078, 0.305882, 0.537255],
[0.192157, 0.305882, 0.533333],[0.192157, 0.301961, 0.52549],[0.192157, 0.298039, 0.517647],[0.192157, 0.294118, 0.509804],
[0.192157, 0.290196, 0.505882],[0.192157, 0.290196, 0.498039],[0.192157, 0.286275, 0.490196],[0.192157, 0.282353, 0.482353],
[0.192157, 0.278431, 0.478431],[0.188235, 0.27451, 0.470588],[0.188235, 0.27451, 0.462745],[0.188235, 0.270588, 0.454902],
[0.188235, 0.266667, 0.45098],[0.188235, 0.262745, 0.443137],[0.188235, 0.258824, 0.435294],[0.184314, 0.258824, 0.431373],
[0.184314, 0.254902, 0.423529],[0.184314, 0.25098, 0.415686],[0.184314, 0.247059, 0.407843],[0.184314, 0.247059, 0.403922],
[0.180392, 0.243137, 0.396078],[0.180392, 0.239216, 0.388235],[0.180392, 0.235294, 0.384314],[0.180392, 0.235294, 0.376471],
[0.176471, 0.231373, 0.368627],[0.176471, 0.227451, 0.364706],[0.176471, 0.223529, 0.356863],[0.172549, 0.219608, 0.34902],
[0.172549, 0.219608, 0.345098],[0.172549, 0.215686, 0.337255],[0.172549, 0.211765, 0.333333],[0.168627, 0.207843, 0.32549],
[0.168627, 0.207843, 0.317647],[0.168627, 0.203922, 0.313725],[0.164706, 0.2, 0.305882],[0.164706, 0.196078, 0.298039],
[0.164706, 0.196078, 0.294118],[0.160784, 0.192157, 0.286275],[0.160784, 0.188235, 0.282353],[0.160784, 0.184314, 0.27451],
[0.156863, 0.184314, 0.266667],[0.156863, 0.180392, 0.262745],[0.152941, 0.176471, 0.254902],[0.152941, 0.176471, 0.25098],
[0.152941, 0.172549, 0.243137],[0.14902, 0.168627, 0.239216],[0.14902, 0.164706, 0.231373],[0.145098, 0.164706, 0.227451],
[0.145098, 0.160784, 0.219608],[0.145098, 0.156863, 0.211765],[0.141176, 0.156863, 0.207843],[0.141176, 0.152941, 0.2],
[0.137255, 0.14902, 0.196078],[0.137255, 0.145098, 0.188235],[0.137255, 0.145098, 0.184314],[0.133333, 0.141176, 0.180392],
[0.133333, 0.137255, 0.172549],[0.133333, 0.137255, 0.168627],[0.129412, 0.133333, 0.160784],[0.129412, 0.133333, 0.156863],
[0.129412, 0.129412, 0.152941],[0.129412, 0.129412, 0.14902],[0.129412, 0.12549, 0.145098],[0.129412, 0.12549, 0.141176],
[0.129412, 0.12549, 0.137255],[0.133333, 0.121569, 0.133333],[0.133333, 0.121569, 0.129412],[0.133333, 0.121569, 0.129412],
[0.137255, 0.121569, 0.12549],[0.141176, 0.121569, 0.12549],[0.145098, 0.121569, 0.121569],[0.14902, 0.121569, 0.121569],
[0.152941, 0.12549, 0.121569],[0.156863, 0.12549, 0.121569],[0.160784, 0.12549, 0.121569],[0.164706, 0.12549, 0.121569],
[0.168627, 0.129412, 0.121569],[0.176471, 0.129412, 0.121569],[0.180392, 0.133333, 0.121569],[0.188235, 0.133333, 0.12549],
[0.192157, 0.133333, 0.12549],[0.196078, 0.137255, 0.12549],[0.203922, 0.137255, 0.12549],[0.207843, 0.141176, 0.129412],
[0.215686, 0.141176, 0.129412],[0.219608, 0.145098, 0.129412],[0.227451, 0.145098, 0.129412],[0.231373, 0.14902, 0.133333],
[0.239216, 0.14902, 0.133333],[0.243137, 0.152941, 0.133333],[0.247059, 0.152941, 0.137255],[0.254902, 0.152941, 0.137255],
[0.258824, 0.156863, 0.137255],[0.266667, 0.156863, 0.137255],[0.270588, 0.160784, 0.141176],[0.278431, 0.160784, 0.141176],
[0.282353, 0.164706, 0.141176],[0.290196, 0.164706, 0.145098],[0.294118, 0.168627, 0.145098],[0.301961, 0.168627, 0.145098],
[0.305882, 0.168627, 0.14902],[0.313725, 0.172549, 0.14902],[0.317647, 0.172549, 0.14902],[0.32549, 0.176471, 0.14902],
[0.329412, 0.176471, 0.152941],[0.337255, 0.180392, 0.152941],[0.341176, 0.180392, 0.152941],[0.34902, 0.180392, 0.156863],
[0.352941, 0.184314, 0.156863],[0.360784, 0.184314, 0.156863],[0.364706, 0.188235, 0.160784],[0.372549, 0.188235, 0.160784],
[0.376471, 0.192157, 0.160784],[0.384314, 0.192157, 0.160784],[0.388235, 0.192157, 0.164706],[0.396078, 0.196078, 0.164706],
[0.4, 0.196078, 0.164706],[0.407843, 0.2, 0.168627],[0.415686, 0.2, 0.168627],[0.419608, 0.203922, 0.168627],
[0.427451, 0.203922, 0.168627],[0.431373, 0.203922, 0.172549],[0.439216, 0.207843, 0.172549],[0.443137, 0.207843, 0.172549],
[0.45098, 0.211765, 0.176471],[0.454902, 0.211765, 0.176471],[0.462745, 0.211765, 0.176471],[0.470588, 0.215686, 0.180392],
[0.47451, 0.215686, 0.180392],[0.482353, 0.219608, 0.180392],[0.486275, 0.219608, 0.180392],[0.494118, 0.219608, 0.184314],
[0.498039, 0.223529, 0.184314],[0.505882, 0.223529, 0.184314],[0.513725, 0.227451, 0.188235],[0.517647, 0.227451, 0.188235],
[0.52549, 0.227451, 0.188235],[0.529412, 0.231373, 0.188235],[0.537255, 0.231373, 0.192157],[0.541176, 0.235294, 0.192157],
[0.54902, 0.235294, 0.192157],[0.556863, 0.235294, 0.196078],[0.560784, 0.239216, 0.196078],[0.568627, 0.239216, 0.196078],
[0.572549, 0.243137, 0.2],[0.580392, 0.243137, 0.2],[0.588235, 0.243137, 0.2],[0.592157, 0.247059, 0.2],
[0.6, 0.247059, 0.203922],[0.607843, 0.25098, 0.203922],[0.611765, 0.25098, 0.203922],[0.619608, 0.25098, 0.207843],
[0.623529, 0.254902, 0.207843],[0.631373, 0.254902, 0.207843],[0.639216, 0.254902, 0.211765],[0.643137, 0.258824, 0.211765],
[0.65098, 0.258824, 0.211765],[0.658824, 0.262745, 0.211765],[0.662745, 0.262745, 0.215686],[0.670588, 0.262745, 0.215686],
[0.678431, 0.266667, 0.215686],[0.682353, 0.266667, 0.219608],[0.690196, 0.266667, 0.219608],[0.698039, 0.270588, 0.219608],
[0.701961, 0.270588, 0.219608],[0.709804, 0.27451, 0.223529],[0.717647, 0.27451, 0.223529],[0.721569, 0.27451, 0.223529],
[0.729412, 0.278431, 0.227451],[0.737255, 0.278431, 0.227451],[0.741176, 0.278431, 0.227451],[0.74902, 0.282353, 0.231373],
[0.756863, 0.282353, 0.231373],[0.760784, 0.282353, 0.231373],[0.768627, 0.286275, 0.231373],[0.776471, 0.286275, 0.235294],
[0.780392, 0.290196, 0.235294],[0.788235, 0.290196, 0.235294],[0.796078, 0.290196, 0.239216],[0.8, 0.294118, 0.239216],
[0.807843, 0.294118, 0.239216],[0.815686, 0.294118, 0.243137],[0.819608, 0.298039, 0.243137],[0.827451, 0.298039, 0.243137],
[0.835294, 0.298039, 0.243137],[0.843137, 0.301961, 0.247059],[0.847059, 0.301961, 0.247059],[0.854902, 0.301961, 0.247059],
[0.862745, 0.305882, 0.25098],[0.866667, 0.305882, 0.25098],[0.87451, 0.305882, 0.25098],[0.882353, 0.309804, 0.254902],
[0.890196, 0.309804, 0.254902],[0.894118, 0.309804, 0.254902],[0.901961, 0.313725, 0.254902] ], [
[0.054902, 0.580392, 0.980392],[0.0666667, 0.576471, 0.972549],[0.0745098, 0.572549, 0.964706],[0.0823529, 0.568627, 0.956863],
[0.0901961, 0.564706, 0.94902],[0.0941176, 0.560784, 0.941176],[0.101961, 0.556863, 0.933333],[0.105882, 0.54902, 0.92549],
[0.109804, 0.545098, 0.917647],[0.113725, 0.541176, 0.913725],[0.121569, 0.537255, 0.905882],[0.12549, 0.533333, 0.898039],
[0.129412, 0.529412, 0.890196],[0.133333, 0.52549, 0.882353],[0.133333, 0.521569, 0.87451],[0.137255, 0.517647, 0.866667],
[0.141176, 0.513725, 0.858824],[0.145098, 0.509804, 0.85098],[0.14902, 0.505882, 0.843137],[0.14902, 0.501961, 0.835294],
[0.152941, 0.498039, 0.827451],[0.152941, 0.494118, 0.819608],[0.156863, 0.490196, 0.811765],[0.160784, 0.486275, 0.803922],
[0.160784, 0.482353, 0.8],[0.164706, 0.478431, 0.792157],[0.164706, 0.47451, 0.784314],[0.168627, 0.470588, 0.776471],
[0.168627, 0.466667, 0.768627],[0.168627, 0.462745, 0.760784],[0.172549, 0.458824, 0.752941],[0.172549, 0.454902, 0.745098],
[0.172549, 0.45098, 0.737255],[0.176471, 0.447059, 0.733333],[0.176471, 0.443137, 0.72549],[0.176471, 0.439216, 0.717647],
[0.180392, 0.435294, 0.709804],[0.180392, 0.431373, 0.701961],[0.180392, 0.427451, 0.694118],[0.180392, 0.423529, 0.686275],
[0.184314, 0.419608, 0.678431],[0.184314, 0.415686, 0.67451],[0.184314, 0.411765, 0.666667],[0.184314, 0.407843, 0.658824],
[0.184314, 0.403922, 0.65098],[0.184314, 0.4, 0.643137],[0.188235, 0.396078, 0.635294],[0.188235, 0.392157, 0.631373],
[0.188235, 0.388235, 0.623529],[0.188235, 0.384314, 0.615686],[0.188235, 0.380392, 0.607843],[0.188235, 0.376471, 0.6],
[0.188235, 0.372549, 0.596078],[0.188235, 0.368627, 0.588235],[0.188235, 0.364706, 0.580392],[0.188235, 0.360784, 0.572549],
[0.188235, 0.356863, 0.564706],[0.188235, 0.352941, 0.560784],[0.188235, 0.34902, 0.552941],[0.188235, 0.345098, 0.545098],
[0.188235, 0.341176, 0.537255],[0.188235, 0.341176, 0.529412],[0.188235, 0.337255, 0.52549],[0.188235, 0.333333, 0.517647],
[0.188235, 0.329412, 0.509804],[0.188235, 0.32549, 0.501961],[0.188235, 0.321569, 0.498039],[0.188235, 0.317647, 0.490196],
[0.188235, 0.313725, 0.482353],[0.188235, 0.309804, 0.47451],[0.184314, 0.305882, 0.470588],[0.184314, 0.301961, 0.462745],
[0.184314, 0.298039, 0.454902],[0.184314, 0.294118, 0.45098],[0.184314, 0.290196, 0.443137],[0.184314, 0.286275, 0.435294],
[0.184314, 0.282353, 0.427451],[0.180392, 0.282353, 0.423529],[0.180392, 0.278431, 0.415686],[0.180392, 0.27451, 0.407843],
[0.180392, 0.270588, 0.403922],[0.180392, 0.266667, 0.396078],[0.176471, 0.262745, 0.388235],[0.176471, 0.258824, 0.384314],
[0.176471, 0.254902, 0.376471],[0.176471, 0.25098, 0.368627],[0.172549, 0.247059, 0.364706],[0.172549, 0.243137, 0.356863],
[0.172549, 0.243137, 0.34902],[0.172549, 0.239216, 0.345098],[0.168627, 0.235294, 0.337255],[0.168627, 0.231373, 0.329412],
[0.168627, 0.227451, 0.32549],[0.164706, 0.223529, 0.317647],[0.164706, 0.219608, 0.309804],[0.164706, 0.215686, 0.305882],
[0.164706, 0.211765, 0.298039],[0.160784, 0.211765, 0.294118],[0.160784, 0.207843, 0.286275],[0.160784, 0.203922, 0.278431],
[0.156863, 0.2, 0.27451],[0.156863, 0.196078, 0.266667],[0.152941, 0.192157, 0.262745],[0.152941, 0.188235, 0.254902],
[0.152941, 0.188235, 0.25098],[0.14902, 0.184314, 0.243137],[0.14902, 0.180392, 0.235294],[0.14902, 0.176471, 0.231373],
[0.145098, 0.172549, 0.223529],[0.145098, 0.168627, 0.219608],[0.141176, 0.164706, 0.211765],[0.141176, 0.164706, 0.207843],
[0.137255, 0.160784, 0.2],[0.137255, 0.156863, 0.196078],[0.137255, 0.152941, 0.188235],[0.133333, 0.14902, 0.184314],
[0.133333, 0.14902, 0.176471],[0.133333, 0.145098, 0.172549],[0.129412, 0.141176, 0.168627],[0.129412, 0.141176, 0.160784],
[0.129412, 0.137255, 0.156863],[0.12549, 0.133333, 0.152941],[0.12549, 0.133333, 0.14902],[0.12549, 0.129412, 0.145098],
[0.12549, 0.129412, 0.137255],[0.12549, 0.129412, 0.137255],[0.12549, 0.129412, 0.133333],[0.129412, 0.12549, 0.129412],
[0.129412, 0.12549, 0.12549],[0.129412, 0.12549, 0.12549],[0.133333, 0.129412, 0.121569],[0.137255, 0.129412, 0.121569],
[0.137255, 0.129412, 0.117647],[0.141176, 0.133333, 0.117647],[0.145098, 0.133333, 0.117647],[0.14902, 0.133333, 0.117647],
[0.14902, 0.137255, 0.117647],[0.152941, 0.141176, 0.117647],[0.156863, 0.141176, 0.117647],[0.160784, 0.145098, 0.117647],
[0.164706, 0.14902, 0.117647],[0.172549, 0.14902, 0.117647],[0.176471, 0.152941, 0.117647],[0.180392, 0.156863, 0.117647],
[0.184314, 0.160784, 0.117647],[0.188235, 0.160784, 0.117647],[0.192157, 0.164706, 0.117647],[0.196078, 0.168627, 0.117647],
[0.2, 0.172549, 0.117647],[0.203922, 0.176471, 0.121569],[0.207843, 0.176471, 0.121569],[0.211765, 0.180392, 0.121569],
[0.219608, 0.184314, 0.121569],[0.223529, 0.188235, 0.121569],[0.227451, 0.192157, 0.121569],[0.231373, 0.192157, 0.121569],
[0.235294, 0.196078, 0.121569],[0.239216, 0.2, 0.121569],[0.243137, 0.203922, 0.121569],[0.247059, 0.207843, 0.12549],
[0.254902, 0.207843, 0.12549],[0.258824, 0.211765, 0.12549],[0.262745, 0.215686, 0.12549],[0.266667, 0.219608, 0.12549],
[0.270588, 0.223529, 0.12549],[0.27451, 0.227451, 0.12549],[0.278431, 0.227451, 0.12549],[0.282353, 0.231373, 0.12549],
[0.290196, 0.235294, 0.12549],[0.294118, 0.239216, 0.12549],[0.298039, 0.243137, 0.12549],[0.301961, 0.247059, 0.129412],
[0.305882, 0.247059, 0.129412],[0.309804, 0.25098, 0.129412],[0.313725, 0.254902, 0.129412],[0.321569, 0.258824, 0.129412],
[0.32549, 0.262745, 0.129412],[0.329412, 0.266667, 0.129412],[0.333333, 0.270588, 0.129412],[0.337255, 0.270588, 0.129412],
[0.341176, 0.27451, 0.129412],[0.345098, 0.278431, 0.129412],[0.352941, 0.282353, 0.129412],[0.356863, 0.286275, 0.129412],
[0.360784, 0.290196, 0.129412],[0.364706, 0.294118, 0.129412],[0.368627, 0.294118, 0.129412],[0.372549, 0.298039, 0.129412],
[0.380392, 0.301961, 0.129412],[0.384314, 0.305882, 0.129412],[0.388235, 0.309804, 0.129412],[0.392157, 0.313725, 0.129412],
[0.396078, 0.317647, 0.129412],[0.403922, 0.321569, 0.129412],[0.407843, 0.321569, 0.129412],[0.411765, 0.32549, 0.129412],
[0.415686, 0.329412, 0.129412],[0.419608, 0.333333, 0.129412],[0.423529, 0.337255, 0.129412],[0.431373, 0.341176, 0.129412],
[0.435294, 0.345098, 0.129412],[0.439216, 0.34902, 0.129412],[0.443137, 0.352941, 0.129412],[0.447059, 0.352941, 0.129412],
[0.454902, 0.356863, 0.129412],[0.458824, 0.360784, 0.129412],[0.462745, 0.364706, 0.129412],[0.466667, 0.368627, 0.129412],
[0.470588, 0.372549, 0.129412],[0.478431, 0.376471, 0.129412],[0.482353, 0.380392, 0.129412],[0.486275, 0.384314, 0.129412],
[0.490196, 0.388235, 0.129412],[0.498039, 0.388235, 0.129412],[0.501961, 0.392157, 0.129412],[0.505882, 0.396078, 0.129412],
[0.509804, 0.4, 0.129412],[0.513725, 0.403922, 0.129412],[0.521569, 0.407843, 0.12549],[0.52549, 0.411765, 0.12549],
[0.529412, 0.415686, 0.12549],[0.533333, 0.419608, 0.12549],[0.541176, 0.423529, 0.12549],[0.545098, 0.427451, 0.12549],
[0.54902, 0.431373, 0.12549],[0.552941, 0.431373, 0.12549],[0.556863, 0.435294, 0.12549],[0.564706, 0.439216, 0.12549],
[0.568627, 0.443137, 0.121569],[0.572549, 0.447059, 0.121569],[0.576471, 0.45098, 0.121569],[0.584314, 0.454902, 0.121569],
[0.588235, 0.458824, 0.121569],[0.592157, 0.462745, 0.121569],[0.596078, 0.466667, 0.121569],[0.603922, 0.470588, 0.117647],
[0.607843, 0.47451, 0.117647],[0.611765, 0.478431, 0.117647],[0.615686, 0.482353, 0.117647],[0.623529, 0.486275, 0.117647],
[0.627451, 0.490196, 0.117647],[0.631373, 0.494118, 0.113725],[0.639216, 0.494118, 0.113725],[0.643137, 0.498039, 0.113725],
[0.647059, 0.501961, 0.113725],[0.65098, 0.505882, 0.113725],[0.658824, 0.509804, 0.109804],[0.662745, 0.513725, 0.109804],
[0.666667, 0.517647, 0.109804],[0.670588, 0.521569, 0.109804],[0.678431, 0.52549, 0.105882],[0.682353, 0.529412, 0.105882],
[0.686275, 0.533333, 0.105882],[0.694118, 0.537255, 0.105882],[0.698039, 0.541176, 0.101961],[0.701961, 0.545098, 0.101961] ], 
[[0.129412, 0.317647, 0.858824],[0.145098, 0.321569, 0.858824],[0.164706, 0.32549, 0.858824],
[0.180392, 0.329412, 0.862745],[0.192157, 0.333333, 0.862745],[0.207843, 0.337255, 0.862745],[0.219608, 0.341176, 0.862745],
[0.231373, 0.345098, 0.862745],[0.243137, 0.34902, 0.866667],[0.254902, 0.356863, 0.866667],[0.262745, 0.360784, 0.866667],
[0.27451, 0.364706, 0.866667],[0.282353, 0.368627, 0.866667],[0.294118, 0.372549, 0.870588],[0.301961, 0.376471, 0.870588],
[0.313725, 0.380392, 0.870588],[0.321569, 0.388235, 0.870588],[0.329412, 0.392157, 0.87451],[0.337255, 0.396078, 0.87451],
[0.345098, 0.4, 0.87451],[0.352941, 0.403922, 0.87451],[0.360784, 0.407843, 0.87451],[0.368627, 0.411765, 0.878431],
[0.376471, 0.419608, 0.878431],[0.384314, 0.423529, 0.878431],[0.392157, 0.427451, 0.878431],[0.4, 0.431373, 0.878431],
[0.407843, 0.435294, 0.882353],[0.411765, 0.439216, 0.882353],[0.419608, 0.447059, 0.882353],[0.427451, 0.45098, 0.882353],
[0.435294, 0.454902, 0.882353],[0.439216, 0.458824, 0.886275],[0.447059, 0.462745, 0.886275],[0.454902, 0.470588, 0.886275],
[0.458824, 0.47451, 0.886275],[0.466667, 0.478431, 0.886275],[0.47451, 0.482353, 0.890196],[0.478431, 0.486275, 0.890196],
[0.486275, 0.494118, 0.890196],[0.490196, 0.498039, 0.890196],[0.498039, 0.501961, 0.890196],[0.505882, 0.505882, 0.894118],
[0.509804, 0.509804, 0.894118],[0.517647, 0.517647, 0.894118],[0.521569, 0.521569, 0.894118],[0.529412, 0.52549, 0.894118],
[0.533333, 0.529412, 0.898039],[0.541176, 0.533333, 0.898039],[0.545098, 0.541176, 0.898039],[0.552941, 0.545098, 0.898039],
[0.556863, 0.54902, 0.898039],[0.560784, 0.552941, 0.898039],[0.568627, 0.560784, 0.901961],[0.572549, 0.564706, 0.901961],
[0.580392, 0.568627, 0.901961],[0.584314, 0.572549, 0.901961],[0.592157, 0.580392, 0.901961],[0.596078, 0.584314, 0.905882],
[0.6, 0.588235, 0.905882],[0.607843, 0.592157, 0.905882],[0.611765, 0.6, 0.905882],[0.619608, 0.603922, 0.905882],
[0.623529, 0.607843, 0.905882],[0.627451, 0.611765, 0.909804],[0.635294, 0.619608, 0.909804],[0.639216, 0.623529, 0.909804],
[0.643137, 0.627451, 0.909804],[0.65098, 0.631373, 0.909804],[0.654902, 0.639216, 0.909804],[0.658824, 0.643137, 0.913725],
[0.666667, 0.647059, 0.913725],[0.670588, 0.65098, 0.913725],[0.67451, 0.658824, 0.913725],[0.682353, 0.662745, 0.913725],
[0.686275, 0.666667, 0.913725],[0.690196, 0.67451, 0.917647],[0.698039, 0.678431, 0.917647],[0.701961, 0.682353, 0.917647],
[0.705882, 0.686275, 0.917647],[0.709804, 0.694118, 0.917647],[0.717647, 0.698039, 0.917647],[0.721569, 0.701961, 0.921569],
[0.72549, 0.705882, 0.921569],[0.733333, 0.713725, 0.921569],[0.737255, 0.717647, 0.921569],[0.741176, 0.721569, 0.921569],
[0.745098, 0.729412, 0.921569],[0.752941, 0.733333, 0.921569],[0.756863, 0.737255, 0.92549],[0.760784, 0.745098, 0.92549],
[0.764706, 0.74902, 0.92549],[0.772549, 0.752941, 0.92549],[0.776471, 0.756863, 0.92549],[0.780392, 0.764706, 0.92549],
[0.784314, 0.768627, 0.92549],[0.792157, 0.772549, 0.929412],[0.796078, 0.780392, 0.929412],[0.8, 0.784314, 0.929412],
[0.803922, 0.788235, 0.929412],[0.811765, 0.796078, 0.929412],[0.815686, 0.8, 0.929412],[0.819608, 0.803922, 0.929412],
[0.823529, 0.811765, 0.933333],[0.827451, 0.815686, 0.933333],[0.835294, 0.819608, 0.933333],[0.839216, 0.827451, 0.933333],
[0.843137, 0.831373, 0.933333],[0.847059, 0.835294, 0.933333],[0.854902, 0.843137, 0.933333],[0.858824, 0.847059, 0.933333],
[0.862745, 0.85098, 0.937255],[0.866667, 0.854902, 0.937255],[0.870588, 0.862745, 0.937255],[0.878431, 0.866667, 0.937255],
[0.882353, 0.870588, 0.937255],[0.886275, 0.878431, 0.937255],[0.890196, 0.882353, 0.937255],[0.894118, 0.886275, 0.937255],
[0.898039, 0.890196, 0.937255],[0.901961, 0.894118, 0.933333],[0.905882, 0.898039, 0.933333],[0.909804, 0.898039, 0.933333],
[0.913725, 0.901961, 0.929412],[0.917647, 0.905882, 0.929412],[0.921569, 0.905882, 0.92549],[0.92549, 0.905882, 0.921569],
[0.929412, 0.905882, 0.917647],[0.933333, 0.905882, 0.913725],[0.933333, 0.905882, 0.909804],[0.937255, 0.901961, 0.905882],
[0.937255, 0.901961, 0.901961],[0.941176, 0.898039, 0.894118],[0.941176, 0.894118, 0.890196],[0.945098, 0.890196, 0.882353],
[0.945098, 0.886275, 0.87451],[0.945098, 0.882353, 0.866667],[0.945098, 0.878431, 0.862745],[0.945098, 0.870588, 0.854902],
[0.94902, 0.866667, 0.847059],[0.94902, 0.858824, 0.839216],[0.94902, 0.854902, 0.831373],[0.94902, 0.847059, 0.823529],
[0.94902, 0.843137, 0.815686],[0.94902, 0.835294, 0.807843],[0.94902, 0.831373, 0.8],[0.94902, 0.823529, 0.796078],
[0.94902, 0.819608, 0.788235],[0.94902, 0.811765, 0.780392],[0.94902, 0.807843, 0.772549],[0.94902, 0.8, 0.764706],
[0.945098, 0.796078, 0.756863],[0.945098, 0.788235, 0.74902],[0.945098, 0.780392, 0.741176],[0.945098, 0.776471, 0.733333],
[0.945098, 0.768627, 0.72549],[0.945098, 0.764706, 0.717647],[0.945098, 0.756863, 0.709804],[0.945098, 0.752941, 0.705882],
[0.941176, 0.745098, 0.698039],[0.941176, 0.741176, 0.690196],[0.941176, 0.733333, 0.682353],[0.941176, 0.72549, 0.67451],
[0.941176, 0.721569, 0.666667],[0.941176, 0.713725, 0.658824],[0.937255, 0.709804, 0.65098],[0.937255, 0.701961, 0.647059],
[0.937255, 0.698039, 0.639216],[0.937255, 0.690196, 0.631373],[0.933333, 0.686275, 0.623529],[0.933333, 0.678431, 0.615686],
[0.933333, 0.67451, 0.607843],[0.933333, 0.666667, 0.6],[0.929412, 0.658824, 0.596078],[0.929412, 0.654902, 0.588235],
[0.929412, 0.647059, 0.580392],[0.929412, 0.643137, 0.572549],[0.92549, 0.635294, 0.564706],[0.92549, 0.631373, 0.556863],
[0.92549, 0.623529, 0.54902],[0.921569, 0.615686, 0.545098],[0.921569, 0.611765, 0.537255],[0.921569, 0.603922, 0.529412],
[0.917647, 0.6, 0.521569],[0.917647, 0.592157, 0.513725],[0.917647, 0.588235, 0.509804],[0.913725, 0.580392, 0.501961],
[0.913725, 0.576471, 0.494118],[0.909804, 0.568627, 0.486275],[0.909804, 0.560784, 0.478431],[0.909804, 0.556863, 0.47451],
[0.905882, 0.54902, 0.466667],[0.905882, 0.545098, 0.458824],[0.901961, 0.537255, 0.45098],[0.901961, 0.533333, 0.443137],
[0.901961, 0.52549, 0.439216],[0.898039, 0.517647, 0.431373],[0.898039, 0.513725, 0.423529],[0.894118, 0.505882, 0.415686],
[0.894118, 0.501961, 0.411765],[0.890196, 0.494118, 0.403922],[0.890196, 0.486275, 0.396078],[0.886275, 0.482353, 0.388235],
[0.886275, 0.47451, 0.384314],[0.882353, 0.470588, 0.376471],[0.882353, 0.462745, 0.368627],[0.878431, 0.454902, 0.360784],
[0.878431, 0.45098, 0.356863],[0.87451, 0.443137, 0.34902],[0.87451, 0.435294, 0.341176],[0.870588, 0.431373, 0.333333],
[0.870588, 0.423529, 0.329412],[0.866667, 0.415686, 0.321569],[0.866667, 0.411765, 0.313725],[0.862745, 0.403922, 0.309804],
[0.858824, 0.4, 0.301961],[0.858824, 0.392157, 0.294118],[0.854902, 0.384314, 0.286275],[0.854902, 0.376471, 0.282353],
[0.85098, 0.372549, 0.27451],[0.847059, 0.364706, 0.266667],[0.847059, 0.356863, 0.262745],[0.843137, 0.352941, 0.254902],
[0.843137, 0.345098, 0.247059],[0.839216, 0.337255, 0.243137],[0.835294, 0.329412, 0.235294],[0.835294, 0.321569, 0.227451],
[0.831373, 0.317647, 0.219608],[0.827451, 0.309804, 0.215686],[0.827451, 0.301961, 0.207843],[0.823529, 0.294118, 0.2],
[0.819608, 0.286275, 0.196078],[0.819608, 0.278431, 0.188235],[0.815686, 0.270588, 0.180392],[0.811765, 0.266667, 0.176471],
[0.811765, 0.258824, 0.168627],[0.807843, 0.25098, 0.160784],[0.803922, 0.239216, 0.156863],[0.803922, 0.231373, 0.14902],
[0.8, 0.223529, 0.141176],[0.796078, 0.215686, 0.133333],[0.792157, 0.207843, 0.129412],[0.792157, 0.2, 0.121569],
[0.788235, 0.188235, 0.113725],[0.784314, 0.180392, 0.105882],[0.784314, 0.168627, 0.101961],[0.780392, 0.160784, 0.0941176],
[0.776471, 0.14902, 0.0862745],[0.772549, 0.137255, 0.0784314],[0.772549, 0.12549, 0.0705882],[0.768627, 0.109804, 0.0627451],
[0.764706, 0.0980392, 0.054902],[0.760784, 0.0784314, 0.0470588],[0.756863, 0.0588235, 0.0392157],[0.756863, 0.0352941, 0.0313725],
[0.752941, 0.00784314, 0.0235294] ], [[0.141176, 0.501961, 1],[0.160784, 0.505882, 1],
[0.180392, 0.505882, 1],[0.196078, 0.509804, 1],[0.211765, 0.513725, 1],[0.227451, 0.517647, 1],
[0.243137, 0.521569, 1],[0.254902, 0.52549, 1],[0.266667, 0.529412, 1],[0.278431, 0.529412, 1],
[0.290196, 0.533333, 0.996078],[0.301961, 0.537255, 0.996078],[0.309804, 0.541176, 0.996078],[0.321569, 0.545098, 0.996078],
[0.329412, 0.54902, 0.996078],[0.341176, 0.552941, 0.996078],[0.34902, 0.556863, 0.996078],[0.356863, 0.556863, 0.996078],
[0.368627, 0.560784, 0.996078],[0.376471, 0.564706, 0.996078],[0.384314, 0.568627, 0.996078],[0.392157, 0.572549, 0.996078],
[0.4, 0.576471, 0.996078],[0.407843, 0.580392, 0.996078],[0.415686, 0.584314, 0.996078],[0.423529, 0.588235, 0.996078],
[0.431373, 0.588235, 0.996078],[0.439216, 0.592157, 0.996078],[0.447059, 0.596078, 0.996078],[0.454902, 0.6, 0.996078],
[0.462745, 0.603922, 0.996078],[0.466667, 0.607843, 0.996078],[0.47451, 0.611765, 0.996078],[0.482353, 0.615686, 0.996078],
[0.490196, 0.619608, 0.996078],[0.494118, 0.623529, 0.996078],[0.501961, 0.623529, 0.996078],[0.509804, 0.627451, 0.996078],
[0.513725, 0.631373, 0.996078],[0.521569, 0.635294, 0.996078],[0.529412, 0.639216, 0.996078],[0.533333, 0.643137, 0.996078],
[0.541176, 0.647059, 0.996078],[0.545098, 0.65098, 0.996078],[0.552941, 0.654902, 0.996078],[0.560784, 0.658824, 0.996078],
[0.564706, 0.662745, 0.996078],[0.572549, 0.666667, 0.996078],[0.576471, 0.670588, 0.996078],[0.584314, 0.670588, 0.996078],
[0.588235, 0.67451, 0.996078],[0.596078, 0.678431, 0.996078],[0.6, 0.682353, 0.996078],[0.607843, 0.686275, 0.996078],
[0.611765, 0.690196, 0.996078],[0.615686, 0.694118, 0.996078],[0.623529, 0.698039, 0.996078],[0.627451, 0.701961, 0.996078],
[0.635294, 0.705882, 0.996078],[0.639216, 0.709804, 0.996078],[0.647059, 0.713725, 0.996078],[0.65098, 0.717647, 0.992157],
[0.654902, 0.721569, 0.992157],[0.662745, 0.72549, 0.992157],[0.666667, 0.729412, 0.992157],[0.670588, 0.733333, 0.992157],
[0.678431, 0.733333, 0.992157],[0.682353, 0.737255, 0.992157],[0.686275, 0.741176, 0.992157],[0.694118, 0.745098, 0.992157],
[0.698039, 0.74902, 0.992157],[0.701961, 0.752941, 0.992157],[0.709804, 0.756863, 0.992157],[0.713725, 0.760784, 0.992157],
[0.717647, 0.764706, 0.992157],[0.72549, 0.768627, 0.992157],[0.729412, 0.772549, 0.992157],[0.733333, 0.776471, 0.992157],
[0.741176, 0.780392, 0.992157],[0.745098, 0.784314, 0.992157],[0.74902, 0.788235, 0.992157],[0.752941, 0.792157, 0.992157],
[0.760784, 0.796078, 0.992157],[0.764706, 0.8, 0.988235],[0.768627, 0.803922, 0.988235],[0.776471, 0.807843, 0.988235],
[0.780392, 0.811765, 0.988235],[0.784314, 0.815686, 0.988235],[0.788235, 0.819608, 0.988235],[0.796078, 0.823529, 0.988235],
[0.8, 0.827451, 0.988235],[0.803922, 0.831373, 0.988235],[0.807843, 0.835294, 0.988235],[0.811765, 0.839216, 0.988235],
[0.819608, 0.843137, 0.988235],[0.823529, 0.847059, 0.988235],[0.827451, 0.85098, 0.988235],[0.831373, 0.854902, 0.988235],
[0.839216, 0.858824, 0.988235],[0.843137, 0.862745, 0.988235],[0.847059, 0.866667, 0.984314],[0.85098, 0.870588, 0.984314],
[0.854902, 0.87451, 0.984314],[0.862745, 0.878431, 0.984314],[0.866667, 0.882353, 0.984314],[0.870588, 0.886275, 0.984314],
[0.87451, 0.890196, 0.984314],[0.878431, 0.894118, 0.984314],[0.886275, 0.898039, 0.984314],[0.890196, 0.901961, 0.984314],
[0.894118, 0.905882, 0.984314],[0.898039, 0.909804, 0.984314],[0.901961, 0.913725, 0.984314],[0.905882, 0.917647, 0.984314],
[0.913725, 0.921569, 0.980392],[0.917647, 0.92549, 0.980392],[0.921569, 0.929412, 0.980392],[0.92549, 0.933333, 0.980392],
[0.929412, 0.937255, 0.980392],[0.933333, 0.941176, 0.980392],[0.941176, 0.945098, 0.980392],[0.945098, 0.94902, 0.980392],
[0.94902, 0.952941, 0.980392],[0.952941, 0.956863, 0.980392],[0.956863, 0.960784, 0.976471],[0.960784, 0.964706, 0.976471],
[0.968627, 0.964706, 0.976471],[0.972549, 0.968627, 0.972549],[0.976471, 0.964706, 0.968627],[0.976471, 0.964706, 0.960784],
[0.980392, 0.960784, 0.956863],[0.980392, 0.956863, 0.94902],[0.980392, 0.94902, 0.941176],[0.984314, 0.945098, 0.937255],
[0.984314, 0.941176, 0.929412],[0.984314, 0.937255, 0.92549],[0.984314, 0.929412, 0.917647],[0.988235, 0.92549, 0.909804],
[0.988235, 0.921569, 0.905882],[0.988235, 0.917647, 0.898039],[0.988235, 0.909804, 0.890196],[0.988235, 0.905882, 0.886275],
[0.988235, 0.901961, 0.878431],[0.988235, 0.898039, 0.87451],[0.992157, 0.890196, 0.866667],[0.992157, 0.886275, 0.858824],
[0.992157, 0.882353, 0.854902],[0.992157, 0.87451, 0.847059],[0.992157, 0.870588, 0.843137],[0.992157, 0.866667, 0.835294],
[0.992157, 0.862745, 0.827451],[0.992157, 0.854902, 0.823529],[0.992157, 0.85098, 0.815686],[0.992157, 0.847059, 0.811765],
[0.996078, 0.843137, 0.803922],[0.996078, 0.835294, 0.796078],[0.996078, 0.831373, 0.792157],[0.996078, 0.827451, 0.784314],
[0.996078, 0.823529, 0.780392],[0.996078, 0.815686, 0.772549],[0.996078, 0.811765, 0.764706],[0.996078, 0.807843, 0.760784],
[0.996078, 0.803922, 0.752941],[0.996078, 0.796078, 0.74902],[0.996078, 0.792157, 0.741176],[0.996078, 0.788235, 0.737255],
[0.996078, 0.780392, 0.729412],[0.996078, 0.776471, 0.72549],[0.996078, 0.772549, 0.717647],[0.996078, 0.768627, 0.709804],
[0.996078, 0.760784, 0.705882],[0.996078, 0.756863, 0.698039],[0.992157, 0.752941, 0.694118],[0.992157, 0.74902, 0.686275],
[0.992157, 0.741176, 0.682353],[0.992157, 0.737255, 0.67451],[0.992157, 0.733333, 0.670588],[0.992157, 0.729412, 0.662745],
[0.992157, 0.721569, 0.654902],[0.992157, 0.717647, 0.65098],[0.992157, 0.713725, 0.643137],[0.992157, 0.705882, 0.639216],
[0.992157, 0.701961, 0.631373],[0.988235, 0.698039, 0.627451],[0.988235, 0.694118, 0.619608],[0.988235, 0.686275, 0.615686],
[0.988235, 0.682353, 0.607843],[0.988235, 0.678431, 0.603922],[0.988235, 0.67451, 0.596078],[0.988235, 0.666667, 0.592157],
[0.984314, 0.662745, 0.584314],[0.984314, 0.658824, 0.580392],[0.984314, 0.65098, 0.572549],[0.984314, 0.647059, 0.568627],
[0.984314, 0.643137, 0.560784],[0.984314, 0.639216, 0.556863],[0.980392, 0.631373, 0.54902],[0.980392, 0.627451, 0.545098],
[0.980392, 0.623529, 0.537255],[0.980392, 0.619608, 0.533333],[0.976471, 0.611765, 0.52549],[0.976471, 0.607843, 0.521569],
[0.976471, 0.603922, 0.513725],[0.976471, 0.596078, 0.509804],[0.976471, 0.592157, 0.501961],[0.972549, 0.588235, 0.498039],
[0.972549, 0.580392, 0.490196],[0.972549, 0.576471, 0.486275],[0.972549, 0.572549, 0.478431],[0.968627, 0.568627, 0.47451],
[0.968627, 0.560784, 0.466667],[0.968627, 0.556863, 0.462745],[0.964706, 0.552941, 0.454902],[0.964706, 0.545098, 0.45098],
[0.964706, 0.541176, 0.443137],[0.964706, 0.537255, 0.439216],[0.960784, 0.529412, 0.431373],[0.960784, 0.52549, 0.427451],
[0.960784, 0.521569, 0.419608],[0.956863, 0.513725, 0.415686],[0.956863, 0.509804, 0.411765],[0.956863, 0.505882, 0.403922],
[0.952941, 0.498039, 0.4],[0.952941, 0.494118, 0.392157],[0.952941, 0.490196, 0.388235],[0.94902, 0.482353, 0.380392],
[0.94902, 0.478431, 0.376471],[0.94902, 0.47451, 0.368627],[0.945098, 0.466667, 0.364706],[0.945098, 0.462745, 0.360784],
[0.945098, 0.458824, 0.352941],[0.941176, 0.45098, 0.34902],[0.941176, 0.447059, 0.341176],[0.941176, 0.443137, 0.337255],
[0.937255, 0.435294, 0.329412],[0.937255, 0.431373, 0.32549],[0.933333, 0.423529, 0.321569],[0.933333, 0.419608, 0.313725],
[0.933333, 0.415686, 0.309804],[0.929412, 0.407843, 0.301961],[0.929412, 0.403922, 0.298039],[0.92549, 0.396078, 0.290196],
[0.92549, 0.392157, 0.286275],[0.92549, 0.384314, 0.282353],[0.921569, 0.380392, 0.27451],[0.921569, 0.376471, 0.270588],
[0.917647, 0.368627, 0.262745],[0.917647, 0.364706, 0.258824],[0.913725, 0.356863, 0.254902],[0.913725, 0.352941, 0.247059],
[0.913725, 0.345098, 0.243137],[0.909804, 0.341176, 0.235294],[0.909804, 0.333333, 0.231373],[0.905882, 0.32549, 0.227451],
[0.905882, 0.321569, 0.219608],[0.901961, 0.313725, 0.215686] ], [[0, 0.85098, 1],
[0, 0.85098, 1],[0.054902, 0.854902, 1],[0.101961, 0.854902, 1],[0.133333, 0.854902, 1],
[0.156863, 0.854902, 1],[0.180392, 0.858824, 1],[0.2, 0.858824, 1],[0.215686, 0.858824, 1],
[0.231373, 0.862745, 1],[0.247059, 0.862745, 1],[0.262745, 0.862745, 1],[0.27451, 0.866667, 1],
[0.286275, 0.866667, 1],[0.301961, 0.866667, 1],[0.313725, 0.870588, 1],[0.321569, 0.870588, 1],
[0.333333, 0.870588, 1],[0.345098, 0.87451, 1],[0.356863, 0.87451, 1],[0.364706, 0.87451, 1],
[0.376471, 0.87451, 1],[0.384314, 0.878431, 1],[0.392157, 0.878431, 1],[0.403922, 0.878431, 1],
[0.411765, 0.882353, 1],[0.419608, 0.882353, 1],[0.427451, 0.882353, 1],[0.435294, 0.886275, 1],
[0.447059, 0.886275, 1],[0.454902, 0.886275, 1],[0.462745, 0.886275, 1],[0.470588, 0.890196, 1],
[0.478431, 0.890196, 1],[0.482353, 0.890196, 1],[0.490196, 0.894118, 1],[0.498039, 0.894118, 1],
[0.505882, 0.894118, 1],[0.513725, 0.898039, 1],[0.521569, 0.898039, 1],[0.52549, 0.898039, 1],
[0.533333, 0.901961, 1],[0.541176, 0.901961, 1],[0.54902, 0.901961, 1],[0.552941, 0.901961, 1],
[0.560784, 0.905882, 1],[0.568627, 0.905882, 1],[0.572549, 0.905882, 1],[0.580392, 0.909804, 1],
[0.588235, 0.909804, 1],[0.592157, 0.909804, 1],[0.6, 0.913725, 1],[0.607843, 0.913725, 1],
[0.611765, 0.913725, 1],[0.619608, 0.913725, 1],[0.623529, 0.917647, 1],[0.631373, 0.917647, 1],
[0.635294, 0.917647, 1],[0.643137, 0.921569, 1],[0.647059, 0.921569, 1],[0.654902, 0.921569, 1],
[0.658824, 0.92549, 1],[0.666667, 0.92549, 1],[0.670588, 0.92549, 1],[0.678431, 0.92549, 1],
[0.682353, 0.929412, 1],[0.690196, 0.929412, 1],[0.694118, 0.929412, 1],[0.701961, 0.933333, 1],
[0.705882, 0.933333, 1],[0.709804, 0.933333, 1],[0.717647, 0.937255, 1],[0.721569, 0.937255, 1],
[0.729412, 0.937255, 1],[0.733333, 0.937255, 1],[0.737255, 0.941176, 1],[0.745098, 0.941176, 1],
[0.74902, 0.941176, 1],[0.752941, 0.945098, 1],[0.760784, 0.945098, 1],[0.764706, 0.945098, 1],
[0.772549, 0.945098, 1],[0.776471, 0.94902, 1],[0.780392, 0.94902, 1],[0.788235, 0.94902, 1],
[0.792157, 0.952941, 1],[0.796078, 0.952941, 1],[0.803922, 0.952941, 1],[0.807843, 0.956863, 1],
[0.811765, 0.956863, 1],[0.815686, 0.956863, 1],[0.823529, 0.956863, 1],[0.827451, 0.960784, 1],
[0.831373, 0.960784, 1],[0.839216, 0.960784, 1],[0.843137, 0.964706, 1],[0.847059, 0.964706, 1],
[0.85098, 0.964706, 1],[0.858824, 0.964706, 1],[0.862745, 0.968627, 1],[0.866667, 0.968627, 1],
[0.870588, 0.968627, 1],[0.878431, 0.972549, 1],[0.882353, 0.972549, 1],[0.886275, 0.972549, 1],
[0.890196, 0.972549, 1],[0.898039, 0.976471, 1],[0.901961, 0.976471, 1],[0.905882, 0.976471, 1],
[0.909804, 0.980392, 1],[0.917647, 0.980392, 1],[0.921569, 0.980392, 1],[0.92549, 0.984314, 1],
[0.929412, 0.984314, 1],[0.933333, 0.984314, 1],[0.941176, 0.984314, 1],[0.945098, 0.988235, 1],
[0.94902, 0.988235, 1],[0.952941, 0.988235, 1],[0.960784, 0.992157, 1],[0.964706, 0.992157, 1],
[0.968627, 0.992157, 1],[0.972549, 0.992157, 1],[0.976471, 0.996078, 1],[0.984314, 0.996078, 1],
[0.988235, 0.996078, 1],[0.992157, 0.996078, 1],[0.996078, 1, 1],[1, 1, 1],
[1, 0.996078, 1],[1, 0.992157, 1],[1, 0.992157, 1],[1, 0.988235, 0.996078],
[1, 0.984314, 0.996078],[1, 0.984314, 0.996078],[1, 0.980392, 0.996078],[1, 0.980392, 0.996078],
[1, 0.976471, 0.996078],[1, 0.972549, 0.996078],[1, 0.972549, 0.996078],[1, 0.968627, 0.996078],
[1, 0.964706, 0.992157],[1, 0.964706, 0.992157],[1, 0.960784, 0.992157],[1, 0.960784, 0.992157],
[1, 0.956863, 0.992157],[1, 0.952941, 0.992157],[1, 0.952941, 0.992157],[1, 0.94902, 0.992157],
[1, 0.945098, 0.992157],[1, 0.945098, 0.988235],[1, 0.941176, 0.988235],[1, 0.937255, 0.988235],
[1, 0.937255, 0.988235],[1, 0.933333, 0.988235],[1, 0.933333, 0.988235],[1, 0.929412, 0.988235],
[1, 0.92549, 0.988235],[1, 0.92549, 0.988235],[1, 0.921569, 0.984314],[1, 0.917647, 0.984314],
[1, 0.917647, 0.984314],[1, 0.913725, 0.984314],[1, 0.913725, 0.984314],[1, 0.909804, 0.984314],
[1, 0.905882, 0.984314],[1, 0.905882, 0.984314],[1, 0.901961, 0.984314],[1, 0.898039, 0.980392],
[1, 0.898039, 0.980392],[1, 0.894118, 0.980392],[1, 0.894118, 0.980392],[1, 0.890196, 0.980392],
[0.996078, 0.886275, 0.980392],[0.996078, 0.886275, 0.980392],[0.996078, 0.882353, 0.980392],[0.996078, 0.878431, 0.980392],
[0.996078, 0.878431, 0.976471],[0.996078, 0.87451, 0.976471],[0.996078, 0.870588, 0.976471],[0.996078, 0.870588, 0.976471],
[0.996078, 0.866667, 0.976471],[0.996078, 0.866667, 0.976471],[0.996078, 0.862745, 0.976471],[0.996078, 0.858824, 0.976471],
[0.996078, 0.858824, 0.972549],[0.996078, 0.854902, 0.972549],[0.996078, 0.85098, 0.972549],[0.996078, 0.85098, 0.972549],
[0.996078, 0.847059, 0.972549],[0.996078, 0.847059, 0.972549],[0.996078, 0.843137, 0.972549],[0.996078, 0.839216, 0.972549],
[0.996078, 0.839216, 0.972549],[0.996078, 0.835294, 0.968627],[0.996078, 0.831373, 0.968627],[0.996078, 0.831373, 0.968627],
[0.996078, 0.827451, 0.968627],[0.996078, 0.823529, 0.968627],[0.996078, 0.823529, 0.968627],[0.992157, 0.819608, 0.968627],
[0.992157, 0.819608, 0.968627],[0.992157, 0.815686, 0.964706],[0.992157, 0.811765, 0.964706],[0.992157, 0.811765, 0.964706],
[0.992157, 0.807843, 0.964706],[0.992157, 0.803922, 0.964706],[0.992157, 0.803922, 0.964706],[0.992157, 0.8, 0.964706],
[0.992157, 0.796078, 0.964706],[0.992157, 0.796078, 0.964706],[0.992157, 0.792157, 0.960784],[0.992157, 0.788235, 0.960784],
[0.992157, 0.788235, 0.960784],[0.992157, 0.784314, 0.960784],[0.992157, 0.784314, 0.960784],[0.992157, 0.780392, 0.960784],
[0.992157, 0.776471, 0.960784],[0.992157, 0.776471, 0.960784],[0.988235, 0.772549, 0.956863],[0.988235, 0.768627, 0.956863],
[0.988235, 0.768627, 0.956863],[0.988235, 0.764706, 0.956863],[0.988235, 0.760784, 0.956863],[0.988235, 0.760784, 0.956863],
[0.988235, 0.756863, 0.956863],[0.988235, 0.752941, 0.956863],[0.988235, 0.752941, 0.956863],[0.988235, 0.74902, 0.952941],
[0.988235, 0.74902, 0.952941],[0.988235, 0.745098, 0.952941],[0.988235, 0.741176, 0.952941],[0.988235, 0.741176, 0.952941],
[0.988235, 0.737255, 0.952941],[0.984314, 0.733333, 0.952941],[0.984314, 0.733333, 0.952941],[0.984314, 0.729412, 0.94902],
[0.984314, 0.72549, 0.94902],[0.984314, 0.72549, 0.94902],[0.984314, 0.721569, 0.94902],[0.984314, 0.717647, 0.94902],
[0.984314, 0.717647, 0.94902],[0.984314, 0.713725, 0.94902],[0.984314, 0.709804, 0.94902],[0.984314, 0.709804, 0.945098],
[0.984314, 0.705882, 0.945098],[0.984314, 0.701961, 0.945098],[0.984314, 0.701961, 0.945098],[0.980392, 0.698039, 0.945098],
[0.980392, 0.694118, 0.945098],[0.980392, 0.694118, 0.945098],[0.980392, 0.690196, 0.945098],[0.980392, 0.686275, 0.941176],
[0.980392, 0.686275, 0.941176],[0.980392, 0.682353, 0.941176],[0.980392, 0.682353, 0.941176] ], [
[0.211765, 0.65098, 0.0862745],[0.211765, 0.647059, 0.0901961],[0.211765, 0.643137, 0.0901961],[0.211765, 0.635294, 0.0941176],
[0.211765, 0.631373, 0.0941176],[0.211765, 0.627451, 0.0941176],[0.211765, 0.623529, 0.0980392],[0.211765, 0.619608, 0.0980392],
[0.211765, 0.615686, 0.101961],[0.211765, 0.607843, 0.101961],[0.211765, 0.603922, 0.101961],[0.211765, 0.6, 0.105882],
[0.211765, 0.596078, 0.105882],[0.211765, 0.592157, 0.105882],[0.211765, 0.584314, 0.105882],[0.211765, 0.580392, 0.109804],
[0.211765, 0.576471, 0.109804],[0.211765, 0.572549, 0.109804],[0.211765, 0.568627, 0.109804],[0.211765, 0.564706, 0.113725],
[0.211765, 0.556863, 0.113725],[0.211765, 0.552941, 0.113725],[0.211765, 0.54902, 0.113725],[0.211765, 0.545098, 0.117647],
[0.211765, 0.541176, 0.117647],[0.211765, 0.537255, 0.117647],[0.211765, 0.529412, 0.117647],[0.211765, 0.52549, 0.117647],
[0.207843, 0.521569, 0.117647],[0.207843, 0.517647, 0.121569],[0.207843, 0.513725, 0.121569],[0.207843, 0.509804, 0.121569],
[0.207843, 0.505882, 0.121569],[0.207843, 0.498039, 0.121569],[0.207843, 0.494118, 0.121569],[0.207843, 0.490196, 0.12549],
[0.207843, 0.486275, 0.12549],[0.207843, 0.482353, 0.12549],[0.207843, 0.478431, 0.12549],[0.203922, 0.47451, 0.12549],
[0.203922, 0.466667, 0.12549],[0.203922, 0.462745, 0.12549],[0.203922, 0.458824, 0.12549],[0.203922, 0.454902, 0.12549],
[0.203922, 0.45098, 0.12549],[0.203922, 0.447059, 0.129412],[0.2, 0.443137, 0.129412],[0.2, 0.439216, 0.129412],
[0.2, 0.431373, 0.129412],[0.2, 0.427451, 0.129412],[0.2, 0.423529, 0.129412],[0.2, 0.419608, 0.129412],
[0.2, 0.415686, 0.129412],[0.196078, 0.411765, 0.129412],[0.196078, 0.407843, 0.129412],[0.196078, 0.403922, 0.129412],
[0.196078, 0.4, 0.129412],[0.196078, 0.392157, 0.129412],[0.192157, 0.388235, 0.129412],[0.192157, 0.384314, 0.129412],
[0.192157, 0.380392, 0.129412],[0.192157, 0.376471, 0.129412],[0.192157, 0.372549, 0.129412],[0.188235, 0.368627, 0.129412],
[0.188235, 0.364706, 0.129412],[0.188235, 0.360784, 0.129412],[0.188235, 0.356863, 0.129412],[0.188235, 0.352941, 0.129412],
[0.184314, 0.345098, 0.129412],[0.184314, 0.341176, 0.129412],[0.184314, 0.337255, 0.129412],[0.184314, 0.333333, 0.129412],
[0.180392, 0.329412, 0.129412],[0.180392, 0.32549, 0.129412],[0.180392, 0.321569, 0.129412],[0.180392, 0.317647, 0.129412],
[0.176471, 0.313725, 0.129412],[0.176471, 0.309804, 0.129412],[0.176471, 0.305882, 0.129412],[0.176471, 0.301961, 0.129412],
[0.172549, 0.298039, 0.129412],[0.172549, 0.294118, 0.129412],[0.172549, 0.290196, 0.129412],[0.172549, 0.286275, 0.129412],
[0.168627, 0.278431, 0.129412],[0.168627, 0.27451, 0.129412],[0.168627, 0.270588, 0.129412],[0.164706, 0.266667, 0.129412],
[0.164706, 0.262745, 0.12549],[0.164706, 0.258824, 0.12549],[0.164706, 0.254902, 0.12549],[0.160784, 0.25098, 0.12549],
[0.160784, 0.247059, 0.12549],[0.160784, 0.243137, 0.12549],[0.156863, 0.239216, 0.12549],[0.156863, 0.235294, 0.12549],
[0.156863, 0.231373, 0.12549],[0.152941, 0.227451, 0.12549],[0.152941, 0.223529, 0.12549],[0.152941, 0.219608, 0.121569],
[0.14902, 0.215686, 0.121569],[0.14902, 0.211765, 0.121569],[0.14902, 0.207843, 0.121569],[0.145098, 0.203922, 0.121569],
[0.145098, 0.2, 0.121569],[0.145098, 0.196078, 0.121569],[0.141176, 0.192157, 0.121569],[0.141176, 0.188235, 0.121569],
[0.141176, 0.184314, 0.117647],[0.137255, 0.180392, 0.117647],[0.137255, 0.176471, 0.117647],[0.137255, 0.172549, 0.117647],
[0.133333, 0.168627, 0.117647],[0.133333, 0.164706, 0.117647],[0.133333, 0.160784, 0.117647],[0.129412, 0.156863, 0.117647],
[0.129412, 0.152941, 0.113725],[0.129412, 0.152941, 0.113725],[0.12549, 0.14902, 0.113725],[0.12549, 0.145098, 0.113725],
[0.12549, 0.141176, 0.113725],[0.12549, 0.137255, 0.113725],[0.12549, 0.137255, 0.113725],[0.12549, 0.133333, 0.113725],
[0.129412, 0.133333, 0.113725],[0.129412, 0.129412, 0.113725],[0.129412, 0.129412, 0.113725],[0.133333, 0.12549, 0.113725],
[0.137255, 0.12549, 0.113725],[0.137255, 0.12549, 0.113725],[0.141176, 0.12549, 0.113725],[0.145098, 0.12549, 0.113725],
[0.14902, 0.12549, 0.113725],[0.156863, 0.12549, 0.113725],[0.160784, 0.129412, 0.117647],[0.164706, 0.129412, 0.117647],
[0.172549, 0.129412, 0.117647],[0.176471, 0.133333, 0.117647],[0.184314, 0.133333, 0.117647],[0.188235, 0.133333, 0.121569],
[0.196078, 0.137255, 0.121569],[0.2, 0.137255, 0.121569],[0.207843, 0.141176, 0.121569],[0.211765, 0.141176, 0.12549],
[0.219608, 0.145098, 0.12549],[0.227451, 0.145098, 0.12549],[0.231373, 0.14902, 0.12549],[0.239216, 0.14902, 0.129412],
[0.243137, 0.152941, 0.129412],[0.25098, 0.152941, 0.129412],[0.258824, 0.156863, 0.129412],[0.262745, 0.156863, 0.133333],
[0.270588, 0.160784, 0.133333],[0.278431, 0.160784, 0.133333],[0.282353, 0.164706, 0.133333],[0.290196, 0.164706, 0.133333],
[0.294118, 0.168627, 0.137255],[0.301961, 0.168627, 0.137255],[0.309804, 0.172549, 0.137255],[0.313725, 0.176471, 0.137255],
[0.321569, 0.176471, 0.141176],[0.329412, 0.180392, 0.141176],[0.333333, 0.180392, 0.141176],[0.341176, 0.180392, 0.141176],
[0.34902, 0.184314, 0.145098],[0.352941, 0.184314, 0.145098],[0.360784, 0.188235, 0.145098],[0.368627, 0.188235, 0.145098],
[0.372549, 0.192157, 0.14902],[0.380392, 0.192157, 0.14902],[0.388235, 0.196078, 0.14902],[0.392157, 0.196078, 0.14902],
[0.4, 0.2, 0.152941],[0.407843, 0.2, 0.152941],[0.415686, 0.203922, 0.152941],[0.419608, 0.203922, 0.152941],
[0.427451, 0.207843, 0.152941],[0.435294, 0.207843, 0.156863],[0.439216, 0.211765, 0.156863],[0.447059, 0.211765, 0.156863],
[0.454902, 0.215686, 0.156863],[0.458824, 0.215686, 0.160784],[0.466667, 0.219608, 0.160784],[0.47451, 0.219608, 0.160784],
[0.482353, 0.223529, 0.160784],[0.486275, 0.223529, 0.164706],[0.494118, 0.227451, 0.164706],[0.501961, 0.227451, 0.164706],
[0.509804, 0.231373, 0.164706],[0.513725, 0.231373, 0.164706],[0.521569, 0.231373, 0.168627],[0.529412, 0.235294, 0.168627],
[0.537255, 0.235294, 0.168627],[0.541176, 0.239216, 0.168627],[0.54902, 0.239216, 0.172549],[0.556863, 0.243137, 0.172549],
[0.564706, 0.243137, 0.172549],[0.568627, 0.247059, 0.172549],[0.576471, 0.247059, 0.172549],[0.584314, 0.25098, 0.176471],
[0.592157, 0.25098, 0.176471],[0.596078, 0.254902, 0.176471],[0.603922, 0.254902, 0.176471],[0.611765, 0.254902, 0.180392],
[0.619608, 0.258824, 0.180392],[0.627451, 0.258824, 0.180392],[0.631373, 0.262745, 0.180392],[0.639216, 0.262745, 0.180392],
[0.647059, 0.266667, 0.184314],[0.654902, 0.266667, 0.184314],[0.662745, 0.270588, 0.184314],[0.666667, 0.270588, 0.184314],
[0.67451, 0.27451, 0.184314],[0.682353, 0.27451, 0.188235],[0.690196, 0.27451, 0.188235],[0.698039, 0.278431, 0.188235],
[0.705882, 0.278431, 0.188235],[0.709804, 0.282353, 0.192157],[0.717647, 0.282353, 0.192157],[0.72549, 0.286275, 0.192157],
[0.733333, 0.286275, 0.192157],[0.741176, 0.286275, 0.192157],[0.74902, 0.290196, 0.196078],[0.752941, 0.290196, 0.196078],
[0.760784, 0.294118, 0.196078],[0.768627, 0.294118, 0.196078],[0.776471, 0.298039, 0.196078],[0.784314, 0.298039, 0.2],
[0.792157, 0.298039, 0.2],[0.8, 0.301961, 0.2],[0.803922, 0.301961, 0.2],[0.811765, 0.305882, 0.203922],
[0.819608, 0.305882, 0.203922],[0.827451, 0.309804, 0.203922],[0.835294, 0.309804, 0.203922],[0.843137, 0.309804, 0.203922],
[0.85098, 0.313725, 0.207843],[0.858824, 0.313725, 0.207843],[0.862745, 0.317647, 0.207843],[0.870588, 0.317647, 0.207843],
[0.878431, 0.321569, 0.207843],[0.886275, 0.321569, 0.211765],[0.894118, 0.321569, 0.211765],[0.901961, 0.32549, 0.211765],
[0.909804, 0.32549, 0.211765],[0.917647, 0.329412, 0.211765],[0.92549, 0.329412, 0.215686],[0.933333, 0.333333, 0.215686],
[0.941176, 0.333333, 0.215686],[0.945098, 0.333333, 0.215686],[0.952941, 0.337255, 0.215686],[0.960784, 0.337255, 0.219608],
[0.968627, 0.341176, 0.219608],[0.976471, 0.341176, 0.219608],[0.984314, 0.341176, 0.219608],[0.992157, 0.345098, 0.219608] ], 
[[0.223529, 0.592157, 0.054902],[0.231373, 0.592157, 0.0705882],[0.239216, 0.596078, 0.0823529],
[0.247059, 0.6, 0.0941176],[0.25098, 0.603922, 0.101961],[0.258824, 0.603922, 0.113725],[0.266667, 0.607843, 0.121569],
[0.27451, 0.611765, 0.133333],[0.282353, 0.611765, 0.141176],[0.290196, 0.615686, 0.14902],[0.294118, 0.619608, 0.156863],
[0.301961, 0.623529, 0.164706],[0.309804, 0.623529, 0.172549],[0.317647, 0.627451, 0.180392],[0.321569, 0.631373, 0.188235],
[0.329412, 0.631373, 0.196078],[0.337255, 0.635294, 0.203922],[0.341176, 0.639216, 0.211765],[0.34902, 0.643137, 0.219608],
[0.352941, 0.643137, 0.223529],[0.360784, 0.647059, 0.231373],[0.368627, 0.65098, 0.239216],[0.372549, 0.65098, 0.247059],
[0.380392, 0.654902, 0.254902],[0.384314, 0.658824, 0.258824],[0.392157, 0.662745, 0.266667],[0.396078, 0.662745, 0.27451],
[0.403922, 0.666667, 0.278431],[0.407843, 0.670588, 0.286275],[0.415686, 0.670588, 0.294118],[0.419608, 0.67451, 0.301961],
[0.427451, 0.678431, 0.305882],[0.431373, 0.682353, 0.313725],[0.439216, 0.682353, 0.321569],[0.443137, 0.686275, 0.32549],
[0.45098, 0.690196, 0.333333],[0.454902, 0.690196, 0.341176],[0.462745, 0.694118, 0.345098],[0.466667, 0.698039, 0.352941],
[0.47451, 0.698039, 0.360784],[0.478431, 0.701961, 0.364706],[0.486275, 0.705882, 0.372549],[0.490196, 0.709804, 0.380392],
[0.494118, 0.709804, 0.384314],[0.501961, 0.713725, 0.392157],[0.505882, 0.717647, 0.4],[0.513725, 0.717647, 0.403922],
[0.517647, 0.721569, 0.411765],[0.521569, 0.72549, 0.415686],[0.529412, 0.729412, 0.423529],[0.533333, 0.729412, 0.431373],
[0.541176, 0.733333, 0.435294],[0.545098, 0.737255, 0.443137],[0.54902, 0.737255, 0.45098],[0.556863, 0.741176, 0.454902],
[0.560784, 0.745098, 0.462745],[0.568627, 0.745098, 0.470588],[0.572549, 0.74902, 0.47451],[0.576471, 0.752941, 0.482353],
[0.584314, 0.756863, 0.486275],[0.588235, 0.756863, 0.494118],[0.592157, 0.760784, 0.501961],[0.6, 0.764706, 0.505882],
[0.603922, 0.764706, 0.513725],[0.607843, 0.768627, 0.521569],[0.615686, 0.772549, 0.52549],[0.619608, 0.772549, 0.533333],
[0.627451, 0.776471, 0.541176],[0.631373, 0.780392, 0.545098],[0.635294, 0.784314, 0.552941],[0.643137, 0.784314, 0.556863],
[0.647059, 0.788235, 0.564706],[0.65098, 0.792157, 0.572549],[0.658824, 0.792157, 0.576471],[0.662745, 0.796078, 0.584314],
[0.666667, 0.8, 0.592157],[0.67451, 0.8, 0.596078],[0.678431, 0.803922, 0.603922],[0.682353, 0.807843, 0.611765],
[0.690196, 0.807843, 0.615686],[0.694118, 0.811765, 0.623529],[0.698039, 0.815686, 0.631373],[0.705882, 0.819608, 0.635294],
[0.709804, 0.819608, 0.643137],[0.713725, 0.823529, 0.65098],[0.721569, 0.827451, 0.654902],[0.72549, 0.827451, 0.662745],
[0.729412, 0.831373, 0.670588],[0.737255, 0.835294, 0.67451],[0.741176, 0.835294, 0.682353],[0.745098, 0.839216, 0.690196],
[0.752941, 0.843137, 0.694118],[0.756863, 0.847059, 0.701961],[0.760784, 0.847059, 0.709804],[0.768627, 0.85098, 0.713725],
[0.772549, 0.854902, 0.721569],[0.776471, 0.854902, 0.729412],[0.780392, 0.858824, 0.733333],[0.788235, 0.862745, 0.741176],
[0.792157, 0.862745, 0.74902],[0.796078, 0.866667, 0.752941],[0.803922, 0.870588, 0.760784],[0.807843, 0.870588, 0.768627],
[0.811765, 0.87451, 0.772549],[0.819608, 0.878431, 0.780392],[0.823529, 0.878431, 0.788235],[0.827451, 0.882353, 0.792157],
[0.835294, 0.886275, 0.8],[0.839216, 0.890196, 0.807843],[0.843137, 0.890196, 0.811765],[0.85098, 0.894118, 0.819608],
[0.854902, 0.898039, 0.827451],[0.858824, 0.898039, 0.831373],[0.862745, 0.901961, 0.839216],[0.870588, 0.905882, 0.847059],
[0.87451, 0.905882, 0.85098],[0.878431, 0.909804, 0.858824],[0.886275, 0.909804, 0.862745],[0.890196, 0.913725, 0.870588],
[0.894118, 0.913725, 0.87451],[0.898039, 0.917647, 0.882353],[0.905882, 0.917647, 0.886275],[0.909804, 0.921569, 0.890196],
[0.913725, 0.921569, 0.894118],[0.917647, 0.921569, 0.898039],[0.921569, 0.921569, 0.901961],[0.92549, 0.921569, 0.901961],
[0.929412, 0.921569, 0.901961],[0.933333, 0.917647, 0.901961],[0.937255, 0.917647, 0.901961],[0.941176, 0.913725, 0.901961],
[0.941176, 0.913725, 0.901961],[0.945098, 0.909804, 0.898039],[0.94902, 0.905882, 0.898039],[0.94902, 0.901961, 0.894118],
[0.952941, 0.901961, 0.890196],[0.952941, 0.898039, 0.886275],[0.956863, 0.894118, 0.882353],[0.956863, 0.886275, 0.87451],
[0.956863, 0.882353, 0.870588],[0.960784, 0.878431, 0.866667],[0.960784, 0.87451, 0.862745],[0.960784, 0.870588, 0.854902],
[0.964706, 0.866667, 0.85098],[0.964706, 0.858824, 0.843137],[0.964706, 0.854902, 0.839216],[0.964706, 0.85098, 0.835294],
[0.968627, 0.847059, 0.827451],[0.968627, 0.843137, 0.823529],[0.968627, 0.835294, 0.815686],[0.968627, 0.831373, 0.811765],
[0.972549, 0.827451, 0.807843],[0.972549, 0.823529, 0.8],[0.972549, 0.815686, 0.796078],[0.972549, 0.811765, 0.788235],
[0.972549, 0.807843, 0.784314],[0.972549, 0.803922, 0.776471],[0.976471, 0.8, 0.772549],[0.976471, 0.792157, 0.768627],
[0.976471, 0.788235, 0.760784],[0.976471, 0.784314, 0.756863],[0.976471, 0.780392, 0.74902],[0.976471, 0.772549, 0.745098],
[0.976471, 0.768627, 0.741176],[0.976471, 0.764706, 0.733333],[0.980392, 0.760784, 0.729412],[0.980392, 0.756863, 0.721569],
[0.980392, 0.74902, 0.717647],[0.980392, 0.745098, 0.713725],[0.980392, 0.741176, 0.705882],[0.980392, 0.737255, 0.701961],
[0.980392, 0.729412, 0.694118],[0.980392, 0.72549, 0.690196],[0.980392, 0.721569, 0.686275],[0.980392, 0.717647, 0.678431],
[0.980392, 0.709804, 0.67451],[0.980392, 0.705882, 0.670588],[0.980392, 0.701961, 0.662745],[0.980392, 0.698039, 0.658824],
[0.980392, 0.690196, 0.65098],[0.980392, 0.686275, 0.647059],[0.980392, 0.682353, 0.643137],[0.980392, 0.678431, 0.635294],
[0.980392, 0.670588, 0.631373],[0.980392, 0.666667, 0.627451],[0.980392, 0.662745, 0.619608],[0.980392, 0.658824, 0.615686],
[0.980392, 0.65098, 0.611765],[0.980392, 0.647059, 0.603922],[0.980392, 0.643137, 0.6],[0.980392, 0.639216, 0.596078],
[0.980392, 0.631373, 0.588235],[0.980392, 0.627451, 0.584314],[0.980392, 0.623529, 0.576471],[0.980392, 0.619608, 0.572549],
[0.980392, 0.611765, 0.568627],[0.980392, 0.607843, 0.560784],[0.980392, 0.603922, 0.556863],[0.980392, 0.6, 0.552941],
[0.976471, 0.592157, 0.545098],[0.976471, 0.588235, 0.541176],[0.976471, 0.584314, 0.537255],[0.976471, 0.576471, 0.529412],
[0.976471, 0.572549, 0.52549],[0.976471, 0.568627, 0.521569],[0.976471, 0.564706, 0.517647],[0.976471, 0.556863, 0.509804],
[0.976471, 0.552941, 0.505882],[0.972549, 0.54902, 0.501961],[0.972549, 0.541176, 0.494118],[0.972549, 0.537255, 0.490196],
[0.972549, 0.533333, 0.486275],[0.972549, 0.52549, 0.478431],[0.972549, 0.521569, 0.47451],[0.968627, 0.517647, 0.470588],
[0.968627, 0.509804, 0.462745],[0.968627, 0.505882, 0.458824],[0.968627, 0.501961, 0.454902],[0.968627, 0.494118, 0.45098],
[0.968627, 0.490196, 0.443137],[0.964706, 0.486275, 0.439216],[0.964706, 0.478431, 0.435294],[0.964706, 0.47451, 0.427451],
[0.964706, 0.470588, 0.423529],[0.964706, 0.462745, 0.419608],[0.960784, 0.458824, 0.415686],[0.960784, 0.454902, 0.407843],
[0.960784, 0.447059, 0.403922],[0.960784, 0.443137, 0.4],[0.956863, 0.435294, 0.392157],[0.956863, 0.431373, 0.388235],
[0.956863, 0.427451, 0.384314],[0.956863, 0.419608, 0.380392],[0.952941, 0.415686, 0.372549],[0.952941, 0.407843, 0.368627],
[0.952941, 0.403922, 0.364706],[0.952941, 0.396078, 0.360784],[0.94902, 0.392157, 0.352941],[0.94902, 0.384314, 0.34902],
[0.94902, 0.380392, 0.345098],[0.94902, 0.372549, 0.341176],[0.945098, 0.368627, 0.333333],[0.945098, 0.360784, 0.329412],
[0.945098, 0.356863, 0.32549],[0.941176, 0.34902, 0.321569],[0.941176, 0.345098, 0.313725],[0.941176, 0.337255, 0.309804],
[0.941176, 0.333333, 0.305882],[0.937255, 0.32549, 0.301961],[0.937255, 0.317647, 0.294118],[0.937255, 0.313725, 0.290196],
[0.933333, 0.305882, 0.286275],[0.933333, 0.298039, 0.282353],[0.933333, 0.294118, 0.278431],[0.929412, 0.286275, 0.270588],
[0.929412, 0.278431, 0.266667] ], [[0.223529, 0.592157, 0.054902],[0.231373, 0.592157, 0.0705882],
[0.239216, 0.596078, 0.0823529],[0.247059, 0.6, 0.0941176],[0.25098, 0.603922, 0.101961],[0.258824, 0.603922, 0.113725],
[0.266667, 0.607843, 0.121569],[0.27451, 0.611765, 0.133333],[0.282353, 0.611765, 0.141176],[0.290196, 0.615686, 0.14902],
[0.294118, 0.619608, 0.156863],[0.301961, 0.623529, 0.164706],[0.309804, 0.623529, 0.172549],[0.317647, 0.627451, 0.180392],
[0.321569, 0.631373, 0.188235],[0.329412, 0.631373, 0.196078],[0.337255, 0.635294, 0.203922],[0.341176, 0.639216, 0.211765],
[0.34902, 0.643137, 0.219608],[0.352941, 0.643137, 0.223529],[0.360784, 0.647059, 0.231373],[0.368627, 0.65098, 0.239216],
[0.372549, 0.65098, 0.247059],[0.380392, 0.654902, 0.254902],[0.384314, 0.658824, 0.258824],[0.392157, 0.662745, 0.266667],
[0.396078, 0.662745, 0.27451],[0.403922, 0.666667, 0.278431],[0.407843, 0.670588, 0.286275],[0.415686, 0.670588, 0.294118],
[0.419608, 0.67451, 0.301961],[0.427451, 0.678431, 0.305882],[0.431373, 0.682353, 0.313725],[0.439216, 0.682353, 0.321569],
[0.443137, 0.686275, 0.32549],[0.45098, 0.690196, 0.333333],[0.454902, 0.690196, 0.341176],[0.462745, 0.694118, 0.345098],
[0.466667, 0.698039, 0.352941],[0.47451, 0.698039, 0.360784],[0.478431, 0.701961, 0.364706],[0.486275, 0.705882, 0.372549],
[0.490196, 0.709804, 0.380392],[0.494118, 0.709804, 0.384314],[0.501961, 0.713725, 0.392157],[0.505882, 0.717647, 0.4],
[0.513725, 0.717647, 0.403922],[0.517647, 0.721569, 0.411765],[0.521569, 0.72549, 0.415686],[0.529412, 0.729412, 0.423529],
[0.533333, 0.729412, 0.431373],[0.541176, 0.733333, 0.435294],[0.545098, 0.737255, 0.443137],[0.54902, 0.737255, 0.45098],
[0.556863, 0.741176, 0.454902],[0.560784, 0.745098, 0.462745],[0.568627, 0.745098, 0.470588],[0.572549, 0.74902, 0.47451],
[0.576471, 0.752941, 0.482353],[0.584314, 0.756863, 0.486275],[0.588235, 0.756863, 0.494118],[0.592157, 0.760784, 0.501961],
[0.6, 0.764706, 0.505882],[0.603922, 0.764706, 0.513725],[0.607843, 0.768627, 0.521569],[0.615686, 0.772549, 0.52549],
[0.619608, 0.772549, 0.533333],[0.627451, 0.776471, 0.541176],[0.631373, 0.780392, 0.545098],[0.635294, 0.784314, 0.552941],
[0.643137, 0.784314, 0.556863],[0.647059, 0.788235, 0.564706],[0.65098, 0.792157, 0.572549],[0.658824, 0.792157, 0.576471],
[0.662745, 0.796078, 0.584314],[0.666667, 0.8, 0.592157],[0.67451, 0.8, 0.596078],[0.678431, 0.803922, 0.603922],
[0.682353, 0.807843, 0.611765],[0.690196, 0.807843, 0.615686],[0.694118, 0.811765, 0.623529],[0.698039, 0.815686, 0.631373],
[0.705882, 0.819608, 0.635294],[0.709804, 0.819608, 0.643137],[0.713725, 0.823529, 0.65098],[0.721569, 0.827451, 0.654902],
[0.72549, 0.827451, 0.662745],[0.729412, 0.831373, 0.670588],[0.737255, 0.835294, 0.67451],[0.741176, 0.835294, 0.682353],
[0.745098, 0.839216, 0.690196],[0.752941, 0.843137, 0.694118],[0.756863, 0.847059, 0.701961],[0.760784, 0.847059, 0.709804],
[0.768627, 0.85098, 0.713725],[0.772549, 0.854902, 0.721569],[0.776471, 0.854902, 0.729412],[0.780392, 0.858824, 0.733333],
[0.788235, 0.862745, 0.741176],[0.792157, 0.862745, 0.74902],[0.796078, 0.866667, 0.752941],[0.803922, 0.870588, 0.760784],
[0.807843, 0.870588, 0.768627],[0.811765, 0.87451, 0.772549],[0.819608, 0.878431, 0.780392],[0.823529, 0.878431, 0.788235],
[0.827451, 0.882353, 0.792157],[0.835294, 0.886275, 0.8],[0.839216, 0.890196, 0.807843],[0.843137, 0.890196, 0.811765],
[0.85098, 0.894118, 0.819608],[0.854902, 0.898039, 0.827451],[0.858824, 0.898039, 0.835294],[0.862745, 0.901961, 0.839216],
[0.870588, 0.905882, 0.847059],[0.87451, 0.905882, 0.854902],[0.878431, 0.909804, 0.858824],[0.882353, 0.909804, 0.866667],
[0.890196, 0.913725, 0.870588],[0.894118, 0.913725, 0.878431],[0.898039, 0.917647, 0.886275],[0.901961, 0.917647, 0.890196],
[0.905882, 0.921569, 0.894118],[0.909804, 0.921569, 0.901961],[0.913725, 0.921569, 0.905882],[0.917647, 0.921569, 0.909804],
[0.917647, 0.921569, 0.913725],[0.921569, 0.921569, 0.917647],[0.92549, 0.921569, 0.921569],[0.92549, 0.917647, 0.92549],
[0.92549, 0.917647, 0.929412],[0.929412, 0.913725, 0.929412],[0.929412, 0.913725, 0.933333],[0.929412, 0.909804, 0.933333],
[0.929412, 0.905882, 0.937255],[0.929412, 0.901961, 0.937255],[0.929412, 0.898039, 0.937255],[0.929412, 0.894118, 0.937255],
[0.92549, 0.890196, 0.937255],[0.92549, 0.886275, 0.941176],[0.92549, 0.882353, 0.941176],[0.92549, 0.878431, 0.941176],
[0.921569, 0.87451, 0.941176],[0.921569, 0.870588, 0.941176],[0.921569, 0.866667, 0.937255],[0.917647, 0.862745, 0.937255],
[0.917647, 0.854902, 0.937255],[0.917647, 0.85098, 0.937255],[0.913725, 0.847059, 0.937255],[0.913725, 0.843137, 0.937255],
[0.913725, 0.839216, 0.937255],[0.909804, 0.835294, 0.937255],[0.909804, 0.831373, 0.937255],[0.905882, 0.823529, 0.937255],
[0.905882, 0.819608, 0.937255],[0.905882, 0.815686, 0.937255],[0.901961, 0.811765, 0.937255],[0.901961, 0.807843, 0.937255],
[0.901961, 0.803922, 0.933333],[0.898039, 0.8, 0.933333],[0.898039, 0.792157, 0.933333],[0.894118, 0.788235, 0.933333],
[0.894118, 0.784314, 0.933333],[0.894118, 0.780392, 0.933333],[0.890196, 0.776471, 0.933333],[0.890196, 0.772549, 0.933333],
[0.886275, 0.768627, 0.933333],[0.886275, 0.760784, 0.933333],[0.886275, 0.756863, 0.933333],[0.882353, 0.752941, 0.933333],
[0.882353, 0.74902, 0.929412],[0.878431, 0.745098, 0.929412],[0.878431, 0.741176, 0.929412],[0.878431, 0.733333, 0.929412],
[0.87451, 0.729412, 0.929412],[0.87451, 0.72549, 0.929412],[0.870588, 0.721569, 0.929412],[0.870588, 0.717647, 0.929412],
[0.870588, 0.713725, 0.929412],[0.866667, 0.709804, 0.929412],[0.866667, 0.701961, 0.929412],[0.862745, 0.698039, 0.929412],
[0.862745, 0.694118, 0.92549],[0.858824, 0.690196, 0.92549],[0.858824, 0.686275, 0.92549],[0.858824, 0.682353, 0.92549],
[0.854902, 0.678431, 0.92549],[0.854902, 0.670588, 0.92549],[0.85098, 0.666667, 0.92549],[0.85098, 0.662745, 0.92549],
[0.847059, 0.658824, 0.92549],[0.847059, 0.654902, 0.92549],[0.847059, 0.65098, 0.921569],[0.843137, 0.643137, 0.921569],
[0.843137, 0.639216, 0.921569],[0.839216, 0.635294, 0.921569],[0.839216, 0.631373, 0.921569],[0.835294, 0.627451, 0.921569],
[0.835294, 0.623529, 0.921569],[0.831373, 0.619608, 0.921569],[0.831373, 0.611765, 0.921569],[0.827451, 0.607843, 0.921569],
[0.827451, 0.603922, 0.917647],[0.827451, 0.6, 0.917647],[0.823529, 0.596078, 0.917647],[0.823529, 0.592157, 0.917647],
[0.819608, 0.584314, 0.917647],[0.819608, 0.580392, 0.917647],[0.815686, 0.576471, 0.917647],[0.815686, 0.572549, 0.917647],
[0.811765, 0.568627, 0.917647],[0.811765, 0.564706, 0.917647],[0.807843, 0.556863, 0.913725],[0.807843, 0.552941, 0.913725],
[0.803922, 0.54902, 0.913725],[0.803922, 0.545098, 0.913725],[0.8, 0.541176, 0.913725],[0.8, 0.533333, 0.913725],
[0.796078, 0.529412, 0.913725],[0.796078, 0.52549, 0.913725],[0.792157, 0.521569, 0.913725],[0.792157, 0.517647, 0.909804],
[0.788235, 0.509804, 0.909804],[0.788235, 0.505882, 0.909804],[0.784314, 0.501961, 0.909804],[0.784314, 0.498039, 0.909804],
[0.780392, 0.494118, 0.909804],[0.780392, 0.486275, 0.909804],[0.776471, 0.482353, 0.909804],[0.776471, 0.478431, 0.905882],
[0.772549, 0.47451, 0.905882],[0.772549, 0.470588, 0.905882],[0.768627, 0.462745, 0.905882],[0.768627, 0.458824, 0.905882],
[0.764706, 0.454902, 0.905882],[0.764706, 0.45098, 0.905882],[0.760784, 0.447059, 0.905882],[0.760784, 0.439216, 0.905882],
[0.756863, 0.435294, 0.901961],[0.756863, 0.431373, 0.901961],[0.752941, 0.427451, 0.901961],[0.74902, 0.419608, 0.901961],
[0.74902, 0.415686, 0.901961],[0.745098, 0.411765, 0.901961],[0.745098, 0.407843, 0.901961],[0.741176, 0.4, 0.901961],
[0.741176, 0.396078, 0.898039],[0.737255, 0.392157, 0.898039],[0.737255, 0.384314, 0.898039],[0.733333, 0.380392, 0.898039],
[0.729412, 0.376471, 0.898039],[0.729412, 0.368627, 0.898039],[0.72549, 0.364706, 0.898039],[0.72549, 0.360784, 0.894118],
[0.721569, 0.356863, 0.894118],[0.721569, 0.34902, 0.894118] ], [[0.215686, 0.717647, 0.92549],
[0.219608, 0.717647, 0.921569],[0.219608, 0.717647, 0.917647],[0.223529, 0.717647, 0.917647],[0.223529, 0.717647, 0.913725],
[0.227451, 0.717647, 0.909804],[0.231373, 0.721569, 0.905882],[0.231373, 0.721569, 0.901961],[0.235294, 0.721569, 0.898039],
[0.235294, 0.721569, 0.894118],[0.239216, 0.721569, 0.890196],[0.239216, 0.721569, 0.890196],[0.243137, 0.721569, 0.886275],
[0.243137, 0.721569, 0.882353],[0.247059, 0.721569, 0.878431],[0.247059, 0.721569, 0.87451],[0.25098, 0.721569, 0.870588],
[0.25098, 0.721569, 0.866667],[0.254902, 0.721569, 0.862745],[0.258824, 0.721569, 0.862745],[0.258824, 0.72549, 0.858824],
[0.262745, 0.72549, 0.854902],[0.262745, 0.72549, 0.85098],[0.266667, 0.72549, 0.847059],[0.266667, 0.72549, 0.843137],
[0.270588, 0.72549, 0.839216],[0.270588, 0.72549, 0.835294],[0.270588, 0.72549, 0.831373],[0.27451, 0.72549, 0.831373],
[0.27451, 0.72549, 0.827451],[0.278431, 0.72549, 0.823529],[0.278431, 0.72549, 0.819608],[0.282353, 0.72549, 0.815686],
[0.282353, 0.72549, 0.811765],[0.286275, 0.72549, 0.807843],[0.286275, 0.72549, 0.803922],[0.290196, 0.729412, 0.8],
[0.290196, 0.729412, 0.796078],[0.294118, 0.729412, 0.796078],[0.294118, 0.729412, 0.792157],[0.298039, 0.729412, 0.788235],
[0.298039, 0.729412, 0.784314],[0.301961, 0.729412, 0.780392],[0.301961, 0.729412, 0.776471],[0.305882, 0.729412, 0.772549],
[0.305882, 0.729412, 0.768627],[0.309804, 0.729412, 0.764706],[0.309804, 0.729412, 0.760784],[0.313725, 0.729412, 0.756863],
[0.313725, 0.729412, 0.756863],[0.313725, 0.729412, 0.752941],[0.317647, 0.729412, 0.74902],[0.317647, 0.729412, 0.745098],
[0.321569, 0.729412, 0.741176],[0.321569, 0.729412, 0.737255],[0.32549, 0.733333, 0.733333],[0.32549, 0.733333, 0.729412],
[0.329412, 0.733333, 0.72549],[0.329412, 0.733333, 0.721569],[0.333333, 0.733333, 0.717647],[0.333333, 0.733333, 0.713725],
[0.337255, 0.733333, 0.709804],[0.337255, 0.733333, 0.705882],[0.341176, 0.733333, 0.701961],[0.341176, 0.733333, 0.701961],
[0.345098, 0.733333, 0.698039],[0.345098, 0.733333, 0.694118],[0.34902, 0.733333, 0.690196],[0.34902, 0.733333, 0.686275],
[0.352941, 0.733333, 0.682353],[0.356863, 0.733333, 0.678431],[0.356863, 0.733333, 0.67451],[0.360784, 0.733333, 0.670588],
[0.360784, 0.733333, 0.666667],[0.364706, 0.733333, 0.662745],[0.364706, 0.733333, 0.658824],[0.368627, 0.733333, 0.654902],
[0.368627, 0.733333, 0.65098],[0.372549, 0.733333, 0.647059],[0.376471, 0.733333, 0.643137],[0.376471, 0.733333, 0.639216],
[0.380392, 0.733333, 0.635294],[0.380392, 0.733333, 0.631373],[0.384314, 0.733333, 0.627451],[0.384314, 0.733333, 0.623529],
[0.388235, 0.733333, 0.619608],[0.392157, 0.733333, 0.615686],[0.392157, 0.733333, 0.611765],[0.396078, 0.733333, 0.607843],
[0.4, 0.733333, 0.603922],[0.4, 0.733333, 0.6],[0.403922, 0.733333, 0.596078],[0.407843, 0.733333, 0.592157],
[0.407843, 0.733333, 0.592157],[0.411765, 0.733333, 0.588235],[0.415686, 0.733333, 0.584314],[0.415686, 0.733333, 0.580392],
[0.419608, 0.733333, 0.576471],[0.423529, 0.733333, 0.572549],[0.427451, 0.733333, 0.568627],[0.427451, 0.733333, 0.564706],
[0.431373, 0.733333, 0.560784],[0.435294, 0.733333, 0.556863],[0.439216, 0.733333, 0.552941],[0.443137, 0.733333, 0.54902],
[0.443137, 0.733333, 0.545098],[0.447059, 0.733333, 0.541176],[0.45098, 0.729412, 0.537255],[0.454902, 0.729412, 0.533333],
[0.458824, 0.729412, 0.529412],[0.462745, 0.729412, 0.52549],[0.462745, 0.729412, 0.521569],[0.466667, 0.729412, 0.517647],
[0.470588, 0.729412, 0.513725],[0.47451, 0.729412, 0.509804],[0.478431, 0.729412, 0.505882],[0.482353, 0.729412, 0.501961],
[0.486275, 0.729412, 0.498039],[0.490196, 0.72549, 0.494118],[0.494118, 0.72549, 0.490196],[0.498039, 0.72549, 0.486275],
[0.501961, 0.72549, 0.486275],[0.505882, 0.72549, 0.482353],[0.509804, 0.72549, 0.478431],[0.513725, 0.72549, 0.47451],
[0.517647, 0.721569, 0.470588],[0.521569, 0.721569, 0.466667],[0.52549, 0.721569, 0.466667],[0.529412, 0.721569, 0.462745],
[0.533333, 0.721569, 0.458824],[0.537255, 0.721569, 0.454902],[0.541176, 0.721569, 0.45098],[0.545098, 0.717647, 0.45098],
[0.54902, 0.717647, 0.447059],[0.556863, 0.717647, 0.443137],[0.560784, 0.717647, 0.443137],[0.564706, 0.717647, 0.439216],
[0.568627, 0.713725, 0.435294],[0.572549, 0.713725, 0.435294],[0.576471, 0.713725, 0.431373],[0.580392, 0.713725, 0.427451],
[0.584314, 0.713725, 0.427451],[0.588235, 0.709804, 0.423529],[0.592157, 0.709804, 0.423529],[0.596078, 0.709804, 0.419608],
[0.6, 0.709804, 0.419608],[0.603922, 0.705882, 0.415686],[0.607843, 0.705882, 0.415686],[0.615686, 0.705882, 0.411765],
[0.619608, 0.705882, 0.411765],[0.623529, 0.701961, 0.407843],[0.627451, 0.701961, 0.407843],[0.631373, 0.701961, 0.403922],
[0.635294, 0.701961, 0.403922],[0.639216, 0.698039, 0.4],[0.643137, 0.698039, 0.4],[0.647059, 0.698039, 0.396078],
[0.65098, 0.698039, 0.396078],[0.654902, 0.694118, 0.392157],[0.658824, 0.694118, 0.392157],[0.662745, 0.694118, 0.392157],
[0.666667, 0.694118, 0.388235],[0.670588, 0.690196, 0.388235],[0.67451, 0.690196, 0.388235],[0.678431, 0.690196, 0.384314],
[0.682353, 0.690196, 0.384314],[0.686275, 0.686275, 0.384314],[0.690196, 0.686275, 0.380392],[0.694118, 0.686275, 0.380392],
[0.698039, 0.686275, 0.380392],[0.701961, 0.682353, 0.376471],[0.705882, 0.682353, 0.376471],[0.709804, 0.682353, 0.376471],
[0.709804, 0.678431, 0.376471],[0.713725, 0.678431, 0.372549],[0.717647, 0.678431, 0.372549],[0.721569, 0.67451, 0.372549],
[0.72549, 0.67451, 0.372549],[0.729412, 0.67451, 0.372549],[0.733333, 0.67451, 0.372549],[0.737255, 0.670588, 0.368627],
[0.741176, 0.670588, 0.368627],[0.745098, 0.670588, 0.368627],[0.74902, 0.666667, 0.368627],[0.752941, 0.666667, 0.368627],
[0.756863, 0.666667, 0.368627],[0.760784, 0.662745, 0.368627],[0.760784, 0.662745, 0.368627],[0.764706, 0.662745, 0.368627],
[0.768627, 0.658824, 0.368627],[0.772549, 0.658824, 0.364706],[0.776471, 0.658824, 0.364706],[0.780392, 0.658824, 0.364706],
[0.784314, 0.654902, 0.364706],[0.788235, 0.654902, 0.364706],[0.788235, 0.654902, 0.364706],[0.792157, 0.65098, 0.368627],
[0.796078, 0.65098, 0.368627],[0.8, 0.65098, 0.368627],[0.803922, 0.647059, 0.368627],[0.807843, 0.647059, 0.368627],
[0.811765, 0.647059, 0.368627],[0.811765, 0.643137, 0.368627],[0.815686, 0.643137, 0.368627],[0.819608, 0.643137, 0.368627],
[0.823529, 0.639216, 0.368627],[0.827451, 0.639216, 0.368627],[0.827451, 0.635294, 0.368627],[0.831373, 0.635294, 0.372549],
[0.835294, 0.635294, 0.372549],[0.839216, 0.631373, 0.372549],[0.843137, 0.631373, 0.372549],[0.847059, 0.631373, 0.372549],
[0.847059, 0.627451, 0.372549],[0.85098, 0.627451, 0.376471],[0.854902, 0.627451, 0.376471],[0.858824, 0.623529, 0.376471],
[0.858824, 0.623529, 0.376471],[0.862745, 0.623529, 0.376471],[0.866667, 0.619608, 0.380392],[0.870588, 0.619608, 0.380392],
[0.87451, 0.615686, 0.380392],[0.87451, 0.615686, 0.380392],[0.878431, 0.615686, 0.384314],[0.882353, 0.611765, 0.384314],
[0.886275, 0.611765, 0.384314],[0.886275, 0.611765, 0.388235],[0.890196, 0.607843, 0.388235],[0.894118, 0.607843, 0.388235],
[0.894118, 0.603922, 0.388235],[0.898039, 0.603922, 0.392157],[0.901961, 0.603922, 0.392157],[0.905882, 0.6, 0.392157],
[0.905882, 0.6, 0.396078],[0.909804, 0.6, 0.396078],[0.913725, 0.596078, 0.396078],[0.917647, 0.596078, 0.4],
[0.917647, 0.592157, 0.4],[0.921569, 0.592157, 0.4],[0.92549, 0.592157, 0.403922],[0.92549, 0.588235, 0.403922],
[0.929412, 0.588235, 0.403922],[0.933333, 0.584314, 0.407843],[0.933333, 0.584314, 0.407843],[0.937255, 0.584314, 0.407843],
[0.941176, 0.580392, 0.411765],[0.941176, 0.580392, 0.411765],[0.945098, 0.576471, 0.415686],[0.94902, 0.576471, 0.415686],
[0.94902, 0.576471, 0.415686],[0.952941, 0.572549, 0.419608],[0.956863, 0.572549, 0.419608],[0.956863, 0.568627, 0.423529],
[0.960784, 0.568627, 0.423529],[0.964706, 0.568627, 0.423529],[0.964706, 0.564706, 0.427451] ], [
[0.439216, 0.819608, 1],[0.439216, 0.823529, 1],[0.439216, 0.823529, 1],[0.439216, 0.823529, 1],
[0.439216, 0.823529, 1],[0.439216, 0.823529, 0.996078],[0.439216, 0.823529, 0.996078],[0.439216, 0.823529, 0.992157],
[0.439216, 0.823529, 0.988235],[0.439216, 0.823529, 0.984314],[0.439216, 0.823529, 0.980392],[0.439216, 0.823529, 0.980392],
[0.443137, 0.827451, 0.976471],[0.443137, 0.827451, 0.972549],[0.443137, 0.827451, 0.968627],[0.443137, 0.827451, 0.968627],
[0.443137, 0.827451, 0.964706],[0.443137, 0.827451, 0.960784],[0.443137, 0.827451, 0.956863],[0.443137, 0.827451, 0.952941],
[0.443137, 0.827451, 0.952941],[0.443137, 0.827451, 0.94902],[0.443137, 0.827451, 0.945098],[0.447059, 0.831373, 0.941176],
[0.447059, 0.831373, 0.937255],[0.447059, 0.831373, 0.933333],[0.447059, 0.831373, 0.933333],[0.447059, 0.831373, 0.929412],
[0.447059, 0.831373, 0.92549],[0.447059, 0.831373, 0.921569],[0.447059, 0.831373, 0.917647],[0.45098, 0.831373, 0.917647],
[0.45098, 0.831373, 0.913725],[0.45098, 0.831373, 0.909804],[0.45098, 0.831373, 0.905882],[0.45098, 0.831373, 0.901961],
[0.45098, 0.835294, 0.898039],[0.45098, 0.835294, 0.898039],[0.454902, 0.835294, 0.894118],[0.454902, 0.835294, 0.890196],
[0.454902, 0.835294, 0.886275],[0.454902, 0.835294, 0.882353],[0.454902, 0.835294, 0.882353],[0.454902, 0.835294, 0.878431],
[0.454902, 0.835294, 0.87451],[0.458824, 0.835294, 0.870588],[0.458824, 0.835294, 0.866667],[0.458824, 0.835294, 0.862745],
[0.458824, 0.835294, 0.858824],[0.458824, 0.835294, 0.858824],[0.462745, 0.835294, 0.854902],[0.462745, 0.839216, 0.85098],
[0.462745, 0.839216, 0.847059],[0.462745, 0.839216, 0.843137],[0.462745, 0.839216, 0.839216],[0.466667, 0.839216, 0.839216],
[0.466667, 0.839216, 0.835294],[0.466667, 0.839216, 0.831373],[0.466667, 0.839216, 0.827451],[0.466667, 0.839216, 0.823529],
[0.470588, 0.839216, 0.819608],[0.470588, 0.839216, 0.815686],[0.470588, 0.839216, 0.815686],[0.470588, 0.839216, 0.811765],
[0.47451, 0.839216, 0.807843],[0.47451, 0.839216, 0.803922],[0.47451, 0.839216, 0.8],[0.478431, 0.839216, 0.796078],
[0.478431, 0.839216, 0.792157],[0.478431, 0.839216, 0.788235],[0.478431, 0.839216, 0.788235],[0.482353, 0.839216, 0.784314],
[0.482353, 0.843137, 0.780392],[0.482353, 0.843137, 0.776471],[0.486275, 0.843137, 0.772549],[0.486275, 0.843137, 0.768627],
[0.486275, 0.843137, 0.764706],[0.490196, 0.843137, 0.760784],[0.490196, 0.843137, 0.756863],[0.490196, 0.843137, 0.756863],
[0.494118, 0.843137, 0.752941],[0.494118, 0.843137, 0.74902],[0.494118, 0.843137, 0.745098],[0.498039, 0.843137, 0.741176],
[0.498039, 0.843137, 0.737255],[0.501961, 0.843137, 0.733333],[0.501961, 0.843137, 0.729412],[0.505882, 0.843137, 0.72549],
[0.505882, 0.843137, 0.721569],[0.505882, 0.843137, 0.721569],[0.509804, 0.843137, 0.717647],[0.509804, 0.843137, 0.713725],
[0.513725, 0.843137, 0.709804],[0.513725, 0.843137, 0.705882],[0.517647, 0.843137, 0.701961],[0.517647, 0.843137, 0.698039],
[0.521569, 0.843137, 0.694118],[0.521569, 0.843137, 0.690196],[0.52549, 0.843137, 0.686275],[0.52549, 0.843137, 0.682353],
[0.529412, 0.843137, 0.678431],[0.533333, 0.843137, 0.67451],[0.533333, 0.843137, 0.67451],[0.537255, 0.843137, 0.670588],
[0.537255, 0.843137, 0.666667],[0.541176, 0.843137, 0.662745],[0.545098, 0.843137, 0.658824],[0.545098, 0.843137, 0.654902],
[0.54902, 0.843137, 0.65098],[0.552941, 0.839216, 0.647059],[0.552941, 0.839216, 0.643137],[0.556863, 0.839216, 0.639216],
[0.560784, 0.839216, 0.635294],[0.564706, 0.839216, 0.631373],[0.564706, 0.839216, 0.627451],[0.568627, 0.839216, 0.627451],
[0.572549, 0.839216, 0.623529],[0.576471, 0.839216, 0.619608],[0.576471, 0.839216, 0.615686],[0.580392, 0.839216, 0.611765],
[0.584314, 0.839216, 0.607843],[0.588235, 0.839216, 0.603922],[0.592157, 0.835294, 0.6],[0.596078, 0.835294, 0.6],
[0.6, 0.835294, 0.596078],[0.603922, 0.835294, 0.592157],[0.603922, 0.835294, 0.588235],[0.607843, 0.835294, 0.584314],
[0.611765, 0.835294, 0.580392],[0.615686, 0.835294, 0.580392],[0.619608, 0.831373, 0.576471],[0.623529, 0.831373, 0.572549],
[0.627451, 0.831373, 0.568627],[0.631373, 0.831373, 0.564706],[0.635294, 0.831373, 0.564706],[0.639216, 0.831373, 0.560784],
[0.643137, 0.827451, 0.556863],[0.647059, 0.827451, 0.556863],[0.65098, 0.827451, 0.552941],[0.654902, 0.827451, 0.54902],
[0.658824, 0.827451, 0.54902],[0.662745, 0.827451, 0.545098],[0.666667, 0.823529, 0.541176],[0.670588, 0.823529, 0.541176],
[0.67451, 0.823529, 0.537255],[0.678431, 0.823529, 0.533333],[0.682353, 0.823529, 0.533333],[0.686275, 0.819608, 0.529412],
[0.690196, 0.819608, 0.529412],[0.694118, 0.819608, 0.52549],[0.698039, 0.819608, 0.52549],[0.701961, 0.819608, 0.521569],
[0.705882, 0.815686, 0.521569],[0.709804, 0.815686, 0.517647],[0.713725, 0.815686, 0.517647],[0.717647, 0.815686, 0.513725],
[0.721569, 0.811765, 0.513725],[0.72549, 0.811765, 0.509804],[0.729412, 0.811765, 0.509804],[0.733333, 0.811765, 0.505882],
[0.737255, 0.811765, 0.505882],[0.741176, 0.807843, 0.501961],[0.745098, 0.807843, 0.501961],[0.74902, 0.807843, 0.498039],
[0.752941, 0.807843, 0.498039],[0.756863, 0.803922, 0.498039],[0.760784, 0.803922, 0.494118],[0.764706, 0.803922, 0.494118],
[0.768627, 0.803922, 0.490196],[0.772549, 0.8, 0.490196],[0.772549, 0.8, 0.490196],[0.776471, 0.8, 0.486275],
[0.780392, 0.8, 0.486275],[0.784314, 0.796078, 0.486275],[0.788235, 0.796078, 0.482353],[0.792157, 0.796078, 0.482353],
[0.796078, 0.796078, 0.482353],[0.8, 0.792157, 0.478431],[0.803922, 0.792157, 0.478431],[0.807843, 0.792157, 0.478431],
[0.811765, 0.792157, 0.47451],[0.815686, 0.788235, 0.47451],[0.819608, 0.788235, 0.47451],[0.819608, 0.788235, 0.47451],
[0.823529, 0.788235, 0.470588],[0.827451, 0.784314, 0.470588],[0.831373, 0.784314, 0.470588],[0.835294, 0.784314, 0.470588],
[0.839216, 0.780392, 0.466667],[0.843137, 0.780392, 0.466667],[0.847059, 0.780392, 0.466667],[0.85098, 0.780392, 0.466667],
[0.85098, 0.776471, 0.466667],[0.854902, 0.776471, 0.466667],[0.858824, 0.776471, 0.462745],[0.862745, 0.772549, 0.462745],
[0.866667, 0.772549, 0.462745],[0.870588, 0.772549, 0.462745],[0.87451, 0.772549, 0.462745],[0.878431, 0.768627, 0.462745],
[0.878431, 0.768627, 0.462745],[0.882353, 0.768627, 0.462745],[0.886275, 0.764706, 0.462745],[0.890196, 0.764706, 0.462745],
[0.894118, 0.764706, 0.462745],[0.898039, 0.760784, 0.462745],[0.898039, 0.760784, 0.462745],[0.901961, 0.760784, 0.462745],
[0.905882, 0.760784, 0.462745],[0.909804, 0.756863, 0.462745],[0.913725, 0.756863, 0.462745],[0.913725, 0.756863, 0.462745],
[0.917647, 0.752941, 0.462745],[0.921569, 0.752941, 0.462745],[0.92549, 0.752941, 0.462745],[0.929412, 0.74902, 0.462745],
[0.929412, 0.74902, 0.462745],[0.933333, 0.74902, 0.462745],[0.937255, 0.745098, 0.462745],[0.941176, 0.745098, 0.466667],
[0.945098, 0.745098, 0.466667],[0.945098, 0.741176, 0.466667],[0.94902, 0.741176, 0.466667],[0.952941, 0.741176, 0.466667],
[0.956863, 0.737255, 0.466667],[0.956863, 0.737255, 0.470588],[0.960784, 0.737255, 0.470588],[0.964706, 0.737255, 0.470588],
[0.968627, 0.733333, 0.470588],[0.968627, 0.733333, 0.47451],[0.972549, 0.733333, 0.47451],[0.976471, 0.729412, 0.47451],
[0.976471, 0.729412, 0.47451],[0.980392, 0.729412, 0.478431],[0.984314, 0.72549, 0.478431],[0.988235, 0.72549, 0.478431],
[0.988235, 0.72549, 0.478431],[0.992157, 0.721569, 0.482353],[0.996078, 0.721569, 0.482353],[0.996078, 0.721569, 0.482353],
[1, 0.717647, 0.486275],[1, 0.717647, 0.486275],[1, 0.717647, 0.486275],[1, 0.713725, 0.490196],
[1, 0.713725, 0.490196],[1, 0.713725, 0.490196],[1, 0.709804, 0.494118],[1, 0.709804, 0.494118],
[1, 0.709804, 0.498039],[1, 0.705882, 0.498039],[1, 0.705882, 0.498039],[1, 0.705882, 0.501961],
[1, 0.701961, 0.501961],[1, 0.701961, 0.505882],[1, 0.701961, 0.505882],[1, 0.698039, 0.505882] ], 
[[0.0784314, 0.729412, 0.901961],[0.0980392, 0.72549, 0.901961],[0.113725, 0.72549, 0.901961],
[0.12549, 0.72549, 0.901961],[0.137255, 0.72549, 0.901961],[0.14902, 0.72549, 0.901961],[0.160784, 0.72549, 0.901961],
[0.172549, 0.721569, 0.905882],[0.180392, 0.721569, 0.905882],[0.188235, 0.721569, 0.905882],[0.2, 0.721569, 0.905882],
[0.207843, 0.721569, 0.905882],[0.215686, 0.721569, 0.905882],[0.223529, 0.721569, 0.905882],[0.231373, 0.717647, 0.909804],
[0.235294, 0.717647, 0.909804],[0.243137, 0.717647, 0.909804],[0.25098, 0.717647, 0.909804],[0.258824, 0.717647, 0.909804],
[0.262745, 0.717647, 0.909804],[0.270588, 0.713725, 0.909804],[0.27451, 0.713725, 0.909804],[0.282353, 0.713725, 0.913725],
[0.290196, 0.713725, 0.913725],[0.294118, 0.713725, 0.913725],[0.298039, 0.713725, 0.913725],[0.305882, 0.713725, 0.913725],
[0.309804, 0.709804, 0.913725],[0.317647, 0.709804, 0.913725],[0.321569, 0.709804, 0.913725],[0.32549, 0.709804, 0.913725],
[0.333333, 0.709804, 0.913725],[0.337255, 0.709804, 0.917647],[0.341176, 0.705882, 0.917647],[0.34902, 0.705882, 0.917647],
[0.352941, 0.705882, 0.917647],[0.356863, 0.705882, 0.917647],[0.360784, 0.705882, 0.917647],[0.364706, 0.705882, 0.917647],
[0.372549, 0.701961, 0.917647],[0.376471, 0.701961, 0.917647],[0.380392, 0.701961, 0.917647],[0.384314, 0.701961, 0.921569],
[0.388235, 0.701961, 0.921569],[0.392157, 0.698039, 0.921569],[0.4, 0.698039, 0.921569],[0.403922, 0.698039, 0.921569],
[0.407843, 0.698039, 0.921569],[0.411765, 0.698039, 0.921569],[0.415686, 0.698039, 0.921569],[0.419608, 0.694118, 0.921569],
[0.423529, 0.694118, 0.921569],[0.427451, 0.694118, 0.921569],[0.431373, 0.694118, 0.921569],[0.435294, 0.694118, 0.921569],
[0.439216, 0.694118, 0.921569],[0.443137, 0.690196, 0.92549],[0.447059, 0.690196, 0.92549],[0.45098, 0.690196, 0.92549],
[0.454902, 0.690196, 0.92549],[0.458824, 0.690196, 0.92549],[0.462745, 0.686275, 0.92549],[0.466667, 0.686275, 0.92549],
[0.470588, 0.686275, 0.92549],[0.47451, 0.686275, 0.92549],[0.478431, 0.686275, 0.92549],[0.482353, 0.682353, 0.92549],
[0.486275, 0.682353, 0.92549],[0.490196, 0.682353, 0.92549],[0.494118, 0.682353, 0.92549],[0.498039, 0.682353, 0.92549],
[0.501961, 0.682353, 0.92549],[0.505882, 0.678431, 0.92549],[0.509804, 0.678431, 0.92549],[0.513725, 0.678431, 0.92549],
[0.517647, 0.678431, 0.92549],[0.521569, 0.678431, 0.92549],[0.52549, 0.67451, 0.92549],[0.529412, 0.67451, 0.92549],
[0.533333, 0.67451, 0.92549],[0.537255, 0.67451, 0.92549],[0.541176, 0.67451, 0.92549],[0.545098, 0.670588, 0.92549],
[0.545098, 0.670588, 0.92549],[0.54902, 0.670588, 0.92549],[0.552941, 0.670588, 0.92549],[0.556863, 0.666667, 0.92549],
[0.560784, 0.666667, 0.92549],[0.564706, 0.666667, 0.92549],[0.568627, 0.666667, 0.92549],[0.572549, 0.666667, 0.92549],
[0.576471, 0.662745, 0.92549],[0.580392, 0.662745, 0.92549],[0.584314, 0.662745, 0.92549],[0.584314, 0.662745, 0.921569],
[0.588235, 0.662745, 0.921569],[0.592157, 0.658824, 0.921569],[0.596078, 0.658824, 0.921569],[0.6, 0.658824, 0.921569],
[0.603922, 0.658824, 0.921569],[0.607843, 0.654902, 0.921569],[0.611765, 0.654902, 0.921569],[0.611765, 0.654902, 0.921569],
[0.615686, 0.654902, 0.921569],[0.619608, 0.654902, 0.917647],[0.623529, 0.65098, 0.917647],[0.627451, 0.65098, 0.917647],
[0.631373, 0.65098, 0.917647],[0.635294, 0.65098, 0.917647],[0.639216, 0.647059, 0.917647],[0.639216, 0.647059, 0.913725],
[0.643137, 0.647059, 0.913725],[0.647059, 0.647059, 0.913725],[0.65098, 0.647059, 0.913725],[0.654902, 0.643137, 0.913725],
[0.658824, 0.643137, 0.909804],[0.662745, 0.643137, 0.909804],[0.662745, 0.643137, 0.909804],[0.666667, 0.639216, 0.909804],
[0.670588, 0.639216, 0.905882],[0.67451, 0.639216, 0.905882],[0.678431, 0.639216, 0.905882],[0.682353, 0.639216, 0.905882],
[0.686275, 0.635294, 0.901961],[0.686275, 0.635294, 0.901961],[0.690196, 0.635294, 0.901961],[0.694118, 0.635294, 0.898039],
[0.698039, 0.631373, 0.898039],[0.701961, 0.631373, 0.898039],[0.701961, 0.631373, 0.894118],[0.705882, 0.631373, 0.894118],
[0.709804, 0.631373, 0.894118],[0.713725, 0.627451, 0.890196],[0.717647, 0.627451, 0.890196],[0.717647, 0.627451, 0.890196],
[0.721569, 0.627451, 0.886275],[0.72549, 0.623529, 0.886275],[0.729412, 0.623529, 0.886275],[0.729412, 0.623529, 0.882353],
[0.733333, 0.623529, 0.882353],[0.737255, 0.623529, 0.878431],[0.741176, 0.619608, 0.878431],[0.741176, 0.619608, 0.878431],
[0.745098, 0.619608, 0.87451],[0.74902, 0.619608, 0.87451],[0.752941, 0.619608, 0.87451],[0.752941, 0.615686, 0.870588],
[0.756863, 0.615686, 0.870588],[0.760784, 0.615686, 0.866667],[0.764706, 0.615686, 0.866667],[0.764706, 0.611765, 0.866667],
[0.768627, 0.611765, 0.862745],[0.772549, 0.611765, 0.862745],[0.772549, 0.611765, 0.858824],[0.776471, 0.611765, 0.858824],
[0.780392, 0.607843, 0.858824],[0.780392, 0.607843, 0.854902],[0.784314, 0.607843, 0.854902],[0.788235, 0.607843, 0.85098],
[0.788235, 0.603922, 0.85098],[0.792157, 0.603922, 0.847059],[0.796078, 0.603922, 0.847059],[0.796078, 0.603922, 0.847059],
[0.8, 0.603922, 0.843137],[0.803922, 0.6, 0.843137],[0.803922, 0.6, 0.839216],[0.807843, 0.6, 0.839216],
[0.811765, 0.6, 0.835294],[0.811765, 0.6, 0.835294],[0.815686, 0.596078, 0.835294],[0.819608, 0.596078, 0.831373],
[0.819608, 0.596078, 0.831373],[0.823529, 0.596078, 0.827451],[0.827451, 0.592157, 0.827451],[0.827451, 0.592157, 0.823529],
[0.831373, 0.592157, 0.823529],[0.831373, 0.592157, 0.823529],[0.835294, 0.592157, 0.819608],[0.839216, 0.588235, 0.819608],
[0.839216, 0.588235, 0.815686],[0.843137, 0.588235, 0.815686],[0.843137, 0.588235, 0.811765],[0.847059, 0.584314, 0.811765],
[0.85098, 0.584314, 0.807843],[0.85098, 0.584314, 0.807843],[0.854902, 0.584314, 0.807843],[0.854902, 0.584314, 0.803922],
[0.858824, 0.580392, 0.803922],[0.862745, 0.580392, 0.8],[0.862745, 0.580392, 0.8],[0.866667, 0.580392, 0.796078],
[0.866667, 0.580392, 0.796078],[0.870588, 0.576471, 0.792157],[0.870588, 0.576471, 0.792157],[0.87451, 0.576471, 0.792157],
[0.878431, 0.576471, 0.788235],[0.878431, 0.572549, 0.788235],[0.882353, 0.572549, 0.784314],[0.882353, 0.572549, 0.784314],
[0.886275, 0.572549, 0.780392],[0.886275, 0.572549, 0.780392],[0.890196, 0.568627, 0.776471],[0.890196, 0.568627, 0.776471],
[0.894118, 0.568627, 0.772549],[0.898039, 0.568627, 0.772549],[0.898039, 0.564706, 0.768627],[0.901961, 0.564706, 0.768627],
[0.901961, 0.564706, 0.768627],[0.905882, 0.564706, 0.764706],[0.905882, 0.564706, 0.764706],[0.909804, 0.560784, 0.760784],
[0.909804, 0.560784, 0.760784],[0.913725, 0.560784, 0.756863],[0.913725, 0.560784, 0.756863],[0.917647, 0.556863, 0.752941],
[0.917647, 0.556863, 0.752941],[0.921569, 0.556863, 0.74902],[0.921569, 0.556863, 0.74902],[0.92549, 0.556863, 0.745098],
[0.92549, 0.552941, 0.745098],[0.929412, 0.552941, 0.745098],[0.929412, 0.552941, 0.741176],[0.933333, 0.552941, 0.741176],
[0.933333, 0.54902, 0.737255],[0.937255, 0.54902, 0.737255],[0.937255, 0.54902, 0.733333],[0.941176, 0.54902, 0.733333],
[0.941176, 0.545098, 0.729412],[0.945098, 0.545098, 0.729412],[0.945098, 0.545098, 0.72549],[0.94902, 0.545098, 0.72549],
[0.94902, 0.545098, 0.721569],[0.952941, 0.541176, 0.721569],[0.952941, 0.541176, 0.717647],[0.956863, 0.541176, 0.717647],
[0.956863, 0.541176, 0.713725],[0.960784, 0.537255, 0.713725],[0.960784, 0.537255, 0.709804],[0.964706, 0.537255, 0.709804],
[0.964706, 0.537255, 0.709804],[0.968627, 0.533333, 0.705882],[0.968627, 0.533333, 0.705882],[0.972549, 0.533333, 0.701961],
[0.972549, 0.533333, 0.701961],[0.976471, 0.533333, 0.698039],[0.976471, 0.529412, 0.698039],[0.976471, 0.529412, 0.694118],
[0.980392, 0.529412, 0.694118],[0.980392, 0.529412, 0.690196],[0.984314, 0.52549, 0.690196],[0.984314, 0.52549, 0.686275],
[0.988235, 0.52549, 0.686275],[0.988235, 0.52549, 0.682353],[0.992157, 0.521569, 0.682353],[0.992157, 0.521569, 0.678431],
[0.996078, 0.521569, 0.678431] ], [[0, 0.0470588, 0.490196],[0, 0.0509804, 0.494118],
[0, 0.0509804, 0.501961],[0, 0.054902, 0.509804],[0, 0.054902, 0.517647],[0, 0.0588235, 0.52549],
[0, 0.0588235, 0.533333],[0, 0.0627451, 0.537255],[0, 0.0627451, 0.545098],[0, 0.0666667, 0.552941],
[0, 0.0666667, 0.560784],[0, 0.0705882, 0.564706],[0, 0.0745098, 0.572549],[0, 0.0745098, 0.580392],
[0, 0.0784314, 0.588235],[0, 0.0823529, 0.592157],[0, 0.0823529, 0.6],[0, 0.0862745, 0.607843],
[0, 0.0901961, 0.611765],[0, 0.0901961, 0.619608],[0, 0.0941176, 0.623529],[0, 0.0980392, 0.631373],
[0, 0.0980392, 0.635294],[0, 0.101961, 0.643137],[0, 0.105882, 0.647059],[0, 0.109804, 0.654902],
[0, 0.113725, 0.658824],[0, 0.113725, 0.666667],[0, 0.117647, 0.670588],[0, 0.121569, 0.67451],
[0, 0.12549, 0.682353],[0, 0.129412, 0.686275],[0, 0.133333, 0.690196],[0, 0.137255, 0.694118],
[0, 0.141176, 0.701961],[0, 0.141176, 0.705882],[0, 0.145098, 0.709804],[0, 0.14902, 0.713725],
[0, 0.152941, 0.717647],[0, 0.160784, 0.717647],[0, 0.164706, 0.721569],[0, 0.168627, 0.72549],
[0, 0.172549, 0.729412],[0, 0.176471, 0.729412],[0, 0.180392, 0.733333],[0, 0.184314, 0.733333],
[0, 0.192157, 0.733333],[0, 0.196078, 0.733333],[0, 0.2, 0.737255],[0, 0.203922, 0.737255],
[0, 0.211765, 0.737255],[0, 0.215686, 0.737255],[0, 0.219608, 0.737255],[0, 0.223529, 0.737255],
[0, 0.227451, 0.737255],[0, 0.235294, 0.737255],[0, 0.239216, 0.733333],[0, 0.243137, 0.733333],
[0, 0.247059, 0.733333],[0, 0.25098, 0.733333],[0, 0.254902, 0.729412],[0, 0.262745, 0.729412],
[0, 0.266667, 0.72549],[0, 0.270588, 0.72549],[0, 0.27451, 0.721569],[0, 0.278431, 0.721569],
[0, 0.286275, 0.717647],[0, 0.290196, 0.713725],[0, 0.294118, 0.709804],[0, 0.298039, 0.709804],
[0, 0.301961, 0.705882],[0, 0.305882, 0.701961],[0, 0.313725, 0.698039],[0, 0.317647, 0.694118],
[0, 0.321569, 0.690196],[0, 0.32549, 0.686275],[0, 0.329412, 0.682353],[0, 0.337255, 0.678431],
[0, 0.341176, 0.67451],[0, 0.345098, 0.670588],[0, 0.34902, 0.662745],[0, 0.352941, 0.658824],
[0, 0.360784, 0.654902],[0, 0.364706, 0.65098],[0, 0.368627, 0.647059],[0, 0.372549, 0.643137],
[0, 0.376471, 0.635294],[0, 0.380392, 0.631373],[0, 0.384314, 0.627451],[0, 0.392157, 0.623529],
[0, 0.396078, 0.619608],[0, 0.4, 0.611765],[0, 0.403922, 0.607843],[0, 0.407843, 0.603922],
[0, 0.411765, 0.6],[0, 0.419608, 0.596078],[0, 0.423529, 0.588235],[0, 0.427451, 0.584314],
[0, 0.431373, 0.580392],[0, 0.435294, 0.572549],[0, 0.439216, 0.568627],[0, 0.443137, 0.564706],
[0, 0.447059, 0.560784],[0, 0.454902, 0.552941],[0, 0.458824, 0.54902],[0, 0.462745, 0.545098],
[0, 0.466667, 0.537255],[0, 0.470588, 0.533333],[0, 0.47451, 0.52549],[0, 0.478431, 0.521569],
[0, 0.482353, 0.513725],[0, 0.490196, 0.509804],[0, 0.494118, 0.501961],[0, 0.498039, 0.498039],
[0, 0.501961, 0.490196],[0, 0.505882, 0.482353],[0, 0.509804, 0.478431],[0, 0.513725, 0.470588],
[0, 0.517647, 0.462745],[0.00784314, 0.521569, 0.454902],[0.0235294, 0.52549, 0.447059],[0.0392157, 0.529412, 0.439216],
[0.0509804, 0.537255, 0.431373],[0.0627451, 0.541176, 0.423529],[0.0745098, 0.545098, 0.415686],[0.0823529, 0.54902, 0.407843],
[0.0901961, 0.552941, 0.4],[0.101961, 0.556863, 0.392157],[0.109804, 0.560784, 0.384314],[0.113725, 0.564706, 0.372549],
[0.121569, 0.568627, 0.364706],[0.129412, 0.572549, 0.356863],[0.133333, 0.576471, 0.345098],[0.141176, 0.580392, 0.333333],
[0.145098, 0.584314, 0.32549],[0.152941, 0.588235, 0.313725],[0.156863, 0.592157, 0.305882],[0.160784, 0.596078, 0.294118],
[0.164706, 0.6, 0.282353],[0.168627, 0.607843, 0.27451],[0.172549, 0.611765, 0.262745],[0.176471, 0.615686, 0.254902],
[0.176471, 0.619608, 0.243137],[0.180392, 0.623529, 0.235294],[0.180392, 0.627451, 0.227451],[0.184314, 0.631373, 0.215686],
[0.184314, 0.635294, 0.207843],[0.188235, 0.639216, 0.2],[0.188235, 0.643137, 0.192157],[0.188235, 0.647059, 0.184314],
[0.188235, 0.65098, 0.176471],[0.192157, 0.654902, 0.168627],[0.192157, 0.658824, 0.160784],[0.192157, 0.662745, 0.152941],
[0.192157, 0.666667, 0.145098],[0.192157, 0.670588, 0.141176],[0.192157, 0.67451, 0.133333],[0.192157, 0.678431, 0.12549],
[0.192157, 0.682353, 0.121569],[0.196078, 0.686275, 0.117647],[0.196078, 0.690196, 0.113725],[0.196078, 0.694118, 0.109804],
[0.196078, 0.701961, 0.105882],[0.196078, 0.705882, 0.101961],[0.196078, 0.709804, 0.101961],[0.196078, 0.713725, 0.101961],
[0.196078, 0.717647, 0.101961],[0.196078, 0.721569, 0.101961],[0.2, 0.72549, 0.101961],[0.2, 0.729412, 0.101961],
[0.2, 0.733333, 0.101961],[0.2, 0.737255, 0.101961],[0.2, 0.741176, 0.101961],[0.2, 0.745098, 0.101961],
[0.2, 0.74902, 0.101961],[0.2, 0.752941, 0.101961],[0.203922, 0.756863, 0.0980392],[0.203922, 0.760784, 0.0980392],
[0.203922, 0.764706, 0.0980392],[0.203922, 0.768627, 0.0980392],[0.203922, 0.772549, 0.0980392],[0.203922, 0.776471, 0.0980392],
[0.203922, 0.780392, 0.0980392],[0.207843, 0.784314, 0.0980392],[0.207843, 0.788235, 0.0980392],[0.207843, 0.792157, 0.0980392],
[0.207843, 0.796078, 0.0980392],[0.207843, 0.8, 0.0980392],[0.207843, 0.803922, 0.0980392],[0.211765, 0.807843, 0.0980392],
[0.211765, 0.811765, 0.0980392],[0.211765, 0.815686, 0.0980392],[0.211765, 0.823529, 0.0980392],[0.211765, 0.827451, 0.101961],
[0.211765, 0.831373, 0.101961],[0.215686, 0.835294, 0.101961],[0.215686, 0.839216, 0.101961],[0.223529, 0.843137, 0.101961],
[0.231373, 0.847059, 0.101961],[0.243137, 0.847059, 0.101961],[0.254902, 0.85098, 0.101961],[0.266667, 0.854902, 0.101961],
[0.282353, 0.858824, 0.101961],[0.294118, 0.862745, 0.101961],[0.309804, 0.866667, 0.101961],[0.32549, 0.866667, 0.101961],
[0.341176, 0.870588, 0.101961],[0.356863, 0.87451, 0.101961],[0.372549, 0.87451, 0.101961],[0.388235, 0.878431, 0.101961],
[0.403922, 0.882353, 0.105882],[0.419608, 0.882353, 0.105882],[0.435294, 0.886275, 0.105882],[0.45098, 0.890196, 0.105882],
[0.466667, 0.890196, 0.105882],[0.482353, 0.894118, 0.105882],[0.498039, 0.894118, 0.105882],[0.509804, 0.898039, 0.105882],
[0.52549, 0.901961, 0.105882],[0.541176, 0.901961, 0.105882],[0.556863, 0.905882, 0.109804],[0.572549, 0.905882, 0.109804],
[0.588235, 0.909804, 0.109804],[0.603922, 0.909804, 0.109804],[0.615686, 0.913725, 0.109804],[0.631373, 0.913725, 0.109804],
[0.647059, 0.917647, 0.109804],[0.662745, 0.917647, 0.113725],[0.67451, 0.917647, 0.113725],[0.690196, 0.921569, 0.113725],
[0.705882, 0.921569, 0.113725],[0.721569, 0.92549, 0.113725],[0.733333, 0.92549, 0.113725],[0.74902, 0.92549, 0.117647],
[0.764706, 0.929412, 0.117647],[0.776471, 0.929412, 0.117647],[0.792157, 0.929412, 0.117647],[0.807843, 0.933333, 0.117647],
[0.819608, 0.933333, 0.121569],[0.835294, 0.933333, 0.121569],[0.85098, 0.937255, 0.121569],[0.862745, 0.937255, 0.121569],
[0.878431, 0.937255, 0.12549],[0.894118, 0.937255, 0.12549],[0.905882, 0.941176, 0.12549],[0.921569, 0.941176, 0.12549],
[0.933333, 0.941176, 0.129412],[0.94902, 0.941176, 0.129412],[0.964706, 0.941176, 0.129412],[0.976471, 0.945098, 0.129412],
[0.992157, 0.945098, 0.133333],[1, 0.945098, 0.133333],[1, 0.945098, 0.133333],[1, 0.945098, 0.133333],
[1, 0.945098, 0.137255],[1, 0.945098, 0.137255] ], [[0.105882, 0, 0.517647],
[0.105882, 0.00392157, 0.52549],[0.105882, 0.00392157, 0.533333],[0.105882, 0.00784314, 0.541176],[0.105882, 0.00784314, 0.54902],
[0.105882, 0.0117647, 0.552941],[0.105882, 0.0156863, 0.560784],[0.105882, 0.0196078, 0.568627],[0.105882, 0.0196078, 0.576471],
[0.105882, 0.0235294, 0.580392],[0.105882, 0.027451, 0.588235],[0.105882, 0.0313725, 0.596078],[0.105882, 0.0352941, 0.603922],
[0.105882, 0.0392157, 0.607843],[0.105882, 0.0431373, 0.615686],[0.105882, 0.0470588, 0.623529],[0.105882, 0.0509804, 0.631373],
[0.105882, 0.054902, 0.635294],[0.105882, 0.0588235, 0.643137],[0.105882, 0.0627451, 0.65098],[0.105882, 0.0627451, 0.654902],
[0.105882, 0.0666667, 0.662745],[0.105882, 0.0705882, 0.670588],[0.105882, 0.0745098, 0.67451],[0.109804, 0.0784314, 0.682353],
[0.109804, 0.0823529, 0.686275],[0.109804, 0.0862745, 0.694118],[0.109804, 0.0901961, 0.701961],[0.109804, 0.0941176, 0.705882],
[0.109804, 0.0980392, 0.713725],[0.109804, 0.101961, 0.717647],[0.109804, 0.105882, 0.72549],[0.109804, 0.109804, 0.729412],
[0.109804, 0.113725, 0.737255],[0.109804, 0.117647, 0.741176],[0.109804, 0.121569, 0.745098],[0.109804, 0.12549, 0.752941],
[0.113725, 0.129412, 0.756863],[0.113725, 0.133333, 0.760784],[0.113725, 0.137255, 0.768627],[0.113725, 0.141176, 0.772549],
[0.113725, 0.14902, 0.776471],[0.113725, 0.152941, 0.784314],[0.117647, 0.156863, 0.788235],[0.117647, 0.160784, 0.792157],
[0.117647, 0.164706, 0.796078],[0.117647, 0.168627, 0.8],[0.121569, 0.172549, 0.803922],[0.121569, 0.176471, 0.807843],
[0.121569, 0.180392, 0.811765],[0.12549, 0.188235, 0.815686],[0.12549, 0.192157, 0.819608],[0.12549, 0.196078, 0.823529],
[0.129412, 0.2, 0.827451],[0.129412, 0.203922, 0.827451],[0.129412, 0.211765, 0.831373],[0.133333, 0.215686, 0.835294],
[0.133333, 0.219608, 0.835294],[0.137255, 0.227451, 0.839216],[0.137255, 0.231373, 0.839216],[0.137255, 0.235294, 0.839216],
[0.141176, 0.243137, 0.843137],[0.141176, 0.247059, 0.843137],[0.145098, 0.254902, 0.843137],[0.145098, 0.258824, 0.843137],
[0.14902, 0.262745, 0.839216],[0.14902, 0.270588, 0.839216],[0.152941, 0.278431, 0.839216],[0.152941, 0.282353, 0.835294],
[0.156863, 0.290196, 0.831373],[0.156863, 0.298039, 0.827451],[0.156863, 0.301961, 0.823529],[0.160784, 0.309804, 0.815686],
[0.160784, 0.317647, 0.807843],[0.160784, 0.32549, 0.8],[0.160784, 0.333333, 0.788235],[0.160784, 0.341176, 0.776471],
[0.164706, 0.352941, 0.764706],[0.164706, 0.356863, 0.752941],[0.164706, 0.364706, 0.741176],[0.168627, 0.372549, 0.729412],
[0.168627, 0.380392, 0.717647],[0.172549, 0.388235, 0.705882],[0.176471, 0.392157, 0.694118],[0.176471, 0.4, 0.682353],
[0.180392, 0.407843, 0.670588],[0.184314, 0.411765, 0.662745],[0.188235, 0.419608, 0.65098],[0.188235, 0.423529, 0.639216],
[0.192157, 0.431373, 0.627451],[0.196078, 0.435294, 0.615686],[0.2, 0.443137, 0.607843],[0.203922, 0.447059, 0.596078],
[0.203922, 0.45098, 0.584314],[0.207843, 0.458824, 0.576471],[0.211765, 0.462745, 0.564706],[0.211765, 0.466667, 0.552941],
[0.215686, 0.47451, 0.541176],[0.219608, 0.478431, 0.533333],[0.223529, 0.482353, 0.521569],[0.223529, 0.490196, 0.513725],
[0.227451, 0.494118, 0.501961],[0.231373, 0.498039, 0.490196],[0.231373, 0.501961, 0.482353],[0.235294, 0.509804, 0.470588],
[0.235294, 0.513725, 0.458824],[0.239216, 0.517647, 0.45098],[0.243137, 0.521569, 0.439216],[0.243137, 0.52549, 0.427451],
[0.247059, 0.529412, 0.419608],[0.247059, 0.537255, 0.407843],[0.25098, 0.541176, 0.4],[0.25098, 0.545098, 0.388235],
[0.254902, 0.54902, 0.376471],[0.254902, 0.552941, 0.368627],[0.258824, 0.556863, 0.356863],[0.258824, 0.560784, 0.34902],
[0.262745, 0.568627, 0.337255],[0.266667, 0.572549, 0.329412],[0.266667, 0.576471, 0.321569],[0.270588, 0.580392, 0.313725],
[0.27451, 0.584314, 0.305882],[0.278431, 0.588235, 0.298039],[0.282353, 0.592157, 0.290196],[0.282353, 0.596078, 0.282353],
[0.286275, 0.6, 0.27451],[0.290196, 0.603922, 0.266667],[0.294118, 0.607843, 0.258824],[0.301961, 0.611765, 0.25098],
[0.305882, 0.615686, 0.243137],[0.309804, 0.619608, 0.239216],[0.313725, 0.623529, 0.231373],[0.317647, 0.627451, 0.227451],
[0.321569, 0.627451, 0.219608],[0.329412, 0.631373, 0.211765],[0.333333, 0.635294, 0.207843],[0.341176, 0.639216, 0.2],
[0.345098, 0.643137, 0.196078],[0.34902, 0.647059, 0.188235],[0.356863, 0.65098, 0.184314],[0.360784, 0.654902, 0.180392],
[0.368627, 0.658824, 0.172549],[0.376471, 0.658824, 0.168627],[0.380392, 0.662745, 0.164706],[0.388235, 0.666667, 0.156863],
[0.392157, 0.670588, 0.152941],[0.4, 0.67451, 0.14902],[0.407843, 0.678431, 0.145098],[0.415686, 0.678431, 0.137255],
[0.419608, 0.682353, 0.133333],[0.427451, 0.686275, 0.129412],[0.435294, 0.690196, 0.12549],[0.443137, 0.694118, 0.121569],
[0.45098, 0.694118, 0.117647],[0.458824, 0.698039, 0.113725],[0.462745, 0.701961, 0.109804],[0.470588, 0.705882, 0.105882],
[0.478431, 0.705882, 0.105882],[0.486275, 0.709804, 0.101961],[0.494118, 0.713725, 0.0980392],[0.501961, 0.717647, 0.0980392],
[0.509804, 0.717647, 0.0941176],[0.517647, 0.721569, 0.0901961],[0.52549, 0.72549, 0.0901961],[0.533333, 0.729412, 0.0901961],
[0.545098, 0.729412, 0.0862745],[0.552941, 0.733333, 0.0862745],[0.560784, 0.737255, 0.0862745],[0.568627, 0.737255, 0.0862745],
[0.576471, 0.741176, 0.0862745],[0.584314, 0.745098, 0.0862745],[0.592157, 0.745098, 0.0862745],[0.603922, 0.74902, 0.0901961],
[0.611765, 0.752941, 0.0901961],[0.619608, 0.752941, 0.0901961],[0.627451, 0.756863, 0.0941176],[0.635294, 0.760784, 0.0941176],
[0.643137, 0.760784, 0.0980392],[0.65098, 0.764706, 0.0980392],[0.662745, 0.768627, 0.101961],[0.670588, 0.768627, 0.101961],
[0.678431, 0.772549, 0.105882],[0.686275, 0.776471, 0.109804],[0.694118, 0.776471, 0.113725],[0.701961, 0.780392, 0.113725],
[0.709804, 0.784314, 0.117647],[0.717647, 0.784314, 0.121569],[0.72549, 0.788235, 0.12549],[0.733333, 0.792157, 0.129412],
[0.741176, 0.792157, 0.129412],[0.74902, 0.796078, 0.133333],[0.752941, 0.8, 0.137255],[0.760784, 0.8, 0.141176],
[0.768627, 0.803922, 0.145098],[0.776471, 0.807843, 0.14902],[0.784314, 0.807843, 0.156863],[0.792157, 0.811765, 0.160784],
[0.796078, 0.815686, 0.164706],[0.803922, 0.815686, 0.168627],[0.811765, 0.819608, 0.172549],[0.819608, 0.823529, 0.180392],
[0.823529, 0.823529, 0.184314],[0.831373, 0.827451, 0.188235],[0.839216, 0.831373, 0.196078],[0.843137, 0.835294, 0.2],
[0.85098, 0.835294, 0.207843],[0.854902, 0.839216, 0.211765],[0.862745, 0.843137, 0.219608],[0.870588, 0.847059, 0.223529],
[0.87451, 0.847059, 0.231373],[0.878431, 0.85098, 0.239216],[0.886275, 0.854902, 0.247059],[0.890196, 0.858824, 0.25098],
[0.898039, 0.862745, 0.258824],[0.901961, 0.862745, 0.266667],[0.905882, 0.866667, 0.27451],[0.909804, 0.870588, 0.282353],
[0.913725, 0.87451, 0.290196],[0.917647, 0.878431, 0.301961],[0.921569, 0.882353, 0.309804],[0.92549, 0.886275, 0.321569],
[0.929412, 0.890196, 0.329412],[0.933333, 0.894118, 0.341176],[0.933333, 0.898039, 0.352941],[0.937255, 0.901961, 0.364706],
[0.937255, 0.905882, 0.376471],[0.937255, 0.909804, 0.388235],[0.941176, 0.913725, 0.403922],[0.941176, 0.917647, 0.415686],
[0.941176, 0.921569, 0.431373],[0.945098, 0.92549, 0.447059],[0.945098, 0.929412, 0.462745],[0.94902, 0.933333, 0.478431],
[0.94902, 0.937255, 0.494118],[0.952941, 0.941176, 0.513725],[0.952941, 0.945098, 0.529412],[0.956863, 0.94902, 0.54902],
[0.960784, 0.952941, 0.568627],[0.960784, 0.956863, 0.588235],[0.964706, 0.956863, 0.607843],[0.968627, 0.960784, 0.627451],
[0.968627, 0.964706, 0.647059],[0.972549, 0.968627, 0.670588],[0.976471, 0.968627, 0.690196],[0.976471, 0.972549, 0.713725],
[0.980392, 0.976471, 0.737255],[0.980392, 0.980392, 0.760784],[0.984314, 0.980392, 0.784314],[0.988235, 0.984314, 0.811765],
[0.988235, 0.988235, 0.835294],[0.992157, 0.988235, 0.862745],[0.992157, 0.992157, 0.886275],[0.996078, 0.992157, 0.913725],
[0.996078, 0.996078, 0.941176],[1, 0.996078, 0.972549],[1, 1, 1] ], [
[0.101961, 0, 0.52549],[0.101961, 0, 0.533333],[0.101961, 0, 0.537255],[0.101961, 0, 0.545098],
[0.101961, 0.00392157, 0.552941],[0.101961, 0.00392157, 0.560784],[0.101961, 0.00784314, 0.568627],[0.101961, 0.0117647, 0.572549],
[0.101961, 0.0156863, 0.580392],[0.101961, 0.0156863, 0.588235],[0.101961, 0.0196078, 0.596078],[0.105882, 0.0235294, 0.6],
[0.105882, 0.027451, 0.607843],[0.105882, 0.0313725, 0.615686],[0.105882, 0.0352941, 0.619608],[0.105882, 0.0392157, 0.627451],
[0.105882, 0.0431373, 0.635294],[0.105882, 0.0470588, 0.643137],[0.105882, 0.0509804, 0.647059],[0.105882, 0.054902, 0.654902],
[0.105882, 0.0588235, 0.658824],[0.105882, 0.0627451, 0.666667],[0.105882, 0.0666667, 0.67451],[0.105882, 0.0745098, 0.678431],
[0.105882, 0.0784314, 0.686275],[0.105882, 0.0823529, 0.690196],[0.105882, 0.0862745, 0.698039],[0.105882, 0.0901961, 0.705882],
[0.105882, 0.0941176, 0.709804],[0.105882, 0.0980392, 0.717647],[0.105882, 0.101961, 0.721569],[0.105882, 0.105882, 0.729412],
[0.105882, 0.109804, 0.733333],[0.105882, 0.113725, 0.737255],[0.105882, 0.117647, 0.745098],[0.105882, 0.121569, 0.74902],
[0.105882, 0.12549, 0.756863],[0.105882, 0.129412, 0.760784],[0.105882, 0.133333, 0.764706],[0.105882, 0.137255, 0.772549],
[0.105882, 0.141176, 0.776471],[0.105882, 0.145098, 0.780392],[0.105882, 0.152941, 0.784314],[0.105882, 0.156863, 0.788235],
[0.109804, 0.160784, 0.796078],[0.109804, 0.164706, 0.8],[0.109804, 0.168627, 0.803922],[0.109804, 0.172549, 0.807843],
[0.109804, 0.176471, 0.811765],[0.109804, 0.184314, 0.815686],[0.109804, 0.188235, 0.819608],[0.113725, 0.192157, 0.823529],
[0.113725, 0.196078, 0.823529],[0.113725, 0.203922, 0.827451],[0.113725, 0.207843, 0.831373],[0.113725, 0.211765, 0.835294],
[0.117647, 0.215686, 0.835294],[0.117647, 0.223529, 0.839216],[0.117647, 0.227451, 0.839216],[0.117647, 0.231373, 0.843137],
[0.121569, 0.239216, 0.843137],[0.121569, 0.243137, 0.843137],[0.121569, 0.25098, 0.843137],[0.121569, 0.254902, 0.843137],
[0.12549, 0.262745, 0.843137],[0.12549, 0.266667, 0.843137],[0.12549, 0.27451, 0.843137],[0.12549, 0.278431, 0.839216],
[0.129412, 0.286275, 0.839216],[0.129412, 0.294118, 0.835294],[0.129412, 0.298039, 0.831373],[0.129412, 0.305882, 0.827451],
[0.129412, 0.313725, 0.819608],[0.129412, 0.321569, 0.811765],[0.129412, 0.329412, 0.803922],[0.12549, 0.337255, 0.792157],
[0.12549, 0.345098, 0.780392],[0.12549, 0.352941, 0.768627],[0.12549, 0.360784, 0.756863],[0.12549, 0.368627, 0.745098],
[0.12549, 0.376471, 0.733333],[0.129412, 0.384314, 0.721569],[0.129412, 0.392157, 0.709804],[0.133333, 0.396078, 0.698039],
[0.137255, 0.403922, 0.686275],[0.141176, 0.411765, 0.678431],[0.145098, 0.415686, 0.666667],[0.14902, 0.423529, 0.654902],
[0.152941, 0.427451, 0.643137],[0.156863, 0.435294, 0.631373],[0.160784, 0.439216, 0.623529],[0.164706, 0.443137, 0.611765],
[0.168627, 0.45098, 0.6],[0.172549, 0.454902, 0.588235],[0.176471, 0.462745, 0.580392],[0.180392, 0.466667, 0.568627],
[0.184314, 0.470588, 0.556863],[0.184314, 0.47451, 0.54902],[0.188235, 0.482353, 0.537255],[0.192157, 0.486275, 0.52549],
[0.196078, 0.490196, 0.517647],[0.2, 0.494118, 0.505882],[0.203922, 0.501961, 0.494118],[0.207843, 0.505882, 0.486275],
[0.211765, 0.509804, 0.47451],[0.215686, 0.513725, 0.462745],[0.215686, 0.517647, 0.454902],[0.219608, 0.52549, 0.443137],
[0.223529, 0.529412, 0.431373],[0.227451, 0.533333, 0.423529],[0.231373, 0.537255, 0.411765],[0.231373, 0.541176, 0.4],
[0.235294, 0.545098, 0.392157],[0.239216, 0.54902, 0.380392],[0.243137, 0.556863, 0.368627],[0.243137, 0.560784, 0.360784],
[0.247059, 0.564706, 0.34902],[0.25098, 0.568627, 0.341176],[0.254902, 0.572549, 0.333333],[0.258824, 0.576471, 0.32549],
[0.262745, 0.580392, 0.313725],[0.266667, 0.584314, 0.305882],[0.266667, 0.588235, 0.298039],[0.270588, 0.592157, 0.290196],
[0.27451, 0.596078, 0.282353],[0.282353, 0.6, 0.27451],[0.286275, 0.603922, 0.266667],[0.290196, 0.607843, 0.258824],
[0.294118, 0.611765, 0.254902],[0.298039, 0.615686, 0.247059],[0.301961, 0.619608, 0.239216],[0.309804, 0.623529, 0.235294],
[0.313725, 0.627451, 0.227451],[0.317647, 0.631373, 0.219608],[0.32549, 0.631373, 0.215686],[0.329412, 0.635294, 0.207843],
[0.337255, 0.639216, 0.203922],[0.341176, 0.643137, 0.196078],[0.34902, 0.647059, 0.192157],[0.352941, 0.65098, 0.184314],
[0.360784, 0.654902, 0.180392],[0.364706, 0.658824, 0.172549],[0.372549, 0.658824, 0.168627],[0.380392, 0.662745, 0.164706],
[0.384314, 0.666667, 0.160784],[0.392157, 0.670588, 0.152941],[0.4, 0.67451, 0.14902],[0.403922, 0.678431, 0.145098],
[0.411765, 0.678431, 0.141176],[0.419608, 0.682353, 0.133333],[0.427451, 0.686275, 0.129412],[0.435294, 0.690196, 0.12549],
[0.443137, 0.694118, 0.121569],[0.45098, 0.694118, 0.117647],[0.454902, 0.698039, 0.113725],[0.462745, 0.701961, 0.109804],
[0.470588, 0.705882, 0.109804],[0.478431, 0.705882, 0.105882],[0.486275, 0.709804, 0.101961],[0.494118, 0.713725, 0.0980392],
[0.501961, 0.717647, 0.0980392],[0.509804, 0.717647, 0.0941176],[0.517647, 0.721569, 0.0901961],[0.52549, 0.72549, 0.0901961],
[0.533333, 0.729412, 0.0901961],[0.545098, 0.729412, 0.0862745],[0.552941, 0.733333, 0.0862745],[0.560784, 0.737255, 0.0862745],
[0.568627, 0.737255, 0.0862745],[0.576471, 0.741176, 0.0862745],[0.584314, 0.745098, 0.0862745],[0.592157, 0.745098, 0.0862745],
[0.603922, 0.74902, 0.0901961],[0.611765, 0.752941, 0.0901961],[0.619608, 0.752941, 0.0901961],[0.627451, 0.756863, 0.0941176],
[0.635294, 0.760784, 0.0941176],[0.643137, 0.760784, 0.0980392],[0.65098, 0.764706, 0.0980392],[0.662745, 0.768627, 0.101961],
[0.670588, 0.768627, 0.101961],[0.678431, 0.772549, 0.105882],[0.686275, 0.776471, 0.109804],[0.694118, 0.776471, 0.113725],
[0.701961, 0.780392, 0.113725],[0.709804, 0.784314, 0.117647],[0.717647, 0.784314, 0.121569],[0.72549, 0.788235, 0.12549],
[0.733333, 0.792157, 0.129412],[0.741176, 0.792157, 0.129412],[0.74902, 0.796078, 0.133333],[0.752941, 0.8, 0.137255],
[0.760784, 0.8, 0.141176],[0.768627, 0.803922, 0.145098],[0.776471, 0.807843, 0.14902],[0.784314, 0.807843, 0.156863],
[0.792157, 0.811765, 0.160784],[0.796078, 0.815686, 0.164706],[0.803922, 0.815686, 0.168627],[0.811765, 0.819608, 0.172549],
[0.819608, 0.823529, 0.180392],[0.823529, 0.823529, 0.184314],[0.831373, 0.827451, 0.188235],[0.839216, 0.831373, 0.196078],
[0.843137, 0.835294, 0.2],[0.85098, 0.835294, 0.207843],[0.854902, 0.839216, 0.211765],[0.862745, 0.843137, 0.219608],
[0.870588, 0.847059, 0.223529],[0.87451, 0.847059, 0.231373],[0.878431, 0.85098, 0.239216],[0.886275, 0.854902, 0.247059],
[0.890196, 0.858824, 0.25098],[0.898039, 0.862745, 0.258824],[0.901961, 0.862745, 0.266667],[0.905882, 0.866667, 0.27451],
[0.909804, 0.870588, 0.282353],[0.913725, 0.87451, 0.290196],[0.917647, 0.878431, 0.301961],[0.921569, 0.882353, 0.309804],
[0.92549, 0.886275, 0.321569],[0.929412, 0.890196, 0.329412],[0.933333, 0.894118, 0.341176],[0.933333, 0.898039, 0.352941],
[0.937255, 0.901961, 0.364706],[0.937255, 0.905882, 0.376471],[0.937255, 0.909804, 0.388235],[0.941176, 0.913725, 0.403922],
[0.941176, 0.917647, 0.415686],[0.941176, 0.921569, 0.431373],[0.945098, 0.92549, 0.447059],[0.945098, 0.929412, 0.462745],
[0.94902, 0.933333, 0.478431],[0.94902, 0.937255, 0.494118],[0.952941, 0.941176, 0.513725],[0.952941, 0.945098, 0.529412],
[0.956863, 0.94902, 0.54902],[0.960784, 0.952941, 0.568627],[0.960784, 0.956863, 0.588235],[0.964706, 0.956863, 0.607843],
[0.968627, 0.960784, 0.627451],[0.968627, 0.964706, 0.647059],[0.972549, 0.968627, 0.670588],[0.976471, 0.968627, 0.690196],
[0.976471, 0.972549, 0.713725],[0.980392, 0.976471, 0.737255],[0.980392, 0.980392, 0.760784],[0.984314, 0.980392, 0.784314],
[0.988235, 0.984314, 0.811765],[0.988235, 0.988235, 0.835294],[0.992157, 0.988235, 0.862745],[0.992157, 0.992157, 0.886275],
[0.996078, 0.992157, 0.913725],[0.996078, 0.996078, 0.941176],[1, 0.996078, 0.972549],[1, 1, 1] ], 
[[0, 0.00392157, 0.305882],[0, 0.00392157, 0.313725],[0, 0.00784314, 0.321569],
[0, 0.00784314, 0.329412],[0, 0.00784314, 0.337255],[0, 0.0117647, 0.345098],[0, 0.0117647, 0.352941],
[0, 0.0117647, 0.360784],[0, 0.0117647, 0.368627],[0, 0.0117647, 0.376471],[0.00392157, 0.0117647, 0.384314],
[0.00784314, 0.0117647, 0.392157],[0.00784314, 0.0117647, 0.4],[0.0117647, 0.0117647, 0.407843],[0.0156863, 0.00784314, 0.415686],
[0.0196078, 0.00784314, 0.423529],[0.0196078, 0.00784314, 0.431373],[0.0235294, 0.00784314, 0.439216],[0.027451, 0.00784314, 0.447059],
[0.0313725, 0.00784314, 0.454902],[0.0313725, 0.00784314, 0.462745],[0.0352941, 0.00784314, 0.470588],[0.0392157, 0.00784314, 0.482353],
[0.0431373, 0.00784314, 0.490196],[0.0431373, 0.00784314, 0.498039],[0.0470588, 0.00784314, 0.505882],[0.0509804, 0.00784314, 0.513725],
[0.0509804, 0.00784314, 0.521569],[0.054902, 0.00784314, 0.533333],[0.054902, 0.00392157, 0.541176],[0.0588235, 0.00392157, 0.54902],
[0.0588235, 0.00392157, 0.556863],[0.0588235, 0.00392157, 0.564706],[0.0627451, 0.00392157, 0.576471],[0.0627451, 0.00392157, 0.584314],
[0.0627451, 0.00392157, 0.592157],[0.0627451, 0.00392157, 0.6],[0.0627451, 0.00784314, 0.611765],[0.0627451, 0.00784314, 0.619608],
[0.0627451, 0.00784314, 0.627451],[0.0627451, 0.00784314, 0.635294],[0.0588235, 0.00784314, 0.647059],[0.0588235, 0.00784314, 0.654902],
[0.054902, 0.0117647, 0.666667],[0.0509804, 0.0117647, 0.67451],[0.0509804, 0.0117647, 0.682353],[0.0470588, 0.0156863, 0.694118],
[0.0431373, 0.0156863, 0.701961],[0.0392157, 0.0156863, 0.709804],[0.0392157, 0.0196078, 0.717647],[0.0352941, 0.0196078, 0.72549],
[0.0352941, 0.0235294, 0.737255],[0.0313725, 0.027451, 0.745098],[0.0313725, 0.027451, 0.752941],[0.0313725, 0.0313725, 0.760784],
[0.0313725, 0.0352941, 0.768627],[0.0313725, 0.0392157, 0.776471],[0.0313725, 0.0392157, 0.784314],[0.0313725, 0.0431373, 0.792157],
[0.0313725, 0.0470588, 0.8],[0.0352941, 0.0509804, 0.807843],[0.0352941, 0.054902, 0.815686],[0.0392157, 0.0588235, 0.823529],
[0.0431373, 0.0627451, 0.831373],[0.0470588, 0.0666667, 0.839216],[0.0509804, 0.0705882, 0.847059],[0.054902, 0.0705882, 0.85098],
[0.0627451, 0.0745098, 0.858824],[0.0666667, 0.0784314, 0.866667],[0.0705882, 0.0823529, 0.87451],[0.0745098, 0.0862745, 0.878431],
[0.0823529, 0.0901961, 0.886275],[0.0862745, 0.0941176, 0.890196],[0.0941176, 0.0980392, 0.898039],[0.0980392, 0.101961, 0.905882],
[0.101961, 0.109804, 0.909804],[0.105882, 0.113725, 0.913725],[0.109804, 0.117647, 0.921569],[0.113725, 0.121569, 0.92549],
[0.121569, 0.129412, 0.929412],[0.12549, 0.133333, 0.933333],[0.129412, 0.141176, 0.941176],[0.133333, 0.145098, 0.945098],
[0.137255, 0.14902, 0.94902],[0.137255, 0.156863, 0.952941],[0.141176, 0.160784, 0.956863],[0.145098, 0.168627, 0.956863],
[0.14902, 0.176471, 0.960784],[0.152941, 0.180392, 0.964706],[0.156863, 0.188235, 0.968627],[0.156863, 0.192157, 0.968627],
[0.160784, 0.2, 0.972549],[0.164706, 0.207843, 0.972549],[0.164706, 0.211765, 0.976471],[0.168627, 0.219608, 0.976471],
[0.168627, 0.227451, 0.980392],[0.172549, 0.231373, 0.980392],[0.172549, 0.239216, 0.980392],[0.176471, 0.247059, 0.980392],
[0.176471, 0.254902, 0.984314],[0.176471, 0.258824, 0.984314],[0.180392, 0.266667, 0.984314],[0.180392, 0.27451, 0.984314],
[0.180392, 0.278431, 0.984314],[0.184314, 0.286275, 0.984314],[0.184314, 0.294118, 0.984314],[0.184314, 0.298039, 0.984314],
[0.184314, 0.305882, 0.984314],[0.184314, 0.309804, 0.984314],[0.184314, 0.317647, 0.984314],[0.184314, 0.321569, 0.988235],
[0.188235, 0.329412, 0.988235],[0.188235, 0.337255, 0.988235],[0.188235, 0.341176, 0.988235],[0.188235, 0.34902, 0.988235],
[0.188235, 0.352941, 0.988235],[0.188235, 0.360784, 0.988235],[0.188235, 0.364706, 0.988235],[0.184314, 0.368627, 0.988235],
[0.184314, 0.376471, 0.988235],[0.184314, 0.380392, 0.988235],[0.184314, 0.388235, 0.992157],[0.184314, 0.392157, 0.992157],
[0.184314, 0.4, 0.992157],[0.180392, 0.403922, 0.992157],[0.180392, 0.411765, 0.992157],[0.180392, 0.415686, 0.992157],
[0.180392, 0.419608, 0.992157],[0.176471, 0.427451, 0.992157],[0.176471, 0.431373, 0.992157],[0.176471, 0.439216, 0.992157],
[0.172549, 0.443137, 0.992157],[0.172549, 0.447059, 0.992157],[0.172549, 0.454902, 0.992157],[0.172549, 0.458824, 0.996078],
[0.172549, 0.462745, 0.996078],[0.168627, 0.470588, 0.996078],[0.168627, 0.47451, 0.996078],[0.168627, 0.482353, 0.996078],
[0.168627, 0.486275, 0.996078],[0.168627, 0.490196, 0.996078],[0.172549, 0.494118, 0.996078],[0.172549, 0.501961, 0.996078],
[0.172549, 0.505882, 0.996078],[0.172549, 0.509804, 0.996078],[0.172549, 0.517647, 0.996078],[0.176471, 0.521569, 0.996078],
[0.176471, 0.52549, 0.996078],[0.176471, 0.529412, 0.996078],[0.180392, 0.537255, 0.996078],[0.180392, 0.541176, 0.996078],
[0.184314, 0.545098, 0.996078],[0.184314, 0.552941, 0.996078],[0.188235, 0.556863, 0.996078],[0.188235, 0.560784, 0.996078],
[0.192157, 0.564706, 0.992157],[0.196078, 0.572549, 0.992157],[0.196078, 0.576471, 0.992157],[0.2, 0.580392, 0.992157],
[0.2, 0.584314, 0.992157],[0.203922, 0.588235, 0.992157],[0.203922, 0.596078, 0.992157],[0.207843, 0.6, 0.992157],
[0.207843, 0.603922, 0.992157],[0.207843, 0.607843, 0.992157],[0.207843, 0.615686, 0.992157],[0.211765, 0.619608, 0.992157],
[0.211765, 0.623529, 0.992157],[0.211765, 0.627451, 0.992157],[0.211765, 0.635294, 0.992157],[0.211765, 0.639216, 0.992157],
[0.211765, 0.643137, 0.992157],[0.211765, 0.65098, 0.992157],[0.211765, 0.654902, 0.992157],[0.207843, 0.658824, 0.992157],
[0.207843, 0.662745, 0.988235],[0.207843, 0.670588, 0.988235],[0.207843, 0.67451, 0.988235],[0.203922, 0.678431, 0.988235],
[0.203922, 0.682353, 0.988235],[0.2, 0.690196, 0.988235],[0.2, 0.694118, 0.988235],[0.196078, 0.698039, 0.988235],
[0.192157, 0.705882, 0.988235],[0.192157, 0.709804, 0.988235],[0.188235, 0.713725, 0.988235],[0.184314, 0.717647, 0.988235],
[0.180392, 0.72549, 0.988235],[0.180392, 0.729412, 0.988235],[0.176471, 0.733333, 0.988235],[0.172549, 0.737255, 0.988235],
[0.172549, 0.745098, 0.988235],[0.168627, 0.74902, 0.988235],[0.168627, 0.752941, 0.988235],[0.164706, 0.760784, 0.988235],
[0.164706, 0.764706, 0.988235],[0.164706, 0.768627, 0.988235],[0.160784, 0.772549, 0.988235],[0.160784, 0.780392, 0.988235],
[0.160784, 0.784314, 0.988235],[0.160784, 0.788235, 0.984314],[0.160784, 0.792157, 0.984314],[0.160784, 0.796078, 0.984314],
[0.160784, 0.803922, 0.984314],[0.160784, 0.807843, 0.984314],[0.160784, 0.811765, 0.984314],[0.160784, 0.815686, 0.984314],
[0.164706, 0.823529, 0.984314],[0.164706, 0.827451, 0.984314],[0.168627, 0.831373, 0.984314],[0.168627, 0.835294, 0.984314],
[0.172549, 0.839216, 0.984314],[0.172549, 0.847059, 0.980392],[0.176471, 0.85098, 0.980392],[0.184314, 0.854902, 0.980392],
[0.188235, 0.858824, 0.980392],[0.2, 0.862745, 0.980392],[0.207843, 0.866667, 0.980392],[0.219608, 0.870588, 0.980392],
[0.235294, 0.87451, 0.980392],[0.247059, 0.882353, 0.980392],[0.258824, 0.886275, 0.976471],[0.27451, 0.890196, 0.976471],
[0.286275, 0.894118, 0.976471],[0.301961, 0.898039, 0.976471],[0.313725, 0.898039, 0.976471],[0.329412, 0.901961, 0.976471],
[0.341176, 0.905882, 0.976471],[0.356863, 0.909804, 0.976471],[0.368627, 0.913725, 0.976471],[0.384314, 0.917647, 0.972549],
[0.396078, 0.921569, 0.972549],[0.411765, 0.92549, 0.972549],[0.423529, 0.929412, 0.972549],[0.439216, 0.933333, 0.972549],
[0.45098, 0.937255, 0.972549],[0.462745, 0.941176, 0.972549],[0.478431, 0.941176, 0.972549],[0.490196, 0.945098, 0.972549],
[0.501961, 0.94902, 0.968627],[0.517647, 0.952941, 0.968627],[0.529412, 0.956863, 0.968627],[0.541176, 0.960784, 0.968627],
[0.556863, 0.960784, 0.968627],[0.568627, 0.964706, 0.968627],[0.580392, 0.968627, 0.968627],[0.592157, 0.972549, 0.968627],
[0.603922, 0.976471, 0.964706],[0.615686, 0.976471, 0.964706],[0.631373, 0.980392, 0.964706],[0.643137, 0.984314, 0.964706],
[0.654902, 0.988235, 0.964706],[0.666667, 0.992157, 0.964706],[0.678431, 0.992157, 0.964706],[0.690196, 0.996078, 0.964706],
[0.701961, 1, 0.964706] ], [[0.945098, 0.945098, 0.945098],[0.941176, 0.941176, 0.945098],
[0.937255, 0.941176, 0.941176],[0.933333, 0.937255, 0.941176],[0.933333, 0.937255, 0.941176],[0.929412, 0.933333, 0.941176],
[0.92549, 0.933333, 0.941176],[0.921569, 0.929412, 0.941176],[0.917647, 0.929412, 0.941176],[0.917647, 0.92549, 0.941176],
[0.913725, 0.92549, 0.941176],[0.909804, 0.921569, 0.941176],[0.905882, 0.921569, 0.941176],[0.905882, 0.917647, 0.941176],
[0.901961, 0.917647, 0.941176],[0.898039, 0.913725, 0.941176],[0.894118, 0.913725, 0.941176],[0.894118, 0.913725, 0.937255],
[0.890196, 0.909804, 0.937255],[0.886275, 0.909804, 0.937255],[0.882353, 0.905882, 0.937255],[0.882353, 0.905882, 0.937255],
[0.878431, 0.901961, 0.937255],[0.87451, 0.901961, 0.937255],[0.870588, 0.898039, 0.937255],[0.870588, 0.898039, 0.937255],
[0.866667, 0.894118, 0.937255],[0.862745, 0.894118, 0.937255],[0.858824, 0.890196, 0.937255],[0.854902, 0.890196, 0.937255],
[0.854902, 0.886275, 0.933333],[0.85098, 0.886275, 0.933333],[0.847059, 0.882353, 0.933333],[0.843137, 0.882353, 0.933333],
[0.843137, 0.882353, 0.933333],[0.839216, 0.878431, 0.933333],[0.835294, 0.878431, 0.933333],[0.831373, 0.87451, 0.933333],
[0.831373, 0.87451, 0.933333],[0.827451, 0.870588, 0.933333],[0.823529, 0.870588, 0.933333],[0.819608, 0.866667, 0.933333],
[0.819608, 0.866667, 0.929412],[0.815686, 0.862745, 0.929412],[0.811765, 0.862745, 0.929412],[0.807843, 0.858824, 0.929412],
[0.803922, 0.858824, 0.929412],[0.803922, 0.854902, 0.929412],[0.8, 0.854902, 0.929412],[0.796078, 0.85098, 0.929412],
[0.792157, 0.85098, 0.929412],[0.792157, 0.85098, 0.929412],[0.788235, 0.847059, 0.92549],[0.784314, 0.847059, 0.92549],
[0.780392, 0.843137, 0.92549],[0.780392, 0.843137, 0.92549],[0.776471, 0.839216, 0.92549],[0.772549, 0.839216, 0.92549],
[0.768627, 0.835294, 0.92549],[0.768627, 0.835294, 0.92549],[0.764706, 0.831373, 0.92549],[0.760784, 0.831373, 0.921569],
[0.756863, 0.827451, 0.921569],[0.752941, 0.827451, 0.921569],[0.752941, 0.823529, 0.921569],[0.74902, 0.823529, 0.921569],
[0.745098, 0.823529, 0.921569],[0.741176, 0.819608, 0.921569],[0.741176, 0.819608, 0.921569],[0.737255, 0.815686, 0.917647],
[0.733333, 0.815686, 0.917647],[0.729412, 0.811765, 0.917647],[0.729412, 0.811765, 0.917647],[0.72549, 0.807843, 0.917647],
[0.721569, 0.807843, 0.917647],[0.717647, 0.803922, 0.917647],[0.717647, 0.803922, 0.913725],[0.713725, 0.8, 0.913725],
[0.709804, 0.8, 0.913725],[0.705882, 0.8, 0.913725],[0.705882, 0.796078, 0.913725],[0.701961, 0.796078, 0.913725],
[0.698039, 0.792157, 0.913725],[0.694118, 0.792157, 0.909804],[0.694118, 0.788235, 0.909804],[0.690196, 0.788235, 0.909804],
[0.686275, 0.784314, 0.909804],[0.682353, 0.784314, 0.909804],[0.682353, 0.780392, 0.905882],[0.678431, 0.780392, 0.905882],
[0.67451, 0.776471, 0.905882],[0.670588, 0.776471, 0.905882],[0.670588, 0.772549, 0.905882],[0.666667, 0.772549, 0.905882],
[0.662745, 0.772549, 0.901961],[0.658824, 0.768627, 0.901961],[0.658824, 0.768627, 0.901961],[0.654902, 0.764706, 0.901961],
[0.65098, 0.764706, 0.901961],[0.65098, 0.760784, 0.898039],[0.647059, 0.760784, 0.898039],[0.643137, 0.756863, 0.898039],
[0.639216, 0.756863, 0.898039],[0.639216, 0.752941, 0.894118],[0.635294, 0.752941, 0.894118],[0.631373, 0.74902, 0.894118],
[0.631373, 0.74902, 0.894118],[0.627451, 0.74902, 0.890196],[0.623529, 0.745098, 0.890196],[0.623529, 0.745098, 0.890196],
[0.619608, 0.741176, 0.886275],[0.615686, 0.741176, 0.886275],[0.615686, 0.737255, 0.886275],[0.611765, 0.737255, 0.886275],
[0.607843, 0.733333, 0.882353],[0.607843, 0.733333, 0.882353],[0.603922, 0.729412, 0.882353],[0.6, 0.729412, 0.878431],
[0.6, 0.72549, 0.878431],[0.596078, 0.72549, 0.87451],[0.596078, 0.721569, 0.87451],[0.592157, 0.721569, 0.87451],
[0.588235, 0.717647, 0.870588],[0.588235, 0.717647, 0.870588],[0.584314, 0.717647, 0.870588],[0.584314, 0.713725, 0.866667],
[0.580392, 0.713725, 0.866667],[0.576471, 0.709804, 0.866667],[0.576471, 0.709804, 0.862745],[0.572549, 0.705882, 0.862745],
[0.572549, 0.705882, 0.858824],[0.568627, 0.701961, 0.858824],[0.568627, 0.701961, 0.858824],[0.564706, 0.698039, 0.854902],
[0.560784, 0.698039, 0.854902],[0.560784, 0.694118, 0.85098],[0.556863, 0.694118, 0.85098],[0.556863, 0.690196, 0.85098],
[0.552941, 0.690196, 0.847059],[0.552941, 0.686275, 0.847059],[0.54902, 0.686275, 0.843137],[0.545098, 0.686275, 0.843137],
[0.545098, 0.682353, 0.843137],[0.541176, 0.682353, 0.839216],[0.541176, 0.678431, 0.839216],[0.537255, 0.678431, 0.835294],
[0.537255, 0.67451, 0.835294],[0.533333, 0.67451, 0.831373],[0.533333, 0.670588, 0.831373],[0.529412, 0.670588, 0.831373],
[0.529412, 0.666667, 0.827451],[0.52549, 0.666667, 0.827451],[0.521569, 0.662745, 0.823529],[0.521569, 0.662745, 0.823529],
[0.517647, 0.658824, 0.819608],[0.517647, 0.658824, 0.819608],[0.513725, 0.658824, 0.815686],[0.513725, 0.654902, 0.815686],
[0.509804, 0.654902, 0.815686],[0.509804, 0.65098, 0.811765],[0.505882, 0.65098, 0.811765],[0.505882, 0.647059, 0.807843],
[0.501961, 0.647059, 0.807843],[0.501961, 0.643137, 0.803922],[0.498039, 0.643137, 0.803922],[0.498039, 0.639216, 0.8],
[0.494118, 0.639216, 0.8],[0.494118, 0.635294, 0.796078],[0.490196, 0.635294, 0.796078],[0.490196, 0.631373, 0.792157],
[0.486275, 0.631373, 0.792157],[0.486275, 0.627451, 0.788235],[0.486275, 0.627451, 0.788235],[0.482353, 0.627451, 0.784314],
[0.482353, 0.623529, 0.784314],[0.478431, 0.623529, 0.780392],[0.47451, 0.619608, 0.780392],[0.47451, 0.619608, 0.780392],
[0.470588, 0.615686, 0.776471],[0.470588, 0.615686, 0.776471],[0.466667, 0.611765, 0.772549],[0.466667, 0.611765, 0.772549],
[0.462745, 0.607843, 0.772549],[0.462745, 0.607843, 0.768627],[0.458824, 0.603922, 0.768627],[0.454902, 0.603922, 0.764706],
[0.454902, 0.603922, 0.764706],[0.45098, 0.6, 0.764706],[0.45098, 0.6, 0.760784],[0.447059, 0.596078, 0.760784],
[0.443137, 0.596078, 0.760784],[0.443137, 0.592157, 0.756863],[0.439216, 0.592157, 0.756863],[0.435294, 0.588235, 0.756863],
[0.435294, 0.588235, 0.752941],[0.431373, 0.588235, 0.752941],[0.427451, 0.584314, 0.752941],[0.427451, 0.584314, 0.74902],
[0.423529, 0.580392, 0.74902],[0.419608, 0.580392, 0.74902],[0.419608, 0.576471, 0.745098],[0.415686, 0.576471, 0.745098],
[0.411765, 0.572549, 0.745098],[0.411765, 0.572549, 0.745098],[0.407843, 0.568627, 0.741176],[0.403922, 0.568627, 0.741176],
[0.403922, 0.568627, 0.741176],[0.4, 0.564706, 0.737255],[0.396078, 0.564706, 0.737255],[0.392157, 0.560784, 0.737255],
[0.392157, 0.560784, 0.737255],[0.388235, 0.556863, 0.733333],[0.384314, 0.556863, 0.733333],[0.380392, 0.556863, 0.733333],
[0.380392, 0.552941, 0.733333],[0.376471, 0.552941, 0.729412],[0.372549, 0.54902, 0.729412],[0.368627, 0.54902, 0.729412],
[0.368627, 0.545098, 0.729412],[0.364706, 0.545098, 0.72549],[0.360784, 0.541176, 0.72549],[0.356863, 0.541176, 0.72549],
[0.352941, 0.541176, 0.72549],[0.352941, 0.537255, 0.721569],[0.34902, 0.537255, 0.721569],[0.345098, 0.533333, 0.721569],
[0.341176, 0.533333, 0.721569],[0.337255, 0.529412, 0.721569],[0.333333, 0.529412, 0.717647],[0.329412, 0.529412, 0.717647],
[0.329412, 0.52549, 0.717647],[0.32549, 0.52549, 0.717647],[0.321569, 0.521569, 0.713725],[0.317647, 0.521569, 0.713725],
[0.313725, 0.517647, 0.713725],[0.309804, 0.517647, 0.713725],[0.305882, 0.517647, 0.713725],[0.301961, 0.513725, 0.709804],
[0.298039, 0.513725, 0.709804],[0.294118, 0.509804, 0.709804],[0.290196, 0.509804, 0.709804],[0.290196, 0.505882, 0.709804],
[0.286275, 0.505882, 0.705882],[0.282353, 0.505882, 0.705882],[0.278431, 0.501961, 0.705882],[0.27451, 0.501961, 0.705882],
[0.270588, 0.498039, 0.705882],[0.262745, 0.498039, 0.701961],[0.258824, 0.494118, 0.701961],[0.254902, 0.494118, 0.701961],
[0.25098, 0.494118, 0.701961],[0.247059, 0.490196, 0.701961],[0.243137, 0.490196, 0.698039],[0.239216, 0.486275, 0.698039],
[0.235294, 0.486275, 0.698039],[0.231373, 0.486275, 0.698039] ], [[0, 0.00784314, 0.294118],
[0, 0.0117647, 0.301961],[0, 0.0117647, 0.309804],[0, 0.0156863, 0.317647],[0, 0.0156863, 0.32549],
[0, 0.0156863, 0.333333],[0, 0.0196078, 0.345098],[0, 0.0196078, 0.352941],[0, 0.0196078, 0.360784],
[0, 0.0196078, 0.368627],[0, 0.0196078, 0.376471],[0, 0.0196078, 0.384314],[0, 0.0196078, 0.392157],
[0, 0.0196078, 0.4],[0, 0.0196078, 0.407843],[0.00392157, 0.0196078, 0.415686],[0.00392157, 0.0196078, 0.427451],
[0.00392157, 0.0196078, 0.435294],[0.00392157, 0.0196078, 0.443137],[0.00784314, 0.0196078, 0.45098],[0.00784314, 0.0196078, 0.458824],
[0.00784314, 0.0196078, 0.466667],[0.0117647, 0.0196078, 0.47451],[0.0117647, 0.0196078, 0.486275],[0.0117647, 0.0196078, 0.494118],
[0.0156863, 0.0196078, 0.501961],[0.0156863, 0.0196078, 0.509804],[0.0156863, 0.0196078, 0.517647],[0.0156863, 0.0235294, 0.529412],
[0.0196078, 0.0235294, 0.537255],[0.0196078, 0.0235294, 0.545098],[0.0196078, 0.0235294, 0.552941],[0.0196078, 0.0235294, 0.564706],
[0.0235294, 0.0235294, 0.572549],[0.0235294, 0.0235294, 0.580392],[0.0235294, 0.0235294, 0.588235],[0.0235294, 0.0235294, 0.6],
[0.0235294, 0.0235294, 0.607843],[0.0235294, 0.0235294, 0.615686],[0.0235294, 0.0235294, 0.623529],[0.0235294, 0.0235294, 0.635294],
[0.0235294, 0.027451, 0.643137],[0.0196078, 0.027451, 0.65098],[0.0196078, 0.027451, 0.662745],[0.0196078, 0.027451, 0.670588],
[0.0196078, 0.027451, 0.678431],[0.0196078, 0.0313725, 0.686275],[0.0196078, 0.0313725, 0.698039],[0.0196078, 0.0313725, 0.705882],
[0.0196078, 0.0313725, 0.713725],[0.0196078, 0.0352941, 0.721569],[0.0235294, 0.0352941, 0.729412],[0.0235294, 0.0392157, 0.741176],
[0.027451, 0.0392157, 0.74902],[0.0313725, 0.0392157, 0.756863],[0.0352941, 0.0431373, 0.764706],[0.0431373, 0.0431373, 0.772549],
[0.0470588, 0.0470588, 0.780392],[0.0509804, 0.0470588, 0.788235],[0.0588235, 0.0509804, 0.796078],[0.0627451, 0.0509804, 0.803922],
[0.0705882, 0.054902, 0.811765],[0.0745098, 0.054902, 0.819608],[0.0823529, 0.0588235, 0.827451],[0.0862745, 0.0588235, 0.835294],
[0.0941176, 0.0627451, 0.839216],[0.0980392, 0.0666667, 0.847059],[0.105882, 0.0666667, 0.854902],[0.113725, 0.0705882, 0.862745],
[0.117647, 0.0705882, 0.870588],[0.12549, 0.0745098, 0.87451],[0.133333, 0.0784314, 0.882353],[0.141176, 0.0784314, 0.890196],
[0.14902, 0.0823529, 0.894118],[0.156863, 0.0823529, 0.901961],[0.164706, 0.0862745, 0.905882],[0.172549, 0.0862745, 0.913725],
[0.184314, 0.0901961, 0.917647],[0.192157, 0.0901961, 0.92549],[0.203922, 0.0941176, 0.929412],[0.215686, 0.0941176, 0.933333],
[0.223529, 0.0980392, 0.937255],[0.235294, 0.0980392, 0.941176],[0.247059, 0.0980392, 0.94902],[0.258824, 0.101961, 0.952941],
[0.270588, 0.101961, 0.956863],[0.282353, 0.101961, 0.960784],[0.294118, 0.105882, 0.960784],[0.301961, 0.105882, 0.964706],
[0.313725, 0.105882, 0.968627],[0.32549, 0.105882, 0.972549],[0.337255, 0.105882, 0.972549],[0.34902, 0.109804, 0.976471],
[0.364706, 0.109804, 0.980392],[0.376471, 0.109804, 0.980392],[0.388235, 0.109804, 0.980392],[0.4, 0.109804, 0.984314],
[0.411765, 0.109804, 0.984314],[0.423529, 0.109804, 0.984314],[0.435294, 0.109804, 0.988235],[0.447059, 0.109804, 0.988235],
[0.458824, 0.109804, 0.988235],[0.470588, 0.109804, 0.988235],[0.482353, 0.109804, 0.988235],[0.494118, 0.109804, 0.988235],
[0.501961, 0.109804, 0.992157],[0.513725, 0.109804, 0.992157],[0.52549, 0.109804, 0.992157],[0.537255, 0.109804, 0.992157],
[0.545098, 0.109804, 0.992157],[0.556863, 0.109804, 0.992157],[0.568627, 0.109804, 0.992157],[0.576471, 0.113725, 0.992157],
[0.588235, 0.113725, 0.996078],[0.6, 0.113725, 0.996078],[0.607843, 0.113725, 0.996078],[0.619608, 0.113725, 0.996078],
[0.627451, 0.113725, 0.996078],[0.639216, 0.113725, 0.996078],[0.647059, 0.113725, 0.996078],[0.658824, 0.113725, 0.996078],
[0.666667, 0.113725, 0.996078],[0.678431, 0.113725, 0.996078],[0.686275, 0.113725, 0.996078],[0.694118, 0.113725, 0.996078],
[0.705882, 0.113725, 0.996078],[0.713725, 0.109804, 0.996078],[0.72549, 0.109804, 0.996078],[0.733333, 0.109804, 0.996078],
[0.741176, 0.113725, 0.996078],[0.752941, 0.113725, 0.996078],[0.760784, 0.113725, 0.996078],[0.768627, 0.113725, 0.996078],
[0.776471, 0.117647, 0.996078],[0.784314, 0.117647, 0.996078],[0.792157, 0.121569, 0.996078],[0.8, 0.12549, 0.996078],
[0.807843, 0.129412, 0.996078],[0.815686, 0.133333, 0.996078],[0.823529, 0.137255, 0.996078],[0.831373, 0.141176, 0.996078],
[0.839216, 0.145098, 0.996078],[0.843137, 0.14902, 0.996078],[0.85098, 0.152941, 0.996078],[0.858824, 0.160784, 0.996078],
[0.862745, 0.164706, 0.996078],[0.870588, 0.172549, 0.996078],[0.878431, 0.176471, 0.996078],[0.882353, 0.180392, 0.996078],
[0.890196, 0.188235, 0.996078],[0.894118, 0.196078, 0.996078],[0.898039, 0.2, 0.996078],[0.905882, 0.207843, 0.996078],
[0.909804, 0.215686, 0.996078],[0.913725, 0.219608, 0.996078],[0.921569, 0.227451, 0.996078],[0.92549, 0.235294, 0.996078],
[0.929412, 0.243137, 0.996078],[0.933333, 0.247059, 0.996078],[0.937255, 0.254902, 0.996078],[0.941176, 0.262745, 0.996078],
[0.945098, 0.270588, 0.996078],[0.94902, 0.278431, 0.996078],[0.952941, 0.286275, 0.996078],[0.956863, 0.294118, 0.996078],
[0.960784, 0.301961, 0.996078],[0.964706, 0.309804, 0.996078],[0.968627, 0.317647, 0.996078],[0.972549, 0.32549, 0.996078],
[0.972549, 0.333333, 0.996078],[0.976471, 0.341176, 0.996078],[0.980392, 0.34902, 0.996078],[0.980392, 0.356863, 0.996078],
[0.984314, 0.364706, 0.996078],[0.984314, 0.372549, 0.996078],[0.988235, 0.380392, 0.996078],[0.988235, 0.388235, 0.996078],
[0.988235, 0.396078, 0.996078],[0.992157, 0.403922, 0.996078],[0.992157, 0.415686, 0.996078],[0.992157, 0.423529, 0.992157],
[0.992157, 0.431373, 0.992157],[0.996078, 0.439216, 0.992157],[0.996078, 0.447059, 0.992157],[0.996078, 0.454902, 0.992157],
[0.996078, 0.462745, 0.992157],[0.996078, 0.470588, 0.992157],[0.996078, 0.482353, 0.992157],[0.996078, 0.490196, 0.992157],
[0.996078, 0.498039, 0.992157],[0.996078, 0.505882, 0.992157],[0.996078, 0.513725, 0.992157],[0.996078, 0.521569, 0.992157],
[0.996078, 0.529412, 0.992157],[0.996078, 0.533333, 0.992157],[0.996078, 0.541176, 0.992157],[0.996078, 0.54902, 0.992157],
[0.996078, 0.556863, 0.992157],[0.996078, 0.564706, 0.992157],[0.996078, 0.572549, 0.992157],[0.996078, 0.580392, 0.992157],
[0.996078, 0.584314, 0.992157],[0.996078, 0.592157, 0.992157],[0.996078, 0.6, 0.992157],[0.996078, 0.607843, 0.992157],
[0.996078, 0.611765, 0.992157],[0.996078, 0.619608, 0.992157],[0.996078, 0.627451, 0.992157],[0.996078, 0.635294, 0.992157],
[0.996078, 0.639216, 0.992157],[0.996078, 0.647059, 0.992157],[0.996078, 0.654902, 0.992157],[0.996078, 0.658824, 0.992157],
[0.996078, 0.666667, 0.992157],[0.996078, 0.67451, 0.992157],[0.996078, 0.678431, 0.992157],[0.996078, 0.686275, 0.992157],
[0.996078, 0.694118, 0.996078],[0.996078, 0.698039, 0.996078],[0.996078, 0.705882, 0.996078],[0.996078, 0.709804, 0.996078],
[0.996078, 0.717647, 0.996078],[0.996078, 0.72549, 0.996078],[0.996078, 0.729412, 0.996078],[0.996078, 0.737255, 0.996078],
[0.996078, 0.741176, 0.996078],[0.996078, 0.74902, 0.996078],[0.996078, 0.756863, 0.996078],[0.996078, 0.760784, 0.996078],
[1, 0.768627, 0.996078],[1, 0.772549, 0.996078],[1, 0.780392, 0.996078],[1, 0.784314, 0.996078],
[1, 0.792157, 0.996078],[1, 0.796078, 0.996078],[1, 0.803922, 0.996078],[1, 0.807843, 0.996078],
[1, 0.815686, 0.996078],[1, 0.819608, 0.996078],[1, 0.827451, 0.996078],[1, 0.831373, 0.996078],
[1, 0.839216, 0.996078],[1, 0.843137, 0.996078],[1, 0.85098, 0.996078],[1, 0.854902, 0.996078],
[1, 0.862745, 0.996078],[1, 0.866667, 0.996078],[1, 0.87451, 0.996078],[1, 0.878431, 0.996078],
[1, 0.886275, 0.996078],[1, 0.890196, 0.996078],[1, 0.898039, 0.996078],[0.996078, 0.901961, 0.996078],
[0.996078, 0.909804, 0.996078],[0.996078, 0.913725, 0.996078],[0.996078, 0.921569, 0.996078] 
], 
[
[0, 0.0196078, 0.345098],[0, 0.0196078, 0.356863],[0, 0.0235294, 0.364706],[0, 0.027451, 0.372549],
[0, 0.027451, 0.384314],[0, 0.0313725, 0.392157],[0, 0.0313725, 0.4],[0, 0.0352941, 0.411765],
[0, 0.0352941, 0.419608],[0, 0.0352941, 0.431373],[0, 0.0352941, 0.439216],[0, 0.0392157, 0.447059],
[0, 0.0392157, 0.458824],[0, 0.0392157, 0.466667],[0, 0.0431373, 0.478431],[0, 0.0431373, 0.486275],
[0, 0.0431373, 0.494118],[0, 0.0431373, 0.505882],[0, 0.0470588, 0.513725],[0, 0.0470588, 0.52549],
[0, 0.0470588, 0.533333],[0, 0.0509804, 0.545098],[0, 0.0509804, 0.552941],[0, 0.0509804, 0.564706],
[0, 0.0509804, 0.572549],[0, 0.054902, 0.584314],[0, 0.054902, 0.592157],[0, 0.054902, 0.603922],
[0, 0.0588235, 0.611765],[0, 0.0588235, 0.623529],[0, 0.0588235, 0.631373],[0, 0.0627451, 0.643137],
[0, 0.0627451, 0.65098],[0, 0.0627451, 0.662745],[0, 0.0666667, 0.670588],[0, 0.0666667, 0.682353],
[0, 0.0705882, 0.690196],[0, 0.0705882, 0.701961],[0, 0.0705882, 0.709804],[0, 0.0745098, 0.721569],
[0, 0.0745098, 0.729412],[0, 0.0745098, 0.741176],[0, 0.0784314, 0.74902],[0, 0.0784314, 0.760784],
[0, 0.0823529, 0.768627],[0, 0.0823529, 0.776471],[0, 0.0862745, 0.788235],[0, 0.0862745, 0.796078],
[0, 0.0862745, 0.803922],[0, 0.0901961, 0.815686],[0, 0.0901961, 0.823529],[0, 0.0941176, 0.831373],
[0, 0.0941176, 0.839216],[0, 0.0941176, 0.847059],[0, 0.0980392, 0.854902],[0, 0.0980392, 0.862745],
[0, 0.0980392, 0.870588],[0, 0.101961, 0.87451],[0, 0.101961, 0.882353],[0, 0.101961, 0.890196],
[0, 0.105882, 0.894118],[0, 0.105882, 0.901961],[0, 0.105882, 0.909804],[0, 0.105882, 0.913725],
[0, 0.109804, 0.921569],[0, 0.109804, 0.92549],[0, 0.109804, 0.929412],[0, 0.109804, 0.937255],
[0, 0.113725, 0.941176],[0, 0.113725, 0.945098],[0, 0.113725, 0.94902],[0, 0.113725, 0.952941],
[0, 0.117647, 0.956863],[0, 0.117647, 0.960784],[0, 0.117647, 0.964706],[0, 0.117647, 0.968627],
[0, 0.117647, 0.972549],[0, 0.117647, 0.976471],[0, 0.121569, 0.980392],[0, 0.121569, 0.980392],
[0, 0.121569, 0.984314],[0, 0.121569, 0.988235],[0, 0.121569, 0.988235],[0.0431373, 0.121569, 0.992157],
[0.0980392, 0.121569, 0.996078],[0.133333, 0.121569, 0.996078],[0.160784, 0.121569, 1],[0.188235, 0.121569, 1],
[0.211765, 0.121569, 1],[0.231373, 0.121569, 1],[0.25098, 0.12549, 1],[0.270588, 0.12549, 1],
[0.290196, 0.12549, 1],[0.305882, 0.12549, 1],[0.321569, 0.12549, 1],[0.337255, 0.12549, 1],
[0.352941, 0.121569, 1],[0.368627, 0.121569, 1],[0.384314, 0.121569, 1],[0.4, 0.121569, 1],
[0.411765, 0.121569, 1],[0.427451, 0.121569, 1],[0.439216, 0.121569, 1],[0.454902, 0.121569, 1],
[0.466667, 0.121569, 1],[0.478431, 0.121569, 1],[0.490196, 0.121569, 1],[0.501961, 0.121569, 1],
[0.513725, 0.121569, 1],[0.52549, 0.121569, 1],[0.537255, 0.121569, 1],[0.54902, 0.121569, 1],
[0.560784, 0.121569, 1],[0.572549, 0.121569, 1],[0.584314, 0.121569, 1],[0.592157, 0.121569, 1],
[0.603922, 0.121569, 1],[0.615686, 0.121569, 1],[0.623529, 0.117647, 1],[0.635294, 0.117647, 1],
[0.647059, 0.117647, 1],[0.654902, 0.117647, 1],[0.666667, 0.117647, 1],[0.67451, 0.117647, 1],
[0.686275, 0.117647, 1],[0.694118, 0.117647, 1],[0.705882, 0.117647, 1],[0.713725, 0.117647, 1],
[0.721569, 0.117647, 1],[0.733333, 0.117647, 1],[0.741176, 0.117647, 1],[0.74902, 0.121569, 1],
[0.760784, 0.121569, 1],[0.768627, 0.121569, 1],[0.776471, 0.12549, 1],[0.784314, 0.12549, 1],
[0.792157, 0.129412, 1],[0.8, 0.133333, 1],[0.807843, 0.137255, 1],[0.815686, 0.137255, 1],
[0.823529, 0.141176, 1],[0.831373, 0.145098, 1],[0.839216, 0.14902, 1],[0.843137, 0.156863, 1],
[0.85098, 0.160784, 1],[0.858824, 0.164706, 1],[0.866667, 0.168627, 1],[0.870588, 0.176471, 1],
[0.878431, 0.180392, 1],[0.882353, 0.184314, 1],[0.890196, 0.192157, 1],[0.898039, 0.196078, 1],
[0.901961, 0.203922, 1],[0.905882, 0.207843, 1],[0.913725, 0.215686, 1],[0.917647, 0.219608, 1],
[0.92549, 0.227451, 1],[0.929412, 0.235294, 1],[0.933333, 0.239216, 1],[0.937255, 0.247059, 1],
[0.945098, 0.254902, 1],[0.94902, 0.262745, 1],[0.952941, 0.266667, 1],[0.956863, 0.27451, 1],
[0.960784, 0.282353, 1],[0.964706, 0.290196, 1],[0.968627, 0.298039, 1],[0.972549, 0.305882, 1],
[0.972549, 0.313725, 1],[0.976471, 0.321569, 1],[0.980392, 0.329412, 1],[0.984314, 0.337255, 1],
[0.984314, 0.345098, 1],[0.988235, 0.352941, 1],[0.988235, 0.360784, 1],[0.992157, 0.368627, 1],
[0.992157, 0.380392, 1],[0.996078, 0.388235, 1],[0.996078, 0.396078, 1],[0.996078, 0.403922, 1],
[1, 0.411765, 1],[1, 0.419608, 1],[1, 0.431373, 1],[1, 0.439216, 1],
[1, 0.447059, 1],[1, 0.454902, 1],[1, 0.462745, 1],[1, 0.470588, 1],
[1, 0.478431, 1],[1, 0.490196, 1],[1, 0.498039, 1],[1, 0.505882, 1],
[1, 0.513725, 1],[1, 0.521569, 1],[1, 0.52549, 1],[1, 0.533333, 1],
[1, 0.541176, 1],[1, 0.54902, 1],[1, 0.556863, 1],[1, 0.564706, 1],
[1, 0.572549, 1],[1, 0.580392, 1],[1, 0.584314, 1],[1, 0.592157, 1],
[1, 0.6, 1],[1, 0.607843, 1],[1, 0.615686, 1],[1, 0.619608, 1],
[1, 0.627451, 1],[1, 0.635294, 1],[1, 0.639216, 1],[1, 0.647059, 1],
[1, 0.654902, 1],[1, 0.658824, 1],[1, 0.666667, 1],[1, 0.67451, 1],
[1, 0.678431, 1],[1, 0.686275, 1],[1, 0.694118, 1],[1, 0.698039, 1],
[1, 0.705882, 1],[1, 0.713725, 1],[1, 0.717647, 1],[1, 0.72549, 1],
[1, 0.729412, 1],[1, 0.737255, 1],[1, 0.745098, 1],[1, 0.74902, 1],
[1, 0.756863, 1],[1, 0.760784, 1],[1, 0.768627, 1],[1, 0.776471, 1],
[1, 0.780392, 1],[1, 0.788235, 1],[1, 0.792157, 1],[1, 0.8, 1],
[1, 0.807843, 1],[1, 0.811765, 1],[1, 0.819608, 1],[1, 0.823529, 1],
[1, 0.831373, 1],[1, 0.835294, 1],[1, 0.843137, 1],[1, 0.847059, 1],
[1, 0.854902, 1],[1, 0.862745, 1],[1, 0.866667, 1],[1, 0.87451, 1],
[1, 0.878431, 1],[1, 0.886275, 1],[1, 0.890196, 1],[1, 0.898039, 1],
[1, 0.901961, 1],[1, 0.909804, 0.996078],[1, 0.913725, 0.996078],[0.996078, 0.921569, 0.996078] 
], 
[
[0, 0.0588235, 0.364706],[0.00392157, 0.0588235, 0.368627],[0.00392157, 0.0627451, 0.376471],
[0.00392157, 0.0627451, 0.384314],[0.00392157, 0.0666667, 0.388235],[0.00784314, 0.0666667, 0.396078],[0.00784314, 0.0705882, 0.403922],
[0.0117647, 0.0705882, 0.407843],[0.0156863, 0.0705882, 0.415686],[0.0196078, 0.0745098, 0.419608],[0.0235294, 0.0745098, 0.427451],
[0.027451, 0.0784314, 0.435294],[0.0313725, 0.0784314, 0.439216],[0.0392157, 0.0823529, 0.447059],[0.0470588, 0.0823529, 0.45098],
[0.0509804, 0.0823529, 0.458824],[0.0588235, 0.0862745, 0.462745],[0.0666667, 0.0862745, 0.466667],[0.0705882, 0.0862745, 0.47451],
[0.0784314, 0.0901961, 0.478431],[0.0862745, 0.0901961, 0.486275],[0.0941176, 0.0901961, 0.490196],[0.0980392, 0.0941176, 0.494118],
[0.105882, 0.0941176, 0.501961],[0.113725, 0.0941176, 0.505882],[0.121569, 0.0980392, 0.509804],[0.129412, 0.0980392, 0.513725],
[0.137255, 0.0980392, 0.521569],[0.145098, 0.0980392, 0.52549],[0.152941, 0.101961, 0.529412],[0.160784, 0.101961, 0.533333],
[0.168627, 0.101961, 0.537255],[0.180392, 0.101961, 0.541176],[0.188235, 0.101961, 0.545098],[0.196078, 0.105882, 0.54902],
[0.203922, 0.105882, 0.552941],[0.215686, 0.105882, 0.556863],[0.223529, 0.105882, 0.560784],[0.231373, 0.105882, 0.560784],
[0.243137, 0.105882, 0.564706],[0.25098, 0.105882, 0.564706],[0.262745, 0.101961, 0.568627],[0.27451, 0.101961, 0.568627],
[0.282353, 0.101961, 0.572549],[0.294118, 0.101961, 0.572549],[0.305882, 0.0980392, 0.572549],[0.317647, 0.0980392, 0.572549],
[0.329412, 0.0980392, 0.572549],[0.337255, 0.0941176, 0.572549],[0.34902, 0.0941176, 0.572549],[0.360784, 0.0901961, 0.572549],
[0.368627, 0.0901961, 0.572549],[0.380392, 0.0901961, 0.572549],[0.388235, 0.0862745, 0.572549],[0.4, 0.0862745, 0.572549],
[0.407843, 0.0823529, 0.572549],[0.419608, 0.0823529, 0.572549],[0.427451, 0.0784314, 0.572549],[0.435294, 0.0784314, 0.572549],
[0.447059, 0.0745098, 0.572549],[0.454902, 0.0745098, 0.572549],[0.462745, 0.0705882, 0.568627],[0.47451, 0.0705882, 0.568627],
[0.482353, 0.0666667, 0.568627],[0.490196, 0.0666667, 0.568627],[0.498039, 0.0627451, 0.568627],[0.505882, 0.0627451, 0.568627],
[0.513725, 0.0588235, 0.564706],[0.52549, 0.0588235, 0.564706],[0.533333, 0.054902, 0.564706],[0.541176, 0.054902, 0.564706],
[0.54902, 0.0509804, 0.560784],[0.556863, 0.0509804, 0.560784],[0.564706, 0.0509804, 0.560784],[0.572549, 0.0470588, 0.556863],
[0.580392, 0.0470588, 0.556863],[0.588235, 0.0431373, 0.556863],[0.596078, 0.0431373, 0.552941],[0.603922, 0.0431373, 0.552941],
[0.611765, 0.0392157, 0.54902],[0.619608, 0.0392157, 0.54902],[0.627451, 0.0392157, 0.54902],[0.635294, 0.0392157, 0.545098],
[0.639216, 0.0392157, 0.545098],[0.647059, 0.0352941, 0.541176],[0.654902, 0.0352941, 0.541176],[0.662745, 0.0352941, 0.541176],
[0.670588, 0.0352941, 0.537255],[0.678431, 0.0352941, 0.537255],[0.682353, 0.0392157, 0.533333],[0.690196, 0.0392157, 0.533333],
[0.698039, 0.0392157, 0.533333],[0.705882, 0.0392157, 0.529412],[0.713725, 0.0392157, 0.529412],[0.717647, 0.0431373, 0.52549],
[0.72549, 0.0431373, 0.52549],[0.733333, 0.0470588, 0.521569],[0.741176, 0.0470588, 0.521569],[0.745098, 0.0509804, 0.521569],
[0.752941, 0.0509804, 0.517647],[0.760784, 0.054902, 0.517647],[0.764706, 0.054902, 0.513725],[0.772549, 0.0588235, 0.513725],
[0.780392, 0.0627451, 0.509804],[0.784314, 0.0627451, 0.509804],[0.792157, 0.0666667, 0.505882],[0.8, 0.0705882, 0.505882],
[0.803922, 0.0745098, 0.501961],[0.811765, 0.0784314, 0.501961],[0.815686, 0.0823529, 0.498039],[0.823529, 0.0862745, 0.498039],
[0.827451, 0.0901961, 0.494118],[0.835294, 0.0980392, 0.490196],[0.839216, 0.101961, 0.490196],[0.847059, 0.109804, 0.486275],
[0.85098, 0.113725, 0.482353],[0.854902, 0.121569, 0.482353],[0.862745, 0.12549, 0.478431],[0.866667, 0.133333, 0.47451],
[0.870588, 0.137255, 0.470588],[0.87451, 0.145098, 0.470588],[0.878431, 0.152941, 0.466667],[0.886275, 0.160784, 0.462745],
[0.890196, 0.164706, 0.458824],[0.894118, 0.172549, 0.454902],[0.898039, 0.180392, 0.45098],[0.901961, 0.188235, 0.447059],
[0.905882, 0.192157, 0.447059],[0.909804, 0.2, 0.443137],[0.913725, 0.207843, 0.439216],[0.917647, 0.215686, 0.435294],
[0.921569, 0.223529, 0.431373],[0.92549, 0.231373, 0.423529],[0.929412, 0.239216, 0.419608],[0.929412, 0.243137, 0.415686],
[0.933333, 0.25098, 0.411765],[0.937255, 0.258824, 0.407843],[0.941176, 0.266667, 0.403922],[0.941176, 0.27451, 0.4],
[0.945098, 0.282353, 0.396078],[0.94902, 0.290196, 0.392157],[0.94902, 0.298039, 0.384314],[0.952941, 0.301961, 0.380392],
[0.952941, 0.309804, 0.376471],[0.956863, 0.317647, 0.372549],[0.956863, 0.32549, 0.368627],[0.960784, 0.333333, 0.364706],
[0.960784, 0.341176, 0.360784],[0.964706, 0.345098, 0.360784],[0.964706, 0.352941, 0.356863],[0.968627, 0.360784, 0.352941],
[0.968627, 0.368627, 0.34902],[0.968627, 0.376471, 0.345098],[0.972549, 0.384314, 0.341176],[0.972549, 0.392157, 0.337255],
[0.972549, 0.396078, 0.333333],[0.972549, 0.403922, 0.333333],[0.972549, 0.411765, 0.329412],[0.976471, 0.419608, 0.32549],
[0.976471, 0.427451, 0.321569],[0.976471, 0.435294, 0.321569],[0.976471, 0.439216, 0.317647],[0.976471, 0.447059, 0.313725],
[0.976471, 0.454902, 0.309804],[0.976471, 0.462745, 0.309804],[0.976471, 0.470588, 0.305882],[0.976471, 0.478431, 0.301961],
[0.976471, 0.482353, 0.301961],[0.976471, 0.490196, 0.298039],[0.976471, 0.498039, 0.298039],[0.976471, 0.505882, 0.294118],
[0.976471, 0.509804, 0.290196],[0.976471, 0.517647, 0.290196],[0.976471, 0.52549, 0.286275],[0.976471, 0.529412, 0.282353],
[0.976471, 0.537255, 0.282353],[0.976471, 0.541176, 0.278431],[0.976471, 0.54902, 0.27451],[0.976471, 0.556863, 0.27451],
[0.976471, 0.560784, 0.270588],[0.976471, 0.568627, 0.266667],[0.976471, 0.572549, 0.266667],[0.976471, 0.580392, 0.262745],
[0.976471, 0.584314, 0.258824],[0.976471, 0.592157, 0.258824],[0.980392, 0.596078, 0.254902],[0.980392, 0.603922, 0.254902],
[0.980392, 0.607843, 0.25098],[0.980392, 0.615686, 0.247059],[0.980392, 0.619608, 0.247059],[0.980392, 0.623529, 0.243137],
[0.980392, 0.631373, 0.239216],[0.984314, 0.635294, 0.239216],[0.984314, 0.643137, 0.235294],[0.984314, 0.647059, 0.231373],
[0.984314, 0.65098, 0.231373],[0.984314, 0.658824, 0.227451],[0.988235, 0.662745, 0.223529],[0.988235, 0.670588, 0.223529],
[0.988235, 0.67451, 0.223529],[0.988235, 0.678431, 0.219608],[0.988235, 0.686275, 0.219608],[0.988235, 0.690196, 0.219608],
[0.988235, 0.694118, 0.219608],[0.988235, 0.701961, 0.219608],[0.988235, 0.705882, 0.219608],[0.988235, 0.713725, 0.219608],
[0.992157, 0.717647, 0.219608],[0.992157, 0.721569, 0.219608],[0.992157, 0.729412, 0.219608],[0.992157, 0.733333, 0.219608],
[0.992157, 0.741176, 0.219608],[0.992157, 0.745098, 0.219608],[0.992157, 0.74902, 0.219608],[0.992157, 0.756863, 0.223529],
[0.992157, 0.760784, 0.223529],[0.988235, 0.764706, 0.223529],[0.988235, 0.772549, 0.223529],[0.988235, 0.776471, 0.227451],
[0.988235, 0.784314, 0.227451],[0.988235, 0.788235, 0.227451],[0.988235, 0.792157, 0.231373],[0.988235, 0.8, 0.231373],
[0.988235, 0.803922, 0.231373],[0.988235, 0.807843, 0.235294],[0.988235, 0.815686, 0.235294],[0.988235, 0.819608, 0.239216],
[0.988235, 0.827451, 0.239216],[0.984314, 0.831373, 0.243137],[0.984314, 0.835294, 0.243137],[0.984314, 0.843137, 0.247059],
[0.984314, 0.847059, 0.247059],[0.984314, 0.85098, 0.25098],[0.984314, 0.858824, 0.25098],[0.984314, 0.862745, 0.254902],
[0.980392, 0.870588, 0.254902],[0.980392, 0.87451, 0.258824],[0.980392, 0.878431, 0.258824],[0.980392, 0.886275, 0.262745],
[0.980392, 0.890196, 0.262745],[0.976471, 0.894118, 0.266667],[0.976471, 0.901961, 0.270588],[0.976471, 0.905882, 0.270588],
[0.976471, 0.909804, 0.27451],[0.972549, 0.917647, 0.27451],[0.972549, 0.921569, 0.278431],[0.972549, 0.929412, 0.282353],
[0.972549, 0.933333, 0.282353],[0.968627, 0.937255, 0.286275],[0.968627, 0.945098, 0.290196],[0.968627, 0.94902, 0.290196],
[0.968627, 0.952941, 0.294118],[0.964706, 0.960784, 0.298039],[0.964706, 0.964706, 0.298039],[0.964706, 0.968627, 0.301961],
[0.960784, 0.976471, 0.305882] 
], 
[
[0, 0.0470588, 0.490196],[0, 0.0509804, 0.494118],
[0, 0.0509804, 0.501961],[0, 0.054902, 0.509804],[0, 0.054902, 0.517647],[0, 0.054902, 0.52549],
[0, 0.0588235, 0.529412],[0, 0.0588235, 0.537255],[0, 0.0627451, 0.545098],[0, 0.0627451, 0.54902],
[0, 0.0666667, 0.556863],[0, 0.0666667, 0.564706],[0, 0.0666667, 0.568627],[0, 0.0705882, 0.576471],
[0, 0.0705882, 0.580392],[0, 0.0705882, 0.588235],[0, 0.0745098, 0.592157],[0, 0.0745098, 0.6],
[0, 0.0745098, 0.603922],[0, 0.0784314, 0.607843],[0, 0.0784314, 0.615686],[0, 0.0784314, 0.619608],
[0, 0.0784314, 0.623529],[0, 0.0823529, 0.627451],[0, 0.0823529, 0.631373],[0, 0.0823529, 0.635294],
[0, 0.0823529, 0.639216],[0, 0.0823529, 0.643137],[0, 0.0823529, 0.647059],[0, 0.0862745, 0.65098],
[0, 0.0862745, 0.654902],[0, 0.0862745, 0.654902],[0, 0.0862745, 0.658824],[0, 0.0862745, 0.662745],
[0, 0.0862745, 0.662745],[0, 0.0862745, 0.662745],[0.0392157, 0.0862745, 0.666667],[0.0823529, 0.0862745, 0.666667],
[0.113725, 0.0823529, 0.666667],[0.137255, 0.0823529, 0.666667],[0.160784, 0.0823529, 0.666667],[0.184314, 0.0823529, 0.662745],
[0.203922, 0.0784314, 0.662745],[0.223529, 0.0784314, 0.658824],[0.243137, 0.0745098, 0.654902],[0.262745, 0.0745098, 0.65098],
[0.278431, 0.0705882, 0.647059],[0.298039, 0.0705882, 0.643137],[0.313725, 0.0666667, 0.639216],[0.32549, 0.0666667, 0.635294],
[0.341176, 0.0627451, 0.631373],[0.352941, 0.0588235, 0.627451],[0.368627, 0.0588235, 0.623529],[0.380392, 0.054902, 0.623529],
[0.392157, 0.054902, 0.619608],[0.403922, 0.0509804, 0.615686],[0.415686, 0.0509804, 0.611765],[0.423529, 0.0470588, 0.607843],
[0.435294, 0.0431373, 0.603922],[0.447059, 0.0431373, 0.6],[0.454902, 0.0392157, 0.6],[0.466667, 0.0392157, 0.596078],
[0.47451, 0.0352941, 0.592157],[0.486275, 0.0313725, 0.592157],[0.494118, 0.0313725, 0.588235],[0.501961, 0.027451, 0.584314],
[0.509804, 0.027451, 0.584314],[0.521569, 0.0235294, 0.580392],[0.529412, 0.0235294, 0.576471],[0.537255, 0.0235294, 0.576471],
[0.545098, 0.0196078, 0.572549],[0.552941, 0.0196078, 0.572549],[0.560784, 0.0156863, 0.568627],[0.568627, 0.0156863, 0.568627],
[0.576471, 0.0156863, 0.564706],[0.584314, 0.0117647, 0.564706],[0.592157, 0.0117647, 0.560784],[0.6, 0.0117647, 0.560784],
[0.607843, 0.00784314, 0.556863],[0.615686, 0.00784314, 0.556863],[0.623529, 0.00784314, 0.556863],[0.631373, 0.00392157, 0.552941],
[0.639216, 0.00392157, 0.552941],[0.647059, 0.00392157, 0.54902],[0.654902, 0.00392157, 0.54902],[0.662745, 0, 0.545098],
[0.670588, 0, 0.545098],[0.67451, 0, 0.541176],[0.682353, 0, 0.541176],[0.690196, 0, 0.537255],
[0.698039, 0, 0.537255],[0.705882, 0, 0.533333],[0.713725, 0, 0.533333],[0.721569, 0, 0.529412],
[0.72549, 0, 0.529412],[0.733333, 0, 0.529412],[0.741176, 0, 0.52549],[0.74902, 0, 0.52549],
[0.756863, 0, 0.521569],[0.764706, 0, 0.521569],[0.768627, 0, 0.517647],[0.776471, 0, 0.517647],
[0.784314, 0, 0.513725],[0.792157, 0, 0.513725],[0.8, 0, 0.509804],[0.807843, 0, 0.509804],
[0.811765, 0, 0.505882],[0.819608, 0, 0.505882],[0.827451, 0, 0.501961],[0.835294, 0, 0.501961],
[0.839216, 0, 0.498039],[0.847059, 0, 0.498039],[0.854902, 0, 0.494118],[0.858824, 0, 0.494118],
[0.866667, 0, 0.490196],[0.870588, 0, 0.486275],[0.878431, 0.00392157, 0.486275],[0.886275, 0.00784314, 0.482353],
[0.890196, 0.0156863, 0.482353],[0.898039, 0.0235294, 0.478431],[0.901961, 0.0313725, 0.478431],[0.909804, 0.0431373, 0.47451],
[0.913725, 0.0509804, 0.470588],[0.917647, 0.0627451, 0.470588],[0.92549, 0.0705882, 0.466667],[0.929412, 0.0784314, 0.466667],
[0.933333, 0.0862745, 0.462745],[0.941176, 0.0941176, 0.458824],[0.945098, 0.101961, 0.458824],[0.94902, 0.109804, 0.454902],
[0.956863, 0.117647, 0.45098],[0.960784, 0.12549, 0.45098],[0.964706, 0.133333, 0.447059],[0.968627, 0.141176, 0.443137],
[0.972549, 0.14902, 0.443137],[0.976471, 0.156863, 0.439216],[0.984314, 0.164706, 0.435294],[0.988235, 0.172549, 0.435294],
[0.992157, 0.180392, 0.431373],[0.996078, 0.188235, 0.427451],[1, 0.196078, 0.423529],[1, 0.203922, 0.423529],
[1, 0.211765, 0.419608],[1, 0.219608, 0.415686],[1, 0.227451, 0.415686],[1, 0.239216, 0.411765],
[1, 0.247059, 0.407843],[1, 0.254902, 0.403922],[1, 0.262745, 0.4],[1, 0.270588, 0.4],
[1, 0.278431, 0.396078],[1, 0.286275, 0.392157],[1, 0.294118, 0.388235],[1, 0.301961, 0.384314],
[1, 0.313725, 0.384314],[1, 0.321569, 0.380392],[1, 0.329412, 0.376471],[1, 0.337255, 0.372549],
[1, 0.345098, 0.368627],[1, 0.352941, 0.364706],[1, 0.364706, 0.360784],[1, 0.372549, 0.356863],
[1, 0.380392, 0.356863],[1, 0.388235, 0.352941],[1, 0.396078, 0.34902],[1, 0.403922, 0.345098],
[1, 0.415686, 0.341176],[1, 0.423529, 0.337255],[1, 0.431373, 0.333333],[1, 0.439216, 0.329412],
[1, 0.447059, 0.32549],[1, 0.454902, 0.321569],[1, 0.462745, 0.317647],[1, 0.470588, 0.313725],
[1, 0.478431, 0.305882],[1, 0.486275, 0.301961],[1, 0.494118, 0.298039],[1, 0.498039, 0.294118],
[1, 0.505882, 0.290196],[1, 0.513725, 0.286275],[1, 0.521569, 0.282353],[1, 0.529412, 0.278431],
[1, 0.533333, 0.270588],[1, 0.541176, 0.266667],[1, 0.54902, 0.262745],[1, 0.556863, 0.258824],
[1, 0.560784, 0.25098],[1, 0.568627, 0.247059],[1, 0.576471, 0.243137],[1, 0.584314, 0.235294],
[1, 0.588235, 0.231373],[1, 0.596078, 0.223529],[1, 0.603922, 0.219608],[1, 0.607843, 0.211765],
[1, 0.615686, 0.207843],[1, 0.623529, 0.2],[1, 0.627451, 0.196078],[1, 0.635294, 0.188235],
[1, 0.639216, 0.184314],[1, 0.647059, 0.180392],[1, 0.65098, 0.172549],[1, 0.658824, 0.168627],
[1, 0.662745, 0.164706],[1, 0.670588, 0.160784],[1, 0.67451, 0.156863],[1, 0.682353, 0.152941],
[1, 0.686275, 0.14902],[1, 0.694118, 0.14902],[1, 0.698039, 0.145098],[1, 0.705882, 0.141176],
[1, 0.709804, 0.137255],[1, 0.713725, 0.137255],[1, 0.721569, 0.133333],[1, 0.72549, 0.133333],
[1, 0.733333, 0.129412],[1, 0.737255, 0.12549],[1, 0.741176, 0.12549],[1, 0.74902, 0.121569],
[1, 0.752941, 0.121569],[1, 0.760784, 0.121569],[1, 0.764706, 0.117647],[1, 0.768627, 0.117647],
[1, 0.776471, 0.117647],[1, 0.780392, 0.113725],[1, 0.784314, 0.113725],[1, 0.792157, 0.113725],
[1, 0.796078, 0.113725],[1, 0.8, 0.113725],[1, 0.807843, 0.113725],[1, 0.811765, 0.109804],
[1, 0.815686, 0.109804],[1, 0.823529, 0.109804],[1, 0.827451, 0.109804],[1, 0.831373, 0.109804],
[1, 0.839216, 0.109804],[1, 0.843137, 0.113725],[1, 0.847059, 0.113725],[1, 0.854902, 0.113725],
[1, 0.858824, 0.113725],[1, 0.862745, 0.113725],[1, 0.870588, 0.113725],[1, 0.87451, 0.117647],
[1, 0.878431, 0.117647],[1, 0.886275, 0.117647],[1, 0.890196, 0.117647],[1, 0.894118, 0.121569],
[1, 0.901961, 0.121569],[1, 0.905882, 0.121569],[1, 0.909804, 0.12549],[1, 0.913725, 0.12549],
[1, 0.921569, 0.129412],[1, 0.92549, 0.129412],[1, 0.929412, 0.133333],[1, 0.937255, 0.133333],
[1, 0.941176, 0.137255],[1, 0.945098, 0.137255] 
], 
[
[0.4, 0.607843, 0.564706],
[0.4, 0.607843, 0.560784],[0.403922, 0.607843, 0.560784],[0.407843, 0.607843, 0.560784],[0.411765, 0.611765, 0.556863],
[0.411765, 0.611765, 0.556863],[0.415686, 0.611765, 0.552941],[0.419608, 0.611765, 0.552941],[0.419608, 0.615686, 0.54902],
[0.423529, 0.615686, 0.54902],[0.427451, 0.615686, 0.545098],[0.431373, 0.615686, 0.545098],[0.431373, 0.615686, 0.541176],
[0.435294, 0.619608, 0.541176],[0.439216, 0.619608, 0.537255],[0.443137, 0.619608, 0.537255],[0.443137, 0.619608, 0.537255],
[0.447059, 0.623529, 0.533333],[0.45098, 0.623529, 0.533333],[0.454902, 0.623529, 0.529412],[0.454902, 0.623529, 0.529412],
[0.458824, 0.623529, 0.52549],[0.462745, 0.627451, 0.52549],[0.466667, 0.627451, 0.521569],[0.466667, 0.627451, 0.521569],
[0.470588, 0.627451, 0.521569],[0.47451, 0.631373, 0.517647],[0.478431, 0.631373, 0.517647],[0.482353, 0.631373, 0.513725],
[0.482353, 0.631373, 0.513725],[0.486275, 0.631373, 0.509804],[0.490196, 0.635294, 0.509804],[0.494118, 0.635294, 0.505882],
[0.498039, 0.635294, 0.505882],[0.498039, 0.635294, 0.505882],[0.501961, 0.635294, 0.501961],[0.505882, 0.635294, 0.501961],
[0.509804, 0.639216, 0.498039],[0.513725, 0.639216, 0.498039],[0.517647, 0.639216, 0.498039],[0.517647, 0.639216, 0.494118],
[0.521569, 0.639216, 0.494118],[0.52549, 0.643137, 0.490196],[0.529412, 0.643137, 0.490196],[0.533333, 0.643137, 0.490196],
[0.537255, 0.643137, 0.486275],[0.541176, 0.643137, 0.486275],[0.545098, 0.643137, 0.482353],[0.54902, 0.643137, 0.482353],
[0.54902, 0.647059, 0.482353],[0.552941, 0.647059, 0.478431],[0.556863, 0.647059, 0.478431],[0.560784, 0.647059, 0.478431],
[0.564706, 0.647059, 0.47451],[0.568627, 0.647059, 0.47451],[0.572549, 0.647059, 0.470588],[0.576471, 0.647059, 0.470588],
[0.580392, 0.65098, 0.470588],[0.584314, 0.65098, 0.466667],[0.588235, 0.65098, 0.466667],[0.592157, 0.65098, 0.466667],
[0.6, 0.65098, 0.462745],[0.603922, 0.65098, 0.462745],[0.607843, 0.65098, 0.462745],[0.611765, 0.65098, 0.462745],
[0.615686, 0.65098, 0.458824],[0.619608, 0.65098, 0.458824],[0.623529, 0.65098, 0.458824],[0.631373, 0.65098, 0.454902],
[0.635294, 0.65098, 0.454902],[0.639216, 0.65098, 0.454902],[0.643137, 0.65098, 0.454902],[0.65098, 0.65098, 0.45098],
[0.654902, 0.65098, 0.45098],[0.658824, 0.65098, 0.45098],[0.666667, 0.65098, 0.45098],[0.670588, 0.65098, 0.45098],
[0.67451, 0.65098, 0.447059],[0.682353, 0.65098, 0.447059],[0.686275, 0.65098, 0.447059],[0.694118, 0.65098, 0.447059],
[0.698039, 0.65098, 0.447059],[0.701961, 0.65098, 0.447059],[0.705882, 0.65098, 0.443137],[0.713725, 0.65098, 0.443137],
[0.717647, 0.65098, 0.443137],[0.721569, 0.65098, 0.443137],[0.72549, 0.65098, 0.443137],[0.729412, 0.65098, 0.443137],
[0.733333, 0.65098, 0.439216],[0.737255, 0.65098, 0.439216],[0.741176, 0.65098, 0.439216],[0.74902, 0.65098, 0.439216],
[0.752941, 0.65098, 0.439216],[0.752941, 0.65098, 0.435294],[0.756863, 0.65098, 0.435294],[0.760784, 0.65098, 0.435294],
[0.764706, 0.65098, 0.435294],[0.768627, 0.65098, 0.435294],[0.772549, 0.65098, 0.435294],[0.776471, 0.65098, 0.431373],
[0.780392, 0.65098, 0.431373],[0.780392, 0.65098, 0.431373],[0.784314, 0.654902, 0.431373],[0.788235, 0.654902, 0.431373],
[0.792157, 0.654902, 0.431373],[0.792157, 0.654902, 0.427451],[0.796078, 0.654902, 0.427451],[0.8, 0.654902, 0.427451],
[0.8, 0.654902, 0.427451],[0.803922, 0.654902, 0.427451],[0.803922, 0.658824, 0.427451],[0.807843, 0.658824, 0.423529],
[0.811765, 0.658824, 0.423529],[0.811765, 0.658824, 0.423529],[0.815686, 0.658824, 0.423529],[0.815686, 0.662745, 0.423529],
[0.815686, 0.662745, 0.423529],[0.819608, 0.662745, 0.419608],[0.819608, 0.662745, 0.419608],[0.823529, 0.666667, 0.419608],
[0.823529, 0.666667, 0.419608],[0.823529, 0.666667, 0.419608],[0.827451, 0.666667, 0.415686],[0.827451, 0.670588, 0.415686],
[0.827451, 0.670588, 0.415686],[0.831373, 0.670588, 0.415686],[0.831373, 0.67451, 0.415686],[0.831373, 0.67451, 0.415686],
[0.831373, 0.67451, 0.411765],[0.831373, 0.678431, 0.411765],[0.835294, 0.678431, 0.411765],[0.835294, 0.678431, 0.411765],
[0.835294, 0.678431, 0.411765],[0.835294, 0.682353, 0.411765],[0.835294, 0.682353, 0.411765],[0.835294, 0.682353, 0.411765],
[0.839216, 0.686275, 0.411765],[0.839216, 0.686275, 0.411765],[0.839216, 0.686275, 0.411765],[0.839216, 0.690196, 0.411765],
[0.839216, 0.690196, 0.415686],[0.839216, 0.690196, 0.415686],[0.843137, 0.694118, 0.415686],[0.843137, 0.694118, 0.415686],
[0.843137, 0.694118, 0.415686],[0.843137, 0.698039, 0.419608],[0.843137, 0.698039, 0.419608],[0.843137, 0.698039, 0.419608],
[0.843137, 0.701961, 0.423529],[0.843137, 0.701961, 0.423529],[0.847059, 0.701961, 0.427451],[0.847059, 0.705882, 0.427451],
[0.847059, 0.705882, 0.427451],[0.847059, 0.705882, 0.431373],[0.847059, 0.709804, 0.431373],[0.847059, 0.709804, 0.435294],
[0.847059, 0.709804, 0.439216],[0.847059, 0.713725, 0.439216],[0.847059, 0.713725, 0.443137],[0.847059, 0.713725, 0.443137],
[0.847059, 0.717647, 0.447059],[0.847059, 0.717647, 0.45098],[0.847059, 0.717647, 0.45098],[0.85098, 0.721569, 0.454902],
[0.85098, 0.721569, 0.458824],[0.85098, 0.721569, 0.462745],[0.85098, 0.721569, 0.462745],[0.85098, 0.72549, 0.466667],
[0.85098, 0.72549, 0.470588],[0.85098, 0.72549, 0.47451],[0.85098, 0.729412, 0.478431],[0.85098, 0.729412, 0.482353],
[0.85098, 0.729412, 0.486275],[0.85098, 0.733333, 0.490196],[0.85098, 0.733333, 0.494118],[0.85098, 0.733333, 0.498039],
[0.85098, 0.737255, 0.501961],[0.847059, 0.737255, 0.505882],[0.847059, 0.737255, 0.509804],[0.847059, 0.741176, 0.513725],
[0.847059, 0.741176, 0.517647],[0.847059, 0.741176, 0.521569],[0.847059, 0.745098, 0.529412],[0.847059, 0.745098, 0.533333],
[0.847059, 0.745098, 0.537255],[0.847059, 0.74902, 0.541176],[0.847059, 0.74902, 0.545098],[0.847059, 0.74902, 0.54902],
[0.847059, 0.752941, 0.552941],[0.847059, 0.752941, 0.556863],[0.847059, 0.752941, 0.560784],[0.847059, 0.756863, 0.564706],
[0.847059, 0.756863, 0.568627],[0.847059, 0.756863, 0.572549],[0.847059, 0.760784, 0.580392],[0.847059, 0.760784, 0.584314],
[0.847059, 0.760784, 0.588235],[0.847059, 0.764706, 0.592157],[0.847059, 0.764706, 0.596078],[0.847059, 0.764706, 0.6],
[0.847059, 0.768627, 0.603922],[0.847059, 0.768627, 0.607843],[0.847059, 0.768627, 0.611765],[0.843137, 0.772549, 0.615686],
[0.843137, 0.772549, 0.619608],[0.843137, 0.772549, 0.623529],[0.843137, 0.772549, 0.631373],[0.843137, 0.776471, 0.635294],
[0.843137, 0.776471, 0.639216],[0.843137, 0.776471, 0.643137],[0.843137, 0.780392, 0.647059],[0.843137, 0.780392, 0.65098],
[0.843137, 0.780392, 0.654902],[0.843137, 0.784314, 0.658824],[0.843137, 0.784314, 0.662745],[0.843137, 0.784314, 0.666667],
[0.843137, 0.788235, 0.670588],[0.843137, 0.788235, 0.67451],[0.843137, 0.788235, 0.678431],[0.843137, 0.792157, 0.686275],
[0.843137, 0.792157, 0.690196],[0.843137, 0.792157, 0.694118],[0.843137, 0.792157, 0.698039],[0.843137, 0.796078, 0.701961],
[0.843137, 0.796078, 0.705882],[0.843137, 0.796078, 0.709804],[0.839216, 0.8, 0.713725],[0.839216, 0.8, 0.717647],
[0.839216, 0.8, 0.721569],[0.839216, 0.803922, 0.72549],[0.839216, 0.803922, 0.729412],[0.839216, 0.803922, 0.733333],
[0.839216, 0.807843, 0.741176],[0.839216, 0.807843, 0.745098],[0.839216, 0.807843, 0.74902],[0.839216, 0.807843, 0.752941],
[0.839216, 0.811765, 0.756863],[0.839216, 0.811765, 0.760784],[0.839216, 0.811765, 0.764706],[0.839216, 0.815686, 0.768627],
[0.839216, 0.815686, 0.772549],[0.835294, 0.815686, 0.776471],[0.835294, 0.819608, 0.780392],[0.835294, 0.819608, 0.784314],
[0.835294, 0.819608, 0.788235],[0.835294, 0.819608, 0.796078],[0.835294, 0.823529, 0.8],[0.835294, 0.823529, 0.803922],
[0.835294, 0.823529, 0.807843],[0.835294, 0.827451, 0.811765],[0.835294, 0.827451, 0.815686],[0.835294, 0.827451, 0.819608],
[0.831373, 0.831373, 0.823529],[0.831373, 0.831373, 0.827451],[0.831373, 0.831373, 0.831373] 
], 
[
[0.439216, 0.678431, 0.360784],[0.443137, 0.678431, 0.360784],[0.447059, 0.678431, 0.360784],[0.454902, 0.678431, 0.360784],
[0.458824, 0.678431, 0.360784],[0.462745, 0.678431, 0.364706],[0.470588, 0.678431, 0.364706],[0.47451, 0.678431, 0.364706],
[0.478431, 0.682353, 0.364706],[0.486275, 0.682353, 0.364706],[0.490196, 0.682353, 0.364706],[0.494118, 0.682353, 0.364706],
[0.501961, 0.682353, 0.364706],[0.505882, 0.682353, 0.364706],[0.509804, 0.682353, 0.368627],[0.513725, 0.682353, 0.368627],
[0.521569, 0.682353, 0.368627],[0.52549, 0.682353, 0.368627],[0.529412, 0.682353, 0.368627],[0.533333, 0.682353, 0.368627],
[0.541176, 0.682353, 0.368627],[0.545098, 0.682353, 0.368627],[0.54902, 0.682353, 0.372549],[0.552941, 0.686275, 0.372549],
[0.556863, 0.686275, 0.372549],[0.564706, 0.686275, 0.372549],[0.568627, 0.686275, 0.372549],[0.572549, 0.686275, 0.372549],
[0.576471, 0.686275, 0.372549],[0.580392, 0.686275, 0.372549],[0.584314, 0.686275, 0.372549],[0.588235, 0.686275, 0.376471],
[0.596078, 0.686275, 0.376471],[0.6, 0.686275, 0.376471],[0.603922, 0.686275, 0.376471],[0.607843, 0.686275, 0.376471],
[0.611765, 0.686275, 0.376471],[0.615686, 0.686275, 0.376471],[0.619608, 0.686275, 0.376471],[0.623529, 0.686275, 0.380392],
[0.627451, 0.690196, 0.380392],[0.631373, 0.690196, 0.380392],[0.639216, 0.690196, 0.380392],[0.643137, 0.690196, 0.380392],
[0.647059, 0.690196, 0.380392],[0.65098, 0.690196, 0.380392],[0.654902, 0.690196, 0.380392],[0.658824, 0.690196, 0.384314],
[0.662745, 0.690196, 0.384314],[0.666667, 0.690196, 0.384314],[0.670588, 0.690196, 0.384314],[0.67451, 0.690196, 0.384314],
[0.678431, 0.690196, 0.384314],[0.682353, 0.690196, 0.384314],[0.686275, 0.690196, 0.384314],[0.690196, 0.690196, 0.388235],
[0.694118, 0.690196, 0.388235],[0.698039, 0.690196, 0.388235],[0.701961, 0.690196, 0.388235],[0.705882, 0.694118, 0.388235],
[0.709804, 0.694118, 0.388235],[0.713725, 0.694118, 0.388235],[0.717647, 0.694118, 0.392157],[0.721569, 0.694118, 0.392157],
[0.72549, 0.694118, 0.392157],[0.729412, 0.694118, 0.392157],[0.733333, 0.694118, 0.392157],[0.737255, 0.694118, 0.392157],
[0.741176, 0.694118, 0.392157],[0.745098, 0.694118, 0.392157],[0.74902, 0.694118, 0.396078],[0.752941, 0.694118, 0.396078],
[0.756863, 0.694118, 0.396078],[0.756863, 0.694118, 0.396078],[0.760784, 0.694118, 0.396078],[0.764706, 0.694118, 0.396078],
[0.768627, 0.694118, 0.396078],[0.772549, 0.698039, 0.4],[0.776471, 0.698039, 0.4],[0.780392, 0.698039, 0.4],
[0.784314, 0.698039, 0.4],[0.788235, 0.698039, 0.4],[0.792157, 0.698039, 0.4],[0.792157, 0.698039, 0.4],
[0.796078, 0.698039, 0.4],[0.8, 0.698039, 0.403922],[0.803922, 0.698039, 0.403922],[0.807843, 0.698039, 0.403922],
[0.811765, 0.698039, 0.403922],[0.811765, 0.698039, 0.403922],[0.815686, 0.698039, 0.403922],[0.819608, 0.698039, 0.403922],
[0.823529, 0.701961, 0.407843],[0.827451, 0.701961, 0.407843],[0.831373, 0.701961, 0.407843],[0.831373, 0.701961, 0.407843],
[0.835294, 0.701961, 0.407843],[0.839216, 0.701961, 0.407843],[0.843137, 0.701961, 0.407843],[0.843137, 0.701961, 0.411765],
[0.847059, 0.701961, 0.411765],[0.85098, 0.701961, 0.411765],[0.854902, 0.701961, 0.411765],[0.854902, 0.705882, 0.411765],
[0.858824, 0.705882, 0.411765],[0.862745, 0.705882, 0.411765],[0.862745, 0.705882, 0.415686],[0.866667, 0.705882, 0.415686],
[0.870588, 0.705882, 0.415686],[0.870588, 0.705882, 0.415686],[0.87451, 0.705882, 0.415686],[0.878431, 0.709804, 0.415686],
[0.878431, 0.709804, 0.419608],[0.882353, 0.709804, 0.419608],[0.882353, 0.709804, 0.419608],[0.886275, 0.709804, 0.419608],
[0.886275, 0.709804, 0.419608],[0.890196, 0.713725, 0.419608],[0.890196, 0.713725, 0.419608],[0.894118, 0.713725, 0.423529],
[0.894118, 0.713725, 0.423529],[0.898039, 0.713725, 0.423529],[0.898039, 0.717647, 0.423529],[0.898039, 0.717647, 0.423529],
[0.901961, 0.717647, 0.423529],[0.901961, 0.717647, 0.427451],[0.901961, 0.721569, 0.427451],[0.901961, 0.721569, 0.427451],
[0.901961, 0.721569, 0.427451],[0.901961, 0.72549, 0.427451],[0.905882, 0.72549, 0.431373],[0.905882, 0.72549, 0.431373],
[0.905882, 0.729412, 0.431373],[0.905882, 0.729412, 0.431373],[0.905882, 0.733333, 0.435294],[0.905882, 0.733333, 0.435294],
[0.905882, 0.733333, 0.435294],[0.905882, 0.737255, 0.439216],[0.905882, 0.737255, 0.439216],[0.905882, 0.737255, 0.443137],
[0.905882, 0.741176, 0.443137],[0.905882, 0.741176, 0.443137],[0.905882, 0.741176, 0.447059],[0.905882, 0.745098, 0.447059],
[0.905882, 0.745098, 0.45098],[0.905882, 0.745098, 0.45098],[0.905882, 0.74902, 0.454902],[0.905882, 0.74902, 0.454902],
[0.905882, 0.752941, 0.458824],[0.905882, 0.752941, 0.458824],[0.905882, 0.752941, 0.462745],[0.905882, 0.756863, 0.466667],
[0.905882, 0.756863, 0.466667],[0.905882, 0.756863, 0.470588],[0.905882, 0.760784, 0.47451],[0.905882, 0.760784, 0.47451],
[0.905882, 0.760784, 0.478431],[0.905882, 0.764706, 0.482353],[0.905882, 0.764706, 0.486275],[0.905882, 0.764706, 0.486275],
[0.905882, 0.768627, 0.490196],[0.905882, 0.768627, 0.494118],[0.905882, 0.768627, 0.498039],[0.905882, 0.772549, 0.498039],
[0.905882, 0.772549, 0.501961],[0.905882, 0.772549, 0.505882],[0.905882, 0.776471, 0.509804],[0.905882, 0.776471, 0.513725],
[0.905882, 0.776471, 0.517647],[0.905882, 0.780392, 0.521569],[0.905882, 0.780392, 0.52549],[0.905882, 0.784314, 0.529412],
[0.905882, 0.784314, 0.533333],[0.905882, 0.784314, 0.537255],[0.905882, 0.788235, 0.541176],[0.905882, 0.788235, 0.545098],
[0.905882, 0.788235, 0.54902],[0.905882, 0.792157, 0.552941],[0.905882, 0.792157, 0.556863],[0.905882, 0.792157, 0.560784],
[0.905882, 0.796078, 0.564706],[0.905882, 0.796078, 0.568627],[0.905882, 0.796078, 0.576471],[0.905882, 0.8, 0.580392],
[0.905882, 0.8, 0.584314],[0.905882, 0.8, 0.588235],[0.905882, 0.803922, 0.592157],[0.901961, 0.803922, 0.596078],
[0.901961, 0.803922, 0.6],[0.901961, 0.807843, 0.603922],[0.901961, 0.807843, 0.607843],[0.901961, 0.807843, 0.611765],
[0.901961, 0.811765, 0.619608],[0.901961, 0.811765, 0.623529],[0.901961, 0.811765, 0.627451],[0.901961, 0.815686, 0.631373],
[0.901961, 0.815686, 0.635294],[0.901961, 0.815686, 0.639216],[0.901961, 0.815686, 0.643137],[0.901961, 0.819608, 0.647059],
[0.901961, 0.819608, 0.65098],[0.901961, 0.819608, 0.654902],[0.901961, 0.823529, 0.662745],[0.901961, 0.823529, 0.666667],
[0.901961, 0.823529, 0.670588],[0.901961, 0.827451, 0.67451],[0.901961, 0.827451, 0.678431],[0.901961, 0.827451, 0.682353],
[0.901961, 0.831373, 0.686275],[0.901961, 0.831373, 0.690196],[0.898039, 0.831373, 0.694118],[0.898039, 0.835294, 0.698039],
[0.898039, 0.835294, 0.701961],[0.898039, 0.835294, 0.709804],[0.898039, 0.839216, 0.713725],[0.898039, 0.839216, 0.717647],
[0.898039, 0.839216, 0.721569],[0.898039, 0.843137, 0.72549],[0.898039, 0.843137, 0.729412],[0.898039, 0.843137, 0.733333],
[0.898039, 0.843137, 0.737255],[0.898039, 0.847059, 0.741176],[0.898039, 0.847059, 0.745098],[0.898039, 0.847059, 0.74902],
[0.898039, 0.85098, 0.756863],[0.898039, 0.85098, 0.760784],[0.898039, 0.85098, 0.764706],[0.898039, 0.854902, 0.768627],
[0.898039, 0.854902, 0.772549],[0.894118, 0.854902, 0.776471],[0.894118, 0.858824, 0.780392],[0.894118, 0.858824, 0.784314],
[0.894118, 0.858824, 0.788235],[0.894118, 0.862745, 0.792157],[0.894118, 0.862745, 0.796078],[0.894118, 0.862745, 0.8],
[0.894118, 0.862745, 0.807843],[0.894118, 0.866667, 0.811765],[0.894118, 0.866667, 0.815686],[0.894118, 0.866667, 0.819608],
[0.894118, 0.870588, 0.823529],[0.894118, 0.870588, 0.827451],[0.894118, 0.870588, 0.831373],[0.890196, 0.87451, 0.835294],
[0.890196, 0.87451, 0.839216],[0.890196, 0.87451, 0.843137],[0.890196, 0.878431, 0.847059],[0.890196, 0.878431, 0.854902],
[0.890196, 0.878431, 0.858824],[0.890196, 0.878431, 0.862745],[0.890196, 0.882353, 0.866667],[0.890196, 0.882353, 0.870588],
[0.890196, 0.882353, 0.87451],[0.890196, 0.886275, 0.878431],[0.886275, 0.886275, 0.882353],[0.886275, 0.886275, 0.886275] 
], 
[
[0.00392157, 0.0823529, 0.0235294],[0.00392157, 0.0862745, 0.0235294],[0.00784314, 0.0901961, 0.0235294],
[0.00784314, 0.0941176, 0.0235294],[0.0117647, 0.0980392, 0.0235294],[0.0156863, 0.0980392, 0.0235294],[0.0156863, 0.101961, 0.0235294],
[0.0196078, 0.105882, 0.0235294],[0.0196078, 0.109804, 0.0235294],[0.0235294, 0.109804, 0.0235294],[0.027451, 0.113725, 0.0235294],
[0.027451, 0.117647, 0.0235294],[0.0313725, 0.121569, 0.0235294],[0.0313725, 0.12549, 0.0235294],[0.0313725, 0.12549, 0.0235294],
[0.0352941, 0.129412, 0.0235294],[0.0352941, 0.133333, 0.0235294],[0.0352941, 0.137255, 0.0196078],[0.0352941, 0.141176, 0.0196078],
[0.0352941, 0.141176, 0.0196078],[0.0352941, 0.145098, 0.0196078],[0.0352941, 0.14902, 0.0196078],[0.0352941, 0.152941, 0.0196078],
[0.0352941, 0.156863, 0.0196078],[0.0352941, 0.160784, 0.0196078],[0.0352941, 0.164706, 0.0196078],[0.0352941, 0.164706, 0.0196078],
[0.0352941, 0.168627, 0.0196078],[0.0352941, 0.172549, 0.0196078],[0.0352941, 0.176471, 0.0196078],[0.0352941, 0.180392, 0.0196078],
[0.0352941, 0.184314, 0.0196078],[0.0352941, 0.188235, 0.0196078],[0.0313725, 0.192157, 0.0196078],[0.0313725, 0.192157, 0.0196078],
[0.0313725, 0.196078, 0.0196078],[0.0313725, 0.2, 0.0196078],[0.0313725, 0.203922, 0.0196078],[0.0352941, 0.207843, 0.0196078],
[0.0352941, 0.211765, 0.0196078],[0.0352941, 0.215686, 0.0196078],[0.0352941, 0.219608, 0.0196078],[0.0352941, 0.219608, 0.0196078],
[0.0352941, 0.223529, 0.0196078],[0.0392157, 0.227451, 0.0196078],[0.0392157, 0.231373, 0.0196078],[0.0392157, 0.235294, 0.0196078],
[0.0392157, 0.239216, 0.0196078],[0.0431373, 0.243137, 0.0196078],[0.0431373, 0.247059, 0.0196078],[0.0431373, 0.25098, 0.0196078],
[0.0431373, 0.25098, 0.0196078],[0.0470588, 0.254902, 0.0196078],[0.0470588, 0.258824, 0.0196078],[0.0470588, 0.262745, 0.0196078],
[0.0509804, 0.266667, 0.0196078],[0.0509804, 0.270588, 0.0196078],[0.0509804, 0.27451, 0.0196078],[0.0509804, 0.278431, 0.0196078],
[0.054902, 0.282353, 0.0196078],[0.054902, 0.286275, 0.0196078],[0.054902, 0.286275, 0.0196078],[0.054902, 0.290196, 0.0196078],
[0.0588235, 0.294118, 0.0196078],[0.0588235, 0.298039, 0.0196078],[0.0588235, 0.301961, 0.0196078],[0.0588235, 0.305882, 0.0196078],
[0.0627451, 0.309804, 0.0196078],[0.0627451, 0.313725, 0.0196078],[0.0627451, 0.317647, 0.0196078],[0.0627451, 0.321569, 0.0196078],
[0.0666667, 0.32549, 0.0196078],[0.0666667, 0.329412, 0.0235294],[0.0666667, 0.333333, 0.0235294],[0.0666667, 0.333333, 0.0235294],
[0.0705882, 0.337255, 0.0235294],[0.0705882, 0.341176, 0.0235294],[0.0705882, 0.345098, 0.0235294],[0.0705882, 0.34902, 0.0235294],
[0.0745098, 0.352941, 0.0235294],[0.0745098, 0.356863, 0.0235294],[0.0745098, 0.360784, 0.0235294],[0.0745098, 0.364706, 0.0235294],
[0.0784314, 0.368627, 0.0235294],[0.0784314, 0.372549, 0.027451],[0.0784314, 0.376471, 0.027451],[0.0784314, 0.380392, 0.027451],
[0.0823529, 0.384314, 0.027451],[0.0823529, 0.388235, 0.027451],[0.0823529, 0.392157, 0.027451],[0.0823529, 0.396078, 0.027451],
[0.0862745, 0.4, 0.027451],[0.0862745, 0.403922, 0.027451],[0.0862745, 0.407843, 0.027451],[0.0862745, 0.407843, 0.0313725],
[0.0901961, 0.411765, 0.0313725],[0.0901961, 0.415686, 0.0313725],[0.0901961, 0.419608, 0.0313725],[0.0901961, 0.423529, 0.0313725],
[0.0941176, 0.427451, 0.0313725],[0.0941176, 0.431373, 0.0313725],[0.0941176, 0.435294, 0.0313725],[0.0941176, 0.439216, 0.0313725],
[0.0980392, 0.443137, 0.0313725],[0.0980392, 0.447059, 0.0352941],[0.0980392, 0.45098, 0.0352941],[0.0980392, 0.454902, 0.0352941],
[0.101961, 0.458824, 0.0352941],[0.101961, 0.462745, 0.0352941],[0.101961, 0.466667, 0.0352941],[0.101961, 0.470588, 0.0352941],
[0.105882, 0.47451, 0.0352941],[0.105882, 0.478431, 0.0352941],[0.105882, 0.482353, 0.0392157],[0.105882, 0.486275, 0.0392157],
[0.109804, 0.490196, 0.0392157],[0.109804, 0.494118, 0.0392157],[0.109804, 0.498039, 0.0392157],[0.109804, 0.501961, 0.0392157],
[0.113725, 0.505882, 0.0392157],[0.113725, 0.509804, 0.0392157],[0.113725, 0.513725, 0.0431373],[0.117647, 0.517647, 0.0431373],
[0.117647, 0.521569, 0.0431373],[0.117647, 0.52549, 0.0431373],[0.117647, 0.529412, 0.0431373],[0.121569, 0.533333, 0.0431373],
[0.121569, 0.537255, 0.0431373],[0.121569, 0.541176, 0.0431373],[0.121569, 0.545098, 0.0431373],[0.12549, 0.54902, 0.0470588],
[0.12549, 0.552941, 0.0470588],[0.12549, 0.556863, 0.0470588],[0.12549, 0.560784, 0.0470588],[0.129412, 0.564706, 0.0470588],
[0.129412, 0.568627, 0.0470588],[0.129412, 0.572549, 0.0470588],[0.129412, 0.576471, 0.0470588],[0.133333, 0.584314, 0.0509804],
[0.133333, 0.588235, 0.0509804],[0.133333, 0.592157, 0.0509804],[0.133333, 0.596078, 0.0509804],[0.137255, 0.6, 0.0509804],
[0.137255, 0.603922, 0.0509804],[0.137255, 0.607843, 0.0509804],[0.137255, 0.611765, 0.0509804],[0.141176, 0.615686, 0.0509804],
[0.141176, 0.619608, 0.054902],[0.141176, 0.623529, 0.054902],[0.145098, 0.627451, 0.054902],[0.145098, 0.631373, 0.054902],
[0.145098, 0.635294, 0.054902],[0.145098, 0.639216, 0.054902],[0.14902, 0.643137, 0.054902],[0.14902, 0.647059, 0.054902],
[0.14902, 0.65098, 0.054902],[0.14902, 0.654902, 0.0588235],[0.152941, 0.658824, 0.0588235],[0.152941, 0.662745, 0.0588235],
[0.152941, 0.666667, 0.0588235],[0.152941, 0.67451, 0.0588235],[0.156863, 0.678431, 0.0588235],[0.156863, 0.682353, 0.0588235],
[0.156863, 0.686275, 0.0588235],[0.156863, 0.690196, 0.0627451],[0.160784, 0.694118, 0.0627451],[0.160784, 0.698039, 0.0627451],
[0.160784, 0.701961, 0.0627451],[0.164706, 0.705882, 0.0627451],[0.164706, 0.709804, 0.0627451],[0.164706, 0.713725, 0.0627451],
[0.164706, 0.717647, 0.0627451],[0.168627, 0.721569, 0.0627451],[0.168627, 0.72549, 0.0666667],[0.168627, 0.729412, 0.0666667],
[0.168627, 0.737255, 0.0666667],[0.172549, 0.741176, 0.0666667],[0.172549, 0.745098, 0.0666667],[0.172549, 0.74902, 0.0666667],
[0.172549, 0.752941, 0.0666667],[0.176471, 0.756863, 0.0666667],[0.176471, 0.760784, 0.0666667],[0.176471, 0.764706, 0.0705882],
[0.180392, 0.768627, 0.0705882],[0.180392, 0.772549, 0.0705882],[0.180392, 0.776471, 0.0705882],[0.180392, 0.780392, 0.0705882],
[0.184314, 0.788235, 0.0705882],[0.184314, 0.792157, 0.0705882],[0.184314, 0.796078, 0.0705882],[0.184314, 0.8, 0.0745098],
[0.188235, 0.803922, 0.0745098],[0.188235, 0.807843, 0.0745098],[0.188235, 0.811765, 0.0745098],[0.192157, 0.815686, 0.0745098],
[0.192157, 0.819608, 0.0745098],[0.192157, 0.823529, 0.0745098],[0.192157, 0.827451, 0.0745098],[0.196078, 0.835294, 0.0745098],
[0.196078, 0.839216, 0.0784314],[0.196078, 0.843137, 0.0784314],[0.196078, 0.847059, 0.0784314],[0.2, 0.85098, 0.0784314],
[0.2, 0.854902, 0.0784314],[0.2, 0.858824, 0.0784314],[0.203922, 0.862745, 0.0784314],[0.203922, 0.866667, 0.0784314],
[0.203922, 0.87451, 0.0784314],[0.203922, 0.878431, 0.0823529],[0.207843, 0.882353, 0.0823529],[0.207843, 0.886275, 0.0823529],
[0.207843, 0.890196, 0.0823529],[0.207843, 0.894118, 0.0823529],[0.211765, 0.898039, 0.0823529],[0.215686, 0.901961, 0.0823529],
[0.223529, 0.905882, 0.0823529],[0.231373, 0.909804, 0.0823529],[0.243137, 0.913725, 0.0823529],[0.254902, 0.917647, 0.0862745],
[0.266667, 0.921569, 0.0862745],[0.282353, 0.92549, 0.0862745],[0.294118, 0.929412, 0.0862745],[0.309804, 0.933333, 0.0862745],
[0.32549, 0.937255, 0.0862745],[0.341176, 0.941176, 0.0862745],[0.360784, 0.941176, 0.0862745],[0.376471, 0.945098, 0.0862745],
[0.392157, 0.94902, 0.0862745],[0.407843, 0.952941, 0.0862745],[0.423529, 0.952941, 0.0862745],[0.443137, 0.956863, 0.0823529],
[0.458824, 0.960784, 0.0823529],[0.47451, 0.964706, 0.0823529],[0.490196, 0.964706, 0.0823529],[0.505882, 0.968627, 0.0823529],
[0.52549, 0.972549, 0.0823529],[0.541176, 0.972549, 0.0823529],[0.556863, 0.976471, 0.0823529],[0.572549, 0.976471, 0.0823529],
[0.588235, 0.980392, 0.0823529],[0.603922, 0.984314, 0.0823529],[0.623529, 0.984314, 0.0823529],[0.639216, 0.988235, 0.0823529],
[0.654902, 0.988235, 0.0823529],[0.670588, 0.992157, 0.0823529],[0.686275, 0.992157, 0.0823529],[0.701961, 0.996078, 0.0823529],
[0.717647, 0.996078, 0.0823529],[0.733333, 1, 0.0823529],[0.752941, 1, 0.0823529],[0.768627, 1, 0.0823529],
[0.784314, 1, 0.0823529],[0.8, 1, 0.0823529],[0.815686, 1, 0.0823529],[0.831373, 1, 0.0823529],
[0.847059, 1, 0.0823529] 
], 
[
 [0, 0, 0],[0.00392157, 0.00392157, 0.00392157],
[0.0117647, 0.0117647, 0.0117647],[0.0156863, 0.0156863, 0.0156863],[0.0235294, 0.0235294, 0.0235294],[0.027451, 0.027451, 0.027451],
[0.0352941, 0.0352941, 0.0352941],[0.0392157, 0.0392157, 0.0392157],[0.0431373, 0.0431373, 0.0431373],[0.0509804, 0.0509804, 0.0509804],
[0.054902, 0.054902, 0.054902],[0.0588235, 0.0588235, 0.0588235],[0.0627451, 0.0627451, 0.0627451],[0.0666667, 0.0666667, 0.0666667],
[0.0705882, 0.0705882, 0.0705882],[0.0745098, 0.0745098, 0.0745098],[0.0784314, 0.0784314, 0.0784314],[0.0823529, 0.0823529, 0.0823529],
[0.0862745, 0.0862745, 0.0862745],[0.0862745, 0.0862745, 0.0862745],[0.0901961, 0.0901961, 0.0901961],[0.0941176, 0.0941176, 0.0941176],
[0.0980392, 0.0980392, 0.0980392],[0.101961, 0.101961, 0.101961],[0.101961, 0.101961, 0.101961],[0.105882, 0.105882, 0.105882],
[0.109804, 0.109804, 0.109804],[0.113725, 0.113725, 0.113725],[0.113725, 0.113725, 0.113725],[0.117647, 0.117647, 0.117647],
[0.121569, 0.121569, 0.121569],[0.12549, 0.12549, 0.12549],[0.129412, 0.129412, 0.129412],[0.129412, 0.129412, 0.129412],
[0.133333, 0.133333, 0.133333],[0.137255, 0.137255, 0.137255],[0.141176, 0.141176, 0.141176],[0.145098, 0.145098, 0.145098],
[0.145098, 0.145098, 0.145098],[0.14902, 0.14902, 0.14902],[0.152941, 0.152941, 0.152941],[0.156863, 0.156863, 0.156863],
[0.160784, 0.160784, 0.160784],[0.164706, 0.164706, 0.164706],[0.164706, 0.164706, 0.164706],[0.168627, 0.168627, 0.168627],
[0.172549, 0.172549, 0.172549],[0.176471, 0.176471, 0.176471],[0.180392, 0.180392, 0.180392],[0.184314, 0.184314, 0.184314],
[0.184314, 0.184314, 0.184314],[0.188235, 0.188235, 0.188235],[0.192157, 0.192157, 0.192157],[0.196078, 0.196078, 0.196078],
[0.2, 0.2, 0.2],[0.203922, 0.203922, 0.203922],[0.207843, 0.207843, 0.207843],[0.207843, 0.207843, 0.207843],
[0.211765, 0.211765, 0.211765],[0.215686, 0.215686, 0.215686],[0.219608, 0.219608, 0.219608],[0.223529, 0.223529, 0.223529],
[0.227451, 0.227451, 0.227451],[0.231373, 0.231373, 0.231373],[0.231373, 0.235294, 0.235294],[0.235294, 0.235294, 0.235294],
[0.239216, 0.239216, 0.239216],[0.243137, 0.243137, 0.243137],[0.247059, 0.247059, 0.247059],[0.25098, 0.25098, 0.25098],
[0.254902, 0.254902, 0.254902],[0.258824, 0.258824, 0.258824],[0.262745, 0.262745, 0.262745],[0.262745, 0.262745, 0.262745],
[0.266667, 0.266667, 0.266667],[0.270588, 0.270588, 0.270588],[0.27451, 0.27451, 0.27451],[0.278431, 0.278431, 0.278431],
[0.282353, 0.282353, 0.282353],[0.286275, 0.286275, 0.286275],[0.290196, 0.290196, 0.290196],[0.294118, 0.294118, 0.294118],
[0.298039, 0.298039, 0.298039],[0.301961, 0.301961, 0.301961],[0.301961, 0.301961, 0.301961],[0.305882, 0.305882, 0.305882],
[0.309804, 0.309804, 0.309804],[0.313725, 0.313725, 0.313725],[0.317647, 0.317647, 0.317647],[0.321569, 0.321569, 0.321569],
[0.32549, 0.32549, 0.32549],[0.329412, 0.329412, 0.329412],[0.333333, 0.333333, 0.333333],[0.337255, 0.337255, 0.337255],
[0.341176, 0.341176, 0.341176],[0.345098, 0.345098, 0.345098],[0.34902, 0.34902, 0.34902],[0.352941, 0.352941, 0.352941],
[0.352941, 0.352941, 0.352941],[0.356863, 0.356863, 0.356863],[0.360784, 0.360784, 0.360784],[0.364706, 0.364706, 0.364706],
[0.368627, 0.368627, 0.368627],[0.372549, 0.372549, 0.372549],[0.376471, 0.376471, 0.376471],[0.380392, 0.380392, 0.380392],
[0.384314, 0.384314, 0.384314],[0.388235, 0.388235, 0.388235],[0.392157, 0.392157, 0.392157],[0.396078, 0.396078, 0.396078],
[0.4, 0.4, 0.4],[0.403922, 0.403922, 0.403922],[0.407843, 0.407843, 0.407843],[0.411765, 0.411765, 0.411765],
[0.415686, 0.415686, 0.415686],[0.419608, 0.419608, 0.419608],[0.423529, 0.423529, 0.423529],[0.427451, 0.427451, 0.427451],
[0.431373, 0.431373, 0.431373],[0.435294, 0.435294, 0.435294],[0.439216, 0.439216, 0.439216],[0.443137, 0.443137, 0.443137],
[0.443137, 0.447059, 0.447059],[0.447059, 0.447059, 0.447059],[0.45098, 0.45098, 0.45098],[0.454902, 0.454902, 0.454902],
[0.458824, 0.458824, 0.458824],[0.462745, 0.462745, 0.462745],[0.466667, 0.466667, 0.466667],[0.470588, 0.470588, 0.470588],
[0.47451, 0.47451, 0.47451],[0.478431, 0.478431, 0.478431],[0.482353, 0.482353, 0.482353],[0.486275, 0.486275, 0.486275],
[0.490196, 0.490196, 0.490196],[0.494118, 0.494118, 0.494118],[0.498039, 0.498039, 0.498039],[0.501961, 0.501961, 0.501961],
[0.505882, 0.505882, 0.505882],[0.509804, 0.509804, 0.509804],[0.513725, 0.513725, 0.513725],[0.517647, 0.517647, 0.517647],
[0.521569, 0.521569, 0.521569],[0.52549, 0.52549, 0.52549],[0.529412, 0.529412, 0.529412],[0.533333, 0.533333, 0.533333],
[0.537255, 0.537255, 0.537255],[0.541176, 0.541176, 0.541176],[0.545098, 0.545098, 0.545098],[0.54902, 0.54902, 0.54902],
[0.552941, 0.552941, 0.552941],[0.556863, 0.560784, 0.560784],[0.564706, 0.564706, 0.564706],[0.568627, 0.568627, 0.568627],
[0.572549, 0.572549, 0.572549],[0.576471, 0.576471, 0.576471],[0.580392, 0.580392, 0.580392],[0.584314, 0.584314, 0.584314],
[0.588235, 0.588235, 0.588235],[0.592157, 0.592157, 0.592157],[0.596078, 0.596078, 0.596078],[0.6, 0.6, 0.6],
[0.603922, 0.603922, 0.603922],[0.607843, 0.607843, 0.607843],[0.611765, 0.611765, 0.611765],[0.615686, 0.615686, 0.615686],
[0.619608, 0.619608, 0.619608],[0.623529, 0.623529, 0.623529],[0.627451, 0.627451, 0.627451],[0.631373, 0.631373, 0.631373],
[0.635294, 0.635294, 0.635294],[0.639216, 0.639216, 0.639216],[0.643137, 0.643137, 0.643137],[0.647059, 0.647059, 0.647059],
[0.65098, 0.65098, 0.65098],[0.654902, 0.654902, 0.654902],[0.658824, 0.658824, 0.658824],[0.662745, 0.662745, 0.662745],
[0.666667, 0.670588, 0.670588],[0.67451, 0.67451, 0.67451],[0.678431, 0.678431, 0.678431],[0.682353, 0.682353, 0.682353],
[0.686275, 0.686275, 0.686275],[0.690196, 0.690196, 0.690196],[0.694118, 0.694118, 0.694118],[0.698039, 0.698039, 0.698039],
[0.701961, 0.701961, 0.701961],[0.705882, 0.705882, 0.705882],[0.709804, 0.709804, 0.709804],[0.713725, 0.713725, 0.713725],
[0.717647, 0.717647, 0.717647],[0.721569, 0.721569, 0.721569],[0.72549, 0.72549, 0.72549],[0.729412, 0.729412, 0.729412],
[0.737255, 0.737255, 0.737255],[0.741176, 0.741176, 0.741176],[0.745098, 0.745098, 0.745098],[0.74902, 0.74902, 0.74902],
[0.752941, 0.752941, 0.752941],[0.756863, 0.756863, 0.756863],[0.760784, 0.760784, 0.760784],[0.764706, 0.764706, 0.764706],
[0.768627, 0.768627, 0.768627],[0.772549, 0.772549, 0.772549],[0.776471, 0.776471, 0.776471],[0.780392, 0.780392, 0.780392],
[0.784314, 0.788235, 0.788235],[0.792157, 0.792157, 0.792157],[0.796078, 0.796078, 0.796078],[0.8, 0.8, 0.8],
[0.803922, 0.803922, 0.803922],[0.807843, 0.807843, 0.807843],[0.811765, 0.811765, 0.811765],[0.815686, 0.815686, 0.815686],
[0.819608, 0.819608, 0.819608],[0.823529, 0.823529, 0.823529],[0.827451, 0.827451, 0.827451],[0.831373, 0.835294, 0.835294],
[0.839216, 0.839216, 0.839216],[0.843137, 0.843137, 0.843137],[0.847059, 0.847059, 0.847059],[0.85098, 0.85098, 0.85098],
[0.854902, 0.854902, 0.854902],[0.858824, 0.858824, 0.858824],[0.862745, 0.862745, 0.862745],[0.866667, 0.866667, 0.866667],
[0.870588, 0.870588, 0.870588],[0.878431, 0.878431, 0.878431],[0.882353, 0.882353, 0.882353],[0.886275, 0.886275, 0.886275],
[0.890196, 0.890196, 0.890196],[0.894118, 0.894118, 0.894118],[0.898039, 0.898039, 0.898039],[0.901961, 0.901961, 0.901961],
[0.905882, 0.905882, 0.905882],[0.909804, 0.909804, 0.909804],[0.917647, 0.917647, 0.917647],[0.921569, 0.921569, 0.921569],
[0.92549, 0.92549, 0.92549],[0.929412, 0.929412, 0.929412],[0.933333, 0.933333, 0.933333],[0.937255, 0.937255, 0.937255],
[0.941176, 0.941176, 0.941176],[0.945098, 0.945098, 0.945098],[0.952941, 0.952941, 0.952941],[0.956863, 0.956863, 0.956863],
[0.960784, 0.960784, 0.960784],[0.964706, 0.964706, 0.964706],[0.968627, 0.968627, 0.968627],[0.972549, 0.972549, 0.972549],
[0.976471, 0.976471, 0.976471],[0.980392, 0.980392, 0.980392],[0.988235, 0.988235, 0.988235],[0.992157, 0.992157, 0.992157],
[0.996078, 0.996078, 0.996078],[1, 1, 1] 
], 
[
[0.105882, 0.105882, 0.105882],
[0.109804, 0.109804, 0.109804],[0.113725, 0.113725, 0.113725],[0.113725, 0.113725, 0.113725],[0.117647, 0.117647, 0.117647],
[0.121569, 0.121569, 0.121569],[0.121569, 0.121569, 0.121569],[0.12549, 0.12549, 0.12549],[0.129412, 0.129412, 0.129412],
[0.133333, 0.133333, 0.133333],[0.133333, 0.133333, 0.133333],[0.137255, 0.137255, 0.137255],[0.141176, 0.141176, 0.141176],
[0.141176, 0.141176, 0.141176],[0.145098, 0.145098, 0.145098],[0.14902, 0.14902, 0.14902],[0.14902, 0.14902, 0.14902],
[0.152941, 0.152941, 0.152941],[0.156863, 0.156863, 0.156863],[0.156863, 0.156863, 0.156863],[0.160784, 0.160784, 0.160784],
[0.164706, 0.164706, 0.164706],[0.168627, 0.168627, 0.168627],[0.168627, 0.168627, 0.168627],[0.172549, 0.172549, 0.172549],
[0.176471, 0.176471, 0.176471],[0.176471, 0.176471, 0.176471],[0.180392, 0.180392, 0.180392],[0.184314, 0.184314, 0.184314],
[0.188235, 0.188235, 0.188235],[0.188235, 0.188235, 0.188235],[0.192157, 0.192157, 0.192157],[0.196078, 0.196078, 0.196078],
[0.196078, 0.196078, 0.196078],[0.2, 0.2, 0.2],[0.203922, 0.203922, 0.203922],[0.207843, 0.207843, 0.207843],
[0.207843, 0.207843, 0.207843],[0.211765, 0.211765, 0.211765],[0.215686, 0.215686, 0.215686],[0.219608, 0.219608, 0.219608],
[0.219608, 0.219608, 0.219608],[0.223529, 0.223529, 0.223529],[0.227451, 0.227451, 0.227451],[0.231373, 0.231373, 0.231373],
[0.231373, 0.231373, 0.231373],[0.235294, 0.235294, 0.235294],[0.239216, 0.239216, 0.239216],[0.243137, 0.243137, 0.243137],
[0.243137, 0.243137, 0.243137],[0.247059, 0.247059, 0.247059],[0.25098, 0.25098, 0.25098],[0.254902, 0.254902, 0.254902],
[0.254902, 0.254902, 0.254902],[0.258824, 0.258824, 0.258824],[0.262745, 0.262745, 0.262745],[0.266667, 0.266667, 0.266667],
[0.266667, 0.266667, 0.266667],[0.270588, 0.270588, 0.270588],[0.27451, 0.27451, 0.27451],[0.278431, 0.278431, 0.278431],
[0.278431, 0.278431, 0.278431],[0.282353, 0.282353, 0.282353],[0.286275, 0.286275, 0.286275],[0.290196, 0.290196, 0.290196],
[0.290196, 0.290196, 0.290196],[0.294118, 0.294118, 0.294118],[0.298039, 0.298039, 0.298039],[0.301961, 0.301961, 0.301961],
[0.305882, 0.305882, 0.305882],[0.305882, 0.305882, 0.305882],[0.309804, 0.309804, 0.309804],[0.313725, 0.313725, 0.313725],
[0.317647, 0.317647, 0.317647],[0.317647, 0.321569, 0.321569],[0.321569, 0.321569, 0.321569],[0.32549, 0.32549, 0.32549],
[0.329412, 0.329412, 0.329412],[0.333333, 0.333333, 0.333333],[0.333333, 0.333333, 0.333333],[0.337255, 0.337255, 0.337255],
[0.341176, 0.341176, 0.341176],[0.345098, 0.345098, 0.345098],[0.34902, 0.34902, 0.34902],[0.34902, 0.34902, 0.34902],
[0.352941, 0.352941, 0.352941],[0.356863, 0.356863, 0.356863],[0.360784, 0.360784, 0.360784],[0.364706, 0.364706, 0.364706],
[0.364706, 0.364706, 0.364706],[0.368627, 0.368627, 0.368627],[0.372549, 0.372549, 0.372549],[0.376471, 0.376471, 0.376471],
[0.380392, 0.380392, 0.380392],[0.380392, 0.380392, 0.380392],[0.384314, 0.384314, 0.384314],[0.388235, 0.388235, 0.388235],
[0.392157, 0.392157, 0.392157],[0.396078, 0.396078, 0.396078],[0.4, 0.4, 0.4],[0.4, 0.4, 0.4],
[0.403922, 0.403922, 0.403922],[0.407843, 0.407843, 0.407843],[0.411765, 0.411765, 0.411765],[0.415686, 0.415686, 0.415686],
[0.415686, 0.415686, 0.415686],[0.419608, 0.419608, 0.419608],[0.423529, 0.423529, 0.423529],[0.427451, 0.427451, 0.427451],
[0.431373, 0.431373, 0.431373],[0.435294, 0.435294, 0.435294],[0.435294, 0.435294, 0.435294],[0.439216, 0.439216, 0.439216],
[0.443137, 0.443137, 0.443137],[0.447059, 0.447059, 0.447059],[0.45098, 0.45098, 0.45098],[0.454902, 0.454902, 0.454902],
[0.454902, 0.454902, 0.454902],[0.458824, 0.458824, 0.458824],[0.462745, 0.462745, 0.462745],[0.466667, 0.466667, 0.466667],
[0.470588, 0.470588, 0.470588],[0.47451, 0.47451, 0.47451],[0.47451, 0.47451, 0.47451],[0.478431, 0.478431, 0.478431],
[0.482353, 0.482353, 0.482353],[0.486275, 0.486275, 0.486275],[0.490196, 0.490196, 0.490196],[0.494118, 0.494118, 0.494118],
[0.494118, 0.498039, 0.498039],[0.498039, 0.498039, 0.498039],[0.501961, 0.501961, 0.501961],[0.505882, 0.505882, 0.505882],
[0.509804, 0.509804, 0.509804],[0.513725, 0.513725, 0.513725],[0.517647, 0.517647, 0.517647],[0.517647, 0.517647, 0.517647],
[0.521569, 0.521569, 0.521569],[0.52549, 0.52549, 0.52549],[0.529412, 0.529412, 0.529412],[0.533333, 0.533333, 0.533333],
[0.537255, 0.537255, 0.537255],[0.541176, 0.541176, 0.541176],[0.541176, 0.541176, 0.541176],[0.545098, 0.545098, 0.545098],
[0.54902, 0.54902, 0.54902],[0.552941, 0.552941, 0.552941],[0.556863, 0.556863, 0.556863],[0.560784, 0.560784, 0.560784],
[0.564706, 0.564706, 0.564706],[0.568627, 0.568627, 0.568627],[0.568627, 0.568627, 0.568627],[0.572549, 0.572549, 0.572549],
[0.576471, 0.576471, 0.576471],[0.580392, 0.580392, 0.580392],[0.584314, 0.584314, 0.584314],[0.588235, 0.588235, 0.588235],
[0.592157, 0.592157, 0.592157],[0.596078, 0.596078, 0.596078],[0.596078, 0.596078, 0.596078],[0.6, 0.6, 0.6],
[0.603922, 0.603922, 0.603922],[0.607843, 0.607843, 0.607843],[0.611765, 0.611765, 0.611765],[0.615686, 0.615686, 0.615686],
[0.619608, 0.619608, 0.619608],[0.623529, 0.623529, 0.623529],[0.623529, 0.623529, 0.623529],[0.627451, 0.627451, 0.627451],
[0.631373, 0.631373, 0.631373],[0.635294, 0.635294, 0.635294],[0.639216, 0.639216, 0.639216],[0.643137, 0.643137, 0.643137],
[0.647059, 0.647059, 0.647059],[0.65098, 0.65098, 0.65098],[0.654902, 0.654902, 0.654902],[0.654902, 0.654902, 0.654902],
[0.658824, 0.658824, 0.658824],[0.662745, 0.662745, 0.662745],[0.666667, 0.666667, 0.666667],[0.670588, 0.670588, 0.670588],
[0.67451, 0.67451, 0.67451],[0.678431, 0.678431, 0.678431],[0.682353, 0.682353, 0.682353],[0.686275, 0.686275, 0.686275],
[0.690196, 0.690196, 0.690196],[0.690196, 0.690196, 0.690196],[0.694118, 0.694118, 0.694118],[0.698039, 0.698039, 0.698039],
[0.701961, 0.701961, 0.701961],[0.705882, 0.705882, 0.705882],[0.709804, 0.709804, 0.709804],[0.713725, 0.713725, 0.713725],
[0.717647, 0.717647, 0.717647],[0.721569, 0.721569, 0.721569],[0.72549, 0.72549, 0.72549],[0.72549, 0.729412, 0.729412],
[0.729412, 0.729412, 0.729412],[0.733333, 0.733333, 0.733333],[0.737255, 0.737255, 0.737255],[0.741176, 0.741176, 0.741176],
[0.745098, 0.745098, 0.745098],[0.74902, 0.74902, 0.74902],[0.752941, 0.752941, 0.752941],[0.756863, 0.756863, 0.756863],
[0.760784, 0.760784, 0.760784],[0.764706, 0.764706, 0.764706],[0.768627, 0.768627, 0.768627],[0.768627, 0.768627, 0.768627],
[0.772549, 0.772549, 0.772549],[0.776471, 0.776471, 0.776471],[0.780392, 0.780392, 0.780392],[0.784314, 0.784314, 0.784314],
[0.788235, 0.788235, 0.788235],[0.792157, 0.792157, 0.792157],[0.796078, 0.796078, 0.796078],[0.8, 0.8, 0.8],
[0.803922, 0.803922, 0.803922],[0.807843, 0.807843, 0.807843],[0.811765, 0.811765, 0.811765],[0.815686, 0.815686, 0.815686],
[0.815686, 0.819608, 0.819608],[0.819608, 0.819608, 0.819608],[0.823529, 0.823529, 0.823529],[0.827451, 0.827451, 0.827451],
[0.831373, 0.831373, 0.831373],[0.835294, 0.835294, 0.835294],[0.839216, 0.839216, 0.839216],[0.843137, 0.843137, 0.843137],
[0.847059, 0.847059, 0.847059],[0.85098, 0.85098, 0.85098],[0.854902, 0.854902, 0.854902],[0.858824, 0.858824, 0.858824],
[0.862745, 0.862745, 0.862745],[0.866667, 0.866667, 0.866667],[0.870588, 0.870588, 0.870588],[0.87451, 0.87451, 0.87451],
[0.87451, 0.878431, 0.878431],[0.878431, 0.878431, 0.878431],[0.882353, 0.882353, 0.882353],[0.886275, 0.886275, 0.886275],
[0.890196, 0.890196, 0.890196],[0.894118, 0.894118, 0.894118],[0.898039, 0.898039, 0.898039],[0.901961, 0.901961, 0.901961],
[0.905882, 0.905882, 0.905882],[0.909804, 0.909804, 0.909804],[0.913725, 0.913725, 0.913725],[0.917647, 0.917647, 0.917647],
[0.921569, 0.921569, 0.921569],[0.92549, 0.92549, 0.92549],[0.929412, 0.929412, 0.929412],[0.933333, 0.933333, 0.933333],
[0.937255, 0.937255, 0.937255],[0.941176, 0.941176, 0.941176],[0.945098, 0.945098, 0.945098] 
], 
[
[0.0666667, 0.0666667, 0.0666667],[0.0784314, 0.0666667, 0.0666667],[0.0901961, 0.0666667, 0.0627451],[0.101961, 0.0666667, 0.0627451],
[0.113725, 0.0705882, 0.0627451],[0.121569, 0.0705882, 0.0588235],[0.129412, 0.0705882, 0.0588235],[0.137255, 0.0705882, 0.0588235],
[0.14902, 0.0705882, 0.054902],[0.156863, 0.0705882, 0.054902],[0.164706, 0.0705882, 0.054902],[0.172549, 0.0705882, 0.0509804],
[0.180392, 0.0666667, 0.0509804],[0.188235, 0.0666667, 0.0509804],[0.196078, 0.0666667, 0.0470588],[0.207843, 0.0666667, 0.0470588],
[0.215686, 0.0627451, 0.0470588],[0.223529, 0.0627451, 0.0431373],[0.231373, 0.0627451, 0.0431373],[0.239216, 0.0588235, 0.0392157],
[0.247059, 0.0588235, 0.0392157],[0.254902, 0.0588235, 0.0392157],[0.262745, 0.054902, 0.0352941],[0.270588, 0.054902, 0.0352941],
[0.278431, 0.0509804, 0.0352941],[0.286275, 0.0470588, 0.0313725],[0.294118, 0.0470588, 0.0313725],[0.301961, 0.0431373, 0.0313725],
[0.309804, 0.0392157, 0.027451],[0.317647, 0.0392157, 0.027451],[0.321569, 0.0352941, 0.027451],[0.329412, 0.0313725, 0.027451],
[0.337255, 0.027451, 0.0235294],[0.345098, 0.027451, 0.0235294],[0.352941, 0.0235294, 0.0235294],[0.360784, 0.0196078, 0.0196078],
[0.364706, 0.0196078, 0.0196078],[0.372549, 0.0156863, 0.0196078],[0.380392, 0.0117647, 0.0196078],[0.384314, 0.0117647, 0.0156863],
[0.392157, 0.00784314, 0.0156863],[0.4, 0.00784314, 0.0156863],[0.403922, 0.00784314, 0.0156863],[0.411765, 0.00784314, 0.0156863],
[0.415686, 0.00784314, 0.0156863],[0.423529, 0.00784314, 0.0117647],[0.427451, 0.00784314, 0.0117647],[0.435294, 0.00784314, 0.0117647],
[0.439216, 0.00784314, 0.0117647],[0.447059, 0.00784314, 0.0117647],[0.45098, 0.00784314, 0.0117647],[0.458824, 0.00784314, 0.0117647],
[0.462745, 0.00784314, 0.0117647],[0.470588, 0.00784314, 0.00784314],[0.47451, 0.00784314, 0.00784314],[0.482353, 0.00784314, 0.00784314],
[0.486275, 0.00784314, 0.00784314],[0.494118, 0.00784314, 0.00784314],[0.498039, 0.00784314, 0.00784314],[0.505882, 0.00784314, 0.00784314],
[0.509804, 0.00784314, 0.00784314],[0.517647, 0.00784314, 0.00784314],[0.521569, 0.00784314, 0.00784314],[0.529412, 0.0117647, 0.00784314],
[0.533333, 0.0117647, 0.00784314],[0.541176, 0.0117647, 0.00784314],[0.545098, 0.0117647, 0.00784314],[0.552941, 0.0117647, 0.00784314],
[0.556863, 0.0117647, 0.00784314],[0.564706, 0.0117647, 0.00784314],[0.568627, 0.0117647, 0.00784314],[0.576471, 0.0117647, 0.0117647],
[0.580392, 0.0156863, 0.0117647],[0.588235, 0.0156863, 0.0117647],[0.592157, 0.0156863, 0.0117647],[0.6, 0.0156863, 0.0117647],
[0.603922, 0.0156863, 0.0117647],[0.611765, 0.0156863, 0.0117647],[0.615686, 0.0196078, 0.0117647],[0.623529, 0.0196078, 0.0117647],
[0.631373, 0.0196078, 0.0117647],[0.635294, 0.0196078, 0.0117647],[0.643137, 0.0196078, 0.0117647],[0.647059, 0.0196078, 0.0117647],
[0.654902, 0.0235294, 0.0117647],[0.658824, 0.0235294, 0.0117647],[0.666667, 0.0235294, 0.0117647],[0.670588, 0.0235294, 0.0117647],
[0.678431, 0.0235294, 0.0117647],[0.686275, 0.027451, 0.0117647],[0.690196, 0.027451, 0.0117647],[0.698039, 0.027451, 0.0117647],
[0.701961, 0.027451, 0.0117647],[0.709804, 0.027451, 0.0117647],[0.717647, 0.0313725, 0.0117647],[0.721569, 0.0313725, 0.0117647],
[0.729412, 0.0313725, 0.0117647],[0.733333, 0.0313725, 0.0117647],[0.741176, 0.0352941, 0.0117647],[0.74902, 0.0352941, 0.0117647],
[0.752941, 0.0352941, 0.0117647],[0.760784, 0.0352941, 0.0117647],[0.764706, 0.0392157, 0.0117647],[0.772549, 0.0392157, 0.0117647],
[0.776471, 0.0431373, 0.0117647],[0.784314, 0.0431373, 0.0117647],[0.792157, 0.0470588, 0.0117647],[0.796078, 0.0509804, 0.0117647],
[0.803922, 0.0509804, 0.0117647],[0.807843, 0.054902, 0.0117647],[0.815686, 0.0588235, 0.0117647],[0.819608, 0.0588235, 0.0117647],
[0.827451, 0.0627451, 0.0117647],[0.831373, 0.0666667, 0.0117647],[0.839216, 0.0705882, 0.0156863],[0.843137, 0.0745098, 0.0156863],
[0.85098, 0.0784314, 0.0156863],[0.854902, 0.0823529, 0.0156863],[0.862745, 0.0862745, 0.0156863],[0.866667, 0.0862745, 0.0156863],
[0.87451, 0.0901961, 0.0156863],[0.878431, 0.0941176, 0.0156863],[0.886275, 0.0980392, 0.0156863],[0.890196, 0.101961, 0.0156863],
[0.898039, 0.109804, 0.0156863],[0.901961, 0.113725, 0.0196078],[0.905882, 0.117647, 0.0196078],[0.913725, 0.121569, 0.0196078],
[0.917647, 0.12549, 0.0196078],[0.92549, 0.129412, 0.0196078],[0.929412, 0.133333, 0.0196078],[0.933333, 0.141176, 0.0235294],
[0.937255, 0.145098, 0.0235294],[0.945098, 0.152941, 0.0235294],[0.94902, 0.160784, 0.0235294],[0.952941, 0.164706, 0.0235294],
[0.956863, 0.172549, 0.0235294],[0.960784, 0.180392, 0.027451],[0.964706, 0.188235, 0.027451],[0.968627, 0.196078, 0.027451],
[0.972549, 0.203922, 0.027451],[0.972549, 0.211765, 0.027451],[0.976471, 0.219608, 0.0313725],[0.980392, 0.227451, 0.0313725],
[0.984314, 0.239216, 0.0313725],[0.984314, 0.247059, 0.0313725],[0.988235, 0.254902, 0.0313725],[0.988235, 0.262745, 0.0352941],
[0.992157, 0.270588, 0.0352941],[0.992157, 0.282353, 0.0352941],[0.992157, 0.290196, 0.0392157],[0.996078, 0.298039, 0.0392157],
[0.996078, 0.309804, 0.0392157],[0.996078, 0.317647, 0.0431373],[0.996078, 0.32549, 0.0431373],[1, 0.337255, 0.0431373],
[1, 0.345098, 0.0470588],[1, 0.352941, 0.0470588],[1, 0.364706, 0.0470588],[1, 0.372549, 0.0509804],
[1, 0.380392, 0.0509804],[1, 0.388235, 0.0509804],[1, 0.396078, 0.054902],[1, 0.403922, 0.054902],
[1, 0.415686, 0.054902],[1, 0.423529, 0.0588235],[1, 0.431373, 0.0588235],[1, 0.439216, 0.0588235],
[1, 0.447059, 0.0627451],[1, 0.454902, 0.0627451],[1, 0.458824, 0.0666667],[0.996078, 0.466667, 0.0666667],
[0.996078, 0.47451, 0.0666667],[0.996078, 0.482353, 0.0705882],[0.996078, 0.490196, 0.0705882],[0.996078, 0.498039, 0.0705882],
[0.996078, 0.505882, 0.0745098],[0.996078, 0.513725, 0.0745098],[0.996078, 0.517647, 0.0745098],[0.996078, 0.52549, 0.0784314],
[0.996078, 0.533333, 0.0784314],[0.996078, 0.541176, 0.0784314],[0.992157, 0.54902, 0.0823529],[0.992157, 0.552941, 0.0823529],
[0.992157, 0.560784, 0.0823529],[0.992157, 0.568627, 0.0862745],[0.992157, 0.576471, 0.0862745],[0.992157, 0.580392, 0.0862745],
[0.988235, 0.588235, 0.0901961],[0.988235, 0.596078, 0.0901961],[0.988235, 0.603922, 0.0941176],[0.988235, 0.607843, 0.0941176],
[0.988235, 0.615686, 0.0980392],[0.984314, 0.623529, 0.0980392],[0.984314, 0.627451, 0.101961],[0.984314, 0.635294, 0.105882],
[0.980392, 0.643137, 0.105882],[0.980392, 0.65098, 0.109804],[0.980392, 0.654902, 0.113725],[0.976471, 0.662745, 0.113725],
[0.976471, 0.670588, 0.117647],[0.976471, 0.67451, 0.121569],[0.972549, 0.682353, 0.12549],[0.972549, 0.690196, 0.12549],
[0.968627, 0.694118, 0.129412],[0.968627, 0.701961, 0.133333],[0.964706, 0.709804, 0.137255],[0.964706, 0.713725, 0.141176],
[0.960784, 0.721569, 0.145098],[0.960784, 0.729412, 0.14902],[0.956863, 0.733333, 0.152941],[0.956863, 0.741176, 0.156863],
[0.952941, 0.74902, 0.160784],[0.952941, 0.752941, 0.164706],[0.94902, 0.760784, 0.164706],[0.945098, 0.768627, 0.168627],
[0.945098, 0.772549, 0.172549],[0.945098, 0.780392, 0.172549],[0.941176, 0.784314, 0.176471],[0.941176, 0.792157, 0.176471],
[0.941176, 0.796078, 0.176471],[0.941176, 0.803922, 0.180392],[0.941176, 0.807843, 0.180392],[0.937255, 0.815686, 0.180392],
[0.937255, 0.819608, 0.180392],[0.937255, 0.823529, 0.180392],[0.937255, 0.831373, 0.180392],[0.937255, 0.835294, 0.180392],
[0.937255, 0.843137, 0.176471],[0.937255, 0.847059, 0.176471],[0.937255, 0.85098, 0.176471],[0.937255, 0.858824, 0.172549],
[0.941176, 0.862745, 0.172549],[0.941176, 0.866667, 0.168627],[0.941176, 0.87451, 0.168627],[0.941176, 0.878431, 0.164706],
[0.941176, 0.882353, 0.164706],[0.941176, 0.886275, 0.160784],[0.945098, 0.894118, 0.156863],[0.945098, 0.898039, 0.152941],
[0.945098, 0.901961, 0.14902],[0.94902, 0.909804, 0.145098],[0.94902, 0.913725, 0.141176],[0.94902, 0.917647, 0.137255],
[0.94902, 0.921569, 0.133333],[0.952941, 0.929412, 0.129412],[0.952941, 0.933333, 0.121569],[0.956863, 0.937255, 0.117647],
[0.956863, 0.941176, 0.109804],[0.956863, 0.94902, 0.101961],[0.960784, 0.952941, 0.0941176],[0.960784, 0.956863, 0.0862745],
[0.964706, 0.960784, 0.0745098],[0.964706, 0.968627, 0.0666667],[0.964706, 0.972549, 0.0509804],[0.968627, 0.976471, 0.0352941] 
], 
[
[0.0666667, 0.0666667, 0.0666667],[0.0823529, 0.0666667, 0.0666667],[0.0941176, 0.0666667, 0.0627451],
[0.105882, 0.0666667, 0.0627451],[0.113725, 0.0666667, 0.0588235],[0.12549, 0.0705882, 0.0588235],[0.133333, 0.0705882, 0.0588235],
[0.141176, 0.0705882, 0.054902],[0.152941, 0.0705882, 0.054902],[0.160784, 0.0666667, 0.054902],[0.168627, 0.0666667, 0.0509804],
[0.180392, 0.0666667, 0.0509804],[0.188235, 0.0666667, 0.0509804],[0.196078, 0.0627451, 0.0470588],[0.203922, 0.0627451, 0.0470588],
[0.215686, 0.0627451, 0.0470588],[0.223529, 0.0588235, 0.0431373],[0.231373, 0.0588235, 0.0431373],[0.239216, 0.054902, 0.0431373],
[0.247059, 0.054902, 0.0392157],[0.254902, 0.0509804, 0.0392157],[0.266667, 0.0470588, 0.0392157],[0.27451, 0.0470588, 0.0352941],
[0.282353, 0.0431373, 0.0352941],[0.290196, 0.0392157, 0.0352941],[0.298039, 0.0352941, 0.0313725],[0.305882, 0.0313725, 0.0313725],
[0.313725, 0.027451, 0.0313725],[0.321569, 0.0235294, 0.027451],[0.329412, 0.0196078, 0.027451],[0.337255, 0.0156863, 0.027451],
[0.345098, 0.0117647, 0.0235294],[0.352941, 0.00784314, 0.0235294],[0.360784, 0.00392157, 0.0235294],[0.368627, 0.00392157, 0.0235294],
[0.376471, 0, 0.0196078],[0.380392, 0, 0.0196078],[0.388235, 0, 0.0196078],[0.396078, 0, 0.0196078],
[0.403922, 0, 0.0196078],[0.407843, 0, 0.0156863],[0.415686, 0, 0.0156863],[0.419608, 0, 0.0156863],
[0.427451, 0, 0.0156863],[0.431373, 0, 0.0156863],[0.439216, 0, 0.0156863],[0.443137, 0, 0.0117647],
[0.45098, 0, 0.0117647],[0.458824, 0, 0.0117647],[0.462745, 0, 0.0117647],[0.470588, 0, 0.0117647],
[0.47451, 0, 0.0117647],[0.482353, 0, 0.00784314],[0.486275, 0, 0.00784314],[0.494118, 0, 0.00784314],
[0.498039, 0, 0.00784314],[0.505882, 0, 0.00784314],[0.513725, 0, 0.00784314],[0.517647, 0, 0.00784314],
[0.52549, 0, 0.00784314],[0.529412, 0, 0.00392157],[0.537255, 0, 0.00392157],[0.541176, 0, 0.00392157],
[0.54902, 0, 0.00392157],[0.556863, 0, 0.00392157],[0.560784, 0, 0.00392157],[0.568627, 0, 0.00392157],
[0.572549, 0, 0.00392157],[0.580392, 0, 0.00392157],[0.584314, 0, 0.00392157],[0.592157, 0, 0.00392157],
[0.6, 0, 0],[0.603922, 0, 0],[0.611765, 0, 0],[0.615686, 0, 0],
[0.623529, 0, 0],[0.631373, 0, 0],[0.635294, 0, 0],[0.643137, 0, 0],
[0.647059, 0, 0],[0.654902, 0, 0],[0.662745, 0, 0],[0.666667, 0, 0],
[0.67451, 0, 0],[0.682353, 0, 0],[0.686275, 0, 0],[0.694118, 0, 0],
[0.701961, 0, 0],[0.705882, 0, 0],[0.713725, 0, 0],[0.721569, 0, 0],
[0.72549, 0, 0],[0.733333, 0, 0],[0.741176, 0, 0],[0.745098, 0, 0],
[0.752941, 0, 0],[0.760784, 0, 0],[0.764706, 0, 0],[0.772549, 0, 0],
[0.780392, 0, 0],[0.788235, 0, 0],[0.792157, 0, 0],[0.8, 0, 0],
[0.807843, 0, 0],[0.811765, 0, 0],[0.819608, 0, 0],[0.827451, 0, 0],
[0.831373, 0, 0],[0.839216, 0, 0],[0.847059, 0, 0],[0.85098, 0, 0],
[0.858824, 0, 0],[0.866667, 0, 0],[0.870588, 0, 0],[0.878431, 0, 0],
[0.882353, 0, 0],[0.890196, 0, 0],[0.898039, 0, 0],[0.901961, 0, 0],
[0.909804, 0, 0],[0.917647, 0, 0],[0.921569, 0, 0],[0.929412, 0.00784314, 0],
[0.933333, 0.0117647, 0],[0.941176, 0.0156863, 0],[0.94902, 0.0235294, 0],[0.952941, 0.0313725, 0],
[0.956863, 0.0431373, 0],[0.964706, 0.054902, 0],[0.968627, 0.0627451, 0],[0.972549, 0.0745098, 0],
[0.980392, 0.0862745, 0],[0.984314, 0.0980392, 0],[0.988235, 0.109804, 0],[0.992157, 0.117647, 0],
[0.996078, 0.129412, 0],[1, 0.141176, 0],[1, 0.152941, 0],[1, 0.160784, 0],
[1, 0.172549, 0],[1, 0.184314, 0],[1, 0.196078, 0],[1, 0.203922, 0],
[1, 0.215686, 0],[1, 0.227451, 0],[1, 0.235294, 0],[1, 0.247059, 0],
[1, 0.258824, 0],[1, 0.270588, 0],[1, 0.278431, 0],[1, 0.290196, 0],
[1, 0.301961, 0],[1, 0.309804, 0],[1, 0.321569, 0],[1, 0.329412, 0],
[1, 0.341176, 0],[1, 0.34902, 0],[1, 0.360784, 0],[1, 0.368627, 0],
[1, 0.376471, 0],[1, 0.388235, 0.00392157],[1, 0.396078, 0.00392157],[1, 0.403922, 0.00784314],
[1, 0.411765, 0.00784314],[1, 0.419608, 0.00784314],[1, 0.427451, 0.0117647],[1, 0.435294, 0.0117647],
[1, 0.443137, 0.0156863],[1, 0.45098, 0.0156863],[1, 0.458824, 0.0156863],[1, 0.466667, 0.0196078],
[1, 0.47451, 0.0196078],[1, 0.482353, 0.0235294],[1, 0.490196, 0.0235294],[1, 0.498039, 0.0235294],
[1, 0.505882, 0.027451],[1, 0.513725, 0.027451],[1, 0.521569, 0.027451],[1, 0.529412, 0.0313725],
[1, 0.537255, 0.0313725],[1, 0.541176, 0.0352941],[1, 0.54902, 0.0352941],[1, 0.556863, 0.0352941],
[1, 0.564706, 0.0392157],[1, 0.572549, 0.0392157],[1, 0.576471, 0.0392157],[1, 0.584314, 0.0431373],
[1, 0.592157, 0.0431373],[1, 0.6, 0.0470588],[1, 0.603922, 0.0470588],[1, 0.611765, 0.0470588],
[1, 0.619608, 0.0509804],[1, 0.623529, 0.0509804],[1, 0.631373, 0.0509804],[1, 0.639216, 0.054902],
[1, 0.643137, 0.054902],[1, 0.65098, 0.054902],[1, 0.658824, 0.0588235],[1, 0.662745, 0.0588235],
[1, 0.670588, 0.0588235],[1, 0.678431, 0.0627451],[1, 0.682353, 0.0627451],[1, 0.690196, 0.0627451],
[1, 0.694118, 0.0666667],[1, 0.701961, 0.0666667],[1, 0.709804, 0.0666667],[1, 0.713725, 0.0705882],
[1, 0.721569, 0.0705882],[1, 0.72549, 0.0705882],[1, 0.733333, 0.0705882],[1, 0.737255, 0.0745098],
[1, 0.745098, 0.0745098],[1, 0.74902, 0.0745098],[1, 0.756863, 0.0745098],[1, 0.760784, 0.0784314],
[1, 0.768627, 0.0784314],[1, 0.772549, 0.0784314],[1, 0.780392, 0.0784314],[1, 0.784314, 0.0784314],
[1, 0.792157, 0.0823529],[1, 0.796078, 0.0823529],[1, 0.803922, 0.0823529],[1, 0.807843, 0.0823529],
[1, 0.815686, 0.0862745],[1, 0.819608, 0.0862745],[1, 0.827451, 0.0862745],[1, 0.831373, 0.0862745],
[1, 0.839216, 0.0862745],[1, 0.843137, 0.0901961],[1, 0.85098, 0.0901961],[1, 0.854902, 0.0901961],
[1, 0.862745, 0.0901961],[1, 0.866667, 0.0901961],[1, 0.87451, 0.0941176],[1, 0.878431, 0.0941176],
[1, 0.886275, 0.0941176],[1, 0.890196, 0.0941176],[1, 0.894118, 0.0941176],[1, 0.901961, 0.0980392],
[1, 0.905882, 0.0980392],[1, 0.913725, 0.0980392],[1, 0.917647, 0.0980392],[1, 0.92549, 0.0980392],
[1, 0.929412, 0.101961],[1, 0.937255, 0.101961],[1, 0.941176, 0.101961],[1, 0.945098, 0.101961],
[1, 0.952941, 0.101961],[1, 0.956863, 0.105882],[1, 0.964706, 0.105882],[1, 0.968627, 0.105882],
[1, 0.976471, 0.105882],[1, 0.980392, 0.105882],[1, 0.988235, 0.109804],[1, 0.992157, 0.109804],
[1, 0.996078, 0.109804] 
], 
[
[0.0666667, 0.0666667, 0.0666667],[0.0784314, 0.0666667, 0.0627451],
[0.0941176, 0.0666667, 0.0627451],[0.105882, 0.0666667, 0.0627451],[0.113725, 0.0705882, 0.0588235],[0.121569, 0.0705882, 0.0588235],
[0.133333, 0.0705882, 0.0588235],[0.141176, 0.0705882, 0.054902],[0.14902, 0.0705882, 0.054902],[0.160784, 0.0705882, 0.054902],
[0.168627, 0.0705882, 0.0509804],[0.176471, 0.0705882, 0.0509804],[0.188235, 0.0666667, 0.0509804],[0.196078, 0.0666667, 0.0470588],
[0.203922, 0.0666667, 0.0470588],[0.211765, 0.0666667, 0.0470588],[0.219608, 0.0627451, 0.0431373],[0.231373, 0.0627451, 0.0431373],
[0.239216, 0.0588235, 0.0392157],[0.247059, 0.0588235, 0.0392157],[0.254902, 0.0588235, 0.0392157],[0.262745, 0.054902, 0.0352941],
[0.270588, 0.0509804, 0.0352941],[0.278431, 0.0509804, 0.0352941],[0.286275, 0.0470588, 0.0313725],[0.294118, 0.0470588, 0.0313725],
[0.305882, 0.0431373, 0.0313725],[0.313725, 0.0392157, 0.027451],[0.321569, 0.0352941, 0.027451],[0.329412, 0.0313725, 0.027451],
[0.337255, 0.0313725, 0.0235294],[0.341176, 0.027451, 0.0235294],[0.34902, 0.0235294, 0.0235294],[0.356863, 0.0196078, 0.0196078],
[0.364706, 0.0196078, 0.0196078],[0.372549, 0.0156863, 0.0196078],[0.380392, 0.0117647, 0.0196078],[0.384314, 0.0117647, 0.0156863],
[0.392157, 0.00784314, 0.0156863],[0.4, 0.00784314, 0.0156863],[0.403922, 0.00784314, 0.0156863],[0.411765, 0.00784314, 0.0156863],
[0.419608, 0.00784314, 0.0156863],[0.423529, 0.00784314, 0.0117647],[0.431373, 0.00784314, 0.0117647],[0.435294, 0.00784314, 0.0117647],
[0.443137, 0.00784314, 0.0117647],[0.447059, 0.00784314, 0.0117647],[0.454902, 0.00784314, 0.0117647],[0.462745, 0.00784314, 0.0117647],
[0.466667, 0.00784314, 0.0117647],[0.47451, 0.00784314, 0.00784314],[0.478431, 0.00784314, 0.00784314],[0.486275, 0.00784314, 0.00784314],
[0.494118, 0.00784314, 0.00784314],[0.498039, 0.00784314, 0.00784314],[0.505882, 0.00784314, 0.00784314],[0.509804, 0.00784314, 0.00784314],
[0.517647, 0.00784314, 0.00784314],[0.521569, 0.00784314, 0.00784314],[0.529412, 0.0117647, 0.00784314],[0.537255, 0.0117647, 0.00784314],
[0.541176, 0.0117647, 0.00784314],[0.54902, 0.0117647, 0.00784314],[0.552941, 0.0117647, 0.00784314],[0.560784, 0.0117647, 0.00784314],
[0.568627, 0.0117647, 0.00784314],[0.572549, 0.0117647, 0.0117647],[0.580392, 0.0156863, 0.0117647],[0.588235, 0.0156863, 0.0117647],
[0.592157, 0.0156863, 0.0117647],[0.6, 0.0156863, 0.0117647],[0.603922, 0.0156863, 0.0117647],[0.611765, 0.0156863, 0.0117647],
[0.619608, 0.0196078, 0.0117647],[0.623529, 0.0196078, 0.0117647],[0.631373, 0.0196078, 0.0117647],[0.639216, 0.0196078, 0.0117647],
[0.643137, 0.0196078, 0.0117647],[0.65098, 0.0196078, 0.0117647],[0.654902, 0.0235294, 0.0117647],[0.662745, 0.0235294, 0.0117647],
[0.670588, 0.0235294, 0.0117647],[0.67451, 0.0235294, 0.0117647],[0.682353, 0.027451, 0.0117647],[0.690196, 0.027451, 0.0117647],
[0.694118, 0.027451, 0.0117647],[0.701961, 0.027451, 0.0117647],[0.709804, 0.027451, 0.0117647],[0.713725, 0.0313725, 0.0117647],
[0.721569, 0.0313725, 0.0117647],[0.729412, 0.0313725, 0.0117647],[0.737255, 0.0313725, 0.0117647],[0.741176, 0.0352941, 0.0117647],
[0.74902, 0.0352941, 0.0117647],[0.756863, 0.0352941, 0.0117647],[0.760784, 0.0392157, 0.0117647],[0.768627, 0.0392157, 0.0117647],
[0.776471, 0.0431373, 0.0117647],[0.780392, 0.0431373, 0.0117647],[0.788235, 0.0470588, 0.0117647],[0.796078, 0.0470588, 0.0117647],
[0.8, 0.0509804, 0.0117647],[0.807843, 0.054902, 0.0117647],[0.811765, 0.054902, 0.0117647],[0.819608, 0.0588235, 0.0117647],
[0.827451, 0.0627451, 0.0117647],[0.831373, 0.0666667, 0.0117647],[0.839216, 0.0705882, 0.0156863],[0.843137, 0.0745098, 0.0156863],
[0.85098, 0.0784314, 0.0156863],[0.858824, 0.0823529, 0.0156863],[0.862745, 0.0862745, 0.0156863],[0.870588, 0.0901961, 0.0156863],
[0.87451, 0.0941176, 0.0156863],[0.882353, 0.0980392, 0.0156863],[0.886275, 0.101961, 0.0156863],[0.894118, 0.105882, 0.0156863],
[0.898039, 0.109804, 0.0196078],[0.905882, 0.113725, 0.0196078],[0.909804, 0.117647, 0.0196078],[0.917647, 0.12549, 0.0196078],
[0.921569, 0.129412, 0.0196078],[0.929412, 0.133333, 0.0196078],[0.933333, 0.141176, 0.0235294],[0.937255, 0.145098, 0.0235294],
[0.945098, 0.152941, 0.0235294],[0.94902, 0.160784, 0.0235294],[0.952941, 0.168627, 0.0235294],[0.956863, 0.176471, 0.0235294],
[0.960784, 0.184314, 0.027451],[0.964706, 0.192157, 0.027451],[0.968627, 0.2, 0.027451],[0.972549, 0.207843, 0.027451],
[0.976471, 0.215686, 0.027451],[0.980392, 0.223529, 0.0313725],[0.980392, 0.235294, 0.0313725],[0.984314, 0.243137, 0.0313725],
[0.988235, 0.25098, 0.0313725],[0.988235, 0.262745, 0.0352941],[0.992157, 0.270588, 0.0352941],[0.992157, 0.278431, 0.0352941],
[0.992157, 0.290196, 0.0392157],[0.996078, 0.298039, 0.0392157],[0.996078, 0.309804, 0.0392157],[0.996078, 0.317647, 0.0431373],
[0.996078, 0.329412, 0.0431373],[1, 0.337255, 0.0431373],[1, 0.345098, 0.0470588],[1, 0.356863, 0.0470588],
[1, 0.364706, 0.0470588],[1, 0.376471, 0.0509804],[1, 0.384314, 0.0509804],[1, 0.392157, 0.0509804],
[1, 0.4, 0.054902],[1, 0.411765, 0.054902],[1, 0.419608, 0.0588235],[1, 0.427451, 0.0588235],
[1, 0.435294, 0.0588235],[1, 0.443137, 0.0627451],[1, 0.45098, 0.0627451],[1, 0.458824, 0.0627451],
[0.996078, 0.466667, 0.0666667],[0.996078, 0.47451, 0.0666667],[0.996078, 0.482353, 0.0705882],[0.996078, 0.490196, 0.0705882],
[0.996078, 0.498039, 0.0705882],[0.996078, 0.505882, 0.0745098],[0.996078, 0.513725, 0.0745098],[0.996078, 0.521569, 0.0745098],
[0.996078, 0.529412, 0.0784314],[0.996078, 0.537255, 0.0784314],[0.992157, 0.545098, 0.0784314],[0.992157, 0.552941, 0.0823529],
[0.992157, 0.560784, 0.0823529],[0.992157, 0.564706, 0.0823529],[0.992157, 0.572549, 0.0862745],[0.992157, 0.580392, 0.0862745],
[0.988235, 0.588235, 0.0901961],[0.988235, 0.596078, 0.0901961],[0.988235, 0.603922, 0.0941176],[0.988235, 0.607843, 0.0941176],
[0.984314, 0.615686, 0.0980392],[0.984314, 0.623529, 0.0980392],[0.984314, 0.631373, 0.101961],[0.984314, 0.639216, 0.105882],
[0.980392, 0.643137, 0.105882],[0.980392, 0.65098, 0.109804],[0.980392, 0.658824, 0.113725],[0.976471, 0.666667, 0.117647],
[0.976471, 0.67451, 0.117647],[0.972549, 0.678431, 0.121569],[0.972549, 0.686275, 0.12549],[0.968627, 0.694118, 0.129412],
[0.968627, 0.701961, 0.133333],[0.964706, 0.705882, 0.137255],[0.964706, 0.713725, 0.141176],[0.960784, 0.721569, 0.145098],
[0.960784, 0.729412, 0.14902],[0.956863, 0.733333, 0.152941],[0.956863, 0.741176, 0.156863],[0.952941, 0.74902, 0.160784],
[0.94902, 0.756863, 0.164706],[0.94902, 0.760784, 0.172549],[0.94902, 0.768627, 0.180392],[0.945098, 0.772549, 0.192157],
[0.945098, 0.780392, 0.203922],[0.945098, 0.784314, 0.215686],[0.945098, 0.792157, 0.231373],[0.94902, 0.796078, 0.243137],
[0.94902, 0.8, 0.258824],[0.94902, 0.807843, 0.270588],[0.94902, 0.811765, 0.286275],[0.94902, 0.815686, 0.301961],
[0.952941, 0.823529, 0.313725],[0.952941, 0.827451, 0.329412],[0.952941, 0.831373, 0.345098],[0.956863, 0.835294, 0.360784],
[0.956863, 0.839216, 0.376471],[0.960784, 0.847059, 0.392157],[0.960784, 0.85098, 0.407843],[0.964706, 0.854902, 0.423529],
[0.964706, 0.858824, 0.439216],[0.964706, 0.862745, 0.454902],[0.968627, 0.866667, 0.470588],[0.968627, 0.87451, 0.486275],
[0.972549, 0.878431, 0.501961],[0.972549, 0.882353, 0.517647],[0.976471, 0.886275, 0.533333],[0.976471, 0.890196, 0.54902],
[0.980392, 0.894118, 0.564706],[0.980392, 0.898039, 0.580392],[0.980392, 0.901961, 0.6],[0.984314, 0.905882, 0.615686],
[0.984314, 0.909804, 0.631373],[0.988235, 0.917647, 0.647059],[0.988235, 0.921569, 0.666667],[0.988235, 0.92549, 0.682353],
[0.992157, 0.929412, 0.698039],[0.992157, 0.933333, 0.717647],[0.992157, 0.937255, 0.733333],[0.996078, 0.941176, 0.74902],
[0.996078, 0.945098, 0.768627],[0.996078, 0.94902, 0.784314],[0.996078, 0.952941, 0.803922],[0.996078, 0.956863, 0.819608],
[1, 0.960784, 0.839216],[1, 0.964706, 0.854902],[1, 0.972549, 0.87451],[1, 0.976471, 0.890196],
[1, 0.980392, 0.909804],[1, 0.984314, 0.92549],[1, 0.988235, 0.945098],[1, 0.992157, 0.964706],
[1, 0.996078, 0.980392],[1, 1, 1] 
], 
[
[0.0666667, 0.0666667, 0.0666667],
[0.0823529, 0.0666667, 0.0627451],[0.0941176, 0.0666667, 0.0627451],[0.105882, 0.0666667, 0.0627451],[0.113725, 0.0666667, 0.0588235],
[0.12549, 0.0705882, 0.0588235],[0.133333, 0.0705882, 0.0588235],[0.145098, 0.0705882, 0.054902],[0.152941, 0.0705882, 0.054902],
[0.160784, 0.0666667, 0.054902],[0.172549, 0.0666667, 0.0509804],[0.180392, 0.0666667, 0.0509804],[0.188235, 0.0666667, 0.0509804],
[0.2, 0.0627451, 0.0470588],[0.207843, 0.0627451, 0.0470588],[0.215686, 0.0627451, 0.0470588],[0.227451, 0.0588235, 0.0431373],
[0.235294, 0.0588235, 0.0431373],[0.243137, 0.054902, 0.0431373],[0.25098, 0.0509804, 0.0392157],[0.258824, 0.0509804, 0.0392157],
[0.270588, 0.0470588, 0.0352941],[0.278431, 0.0431373, 0.0352941],[0.286275, 0.0392157, 0.0352941],[0.294118, 0.0392157, 0.0313725],
[0.301961, 0.0352941, 0.0313725],[0.309804, 0.0313725, 0.0313725],[0.317647, 0.027451, 0.027451],[0.32549, 0.0235294, 0.027451],
[0.333333, 0.0196078, 0.027451],[0.341176, 0.0156863, 0.027451],[0.34902, 0.0117647, 0.0235294],[0.356863, 0.00784314, 0.0235294],
[0.364706, 0.00392157, 0.0235294],[0.372549, 0, 0.0196078],[0.380392, 0, 0.0196078],[0.388235, 0, 0.0196078],
[0.392157, 0, 0.0196078],[0.4, 0, 0.0196078],[0.407843, 0, 0.0156863],[0.411765, 0, 0.0156863],
[0.419608, 0, 0.0156863],[0.427451, 0, 0.0156863],[0.431373, 0, 0.0156863],[0.439216, 0, 0.0156863],
[0.443137, 0, 0.0117647],[0.45098, 0, 0.0117647],[0.458824, 0, 0.0117647],[0.462745, 0, 0.0117647],
[0.470588, 0, 0.0117647],[0.47451, 0, 0.0117647],[0.482353, 0, 0.00784314],[0.486275, 0, 0.00784314],
[0.494118, 0, 0.00784314],[0.501961, 0, 0.00784314],[0.505882, 0, 0.00784314],[0.513725, 0, 0.00784314],
[0.517647, 0, 0.00784314],[0.52549, 0, 0.00784314],[0.533333, 0, 0.00392157],[0.537255, 0, 0.00392157],
[0.545098, 0, 0.00392157],[0.54902, 0, 0.00392157],[0.556863, 0, 0.00392157],[0.564706, 0, 0.00392157],
[0.568627, 0, 0.00392157],[0.576471, 0, 0.00392157],[0.584314, 0, 0.00392157],[0.588235, 0, 0.00392157],
[0.596078, 0, 0],[0.6, 0, 0],[0.607843, 0, 0],[0.615686, 0, 0],
[0.619608, 0, 0],[0.627451, 0, 0],[0.635294, 0, 0],[0.639216, 0, 0],
[0.647059, 0, 0],[0.654902, 0, 0],[0.658824, 0, 0],[0.666667, 0, 0],
[0.67451, 0, 0],[0.678431, 0, 0],[0.686275, 0, 0],[0.694118, 0, 0],
[0.698039, 0, 0],[0.705882, 0, 0],[0.713725, 0, 0],[0.717647, 0, 0],
[0.72549, 0, 0],[0.733333, 0, 0],[0.741176, 0, 0],[0.745098, 0, 0],
[0.752941, 0, 0],[0.760784, 0, 0],[0.768627, 0, 0],[0.772549, 0, 0],
[0.780392, 0, 0],[0.788235, 0, 0],[0.792157, 0, 0],[0.8, 0, 0],
[0.807843, 0, 0],[0.815686, 0, 0],[0.819608, 0, 0],[0.827451, 0, 0],
[0.835294, 0, 0],[0.839216, 0, 0],[0.847059, 0, 0],[0.854902, 0, 0],
[0.858824, 0, 0],[0.866667, 0, 0],[0.87451, 0, 0],[0.878431, 0, 0],
[0.886275, 0, 0],[0.894118, 0, 0],[0.898039, 0, 0],[0.905882, 0, 0],
[0.913725, 0, 0],[0.917647, 0, 0],[0.92549, 0.00392157, 0],[0.933333, 0.00784314, 0],
[0.937255, 0.0156863, 0],[0.945098, 0.0196078, 0],[0.94902, 0.027451, 0],[0.956863, 0.0392157, 0],
[0.960784, 0.0509804, 0],[0.968627, 0.0627451, 0],[0.972549, 0.0705882, 0],[0.976471, 0.0823529, 0],
[0.984314, 0.0941176, 0],[0.988235, 0.105882, 0],[0.992157, 0.117647, 0],[0.996078, 0.129412, 0],
[1, 0.137255, 0],[1, 0.14902, 0],[1, 0.160784, 0],[1, 0.172549, 0],
[1, 0.184314, 0],[1, 0.192157, 0],[1, 0.203922, 0],[1, 0.215686, 0],
[1, 0.227451, 0],[1, 0.239216, 0],[1, 0.247059, 0],[1, 0.258824, 0],
[1, 0.270588, 0],[1, 0.282353, 0],[1, 0.290196, 0],[1, 0.301961, 0],
[1, 0.313725, 0],[1, 0.321569, 0],[1, 0.333333, 0],[1, 0.345098, 0],
[1, 0.352941, 0],[1, 0.360784, 0],[1, 0.372549, 0],[1, 0.380392, 0.00392157],
[1, 0.388235, 0.00392157],[1, 0.4, 0.00392157],[1, 0.407843, 0.00784314],[1, 0.415686, 0.00784314],
[1, 0.423529, 0.0117647],[1, 0.431373, 0.0117647],[1, 0.439216, 0.0117647],[1, 0.447059, 0.0156863],
[1, 0.458824, 0.0156863],[1, 0.466667, 0.0196078],[1, 0.47451, 0.0196078],[1, 0.482353, 0.0196078],
[1, 0.490196, 0.0235294],[1, 0.498039, 0.0235294],[1, 0.505882, 0.027451],[1, 0.509804, 0.027451],
[1, 0.517647, 0.027451],[1, 0.52549, 0.0313725],[1, 0.533333, 0.0313725],[1, 0.541176, 0.0313725],
[1, 0.54902, 0.0352941],[1, 0.556863, 0.0352941],[1, 0.564706, 0.0392157],[1, 0.568627, 0.0392157],
[1, 0.576471, 0.0392157],[1, 0.584314, 0.0431373],[1, 0.592157, 0.0431373],[1, 0.6, 0.0470588],
[1, 0.603922, 0.0470588],[1, 0.611765, 0.0470588],[1, 0.619608, 0.0509804],[1, 0.627451, 0.0509804],
[1, 0.631373, 0.0509804],[1, 0.639216, 0.054902],[1, 0.647059, 0.054902],[1, 0.65098, 0.054902],
[1, 0.658824, 0.0588235],[1, 0.666667, 0.0588235],[1, 0.670588, 0.0588235],[1, 0.678431, 0.0627451],
[1, 0.686275, 0.0627451],[1, 0.690196, 0.0627451],[1, 0.698039, 0.0666667],[1, 0.701961, 0.0666667],
[1, 0.709804, 0.0666667],[1, 0.717647, 0.0705882],[1, 0.721569, 0.0784314],[1, 0.729412, 0.0901961],
[1, 0.733333, 0.105882],[1, 0.741176, 0.121569],[1, 0.745098, 0.137255],[1, 0.752941, 0.152941],
[1, 0.756863, 0.172549],[1, 0.760784, 0.188235],[1, 0.768627, 0.203922],[1, 0.772549, 0.223529],
[1, 0.780392, 0.239216],[1, 0.784314, 0.254902],[1, 0.788235, 0.27451],[1, 0.796078, 0.290196],
[1, 0.8, 0.305882],[1, 0.803922, 0.32549],[1, 0.811765, 0.341176],[1, 0.815686, 0.356863],
[1, 0.819608, 0.376471],[1, 0.827451, 0.392157],[1, 0.831373, 0.407843],[1, 0.835294, 0.427451],
[1, 0.843137, 0.443137],[1, 0.847059, 0.458824],[1, 0.85098, 0.478431],[1, 0.858824, 0.494118],
[1, 0.862745, 0.513725],[1, 0.866667, 0.529412],[1, 0.87451, 0.54902],[1, 0.878431, 0.564706],
[1, 0.882353, 0.584314],[1, 0.890196, 0.6],[1, 0.894118, 0.619608],[1, 0.898039, 0.639216],
[1, 0.905882, 0.654902],[1, 0.909804, 0.67451],[1, 0.913725, 0.694118],[1, 0.921569, 0.709804],
[1, 0.92549, 0.729412],[1, 0.929412, 0.74902],[1, 0.937255, 0.768627],[1, 0.941176, 0.784314],
[1, 0.945098, 0.803922],[1, 0.952941, 0.823529],[1, 0.956863, 0.843137],[1, 0.960784, 0.862745],
[1, 0.968627, 0.882353],[1, 0.972549, 0.901961],[1, 0.980392, 0.921569],[1, 0.984314, 0.941176],
[1, 0.988235, 0.960784],[1, 0.996078, 0.980392],[1, 1, 1] 
], 
[
[0, 0, 0],[0, 0.00392157, 0.00784314],[0, 0.00392157, 0.0156863],[0.00392157, 0.00784314, 0.0235294],
[0.00392157, 0.0117647, 0.0313725],[0.00392157, 0.0117647, 0.0392157],[0.00392157, 0.0156863, 0.0431373],[0.00392157, 0.0156863, 0.0509804],
[0.00784314, 0.0196078, 0.0588235],[0.00784314, 0.0235294, 0.0666667],[0.00784314, 0.0235294, 0.0745098],[0.00784314, 0.027451, 0.0784314],
[0.00784314, 0.027451, 0.0862745],[0.00784314, 0.0313725, 0.0941176],[0.0117647, 0.0313725, 0.0980392],[0.0117647, 0.0352941, 0.105882],
[0.0117647, 0.0352941, 0.109804],[0.0117647, 0.0392157, 0.117647],[0.0117647, 0.0392157, 0.121569],[0.0117647, 0.0431373, 0.129412],
[0.0117647, 0.0431373, 0.133333],[0.0156863, 0.0470588, 0.137255],[0.0156863, 0.0470588, 0.145098],[0.0156863, 0.0509804, 0.14902],
[0.0156863, 0.0509804, 0.152941],[0.0156863, 0.0509804, 0.156863],[0.0156863, 0.054902, 0.160784],[0.0156863, 0.054902, 0.168627],
[0.0156863, 0.054902, 0.172549],[0.0156863, 0.0588235, 0.176471],[0.0196078, 0.0588235, 0.180392],[0.0196078, 0.0588235, 0.184314],
[0.0196078, 0.0627451, 0.188235],[0.0196078, 0.0627451, 0.192157],[0.0196078, 0.0666667, 0.196078],[0.0196078, 0.0666667, 0.2],
[0.0196078, 0.0666667, 0.203922],[0.0196078, 0.0666667, 0.207843],[0.0196078, 0.0705882, 0.211765],[0.0196078, 0.0705882, 0.215686],
[0.0235294, 0.0705882, 0.219608],[0.0235294, 0.0745098, 0.223529],[0.0235294, 0.0745098, 0.223529],[0.0235294, 0.0745098, 0.227451],
[0.0235294, 0.0784314, 0.231373],[0.0235294, 0.0784314, 0.235294],[0.0235294, 0.0784314, 0.239216],[0.0235294, 0.0784314, 0.243137],
[0.0235294, 0.0823529, 0.243137],[0.0235294, 0.0823529, 0.247059],[0.0235294, 0.0823529, 0.25098],[0.0235294, 0.0823529, 0.254902],
[0.027451, 0.0862745, 0.258824],[0.027451, 0.0862745, 0.262745],[0.027451, 0.0862745, 0.262745],[0.027451, 0.0901961, 0.266667],
[0.027451, 0.0901961, 0.270588],[0.027451, 0.0901961, 0.27451],[0.027451, 0.0901961, 0.278431],[0.027451, 0.0941176, 0.282353],
[0.027451, 0.0941176, 0.282353],[0.027451, 0.0941176, 0.286275],[0.027451, 0.0941176, 0.290196],[0.027451, 0.0980392, 0.294118],
[0.0313725, 0.0980392, 0.298039],[0.0313725, 0.0980392, 0.301961],[0.0313725, 0.101961, 0.305882],[0.0313725, 0.101961, 0.305882],
[0.0313725, 0.101961, 0.309804],[0.0313725, 0.101961, 0.313725],[0.0313725, 0.105882, 0.317647],[0.0313725, 0.105882, 0.321569],
[0.0313725, 0.105882, 0.32549],[0.0313725, 0.109804, 0.32549],[0.0313725, 0.109804, 0.329412],[0.0352941, 0.109804, 0.333333],
[0.0352941, 0.109804, 0.337255],[0.0352941, 0.113725, 0.341176],[0.0352941, 0.113725, 0.345098],[0.0352941, 0.113725, 0.34902],
[0.0352941, 0.117647, 0.352941],[0.0352941, 0.117647, 0.352941],[0.0352941, 0.117647, 0.356863],[0.0352941, 0.117647, 0.360784],
[0.0352941, 0.121569, 0.364706],[0.0352941, 0.121569, 0.368627],[0.0352941, 0.121569, 0.372549],[0.0392157, 0.12549, 0.376471],
[0.0392157, 0.12549, 0.376471],[0.0392157, 0.12549, 0.380392],[0.0392157, 0.12549, 0.384314],[0.0392157, 0.129412, 0.388235],
[0.0392157, 0.129412, 0.392157],[0.0392157, 0.129412, 0.396078],[0.0392157, 0.133333, 0.4],[0.0392157, 0.133333, 0.403922],
[0.0392157, 0.133333, 0.403922],[0.0392157, 0.133333, 0.407843],[0.0431373, 0.137255, 0.411765],[0.0431373, 0.137255, 0.415686],
[0.0431373, 0.137255, 0.419608],[0.0431373, 0.141176, 0.423529],[0.0431373, 0.141176, 0.427451],[0.0431373, 0.141176, 0.431373],
[0.0431373, 0.141176, 0.435294],[0.0431373, 0.145098, 0.435294],[0.0431373, 0.145098, 0.439216],[0.0431373, 0.145098, 0.443137],
[0.0431373, 0.14902, 0.447059],[0.0470588, 0.14902, 0.45098],[0.0470588, 0.14902, 0.454902],[0.0470588, 0.152941, 0.458824],
[0.0470588, 0.152941, 0.462745],[0.0470588, 0.152941, 0.466667],[0.0470588, 0.152941, 0.466667],[0.0470588, 0.156863, 0.470588],
[0.0470588, 0.156863, 0.47451],[0.0470588, 0.156863, 0.478431],[0.0470588, 0.160784, 0.482353],[0.0470588, 0.160784, 0.486275],
[0.0470588, 0.160784, 0.490196],[0.0509804, 0.164706, 0.494118],[0.0509804, 0.164706, 0.498039],[0.0509804, 0.164706, 0.501961],
[0.0509804, 0.164706, 0.505882],[0.0509804, 0.168627, 0.505882],[0.0509804, 0.168627, 0.509804],[0.0509804, 0.168627, 0.513725],
[0.0509804, 0.172549, 0.517647],[0.0509804, 0.172549, 0.521569],[0.0509804, 0.172549, 0.52549],[0.0509804, 0.176471, 0.529412],
[0.054902, 0.176471, 0.533333],[0.054902, 0.176471, 0.537255],[0.054902, 0.176471, 0.541176],[0.054902, 0.180392, 0.545098],
[0.054902, 0.180392, 0.54902],[0.054902, 0.180392, 0.54902],[0.054902, 0.184314, 0.552941],[0.054902, 0.184314, 0.556863],
[0.054902, 0.184314, 0.560784],[0.054902, 0.188235, 0.564706],[0.0588235, 0.188235, 0.568627],[0.0588235, 0.188235, 0.572549],
[0.0588235, 0.188235, 0.576471],[0.0588235, 0.192157, 0.580392],[0.0588235, 0.192157, 0.584314],[0.0588235, 0.192157, 0.588235],
[0.0588235, 0.196078, 0.592157],[0.0588235, 0.196078, 0.596078],[0.0588235, 0.196078, 0.596078],[0.0588235, 0.2, 0.6],
[0.0588235, 0.2, 0.603922],[0.0627451, 0.2, 0.607843],[0.0627451, 0.203922, 0.611765],[0.0627451, 0.203922, 0.615686],
[0.0627451, 0.203922, 0.619608],[0.0627451, 0.203922, 0.623529],[0.0627451, 0.207843, 0.627451],[0.0627451, 0.207843, 0.631373],
[0.0627451, 0.207843, 0.635294],[0.0627451, 0.211765, 0.639216],[0.0627451, 0.211765, 0.643137],[0.0627451, 0.211765, 0.647059],
[0.0666667, 0.215686, 0.65098],[0.0666667, 0.215686, 0.654902],[0.0666667, 0.215686, 0.658824],[0.0666667, 0.219608, 0.658824],
[0.0666667, 0.219608, 0.662745],[0.0666667, 0.219608, 0.666667],[0.0666667, 0.223529, 0.670588],[0.0666667, 0.223529, 0.67451],
[0.0666667, 0.223529, 0.678431],[0.0666667, 0.223529, 0.682353],[0.0705882, 0.227451, 0.686275],[0.0705882, 0.227451, 0.690196],
[0.0705882, 0.227451, 0.694118],[0.0705882, 0.231373, 0.698039],[0.0705882, 0.231373, 0.701961],[0.0705882, 0.231373, 0.705882],
[0.0705882, 0.235294, 0.709804],[0.0705882, 0.235294, 0.713725],[0.0705882, 0.235294, 0.717647],[0.0705882, 0.239216, 0.721569],
[0.0705882, 0.239216, 0.72549],[0.0745098, 0.239216, 0.729412],[0.0745098, 0.243137, 0.733333],[0.0745098, 0.243137, 0.737255],
[0.0745098, 0.243137, 0.741176],[0.0745098, 0.247059, 0.745098],[0.0745098, 0.247059, 0.745098],[0.0745098, 0.247059, 0.74902],
[0.0745098, 0.25098, 0.752941],[0.0745098, 0.25098, 0.756863],[0.0745098, 0.25098, 0.760784],[0.0784314, 0.25098, 0.764706],
[0.0784314, 0.254902, 0.768627],[0.0784314, 0.254902, 0.772549],[0.0784314, 0.254902, 0.776471],[0.0784314, 0.258824, 0.780392],
[0.0784314, 0.258824, 0.784314],[0.0784314, 0.258824, 0.788235],[0.0784314, 0.262745, 0.792157],[0.0784314, 0.262745, 0.796078],
[0.0784314, 0.262745, 0.8],[0.0823529, 0.266667, 0.803922],[0.0823529, 0.266667, 0.807843],[0.0823529, 0.266667, 0.811765],
[0.0823529, 0.270588, 0.815686],[0.0823529, 0.270588, 0.819608],[0.0823529, 0.270588, 0.823529],[0.0823529, 0.27451, 0.827451],
[0.0823529, 0.27451, 0.831373],[0.0823529, 0.27451, 0.835294],[0.0823529, 0.278431, 0.839216],[0.0823529, 0.278431, 0.843137],
[0.0862745, 0.278431, 0.847059],[0.0862745, 0.282353, 0.85098],[0.0862745, 0.282353, 0.854902],[0.0862745, 0.282353, 0.858824],
[0.0862745, 0.286275, 0.862745],[0.0862745, 0.286275, 0.866667],[0.0862745, 0.286275, 0.870588],[0.0862745, 0.290196, 0.87451],
[0.0862745, 0.290196, 0.878431],[0.0862745, 0.290196, 0.882353],[0.0901961, 0.294118, 0.886275],[0.0901961, 0.294118, 0.890196],
[0.0901961, 0.294118, 0.894118],[0.0901961, 0.298039, 0.898039],[0.0901961, 0.298039, 0.901961],[0.0901961, 0.298039, 0.905882],
[0.0901961, 0.298039, 0.909804],[0.0901961, 0.301961, 0.913725],[0.0901961, 0.301961, 0.917647],[0.0901961, 0.301961, 0.921569],
[0.0941176, 0.305882, 0.92549],[0.0941176, 0.305882, 0.929412],[0.0941176, 0.305882, 0.933333],[0.0941176, 0.309804, 0.937255],
[0.0941176, 0.309804, 0.941176],[0.0941176, 0.309804, 0.945098],[0.0941176, 0.313725, 0.94902],[0.0941176, 0.313725, 0.952941],
[0.0941176, 0.313725, 0.956863],[0.0941176, 0.317647, 0.960784],[0.0980392, 0.317647, 0.964706],[0.0980392, 0.317647, 0.968627],
[0.0980392, 0.321569, 0.972549],[0.0980392, 0.321569, 0.976471],[0.0980392, 0.321569, 0.980392],[0.0980392, 0.32549, 0.984314],
[0.0980392, 0.32549, 0.988235],[0.0980392, 0.32549, 0.992157],[0.0980392, 0.329412, 0.996078],[0.101961, 0.329412, 1] 
], 
[
[0, 0, 0],[0, 0.00392157, 0],[0, 0.00784314, 0],
[0, 0.0117647, 0],[0, 0.0156863, 0],[0, 0.0196078, 0],[0, 0.0235294, 0],
[0, 0.0235294, 0],[0, 0.027451, 0],[0, 0.0313725, 0],[0, 0.0352941, 0],
[0, 0.0392157, 0],[0, 0.0431373, 0],[0, 0.0470588, 0],[0, 0.0509804, 0],
[0, 0.0509804, 0],[0, 0.054902, 0],[0, 0.0588235, 0],[0, 0.0627451, 0],
[0, 0.0627451, 0],[0, 0.0666667, 0],[0, 0.0705882, 0],[0, 0.0705882, 0],
[0, 0.0745098, 0],[0, 0.0745098, 0],[0, 0.0784314, 0],[0, 0.0823529, 0],
[0, 0.0823529, 0],[0, 0.0862745, 0],[0, 0.0862745, 0],[0, 0.0901961, 0],
[0, 0.0901961, 0],[0, 0.0941176, 0],[0, 0.0941176, 0],[0, 0.0980392, 0],
[0, 0.0980392, 0],[0, 0.101961, 0],[0, 0.101961, 0],[0, 0.101961, 0],
[0, 0.105882, 0],[0, 0.105882, 0],[0, 0.109804, 0],[0, 0.109804, 0],
[0, 0.113725, 0],[0, 0.113725, 0],[0, 0.113725, 0],[0, 0.117647, 0],
[0, 0.117647, 0],[0, 0.121569, 0],[0, 0.121569, 0],[0, 0.121569, 0],
[0, 0.12549, 0],[0, 0.12549, 0],[0, 0.129412, 0],[0, 0.129412, 0],
[0, 0.133333, 0],[0, 0.133333, 0],[0, 0.133333, 0],[0, 0.137255, 0],
[0, 0.137255, 0],[0, 0.141176, 0],[0, 0.141176, 0],[0, 0.141176, 0],
[0, 0.145098, 0],[0, 0.145098, 0],[0, 0.14902, 0],[0, 0.14902, 0],
[0, 0.152941, 0],[0, 0.152941, 0],[0, 0.152941, 0],[0, 0.156863, 0],
[0, 0.156863, 0],[0, 0.160784, 0],[0, 0.160784, 0],[0, 0.164706, 0],
[0, 0.164706, 0],[0, 0.164706, 0],[0, 0.168627, 0],[0, 0.168627, 0],
[0, 0.172549, 0],[0, 0.172549, 0],[0, 0.176471, 0],[0, 0.176471, 0],
[0, 0.176471, 0],[0, 0.180392, 0],[0, 0.180392, 0],[0, 0.184314, 0],
[0, 0.184314, 0],[0, 0.188235, 0],[0, 0.188235, 0],[0, 0.188235, 0],
[0, 0.192157, 0],[0, 0.192157, 0],[0, 0.196078, 0],[0, 0.196078, 0],
[0, 0.2, 0],[0, 0.2, 0],[0, 0.203922, 0],[0, 0.203922, 0],
[0, 0.203922, 0],[0, 0.207843, 0],[0, 0.207843, 0],[0, 0.211765, 0],
[0, 0.211765, 0],[0, 0.215686, 0],[0, 0.215686, 0],[0, 0.219608, 0],
[0, 0.219608, 0],[0, 0.219608, 0],[0, 0.223529, 0],[0, 0.223529, 0],
[0, 0.227451, 0],[0, 0.227451, 0],[0, 0.231373, 0],[0, 0.231373, 0],
[0, 0.235294, 0],[0, 0.235294, 0],[0, 0.235294, 0],[0, 0.239216, 0],
[0, 0.239216, 0],[0, 0.243137, 0],[0, 0.243137, 0],[0, 0.247059, 0],
[0, 0.247059, 0],[0, 0.25098, 0],[0, 0.25098, 0],[0, 0.254902, 0],
[0, 0.254902, 0],[0, 0.254902, 0],[0, 0.258824, 0],[0, 0.258824, 0],
[0, 0.262745, 0],[0, 0.262745, 0],[0, 0.266667, 0],[0, 0.266667, 0],
[0, 0.270588, 0],[0, 0.270588, 0],[0, 0.27451, 0],[0, 0.27451, 0],
[0, 0.278431, 0],[0, 0.278431, 0],[0, 0.278431, 0],[0, 0.282353, 0],
[0, 0.282353, 0],[0, 0.286275, 0],[0, 0.286275, 0],[0, 0.290196, 0],
[0, 0.290196, 0],[0, 0.294118, 0],[0, 0.294118, 0],[0, 0.298039, 0],
[0, 0.298039, 0],[0, 0.301961, 0],[0, 0.301961, 0],[0, 0.305882, 0],
[0, 0.305882, 0],[0, 0.309804, 0],[0, 0.309804, 0],[0, 0.309804, 0],
[0, 0.313725, 0],[0, 0.313725, 0],[0, 0.317647, 0],[0, 0.317647, 0],
[0, 0.321569, 0],[0, 0.321569, 0],[0, 0.32549, 0],[0, 0.32549, 0],
[0, 0.329412, 0],[0, 0.329412, 0],[0, 0.333333, 0],[0, 0.333333, 0],
[0, 0.337255, 0],[0, 0.337255, 0],[0, 0.341176, 0],[0, 0.341176, 0],
[0, 0.345098, 0],[0, 0.345098, 0],[0, 0.34902, 0],[0, 0.34902, 0],
[0, 0.352941, 0],[0, 0.352941, 0],[0, 0.352941, 0],[0, 0.356863, 0],
[0, 0.356863, 0],[0, 0.360784, 0],[0, 0.360784, 0],[0, 0.364706, 0],
[0, 0.364706, 0],[0, 0.368627, 0],[0, 0.368627, 0],[0, 0.372549, 0],
[0, 0.372549, 0],[0, 0.376471, 0],[0, 0.376471, 0],[0, 0.380392, 0],
[0, 0.380392, 0],[0, 0.384314, 0],[0, 0.384314, 0],[0, 0.388235, 0],
[0, 0.388235, 0],[0, 0.392157, 0],[0, 0.392157, 0],[0, 0.396078, 0],
[0, 0.396078, 0],[0, 0.4, 0],[0, 0.4, 0],[0, 0.403922, 0],
[0, 0.403922, 0],[0, 0.407843, 0],[0, 0.407843, 0],[0, 0.411765, 0],
[0, 0.411765, 0],[0, 0.415686, 0],[0, 0.415686, 0],[0, 0.419608, 0],
[0, 0.419608, 0],[0, 0.423529, 0],[0, 0.423529, 0],[0, 0.427451, 0],
[0, 0.427451, 0],[0, 0.431373, 0],[0, 0.431373, 0],[0, 0.435294, 0],
[0, 0.435294, 0],[0, 0.439216, 0],[0, 0.439216, 0],[0, 0.443137, 0],
[0, 0.443137, 0],[0, 0.447059, 0],[0, 0.447059, 0],[0, 0.45098, 0],
[0, 0.45098, 0],[0, 0.454902, 0],[0, 0.454902, 0],[0, 0.458824, 0],
[0, 0.458824, 0],[0, 0.462745, 0],[0, 0.462745, 0],[0, 0.466667, 0],
[0, 0.466667, 0],[0, 0.470588, 0],[0, 0.470588, 0],[0, 0.47451, 0],
[0, 0.47451, 0],[0, 0.478431, 0],[0, 0.478431, 0],[0, 0.482353, 0],
[0, 0.482353, 0],[0, 0.486275, 0],[0, 0.486275, 0],[0, 0.490196, 0],
[0, 0.490196, 0],[0, 0.494118, 0],[0, 0.494118, 0],[0, 0.498039, 0],
[0, 0.501961, 0] 
], 
[
[0, 0, 0],[0.00784314, 0, 0],
[0.0156863, 0.00392157, 0],[0.0235294, 0.00392157, 0],[0.0313725, 0.00784314, 0],[0.0392157, 0.00784314, 0],
[0.0470588, 0.00784314, 0],[0.054902, 0.0117647, 0],[0.0627451, 0.0117647, 0],[0.0705882, 0.0117647, 0],
[0.0745098, 0.0156863, 0],[0.0823529, 0.0156863, 0],[0.0862745, 0.0156863, 0],[0.0901961, 0.0156863, 0],
[0.0980392, 0.0196078, 0],[0.101961, 0.0196078, 0],[0.105882, 0.0196078, 0],[0.109804, 0.0196078, 0],
[0.117647, 0.0235294, 0],[0.121569, 0.0235294, 0],[0.12549, 0.0235294, 0],[0.129412, 0.0235294, 0],
[0.133333, 0.0235294, 0],[0.137255, 0.027451, 0],[0.141176, 0.027451, 0],[0.145098, 0.027451, 0],
[0.14902, 0.027451, 0],[0.152941, 0.027451, 0],[0.156863, 0.0313725, 0],[0.160784, 0.0313725, 0],
[0.160784, 0.0313725, 0],[0.164706, 0.0313725, 0],[0.168627, 0.0313725, 0],[0.172549, 0.0313725, 0],
[0.176471, 0.0313725, 0],[0.180392, 0.0352941, 0],[0.180392, 0.0352941, 0],[0.184314, 0.0352941, 0],
[0.188235, 0.0352941, 0],[0.192157, 0.0352941, 0],[0.196078, 0.0352941, 0],[0.196078, 0.0392157, 0],
[0.2, 0.0392157, 0],[0.203922, 0.0392157, 0],[0.207843, 0.0392157, 0],[0.207843, 0.0392157, 0],
[0.211765, 0.0392157, 0],[0.215686, 0.0392157, 0],[0.215686, 0.0392157, 0],[0.219608, 0.0431373, 0],
[0.223529, 0.0431373, 0],[0.227451, 0.0431373, 0],[0.227451, 0.0431373, 0],[0.231373, 0.0431373, 0],
[0.235294, 0.0431373, 0],[0.239216, 0.0431373, 0],[0.239216, 0.0470588, 0],[0.243137, 0.0470588, 0],
[0.247059, 0.0470588, 0],[0.25098, 0.0470588, 0],[0.25098, 0.0470588, 0],[0.254902, 0.0470588, 0],
[0.258824, 0.0470588, 0],[0.262745, 0.0509804, 0],[0.266667, 0.0509804, 0],[0.266667, 0.0509804, 0],
[0.270588, 0.0509804, 0],[0.27451, 0.0509804, 0],[0.278431, 0.0509804, 0],[0.278431, 0.0509804, 0],
[0.282353, 0.054902, 0],[0.286275, 0.054902, 0],[0.290196, 0.054902, 0],[0.290196, 0.054902, 0],
[0.294118, 0.054902, 0],[0.298039, 0.054902, 0],[0.301961, 0.0588235, 0],[0.305882, 0.0588235, 0],
[0.305882, 0.0588235, 0],[0.309804, 0.0588235, 0],[0.313725, 0.0588235, 0],[0.317647, 0.0588235, 0],
[0.317647, 0.0588235, 0],[0.321569, 0.0627451, 0],[0.32549, 0.0627451, 0],[0.329412, 0.0627451, 0],
[0.333333, 0.0627451, 0],[0.333333, 0.0627451, 0],[0.337255, 0.0627451, 0],[0.341176, 0.0627451, 0],
[0.345098, 0.0666667, 0],[0.345098, 0.0666667, 0],[0.34902, 0.0666667, 0],[0.352941, 0.0666667, 0],
[0.356863, 0.0666667, 0],[0.360784, 0.0666667, 0],[0.360784, 0.0666667, 0],[0.364706, 0.0705882, 0],
[0.368627, 0.0705882, 0],[0.372549, 0.0705882, 0],[0.376471, 0.0705882, 0],[0.376471, 0.0705882, 0],
[0.380392, 0.0705882, 0],[0.384314, 0.0745098, 0],[0.388235, 0.0745098, 0],[0.392157, 0.0745098, 0],
[0.392157, 0.0745098, 0],[0.396078, 0.0745098, 0],[0.4, 0.0745098, 0],[0.403922, 0.0745098, 0],
[0.407843, 0.0784314, 0],[0.407843, 0.0784314, 0],[0.411765, 0.0784314, 0],[0.415686, 0.0784314, 0],
[0.419608, 0.0784314, 0],[0.423529, 0.0784314, 0],[0.427451, 0.0784314, 0],[0.427451, 0.0823529, 0],
[0.431373, 0.0823529, 0],[0.435294, 0.0823529, 0],[0.439216, 0.0823529, 0],[0.443137, 0.0823529, 0],
[0.443137, 0.0823529, 0],[0.447059, 0.0862745, 0],[0.45098, 0.0862745, 0],[0.454902, 0.0862745, 0],
[0.458824, 0.0862745, 0],[0.462745, 0.0862745, 0],[0.462745, 0.0862745, 0],[0.466667, 0.0901961, 0],
[0.470588, 0.0901961, 0],[0.47451, 0.0901961, 0],[0.478431, 0.0901961, 0],[0.478431, 0.0901961, 0],
[0.482353, 0.0901961, 0],[0.486275, 0.0901961, 0],[0.490196, 0.0941176, 0],[0.494118, 0.0941176, 0],
[0.498039, 0.0941176, 0],[0.498039, 0.0941176, 0],[0.501961, 0.0941176, 0],[0.505882, 0.0941176, 0],
[0.509804, 0.0980392, 0],[0.513725, 0.0980392, 0],[0.517647, 0.0980392, 0],[0.517647, 0.0980392, 0],
[0.521569, 0.0980392, 0],[0.52549, 0.0980392, 0],[0.529412, 0.0980392, 0],[0.533333, 0.101961, 0],
[0.537255, 0.101961, 0],[0.541176, 0.101961, 0],[0.541176, 0.101961, 0],[0.545098, 0.101961, 0],
[0.54902, 0.101961, 0],[0.552941, 0.105882, 0],[0.556863, 0.105882, 0],[0.560784, 0.105882, 0],
[0.560784, 0.105882, 0],[0.564706, 0.105882, 0],[0.568627, 0.105882, 0],[0.572549, 0.109804, 0],
[0.576471, 0.109804, 0],[0.580392, 0.109804, 0],[0.584314, 0.109804, 0],[0.584314, 0.109804, 0],
[0.588235, 0.109804, 0],[0.592157, 0.113725, 0],[0.596078, 0.113725, 0],[0.6, 0.113725, 0],
[0.603922, 0.113725, 0],[0.607843, 0.113725, 0],[0.607843, 0.113725, 0],[0.611765, 0.117647, 0],
[0.615686, 0.117647, 0],[0.619608, 0.117647, 0],[0.623529, 0.117647, 0],[0.627451, 0.117647, 0],
[0.631373, 0.117647, 0],[0.631373, 0.117647, 0],[0.635294, 0.121569, 0],[0.639216, 0.121569, 0],
[0.643137, 0.121569, 0],[0.647059, 0.121569, 0],[0.65098, 0.121569, 0],[0.654902, 0.121569, 0],
[0.658824, 0.12549, 0],[0.658824, 0.12549, 0],[0.662745, 0.12549, 0],[0.666667, 0.12549, 0],
[0.670588, 0.12549, 0],[0.67451, 0.12549, 0],[0.678431, 0.129412, 0],[0.682353, 0.129412, 0],
[0.686275, 0.129412, 0],[0.686275, 0.129412, 0],[0.690196, 0.129412, 0],[0.694118, 0.129412, 0],
[0.698039, 0.133333, 0],[0.701961, 0.133333, 0],[0.705882, 0.133333, 0],[0.709804, 0.133333, 0],
[0.713725, 0.133333, 0],[0.713725, 0.133333, 0],[0.717647, 0.137255, 0],[0.721569, 0.137255, 0],
[0.72549, 0.137255, 0],[0.729412, 0.137255, 0],[0.733333, 0.137255, 0],[0.737255, 0.137255, 0],
[0.741176, 0.141176, 0],[0.745098, 0.141176, 0],[0.745098, 0.141176, 0],[0.74902, 0.141176, 0],
[0.752941, 0.141176, 0],[0.756863, 0.141176, 0],[0.760784, 0.145098, 0],[0.764706, 0.145098, 0],
[0.768627, 0.145098, 0],[0.772549, 0.145098, 0],[0.776471, 0.145098, 0],[0.780392, 0.14902, 0],
[0.780392, 0.14902, 0],[0.784314, 0.14902, 0],[0.788235, 0.14902, 0],[0.792157, 0.14902, 0],
[0.796078, 0.14902, 0],[0.8, 0.152941, 0],[0.803922, 0.152941, 0],[0.807843, 0.152941, 0],
[0.811765, 0.152941, 0],[0.815686, 0.152941, 0],[0.815686, 0.152941, 0],[0.819608, 0.156863, 0],
[0.823529, 0.156863, 0],[0.827451, 0.156863, 0],[0.831373, 0.156863, 0],[0.835294, 0.156863, 0],
[0.839216, 0.156863, 0],[0.843137, 0.160784, 0],[0.847059, 0.160784, 0],[0.85098, 0.160784, 0],
[0.854902, 0.160784, 0],[0.854902, 0.160784, 0],[0.858824, 0.160784, 0],[0.862745, 0.164706, 0],
[0.866667, 0.164706, 0],[0.870588, 0.164706, 0],[0.87451, 0.164706, 0],[0.878431, 0.164706, 0],
[0.882353, 0.164706, 0],[0.886275, 0.168627, 0],[0.890196, 0.168627, 0],[0.894118, 0.168627, 0],
[0.898039, 0.168627, 0],[0.901961, 0.168627, 0] 
], 
[
[0, 0.203922, 0.960784],
[0, 0.215686, 0.952941],[0, 0.227451, 0.941176],[0, 0.235294, 0.929412],[0, 0.247059, 0.917647],
[0, 0.254902, 0.905882],[0, 0.266667, 0.898039],[0, 0.27451, 0.886275],[0, 0.282353, 0.87451],
[0, 0.290196, 0.862745],[0, 0.301961, 0.85098],[0, 0.309804, 0.839216],[0, 0.317647, 0.831373],
[0, 0.32549, 0.819608],[0, 0.333333, 0.807843],[0, 0.341176, 0.796078],[0, 0.34902, 0.788235],
[0, 0.352941, 0.776471],[0, 0.360784, 0.764706],[0, 0.368627, 0.752941],[0, 0.376471, 0.741176],
[0, 0.384314, 0.733333],[0, 0.388235, 0.721569],[0, 0.396078, 0.709804],[0, 0.403922, 0.698039],
[0, 0.407843, 0.690196],[0, 0.415686, 0.678431],[0, 0.423529, 0.666667],[0, 0.427451, 0.654902],
[0, 0.435294, 0.647059],[0, 0.439216, 0.635294],[0, 0.443137, 0.627451],[0, 0.45098, 0.615686],
[0, 0.454902, 0.603922],[0, 0.458824, 0.596078],[0.00392157, 0.462745, 0.584314],[0.0313725, 0.466667, 0.576471],
[0.054902, 0.470588, 0.568627],[0.0784314, 0.478431, 0.556863],[0.0941176, 0.482353, 0.54902],[0.109804, 0.486275, 0.537255],
[0.12549, 0.490196, 0.529412],[0.137255, 0.494118, 0.521569],[0.14902, 0.498039, 0.509804],[0.160784, 0.501961, 0.501961],
[0.168627, 0.505882, 0.494118],[0.176471, 0.509804, 0.482353],[0.184314, 0.513725, 0.47451],[0.192157, 0.517647, 0.462745],
[0.2, 0.521569, 0.454902],[0.203922, 0.52549, 0.447059],[0.207843, 0.529412, 0.435294],[0.215686, 0.533333, 0.427451],
[0.219608, 0.537255, 0.415686],[0.223529, 0.541176, 0.407843],[0.227451, 0.545098, 0.396078],[0.231373, 0.54902, 0.388235],
[0.231373, 0.552941, 0.376471],[0.235294, 0.556863, 0.368627],[0.239216, 0.560784, 0.356863],[0.239216, 0.564706, 0.345098],
[0.243137, 0.568627, 0.337255],[0.243137, 0.572549, 0.32549],[0.243137, 0.576471, 0.313725],[0.247059, 0.580392, 0.305882],
[0.247059, 0.584314, 0.294118],[0.247059, 0.588235, 0.282353],[0.247059, 0.596078, 0.270588],[0.247059, 0.6, 0.258824],
[0.247059, 0.603922, 0.247059],[0.247059, 0.607843, 0.235294],[0.247059, 0.611765, 0.223529],[0.247059, 0.615686, 0.207843],
[0.247059, 0.619608, 0.196078],[0.247059, 0.623529, 0.184314],[0.247059, 0.627451, 0.168627],[0.247059, 0.631373, 0.156863],
[0.25098, 0.635294, 0.141176],[0.25098, 0.639216, 0.129412],[0.254902, 0.643137, 0.117647],[0.258824, 0.647059, 0.105882],
[0.266667, 0.65098, 0.0941176],[0.27451, 0.65098, 0.0862745],[0.282353, 0.654902, 0.0784314],[0.290196, 0.658824, 0.0705882],
[0.301961, 0.662745, 0.0627451],[0.309804, 0.662745, 0.0627451],[0.321569, 0.666667, 0.0588235],[0.333333, 0.670588, 0.0588235],
[0.341176, 0.670588, 0.0588235],[0.352941, 0.67451, 0.0588235],[0.364706, 0.67451, 0.0588235],[0.376471, 0.678431, 0.0627451],
[0.384314, 0.682353, 0.0627451],[0.396078, 0.682353, 0.0666667],[0.407843, 0.686275, 0.0666667],[0.415686, 0.686275, 0.0705882],
[0.427451, 0.690196, 0.0705882],[0.435294, 0.694118, 0.0705882],[0.447059, 0.694118, 0.0745098],[0.454902, 0.698039, 0.0745098],
[0.466667, 0.698039, 0.0784314],[0.47451, 0.701961, 0.0784314],[0.486275, 0.705882, 0.0823529],[0.494118, 0.705882, 0.0823529],
[0.505882, 0.709804, 0.0862745],[0.513725, 0.709804, 0.0862745],[0.521569, 0.713725, 0.0901961],[0.533333, 0.713725, 0.0901961],
[0.541176, 0.717647, 0.0941176],[0.54902, 0.717647, 0.0941176],[0.560784, 0.721569, 0.0980392],[0.568627, 0.72549, 0.0980392],
[0.576471, 0.72549, 0.101961],[0.588235, 0.729412, 0.101961],[0.596078, 0.729412, 0.101961],[0.603922, 0.733333, 0.105882],
[0.611765, 0.733333, 0.105882],[0.623529, 0.737255, 0.109804],[0.631373, 0.737255, 0.109804],[0.639216, 0.741176, 0.113725],
[0.647059, 0.745098, 0.113725],[0.658824, 0.745098, 0.117647],[0.666667, 0.74902, 0.117647],[0.67451, 0.74902, 0.121569],
[0.682353, 0.752941, 0.121569],[0.690196, 0.752941, 0.121569],[0.701961, 0.756863, 0.12549],[0.709804, 0.756863, 0.12549],
[0.717647, 0.760784, 0.129412],[0.72549, 0.760784, 0.129412],[0.733333, 0.764706, 0.133333],[0.745098, 0.764706, 0.133333],
[0.752941, 0.768627, 0.137255],[0.760784, 0.768627, 0.137255],[0.768627, 0.772549, 0.137255],[0.776471, 0.772549, 0.141176],
[0.784314, 0.776471, 0.141176],[0.796078, 0.776471, 0.145098],[0.803922, 0.780392, 0.145098],[0.811765, 0.780392, 0.14902],
[0.819608, 0.784314, 0.14902],[0.827451, 0.784314, 0.152941],[0.835294, 0.788235, 0.152941],[0.843137, 0.788235, 0.152941],
[0.854902, 0.792157, 0.156863],[0.862745, 0.792157, 0.156863],[0.870588, 0.796078, 0.160784],[0.878431, 0.796078, 0.160784],
[0.886275, 0.796078, 0.164706],[0.894118, 0.8, 0.164706],[0.901961, 0.8, 0.168627],[0.909804, 0.803922, 0.168627],
[0.917647, 0.803922, 0.168627],[0.92549, 0.803922, 0.172549],[0.933333, 0.803922, 0.172549],[0.941176, 0.803922, 0.172549],
[0.94902, 0.803922, 0.172549],[0.956863, 0.803922, 0.172549],[0.960784, 0.803922, 0.172549],[0.964706, 0.8, 0.172549],
[0.972549, 0.8, 0.172549],[0.976471, 0.796078, 0.172549],[0.976471, 0.792157, 0.172549],[0.980392, 0.788235, 0.168627],
[0.984314, 0.784314, 0.168627],[0.984314, 0.780392, 0.164706],[0.988235, 0.772549, 0.164706],[0.988235, 0.768627, 0.160784],
[0.988235, 0.764706, 0.160784],[0.988235, 0.760784, 0.156863],[0.992157, 0.752941, 0.156863],[0.992157, 0.74902, 0.152941],
[0.992157, 0.745098, 0.152941],[0.992157, 0.737255, 0.14902],[0.992157, 0.733333, 0.14902],[0.992157, 0.729412, 0.145098],
[0.996078, 0.721569, 0.145098],[0.996078, 0.717647, 0.141176],[0.996078, 0.709804, 0.137255],[0.996078, 0.705882, 0.137255],
[0.996078, 0.701961, 0.133333],[0.996078, 0.694118, 0.133333],[0.996078, 0.690196, 0.129412],[0.996078, 0.686275, 0.129412],
[1, 0.678431, 0.12549],[1, 0.67451, 0.121569],[1, 0.666667, 0.121569],[1, 0.662745, 0.117647],
[1, 0.658824, 0.117647],[1, 0.65098, 0.113725],[1, 0.647059, 0.113725],[1, 0.639216, 0.109804],
[1, 0.635294, 0.109804],[1, 0.631373, 0.105882],[1, 0.623529, 0.101961],[1, 0.619608, 0.101961],
[1, 0.611765, 0.0980392],[1, 0.607843, 0.0980392],[1, 0.6, 0.0941176],[1, 0.596078, 0.0941176],
[1, 0.588235, 0.0901961],[1, 0.584314, 0.0862745],[1, 0.576471, 0.0862745],[1, 0.572549, 0.0823529],
[1, 0.568627, 0.0823529],[1, 0.560784, 0.0784314],[1, 0.556863, 0.0784314],[1, 0.54902, 0.0745098],
[1, 0.545098, 0.0705882],[1, 0.537255, 0.0705882],[1, 0.533333, 0.0666667],[1, 0.52549, 0.0666667],
[1, 0.517647, 0.0627451],[1, 0.513725, 0.0627451],[1, 0.505882, 0.0588235],[1, 0.501961, 0.054902],
[1, 0.494118, 0.054902],[1, 0.490196, 0.0509804],[1, 0.482353, 0.0509804],[1, 0.47451, 0.0470588],
[1, 0.470588, 0.0431373],[1, 0.462745, 0.0431373],[1, 0.454902, 0.0392157],[1, 0.45098, 0.0392157],
[1, 0.443137, 0.0352941],[1, 0.435294, 0.0313725],[1, 0.431373, 0.0313725],[1, 0.423529, 0.027451],
[1, 0.415686, 0.027451],[1, 0.411765, 0.0235294],[1, 0.403922, 0.0235294],[1, 0.396078, 0.0196078],
[1, 0.388235, 0.0196078],[1, 0.380392, 0.0156863],[1, 0.372549, 0.0156863],[1, 0.368627, 0.0117647],
[1, 0.360784, 0.0117647],[1, 0.352941, 0.0117647],[1, 0.345098, 0.00784314],[0.996078, 0.337255, 0.00784314],
[0.996078, 0.329412, 0.00784314],[0.996078, 0.321569, 0.00392157],[0.996078, 0.309804, 0.00392157],[0.996078, 0.301961, 0.00392157],
[0.996078, 0.294118, 0],[0.996078, 0.286275, 0],[0.996078, 0.27451, 0],[0.996078, 0.266667, 0],
[0.992157, 0.254902, 0],[0.992157, 0.247059, 0],[0.992157, 0.235294, 0],[0.992157, 0.223529, 0],
[0.992157, 0.211765, 0],[0.992157, 0.2, 0],[0.992157, 0.188235, 0] 
], 
[
[0, 0.207843, 0.976471],[0, 0.219608, 0.964706],[0, 0.227451, 0.952941],[0, 0.239216, 0.941176],
[0, 0.247059, 0.929412],[0, 0.258824, 0.917647],[0, 0.266667, 0.905882],[0, 0.278431, 0.894118],
[0, 0.286275, 0.882353],[0, 0.294118, 0.87451],[0, 0.301961, 0.862745],[0, 0.309804, 0.85098],
[0, 0.317647, 0.839216],[0, 0.32549, 0.827451],[0, 0.333333, 0.815686],[0, 0.341176, 0.803922],
[0, 0.34902, 0.792157],[0, 0.356863, 0.780392],[0, 0.360784, 0.768627],[0, 0.368627, 0.760784],
[0, 0.376471, 0.74902],[0, 0.384314, 0.737255],[0, 0.388235, 0.72549],[0, 0.396078, 0.713725],
[0, 0.403922, 0.701961],[0, 0.407843, 0.690196],[0, 0.415686, 0.682353],[0, 0.423529, 0.670588],
[0, 0.427451, 0.658824],[0, 0.435294, 0.647059],[0, 0.439216, 0.639216],[0, 0.443137, 0.627451],
[0, 0.45098, 0.615686],[0, 0.454902, 0.607843],[0, 0.458824, 0.596078],[0, 0.462745, 0.588235],
[0.027451, 0.466667, 0.576471],[0.054902, 0.470588, 0.568627],[0.0784314, 0.478431, 0.556863],[0.0941176, 0.482353, 0.54902],
[0.109804, 0.486275, 0.537255],[0.12549, 0.490196, 0.529412],[0.137255, 0.494118, 0.521569],[0.14902, 0.498039, 0.509804],
[0.160784, 0.501961, 0.501961],[0.168627, 0.505882, 0.494118],[0.176471, 0.509804, 0.482353],[0.184314, 0.513725, 0.47451],
[0.192157, 0.517647, 0.462745],[0.2, 0.521569, 0.454902],[0.203922, 0.52549, 0.447059],[0.207843, 0.529412, 0.435294],
[0.215686, 0.533333, 0.427451],[0.219608, 0.537255, 0.415686],[0.223529, 0.541176, 0.407843],[0.227451, 0.545098, 0.396078],
[0.231373, 0.54902, 0.388235],[0.231373, 0.552941, 0.376471],[0.235294, 0.556863, 0.368627],[0.239216, 0.560784, 0.356863],
[0.239216, 0.564706, 0.345098],[0.243137, 0.568627, 0.337255],[0.243137, 0.572549, 0.32549],[0.243137, 0.576471, 0.313725],
[0.247059, 0.580392, 0.305882],[0.247059, 0.584314, 0.294118],[0.247059, 0.588235, 0.282353],[0.247059, 0.596078, 0.270588],
[0.247059, 0.6, 0.258824],[0.247059, 0.603922, 0.247059],[0.247059, 0.607843, 0.235294],[0.247059, 0.611765, 0.223529],
[0.247059, 0.615686, 0.207843],[0.247059, 0.619608, 0.196078],[0.247059, 0.623529, 0.184314],[0.247059, 0.627451, 0.168627],
[0.247059, 0.631373, 0.156863],[0.25098, 0.635294, 0.141176],[0.25098, 0.639216, 0.129412],[0.254902, 0.643137, 0.117647],
[0.258824, 0.647059, 0.105882],[0.266667, 0.65098, 0.0941176],[0.27451, 0.65098, 0.0823529],[0.282353, 0.654902, 0.0745098],
[0.290196, 0.658824, 0.0666667],[0.301961, 0.662745, 0.0627451],[0.309804, 0.662745, 0.0588235],[0.321569, 0.666667, 0.054902],
[0.333333, 0.670588, 0.0509804],[0.341176, 0.670588, 0.0509804],[0.352941, 0.67451, 0.0509804],[0.364706, 0.67451, 0.054902],
[0.376471, 0.678431, 0.054902],[0.388235, 0.682353, 0.054902],[0.396078, 0.682353, 0.054902],[0.407843, 0.686275, 0.0588235],
[0.415686, 0.686275, 0.0588235],[0.427451, 0.690196, 0.0627451],[0.439216, 0.694118, 0.0627451],[0.447059, 0.694118, 0.0627451],
[0.458824, 0.698039, 0.0666667],[0.466667, 0.698039, 0.0666667],[0.478431, 0.701961, 0.0666667],[0.486275, 0.705882, 0.0705882],
[0.494118, 0.705882, 0.0705882],[0.505882, 0.709804, 0.0745098],[0.513725, 0.709804, 0.0745098],[0.52549, 0.713725, 0.0745098],
[0.533333, 0.713725, 0.0784314],[0.541176, 0.717647, 0.0784314],[0.552941, 0.717647, 0.0784314],[0.560784, 0.721569, 0.0823529],
[0.568627, 0.72549, 0.0823529],[0.580392, 0.72549, 0.0823529],[0.588235, 0.729412, 0.0862745],[0.596078, 0.729412, 0.0862745],
[0.603922, 0.733333, 0.0901961],[0.615686, 0.733333, 0.0901961],[0.623529, 0.737255, 0.0901961],[0.631373, 0.737255, 0.0941176],
[0.639216, 0.741176, 0.0941176],[0.65098, 0.741176, 0.0941176],[0.658824, 0.745098, 0.0980392],[0.666667, 0.74902, 0.0980392],
[0.67451, 0.74902, 0.101961],[0.686275, 0.752941, 0.101961],[0.694118, 0.752941, 0.101961],[0.701961, 0.756863, 0.105882],
[0.709804, 0.756863, 0.105882],[0.717647, 0.760784, 0.105882],[0.729412, 0.760784, 0.109804],[0.737255, 0.764706, 0.109804],
[0.745098, 0.764706, 0.109804],[0.752941, 0.768627, 0.113725],[0.760784, 0.768627, 0.113725],[0.768627, 0.772549, 0.117647],
[0.780392, 0.772549, 0.117647],[0.788235, 0.776471, 0.117647],[0.796078, 0.776471, 0.121569],[0.803922, 0.780392, 0.121569],
[0.811765, 0.780392, 0.121569],[0.819608, 0.784314, 0.12549],[0.827451, 0.784314, 0.12549],[0.839216, 0.788235, 0.129412],
[0.847059, 0.788235, 0.129412],[0.854902, 0.792157, 0.129412],[0.862745, 0.792157, 0.133333],[0.870588, 0.796078, 0.133333],
[0.878431, 0.796078, 0.133333],[0.886275, 0.796078, 0.137255],[0.898039, 0.8, 0.137255],[0.905882, 0.8, 0.137255],
[0.913725, 0.803922, 0.141176],[0.921569, 0.803922, 0.141176],[0.929412, 0.803922, 0.141176],[0.937255, 0.803922, 0.145098],
[0.945098, 0.803922, 0.145098],[0.94902, 0.803922, 0.145098],[0.956863, 0.803922, 0.145098],[0.960784, 0.803922, 0.145098],
[0.968627, 0.8, 0.145098],[0.972549, 0.8, 0.141176],[0.976471, 0.796078, 0.141176],[0.980392, 0.792157, 0.141176],
[0.984314, 0.788235, 0.137255],[0.984314, 0.784314, 0.137255],[0.988235, 0.780392, 0.133333],[0.988235, 0.772549, 0.133333],
[0.992157, 0.768627, 0.129412],[0.992157, 0.764706, 0.129412],[0.992157, 0.756863, 0.12549],[0.992157, 0.752941, 0.12549],
[0.996078, 0.74902, 0.121569],[0.996078, 0.741176, 0.117647],[0.996078, 0.737255, 0.117647],[0.996078, 0.733333, 0.113725],
[0.996078, 0.72549, 0.113725],[1, 0.721569, 0.109804],[1, 0.713725, 0.105882],[1, 0.709804, 0.105882],
[1, 0.705882, 0.101961],[1, 0.698039, 0.0980392],[1, 0.694118, 0.0980392],[1, 0.690196, 0.0941176],
[1, 0.682353, 0.0941176],[1, 0.678431, 0.0901961],[1, 0.670588, 0.0862745],[1, 0.666667, 0.0862745],
[1, 0.658824, 0.0823529],[1, 0.654902, 0.0784314],[1, 0.65098, 0.0784314],[1, 0.643137, 0.0745098],
[1, 0.639216, 0.0705882],[1, 0.631373, 0.0705882],[1, 0.627451, 0.0666667],[1, 0.619608, 0.0627451],
[1, 0.615686, 0.0627451],[1, 0.611765, 0.0588235],[1, 0.603922, 0.054902],[1, 0.6, 0.0509804],
[1, 0.592157, 0.0509804],[1, 0.588235, 0.0470588],[1, 0.580392, 0.0431373],[1, 0.576471, 0.0392157],
[1, 0.568627, 0.0392157],[1, 0.564706, 0.0352941],[1, 0.556863, 0.0313725],[1, 0.552941, 0.027451],
[1, 0.545098, 0.027451],[1, 0.541176, 0.0235294],[1, 0.533333, 0.0196078],[1, 0.52549, 0.0196078],
[1, 0.521569, 0.0156863],[1, 0.513725, 0.0117647],[1, 0.509804, 0.0117647],[1, 0.501961, 0.00784314],
[1, 0.498039, 0.00784314],[1, 0.490196, 0.00392157],[1, 0.482353, 0.00392157],[1, 0.478431, 0],
[1, 0.470588, 0],[1, 0.462745, 0],[1, 0.458824, 0],[1, 0.45098, 0],
[1, 0.443137, 0],[1, 0.439216, 0],[1, 0.431373, 0],[1, 0.423529, 0],
[1, 0.415686, 0],[1, 0.407843, 0],[1, 0.403922, 0],[1, 0.396078, 0],
[1, 0.388235, 0],[1, 0.380392, 0],[1, 0.372549, 0],[1, 0.364706, 0],
[1, 0.356863, 0],[1, 0.34902, 0],[1, 0.341176, 0],[1, 0.333333, 0],
[1, 0.32549, 0],[1, 0.317647, 0],[1, 0.309804, 0],[1, 0.298039, 0],
[1, 0.290196, 0],[1, 0.282353, 0],[1, 0.270588, 0],[1, 0.262745, 0],
[1, 0.25098, 0],[1, 0.239216, 0],[1, 0.231373, 0],[1, 0.219608, 0],
[1, 0.207843, 0],[1, 0.192157, 0],[1, 0.180392, 0],[1, 0.164706, 0] 
], 
[
[0, 0.188235, 0.960784],[0, 0.203922, 0.94902],[0, 0.215686, 0.933333],
[0, 0.231373, 0.921569],[0.0117647, 0.243137, 0.905882],[0.0352941, 0.258824, 0.894118],[0.054902, 0.270588, 0.882353],
[0.0705882, 0.282353, 0.866667],[0.0784314, 0.290196, 0.854902],[0.0862745, 0.301961, 0.839216],[0.0901961, 0.313725, 0.827451],
[0.0941176, 0.321569, 0.811765],[0.0980392, 0.333333, 0.8],[0.0980392, 0.341176, 0.784314],[0.0980392, 0.352941, 0.772549],
[0.0980392, 0.360784, 0.756863],[0.0980392, 0.368627, 0.745098],[0.0980392, 0.376471, 0.733333],[0.0941176, 0.388235, 0.717647],
[0.0941176, 0.396078, 0.705882],[0.0941176, 0.403922, 0.694118],[0.0901961, 0.411765, 0.678431],[0.0901961, 0.415686, 0.666667],
[0.0941176, 0.423529, 0.654902],[0.0941176, 0.431373, 0.643137],[0.0980392, 0.439216, 0.627451],[0.105882, 0.443137, 0.615686],
[0.109804, 0.45098, 0.603922],[0.117647, 0.458824, 0.592157],[0.12549, 0.462745, 0.580392],[0.133333, 0.470588, 0.568627],
[0.141176, 0.47451, 0.556863],[0.152941, 0.478431, 0.545098],[0.160784, 0.486275, 0.533333],[0.168627, 0.490196, 0.521569],
[0.176471, 0.494118, 0.509804],[0.184314, 0.501961, 0.498039],[0.192157, 0.505882, 0.486275],[0.2, 0.509804, 0.47451],
[0.207843, 0.517647, 0.462745],[0.211765, 0.521569, 0.45098],[0.219608, 0.52549, 0.439216],[0.223529, 0.533333, 0.427451],
[0.227451, 0.537255, 0.415686],[0.231373, 0.541176, 0.403922],[0.235294, 0.545098, 0.388235],[0.239216, 0.552941, 0.376471],
[0.243137, 0.556863, 0.364706],[0.243137, 0.560784, 0.352941],[0.247059, 0.568627, 0.341176],[0.247059, 0.572549, 0.32549],
[0.25098, 0.576471, 0.313725],[0.25098, 0.584314, 0.301961],[0.25098, 0.588235, 0.290196],[0.254902, 0.592157, 0.27451],
[0.254902, 0.6, 0.262745],[0.254902, 0.603922, 0.247059],[0.254902, 0.607843, 0.235294],[0.258824, 0.611765, 0.219608],
[0.258824, 0.619608, 0.207843],[0.262745, 0.623529, 0.196078],[0.266667, 0.627451, 0.180392],[0.270588, 0.631373, 0.168627],
[0.27451, 0.635294, 0.156863],[0.278431, 0.639216, 0.145098],[0.286275, 0.643137, 0.133333],[0.294118, 0.647059, 0.121569],
[0.301961, 0.65098, 0.109804],[0.309804, 0.654902, 0.101961],[0.321569, 0.658824, 0.0941176],[0.329412, 0.662745, 0.0862745],
[0.341176, 0.666667, 0.0784314],[0.352941, 0.670588, 0.0745098],[0.364706, 0.67451, 0.0705882],[0.376471, 0.678431, 0.0666667],
[0.388235, 0.678431, 0.0666667],[0.4, 0.682353, 0.0627451],[0.411765, 0.686275, 0.0627451],[0.423529, 0.690196, 0.0627451],
[0.435294, 0.690196, 0.0627451],[0.447059, 0.694118, 0.0666667],[0.458824, 0.698039, 0.0666667],[0.47451, 0.701961, 0.0666667],
[0.486275, 0.701961, 0.0705882],[0.498039, 0.705882, 0.0705882],[0.509804, 0.709804, 0.0745098],[0.517647, 0.713725, 0.0745098],
[0.529412, 0.713725, 0.0784314],[0.541176, 0.717647, 0.0784314],[0.552941, 0.721569, 0.0784314],[0.564706, 0.721569, 0.0823529],
[0.576471, 0.72549, 0.0823529],[0.588235, 0.729412, 0.0862745],[0.6, 0.729412, 0.0862745],[0.607843, 0.733333, 0.0901961],
[0.619608, 0.737255, 0.0901961],[0.631373, 0.737255, 0.0941176],[0.643137, 0.741176, 0.0941176],[0.65098, 0.745098, 0.0980392],
[0.662745, 0.745098, 0.0980392],[0.67451, 0.74902, 0.0980392],[0.686275, 0.752941, 0.101961],[0.694118, 0.752941, 0.101961],
[0.705882, 0.756863, 0.105882],[0.717647, 0.760784, 0.105882],[0.729412, 0.760784, 0.109804],[0.737255, 0.764706, 0.109804],
[0.74902, 0.764706, 0.113725],[0.760784, 0.768627, 0.113725],[0.768627, 0.772549, 0.117647],[0.780392, 0.772549, 0.117647],
[0.792157, 0.776471, 0.117647],[0.8, 0.780392, 0.121569],[0.811765, 0.780392, 0.121569],[0.823529, 0.784314, 0.12549],
[0.831373, 0.784314, 0.12549],[0.843137, 0.788235, 0.129412],[0.85098, 0.788235, 0.129412],[0.862745, 0.792157, 0.133333],
[0.87451, 0.792157, 0.133333],[0.882353, 0.792157, 0.133333],[0.890196, 0.796078, 0.137255],[0.901961, 0.796078, 0.137255],
[0.909804, 0.796078, 0.137255],[0.917647, 0.796078, 0.141176],[0.92549, 0.796078, 0.141176],[0.933333, 0.796078, 0.141176],
[0.941176, 0.796078, 0.141176],[0.945098, 0.792157, 0.141176],[0.952941, 0.792157, 0.141176],[0.956863, 0.788235, 0.141176],
[0.960784, 0.784314, 0.141176],[0.964706, 0.784314, 0.141176],[0.968627, 0.780392, 0.141176],[0.972549, 0.772549, 0.141176],
[0.972549, 0.768627, 0.141176],[0.976471, 0.764706, 0.141176],[0.976471, 0.760784, 0.137255],[0.976471, 0.752941, 0.137255],
[0.980392, 0.74902, 0.137255],[0.980392, 0.745098, 0.137255],[0.980392, 0.737255, 0.133333],[0.980392, 0.733333, 0.133333],
[0.980392, 0.72549, 0.133333],[0.980392, 0.721569, 0.129412],[0.980392, 0.713725, 0.129412],[0.980392, 0.705882, 0.129412],
[0.980392, 0.701961, 0.12549],[0.976471, 0.694118, 0.12549],[0.976471, 0.690196, 0.12549],[0.976471, 0.682353, 0.121569],
[0.976471, 0.678431, 0.121569],[0.976471, 0.670588, 0.121569],[0.976471, 0.662745, 0.117647],[0.976471, 0.658824, 0.117647],
[0.976471, 0.65098, 0.117647],[0.972549, 0.647059, 0.113725],[0.972549, 0.639216, 0.113725],[0.972549, 0.631373, 0.113725],
[0.972549, 0.627451, 0.113725],[0.972549, 0.619608, 0.109804],[0.972549, 0.615686, 0.109804],[0.972549, 0.607843, 0.109804],
[0.968627, 0.6, 0.105882],[0.968627, 0.596078, 0.105882],[0.968627, 0.588235, 0.105882],[0.968627, 0.580392, 0.101961],
[0.968627, 0.576471, 0.101961],[0.964706, 0.568627, 0.101961],[0.964706, 0.560784, 0.101961],[0.964706, 0.556863, 0.0980392],
[0.964706, 0.54902, 0.0980392],[0.964706, 0.541176, 0.0980392],[0.960784, 0.537255, 0.0941176],[0.960784, 0.529412, 0.0941176],
[0.960784, 0.521569, 0.0941176],[0.960784, 0.517647, 0.0941176],[0.956863, 0.509804, 0.0901961],[0.956863, 0.501961, 0.0901961],
[0.956863, 0.498039, 0.0901961],[0.956863, 0.490196, 0.0901961],[0.956863, 0.482353, 0.0862745],[0.952941, 0.47451, 0.0862745],
[0.952941, 0.466667, 0.0862745],[0.952941, 0.462745, 0.0862745],[0.952941, 0.454902, 0.0823529],[0.94902, 0.447059, 0.0823529],
[0.94902, 0.439216, 0.0823529],[0.94902, 0.431373, 0.0823529],[0.945098, 0.427451, 0.0823529],[0.945098, 0.419608, 0.0823529],
[0.945098, 0.411765, 0.0823529],[0.945098, 0.403922, 0.0823529],[0.941176, 0.396078, 0.0823529],[0.941176, 0.392157, 0.0862745],
[0.941176, 0.384314, 0.0862745],[0.941176, 0.376471, 0.0901961],[0.941176, 0.372549, 0.0941176],[0.941176, 0.364706, 0.101961],
[0.941176, 0.360784, 0.105882],[0.941176, 0.352941, 0.113725],[0.941176, 0.34902, 0.121569],[0.941176, 0.345098, 0.129412],
[0.941176, 0.341176, 0.141176],[0.941176, 0.341176, 0.14902],[0.945098, 0.337255, 0.160784],[0.945098, 0.337255, 0.172549],
[0.94902, 0.337255, 0.184314],[0.94902, 0.337255, 0.2],[0.952941, 0.337255, 0.211765],[0.952941, 0.341176, 0.227451],
[0.956863, 0.345098, 0.243137],[0.960784, 0.345098, 0.254902],[0.960784, 0.34902, 0.270588],[0.964706, 0.352941, 0.286275],
[0.968627, 0.356863, 0.301961],[0.968627, 0.360784, 0.321569],[0.972549, 0.368627, 0.337255],[0.976471, 0.372549, 0.352941],
[0.976471, 0.376471, 0.368627],[0.980392, 0.380392, 0.384314],[0.984314, 0.388235, 0.4],[0.984314, 0.392157, 0.415686],
[0.988235, 0.396078, 0.435294],[0.988235, 0.403922, 0.45098],[0.992157, 0.407843, 0.466667],[0.992157, 0.411765, 0.482353],
[0.996078, 0.419608, 0.501961],[0.996078, 0.423529, 0.517647],[1, 0.427451, 0.533333],[1, 0.435294, 0.54902],
[1, 0.439216, 0.568627],[1, 0.447059, 0.584314],[1, 0.45098, 0.6],[1, 0.454902, 0.615686],
[1, 0.462745, 0.635294],[1, 0.466667, 0.65098],[1, 0.470588, 0.666667],[1, 0.478431, 0.686275],
[1, 0.482353, 0.701961],[1, 0.490196, 0.717647],[1, 0.494118, 0.737255],[1, 0.498039, 0.752941],
[1, 0.505882, 0.768627],[1, 0.509804, 0.788235],[1, 0.517647, 0.803922],[1, 0.521569, 0.823529],
[1, 0.52549, 0.839216],[1, 0.533333, 0.858824],[1, 0.537255, 0.87451],[1, 0.545098, 0.890196],
[1, 0.54902, 0.909804],[1, 0.552941, 0.92549],[0.996078, 0.560784, 0.945098],[0.996078, 0.564706, 0.960784],
[0.992157, 0.572549, 0.980392] 
], 
[
[0, 0.207843, 0.976471],[0, 0.219608, 0.960784],
[0, 0.235294, 0.94902],[0, 0.247059, 0.933333],[0, 0.258824, 0.917647],[0, 0.270588, 0.905882],
[0, 0.278431, 0.890196],[0, 0.290196, 0.878431],[0, 0.301961, 0.862745],[0, 0.309804, 0.847059],
[0, 0.321569, 0.835294],[0, 0.329412, 0.819608],[0, 0.341176, 0.807843],[0, 0.34902, 0.792157],
[0, 0.356863, 0.776471],[0, 0.364706, 0.764706],[0, 0.372549, 0.74902],[0, 0.384314, 0.737255],
[0, 0.392157, 0.721569],[0, 0.4, 0.709804],[0, 0.403922, 0.694118],[0, 0.411765, 0.682353],
[0.00784314, 0.419608, 0.670588],[0.0156863, 0.427451, 0.654902],[0.027451, 0.435294, 0.643137],[0.0431373, 0.439216, 0.631373],
[0.054902, 0.447059, 0.619608],[0.0705882, 0.454902, 0.603922],[0.0823529, 0.458824, 0.592157],[0.0941176, 0.466667, 0.580392],
[0.105882, 0.470588, 0.568627],[0.117647, 0.47451, 0.556863],[0.129412, 0.482353, 0.545098],[0.141176, 0.486275, 0.533333],
[0.152941, 0.490196, 0.521569],[0.160784, 0.498039, 0.509804],[0.172549, 0.501961, 0.498039],[0.180392, 0.505882, 0.486275],
[0.188235, 0.513725, 0.47451],[0.196078, 0.517647, 0.462745],[0.203922, 0.521569, 0.45098],[0.207843, 0.529412, 0.439216],
[0.215686, 0.533333, 0.427451],[0.219608, 0.537255, 0.415686],[0.223529, 0.541176, 0.403922],[0.227451, 0.54902, 0.388235],
[0.231373, 0.552941, 0.376471],[0.235294, 0.556863, 0.364706],[0.239216, 0.564706, 0.352941],[0.243137, 0.568627, 0.341176],
[0.243137, 0.572549, 0.32549],[0.247059, 0.576471, 0.313725],[0.247059, 0.584314, 0.301961],[0.247059, 0.588235, 0.286275],
[0.25098, 0.592157, 0.27451],[0.25098, 0.6, 0.262745],[0.25098, 0.603922, 0.247059],[0.254902, 0.607843, 0.235294],
[0.254902, 0.611765, 0.219608],[0.258824, 0.619608, 0.207843],[0.258824, 0.623529, 0.196078],[0.262745, 0.627451, 0.180392],
[0.266667, 0.631373, 0.168627],[0.27451, 0.635294, 0.156863],[0.278431, 0.639216, 0.145098],[0.286275, 0.643137, 0.133333],
[0.294118, 0.647059, 0.121569],[0.301961, 0.65098, 0.109804],[0.309804, 0.654902, 0.101961],[0.321569, 0.658824, 0.0941176],
[0.329412, 0.662745, 0.0862745],[0.341176, 0.666667, 0.0784314],[0.352941, 0.670588, 0.0745098],[0.364706, 0.67451, 0.0705882],
[0.376471, 0.678431, 0.0666667],[0.388235, 0.678431, 0.0666667],[0.4, 0.682353, 0.0627451],[0.411765, 0.686275, 0.0627451],
[0.423529, 0.690196, 0.0627451],[0.435294, 0.690196, 0.0627451],[0.447059, 0.694118, 0.0666667],[0.458824, 0.698039, 0.0666667],
[0.47451, 0.701961, 0.0666667],[0.486275, 0.701961, 0.0705882],[0.498039, 0.705882, 0.0705882],[0.509804, 0.709804, 0.0745098],
[0.517647, 0.713725, 0.0745098],[0.529412, 0.713725, 0.0784314],[0.541176, 0.717647, 0.0784314],[0.552941, 0.721569, 0.0784314],
[0.564706, 0.721569, 0.0823529],[0.576471, 0.72549, 0.0823529],[0.588235, 0.729412, 0.0862745],[0.6, 0.729412, 0.0862745],
[0.607843, 0.733333, 0.0901961],[0.619608, 0.737255, 0.0901961],[0.631373, 0.737255, 0.0941176],[0.643137, 0.741176, 0.0941176],
[0.65098, 0.745098, 0.0980392],[0.662745, 0.745098, 0.0980392],[0.67451, 0.74902, 0.0980392],[0.686275, 0.752941, 0.101961],
[0.694118, 0.752941, 0.101961],[0.705882, 0.756863, 0.105882],[0.717647, 0.760784, 0.105882],[0.729412, 0.760784, 0.109804],
[0.737255, 0.764706, 0.109804],[0.74902, 0.764706, 0.113725],[0.760784, 0.768627, 0.113725],[0.768627, 0.772549, 0.117647],
[0.780392, 0.772549, 0.117647],[0.792157, 0.776471, 0.117647],[0.8, 0.776471, 0.121569],[0.811765, 0.780392, 0.121569],
[0.823529, 0.784314, 0.12549],[0.831373, 0.784314, 0.12549],[0.843137, 0.788235, 0.129412],[0.85098, 0.788235, 0.129412],
[0.862745, 0.792157, 0.133333],[0.87451, 0.792157, 0.133333],[0.882353, 0.792157, 0.133333],[0.894118, 0.796078, 0.137255],
[0.901961, 0.796078, 0.137255],[0.909804, 0.796078, 0.137255],[0.917647, 0.796078, 0.137255],[0.92549, 0.796078, 0.141176],
[0.933333, 0.796078, 0.141176],[0.941176, 0.796078, 0.141176],[0.94902, 0.792157, 0.141176],[0.956863, 0.792157, 0.141176],
[0.960784, 0.788235, 0.141176],[0.964706, 0.784314, 0.141176],[0.968627, 0.780392, 0.141176],[0.972549, 0.776471, 0.137255],
[0.976471, 0.772549, 0.137255],[0.980392, 0.768627, 0.137255],[0.984314, 0.760784, 0.137255],[0.984314, 0.756863, 0.133333],
[0.988235, 0.752941, 0.133333],[0.988235, 0.745098, 0.129412],[0.988235, 0.737255, 0.129412],[0.988235, 0.733333, 0.129412],
[0.992157, 0.72549, 0.12549],[0.992157, 0.721569, 0.12549],[0.992157, 0.713725, 0.121569],[0.992157, 0.705882, 0.121569],
[0.992157, 0.701961, 0.117647],[0.992157, 0.694118, 0.117647],[0.992157, 0.686275, 0.117647],[0.996078, 0.682353, 0.113725],
[0.996078, 0.67451, 0.113725],[0.996078, 0.666667, 0.109804],[0.996078, 0.662745, 0.109804],[0.996078, 0.654902, 0.105882],
[0.996078, 0.647059, 0.105882],[0.996078, 0.643137, 0.101961],[0.996078, 0.635294, 0.101961],[0.996078, 0.627451, 0.101961],
[0.996078, 0.619608, 0.0980392],[0.996078, 0.615686, 0.0980392],[0.996078, 0.607843, 0.0941176],[0.996078, 0.6, 0.0941176],
[0.996078, 0.592157, 0.0901961],[0.996078, 0.588235, 0.0901961],[0.996078, 0.580392, 0.0901961],[0.996078, 0.572549, 0.0862745],
[0.996078, 0.564706, 0.0862745],[0.996078, 0.556863, 0.0823529],[0.996078, 0.552941, 0.0823529],[0.996078, 0.545098, 0.0823529],
[0.996078, 0.537255, 0.0784314],[0.996078, 0.529412, 0.0784314],[0.996078, 0.521569, 0.0745098],[0.992157, 0.513725, 0.0745098],
[0.992157, 0.509804, 0.0745098],[0.992157, 0.501961, 0.0705882],[0.992157, 0.494118, 0.0705882],[0.992157, 0.486275, 0.0705882],
[0.992157, 0.478431, 0.0666667],[0.992157, 0.470588, 0.0666667],[0.992157, 0.462745, 0.0627451],[0.992157, 0.454902, 0.0627451],
[0.992157, 0.447059, 0.0627451],[0.992157, 0.439216, 0.0588235],[0.988235, 0.431373, 0.0588235],[0.988235, 0.423529, 0.0588235],
[0.988235, 0.415686, 0.0588235],[0.988235, 0.407843, 0.054902],[0.988235, 0.4, 0.054902],[0.988235, 0.388235, 0.054902],
[0.988235, 0.380392, 0.054902],[0.988235, 0.372549, 0.054902],[0.984314, 0.364706, 0.054902],[0.984314, 0.356863, 0.0588235],
[0.984314, 0.34902, 0.0588235],[0.984314, 0.341176, 0.0627451],[0.984314, 0.333333, 0.0666667],[0.984314, 0.32549, 0.0705882],
[0.984314, 0.317647, 0.0784314],[0.984314, 0.309804, 0.0862745],[0.984314, 0.301961, 0.0941176],[0.984314, 0.298039, 0.101961],
[0.984314, 0.294118, 0.109804],[0.988235, 0.290196, 0.121569],[0.988235, 0.286275, 0.133333],[0.988235, 0.282353, 0.145098],
[0.992157, 0.282353, 0.156863],[0.992157, 0.282353, 0.168627],[0.992157, 0.282353, 0.184314],[0.996078, 0.286275, 0.196078],
[0.996078, 0.286275, 0.211765],[1, 0.290196, 0.227451],[1, 0.294118, 0.243137],[1, 0.301961, 0.258824],
[1, 0.305882, 0.27451],[1, 0.313725, 0.290196],[1, 0.317647, 0.305882],[1, 0.32549, 0.321569],
[1, 0.329412, 0.341176],[1, 0.337255, 0.356863],[1, 0.345098, 0.372549],[1, 0.34902, 0.388235],
[1, 0.356863, 0.407843],[1, 0.364706, 0.423529],[1, 0.368627, 0.439216],[1, 0.376471, 0.454902],
[1, 0.384314, 0.47451],[1, 0.388235, 0.490196],[1, 0.396078, 0.505882],[1, 0.403922, 0.52549],
[1, 0.411765, 0.541176],[1, 0.415686, 0.556863],[1, 0.423529, 0.576471],[1, 0.431373, 0.592157],
[1, 0.435294, 0.607843],[1, 0.443137, 0.627451],[1, 0.45098, 0.643137],[1, 0.454902, 0.662745],
[1, 0.462745, 0.678431],[1, 0.466667, 0.694118],[1, 0.47451, 0.713725],[1, 0.482353, 0.729412],
[1, 0.486275, 0.74902],[1, 0.494118, 0.764706],[1, 0.501961, 0.784314],[1, 0.505882, 0.8],
[1, 0.513725, 0.819608],[1, 0.521569, 0.835294],[1, 0.52549, 0.854902],[1, 0.533333, 0.870588],
[1, 0.541176, 0.890196],[1, 0.545098, 0.909804],[1, 0.552941, 0.92549],[1, 0.556863, 0.945098],
[0.996078, 0.564706, 0.960784],[0.992157, 0.572549, 0.980392] 
]
]

end # module
